@@ -306,7 +306,7 @@ order for the change to take effect."
   ("*"         . ("⋆"))
   (".+"        . ("∔"))
   (".-"        . ("∸"))
-  (":"         . ("∶"))
+  (":"         . ,(agda-input-to-string-list "∶⦂ː꞉˸፥፦：﹕︓"))
   ("::"        . ("∷"))
   ("::-"       . ("∺"))
   ("-:"        . ("∹"))
@@ -541,6 +541,19 @@ order for the change to take effect."
   ("b8"   . ("𝟠"))
   ("b9"   . ("𝟡"))
 
+  ;; Mathematical bold digits.
+
+  ("B0"   . ("𝟎"))
+  ("B1"   . ("𝟏"))
+  ("B2"   . ("𝟐"))
+  ("B3"   . ("𝟑"))
+  ("B4"   . ("𝟒"))
+  ("B5"   . ("𝟓"))
+  ("B6"   . ("𝟔"))
+  ("B7"   . ("𝟕"))
+  ("B8"   . ("𝟖"))
+  ("B9"   . ("𝟗"))
+
   ;; Parentheses.
 
   ("(" . ,(agda-input-to-string-list "([{⁅⁽₍〈⎴⟅⟦⟨⟪⦃〈《「『【〔〖〚︵︷︹︻︽︿﹁﹃﹙﹛﹝（［｛｢"))
