@@ -1,27 +1,31 @@
 {-# LANGUAGE CPP #-}
 
-module Agda.Auto.Auto (auto) where
+module Agda.Auto.Auto
+      (auto
+      , AutoResult(..)
+      , AutoProgress(..)
+      ) where
 
 import Prelude hiding (null)
 
+#if __GLASGOW_HASKELL__ <= 708
+import Control.Applicative ( (<$), pure )
+#endif
+
+import Data.Functor
 import Control.Monad.State
-import Data.List hiding (null)
+import qualified Data.List as List
 import qualified Data.Map as Map
 import Data.IORef
 import qualified System.Timeout
 import Data.Maybe
-import Data.Functor
 import qualified Data.Traversable as Trav
 
 import Agda.Utils.Permutation (permute, takeP)
 import Agda.TypeChecking.Monad hiding (withCurrentModule)
--- import Agda.TypeChecking.Monad.Base
--- import Agda.TypeChecking.Monad.MetaVars
--- import Agda.TypeChecking.Monad.Context
--- import Agda.TypeChecking.Monad.Signature
-import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 
+import Agda.Syntax.Common (Hiding(..))
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Abstract.Pretty (prettyA)
 import qualified Text.PrettyPrint as PP
@@ -40,6 +44,7 @@ import qualified Agda.TypeChecking.Monad.Base as TCM
 import Agda.TypeChecking.EtaContract (etaContract)
 import qualified Agda.Utils.HashMap as HMap
 
+import Agda.Auto.Options
 import Agda.Auto.Convert
 import Agda.Auto.NarrowingSearch
 import Agda.Auto.Syntax
@@ -63,61 +68,68 @@ insertAbsurdPattern [] = []
 insertAbsurdPattern s@(_:_) | take (length abspatvarname) s == abspatvarname = "()" ++ drop (length abspatvarname) s
 insertAbsurdPattern (c:s) = c : insertAbsurdPattern s
 
-getName :: A.Expr -> Maybe (Bool, I.QName)
-getName (A.ScopedExpr _ e) = getName e
-getName (A.Def qname) = Just (False, qname)
-getName (A.Proj _ qname) = Just (False, head $ I.unAmbQ qname)
-getName (A.Con qname) = Just (True, head $ I.unAmbQ qname)
-getName _ = Nothing
+getHeadAsHint :: A.Expr -> Maybe Hint
+getHeadAsHint (A.ScopedExpr _ e) = getHeadAsHint e
+getHeadAsHint (A.Def qname)      = Just $ Hint False qname
+getHeadAsHint (A.Proj _ qname)   = Just $ Hint False $ head $ I.unAmbQ qname
+getHeadAsHint (A.Con qname)      = Just $ Hint True  $ head $ I.unAmbQ qname
+getHeadAsHint _ = Nothing
 
-dispmsg :: String ->
-           TCM (Either [(InteractionId, String)]
-                       (Either [String] String)
-               , Maybe String)
-dispmsg msg = return (Left [], Just msg)
+-- | Result type: Progress & potential Message for the user
+--
+--   The  of the Auto tactic can be one of the following three:
+--
+--   1. @Solutions [(ii,s)]@
+--      A list of solutions @s@ for interaction ids @ii@.
+--      In particular, @Solutions []@ means Agsy found no solution.
+--
+--   2. @FunClauses cs@
+--      A list of clauses for the interaction id @ii@ in which Auto
+--      was invoked with case-splitting turned on.
+--
+--   3. @Refinement s@
+--      A refinement for the interaction id @ii@ in which Auto was invoked.
+
+data AutoProgress =
+    Solutions  [(InteractionId, String)]
+  | FunClauses [String]
+  | Refinement String
+
+data AutoResult = AutoResult
+  { autoProgress :: AutoProgress
+  , autoMessage  :: Maybe String
+  }
+
+stopWithMsg :: String -> TCM AutoResult
+stopWithMsg msg = return $ AutoResult (Solutions []) (Just msg)
 
 -- | Entry point for Auto tactic (Agsy).
 --
---     @auto ii rng s = return (res, mmsg)@
---
---   If @mmsg = Just msg@, the message @msg@ produced by Agsy should
---   be displayed to the user.
---
---   The result @res@ of the Auto tactic can be one of the following three:
---
---   1. @Left [(ii,s)]@
---      A list of solutions @s@ for interaction ids @ii@.
---      In particular, @Left []@ means Agsy found no solution.
---
---   2. @Right (Left cs)@
---      A list of clauses (the user allowed case-split).
---
---   3. @Right (Right s)@
---      A refinement for the interaction id @ii@ in which Auto was invoked.
+--   If the @autoMessage@ part of the result is set to @Just msg@, the
+--   message @msg@ produced by Agsy should be displayed to the user.
 
 auto
   :: InteractionId
   -> Range
   -> String
-  -> TCM ( Either [(InteractionId, String)]
-                  (Either [String] String)
-         , Maybe String)
+  -> TCM AutoResult
 auto ii rng argstr = do
 
   -- Parse hints and other configuration.
-  let (hints, timeout, pick, mode, hintmode) = parseargs argstr
+  let autoOptions = parseArgs argstr
+  let hints    = autoOptions ^. aoHints
+  let timeout  = autoOptions ^. aoTimeOut
+  let pick     = autoOptions ^. aoPick
+  let mode     = autoOptions ^. aoMode
+  let hintmode = autoOptions ^. aoHintMode
   ahints <- case mode of
     MRefine{} -> return []
     _         -> mapM (parseExprIn ii rng) hints
-  let failHints = dispmsg "Hints must be a list of constant names"
-  caseMaybe (mapM getName ahints) failHints $ \ ehints -> do
+  let failHints = stopWithMsg "Hints must be a list of constant names"
 
-    -- Get names for equality reasoning.
-    -- @eqstuff == []@ if any of these names is not defined.
-    eqstuffExprs <- mapM (parseExprIn ii rng) ["_≡_", "begin_", "_≡⟨_⟩_", "_∎", "sym", "cong"]
-      `catchError`
-        (\_ -> return [])
-    let eqstuff = fromMaybe [] $ mapM getName eqstuffExprs
+  eqstuff <- getEqCombinators ii rng
+
+  caseMaybe (mapM getHeadAsHint ahints) failHints $ \ ehints -> do
 
     -- Get the meta variable for the interaction point we are trying to fill.
     -- Add the @autohints@ for that meta to the hints collection.
@@ -156,7 +168,9 @@ auto ii rng argstr = do
              nsol' <- readIORef nsol
              let cond = nsol' <= numsols
              when cond $ do
-               trms <- runExceptT $ mapM (\ (m, _, _, _) -> frommy (Meta m)) $ Map.elems tccons
+               trms <- runExceptT
+                       $ mapM (\ (m , _, _, _) -> convert (Meta m) :: MOT I.Term)
+                       $ Map.elems tccons
                case trms of
                  Left{}     -> writeIORef nsol $! nsol' + 1
                  Right trms -> modifyIORef sols (trms :)
@@ -165,7 +179,8 @@ auto ii rng argstr = do
         ticks <- liftIO $ newIORef 0
 
         let exsearch initprop recinfo defdfv =
-             liftIO $ System.Timeout.timeout (timeout * 1000000) $ loop 0
+             liftIO $ System.Timeout.timeout (getTimeOut timeout * 1000000)
+                    $ loop 0
              where
                loop d = do
                  let rechint x = case recinfo of
@@ -207,7 +222,7 @@ auto ii rng argstr = do
                            Nothing -> return 0
                 ee <- liftIO $ newIORef $ ConstDef {cdname = "T", cdorigin = __IMPOSSIBLE__, cdtype = NotM $ Sort (Set 0), cdcont = Postulate, cddeffreevars = 0}
                 let (restargs, modargs) = splitAt (length mylocalVars - defdfv) mylocalVars
-                    mytype' = foldl (\x y -> NotM $ Pi Nothing Agda.Auto.Syntax.NotHidden (freeIn 0 y) y (Abs NoId x)) mytype restargs
+                    mytype' = foldl (\x y -> NotM $ Pi Nothing NotHidden (freeIn 0 y) y (Abs NoId x)) mytype restargs
                     htyp = negtype ee mytype'
                     sctx = (Id "h", closify htyp) : map (\x -> (NoId, closify x)) modargs
                     ntt = closify (NotM $ App Nothing (NotM OKVal) (Const ee) (NotM ALNil))
@@ -215,7 +230,7 @@ auto ii rng argstr = do
                 rsols <- liftM reverse $ liftIO $ readIORef sols
                 if null rsols then do
                   nsol' <- liftIO $ readIORef nsol
-                  dispmsg $ insuffsols (pick + numsols - nsol')
+                  stopWithMsg $ insuffsols (pick + numsols - nsol')
                  else do
                   aexprss <- mapM getsols rsols
                   cexprss <- forM aexprss $ mapM $ \(mi, e) -> do
@@ -226,11 +241,11 @@ auto ii rng argstr = do
                       disp [(_, cexpr)] = ss cexpr
                       disp cexprs = concat $ map (\ (mi, cexpr) -> ss cexpr ++ " ") cexprs
                   ticks <- liftIO $ readIORef ticks
-                  dispmsg $ unlines $
+                  stopWithMsg $ unlines $
                     ("Listing disproof(s) " ++ show pick ++ "-" ++ show (pick + length rsols - 1)) :
                     for (zip cexprss [pick..]) (\ (x, y) -> show y ++ "  " ++ disp x)
-            _ -> dispmsg "Metavariable dependencies not allowed in disprove mode"
-           _ -> dispmsg "Metavariable dependencies not allowed in disprove mode"
+            _ -> stopWithMsg "Metavariable dependencies not allowed in disprove mode"
+           _ -> stopWithMsg "Metavariable dependencies not allowed in disprove mode"
          else do
           (recinfo, defdfv) <-
            case thisdefinfo of
@@ -269,7 +284,7 @@ auto ii rng argstr = do
             rsols <- liftM reverse $ liftIO $ readIORef sols
             if null rsols then do
               nsol' <- liftIO $ readIORef nsol
-              dispmsg $ insuffsols (pick + numsols - nsol') ++ timeoutString
+              stopWithMsg $ insuffsols (pick + numsols - nsol') ++ timeoutString
              else do
               aexprss <- mapM getsols rsols
               -- cexprss <- mapM (mapM (\(mi, e) -> lookupMeta mi >>= \mv -> withMetaInfo (getMetaInfo mv) $ abstractToConcrete_ e >>= \e' -> return (mi, e'))) aexprss
@@ -284,20 +299,20 @@ auto ii rng argstr = do
                     maybe (show mi) show (lookup mi riis)
                       ++ " := " ++ show cexpr ++ " "
               ticks <- liftIO $ readIORef ticks
-              dispmsg $ "Listing solution(s) " ++ show pick ++ "-" ++ show (pick + length rsols - 1) ++ timeoutString ++
+              stopWithMsg $ "Listing solution(s) " ++ show pick ++ "-" ++ show (pick + length rsols - 1) ++ timeoutString ++
                         "\n" ++ unlines (map (\(x, y) -> show y ++ "  " ++ disp x) $ zip cexprss [pick..])
            else {- not listmode -}
             case res of
              Nothing -> do
               nsol' <- liftIO $ readIORef nsol
-              dispmsg $ insuffsols (pick + numsols - nsol') ++ timeoutString
+              stopWithMsg $ insuffsols (pick + numsols - nsol') ++ timeoutString
              Just depthreached -> do
               ticks <- liftIO $ readIORef ticks
               rsols <- liftIO $ readIORef sols
               case rsols of
                 [] -> do
                   nsol' <- liftIO $ readIORef nsol
-                  dispmsg $ insuffsols (pick + numsols - nsol')
+                  stopWithMsg $ insuffsols (pick + numsols - nsol')
                 terms -> loop terms where
                   -- Andreas, 2015-05-17  Issue 1504
                   -- If giving a solution failed (e.g. ill-typed)
@@ -305,7 +320,8 @@ auto ii rng argstr = do
                   -- However, currently @terms@ is always a singleton list.
                   -- Thus, the following @loop@ is not doing something very
                   -- meaningful.
-                  loop [] = return (Left [], Just "")
+                  loop :: [[I.Term]] -> TCM AutoResult
+                  loop [] = return $ AutoResult (Solutions []) (Just "")
                   loop (term : terms') = do
                     -- On exception, try next solution
                     flip catchError (const $ loop terms') $ do
@@ -316,10 +332,10 @@ auto ii rng argstr = do
                         case lookup mi riis of
                          Nothing ->
                           -- catchError
-                           (giveExpr Nothing mi expr >> return (Nothing, Nothing))
+                           (giveExpr WithoutForce Nothing mi expr >> return (Nothing, Nothing))
                            -- (const retry)
                            -- (\_ -> return (Nothing, Just ("Failed to give expr for side solution of " ++ show mi)))
-                         Just ii' -> do ae <- give ii' Nothing expr
+                         Just ii' -> do ae <- give WithoutForce ii' Nothing expr
                                         mv <- lookupMeta mi
                                         let scope = getMetaScope mv
                                         ce <- abstractToConcreteEnv (makeEnv scope) ae
@@ -337,10 +353,8 @@ auto ii rng argstr = do
                                           if mi' == mi then "" else (" " ++ case lookup mi' riis of {Nothing -> show mi'; Just ii -> show ii})
                                          ) exprs
                       let msgs = catMaybes $ msg : map snd giveress
-                          msg' = case msgs of
-                                  [] -> Nothing
-                                  _ -> Just $ unlines msgs
-                      return (Left $ catMaybes $ map fst giveress, msg')
+                          msg' = unlines msgs <$ guard (not $ null msgs)
+                      return $ AutoResult (Solutions $ catMaybes $ map fst giveress) msg'
 
      MCaseSplit -> do
       case thisdefinfo of
@@ -353,7 +367,7 @@ auto ii rng argstr = do
           let [rectyp'] = mymrectyp
           defdfv <- getdfv mi def
           myrecdef <- liftIO $ newIORef $ ConstDef {cdname = "", cdorigin = (Nothing, def), cdtype = rectyp', cdcont = Postulate, cddeffreevars = defdfv}
-          sols <- liftIO $ System.Timeout.timeout (timeout * 1000000) (
+          sols <- liftIO $ System.Timeout.timeout (getTimeOut timeout * 1000000) (
              let r d = do
                   sols <- liftIO $ caseSplitSearch ticks __IMPOSSIBLE__ myhints meqr __IMPOSSIBLE__ d myrecdef ctx mytype pats
                   case sols of
@@ -364,25 +378,25 @@ auto ii rng argstr = do
            Just (cls : _) -> withInteractionId ii $ do
             cls' <- liftIO $ runExceptT (mapM frommyClause cls)
             case cls' of
-             Left{} -> dispmsg "No solution found"
+             Left{} -> stopWithMsg "No solution found"
              Right cls' -> do
-              cls'' <- forM cls' $ \ (I.Clause _ tel ps body t catchall) -> do
+              cls'' <- forM cls' $ \ (I.Clause _ _ tel ps body t catchall reachable) -> do
                 withCurrentModule (AN.qnameModule def) $ do
                  -- Normalise the dot patterns
                  ps <- addContext tel $ normalise ps
                  body <- etaContract body
-                 liftM modifyAbstractClause $ inTopContext $ reify $ AN.QNamed def $ I.Clause noRange tel ps body t catchall
+                 liftM modifyAbstractClause $ inTopContext $ reify $ AN.QNamed def $ I.Clause noRange noRange tel ps body t catchall reachable
               moduleTel <- lookupSection (AN.qnameModule def)
               pcs <- withInteractionId ii $ inTopContext $ addContext moduleTel $ mapM prettyA cls''
               ticks <- liftIO $ readIORef ticks
 
 
-              return (Right $ Left (map (insertAbsurdPattern . PP.renderStyle (PP.style { PP.mode = PP.OneLineMode })) pcs), Nothing)
+              return $ AutoResult (FunClauses $ map (insertAbsurdPattern . PP.renderStyle (PP.style { PP.mode = PP.OneLineMode })) pcs) Nothing
 
-           Just [] -> dispmsg "No solution found" -- case not possible at the moment because case split doesnt care about search exhaustiveness
-           Nothing -> dispmsg $ "No solution found at time out (" ++ show timeout ++ "s)"
-         _ -> dispmsg "Metavariable dependencies not allowed in case split mode"
-       _ -> dispmsg "Metavariable is not at top level of clause RHS"
+           Just [] -> stopWithMsg "No solution found" -- case not possible at the moment because case split doesnt care about search exhaustiveness
+           Nothing -> stopWithMsg $ "No solution found at time out (" ++ show timeout ++ "s)"
+         _ -> stopWithMsg "Metavariable dependencies not allowed in case split mode"
+       _ -> stopWithMsg "Metavariable is not at top level of clause RHS"
 
      MRefine listmode -> do
       mv <- lookupMeta mi
@@ -431,32 +445,31 @@ auto ii rng argstr = do
            Just score -> Just (show cn, score)
          ) modnames
 
-      let sorthits = sortBy (\(_, (pa1, pb1)) (_, (pa2, pb2)) -> case compare pa2 pa1 of {EQ -> compare pb1 pb2; o -> o}) hits
+      let sorthits = List.sortBy (\(_, (pa1, pb1)) (_, (pa2, pb2)) -> case compare pa2 pa1 of {EQ -> compare pb1 pb2; o -> o}) hits
       if listmode || pick == (-1) then
         let pick' = max 0 pick
         in if pick' >= length sorthits then
-             dispmsg $ insuffcands $ length sorthits
+             stopWithMsg $ insuffcands $ length sorthits
             else
              let showhits = take 10 $ drop pick' sorthits
-             in dispmsg $ "Listing candidate(s) " ++ show pick' ++ "-" ++ show (pick' + length showhits - 1) ++ " (found " ++ show (length sorthits) ++ " in total)\n" ++
+             in stopWithMsg $ "Listing candidate(s) " ++ show pick' ++ "-" ++ show (pick' + length showhits - 1) ++ " (found " ++ show (length sorthits) ++ " in total)\n" ++
                            unlines (map (\(i, (cn, _)) -> show i ++ "  " ++ cn) (zip [pick'..pick' + length showhits - 1] showhits))
        else
         if pick >= length sorthits then
-         dispmsg $ insuffcands $ length sorthits
+         stopWithMsg $ insuffcands $ length sorthits
         else
-         return (Right $ Right (fst $ sorthits !! pick), Nothing)
+         return $ AutoResult (Refinement $ fst $ sorthits !! pick) Nothing
   where
     agsyinfo ticks = ""
 
 -- Get the functions and axioms defined in the same module as @def@.
-autohints :: AutoHintMode -> I.MetaId -> Maybe AN.QName ->
-             TCM [(Bool, AN.QName)]
+autohints :: AutoHintMode -> I.MetaId -> Maybe AN.QName -> TCM [Hint]
 autohints AHMModule mi (Just def) = do
   scope <- clScope . getMetaInfo <$> lookupMeta mi
   let names     = Scope.nsNames $ Scope.everythingInScope scope
       qnames    = map (Scope.anameName . head) $ Map.elems names
       modnames  = filter (\n -> AN.qnameModule n == AN.qnameModule def && n /= def) qnames
-  map (False,) <$> do
+  map (Hint False) <$> do
     (`filterM` modnames) $ \ n -> do
       c <- getConstInfo n
       case theDef c of
@@ -467,48 +480,27 @@ autohints AHMModule mi (Just def) = do
 
 autohints _ _ _ = return []
 
+
+
+-- | Names for the equality reasoning combinators
+--   Empty if any of these names is not defined.
+
+getEqCombinators :: InteractionId -> Range -> TCM [Hint]
+getEqCombinators ii rng = do
+  let eqCombinators = ["_≡_", "begin_", "_≡⟨_⟩_", "_∎", "sym", "cong"]
+  raw <- mapM (parseExprIn ii rng) eqCombinators `catchError` const (pure [])
+  return $ fromMaybe [] $ mapM getHeadAsHint raw
+
+-- | Templates for error messages
+
+genericNotEnough :: String -> Int -> String
+genericNotEnough str n = List.intercalate " " $ case n of
+  0 -> [ "No"    , str, "found"]
+  1 -> [ "Only 1", str, "found" ]
+  _ -> [ "Only", show n, str ++ "s", "found" ]
+
 insuffsols :: Int -> String
-insuffsols 0 = "No solution found"
-insuffsols n = "Only " ++ show n ++ " solution(s) found"
+insuffsols  = genericNotEnough "solution"
 
 insuffcands :: Int -> String
-insuffcands 0 = "No candidate found"
-insuffcands n = "Only " ++ show n ++ " candidate(s) found"
-
-data Mode = MNormal Bool Bool -- true if list mode, true if disprove
-
-          | MCaseSplit
-
-          | MRefine Bool -- true if list mode
-
-
-data AutoHintMode = AHMNone
-                  | AHMModule
-
-parseargs :: String -> ([String], Int, Int, Mode, AutoHintMode)
-parseargs s =
- let r ("-t" : timeout : ws) (_, pick, mode, hintmode) =
-      r ws (read timeout, pick, mode, hintmode)
-     r ("-s" : pick : ws) (timeout, _, mode, hintmode) =
-      r ws (timeout, read pick, mode, hintmode)
-
-
-     r ("-l" : ws) (timeout, pick, MNormal _ disprove, hintmode) =
-      r ws (timeout, pick, MNormal True disprove, hintmode)
-     r ("-l" : ws) (timeout, pick, MRefine _, hintmode) =
-      r ws (timeout, pick, MRefine True, hintmode)
-     r ("-d" : ws) (timeout, pick, MNormal listmode _, hintmode) =
-      r ws (timeout, pick, MNormal listmode True, hintmode)
-     r ("-m" : ws) (timeout, pick, mode, _) =
-      r ws (timeout, pick, mode, AHMModule)
-
-     r ("-c" : ws) (timeout, pick, _, hintmode) =
-      r ws (timeout, pick, MCaseSplit, hintmode)
-
-     r ("-r" : ws) (timeout, pick, _, hintmode) =
-      r ws (timeout, (-1), MRefine False, hintmode)
-     r (h : ws) x =
-      let (hints, timeout, pick, mode, hintmode) = r ws x
-      in (h : hints, timeout, pick, mode, hintmode)
-     r [] (x,y,z,w) = ([],x,y,z,w)
- in r (words s) (5, 0, MNormal False False, AHMNone)
+insuffcands = genericNotEnough "candidate"
