@@ -3,7 +3,13 @@
 module Agda.Auto.Syntax where
 
 import Data.IORef
+import qualified Data.Set as Set
 
+#if __GLASGOW_HASKELL__ <= 708
+import Control.Applicative ( (<$>), (<*>) )
+#endif
+
+import Agda.Syntax.Common (Hiding)
 import Agda.Auto.NarrowingSearch
 
 #include "undefined.h"
@@ -15,8 +21,15 @@ type UId o = Metavar (Exp o) (RefInfo o)
 data HintMode = HMNormal
               | HMRecCall
 
-
-data EqReasoningConsts o = EqReasoningConsts {eqrcId, eqrcBegin, eqrcStep, eqrcEnd, eqrcSym, eqrcCong :: ConstRef o} -- "_≡_", "begin_", "_≡⟨_⟩_", "_∎", "sym", "cong"
+data EqReasoningConsts o = EqReasoningConsts
+  { eqrcId    -- "_≡_"
+  , eqrcBegin -- "begin_"
+  , eqrcStep  -- "_≡⟨_⟩_"
+  , eqrcEnd   -- "_∎"
+  , eqrcSym   -- "sym"
+  , eqrcCong  -- "cong"
+  :: ConstRef o
+  }
 
 data EqReasoningState = EqRSNone | EqRSChain | EqRSPrf1 | EqRSPrf2 | EqRSPrf3
  deriving (Eq, Show)
@@ -42,7 +55,8 @@ data RefInfo o
        --   (used to put cost when traversing a definition
        --    by construction instantiation).
     }
-  | RIUnifInfo [CAction o] (HNExp o) -- meta environment, opp hne
+  | RIUnifInfo [CAction o] (HNExp o)
+    -- meta environment, opp hne
   | RICopyInfo (ICExp o)
   | RIIotaStep Bool -- True - semiflex
   | RIInferredTypeUnknown
@@ -62,19 +76,9 @@ type MyMB a o = MB a (RefInfo o)
 
 type Nat = Int
 
--- | 'Hiding' in Agda.
-data FMode = Hidden
-           | Instance
-           | NotHidden
- deriving Eq
-
-
 data MId = Id String
          | NoId
 
-stringToMyId :: String -> MId
-stringToMyId = Id
-
 -- | Abstraction with maybe a name.
 --
 --   Different from Agda, where there is also info
@@ -99,7 +103,8 @@ data ConstDef o = ConstDef
 -- | Constant definitions.
 
 data DeclCont o
-  = Def Nat [Clause o] (Maybe Nat) (Maybe Nat) -- maybe an index to elimand argument, maybe index to elim arg if semiflex
+  = Def Nat [Clause o] (Maybe Nat) -- maybe an index to elimand argument
+                       (Maybe Nat) -- maybe index to elim arg if semiflex
   | Datatype [ConstRef o] -- constructors
              [ConstRef o] -- projection functions (in case it is a record)
 
@@ -124,6 +129,15 @@ type ConstRef o = IORef (ConstDef o)
 data Elr o
   = Var Nat
   | Const (ConstRef o)
+  deriving (Eq)
+
+getVar :: Elr o -> Maybe Nat
+getVar (Var n) = Just n
+getVar Const{} = Nothing
+
+getConst :: Elr o -> Maybe (ConstRef o)
+getConst (Const c) = Just c
+getConst Var{}     = Nothing
 
 data Sort
   = Set Nat
@@ -142,13 +156,13 @@ data Exp o
     , appElims :: MArgList o
       -- ^ Arguments.
     }
-  | Lam FMode (Abs (MExp o))
+  | Lam Hiding (Abs (MExp o))
     -- ^ Lambda with hiding information.
-  | Pi (Maybe (UId o)) FMode Bool (MExp o) (Abs (MExp o))
+  | Pi (Maybe (UId o)) Hiding Bool (MExp o) (Abs (MExp o))
     -- ^ @True@ if possibly dependent (var not known to not occur).
     --   @False@ if non-dependent.
   | Sort Sort
-  | AbsurdLambda FMode
+  | AbsurdLambda Hiding
     -- ^ Absurd lambda with hiding information.
 
 dontCare :: Exp o
@@ -160,10 +174,10 @@ type MExp o = MM (Exp o) (RefInfo o)
 data ArgList o
   = ALNil
     -- ^ No more eliminations.
-  | ALCons FMode (MExp o) (MArgList o)
+  | ALCons Hiding (MExp o) (MArgList o)
     -- ^ Application and tail.
 
-  | ALProj (MArgList o) (MM (ConstRef o) (RefInfo o)) FMode (MArgList o)
+  | ALProj (MArgList o) (MM (ConstRef o) (RefInfo o)) Hiding (MArgList o)
     -- ^ proj pre args, projfcn idx, tail
 
   | ALConPar (MArgList o)
@@ -172,19 +186,26 @@ data ArgList o
     --   Inserted to cover glitch of polymorphic constructor
     --   applications coming from Agda
 
-
 type MArgList o = MM (ArgList o) (RefInfo o)
 
-data HNExp o = HNApp [Maybe (UId o)] (Elr o) (ICArgList o)
-             | HNLam [Maybe (UId o)] FMode (Abs (ICExp o))
-             | HNPi [Maybe (UId o)] FMode Bool (ICExp o) (Abs (ICExp o))
-             | HNSort Sort
+data WithSeenUIds a o = WithSeenUIds
+  { seenUIds :: [Maybe (UId o)]
+  , rawValue :: a
+  }
+
+type HNExp o = WithSeenUIds (HNExp' o) o
+
+data HNExp' o =
+    HNApp  (Elr o) (ICArgList o)
+  | HNLam  Hiding (Abs (ICExp o))
+  | HNPi   Hiding Bool (ICExp o) (Abs (ICExp o))
+  | HNSort Sort
 
 -- | Head-normal form of 'ICArgList'.  First entry is exposed.
 --
 --   Q: Why are there no projection eliminations?
 data HNArgList o = HNALNil
-                 | HNALCons FMode (ICExp o) (ICArgList o)
+                 | HNALCons Hiding (ICExp o) (ICArgList o)
                  | HNALConPar (ICArgList o)
 
 -- | Lazy concatenation of argument lists under explicit substitutions.
@@ -233,6 +254,7 @@ detecteliminand cls =
 
 detectsemiflex :: ConstRef o -> [Clause o] -> IO Bool
 detectsemiflex _ _ = return False -- disabled
+
 categorizedecl :: ConstRef o -> IO ()
 categorizedecl c = do
  cd <- readIORef c
@@ -248,103 +270,72 @@ categorizedecl c = do
 
 -- -------------------------------------------
 
+class MetaliseOKH t where
+  metaliseOKH :: t -> IO t
+
+instance MetaliseOKH t => MetaliseOKH (MM t a) where
+  metaliseOKH e = case e of
+    Meta m -> return $ Meta m
+    NotM e -> NotM <$> metaliseOKH e
+
+instance MetaliseOKH t => MetaliseOKH (Abs t) where
+  metaliseOKH (Abs id b) = Abs id <$> metaliseOKH b
+
+instance MetaliseOKH (Exp o) where
+  metaliseOKH e = case e of
+    App uid okh elr args ->
+      (\ m -> App uid m elr) <$> (Meta <$> initMeta) <*> metaliseOKH args
+    Lam hid b -> Lam hid <$> metaliseOKH b
+    Pi uid hid dep it ot ->
+      Pi uid hid dep <$> metaliseOKH it <*> metaliseOKH ot
+    Sort{} -> return e
+    AbsurdLambda{} -> return e
+
+instance MetaliseOKH (ArgList o) where
+  metaliseOKH e = case e of
+    ALNil -> return ALNil
+    ALCons hid a as -> ALCons hid <$> metaliseOKH a <*> metaliseOKH as
+    ALProj eas idx hid as ->
+      (\ eas -> ALProj eas idx hid) <$> metaliseOKH eas <*> metaliseOKH as
+    ALConPar as -> ALConPar <$> metaliseOKH as
+
 metaliseokh :: MExp o -> IO (MExp o)
-metaliseokh = fm
- where
-  fm (Meta m) = return $ Meta m
-  fm (NotM e) = do
-   e <- f e
-   return $ NotM e
-  f (App uid _ elr args) = do
-   m <- initMeta
-   args <- fms args
-   return $ App uid (Meta m) elr args
-  f (Lam hid (Abs id b)) = do
-   b <- fm b
-   return $ Lam hid (Abs id b)
-  f (Pi uid hid posdep it (Abs id ot)) = do
-   it <- fm it
-   ot <- fm ot
-   return $ Pi uid hid posdep it (Abs id ot)
-  f e@(Sort{}) = return e
-
-  f e@(AbsurdLambda{}) = return e
-
-
-  fms (Meta m) = return $ Meta m
-  fms (NotM es) = do
-   es <- fs es
-   return $ NotM es
-  fs ALNil = return ALNil
-  fs (ALCons hid a as) = do
-   a <- fm a
-   as <- fms as
-   return $ ALCons hid a as
-
-  fs (ALProj eas idx hid as) = do
-   eas <- fms eas
-   as <- fms as
-   return $ ALProj eas idx hid as
-
-
-  fs (ALConPar as) = do
-   as <- fms as
-   return $ ALConPar as
-
+metaliseokh = metaliseOKH
 
 -- -------------------------------------------
 
-expandExp :: MExp o -> IO (MExp o)
-expandExp = fm
- where
-  fm (Meta m) = do
-   mb <- readIORef $ mbind m
-   case mb of
-    Nothing -> return $ Meta m
-    Just e -> fm (NotM e)
-  fm (NotM e) = do
-   e <- f e
-   return $ NotM e
-  f (App uid okh elr args) = do
-   args <- fms args
-   return $ App uid okh elr args
-  f (Lam hid (Abs id b)) = do
-   b <- fm b
-   return $ Lam hid (Abs id b)
-  f (Pi uid hid posdep it (Abs id ot)) = do
-   it <- fm it
-   ot <- fm ot
-   return $ Pi uid hid posdep it (Abs id ot)
-  f e@(Sort{}) = return e
+class ExpandMetas t where
+  expandMetas :: t -> IO t
 
-  f e@(AbsurdLambda{}) = return e
+instance ExpandMetas t => ExpandMetas (MM t a) where
+  expandMetas t = case t of
+    NotM e -> NotM <$> expandMetas e
+    Meta m -> do
+      mb <- readIORef (mbind m)
+      case mb of
+        Nothing -> return $ Meta m
+        Just e  -> NotM <$> expandMetas e
 
+instance ExpandMetas t => ExpandMetas (Abs t) where
+  expandMetas (Abs id b) = Abs id <$> expandMetas b
 
-  fms (Meta m) = do
-   mb <- readIORef $ mbind m
-   case mb of
-    Nothing -> return $ Meta m
-    Just es -> fms (NotM es)
-  fms (NotM es) = do
-   es <- fs es
-   return $ NotM es
-  fs ALNil = return ALNil
-  fs (ALCons hid a as) = do
-   a <- fm a
-   as <- fms as
-   return $ ALCons hid a as
-
-  fs (ALProj eas idx hid as) = do
-   idx <- expandbind idx
-   eas <- fms eas
-   as <- fms as
-   return $ ALProj eas idx hid as
-
-
-  fs (ALConPar as) = do
-   as <- fms as
-   return $ ALConPar as
+instance ExpandMetas (Exp o) where
+  expandMetas t = case t of
+    App uid okh elr args -> App uid okh elr <$> expandMetas args
+    Lam hid b -> Lam hid <$> expandMetas b
+    Pi uid hid dep it ot ->
+      Pi uid hid dep <$> expandMetas it <*> expandMetas ot
+    Sort{} -> return t
+    AbsurdLambda{} -> return t
 
+instance ExpandMetas (ArgList o) where
+  expandMetas e = case e of
+    ALNil -> return ALNil
+    ALCons hid a as -> ALCons hid <$> expandMetas a <*> expandMetas as
+    ALProj eas idx hid as ->
+      (\ a b -> ALProj a b hid) <$> expandMetas eas
+      <*> expandbind idx <*> expandMetas as
+    ALConPar as -> ALConPar <$> expandMetas as
 
 -- ---------------------------------
 
@@ -370,22 +361,26 @@ subi :: MExp o -> ICExp o -> ICExp o
 subi e (Clos (Skip : as) x) = Clos (Sub (Clos [] e) : as) x
 subi _ _ = __IMPOSSIBLE__
 
-weak :: Nat -> CExp o -> CExp o
-weak n (TrBr trs e) = TrBr trs (weaki n e)
+weak :: Weakening t => Nat -> t -> t
+weak 0 = id
+weak n = weak' n
 
-weaki :: Nat -> Clos a o -> Clos a o
-weaki 0 x = x
-weaki n (Clos as x) = Clos (Weak n : as) x
+class Weakening t where
+  weak' :: Nat -> t -> t
 
-weakarglist :: Nat -> ICArgList o -> ICArgList o
-weakarglist 0 = id
-weakarglist n = f
- where f CALNil = CALNil
-       f (CALConcat (Clos cl as) as2) = CALConcat (Clos (Weak n : cl) as) (f as2)
-weakelr :: Nat -> Elr o -> Elr o
-weakelr 0 elr = elr
-weakelr n (Var v) = Var (v + n)
-weakelr _ elr@(Const _) = elr
+instance Weakening a => Weakening (TrBr a o) where
+  weak' n (TrBr trs e) = TrBr trs (weak' n e)
+
+instance Weakening (Clos a o) where
+  weak' n (Clos as x) = Clos (Weak n : as) x
+
+instance Weakening (ICArgList o) where
+  weak' n e = case e of
+    CALNil         -> CALNil
+    CALConcat a as -> CALConcat (weak' n a) (weak' n as)
+
+instance Weakening (Elr o) where
+  weak' n = rename (n+)
 
 -- | Substituting for a variable.
 doclos :: [CAction o] -> Nat -> Either Nat (ICExp o)
@@ -394,7 +389,82 @@ doclos = f 0
   -- ns is the number of weakenings
   f ns []            i = Left (ns + i)
   f ns (Weak n : xs) i = f (ns + n) xs i
-  f ns (Sub s  : _ ) 0 = Right (weaki ns s)
+  f ns (Sub s  : _ ) 0 = Right (weak ns s)
   f ns (Skip   : _ ) 0 = Left ns
   f ns (Skip   : xs) i = f (ns + 1) xs (i - 1)
   f ns (Sub _  : xs) i = f ns xs (i - 1)
+
+
+-- | FreeVars class and instances
+
+freeVars :: FreeVars t => t -> Set.Set Nat
+freeVars = freeVarsOffset 0
+
+class FreeVars t where
+  freeVarsOffset :: Nat -> t -> Set.Set Nat
+
+instance (FreeVars a, FreeVars b) => FreeVars (a, b) where
+  freeVarsOffset n (a, b) = Set.union (freeVarsOffset n a) (freeVarsOffset n b)
+
+instance FreeVars t => FreeVars (MM t a) where
+  freeVarsOffset n e = freeVarsOffset n (rm __IMPOSSIBLE__ e)
+
+instance FreeVars t => FreeVars (Abs t) where
+  freeVarsOffset n (Abs id e) = freeVarsOffset (n + 1) e
+
+instance FreeVars (Elr o) where
+  freeVarsOffset n e = case e of
+    Var v   -> Set.singleton (v - n)
+    Const{} -> Set.empty
+
+instance FreeVars (Exp o) where
+  freeVarsOffset n e = case e of
+   App _ _ elr args -> freeVarsOffset n (elr, args)
+   Lam _ b          -> freeVarsOffset n b
+   Pi _ _ _ it ot   -> freeVarsOffset n (it, ot)
+   Sort{}           -> Set.empty
+   AbsurdLambda{}   -> Set.empty
+
+instance FreeVars (ArgList o) where
+  freeVarsOffset n es = case es of
+    ALNil         -> Set.empty
+    ALCons _ e es -> freeVarsOffset n (e, es)
+    ALConPar es   -> freeVarsOffset n es
+    ALProj{}      -> __IMPOSSIBLE__
+
+
+-- | Renaming Typeclass and instances
+rename :: Renaming t => (Nat -> Nat) -> t -> t
+rename = renameOffset 0
+
+class Renaming t where
+  renameOffset :: Nat -> (Nat -> Nat) -> t -> t
+
+instance (Renaming a, Renaming b) => Renaming (a, b) where
+  renameOffset j ren (a, b) = (renameOffset j ren a, renameOffset j ren b)
+
+instance Renaming t => Renaming (MM t a) where
+  renameOffset j ren e = NotM $ renameOffset j ren (rm __IMPOSSIBLE__ e)
+
+instance Renaming t => Renaming (Abs t) where
+  renameOffset j ren (Abs id e) = Abs id $ renameOffset (j + 1) ren e
+
+instance Renaming (Elr o) where
+  renameOffset j ren e = case e of
+    Var v | v >= j -> Var (ren (v - j) + j)
+    _              -> e
+
+instance Renaming (Exp o) where
+  renameOffset j ren e = case e of
+    App uid ok elr args -> uncurry (App uid ok) $ renameOffset j ren (elr, args)
+    Lam hid e -> Lam hid (renameOffset j ren e)
+    Pi a b c it ot -> uncurry (Pi a b c) $ renameOffset j ren (it, ot)
+    Sort{}         -> e
+    AbsurdLambda{} -> e
+
+instance Renaming (ArgList o) where
+  renameOffset j ren e = case e of
+    ALNil           -> ALNil
+    ALCons hid a as -> uncurry (ALCons hid) $ renameOffset j ren (a, as)
+    ALConPar as     -> ALConPar (renameOffset j ren as)
+    ALProj{}        -> __IMPOSSIBLE__
