@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Compiler.Common where
 
@@ -14,6 +15,8 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Char
 import Data.Function
+import Data.Semigroup
+import Data.Monoid hiding ((<>))
 
 import Control.Monad
 import Control.Monad.State  hiding (mapM_, forM_, mapM, forM, sequence)
@@ -28,7 +31,7 @@ import Agda.Interaction.Imports
 import Agda.Interaction.Options
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Pretty hiding ((<>))
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
@@ -38,7 +41,7 @@ import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
-import Agda.Utils.Pretty
+import Agda.Utils.Pretty hiding ((<>))
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -46,7 +49,16 @@ import Agda.Utils.Impossible
 data IsMain = IsMain | NotMain
   deriving (Eq, Show)
 
-doCompile :: IsMain -> Interface -> (IsMain -> Interface -> TCM ()) -> TCM ()
+instance Semigroup IsMain where
+  NotMain <> _ = NotMain
+  _       <> NotMain = NotMain
+  IsMain  <> IsMain = IsMain
+
+instance Monoid IsMain where
+  mempty = IsMain
+  mappend = (<>)
+
+doCompile :: forall r. Monoid r => IsMain -> Interface -> (IsMain -> Interface -> TCM r) -> TCM r
 doCompile isMain i f = do
   -- The Agda.Primitive module is implicitly assumed to be always imported,
   -- even though it not necesseraly occurs in iImportedModules.
@@ -54,26 +66,27 @@ doCompile isMain i f = do
   [agdaPrimInter] <- filter (("Agda.Primitive"==) . prettyShow . iModuleName)
     . map miInterface . Map.elems
       <$> getVisitedModules
-  flip evalStateT Set.empty $ comp NotMain agdaPrimInter >> comp isMain i
+  flip evalStateT Set.empty $ mappend <$> comp NotMain agdaPrimInter <*> comp isMain i
   where
-    comp :: IsMain -> Interface -> StateT (Set ModuleName) TCM ()
+    comp :: IsMain -> Interface -> StateT (Set ModuleName) TCM r
     comp isMain i = do
       alreadyDone <- Set.member (iModuleName i) <$> get
-      when (not alreadyDone) $ do
+      if alreadyDone then return mempty else do
         imps <- lift $
           map miInterface . catMaybes <$>
             mapM (getVisitedModule . toTopLevelModuleName . fst) (iImportedModules i)
-        mapM_ (comp NotMain) imps
+        ri <- mconcat <$> mapM (comp NotMain) imps
         lift $ setInterface i
-        lift $ f isMain i
+        r <- lift $ f isMain i
         modify (Set.insert $ iModuleName i)
+        return $ mappend ri r
 
 setInterface :: Interface -> TCM ()
 setInterface i = do
   opts <- gets (stPersistentOptions . stPersistentState)
   setCommandLineOptions opts
   mapM_ setOptionsFromPragma (iPragmaOptions i)
-  stImportedModules .= Set.empty
+  stImportedModules .= Set.fromList (map fst $ iImportedModules i)
   stCurrentModule   .= Just (iModuleName i)
 
 curIF :: TCM Interface
@@ -132,18 +145,19 @@ repl subs = go where
 conArityAndPars :: QName -> TCM (Nat, Nat)
 conArityAndPars q = do
   def <- getConstInfo q
-  TelV tel _ <- telView $ defType def
+  n   <- typeArity (defType def)
   let Constructor{ conPars = np } = theDef def
-      n = length (telToList tel)
   return (n - np, np)
 
 -- | Sets up the compilation environment.
 inCompilerEnv :: Interface -> TCM a -> TCM a
-inCompilerEnv mainI cont =
+inCompilerEnv mainI cont = do
   -- Preserve the state (the compiler modifies the state).
   -- Andreas, 2014-03-23 But we might want to collect Benchmark info,
   -- so use localTCState.
-  localTCState $ do
+  -- FNF, 2017-02-22 we also want to keep the warnings we have encountered,
+  -- so use localTCStateSaving and pick them out.
+  (a , s) <- localTCStateSaving $ do
 
     -- Compute the output directory. Note: using commandLineOptions would make
     -- the current pragma options persistent when we setCommandLineOptions
@@ -159,8 +173,20 @@ inCompilerEnv mainI cont =
     setCommandLineOptions $
       opts { optCompileDir = Just compileDir }
 
+    -- Andreas, 2017-08-23, issue #2714 recover pragma option --no-main
+    -- Unfortunately, a pragma option is stored in the interface file as
+    -- just a list of strings, thus, the solution is a bit of hack:
+    -- We match on whether @["--no-main"]@ is one of the stored options.
+    when (["--no-main"] `elem` iPragmaOptions mainI) $
+      stPragmaOptions %= \ o -> o { optCompileNoMain = True }
+
+    setScope (iInsideScope mainI) -- so that compiler errors don't use overly qualified names
     ignoreAbstractMode $ do
       cont
+  -- keep generated warnings
+  let newWarnings = stPostTCWarnings $  stPostScopeState $ s
+  stTCWarnings .= newWarnings
+  return a
 
 topLevelModuleName :: ModuleName -> TCM ModuleName
 topLevelModuleName m = do
