@@ -6,6 +6,7 @@ import Prelude hiding ( null, writeFile )
 import Control.Applicative
 import Control.Monad.Reader ( liftIO )
 import Control.Monad.Trans
+import Data.Char ( isSpace )
 import Data.List ( intercalate, genericLength, partition )
 import Data.Maybe ( isJust )
 import Data.Set ( Set, null, insert, difference, delete )
@@ -20,7 +21,7 @@ import Agda.Interaction.FindFile ( findFile, findInterfaceFile )
 import Agda.Interaction.Imports ( isNewerThan )
 import Agda.Interaction.Options ( optCompileDir )
 import Agda.Syntax.Common ( Nat, unArg, namedArg, NameId(..) )
-import Agda.Syntax.Concrete.Name ( projectRoot )
+import Agda.Syntax.Concrete.Name ( projectRoot , isNoName )
 import Agda.Syntax.Abstract.Name
   ( ModuleName(MName), QName,
     mnameToConcrete,
@@ -29,6 +30,7 @@ import Agda.Syntax.Internal
   ( Name, Args, Type,
     conName,
     toTopLevelModuleName, arity, unEl, unAbs, nameFixity )
+import Agda.Syntax.Position
 import Agda.Syntax.Literal ( Literal(LitNat,LitFloat,LitString,LitChar,LitQName,LitMeta) )
 import Agda.Syntax.Fixity
 import qualified Agda.Syntax.Treeless as T
@@ -36,7 +38,8 @@ import Agda.TypeChecking.Substitute ( absBody )
 import Agda.TypeChecking.Level ( reallyUnLevelView )
 import Agda.TypeChecking.Monad hiding (Global, Local)
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Options ( setCommandLineOptions, commandLineOptions, reportSLn )
+import Agda.TypeChecking.Monad.Debug ( reportSLn )
+import Agda.TypeChecking.Monad.Options ( setCommandLineOptions )
 import Agda.TypeChecking.Reduce ( instantiateFull, normalise )
 import Agda.TypeChecking.Substitute (TelV(..))
 import Agda.TypeChecking.Telescope
@@ -53,9 +56,10 @@ import qualified Agda.Utils.HashMap as HMap
 
 import Agda.Compiler.Common
 import Agda.Compiler.ToTreeless
-import Agda.Compiler.Treeless.DelayCoinduction
+import Agda.Compiler.Treeless.EliminateDefaults
 import Agda.Compiler.Treeless.EliminateLiteralPatterns
 import Agda.Compiler.Treeless.GuardsToPrims
+import Agda.Compiler.Backend (Backend(..), Backend'(..), Recompile(..))
 
 import Agda.Compiler.JS.Syntax
   ( Exp(Self,Local,Global,Undefined,String,Char,Integer,Double,Lambda,Object,Apply,Lookup,If,BinOp,PlainJS),
@@ -65,6 +69,8 @@ import Agda.Compiler.JS.Substitution
   ( curriedLambda, curriedApply, emp, subst, apply )
 import qualified Agda.Compiler.JS.Pretty as JSPretty
 
+import Agda.Interaction.Options
+
 import Paths_Agda
 
 #include "undefined.h"
@@ -74,24 +80,80 @@ import Agda.Utils.Impossible ( Impossible(Impossible), throwImpossible )
 -- Entry point into the compiler
 --------------------------------------------------
 
-compilerMain :: Interface -> TCM ()
-compilerMain mainI = inCompilerEnv mainI $ do
-  doCompile IsMain mainI $ do
-    compile
-  copyRTEModules
+jsBackend :: Backend
+jsBackend = Backend jsBackend'
 
-compile :: IsMain -> Interface -> TCM ()
-compile isMain i = do
-  ifM uptodate noComp $ do
-    yesComp
-    writeModule =<< curModule isMain
+jsBackend' :: Backend' JSOptions JSOptions JSModuleEnv () (Maybe Export)
+jsBackend' = Backend'
+  { backendName           = jsBackendName
+  , backendVersion        = Nothing
+  , options               = defaultJSOptions
+  , commandLineFlags      = jsCommandLineFlags
+  , isEnabled             = optJSCompile
+  , preCompile            = jsPreCompile
+  , postCompile           = jsPostCompile
+  , preModule             = jsPreModule
+  , postModule            = jsPostModule
+  , compileDef            = jsCompileDef
+  , scopeCheckingSuffices = False
+  }
+
+--- Options ---
+
+data JSOptions = JSOptions
+  { optJSCompile :: Bool }
+
+defaultJSOptions :: JSOptions
+defaultJSOptions = JSOptions
+  { optJSCompile = False }
+
+jsCommandLineFlags :: [OptDescr (Flag JSOptions)]
+jsCommandLineFlags =
+    [ Option [] ["js"] (NoArg enable) "compile program using the JS backend"
+    ]
   where
-  uptodate = liftIO =<< (isNewerThan <$> outFile_ <*> ifile)
-  ifile    = maybe __IMPOSSIBLE__ filePath <$>
-               (findInterfaceFile . toTopLevelModuleName =<< curMName)
-  noComp   = reportSLn "compile.js" 2 . (++ " : no compilation is needed.") . prettyShow =<< curMName
-  yesComp  = reportSLn "compile.js" 1 . (`repl` "Compiling <<0>> in <<1>> to <<2>>") =<<
-             sequence [prettyShow <$> curMName, ifile, outFile_] :: TCM ()
+    enable o = pure o{ optJSCompile = True }
+
+--- Top-level compilation ---
+
+jsPreCompile :: JSOptions -> TCM JSOptions
+jsPreCompile opts = return opts
+
+jsPostCompile :: JSOptions -> IsMain -> a -> TCM ()
+jsPostCompile _ _ _ = copyRTEModules
+
+--- Module compilation ---
+
+type JSModuleEnv = Maybe CoinductionKit
+
+jsPreModule :: JSOptions -> ModuleName -> FilePath -> TCM (Recompile JSModuleEnv ())
+jsPreModule _ m ifile = ifM uptodate noComp yesComp
+  where
+    uptodate = liftIO =<< isNewerThan <$> outFile_ <*> pure ifile
+
+    noComp = do
+      reportSLn "compile.js" 2 . (++ " : no compilation is needed.") . prettyShow =<< curMName
+      return $ Skip ()
+
+    yesComp = do
+      m   <- prettyShow <$> curMName
+      out <- outFile_
+      reportSLn "compile.js" 1 $ repl [m, ifile, out] "Compiling <<0>> in <<1>> to <<2>>"
+      Recompile <$> coinductionKit
+
+jsPostModule :: JSOptions -> JSModuleEnv -> IsMain -> ModuleName -> [Maybe Export] -> TCM ()
+jsPostModule _ _ isMain _ defs = do
+  m             <- jsMod <$> curMName
+  is            <- map (jsMod . fst) . iImportedModules <$> curIF
+  let es = catMaybes defs
+  writeModule $ Module m (reorder es) main
+  where
+    main = case isMain of
+      IsMain  -> Just $ Apply (Lookup Self $ MemberId "main") [Lambda 1 emp]
+      NotMain -> Nothing
+
+jsCompileDef :: JSOptions -> JSModuleEnv -> Definition -> TCM (Maybe Export)
+jsCompileDef _ kit def = definition kit (defName def, def)
 
 --------------------------------------------------
 -- Naming
@@ -101,19 +163,18 @@ prefix :: [Char]
 prefix = "jAgda"
 
 jsMod :: ModuleName -> GlobalId
-jsMod m = GlobalId (prefix : map show (mnameToList m))
+jsMod m = GlobalId (prefix : map prettyShow (mnameToList m))
 
 jsFileName :: GlobalId -> String
 jsFileName (GlobalId ms) = intercalate "." ms ++ ".js"
 
 jsMember :: Name -> MemberId
-jsMember n =
+jsMember n
   -- Anonymous fields are used for where clauses,
   -- and they're all given the concrete name "_",
   -- so we disambiguate them using their name id.
-  case show n of
-    "_" -> MemberId ("_" ++ show (nameId n))
-    l   -> MemberId l
+  | isNoName n = MemberId ("_" ++ show (nameId n))
+  | otherwise  = MemberId $ prettyShow n
 
 -- Rather annoyingly, the anonymous construtor of a record R in module M
 -- is given the name M.recCon, but a named constructor C
@@ -169,13 +230,13 @@ reorder' defs (e : es) =
     False -> reorder' defs (insertAfter us e es)
 
 isTopLevelValue :: Export -> Bool
-isTopLevelValue (Export _ isCoind e) = case e of
-  _ | isCoind -> False
+isTopLevelValue (Export _ e) = case e of
+  Object m | flatName `Map.member` m -> False
   Lambda{} -> False
   _        -> True
 
 isEmptyObject :: Export -> Bool
-isEmptyObject (Export _ _ e) = case e of
+isEmptyObject (Export _ e) = case e of
   Object m -> Map.null m
   Lambda{} -> True
   _        -> False
@@ -209,30 +270,49 @@ definition kit (q,d) = do
 
   definition' kit q d (defType d) ls
 
+-- | Ensure that there is at most one pragma for a name.
+checkCompilerPragmas :: QName -> TCM ()
+checkCompilerPragmas q =
+  caseMaybeM (getUniqueCompilerPragma jsBackendName q) (return ()) $ \ (CompilerPragma r s) ->
+  setCurrentRange r $ case words s of
+    "=" : _ -> return ()
+    _       -> genericDocError $ P.sep [ P.text "Badly formed COMPILE JS pragma. Expected",
+                                         P.text "{-# COMPILE JS <name> = <js> #-}" ]
+
+defJSDef :: Definition -> Maybe String
+defJSDef def =
+  case defCompilerPragmas jsBackendName def of
+    [CompilerPragma _ s] -> Just (dropEquals s)
+    []                   -> Nothing
+    _:_:_                -> __IMPOSSIBLE__
+  where
+    dropEquals = dropWhile $ \ c -> isSpace c || c == '='
+
 definition' :: Maybe CoinductionKit -> QName -> Definition -> Type -> [MemberId] -> TCM (Maybe Export)
-definition' kit q d t ls =
+definition' kit q d t ls = do
+  checkCompilerPragmas q
   case theDef d of
     -- coinduction
     Constructor{} | Just q == (nameOfSharp <$> kit) -> do
-      ret $  Lambda 1 $ local 0
+      return Nothing
     Function{} | Just q == (nameOfFlat <$> kit) -> do
-      ret $ Lambda 1 $ Apply (local 0) [Integer 0]
+      ret $ Lambda 1 $ Apply (Lookup (local 0) flatName) []
 
     Axiom | Just e <- defJSDef d -> plainJS e
     Axiom | otherwise -> ret Undefined
 
     Function{} | Just e <- defJSDef d -> plainJS e
     Function{} | otherwise -> do
-      isInfVal <- outputIsInf kit t
 
       reportSDoc "compile.js" 5 $ text "compiling fun:" <+> prettyTCM q
       caseMaybeM (toTreeless q) (pure Nothing) $ \ treeless -> do
-        funBody <- eliminateLiteralPatterns $ convertGuards $ treeless
-        funBody' <- delayCoinduction funBody t
-        reportSDoc "compile.js" 30 $ text " compiled treeless fun:" <+> pretty funBody'
-        funBody'' <- compileTerm funBody'
-        reportSDoc "compile.js" 30 $ text " compiled JS fun:" <+> (text . show) funBody''
-        return $ Just $ Export ls isInfVal funBody''
+        funBody <- eliminateCaseDefaults =<<
+          eliminateLiteralPatterns
+          (convertGuards treeless)
+        reportSDoc "compile.js" 30 $ text " compiled treeless fun:" <+> pretty funBody
+        funBody' <- compileTerm funBody
+        reportSDoc "compile.js" 30 $ text " compiled JS fun:" <+> (text . show) funBody'
+        return $ Just $ Export ls funBody'
 
     Primitive{primName = p} | p `Set.member` primitives ->
       plainJS $ "agdaRTS." ++ p
@@ -259,53 +339,72 @@ definition' kit q d t ls =
             (Apply (Lookup (Local (LocalId 0)) (last ls))
               [ Local (LocalId (np - i)) | i <- [0 .. np-1] ]))
 
-    AbstractDefn -> __IMPOSSIBLE__
+    AbstractDefn{} -> __IMPOSSIBLE__
   where
-    ret = return . Just . Export ls False
-    plainJS = return . Just . Export ls False . PlainJS
-
+    ret = return . Just . Export ls
+    plainJS = return . Just . Export ls . PlainJS
 
 compileTerm :: T.TTerm -> TCM Exp
-compileTerm term = do
-  case term of
-    T.TVar x -> return $ Local $ LocalId x
-    T.TDef q -> do
-      d <- getConstInfo q
-      case theDef d of
-        -- Datatypes and records are erased
-        Datatype {} -> return (String "*")
-        Record {} -> return (String "*")
-        _ -> qname q
-    T.TApp t xs -> curriedApply <$> compileTerm t <*> mapM compileTerm xs
-    T.TLam t -> Lambda 1 <$> compileTerm t
-    -- TODO This is not a lazy let, but it should be...
-    T.TLet t e -> apply <$> (Lambda 1 <$> compileTerm e) <*> traverse compileTerm [t]
-    T.TLit l -> return $ literal l
-    T.TCon q -> do
-      d <- getConstInfo q
-      qname q
-    T.TCase sc (T.CTData dt) def alts -> do
-      dt <- getConstInfo dt
-      alts' <- traverse compileAlt alts
-      let obj = Object $ Map.fromList alts'
-      case (theDef dt, defJSDef dt) of
-        (_, Just e) -> do
-          return $ apply (PlainJS e) [Local (LocalId sc), obj]
-        (Record{}, _) -> do
-          memId <- visitorName $ recCon $ theDef dt
-          return $ apply (Lookup (Local $ LocalId sc) memId) [obj]
-        (Datatype{}, _) -> do
-          return $ curriedApply (Local (LocalId sc)) [obj]
-        _ -> __IMPOSSIBLE__
-    T.TCase _ _ _ _ -> __IMPOSSIBLE__
-
-    T.TPrim p -> return $ compilePrim p
-    T.TUnit -> unit
-    T.TSort -> unit
-    T.TErased -> unit
-    T.TError T.TUnreachable -> return Undefined
+compileTerm t = do
+  kit <- coinductionKit
+  compileTerm' kit t
 
+compileTerm' :: Maybe CoinductionKit -> T.TTerm -> TCM Exp
+compileTerm' kit t = go t
   where
+    go :: T.TTerm -> TCM Exp
+    go t = case t of
+      T.TVar x -> return $ Local $ LocalId x
+      T.TDef q -> do
+        d <- getConstInfo q
+        case theDef d of
+          -- Datatypes and records are erased
+          Datatype {} -> return (String "*")
+          Record {} -> return (String "*")
+          _ -> qname q
+      T.TApp (T.TCon q) [x] | Just q == (nameOfSharp <$> kit) -> do
+        x <- go x
+        let evalThunk = unlines
+              [ "function() {"
+              , "  delete this.flat;"
+              , "  var result = this.__flat_helper();"
+              , "  delete this.__flat_helper;"
+              , "  this.flat = function() { return result; };"
+              , "  return result;"
+              , "}"
+              ]
+        return $ Object $ Map.fromList
+          [(flatName, PlainJS evalThunk)
+          ,(MemberId "__flat_helper", Lambda 0 x)]
+      T.TApp t xs -> curriedApply <$> go t <*> mapM go xs
+      T.TLam t -> Lambda 1 <$> go t
+      -- TODO This is not a lazy let, but it should be...
+      T.TLet t e -> apply <$> (Lambda 1 <$> go e) <*> traverse go [t]
+      T.TLit l -> return $ literal l
+      T.TCon q -> do
+        d <- getConstInfo q
+        qname q
+      T.TCase sc (T.CTData dt) def alts -> do
+        dt <- getConstInfo dt
+        alts' <- traverse compileAlt alts
+        let obj = Object $ Map.fromList alts'
+        case (theDef dt, defJSDef dt) of
+          (_, Just e) -> do
+            return $ apply (PlainJS e) [Local (LocalId sc), obj]
+          (Record{}, _) -> do
+            memId <- visitorName $ recCon $ theDef dt
+            return $ apply (Lookup (Local $ LocalId sc) memId) [obj]
+          (Datatype{}, _) -> do
+            return $ curriedApply (Local (LocalId sc)) [obj]
+          _ -> __IMPOSSIBLE__
+      T.TCase _ _ _ _ -> __IMPOSSIBLE__
+
+      T.TPrim p -> return $ compilePrim p
+      T.TUnit -> unit
+      T.TSort -> unit
+      T.TErased -> unit
+      T.TError T.TUnreachable -> return Undefined
+
     unit = return $ Integer 0
 
 compilePrim :: T.TPrim -> Exp
@@ -321,6 +420,8 @@ compilePrim p =
     T.PAdd -> binOp "agdaRTS.uprimIntegerPlus"
     T.PSub -> binOp "agdaRTS.uprimIntegerMinus"
     T.PMul -> binOp "agdaRTS.uprimIntegerMultiply"
+    T.PRem -> binOp "agdaRTS.uprimIntegerRem"
+    T.PQuot -> binOp "agdaRTS.uprimIntegerQuot"
     T.PSeq -> binOp "agdaRTS.primSeq"
     _ -> __IMPOSSIBLE__
   where binOp js = curriedLambda 2 $ apply (PlainJS js) [local 1, local 0]
@@ -337,6 +438,9 @@ compileAlt a = case a of
 visitorName :: QName -> TCM MemberId
 visitorName q = do (m,ls) <- global q; return (last ls)
 
+flatName :: MemberId
+flatName = MemberId "flat"
+
 local :: Nat -> Exp
 local = Local . LocalId
 
@@ -359,7 +463,7 @@ litqname q =
   Object $ Map.fromList
     [ (mem "id", Integer $ fromIntegral n)
     , (mem "moduleId", Integer $ fromIntegral m)
-    , (mem "name", String $ show q)
+    , (mem "name", String $ prettyShow q)
     , (mem "fixity", litfixity fx)]
   where
     mem = MemberId
@@ -372,11 +476,11 @@ litqname q =
       , (mem "prec", litPrec $ fixityLevel fx)]
 
     -- TODO this will probably not work well together with the necessary FFI bindings
-    litAssoc NonAssoc   = String "NonAssoc"
-    litAssoc LeftAssoc  = String "LeftAssoc"
-    litAssoc RightAssoc = String "RightAssoc"
+    litAssoc NonAssoc   = String "non-assoc"
+    litAssoc LeftAssoc  = String "left-assoc"
+    litAssoc RightAssoc = String "right-assoc"
 
-    litPrec Unrelated   = String "Unrelated"
+    litPrec Unrelated   = String "unrelated"
     litPrec (Related l) = Integer l
 
 --------------------------------------------------
@@ -434,4 +538,6 @@ primitives = Set.fromList
   , "primATan2"
   , "primShowQName"
   , "primQNameEquality"
+  , "primQNameLess"
+  , "primQNameFixity"
   ]
