@@ -2,8 +2,10 @@
 
 module Agda.Compiler.JS.Syntax where
 
+import Data.Map ( Map )
+import qualified Data.Map as Map
+
 import Data.Typeable ( Typeable )
-import Data.Map ( Map, fold )
 import Data.Set ( Set, empty, singleton, union )
 
 import Agda.Syntax.Common ( Nat )
@@ -47,7 +49,7 @@ newtype MemberId = MemberId String
 -- The top-level compilation unit is a module, which names
 -- the GId of its exports, and a list of definitions
 
-data Export = Export { expName :: [MemberId], isCoind :: Bool, defn :: Exp }
+data Export = Export { expName :: [MemberId], defn :: Exp }
   deriving (Typeable, Show)
 
 data Module = Module { modName :: GlobalId, exports :: [Export], postscript :: Maybe Exp }
@@ -65,10 +67,10 @@ instance Uses a => Uses [a] where
   uses = foldr (union . uses) empty
 
 instance Uses a => Uses (Map k a) where
-  uses = fold (union . uses) empty
+  uses = Map.foldr (union . uses) empty
 
 instance Uses Exp where
-  uses (Object o)     = fold (union . uses) empty o
+  uses (Object o)     = Map.foldr (union . uses) empty o
   uses (Apply e es)   = foldr (union . uses) (uses e) es
   uses (Lookup e l)   = uses' e [l] where
       uses' Self         ls = singleton ls
@@ -80,7 +82,7 @@ instance Uses Exp where
   uses e              = empty
 
 instance Uses Export where
-  uses (Export _ _ e) = uses e
+  uses (Export _ e) = uses e
 
 -- All global ids
 
@@ -91,7 +93,7 @@ instance Globals a => Globals [a] where
   globals = foldr (union . globals) empty
 
 instance Globals a => Globals (Map k a) where
-  globals = fold (union . globals) empty
+  globals = Map.foldr (union . globals) empty
 
 instance Globals Exp where
   globals (Global i) = singleton i
@@ -105,7 +107,7 @@ instance Globals Exp where
   globals _ = empty
 
 instance Globals Export where
-  globals (Export _ _ e) = globals e
+  globals (Export _ e) = globals e
 
 instance Globals Module where
   globals (Module m es _) = globals es
