@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
 
 module Agda.Compiler.MAlonzo.Compiler where
 
@@ -21,6 +20,7 @@ import Data.Maybe
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable hiding (for)
+import Data.Monoid hiding ((<>))
 
 import Numeric.IEEE
 
@@ -34,9 +34,12 @@ import Agda.Compiler.Common
 import Agda.Compiler.MAlonzo.Misc
 import Agda.Compiler.MAlonzo.Pretty
 import Agda.Compiler.MAlonzo.Primitives
+import Agda.Compiler.MAlonzo.HaskellTypes
+import Agda.Compiler.MAlonzo.Pragmas
 import Agda.Compiler.ToTreeless
 import Agda.Compiler.Treeless.Unused
 import Agda.Compiler.Treeless.Erase
+import Agda.Compiler.Backend
 
 import Agda.Interaction.FindFile
 import Agda.Interaction.Imports
@@ -60,6 +63,7 @@ import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Level (reallyUnLevelView)
+import Agda.TypeChecking.Warnings
 
 import Agda.TypeChecking.CompiledClause
 
@@ -74,6 +78,7 @@ import Agda.Utils.Pretty (prettyShow, Pretty)
 import qualified Agda.Utils.IO.UTF8 as UTF8
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Singleton
+import Agda.Utils.Size
 import Agda.Utils.Tuple
 
 import Paths_Agda
@@ -81,27 +86,102 @@ import Paths_Agda
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-compilerMain :: IsMain -> Interface -> TCM ()
-compilerMain isMain i =
-  inCompilerEnv i $ do
-    doCompile isMain i $
-      \_ -> compile
-    copyRTEModules
-    callGHC isMain i
+-- The backend callbacks --------------------------------------------------
 
-compile :: Interface -> TCM ()
-compile i = do
-  ifM uptodate noComp $ {- else -} do
-    yesComp
-    writeModule =<< decl <$> curHsMod <*> (definitions =<< curDefs) <*> imports
+ghcBackend :: Backend
+ghcBackend = Backend ghcBackend'
+
+ghcBackend' :: Backend' GHCOptions GHCOptions GHCModuleEnv IsMain [HS.Decl]
+ghcBackend' = Backend'
+  { backendName           = "GHC"
+  , backendVersion        = Nothing
+  , options               = defaultGHCOptions
+  , commandLineFlags      = ghcCommandLineFlags
+  , isEnabled             = optGhcCompile
+  , preCompile            = ghcPreCompile
+  , postCompile           = ghcPostCompile
+  , preModule             = ghcPreModule
+  , postModule            = ghcPostModule
+  , compileDef            = ghcCompileDef
+  , scopeCheckingSuffices = False
+  }
+
+--- Options ---
+
+data GHCOptions = GHCOptions
+  { optGhcCompile :: Bool
+  , optGhcCallGhc :: Bool
+  , optGhcFlags   :: [String]
+  }
+
+defaultGHCOptions :: GHCOptions
+defaultGHCOptions = GHCOptions
+  { optGhcCompile = False
+  , optGhcCallGhc = True
+  , optGhcFlags   = []
+  }
+
+ghcCommandLineFlags :: [OptDescr (Flag GHCOptions)]
+ghcCommandLineFlags =
+    [ Option ['c']  ["compile", "ghc"] (NoArg enable)
+                    "compile program using the GHC backend"
+    , Option []     ["ghc-dont-call-ghc"] (NoArg dontCallGHC)
+                    "don't call GHC, just write the GHC Haskell files."
+    , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
+                    "give the flag GHC-FLAG to GHC"
+    ]
   where
-  decl mn ds imp = HS.Module mn [] imp (map fakeDecl (reverse $ iHaskellCode i) ++ ds)
-  uptodate = liftIO =<< (isNewerThan <$> outFile_ <*> ifile)
-  ifile    = maybe __IMPOSSIBLE__ filePath <$>
-               (findInterfaceFile . toTopLevelModuleName =<< curMName)
-  noComp   = reportSLn "compile.ghc" 2 . (++ " : no compilation is needed.") . show . A.mnameToConcrete =<< curMName
-  yesComp  = reportSLn "compile.ghc" 1 . (`repl` "Compiling <<0>> in <<1>> to <<2>>") =<<
-             sequence [show . A.mnameToConcrete <$> curMName, ifile, outFile_] :: TCM ()
+    enable      o = pure o{ optGhcCompile = True }
+    dontCallGHC o = pure o{ optGhcCallGhc = False }
+    ghcFlag f   o = pure o{ optGhcFlags   = optGhcFlags o ++ [f] }
+
+--- Top-level compilation ---
+
+ghcPreCompile :: GHCOptions -> TCM GHCOptions
+ghcPreCompile ghcOpts = do
+  allowUnsolved <- optAllowUnsolved <$> pragmaOptions
+  when allowUnsolved $ genericError $ "Unsolved meta variables are not allowed when compiling."
+  return ghcOpts
+
+ghcPostCompile :: GHCOptions -> IsMain -> Map ModuleName IsMain -> TCM ()
+ghcPostCompile opts isMain mods = copyRTEModules >> callGHC opts isMain mods
+
+--- Module compilation ---
+
+type GHCModuleEnv = Maybe CoinductionKit
+
+ghcPreModule :: GHCOptions -> ModuleName -> FilePath -> TCM (Recompile GHCModuleEnv IsMain)
+ghcPreModule _ m ifile = ifM uptodate noComp yesComp
+  where
+    uptodate = liftIO =<< isNewerThan <$> outFile_ <*> pure ifile
+
+    noComp = do
+      reportSLn "compile.ghc" 2 . (++ " : no compilation is needed.") . show . A.mnameToConcrete =<< curMName
+      Skip . hasMainFunction <$> curIF
+
+    yesComp = do
+      m   <- show . A.mnameToConcrete <$> curMName
+      out <- outFile_
+      reportSLn "compile.ghc" 1 $ repl [m, ifile, out] "Compiling <<0>> in <<1>> to <<2>>"
+      stImportedModules .= Set.empty  -- we use stImportedModules to accumulate the required Haskell imports
+      Recompile <$> coinductionKit
+
+ghcPostModule :: GHCOptions -> GHCModuleEnv -> IsMain -> ModuleName -> [[HS.Decl]] -> TCM IsMain
+ghcPostModule _ _ _ _ defs = do
+  m      <- curHsMod
+  imps   <- imports
+  -- Get content of FOREIGN pragmas.
+  (headerPragmas, hsImps, code) <- foreignHaskell
+  writeModule $ HS.Module m
+    (map HS.OtherPragma headerPragmas)
+    imps
+    (map fakeDecl (hsImps ++ code) ++ concat defs)
+  hasMainFunction <$> curIF
+
+ghcCompileDef :: GHCOptions -> GHCModuleEnv -> Definition -> TCM [HS.Decl]
+ghcCompileDef _ = definition
+
+-- Compilation ------------------------------------------------------------
 
 --------------------------------------------------
 -- imported modules
@@ -110,11 +190,9 @@ compile i = do
 --------------------------------------------------
 
 imports :: TCM [HS.ImportDecl]
-imports = (++) <$> hsImps <*> imps where
-  hsImps :: TCM [HS.ImportDecl]
-  hsImps = ((unqualRTE :) . List.map decl . Set.toList .
-            Set.insert mazRTE . Set.map HS.ModuleName) <$>
-             getHaskellImports
+imports = (hsImps ++) <$> imps where
+  hsImps :: [HS.ImportDecl]
+  hsImps = [unqualRTE, decl mazRTE]
 
   unqualRTE :: HS.ImportDecl
   unqualRTE = HS.ImportDecl mazRTE False $ Just $
@@ -139,15 +217,6 @@ imports = (++) <$> hsImps <*> imps where
 -- Main compiling clauses
 --------------------------------------------------
 
-definitions :: Definitions -> TCM [HS.Decl]
-definitions defs = do
-  kit <- coinductionKit
-  concat <$>
-    (mapM (\(_, d) -> definition kit =<< instantiateFull d) $
-    -- sort defs to ensure consistent order (see Issue 1900)
-     sortDefs defs
-    )
-
 -- | Note that the INFINITY, SHARP and FLAT builtins are translated as
 -- follows (if a 'CoinductionKit' is given):
 --
@@ -172,18 +241,26 @@ definition kit Defn{defArgInfo = info, defName = q} | isIrrelevant info = do
   reportSDoc "compile.ghc.definition" 10 $
     text "Not compiling" <+> prettyTCM q <> text "."
   return []
-definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef = d} = do
+definition kit Defn{defName = q, defType = ty, theDef = d} = do
   reportSDoc "compile.ghc.definition" 10 $ vcat
     [ text "Compiling" <+> prettyTCM q <> text ":"
     , nest 2 $ text (show d)
     ]
+  pragma <- getHaskellPragma q
   checkTypeOfMain q ty $ do
     infodecl q <$> case d of
 
-      _ | Just (HsDefn hsty hs) <- compiledHaskell compiled -> do
+      _ | Just (HsDefn r hs) <- pragma -> setCurrentRange r $ do
         -- Make sure we have imports for all names mentioned in the type.
-        ty <- normalise ty
+        hsty <- haskellType q
+        ty   <- normalise ty
         sequence_ [ xqual x (HS.Ident "_") | x <- Set.toList (namesIn ty) ]
+
+        -- Check that the function isn't INLINE (since that will make this
+        -- definition pointless).
+        inline <- (^. funInline) . theDef <$> getConstInfo q
+        when inline $ warning $ UselessInline q
+
         return $ fbWithType hsty (fakeExp hs)
 
       -- Special treatment of coinductive builtins.
@@ -224,47 +301,56 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
                                  emptyBinds]
           ]
 
-      Axiom{} -> return $ fb axiomErr
+      Axiom{} -> do
+        ar <- typeArity ty
+        return $ [ compiledTypeSynonym q ty ar | Just (HsType r ty) <- [pragma] ] ++
+                 fb axiomErr
       Primitive{ primName = s } -> fb <$> primBody s
 
-      Function{} -> function (exportHaskell compiled) $ functionViaTreeless q
+      Function{} -> function pragma $ functionViaTreeless q
 
       Datatype{ dataPars = np, dataIxs = ni, dataClause = cl, dataCons = cs }
-        | Just (HsType ty) <- compiledHaskell compiled -> do
-        ccscov <- ifM (noCheckCover q) (return []) $ do
-            ccs <- List.concat <$> mapM checkConstructorType cs
-            cov <- checkCover q ty (np + ni) cs
-            return $ ccs ++ cov
-        return $ tvaldecl q (dataInduction d) 0 (np + ni) [] (Just __IMPOSSIBLE__) ++ ccscov
+        | Just (HsData r ty hsCons) <- pragma -> setCurrentRange r $ do
+        computeErasedConstructorArgs q
+        ccscov <- constructorCoverageCode q (np + ni) cs ty hsCons
+        cds <- mapM compiledcondecl cs
+        return $ tvaldecl q (dataInduction d) 0 (np + ni) [] (Just __IMPOSSIBLE__) ++
+                 [compiledTypeSynonym q ty np] ++ cds ++ ccscov
       Datatype{ dataPars = np, dataIxs = ni, dataClause = cl, dataCons = cs } -> do
         computeErasedConstructorArgs q
         (ars, cds) <- unzip <$> mapM condecl cs
         return $ tvaldecl q (dataInduction d) (List.maximum (np:ars) - np) (np + ni) cds cl
       Constructor{} -> return []
+      Record{ recPars = np, recClause = cl, recConHead = con }
+        | Just (HsData r ty hsCons) <- pragma -> setCurrentRange r $ do
+        let cs = [conName con]
+        computeErasedConstructorArgs q
+        ccscov <- constructorCoverageCode q np cs ty hsCons
+        cds <- mapM compiledcondecl cs
+        return $ tvaldecl q Inductive 0 np [] (Just __IMPOSSIBLE__) ++
+                 [compiledTypeSynonym q ty np] ++ cds ++ ccscov
       Record{ recClause = cl, recConHead = con, recFields = flds } -> do
         computeErasedConstructorArgs q
         let c = conName con
         let noFields = length flds
         let ar = I.arity ty
         cd <- snd <$> condecl c
-  --       cd <- case c of
-  --         Nothing -> return $ cdecl q noFields
-  --         Just c  -> snd <$> condecl c
         return $ tvaldecl q Inductive noFields ar [cd] cl
-      AbstractDefn -> __IMPOSSIBLE__
+      AbstractDefn{} -> __IMPOSSIBLE__
   where
-  function :: Maybe HaskellExport -> TCM [HS.Decl] -> TCM [HS.Decl]
+  function :: Maybe HaskellPragma -> TCM [HS.Decl] -> TCM [HS.Decl]
   function mhe fun = do
     ccls <- mkwhere <$> fun
     case mhe of
-      Nothing -> return ccls
-      Just (HsExport t name) -> do
+      Just (HsExport r name) -> do
+        t <- setCurrentRange r $ haskellType q
         let tsig :: HS.Decl
             tsig = HS.TypeSig [HS.Ident name] (fakeType t)
 
             def :: HS.Decl
             def = HS.FunBind [HS.Match (HS.Ident name) [] (HS.UnGuardedRhs (hsVarUQ $ dname q)) emptyBinds]
         return ([tsig,def] ++ ccls)
+      _ -> return ccls
 
   functionViaTreeless :: QName -> TCM [HS.Decl]
   functionViaTreeless q = caseMaybeM (toTreeless q) (pure []) $ \ treeless -> do
@@ -314,7 +400,15 @@ definition kit Defn{defName = q, defType = ty, defCompiledRep = compiled, theDef
                                 (HS.UnGuardedRhs $ e) emptyBinds]]
 
   axiomErr :: HS.Exp
-  axiomErr = rtmError $ "postulate evaluated: " ++ show (A.qnameToConcrete q)
+  axiomErr = rtmError $ "postulate evaluated: " ++ prettyShow q
+
+constructorCoverageCode :: QName -> Int -> [QName] -> HaskellType -> [HaskellCode] -> TCM [HS.Decl]
+constructorCoverageCode q np cs hsTy hsCons = do
+  checkConstructorCount q cs hsCons
+  ifM (noCheckCover q) (return []) $ do
+    ccs <- List.concat <$> zipWithM checkConstructorType cs hsCons
+    cov <- checkCover q hsTy np cs hsCons
+    return $ ccs ++ cov
 
 -- | Environment for naming of local variables.
 --   Invariant: @reverse ccCxt ++ ccNameSupply@
@@ -356,24 +450,23 @@ intros :: Int -> ([HS.Name] -> CC a) -> CC a
 intros n cont = freshNames n $ \xs ->
   local (mapContext (reverse xs ++)) $ cont xs
 
-checkConstructorType :: QName -> TCM [HS.Decl]
-checkConstructorType q = do
-  Just (HsDefn ty hs) <- compiledHaskell . defCompiledRep <$> getConstInfo q
+checkConstructorType :: QName -> HaskellCode -> TCM [HS.Decl]
+checkConstructorType q hs = do
+  ty <- haskellType q
   return [ HS.TypeSig [unqhname "check" q] $ fakeType ty
          , HS.FunBind [HS.Match (unqhname "check" q) []
                                 (HS.UnGuardedRhs $ fakeExp hs) emptyBinds]
          ]
 
-checkCover :: QName -> HaskellType -> Nat -> [QName] -> TCM [HS.Decl]
-checkCover q ty n cs = do
+checkCover :: QName -> HaskellType -> Nat -> [QName] -> [HaskellCode] -> TCM [HS.Decl]
+checkCover q ty n cs hsCons = do
   let tvs = [ "a" ++ show i | i <- [1..n] ]
-      makeClause c = do
+      makeClause c hsc = do
         a <- erasedArity c
-        Just (HsDefn _ hsc) <- compiledHaskell . defCompiledRep <$> getConstInfo c
         let pat = HS.PApp (HS.UnQual $ HS.Ident hsc) $ replicate a HS.PWildCard
         return $ HS.Alt pat (HS.UnGuardedRhs $ HS.unit_con) emptyBinds
 
-  cs <- mapM makeClause cs
+  cs <- zipWithM makeClause cs hsCons
   let rhs = case cs of
               [] -> fakeExp "()" -- There is no empty case statement in Haskell
               _  -> HS.Case (HS.Var $ HS.UnQual $ HS.Ident "x") cs
@@ -396,10 +489,11 @@ term tm0 = case tm0 of
     return $ hsVarUQ x
   T.TApp (T.TDef f) ts -> do
     used <- lift $ getCompiledArgUse f
+    isCompiled <- lift $ isJust <$> getHaskellPragma f  -- #2248: no unused argument pruning for COMPILE'd functions
     let given   = length ts
         needed  = length used
         missing = drop given used
-    if any not used
+    if not isCompiled && any not used
       then if any not missing then term (etaExpand (needed - given) tm0) else do
         f <- lift $ HS.Var <$> xhqn "du" f  -- used stripped function
         f `apps` [ t | (t, True) <- zip ts $ used ++ repeat True ]
@@ -542,7 +636,7 @@ litqname x =
   rteCon "QName" `apps`
   [ hsTypedInt n
   , hsTypedInt m
-  , HS.Lit $ HS.String $ show x
+  , HS.Lit $ HS.String $ prettyShow x
   , rteCon "Fixity" `apps`
     [ litAssoc (fixityAssoc fx)
     , litPrec  (fixityLevel fx) ] ]
@@ -585,6 +679,20 @@ cdecl :: QName -> Nat -> HS.ConDecl
 cdecl q n = HS.ConDecl (unqhname "C" q)
             [ HS.TyVar $ ihname "a" i | i <- [0 .. n - 1] ]
 
+compiledcondecl :: QName -> TCM HS.Decl
+compiledcondecl q = do
+  (ar, np) <- conArityAndPars q
+  hsCon <- fromMaybe __IMPOSSIBLE__ <$> getHaskellConstructor q
+  let patVars = map (HS.PVar . ihname "a") [0 .. ar - 1]
+  return $ HS.PatSyn (HS.PApp (HS.UnQual $ unqhname "C" q) patVars) (HS.PApp (hsName hsCon) patVars)
+
+compiledTypeSynonym :: QName -> String -> Nat -> HS.Decl
+compiledTypeSynonym q hsT arity =
+  HS.TypeDecl (unqhname "T" q) (map HS.UnkindedVar vs)
+              (foldl HS.TyApp (HS.FakeType hsT) $ map HS.TyVar vs)
+  where
+    vs = [ ihname "a" i | i <- [0 .. arity - 1]]
+
 tvaldecl :: QName
          -> Induction
             -- ^ Is the type inductive or coinductive?
@@ -640,6 +748,7 @@ hsCast' e =
 -- functions.
 hsCastApp :: HS.Exp -> HS.Exp
 hsCastApp (HS.Lambda ps b) = HS.Lambda ps (hsCastApp b)
+hsCastApp (HS.Let bs e) = HS.Let bs $ hsCastApp e
 hsCastApp (HS.Case sc bs) = HS.Case (hsCastApp sc) (map (hsMapAlt hsCastApp) bs)
 hsCastApp (HS.InfixApp e1 op e2) = HS.InfixApp (hsCastApp e1) op (hsCastApp e2)
 hsCastApp e =
@@ -680,6 +789,7 @@ writeModule (HS.Module m ps imp ds) = do
         , "ScopedTypeVariables"
         , "NoMonomorphismRestriction"
         , "Rank2Types"
+        , "PatternSynonyms"
         ]
 
 
@@ -701,35 +811,42 @@ outFile m = snd <$> outFile' m
 outFile_ :: TCM FilePath
 outFile_ = outFile =<< curHsMod
 
-callGHC :: IsMain -> Interface -> TCM ()
-callGHC modIsMain i = do
-  setInterface i
-  mdir          <- compileDir
-  hsmod         <- prettyPrint <$> curHsMod
-  MName agdaMod <- curMName
-  let outputName = case agdaMod of
+callGHC :: GHCOptions -> IsMain -> Map ModuleName IsMain -> TCM ()
+callGHC opts modIsMain mods = do
+  mdir    <- compileDir
+  hsmod   <- prettyPrint <$> curHsMod
+  agdaMod <- curMName
+  let outputName = case mnameToList agdaMod of
         [] -> __IMPOSSIBLE__
         ms -> last ms
   (mdir, fp) <- outFile' =<< curHsMod
-  opts       <- optGhcFlags <$> commandLineOptions
+  let ghcopts = optGhcFlags opts
+
+  let modIsReallyMain = fromMaybe __IMPOSSIBLE__ $ Map.lookup agdaMod mods
+      isMain = mappend modIsMain modIsReallyMain  -- both need to be IsMain
+
+  -- Warn if no main function and not --no-main
+  when (modIsMain /= isMain) $
+    genericWarning =<< fsep (pwords "No main function defined in" ++ [prettyTCM agdaMod <> text "."] ++
+                             pwords "Use --no-main to suppress this warning.")
 
   let overridableArgs =
         [ "-O"] ++
-        (if modIsMain == IsMain then ["-o", mdir </> show (nameConcrete outputName)] else []) ++
+        (if isMain == IsMain then ["-o", mdir </> show (nameConcrete outputName)] else []) ++
         [ "-Werror"]
       otherArgs       =
         [ "-i" ++ mdir] ++
-        (if modIsMain == IsMain then ["-main-is", hsmod] else []) ++
+        (if isMain == IsMain then ["-main-is", hsmod] else []) ++
         [ fp
         , "--make"
         , "-fwarn-incomplete-patterns"
         , "-fno-warn-overlapping-patterns"
         ]
-      args     = overridableArgs ++ opts ++ otherArgs
+      args     = overridableArgs ++ ghcopts ++ otherArgs
       compiler = "ghc"
 
   -- Note: Some versions of GHC use stderr for progress reports. For
   -- those versions of GHC we don't print any progress information
   -- unless an error is encountered.
-  doCall <- optGhcCallGhc <$> commandLineOptions
+  let doCall = optGhcCallGhc opts
   callCompiler doCall compiler args
