@@ -4,7 +4,7 @@ module Agda.Compiler.MAlonzo.Misc where
 
 import Control.Monad.State (gets)
 import Data.Char
-import Data.List as List
+import qualified Data.List as List
 import Data.Map as Map
 import Data.Set as Set
 import Data.Function
@@ -12,6 +12,7 @@ import Data.Function
 import qualified Agda.Utils.Haskell.Syntax as HS
 
 import Agda.Compiler.Common
+import Agda.Compiler.MAlonzo.Pragmas
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -22,6 +23,7 @@ import Agda.TypeChecking.Monad.Builtin
 import Agda.Utils.Lens
 import Agda.Utils.Monad
 import Agda.Utils.Pretty
+import Agda.Utils.Maybe
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -82,11 +84,7 @@ conhqn :: QName -> TCM HS.QName
 conhqn q = do
     cq  <- canonicalName q
     def <- getConstInfo cq
-    cname <- xhqn "C" cq  -- Do this even if it has custom compiledHaskell code
-                          -- to make sure we get the import.
-    case (compiledHaskell (defCompiledRep def), theDef def) of
-      (Just (HsDefn _ hs), Constructor{}) -> return $ hsName hs
-      _                                   -> return cname
+    xhqn "C" cq
 
 -- qualify name s by the module of builtin b
 bltQual :: String -> String -> TCM HS.QName
