@@ -31,6 +31,8 @@ instance Pretty HS.Module where
 instance Pretty HS.ModulePragma where
   pretty (HS.LanguagePragma ps) =
     text "{-#" <+> text "LANGUAGE" <+> fsep (punctuate comma $ map pretty ps) <+> text "#-}"
+  pretty (HS.OtherPragma p) =
+    text p
 
 instance Pretty HS.ImportDecl where
   pretty HS.ImportDecl{ HS.importModule    = m
@@ -65,6 +67,7 @@ instance Pretty HS.Decl where
       sep [ hsep (punctuate comma (map pretty fs)) <+> text "::"
           , nest 2 $ pretty t ]
     HS.FunBind ms -> vcat $ map pretty ms
+    HS.PatSyn p1 p2 -> sep [ text "pattern" <+> pretty p1 <+> text "=" <+> pretty p2 ]
     HS.FakeDecl s -> text s
 
 instance Pretty HS.ConDecl where
@@ -211,4 +214,3 @@ isOperator q =
 prettyQName :: HS.QName -> Doc
 prettyQName (HS.Qual m x)           = pretty m <> text "." <> pretty x
 prettyQName (HS.UnQual x)           = pretty x
-
