@@ -4,9 +4,9 @@ module Agda.Compiler.MAlonzo.Primitives where
 
 import Control.Monad.State
 import Data.Char
-import Data.List as L
+import qualified Data.List as List
 import Data.Map as M
-import qualified Agda.Utils.Haskell.Syntax as HS
+import Data.Maybe
 
 import Agda.Compiler.Common
 import Agda.Compiler.ToTreeless
@@ -21,34 +21,31 @@ import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Pretty
-import Agda.Utils.Monad
+
+import Agda.Utils.Either
 import Agda.Utils.Except
+import Agda.Utils.Lens
+import Agda.Utils.Monad
 import qualified Agda.Utils.HashMap as HMap
+import qualified Agda.Utils.Haskell.Syntax as HS
 
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-{- OLD
--- | Check that the main function has type IO a, for some a.
-checkTypeOfMain :: QName -> Type -> TCM ()
-checkTypeOfMain q ty
-  | show (qnameName q) /= "main" = return ()
-  | otherwise = do
-    Def io _ <- ignoreSharing <$> primIO
-    ty <- normalise ty
-    case ignoreSharing $ unEl ty of
-      Def d _ | d == io -> return ()
-      _                 -> do
-        err <- fsep $
-          pwords "The type of main should be" ++
-          [prettyTCM io] ++ pwords " A, for some A. The given type is" ++ [prettyTCM ty]
-        typeError $ GenericError $ show err
--}
+isMainFunction :: QName -> Bool
+isMainFunction q = "main" == show (nameConcrete $ qnameName q)
+
+hasMainFunction :: Interface -> IsMain
+hasMainFunction i
+  | List.any isMainFunction names = IsMain
+  | otherwise                     = NotMain
+  where
+    names = HMap.keys $ iSignature i ^. sigDefinitions
 
 -- | Check that the main function has type IO a, for some a.
 checkTypeOfMain :: QName -> Type -> TCM [HS.Decl] -> TCM [HS.Decl]
 checkTypeOfMain q ty ret
-  | show (nameConcrete $ qnameName q) /= "main" = ret
+  | not (isMainFunction q) = ret
   | otherwise = do
     Def io _ <- ignoreSharing <$> primIO
     ty <- normalise ty
@@ -84,12 +81,12 @@ treelessPrimName p =
     -- primitives only used by GuardsToPrims transformation, which MAlonzo doesn't use
     PIf   -> __IMPOSSIBLE__
 
--- Haskell modules to be imported for BUILT-INs
+-- | Haskell modules to be imported for BUILT-INs
 importsForPrim :: TCM [HS.ModuleName]
 importsForPrim =
   fmap (++ [HS.ModuleName "Data.Text"]) $
   xForPrim $
-  L.map (\(s, ms) -> (s, return (L.map HS.ModuleName ms))) $
+  List.map (\(s, ms) -> (s, return (List.map HS.ModuleName ms))) $
   [ "CHAR"              |-> ["Data.Char"]
   , "primIsAlpha"       |-> ["Data.Char"]
   , "primIsAscii"       |-> ["Data.Char"]
@@ -116,14 +113,14 @@ xForPrim table = do
       getName (Builtin (Lam _ b))    = getName (Builtin $ unAbs b)
       getName (Builtin _)            = __IMPOSSIBLE__
       getName (Prim (PrimFun q _ _)) = q
-  concat <$> sequence [ maybe (return []) id $ L.lookup s table
+  concat <$> sequence [ fromMaybe (return []) $ List.lookup s table
                         | (s, def) <- bs, getName def `elem` qs ]
 
 
--- Definition bodies for primitive functions
+-- | Definition bodies for primitive functions
 primBody :: String -> TCM HS.Exp
-primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
-             L.lookup s $
+primBody s = maybe unimplemented (fromRight (hsVarUQ . HS.Ident) <$>) $
+             List.lookup s $
   [
   -- Integer functions
     "primIntegerPlus"    |-> binAsis "(+)" "Integer"
@@ -200,7 +197,7 @@ primBody s = maybe unimplemented (either (hsVarUQ . HS.Ident) id <$>) $
   , "primStringFromList" |-> return "Data.Text.pack"
   , "primStringAppend"   |-> binAsis "Data.Text.append" "Data.Text.Text"
   , "primStringEquality" |-> rel "(==)" "Data.Text.Text"
-  , "primShowString"     |-> return "id"
+  , "primShowString"     |-> return "(Data.Text.pack . show :: Data.Text.Text -> Data.Text.Text)"
 
   -- Reflection
   , "primQNameEquality"   |-> rel "(==)" "MAlonzo.RTE.QName"
