@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
 
 module Agda.Compiler.ToTreeless
   ( toTreeless
@@ -44,6 +43,7 @@ import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Lens
+import Agda.Utils.Pretty (prettyShow)
 import qualified Agda.Utils.Pretty as P
 
 #include "undefined.h"
@@ -63,7 +63,7 @@ toTreeless q = ifM (alwaysInline q) (pure Nothing) $ Just <$> toTreeless' q
 
 toTreeless' :: QName -> TCM C.TTerm
 toTreeless' q =
-  flip fromMaybeM (getTreeless q) $ verboseBracket "treeless.convert" 20 ("compiling " ++ show q) $ do
+  flip fromMaybeM (getTreeless q) $ verboseBracket "treeless.convert" 20 ("compiling " ++ prettyShow q) $ do
     Just cc <- defCompiled <$> getConstInfo q
     unlessM (alwaysInline q) $ setTreeless q (C.TDef q)
       -- so recursive inlining doesn't loop, but not for always inlined
@@ -81,7 +81,7 @@ cacheTreeless q = do
 ccToTreeless :: QName -> CC.CompiledClauses -> TCM C.TTerm
 ccToTreeless q cc = do
   let pbody b = pbody' "" b
-      pbody' suf b = sep [ text (show q ++ suf) <+> text "=", nest 2 $ prettyPure b ]
+      pbody' suf b = sep [ text (prettyShow q ++ suf) <+> text "=", nest 2 $ prettyPure b ]
   v <- ifM (alwaysInline q) (return 20) (return 0)
   reportSDoc "treeless.convert" (30 + v) $ text "-- compiled clauses of" <+> prettyTCM q $$ nest 2 (prettyPure cc)
   body <- casetreeTop cc
@@ -168,9 +168,11 @@ casetree :: CC.CompiledClauses -> CC C.TTerm
 casetree cc = do
   case cc of
     CC.Fail -> return C.tUnreachable
-    CC.Done xs v -> lambdasUpTo (length xs) $ do
-        v <- lift $ putAllowedReductions [ProjectionReductions, CopatternReductions] $ normalise v
-        substTerm v
+    CC.Done xs v -> withContextSize (length xs) $ do
+      -- Issue 2469: Body context size (`length xs`) may be smaller than current context size
+      -- if some arguments are not used in the body.
+      v <- lift (putAllowedReductions [ProjectionReductions, CopatternReductions] $ normalise v)
+      substTerm v
     CC.Case _ (CC.Branches True _ _ Just{}) -> __IMPOSSIBLE__
       -- Andreas, 2016-06-03, issue #1986: Ulf: "no catch-all for copatterns!"
       -- lift $ do
@@ -231,6 +233,28 @@ updateCatchAll (Just cc) cont = do
   local (\e -> e { ccCatchAll = Just 0, ccCxt = shift 1 (ccCxt e) }) $ do
     C.mkLet def <$> cont
 
+-- | Shrinks or grows the context to the given size.
+-- Does not update the catchAll expression, the catchAll expression
+-- MUST NOT be used inside `cont`.
+withContextSize :: Int -> CC C.TTerm -> CC C.TTerm
+withContextSize n cont = do
+  diff <- (n -) . length <$> asks ccCxt
+
+  if diff <= 0
+  then do
+    local (\e -> e { ccCxt = shift diff $ drop (-diff) (ccCxt e)}) $
+      C.mkTApp <$> cont <*> pure [C.TVar i | i <- reverse [0..(-diff - 1)]]
+  else do
+    local (\e -> e { ccCxt = [0..(diff - 1)] ++ shift diff (ccCxt e)}) $ do
+      createLambdas diff <$> do
+        cont
+  where createLambdas :: Int -> C.TTerm -> C.TTerm
+        createLambdas 0 cont' = cont'
+        createLambdas i cont' | i > 0 = C.TLam (createLambdas (i - 1) cont')
+        createLambdas _ _ = __IMPOSSIBLE__
+
+-- | Adds lambdas until the context has at least the given size.
+-- Updates the catchAll expression to take the additional lambdas into account.
 lambdasUpTo :: Int -> CC C.TTerm -> CC C.TTerm
 lambdasUpTo n cont = do
   diff <- (n -) . length <$> asks ccCxt
@@ -239,23 +263,18 @@ lambdasUpTo n cont = do
   else do
     catchAll <- asks ccCatchAll
 
-    local (\e -> e { ccCxt = [0..(diff - 1)] ++ shift diff (ccCxt e)}) $ do
-      createLambdas diff <$> do
-        case catchAll of
-          Just catchAll' -> do
-            -- the catch all doesn't know about the additional lambdas, so just directly
-            -- apply it again to the newly introduced lambda arguments.
-            -- we also bind the catch all to a let, to avoid code duplication
-            local (\e -> e { ccCatchAll = Just 0
-                           , ccCxt = shift 1 (ccCxt e)}) $ do
-              let catchAllArgs = map C.TVar $ reverse [0..(diff - 1)]
-              C.mkLet (C.mkTApp (C.TVar $ catchAll' + diff) catchAllArgs)
-                <$> cont
-          Nothing -> cont
-  where createLambdas :: Int -> C.TTerm -> C.TTerm
-        createLambdas 0 cont' = cont'
-        createLambdas i cont' | i > 0 = C.TLam (createLambdas (i - 1) cont')
-        createLambdas _ _ = __IMPOSSIBLE__
+    withContextSize n $ do
+      case catchAll of
+        Just catchAll' -> do
+          -- the catch all doesn't know about the additional lambdas, so just directly
+          -- apply it again to the newly introduced lambda arguments.
+          -- we also bind the catch all to a let, to avoid code duplication
+          local (\e -> e { ccCatchAll = Just 0
+                         , ccCxt = shift 1 (ccCxt e)}) $ do
+            let catchAllArgs = map C.TVar $ reverse [0..(diff - 1)]
+            C.mkLet (C.mkTApp (C.TVar $ catchAll' + diff) catchAllArgs)
+              <$> cont
+        Nothing -> cont
 
 conAlts :: Int -> Map QName (CC.WithArity CC.CompiledClauses) -> CC [C.TAlt]
 conAlts x br = forM (Map.toList br) $ \ (c, CC.WithArity n cc) -> do
@@ -372,6 +391,5 @@ substArg x | erasable x = return C.TErased
       case getRelevance x of
         Irrelevant -> True
         NonStrict  -> True
-        UnusedArg  -> True
         Forced{}   -> False -- TODO: would like this to be True
         Relevant   -> False
