@@ -15,9 +15,6 @@
 --
 -- Ulf, 2015-10-30: Guards are actually a better primitive. Fixed that.
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE PatternGuards #-}
-{-# LANGUAGE RecordWildCards #-}
-{-# LANGUAGE FlexibleContexts #-}
 module Agda.Compiler.Treeless.Builtin (translateBuiltins) where
 
 import Control.Applicative
@@ -64,7 +61,7 @@ builtinKit =
              <*> isB def builtinNatEquals
              <*> isP pf  "primForce"
   where
-    con (I.Con c _) = pure $ I.conName c
+    con (I.Con c _ _) = pure $ I.conName c
     con _           = Nothing
     def (I.Def d _) = pure d
     def _           = Nothing
@@ -93,7 +90,7 @@ transform BuiltinKit{..} = tr
       TDef f | isPlus f   -> TPrim PAdd
              | isTimes f  -> TPrim PMul
              | isLess f   -> TPrim PLt
-             | isEqual f  -> TPrim PEq
+             | isEqual f  -> TPrim PEqI
         -- Note: Don't do this for builtinNatMinus! PSub is integer minus and
         --       builtin minus is monus. The simplifier will do it if it can see
         --       that it won't underflow.
@@ -115,7 +112,7 @@ transform BuiltinKit{..} = tr
           e | Just (i, e) <- plusKView e -> tNegPlusK (i + 1) e
           e -> tNegPlusK 1 e
 
-      TCase e t d bs -> TCase e t (tr d) $ concatMap trAlt bs
+      TCase e t d bs -> TCase e (caseType t bs) (tr d) $ concatMap trAlt bs
         where
           trAlt b = case b of
             TACon c 0 b | isZero c -> [TALit (LitNat noRange 0) (tr b)]
@@ -125,8 +122,10 @@ transform BuiltinKit{..} = tr
                 TCase 0 _ d bs' -> map sucBranch bs' ++ [nPlusKAlt 1 d]
                 b -> [nPlusKAlt 1 b]
               where
-                sucBranch (TALit (LitNat r i) b) = TALit (LitNat r (i + 1)) $ applySubst (str __IMPOSSIBLE__) b
-                sucBranch alt | Just (k, b) <- nPlusKView alt = nPlusKAlt (k + 1) $ applySubst (liftS 1 $ str __IMPOSSIBLE__) b
+                sucBranch (TALit (LitNat r i) b) = TALit (LitNat r (i + 1)) $ TLet (tInt i) b
+                sucBranch alt | Just (k, b) <- nPlusKView alt =
+                  nPlusKAlt (k + 1) $ TLet (tOp PAdd (TVar 0) (tInt 1)) $
+                    applySubst ([TVar 1, TVar 0] ++# wkS 2 idS) b
                 sucBranch _ = __IMPOSSIBLE__
 
                 nPlusKAlt k b = TAGuard (tOp PGeq (TVar e) (tInt k)) $
@@ -179,8 +178,15 @@ transform BuiltinKit{..} = tr
       TLam b                  -> TLam (tr b)
       TApp a bs               -> TApp (tr a) (map tr bs)
       TLet e b                -> TLet (tr e) (tr b)
+
+    caseType t (TACon c _ _ : _)
+      | isZero c   = CTNat
+      | isSuc c    = CTNat
+      | isPos c    = CTInt
+      | isNegSuc c = CTInt
+    caseType t _ = t
+
     nPlusKView (TAGuard (TApp (TPrim PGeq) [TVar 0, (TLit (LitNat _ k))])
                         (TLet (TApp (TPrim PSub) [TVar 0, (TLit (LitNat _ j))]) b))
       | k == j = Just (k, b)
     nPlusKView _ = Nothing
-
