@@ -3,7 +3,6 @@
 module Agda.Compiler.Treeless.EliminateLiteralPatterns where
 
 import Control.Applicative
-import Data.List
 import Data.Maybe
 
 import Agda.Syntax.Abstract.Name (QName)
@@ -49,7 +48,10 @@ transform kit = tr
           litAlt _ _ = __IMPOSSIBLE__
       TCase sc t@(CTData dt) def alts -> TCase sc t (tr def) (map trAlt alts)
         where
-          trAlt a = a { aBody = tr (aBody a) }
+          trAlt a = case a of
+            TAGuard g b -> TAGuard (tr g) (tr b)
+            TACon q a b -> TACon q a (tr b)
+            TALit l b   -> TALit l (tr b)
       TCase _ _ _ _ -> __IMPOSSIBLE__
 
       TVar{}    -> t
