@@ -6,7 +6,7 @@ import Control.Applicative
 import Control.Monad.Reader
 import Control.Monad.Writer
 import Data.Traversable (traverse)
-import Data.List
+import qualified Data.List as List
 
 import Agda.Syntax.Treeless
 import Agda.Syntax.Internal (Substitution'(..))
@@ -55,6 +55,9 @@ underLam = underLams 1
 underLet :: TTerm -> S a -> S a
 underLet u = onRewrite (raiseS 1) . onSubst (\rho -> wkS 1 $ u :# rho)
 
+bindVar :: Int -> TTerm -> S a -> S a
+bindVar x u = onSubst (inplaceS x u `composeS`)
+
 rewrite :: TTerm -> S TTerm
 rewrite t = do
   rules <- asks envRewrite
@@ -251,10 +254,11 @@ simplify FunctionKit{..} = simpl
     simplAlt x (TALit l b)   = TALit l   <$> maybeAddRewrite x (TLit l) (simpl b)
     simplAlt x (TAGuard g b) = TAGuard   <$> simpl g <*> simpl b
 
+    -- If x is already bound we add a rewrite, otherwise we bind x to rhs.
     maybeAddRewrite x rhs cont = do
       v <- lookupVar x
       case v of
-        TVar y | x == y -> cont
+        TVar y | x == y -> bindVar x rhs $ cont
         _ -> addRewrite v rhs cont
 
     isTrue (TCon c) = Just c == true
@@ -268,7 +272,7 @@ simplify FunctionKit{..} = simpl
       b_a  <- rewrite' (tOp PLt b a)
       b_sa <- rewrite' (tOp PLt b (tOp PAdd (tInt 1) a))
       a_b  <- rewrite' (tOp PLt a b)
-      if isTrue b_a || isTrue b_sa || isFalse b_a && isFalse a_b
+      if isTrue b_a || isTrue b_sa || isFalse a_b
         then pure $ tOp PSub a b
         else tApp f es
 
@@ -312,7 +316,7 @@ simplify FunctionKit{..} = simpl
         Nothing  -> return $ TCase x CTNat d bs
       where
         complete bs small (Just upper)
-          | null $ [0..upper - 1] \\ small = Just []
+          | null $ [0..upper - 1] List.\\ small = Just []
         complete (b@(TALit (LitNat _ n) _) : bs) small upper =
           (b :) <$> complete bs (n : small) upper
         complete (b@(TAGuard (TApp (TPrim PGeq) [TVar y, TLit (LitNat _ j)]) _) : bs) small upper | x == y =
@@ -367,12 +371,12 @@ aNeg (Neg a) = Pos a
 
 aCancel :: [Atom] -> [Atom]
 aCancel (a : as)
-  | elem (aNeg a) as = aCancel (delete (aNeg a) as)
+  | elem (aNeg a) as = aCancel (List.delete (aNeg a) as)
   | otherwise        = a : aCancel as
 aCancel [] = []
 
 sortR :: Ord a => [a] -> [a]
-sortR = sortBy (flip compare)
+sortR = List.sortBy (flip compare)
 
 aAdd :: Arith -> Arith -> Arith
 aAdd (a, xs) (b, ys) = (a + b, aCancel $ sortR $ xs ++ ys)
@@ -405,4 +409,3 @@ toArith t = (0, [Pos t])
 
 simplArith :: TTerm -> TTerm
 simplArith = fromArith . toArith
-
