@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP                   #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
 {-# LANGUAGE UndecidableInstances  #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
@@ -15,7 +14,8 @@ import Control.Monad.State
 import Control.Monad.Identity
 
 import qualified Data.Map as Map
-import Data.List hiding (null)
+import qualified Data.Set as Set
+import qualified Data.List as List
 import Data.Maybe
 import Data.Traversable hiding (mapM, forM, for)
 import Data.Monoid
@@ -41,6 +41,7 @@ import Agda.Syntax.Parser
 import Agda.TheTypeChecker
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
+import Agda.TypeChecking.Errors ( stringTCErr )
 import Agda.TypeChecking.Monad as M hiding (MetaInfo)
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Reduce
@@ -55,10 +56,11 @@ import Agda.TypeChecking.Free
 import Agda.TypeChecking.CheckInternal
 import Agda.TypeChecking.SizedTypes.Solve
 import qualified Agda.TypeChecking.Pretty as TP
+import Agda.TypeChecking.Warnings (runPM)
 
 import Agda.Termination.TermCheck (termMutual)
 
-import Agda.Utils.Except ( Error(strMsg), MonadError(catchError, throwError) )
+import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
@@ -90,10 +92,10 @@ parseExprIn ii rng s = do
     e   <- parseExpr rng s
     concreteToAbstract (clScope mi) e
 
-giveExpr :: Maybe InteractionId -> MetaId -> Expr -> TCM ()
+giveExpr :: UseForce -> Maybe InteractionId -> MetaId -> Expr -> TCM ()
 -- When translator from internal to abstract is given, this function might return
 -- the expression returned by the type checker.
-giveExpr mii mi e = do
+giveExpr force mii mi e = do
     mv <- lookupMeta mi
     -- In the context (incl. signature) of the meta variable,
     -- type check expression and assign meta
@@ -127,7 +129,7 @@ giveExpr mii mi e = do
             -- if there are more free variables than the context has
             -- we need to abstract over the additional ones (xs2)
             let (_xs1, xs2) = splitAt (size ctx) xs
-            v' <- return $ foldr (\ (Arg ai x) -> I.Lam ai . I.Abs x) v' xs2
+            v' <- return $ foldr mkLam v' xs2
             reportSDoc "interaction.give" 20 $ TP.sep
               [ TP.text "in meta context, v' = " TP.<+> prettyTCM v'
               ]
@@ -139,12 +141,14 @@ giveExpr mii mi e = do
             nowSolvingConstraints $ assign DirEq mi args v
 
         reportSDoc "interaction.give" 20 $ TP.text "give: meta variable updated!"
-        redoChecks mii
+        unless (force == WithForce) $ redoChecks mii
         wakeupConstraints mi
         solveSizeConstraints DontDefaultToInfty
-        -- Double check.
-        vfull <- instantiateFull v
-        checkInternal vfull t'
+        unless (force == WithForce) $ do
+          -- Double check.
+          reportSDoc "interaction.give" 20 $ TP.text "give: double checking"
+          vfull <- instantiateFull v
+          checkInternal vfull t'
 
 -- | After a give, redo termination etc. checks for function which was complemented.
 redoChecks :: Maybe InteractionId -> TCM ()
@@ -166,22 +170,25 @@ redoChecks (Just ii) = do
 --   Returns the given expression unchanged
 --   (for convenient generalization to @'refine'@).
 give
-  :: InteractionId  -- ^ Hole.
+  :: UseForce       -- ^ Skip safety checks?
+  -> InteractionId  -- ^ Hole.
   -> Maybe Range
   -> Expr           -- ^ The expression to give.
   -> TCM Expr       -- ^ If successful, the very expression is returned unchanged.
-give ii mr e = liftTCM $ do
+give force ii mr e = liftTCM $ do
   -- if Range is given, update the range of the interaction meta
   mi  <- lookupInteractionId ii
   whenJust mr $ updateMetaVarRange mi
   reportSDoc "interaction.give" 10 $ TP.text "giving expression" TP.<+> prettyTCM e
   reportSDoc "interaction.give" 50 $ TP.text $ show $ deepUnscope e
   -- Try to give mi := e
-  catchError (giveExpr (Just ii) mi e) $ \ err -> case err of
-    -- Turn PatternErr into proper error:
-    PatternErr{} -> typeError . GenericDocError =<< do
-      withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
-    _ -> throwError err
+  do setMetaOccursCheck mi DontRunMetaOccursCheck -- #589, #2710: Allow giving recursive solutions.
+     giveExpr force (Just ii) mi e
+    `catchError` \ case
+      -- Turn PatternErr into proper error:
+      PatternErr{} -> typeError . GenericDocError =<< do
+        withInteractionId ii $ TP.text "Failed to give" TP.<+> prettyTCM e
+      err -> throwError err
   removeInteractionPoint ii
   return e
 
@@ -191,11 +198,12 @@ give ii mr e = liftTCM $ do
 --   This amounts to successively try to give @e@, @e ?@, @e ? ?@, ...
 --   Returns the successfully given expression.
 refine
-  :: InteractionId  -- ^ Hole.
+  :: UseForce       -- ^ Skip safety checks when giving?
+  -> InteractionId  -- ^ Hole.
   -> Maybe Range
   -> Expr           -- ^ The expression to refine the hole with.
   -> TCM Expr       -- ^ The successfully given expression.
-refine ii mr e = do
+refine force ii mr e = do
   mi <- lookupInteractionId ii
   mv <- lookupMeta mi
   let range = fromMaybe (getRange mv) mr
@@ -211,8 +219,8 @@ refine ii mr e = do
     tryRefine nrOfMetas r scope e = try nrOfMetas e
       where
         try :: Int -> Expr -> TCM Expr
-        try 0 e = throwError $ strMsg "Cannot refine"
-        try n e = give ii (Just r) e `catchError` (\_ -> try (n - 1) =<< appMeta e)
+        try 0 e = throwError $ stringTCErr "Cannot refine"
+        try n e = give force ii (Just r) e `catchError` (\_ -> try (n - 1) =<< appMeta e)
 
         -- Apply A.Expr to a new meta
         appMeta :: Expr -> TCM Expr
@@ -222,7 +230,10 @@ refine ii mr e = do
           ii <- registerInteractionPoint False rng Nothing
           let info = Info.MetaInfo
                 { Info.metaRange = rng
-                , Info.metaScope = scope { scopePrecedence = ArgumentCtx }
+                , Info.metaScope = scope { scopePrecedence = [ArgumentCtx] }
+                    -- Ulf, 2017-09-07: The `ArgumentCtx` above is causing #737.
+                    -- If we're building an operator application the precedence
+                    -- should be something else.
                 , metaNumber = Nothing -- in order to print just as ?, not ?n
                 , metaNameSuggestion = ""
                 }
@@ -264,9 +275,11 @@ evalInMeta ii e =
         withMetaInfo mi $
             evalInCurrent e
 
-
+-- | Modifier for interactive commands,
+--   specifying the amount of normalization in the output.
+--
 data Rewrite =  AsIs | Instantiated | HeadNormal | Simplified | Normalised
-    deriving (Read)
+    deriving (Show, Read)
 
 normalForm :: (Reduce t, Simplify t, Normalise t) => Rewrite -> t -> TCM t
 normalForm AsIs         t = return t
@@ -275,8 +288,11 @@ normalForm HeadNormal   t = {- etaContract =<< -} reduce t
 normalForm Simplified   t = {- etaContract =<< -} simplify t
 normalForm Normalised   t = {- etaContract =<< -} normalise t
 
+-- | Modifier for the interactive computation command,
+--   specifying the mode of computation and result display.
+--
 data ComputeMode = DefaultCompute | IgnoreAbstract | UseShowInstance
-  deriving (Read, Eq)
+  deriving (Show, Read, Eq)
 
 computeIgnoreAbstract :: ComputeMode -> Bool
 computeIgnoreAbstract DefaultCompute  = False
@@ -296,6 +312,13 @@ showComputed UseShowInstance e =
     _                     -> (text "Not a string:" $$) <$> prettyATop e
 showComputed _ e = prettyATop e
 
+-- | Modifier for interactive commands,
+--   specifying whether safety checks should be ignored.
+data UseForce
+  = WithForce     -- ^ Ignore additional checks, like termination/positivity...
+  | WithoutForce  -- ^ Don't ignore any checks.
+  deriving (Eq, Read, Show)
+
 data OutputForm a b = OutputForm Range [ProblemId] (OutputConstraint a b)
   deriving (Functor)
 
@@ -342,7 +365,7 @@ outputFormId (OutputForm _ _ o) = out o
       FindInScopeOF _ _ _        -> __IMPOSSIBLE__
 
 instance Reify ProblemConstraint (Closure (OutputForm Expr Expr)) where
-  reify (PConstr pids cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm (getRange c) pids <$> reify c)
+  reify (PConstr pids cl) = enterClosure cl $ \c -> buildClosure =<< (OutputForm (getRange c) (Set.toList pids) <$> reify c)
 
 reifyElimToExpr :: I.Elim -> TCM Expr
 reifyElimToExpr e = case e of
@@ -380,7 +403,7 @@ instance Reify Constraint (OutputConstraint Expr Expr) where
               target  <- reify target
               let bs = TypedBindings noRange $ Arg ai $
                        TBind noRange xs domType
-                  e  = A.Lam Info.exprNoRange (DomainFull bs) body
+                  e  = A.Lam Info.defaultLamInfo_ (DomainFull bs) body
               return $ TypedAssign m' e target
             CheckArgs _ _ args t0 t1 _ -> do
               t0 <- reify t0
@@ -436,10 +459,10 @@ instance (Show a,Show b) => Show (OutputConstraint a b) where
     show (IsEmptyType a)        = "Is empty: " ++ show a
     show (SizeLtSat a)          = "Not empty type of sizes: " ++ show a
     show (FindInScopeOF s t cs) = "Resolve instance argument " ++ showCand (s,t) ++ ".\n  Candidates:\n    [ " ++
-                                    intercalate "\n    , " (map showCand cs) ++ " ]"
-      where showCand (tm,ty) = indent 6 $ show tm ++ " : " ++ show ty
-            indent n s = intercalate ("\n" ++ replicate n ' ') (l:ls)
-              where l:ls = lines s
+                                    List.intercalate "\n    , " (map showCand cs) ++ " ]"
+      where
+      showCand (tm,ty) = indent 6 $ show tm ++ " : " ++ show ty
+      indent n s = List.intercalate ("\n" ++ replicate n ' ') $ lines s
 
 instance (ToConcrete a c, ToConcrete b d) =>
          ToConcrete (OutputForm a b) (OutputForm c d) where
@@ -831,7 +854,7 @@ introTactic pmLambda ii = do
       hfs <- getRecordFieldNames d
       fs <- ifM showImplicitArguments
             (return $ map unArg hfs)
-            (return [ unArg a | a <- hfs, getHiding a == NotHidden ])
+            (return [ unArg a | a <- hfs, visible a ])
       let e = C.Rec noRange $ for fs $ \ f ->
             Left $ C.FieldAssignment f $ C.QuestionMark noRange Nothing
       return [ prettyShow e ]
@@ -849,12 +872,36 @@ atTopLevel m = inConcreteMode $ do
       tel <- lookupSection current
       -- Get the names of the local variables from @scope@
       -- and put them into the context.
+      --
+      -- Andreas, 2017-04-24, issue #2552:
+      --
+      -- Delete the let-bound ones, since they are not represented
+      -- in the module telescope.
+      --
+      -- This is a temporary fix until a better solution is available,
+      -- e.g., when the module telescope represents let-bound variables.
+      --
+      -- Unfortunately, referring to let-bound variables
+      -- from the top level module telescope will for now result in a not-in-scope error.
       let names :: [A.Name]
-          names = reverse $ map snd $ notShadowedLocals $ scopeLocals scope
-          types :: [Dom I.Type]
+          names = map localVar $ filter ((False ==) . localLetBound) $ map snd $ reverse $ scopeLocals scope
+      -- Andreas, 2016-12-31, issue #2371
+      -- The following is an unnecessary complication, as shadowed locals
+      -- are not in scope anyway (they are ambiguous).
+      -- -- Replace the shadowed names by fresh names (such that they do not shadow imports)
+      -- let mnames :: [Maybe A.Name]
+      --     mnames = map (notShadowedLocal . snd) $ reverse $ scopeLocals scope
+      -- names <- mapM (maybe freshNoName_ return) mnames
+      let types :: [Dom I.Type]
           types = map (snd <$>) $ telToList tel
           gamma :: ListTel' A.Name
-          gamma = zipWith' (\ x dom -> (x,) <$> dom) names types
+          gamma = fromMaybe __IMPOSSIBLE__ $
+                    zipWith' (\ x dom -> (x,) <$> dom) names types
+      reportSDoc "interaction.top" 20 $ TP.vcat
+        [ TP.text "BasicOps.atTopLevel"
+        , TP.text "  names = " TP.<+> TP.sep (map prettyA   names)
+        , TP.text "  types = " TP.<+> TP.sep (map prettyTCM types)
+        ]
       M.withCurrentModule current $
         withScope_ scope $
           addContext gamma $
