@@ -5,7 +5,7 @@ module Agda.Interaction.CommandLine where
 import Control.Monad.Reader
 import Control.Applicative
 
-import Data.List as List
+import qualified Data.List as List
 import Data.Maybe
 
 import Agda.Interaction.BasicOps as BasicOps hiding (parseExpr)
@@ -13,6 +13,7 @@ import Agda.Interaction.Monad
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Common
+import Agda.Syntax.Internal (telToList)
 import qualified Agda.Syntax.Internal as I
 import Agda.Syntax.Parser
 import Agda.Syntax.Position
@@ -29,6 +30,7 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Pretty ( PrettyTCM(prettyTCM) )
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Warnings (runPM)
 
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Monad
@@ -42,7 +44,7 @@ type Command a = (String, [String] -> TCM (ExitCode a))
 
 matchCommand :: String -> [Command a] -> Either [String] ([String] -> TCM (ExitCode a))
 matchCommand x cmds =
-    case List.filter (isPrefixOf x . fst) cmds of
+    case List.filter (List.isPrefixOf x . fst) cmds of
         [(_,m)] -> Right m
         xs      -> Left $ List.map fst xs
 
@@ -65,7 +67,7 @@ interaction prompt cmds eval = loop
                                     do  liftIO $ putStrLn $ "Unknown command '" ++ cmd ++ "'"
                                         loop
                                 Left xs ->
-                                    do  liftIO $ putStrLn $ "More than one command match: " ++ concat (intersperse ", " xs)
+                                    do  liftIO $ putStrLn $ "More than one command match: " ++ concat (List.intersperse ", " xs)
                                         loop
                     Just _ ->
                         do  go =<< liftTCM (eval $ fromJust ms)
@@ -202,7 +204,7 @@ actOnMeta _ _ = __IMPOSSIBLE__
 
 giveMeta :: [String] -> TCM ()
 giveMeta s | length s >= 2 = do
-  _ <- actOnMeta s (\ii -> \e -> give ii Nothing e)
+  _ <- actOnMeta s $ \ ii e -> give WithoutForce ii Nothing e
   return ()
 giveMeta _ = liftIO $ putStrLn $ ": give" ++ " metaid expr"
 
@@ -210,7 +212,7 @@ giveMeta _ = liftIO $ putStrLn $ ": give" ++ " metaid expr"
 
 refineMeta :: [String] -> TCM ()
 refineMeta s | length s >= 2 = do
-  _ <- actOnMeta s (\ii -> \e -> refine ii Nothing e)
+  _ <- actOnMeta s $ \ ii e -> refine WithoutForce ii Nothing e
   return ()
 refineMeta _ = liftIO $ putStrLn $ ": refine" ++ " metaid expr"
 
@@ -289,7 +291,8 @@ splashScreen = unlines
     , "        __/ /"
     , "        \\__/"
     , ""
-    , "The interactive mode is no longer supported. Don't complain if it doesn't work."
+    -- , "The interactive mode is no longer supported. Don't complain if it doesn't work."
+    , "The interactive mode is no longer under active development. Use at your own risk."
     ]
 
 -- | The help message
