@@ -7,12 +7,13 @@ import Control.Applicative
 import Control.Monad.State
 
 import Data.Char
-import Data.List
+import qualified Data.List as List
 import Data.Maybe
 
 import System.IO
 
 import Agda.Utils.Monad
+import Agda.Utils.Maybe
 import Agda.Utils.Pretty
 import Agda.Utils.String
 
@@ -23,7 +24,7 @@ import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
 import Agda.Interaction.Response as R
 import Agda.Interaction.InteractionTop
-import Agda.Interaction.EmacsCommand
+import Agda.Interaction.EmacsCommand hiding (putResponse)
 import Agda.Interaction.Highlighting.Emacs
 import Agda.Interaction.Options
 
@@ -46,70 +47,89 @@ mimicGHCi setup = do
       hSetEncoding  stdin  utf8
 
     setInteractionOutputCallback $
-        liftIO . mapM_ print <=< lispifyResponse
+        mapM_ print <=< lispifyResponse
 
-    handleCommand_ (lift setup) `evalStateT` initCommandState
+    commands <- liftIO $ initialiseCommandQueue readCommand
+
+    handleCommand_ (lift setup) `evalStateT` initCommandState commands
 
     opts <- commandLineOptions
-    _ <- interact' `runStateT` initCommandState { optionsOnReload = opts{ optAbsoluteIncludePaths = [] } }
+    _ <- interact' `runStateT`
+           (initCommandState commands)
+             { optionsOnReload = opts{ optAbsoluteIncludePaths = [] } }
     return ()
   where
+  interact' :: CommandM ()
+  interact' = do
+    Bench.reset
+    done <- Bench.billTo [] $ do
 
-    interact' :: CommandM ()
-    interact' = do
-      Bench.reset
-      done <- Bench.billTo [] $ do
+      liftIO $ do
+        putStr "Agda2> "
+        hFlush stdout
+      c <- nextCommand
+      case c of
+        Done      -> return True -- Done.
+        Error s   -> liftIO (putStrLn s) >> return False
+        Command c -> do
+          maybeAbort (runInteraction c)
+          return False
 
-        liftIO $ do
-          putStr "Agda2> "
-          hFlush stdout
-        done <- liftIO isEOF
-        unless done $ do
-            r <- liftIO getLine
-            _ <- return $! length r     -- force to read the full input line
-            case dropWhile isSpace r of
-                ""  -> return ()
-                ('-':'-':_) -> return ()
-                _ -> case listToMaybe $ reads r of
-                    Just (x, "")  -> runInteraction x
-                    Just (_, rem) -> liftIO $ putStrLn $ "not consumed: " ++ rem
-                    _ ->             liftIO $ putStrLn $ "cannot read: " ++ r
-        return done
+    lift Bench.print
+    unless done interact'
 
-      lift Bench.print
-      unless done interact'
+  -- Reads the next command from stdin.
+
+  readCommand :: IO Command
+  readCommand = do
+    done <- isEOF
+    if done then
+      return Done
+    else do
+      r <- getLine
+      _ <- return $! length r     -- force to read the full input line
+      case dropWhile isSpace r of
+        ""          -> readCommand
+        ('-':'-':_) -> readCommand
+        _           -> case listToMaybe $ reads r of
+          Just (x, "")  -> return $ Command x
+          Just (_, rem) -> return $ Error $ "not consumed: " ++ rem
+          _             -> return $ Error $ "cannot read: " ++ r
+
+-- | Given strings of goals, warnings and errors, return a pair of the
+--   body and the title for the info buffer
+formatWarningsAndErrors :: String -> String -> String -> (String, String)
+formatWarningsAndErrors g w e = (body, title)
+  where
+    isG = not $ null g
+    isW = not $ null w
+    isE = not $ null e
+    title = List.intercalate "," $ catMaybes
+              [ " Goals"    <$ guard isG
+              , " Warnings" <$ guard isW
+              , " Errors"   <$ guard isE
+              , " Done"     <$ guard (not (isG || isW || isE))
+              ]
+
+    body = List.intercalate "\n" $ catMaybes
+             [ g                    <$ guard isG
+             , delimiter "Warnings" <$ guard (isW && (isG || isE))
+             , w                    <$ guard isW
+             , delimiter "Errors"   <$ guard (isE && (isG || isW))
+             , e                    <$ guard isE
+             ]
 
 -- | Convert Response to an elisp value for the interactive emacs frontend.
 
-lispifyResponse :: Response -> TCM [Lisp String]
-lispifyResponse (Resp_HighlightingInfo info modFile) =
-  (:[]) <$> lispifyHighlightingInfo info modFile
+lispifyResponse :: Response -> IO [Lisp String]
+lispifyResponse (Resp_HighlightingInfo info method modFile) =
+  (:[]) <$> lispifyHighlightingInfo info method modFile
 lispifyResponse (Resp_DisplayInfo info) = return $ case info of
-    Info_CompilationOk -> f "The module was successfully compiled." "*Compilation result*"
+    Info_CompilationOk w e -> f body "*Compilation result*"
+      where (body, _) = formatWarningsAndErrors "The module was successfully compiled.\n" w e -- abusing the goals field since we ignore the title
     Info_Constraints s -> f s "*Constraints*"
-    Info_AllGoalsWarnings g  w e -> f body ("*All" ++ title ++ "*")
-      where
-        isG = not $ null g
-        isW = not $ null w
-        isE = not $ null e
-        title = intercalate "," $ catMaybes
-                  [ " Goals"    <$ guard isG
-                  , " Warnings" <$ guard isW
-                  , " Errors"   <$ guard isE
-                  , " Done"     <$ guard (not (isG || isW || isE))
-                  ]
-        delimiter s = concat [ replicate 4 '\x2014'
-                             , " ", s, " "
-                             , replicate (54 - length s) '\x2014'
-                             ]
-
-        body = intercalate "\n" $ catMaybes
-                 [ g                    <$ guard isG
-                 , delimiter "Warnings" <$ guard (isW && (isG || isE))
-                 , w                    <$ guard isW
-                 , delimiter "Errors"   <$ guard (isE && (isG || isW))
-                 , e                    <$ guard isE
-                 ]
+    Info_AllGoalsWarnings g w e -> f body ("*All" ++ title ++ "*")
+      where (body, title) = formatWarningsAndErrors g w e
     Info_Auto s -> f s "*Auto*"
     Info_Error s -> f s "*Error*"
     -- FNF: if Info_Warning comes back into use, the above should be
@@ -134,6 +154,7 @@ lispifyResponse (Resp_DisplayInfo info) = return $ case info of
     Info_Version -> f ("Agda version " ++ versionWithCommitInfo) "*Agda Version*"
   where f content bufname = [ display_info' False bufname content ]
 lispifyResponse Resp_ClearHighlighting = return [ L [ A "agda2-highlight-clear" ] ]
+lispifyResponse Resp_DoneAborting = return [ L [ A "agda2-abort-done" ] ]
 lispifyResponse Resp_ClearRunningInfo = return [ clearRunningInfo ]
 -- FNF: if Info_Warning comes back into use, the above should be
 -- return [ clearRunningInfo, clearWarning ]
@@ -142,12 +163,10 @@ lispifyResponse (Resp_RunningInfo n s)
   | otherwise = return [ L [A "agda2-verbose", A (quote s)] ]
 lispifyResponse (Resp_Status s)
     = return [ L [ A "agda2-status-action"
-                 , A (quote $ intercalate "," $ catMaybes [checked, showImpl])
+                 , A (quote $ List.intercalate "," $ catMaybes [checked, showImpl])
                  ]
              ]
   where
-    boolToMaybe b x = if b then Just x else Nothing
-
     checked  = boolToMaybe (sChecked               s) "Checked"
     showImpl = boolToMaybe (sShowImplicitArguments s) "ShowImplicit"
 
