@@ -38,7 +38,8 @@ import Agda.Utils.Impossible
 toAtoms :: Aspects -> [String]
 toAtoms m = map toAtom (otherAspects m) ++ toAtoms' (aspect m)
   where
-  toAtom x = map toLower (show x)
+  toAtom :: Show a => a -> String
+  toAtom = map toLower . show
 
   kindToAtom (Constructor Inductive)   = "inductiveconstructor"
   kindToAtom (Constructor CoInductive) = "coinductiveconstructor"
@@ -58,20 +59,18 @@ showAspects
   :: ModuleToSource
      -- ^ Must contain a mapping for the definition site's module, if any.
   -> (Range, Aspects) -> Lisp String
-showAspects modFile (r, m) =
-    L $ ((map (A . show) [from r, to r])
-           ++
-         [L $ map A $ toAtoms m]
-           ++
-         (A $ maybe "nil" quote $ note m)
-         :
-          defSite)
+showAspects modFile (r, m) = L $
+    (map (A . show) [from r, to r])
+      ++
+    [L $ map A $ toAtoms m]
+      ++
+    [A $ maybe "nil" quote $ note m]
+      ++
+    (maybeToList $ fmap defSite $ definitionSite m)
   where
-  defSite = case definitionSite m of
-    Nothing     -> []
-    Just (m, p) -> case Map.lookup m modFile of
-      Nothing -> __IMPOSSIBLE__
-      Just f  -> [Cons (A $ quote $ filePath f) (A $ show p)]
+  defSite (DefinitionSite m p _ _) =
+    Cons (A $ quote $ filePath f) (A $ show p)
+    where f = Map.findWithDefault __IMPOSSIBLE__ m modFile
 
 -- | Turns syntax highlighting information into a list of
 -- S-expressions.
@@ -83,12 +82,12 @@ showAspects modFile (r, m) =
 
 lispifyHighlightingInfo
   :: HighlightingInfo
+  -> HighlightingMethod
   -> ModuleToSource
      -- ^ Must contain a mapping for every definition site's module.
-  -> TCM (Lisp String)
-lispifyHighlightingInfo h modFile = do
-  method <- envHighlightingMethod <$> ask
-  liftIO $ case ranges h of
+  -> IO (Lisp String)
+lispifyHighlightingInfo h method modFile = do
+  case ranges h of
     _             | method == Direct                   -> direct
     ((_, mi) : _) | otherAspects mi == [TypeChecks] ||
                     mi == mempty                       -> direct
