@@ -43,10 +43,12 @@ import Agda.TypeChecking.Monad
   hiding (MetaInfo, Primitive, Constructor, Record, Function, Datatype)
 import qualified Agda.TypeChecking.Monad as M
 import Agda.TypeChecking.Positivity.Occurrence
+import Agda.TypeChecking.Warnings (runPM)
 
 import qualified Agda.Syntax.Abstract as A
 import Agda.Syntax.Concrete (FieldAssignment'(..))
 import qualified Agda.Syntax.Common as Common
+import qualified Agda.Syntax.Concrete.Name as C
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Fixity
 import qualified Agda.Syntax.Info as SI
@@ -64,6 +66,7 @@ import Agda.Utils.List
 import Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Null
+import Agda.Utils.Pretty
 import Agda.Utils.HashMap (HashMap)
 import qualified Agda.Utils.HashMap as HMap
 
@@ -109,14 +112,17 @@ highlightAsTypeChecked rPre r m
 -- | Lispify and print the given highlighting information.
 
 printHighlightingInfo :: MonadTCM tcm => HighlightingInfo -> tcm ()
-printHighlightingInfo x = do
+printHighlightingInfo info = do
   modToSrc <- use stModuleToSource
-  liftTCM $ reportSLn "highlighting" 50 $
-    "Printing highlighting info:\n" ++ show x ++ "\n" ++
-    "  modToSrc = " ++ show modToSrc
-  unless (null $ ranges x) $ do
+  method   <- view eHighlightingMethod
+  liftTCM $ reportSLn "highlighting" 50 $ unlines
+    [ "Printing highlighting info:"
+    , show info
+    , "  modToSrc = " ++ show modToSrc
+    ]
+  unless (null $ ranges info) $ do
     liftTCM $ appInteractionOutputCallback $
-        Resp_HighlightingInfo x modToSrc
+        Resp_HighlightingInfo info method modToSrc
 
 -- | Highlighting levels.
 
@@ -129,20 +135,24 @@ data Level
     --   constructors.
 
 -- | Generate syntax highlighting information for the given
--- declaration, and (if appropriate) print it. If the
--- 'HighlightingLevel' is @'Full'@, then the state is
--- additionally updated with the new highlighting info (in case of a
--- conflict new info takes precedence over old info).
+-- declaration, and (if appropriate) print it. If the boolean is
+-- 'True', then the state is additionally updated with the new
+-- highlighting info (in case of a conflict new info takes precedence
+-- over old info).
 --
 -- The procedure makes use of some of the token highlighting info in
 -- 'stTokens' (that corresponding to the interval covered by the
--- declaration). If the 'HighlightingLevel' is @'Full'@,
--- then this token highlighting info is additionally removed from
--- 'stTokens'.
+-- declaration). If the boolean is 'True', then this token
+-- highlighting info is additionally removed from 'stTokens'.
 
-generateAndPrintSyntaxInfo :: A.Declaration -> Level -> TCM ()
-generateAndPrintSyntaxInfo decl _ | null $ P.getRange decl = return ()
-generateAndPrintSyntaxInfo decl hlLevel = do
+generateAndPrintSyntaxInfo
+  :: A.Declaration
+  -> Level
+  -> Bool
+     -- ^ Update the state?
+  -> TCM ()
+generateAndPrintSyntaxInfo decl _ _ | null $ P.getRange decl = return ()
+generateAndPrintSyntaxInfo decl hlLevel updateState = do
   file <- fromMaybe __IMPOSSIBLE__ <$> asks envCurrentPath
 
   reportSLn "import.iface.create" 15 $
@@ -152,7 +162,7 @@ generateAndPrintSyntaxInfo decl hlLevel = do
             Partial {} -> "(first approximation)"
         ++ "."
 
-  reportSLn "highlighting.names" 60 $ "highlighting names = " ++ show names
+  reportSLn "highlighting.names" 60 $ "highlighting names = " ++ prettyShow names
 
   M.ignoreAbstractMode $ do
     modMap <- sourceToModule
@@ -166,7 +176,11 @@ generateAndPrintSyntaxInfo decl hlLevel = do
       Full{} -> generateConstructorInfo modMap file kinds decl
       _      -> return mempty
 
-    warnInfo <- Fold.fold . map (warningHighlighting . tcWarning) <$> use stTCWarnings
+    cm <- P.rangeFile <$> view eRange
+    reportSLn "highlighting.warning" 60 $ "current path = " ++ show cm
+
+    warnInfo <- Fold.foldMap warningHighlighting
+                 . filter ((cm ==) . tcWarningOrigin) <$> use stTCWarnings
 
     let (from, to) = case P.rangeToInterval (P.getRange decl) of
           Nothing -> __IMPOSSIBLE__
@@ -189,10 +203,9 @@ generateAndPrintSyntaxInfo decl hlLevel = do
                        `mappend`
                      curTokens
 
-    case hlLevel of
-      Full{} -> do stSyntaxInfo %= mappend syntaxInfo
-                   stTokens     .= prevTokens `mappend` postTokens
-      _      -> return ()
+    when updateState $ do
+      stSyntaxInfo %= mappend syntaxInfo
+      stTokens     .= prevTokens `mappend` postTokens
 
     ifTopLevelAndHighlightingLevelIs NonInteractive $
       printHighlightingInfo syntaxInfo
@@ -243,12 +256,14 @@ generateAndPrintSyntaxInfo decl hlLevel = do
                           (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
                           Nothing
 
+    -- For top level modules, we set the binding site to the beginning of the file
+    -- so that clicking on an imported module will jump to the beginning of the file
+    -- which defines this module.
     mod isTopLevelModule n =
       nameToFile modMap file []
                  (A.nameConcrete n) P.noRange
                  (\isOp -> mempty { aspect = Just $ Name (Just Module) isOp })
-                 (Just $ (if isTopLevelModule then P.beginningOfFile else id)
-                           (A.nameBindingSite n))
+                 (Just $ applyWhen isTopLevelModule P.beginningOfFile $ A.nameBindingSite n)
 
     getVarAndField :: A.Expr -> File
     getVarAndField (A.Var x)            = bound x
@@ -375,7 +390,9 @@ tokenHighlighting = merge . map tokenToCFile
   tokenToCFile (T.TokTeX (i, _))                = aToF Comment (P.getRange i)
   tokenToCFile (T.TokId {})                     = mempty
   tokenToCFile (T.TokQId {})                    = mempty
-  tokenToCFile (T.TokString {})                 = mempty
+  tokenToCFile (T.TokString (i,s))
+    | "--" `isPrefixOf` s                       = aToF Option (P.getRange i)
+    | otherwise                                 = mempty
   tokenToCFile (T.TokDummy {})                  = mempty
   tokenToCFile (T.TokEOF {})                    = mempty
 
@@ -424,7 +441,7 @@ nameKinds hlLevel decl = do
   defnToKind   M.Record{}                          = Record
   defnToKind   M.Constructor{ M.conInd = i }       = Constructor i
   defnToKind   M.Primitive{}                       = Primitive
-  defnToKind   M.AbstractDefn                      = __IMPOSSIBLE__
+  defnToKind   M.AbstractDefn{}                    = __IMPOSSIBLE__
 
   declToKind :: A.Declaration ->
                 HashMap A.QName NameKind -> HashMap A.QName NameKind
@@ -530,12 +547,33 @@ errorHighlighting e = do
 
 -- | Generate syntax highlighting for warnings.
 
-warningHighlighting :: Warning -> File
-warningHighlighting w = case w of
-  TerminationIssue terrs    -> terminationErrorHighlighting terrs
-  NotStrictlyPositive d ocs -> positivityErrorHighlighting d ocs
-  _ -> mempty
-
+warningHighlighting :: TCWarning -> File
+warningHighlighting w = case tcWarning w of
+  TerminationIssue terrs     -> terminationErrorHighlighting terrs
+  NotStrictlyPositive d ocs  -> positivityErrorHighlighting d ocs
+  UnreachableClauses{}       -> unreachableErrorHighlighting $ P.getRange w
+  CoverageIssue{}            -> coverageErrorHighlighting $ P.getRange w
+  CoverageNoExactSplit{}     -> catchallHighlighting $ P.getRange w
+  -- expanded catch-all case to get a warning for new constructors
+  UnsolvedMetaVariables{}    -> mempty
+  UnsolvedInteractionMetas{} -> mempty
+  UnsolvedConstraints{}      -> mempty
+  OldBuiltin{}               -> mempty
+  EmptyRewritePragma{}       -> mempty
+  UselessPublic{}            -> mempty
+  UselessInline{}            -> mempty
+  ParseWarning{}             -> mempty
+  GenericWarning{}           -> mempty
+  GenericNonFatalError{}     -> mempty
+  SafeFlagPostulate{}        -> mempty
+  SafeFlagPragma{}           -> mempty
+  SafeFlagNonTerminating     -> mempty
+  SafeFlagTerminating        -> mempty
+  SafeFlagPrimTrustMe        -> mempty
+  SafeFlagNoPositivityCheck  -> mempty
+  SafeFlagPolarity           -> mempty
+  DeprecationWarning{}       -> mempty
+  NicifierIssue{}            -> mempty
 
 -- | Generate syntax highlighting for termination errors.
 
@@ -558,6 +596,19 @@ positivityErrorHighlighting q o = several (rToR <$> P.getRange q : rs) m
     rs = case o of Unknown -> []; Known r _ -> [r]
     m  = mempty { otherAspects = [PositivityProblem] }
 
+unreachableErrorHighlighting :: P.Range -> File
+unreachableErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+  where m = mempty { otherAspects = [ReachabilityProblem] }
+
+coverageErrorHighlighting :: P.Range -> File
+coverageErrorHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+  where m = mempty { otherAspects = [CoverageProblem] }
+
+
+catchallHighlighting :: P.Range -> File
+catchallHighlighting r = singleton (rToR $ P.continuousPerLine r) m
+  where m = mempty { otherAspects = [CatchallClause] }
+
 
 -- | Generates and prints syntax highlighting information for unsolved
 -- meta-variables and certain unsolved constraints.
@@ -650,18 +701,65 @@ nameToFile :: SourceToModule
 nameToFile modMap file xs x fr m mR =
   -- We don't care if we get any funny ranges.
   if all (== Strict.Just file) fileNames then
+    frFile `mappend`
     several (map rToR rs)
-            ((m $ C.isOperator x) { definitionSite = mFilePos })
+            (aspects { definitionSite = mFilePos })
    else
     mempty
   where
+  aspects    = m $ C.isOperator x
   fileNames  = catMaybes $ map (fmap P.srcFile . P.rStart . P.getRange) (x : xs)
-  rs         = applyWhen (not $ null fr) (fr :) $ map P.getRange (x : xs)
+  frFile     = singleton (rToR fr) (aspects { definitionSite = notHere <$> mFilePos })
+  rs         = map P.getRange (x : xs)
+
+  -- The fixity declaration should not get a symbolic anchor.
+  notHere d = d { defSiteHere = False }
+
+  mFilePos  :: Maybe DefinitionSite
   mFilePos   = do
     r <- mR
     P.Pn { P.srcFile = Strict.Just f, P.posPos = p } <- P.rStart r
     mod <- Map.lookup f modMap
-    return (mod, fromIntegral p)
+    -- Andreas, 2017-06-16, Issue #2604: Symbolic anchors.
+    -- We drop the file name part from the qualifiers, since
+    -- this is contained in the html file name already.
+    -- We want to get anchors of the form:
+    -- @<a name="TopLevelModule.html#LocalModule.NestedModule.identifier">@
+    let qualifiers = drop (length $ C.moduleNameParts mod) xs
+    -- For bound variables, we do not create symbolic anchors.
+        local = maybe True isLocalAspect $ aspect aspects
+    return $ DefinitionSite
+      { defSiteModule = mod
+      , defSitePos    = fromIntegral p
+        -- Is our current position the definition site?
+      , defSiteHere   = r == P.getRange x
+        -- For bound variables etc. we do not create a symbolic anchor name.
+        -- Also not for names that include anonymous modules,
+        -- otherwise, we do not get unique anchors.
+      , defSiteAnchor = if local || C.isNoName x || any Common.isUnderscore qualifiers
+          then Nothing
+          else Just $ prettyShow $ foldr C.Qual (C.QName x) qualifiers
+      }
+
+  -- Is the name a bound variable or similar? If in doubt, yes.
+  isLocalAspect :: Aspect -> Bool
+  isLocalAspect = \case
+    Name mkind _ -> maybe True isLocal mkind
+    _ -> True
+  isLocal :: NameKind -> Bool
+  isLocal = \case
+    Bound         -> True
+    Argument      -> True
+    Constructor{} -> False
+    Datatype      -> False
+    Field         -> False
+    Function      -> False
+    Module        -> False
+    Postulate     -> False
+    Primitive     -> False
+    Record        -> False
+    Macro         -> False
+
 
 -- | A variant of 'nameToFile' for qualified abstract names.
 
