@@ -1,4 +1,5 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP               #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 -- | Function for generating highlighted, hyperlinked HTML from Agda
 -- sources.
@@ -14,21 +15,31 @@ module Agda.Interaction.Highlighting.HTML
   , code
   ) where
 
+import Prelude hiding ((!!), concatMap)
 import Control.Applicative
 import Control.Monad
 import Control.Monad.Trans
 
 import Data.Function
 import Data.Monoid
+import Data.Foldable (toList, concatMap)
 import Data.Maybe
 import qualified Data.IntMap as IntMap
 import qualified Data.Map    as Map
 import qualified Data.List   as List
+import Data.Text.Lazy (Text)
+
+import qualified Network.URI.Encode
 
 import System.FilePath
 import System.Directory
 
-import Text.XHtml.Strict
+import Text.Blaze.Html5 hiding (code, map, title)
+import qualified Text.Blaze.Html5 as Html5
+import Text.Blaze.Html5.Attributes as Attr
+import Text.Blaze.Html.Renderer.Text
+  -- The imported operator (!) attaches an Attribute to an Html value
+  -- The defined operator (!!) attaches a list of such Attributes
 
 import Paths_Agda
 
@@ -42,9 +53,10 @@ import Agda.TypeChecking.Monad (TCM)
 import qualified Agda.TypeChecking.Monad as TCM
 
 import Agda.Utils.FileName (filePath)
+import Agda.Utils.Function
 import Agda.Utils.Lens
 import qualified Agda.Utils.IO.UTF8 as UTF8
-import Agda.Utils.Pretty
+import Agda.Utils.Pretty hiding ((<>))
 import Agda.Utils.Tuple
 
 #include "undefined.h"
@@ -64,8 +76,10 @@ defaultCSSFile = "Agda.css"
 generateHTML :: TCM ()
 generateHTML = generateHTMLWithPageGen pageGen
   where
+  pageGen :: FilePath -> C.TopLevelModuleName -> CompressedFile -> TCM ()
   pageGen dir mod hinfo = generatePage renderer dir mod
     where
+    renderer :: FilePath -> FilePath -> String -> Text
     renderer css _ contents = page css mod $ code $ tokenStream contents hinfo
 
 -- | Prepare information for HTML page generation.
@@ -105,53 +119,54 @@ generateHTMLWithPageGen generatePage = do
 -- | Converts module names to the corresponding HTML file names.
 
 modToFile :: C.TopLevelModuleName -> FilePath
-modToFile m = render (pretty m) <.> "html"
+modToFile m =
+  Network.URI.Encode.encode $
+    render (pretty m) <.> "html"
 
 -- | Generates a highlighted, hyperlinked version of the given module.
 
 generatePage
-  :: (FilePath -> FilePath -> String -> String)  -- ^ Page renderer
+  :: (FilePath -> FilePath -> String -> Text)  -- ^ Page renderer
   -> FilePath              -- ^ Directory in which to create files.
   -> C.TopLevelModuleName  -- ^ Module to be highlighted.
   -> TCM ()
 generatePage renderpage dir mod = do
-  mf <- Map.lookup mod <$> use TCM.stModuleToSource
-  case mf of
-    Nothing -> __IMPOSSIBLE__
-    Just f  -> do
-      contents <- liftIO $ UTF8.readTextFile $ filePath f
-      css      <- maybe defaultCSSFile id . optCSSFile <$>
-                    TCM.commandLineOptions
-      let html = renderpage css (filePath f) contents
-      TCM.reportSLn "html" 1 $ "Generating HTML for " ++
-                               render (pretty mod) ++
-                               " (" ++ target ++ ")."
-      liftIO $ UTF8.writeFile target html
+  f <- fromMaybe __IMPOSSIBLE__ . Map.lookup mod <$> use TCM.stModuleToSource
+  contents <- liftIO $ UTF8.readTextFile $ filePath f
+  css      <- fromMaybe defaultCSSFile . optCSSFile <$>
+                TCM.commandLineOptions
+  let html = renderpage css (filePath f) contents
+  TCM.reportSLn "html" 1 $ "Generating HTML for " ++
+                           render (pretty mod) ++
+                           " (" ++ target ++ ")."
+  liftIO $ UTF8.writeTextToFile target html
   where target = dir </> modToFile mod
 
+
+-- | Attach multiple Attributes
+
+(!!) :: Html -> [Attribute] -> Html
+h !! as = h ! mconcat as
+
 -- | Constructs the web page, including headers.
 
 page :: FilePath              -- ^ URL to the CSS file.
      -> C.TopLevelModuleName  -- ^ Module to be highlighted.
      -> Html
-     -> String
-page css modName pagecontent = renderHtml $
-  header (thetitle << render (pretty modName)
-            +++
-          meta ! [ httpequiv "Content-Type"
-                 , content "text/html; charset=UTF-8"
-                 ]
-            +++
-          meta ! [ httpequiv "Content-Style-Type"
-                 , content "text/css"
-                 ]
-            +++
-          thelink noHtml ! [ href css
-                           , rel "stylesheet"
-                           , thetype "text/css"
-                           ])
-  +++
-  body << pre << pagecontent
+     -> Text
+page css modName pagecontent = renderHtml $ docTypeHtml $ hdr <> rest
+
+  where
+
+    hdr = Html5.head $ mconcat
+      [ meta !! [ charset "utf-8" ]
+      , Html5.title (toHtml $ render $ pretty modName)
+      , link !! [ rel "stylesheet"
+                , href (stringValue css)
+                ]
+      ]
+
+    rest = body (pre pagecontent)
 
 -- | Constructs token stream ready to print.
 
@@ -162,7 +177,7 @@ tokenStream
 tokenStream contents info =
   map (\cs -> case cs of
           (mi, (pos, _)) : _ ->
-            (pos, map (snd . snd) cs, maybe mempty id mi)
+            (pos, map (snd . snd) cs, fromMaybe mempty mi)
           [] -> __IMPOSSIBLE__) $
   List.groupBy ((==) `on` fst) $
   map (\(pos, c) -> (IntMap.lookup pos infoMap, (pos, c))) $
@@ -174,26 +189,32 @@ tokenStream contents info =
 
 code :: [(Int, String, Aspects)]
      -> Html
-code = mconcat . map (\(pos, s, mi) -> annotate pos mi (stringToHtml s))
+code = mconcat . map mkHtml
   where
-  annotate :: Int -> Aspects -> Html -> Html
-  annotate pos mi = anchor ! attributes
-    where
-    attributes =
-      [name (show pos)] ++
-      maybe [] link (definitionSite mi) ++
-      (case classes of
-        [] -> []
-        cs -> [theclass $ unwords cs])
+  mkHtml :: (Int, String, Aspects) -> Html
+  mkHtml (pos, s, mi) =
+    -- Andreas, 2017-06-16, issue #2605:
+    -- Do not create anchors for whitespace.
+    applyUnless (mi == mempty) (annotate pos mi) $ toHtml s
 
-    classes =
-      maybe [] noteClasses (note mi)
-      ++ otherAspectClasses (otherAspects mi)
-      ++ maybe [] aspectClasses (aspect mi)
+  annotate :: Int -> Aspects -> Html -> Html
+  annotate pos mi content = a content !! attributes
+    where
+    attributes = concat
+      [ [Attr.id $ stringValue $ applyWhen here anchorName $ show pos ]
+      , toList $ fmap link mDefinitionSite
+      , class_ (stringValue $ unwords classes) <$ guard (not $ null classes)
+      ]
+
+    classes = concat
+      [ concatMap noteClasses (note mi)
+      , otherAspectClasses (otherAspects mi)
+      , concatMap aspectClasses (aspect mi)
+      ]
 
     aspectClasses (Name mKind op) = kindClass ++ opClass
       where
-      kindClass = maybe [] ((: []) . showKind) mKind
+      kindClass = toList $ fmap showKind mKind
 
       showKind (Constructor Inductive)   = "InductiveConstructor"
       showKind (Constructor CoInductive) = "CoinductiveConstructor"
@@ -207,4 +228,13 @@ code = mconcat . map (\(pos, s, mi) -> annotate pos mi (stringToHtml s))
     -- Notes are not included.
     noteClasses s = []
 
-    link (m, pos) = [href $ modToFile m ++ "#" ++ show pos]
+    mDefinitionSite = definitionSite mi
+    here       = maybe False defSiteHere mDefinitionSite
+    anchorName = (`fromMaybe` maybe __IMPOSSIBLE__ defSiteAnchor mDefinitionSite)
+    link (DefinitionSite m pos _here aName) = href $ stringValue $
+      -- If the definition site points to the top of a file,
+      -- we drop the anchor part and just link to the file.
+      applyUnless (pos <= 1)
+        (++ "#" ++
+         Network.URI.Encode.encode (fromMaybe (show pos) aName))
+        (Network.URI.Encode.encode $ modToFile m)
