@@ -24,7 +24,7 @@ import qualified Data.Map as Map
 import qualified Data.List as List
 import qualified Data.Set as Set
 import qualified Data.Foldable as Fold (toList)
-import Data.List hiding (null)
+import qualified Data.List as List
 import Data.Maybe
 import Data.Monoid (mempty, mappend)
 import Data.Map (Map)
@@ -47,6 +47,7 @@ import Agda.Syntax.Translation.ConcreteToAbstract
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Errors
+import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.MetaVars ( openMetasToPostulates )
 import Agda.TypeChecking.Monad
@@ -54,7 +55,6 @@ import Agda.TypeChecking.Serialise
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Pretty as P
-import Agda.TypeChecking.Rewriting (killCtxId)
 import Agda.TypeChecking.DeadCode
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
 
@@ -72,10 +72,11 @@ import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.FileName
 import Agda.Utils.Lens
 import Agda.Utils.Maybe
+import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.IO.Binary
-import Agda.Utils.Pretty
+import Agda.Utils.Pretty hiding (Mode)
 import Agda.Utils.Time
 import Agda.Utils.Hash
 import qualified Agda.Utils.HashMap as HMap
@@ -84,13 +85,33 @@ import qualified Agda.Utils.Trie as Trie
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+-- | Is the aim to type-check the top-level module, or only to
+-- scope-check it?
+
+data Mode
+  = ScopeCheck
+  | TypeCheck
+  deriving (Eq, Show)
+
 -- | Are we loading the interface for the user-loaded file
 --   or for an import?
 data MainInterface
-  = MainInterface     -- ^ Interface for main file.
-  | NotMainInterface  -- ^ Interface for imported file.
+  = MainInterface Mode -- ^ For the main file.
+                       --
+                       --   In this case state changes inflicted by
+                       --   'createInterface' are preserved.
+  | NotMainInterface   -- ^ For an imported file.
+                       --
+                       --   In this case state changes inflicted by
+                       --   'createInterface' are not preserved.
   deriving (Eq, Show)
 
+-- | Should state changes inflicted by 'createInterface' be preserved?
+
+includeStateChanges :: MainInterface -> Bool
+includeStateChanges (MainInterface _) = True
+includeStateChanges NotMainInterface  = False
+
 -- | Merge an interface into the current proof state.
 mergeInterface :: Interface -> TCM ()
 mergeInterface i = do
@@ -98,6 +119,7 @@ mergeInterface i = do
         builtin = Map.toList $ iBuiltin i
         prim    = [ x | (_,Prim x) <- builtin ]
         bi      = Map.fromList [ (x,Builtin t) | (x,Builtin t) <- builtin ]
+        warns   = iWarnings i
     bs <- gets stBuiltinThings
     reportSLn "import.iface.merge" 10 $ "Merging interface"
     reportSLn "import.iface.merge" 20 $
@@ -112,7 +134,7 @@ mergeInterface i = do
             Just b1 = Map.lookup b bs
             Just b2 = Map.lookup b bi
     mapM_ check (map fst $ Map.toList $ Map.intersection bs bi)
-    addImportedThings sig bi (iHaskellImports i) (iHaskellImportsUHC i) (iPatternSyns i) (iDisplayForms i)
+    addImportedThings sig bi (iPatternSyns i) (iDisplayForms i) warns
     reportSLn "import.iface.merge" 20 $
       "  Rebinding primitives " ++ show prim
     mapM_ rebind prim
@@ -123,16 +145,13 @@ mergeInterface i = do
 
 addImportedThings ::
   Signature -> BuiltinThings PrimFun ->
-  Set String -> -- MAlonzo imoprts
-  Set String -> -- UHC backend imports
-  A.PatternSynDefns -> DisplayForms -> TCM ()
-addImportedThings isig ibuiltin hsImports hsImportsUHC patsyns display = do
-  stImports              %= \ imp -> unionSignatures [imp, over sigRewriteRules killCtxId isig]
+  A.PatternSynDefns -> DisplayForms -> [TCWarning] -> TCM ()
+addImportedThings isig ibuiltin patsyns display warnings = do
+  stImports              %= \ imp -> unionSignatures [imp, isig]
   stImportedBuiltins     %= \ imp -> Map.union imp ibuiltin
-  stHaskellImports       %= \ imp -> Set.union imp hsImports
-  stHaskellImportsUHC    %= \ imp -> Set.union imp hsImportsUHC
   stPatternSynImports    %= \ imp -> Map.union imp patsyns
   stImportedDisplayForms %= \ imp -> HMap.unionWith (++) imp display
+  stTCWarnings           %= \ imp -> List.union imp warnings
   addImportedInstances isig
 
 -- | Scope checks the given module. A proper version of the module
@@ -144,7 +163,7 @@ scopeCheckImport x = do
     verboseS "import.scope" 10 $ do
       visited <- Map.keys <$> getVisitedModules
       reportSLn "import.scope" 10 $
-        "  visited: " ++ intercalate ", " (map prettyShow visited)
+        "  visited: " ++ List.intercalate ", " (map prettyShow visited)
     -- Since scopeCheckImport is called from the scope checker,
     -- we need to reimburse her account.
     i <- Bench.billTo [] $ getInterface x
@@ -174,12 +193,14 @@ hasWarnings = not . null
 -- | If the module has already been visited (without warnings), then
 -- its interface is returned directly. Otherwise the computation is
 -- used to find the interface and the computed interface is stored for
--- potential later use.
+-- potential later use (unless the 'MainInterface' is @'MainInterface'
+-- 'ScopeCheck'@).
 
 alreadyVisited :: C.TopLevelModuleName ->
+                  MainInterface ->
                   TCM (Interface, MaybeWarnings) ->
                   TCM (Interface, MaybeWarnings)
-alreadyVisited x getIface = do
+alreadyVisited x isMain getIface = do
     mm <- getVisitedModule x
     case mm of
         -- A module with warnings should never be allowed to be
@@ -191,10 +212,11 @@ alreadyVisited x getIface = do
           reportSLn "import.visit" 5 $ "  Getting interface for " ++ prettyShow x
           r@(i, wt) <- getIface
           reportSLn "import.visit" 5 $ "  Now we've looked at " ++ prettyShow x
-          visitModule $ ModuleInfo
-            { miInterface  = i
-            , miWarnings   = hasWarnings wt
-            }
+          unless (isMain == MainInterface ScopeCheck) $
+            visitModule $ ModuleInfo
+              { miInterface  = i
+              , miWarnings   = hasWarnings wt
+              }
           return r
 
 -- | Type checks the main file of the interaction.
@@ -203,9 +225,16 @@ alreadyVisited x getIface = do
 --
 --   First, the primitive modules are imported.
 --   Then, @getInterface'@ is called to do the main work.
+--
+--   If the 'Mode' is 'ScopeCheck', then type-checking is not
+--   performed, only scope-checking. (This may include type-checking
+--   of imported modules.) In this case the generated, partial
+--   interface is not stored in the state ('stDecodedModules'). Note,
+--   however, that if the file has already been type-checked, then a
+--   complete interface is returned.
 
-typeCheckMain :: AbsolutePath -> TCM (Interface, MaybeWarnings)
-typeCheckMain f = do
+typeCheckMain :: AbsolutePath -> Mode -> TCM (Interface, MaybeWarnings)
+typeCheckMain f mode = do
   -- liftIO $ putStrLn $ "This is typeCheckMain " ++ prettyShow f
   -- liftIO . putStrLn . show =<< getVerbosity
   reportSLn "import.main" 10 $ "Importing the primitive modules."
@@ -229,7 +258,7 @@ typeCheckMain f = do
 
   -- Now do the type checking via getInterface.
   m <- moduleName f
-  getInterface' m MainInterface
+  getInterface' m (MainInterface mode)
 
 -- | Tries to return the interface associated to the given (imported) module.
 --   The time stamp of the relevant interface file is also returned.
@@ -263,19 +292,17 @@ getInterface_ x = do
 getInterface'
   :: C.TopLevelModuleName
   -> MainInterface
-     -- ^ If type checking is necessary,
-     --   should all state changes inflicted by 'createInterface' be preserved?
-     --   Yes, if we are the 'MainInterface'.  No, if we are 'NotMainInterface'.
   -> TCM (Interface, MaybeWarnings)
 getInterface' x isMain = do
   withIncreasedModuleNestingLevel $ do
-    -- Preserve the pragma options unless includeStateChanges is True.
+    -- Preserve the pragma options unless we are checking the main
+    -- interface.
     bracket_ (use stPragmaOptions)
-             (unless includeStateChanges . setPragmaOptions) $ do
+             (unless (includeStateChanges isMain) . setPragmaOptions) $ do
      -- Forget the pragma options (locally).
      setCommandLineOptions . stPersistentOptions . stPersistentState =<< get
 
-     alreadyVisited x $ addImportCycleCheck x $ do
+     alreadyVisited x isMain $ addImportCycleCheck x $ do
       file <- findFile x  -- requires source to exist
 
       reportSLn "import.iface" 10 $ "  Check for cycle"
@@ -305,9 +332,10 @@ getInterface' x isMain = do
         -- let maySkip = isMain == NotMainInterface
         -- Andreas, 2015-07-13: Serialize iInsideScope again.
         let maySkip = True
+
         if uptodate && maySkip
-          then getStoredInterface x file includeStateChanges
-          else typeCheck          x file includeStateChanges
+          then getStoredInterface x file isMain
+          else typeCheck          x file isMain
 
       -- Ensure that the given module name matches the one in the file.
       let topLevelName = toTopLevelModuleName $ iModuleName i
@@ -327,16 +355,15 @@ getInterface' x isMain = do
 
       stCurrentModule .= Just (iModuleName i)
 
-      -- Interfaces are only stored if no warnings were encountered.
-      case wt of
-        SomeWarnings w -> return ()
-        NoWarnings     -> storeDecodedModule i
+      -- Interfaces are not stored if we are only scope-checking, or
+      -- if any warnings were encountered.
+      case (isMain, wt) of
+        (MainInterface ScopeCheck, _) -> return ()
+        (_, SomeWarnings w)           -> return ()
+        _                             -> storeDecodedModule i
 
       return (i, wt)
 
-    where
-      includeStateChanges = isMain == MainInterface
-
 -- | Check whether interface file exists and is in cache
 --   in the correct version (as testified by the interface file hash).
 
@@ -364,7 +391,6 @@ isCached x file = do
 
   return mi
 
-
 -- | Try to get the interface from interface file or cache.
 
 getStoredInterface
@@ -372,16 +398,13 @@ getStoredInterface
      -- ^ Module name of file we process.
   -> AbsolutePath
      -- ^ File we process.
-  -> Bool
-     -- ^ If type checking is necessary,
-     --   should all state changes inflicted by 'createInterface' be preserved?
-     --   @True@, if we are the 'MainInterface'.  @False@, if we are 'NotMainInterface'.
+  -> MainInterface
   -> TCM (Bool, (Interface, MaybeWarnings))
      -- ^ @Bool@ is: do we have to merge the interface?
-getStoredInterface x file includeStateChanges = do
+getStoredInterface x file isMain = do
   -- If something goes wrong (interface outdated etc.)
   -- we revert to fresh type checking.
-  let fallback = typeCheck x file includeStateChanges
+  let fallback = typeCheck x file isMain
 
   -- Examine the hash of the interface file. If it is different from the
   -- stored version (in stDecodedModules), or if there is no stored version,
@@ -421,7 +444,12 @@ getStoredInterface x file includeStateChanges = do
           -- liftIO close -- Close the interface file. See above.
           fallback
         else do
-          unless cached $ chaseMsg "Skipping" x $ Just ifile
+          unless cached $ do
+            chaseMsg "Loading " x $ Just ifile
+            -- print imported warnings
+            let ws = filter ((Strict.Just file ==) . tcWarningOrigin) (iWarnings i)
+            unless (null ws) $ reportSDoc "warning" 1 $ P.vcat $ P.prettyTCM <$> ws
+
           -- We set the pragma options of the skipped file here,
           -- because if the top-level file is skipped we want the
           -- pragmas to apply to interactive commands in the UI.
@@ -440,105 +468,107 @@ typeCheck
      -- ^ Module name of file we process.
   -> AbsolutePath
      -- ^ File we process.
-  -> Bool
-     -- ^ If type checking is necessary,
-     --   should all state changes inflicted by 'createInterface' be preserved?
-     --   @True@, if we are the 'MainInterface'.  @False@, if we are 'NotMainInterface'.
+  -> MainInterface
   -> TCM (Bool, (Interface, MaybeWarnings))
      -- ^ @Bool@ is: do we have to merge the interface?
-typeCheck x file includeStateChanges = do
-  unless includeStateChanges cleanCachedLog
-  let withMsgs = bracket_
-       (chaseMsg "Checking" x $ Just $ filePath file)
+typeCheck x file isMain = do
+  unless (includeStateChanges isMain) cleanCachedLog
+  let checkMsg = case isMain of
+                   MainInterface ScopeCheck -> "Reading "
+                   _                        -> "Checking"
+      withMsgs = bracket_
+       (chaseMsg checkMsg x $ Just $ filePath file)
        (const $ do ws <- getAllWarnings' AllWarnings RespectFlags
                    let (we, wa) = classifyWarnings ws
-                   unless (null wa) $ reportSDoc "warning" 1
-                                    $ P.vcat $ P.prettyTCM <$> wa
-                   unless (not $ null we) $ chaseMsg "Finished" x Nothing)
+                   let wa' = filter ((Strict.Just file ==) . tcWarningOrigin) wa
+                   unless (null wa') $
+                     reportSDoc "warning" 1 $ P.vcat $ P.prettyTCM <$> wa'
+                   when (null we) $ chaseMsg "Finished" x Nothing)
 
   -- Do the type checking.
 
-  if includeStateChanges then do
-     r <- withMsgs $ createInterface file x includeStateChanges
+  case isMain of
+    MainInterface _ -> do
+      r <- withMsgs $ createInterface file x isMain
 
-     -- Merge the signature with the signature for imported
-     -- things.
-     reportSLn "import.iface" 40 $ "Merging with state changes included."
-     sig     <- getSignature
-     patsyns <- getPatternSyns
-     display <- use stImportsDisplayForms
-     addImportedThings sig Map.empty Set.empty Set.empty patsyns display
-     setSignature emptySignature
-     setPatternSyns Map.empty
+      -- Merge the signature with the signature for imported
+      -- things.
+      reportSLn "import.iface" 40 $ "Merging with state changes included."
+      sig     <- getSignature
+      patsyns <- getPatternSyns
+      display <- use stImportsDisplayForms
+      addImportedThings sig Map.empty patsyns display []
+      setSignature emptySignature
+      setPatternSyns Map.empty
 
-     return (True, r)
-   else do
-    ms       <- getImportPath
-    nesting  <- asks envModuleNestingLevel
-    range    <- asks envRange
-    call     <- asks envCall
-    mf       <- use stModuleToSource
-    vs       <- getVisitedModules
-    ds       <- getDecodedModules
-    opts     <- stPersistentOptions . stPersistentState <$> get
-    isig     <- use stImports
-    ibuiltin <- use stImportedBuiltins
-    display  <- use stImportsDisplayForms
-    ipatsyns <- getPatternSynImports
-    ho       <- getInteractionOutputCallback
-    -- Every interface is treated in isolation. Note: Changes
-    -- to stDecodedModules are not preserved if an error is
-    -- encountered in an imported module.
-    -- Andreas, 2014-03-23: freshTCM spawns a new TCM computation
-    -- with initial state and environment
-    -- but on the same Benchmark accounts.
-    r <- freshTCM $
-           withImportPath ms $
-           local (\e -> e { envModuleNestingLevel = nesting
-                            -- Andreas, 2014-08-18:
-                            -- Preserve the range of import statement
-                            -- for reporting termination errors in
-                            -- imported modules:
-                          , envRange              = range
-                          , envCall               = call
-                          }) $ do
-             setDecodedModules ds
-             setCommandLineOptions opts
-             setInteractionOutputCallback ho
-             stModuleToSource .= mf
-             setVisitedModules vs
-             addImportedThings isig ibuiltin Set.empty Set.empty ipatsyns display
+      return (True, r)
+    NotMainInterface -> do
+      ms       <- getImportPath
+      nesting  <- asks envModuleNestingLevel
+      range    <- asks envRange
+      call     <- asks envCall
+      mf       <- use stModuleToSource
+      vs       <- getVisitedModules
+      ds       <- getDecodedModules
+      opts     <- stPersistentOptions . stPersistentState <$> get
+      isig     <- use stImports
+      ibuiltin <- use stImportedBuiltins
+      display  <- use stImportsDisplayForms
+      ipatsyns <- getPatternSynImports
+      ho       <- getInteractionOutputCallback
+      -- Every interface is treated in isolation. Note: Some changes to
+      -- the persistent state may not be preserved if an error other
+      -- than a type error or an IO exception is encountered in an
+      -- imported module.
+      r <- noCacheForImportedModule $
+           freshTCM $
+             withImportPath ms $
+             local (\e -> e { envModuleNestingLevel = nesting
+                              -- Andreas, 2014-08-18:
+                              -- Preserve the range of import statement
+                              -- for reporting termination errors in
+                              -- imported modules:
+                            , envRange              = range
+                            , envCall               = call
+                            }) $ do
+               setDecodedModules ds
+               setCommandLineOptions opts
+               setInteractionOutputCallback ho
+               stModuleToSource .= mf
+               setVisitedModules vs
+               addImportedThings isig ibuiltin ipatsyns display []
 
-             r  <- withMsgs $ createInterface file x includeStateChanges
-             mf <- use stModuleToSource
-             ds <- getDecodedModules
-             return (r, do
-                stModuleToSource .= mf
-                setDecodedModules ds
-                case r of
-                  (i, NoWarnings) -> storeDecodedModule i
-                  _               -> return ()
-                )
+               r  <- withMsgs $ createInterface file x isMain
+               mf <- use stModuleToSource
+               ds <- getDecodedModules
+               return (r, do
+                  stModuleToSource .= mf
+                  setDecodedModules ds
+                  case r of
+                    (i, NoWarnings) -> storeDecodedModule i
+                    _               -> return ()
+                  )
 
-    case r of
-        Left err          -> throwError err
-        Right (r, update) -> do
-          update
-          case r of
-            (_, NoWarnings) ->
-              -- We skip the file which has just been type-checked to
-              -- be able to forget some of the local state from
-              -- checking the module.
-              -- Note that this doesn't actually read the interface
-              -- file, only the cached interface.
-              getStoredInterface x file includeStateChanges
-            _ -> return (False, r)
+      case r of
+          Left err          -> throwError err
+          Right (r, update) -> do
+            update
+            case r of
+              (_, NoWarnings) ->
+                -- We skip the file which has just been type-checked to
+                -- be able to forget some of the local state from
+                -- checking the module.
+                -- Note that this doesn't actually read the interface
+                -- file, only the cached interface. (This comment is not
+                -- correct, see
+                -- test/Fail/customised/NestedProjectRoots.err.)
+                getStoredInterface x file isMain
+              _ -> return (False, r)
 
-
--- | Formats and outputs the "Checking", "Finished" and "Skipping" messages.
+-- | Formats and outputs the "Checking", "Finished" and "Loading " messages.
 
 chaseMsg
-  :: String               -- ^ The prefix, like @Checking@, @Finished@, @Skipping@.
+  :: String               -- ^ The prefix, like @Checking@, @Finished@, @Loading @.
   -> C.TopLevelModuleName -- ^ The module name.
   -> Maybe String         -- ^ Optionally: the file name.
   -> TCM ()
@@ -583,12 +613,12 @@ readInterface file = do
   `catchError` handler
   where
     handler e = case e of
-      IOException _ e -> do
+      IOException _ _ e -> do
         reportSLn "" 0 $ "IO exception: " ++ show e
         return Nothing   -- Work-around for file locking bug.
                          -- TODO: What does this refer to? Please
                          -- document.
-      _               -> throwError e
+      _ -> throwError e
 
 -- | Writes the given interface to the given file.
 
@@ -629,9 +659,9 @@ removePrivates si = si { scopeModules = restrictPrivate <$> scopeModules si }
 createInterface
   :: AbsolutePath          -- ^ The file to type check.
   -> C.TopLevelModuleName  -- ^ The expected module name.
-  -> Bool
+  -> MainInterface
   -> TCM (Interface, MaybeWarnings)
-createInterface file mname isMain =
+createInterface file mname isMain = Bench.billTo [Bench.TopModule mname] $
   local (\e -> e { envCurrentPath = Just file }) $ do
     modFile       <- use stModuleToSource
     fileTokenInfo <- Bench.billTo [Bench.Highlighting] $
@@ -643,10 +673,7 @@ createInterface file mname isMain =
     verboseS "import.iface.create" 10 $ do
       visited <- Map.keys <$> getVisitedModules
       reportSLn "import.iface.create" 10 $
-        "  visited: " ++ intercalate ", " (map prettyShow visited)
-
-    previousHsImports <- getHaskellImports
-    previousHsImportsUHC <- getHaskellImportsUHC
+        "  visited: " ++ List.intercalate ", " (map prettyShow visited)
 
     -- Parsing.
     (pragmas, top) <- Bench.billTo [Bench.Parsing] $
@@ -672,10 +699,12 @@ createInterface file mname isMain =
     -- Highlighting from scope checker.
     reportSLn "import.iface.create" 7 $ "Starting highlighting from scope."
     Bench.billTo [Bench.Highlighting] $ do
-      ifTopLevelAndHighlightingLevelIs NonInteractive $ do
-        -- Generate and print approximate syntax highlighting info.
+      -- Generate and print approximate syntax highlighting info.
+      ifTopLevelAndHighlightingLevelIs NonInteractive $
         printHighlightingInfo fileTokenInfo
-        mapM_ (\ d -> generateAndPrintSyntaxInfo d Partial) ds
+      let onlyScope = isMain == MainInterface ScopeCheck
+      ifTopLevelAndHighlightingLevelIsOr NonInteractive onlyScope $
+        mapM_ (\ d -> generateAndPrintSyntaxInfo d Partial onlyScope) ds
     reportSLn "import.iface.create" 7 $ "Finished highlighting from scope."
 
 
@@ -693,9 +722,14 @@ createInterface file mname isMain =
         cleanCachedLog
     writeToCurrentLog $ Pragmas opts
 
-    reportSLn "import.iface.create" 7 $ "Starting type checking."
-    Bench.billTo [Bench.Typing] $ mapM_ checkDeclCached ds `finally_` cacheCurrentLog
-    reportSLn "import.iface.create" 7 $ "Finished type checking."
+    case isMain of
+      MainInterface ScopeCheck -> do
+        reportSLn "import.iface.create" 7 $ "Skipping type checking."
+        cacheCurrentLog
+      _ -> do
+        reportSLn "import.iface.create" 7 $ "Starting type checking."
+        Bench.billTo [Bench.Typing] $ mapM_ checkDeclCached ds `finally_` cacheCurrentLog
+        reportSLn "import.iface.create" 7 $ "Finished type checking."
 
     -- Ulf, 2013-11-09: Since we're rethrowing the error, leave it up to the
     -- code that handles that error to reset the state.
@@ -746,7 +780,7 @@ createInterface file mname isMain =
     -- permanently freeze them now by turning them into postulates.
     -- This will enable serialization.
     -- savedMetaStore <- use stMetaStore
-    unless isMain $
+    unless (includeStateChanges isMain) $
       whenM (optAllowUnsolved <$> pragmaOptions) $ do
         withCurrentModule (scopeCurrent scope) $
           openMetasToPostulates
@@ -759,24 +793,30 @@ createInterface file mname isMain =
     reportSLn "import.iface.create" 7 $ "Starting serialization."
     syntaxInfo <- use stSyntaxInfo
     i <- Bench.billTo [Bench.Serialization, Bench.BuildInterface] $ do
-      buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC options
+      buildInterface file topLevel syntaxInfo options
 
-    reportSLn "tc.top" 101 $ concat $
-      "Signature:\n" :
-      [ show x ++ "\n  type: " ++ show (defType def)
-               ++ "\n  def:  " ++ show cc ++ "\n"
+    reportSLn "tc.top" 101 $ unlines $
+      "Signature:" :
+      [ unlines
+          [ prettyShow x
+          , "  type: " ++ show (defType def)
+          , "  def:  " ++ show cc
+          ]
       | (x, def) <- HMap.toList $ iSignature i ^. sigDefinitions,
         Function{ funCompiled = cc } <- [theDef def]
       ]
     reportSLn "import.iface.create" 7 $ "Finished serialization."
 
     mallWarnings <- getAllWarnings ErrorWarnings
-                      $ if isMain then IgnoreFlags else RespectFlags
+                      $ case isMain of
+                          MainInterface _  -> IgnoreFlags
+                          NotMainInterface -> RespectFlags
 
     reportSLn "import.iface.create" 7 $ "Considering writing to interface file."
-    case mallWarnings of
-      SomeWarnings allWarnings -> return ()
-      NoWarnings               -> Bench.billTo [Bench.Serialization] $ do
+    case (mallWarnings, isMain) of
+      (SomeWarnings allWarnings, _) -> return ()
+      (_, MainInterface ScopeCheck) -> return ()
+      _ -> Bench.billTo [Bench.Serialization] $ do
         -- The file was successfully type-checked (and no warnings were
         -- encountered), so the interface should be written out.
         let ifile = filePath $ toIFile file
@@ -804,24 +844,6 @@ createInterface file mname isMain =
 -- in by the user, or not (for instance when deciding if we are
 -- writing an interface file or not)
 
-data WhichWarnings = ErrorWarnings | AllWarnings
-  -- ^ order of constructors important for derived Ord instance
-  deriving (Eq, Ord)
-
-classifyWarning :: Warning -> WhichWarnings
-classifyWarning w = case w of
-  OldBuiltin{}               -> AllWarnings
-  EmptyRewritePragma         -> AllWarnings
-  TerminationIssue{}         -> ErrorWarnings
-  NotStrictlyPositive{}      -> ErrorWarnings
-  UnsolvedMetaVariables{}    -> ErrorWarnings
-  UnsolvedInteractionMetas{} -> ErrorWarnings
-  UnsolvedConstraints{}      -> ErrorWarnings
-  ParseWarning{}             -> ErrorWarnings
-
-classifyWarnings :: [TCWarning] -> ([TCWarning], [TCWarning])
-classifyWarnings = partition $ (< AllWarnings) . classifyWarning . tcWarning
-
 getAllWarnings' :: WhichWarnings -> IgnoreFlags -> TCM [TCWarning]
 getAllWarnings' ww ifs = do
   openMetas            <- getOpenMetas
@@ -850,9 +872,24 @@ getAllWarnings ww ifs = do
            then NoWarnings
            else SomeWarnings allWarnings
 
--- constructIScope :: ScopeInfo -> Map ModuleName Scope
+errorWarningsOfTCErr :: TCErr -> TCM [TCWarning]
+errorWarningsOfTCErr err = case err of
+  TypeError tcst cls -> case clValue cls of
+    NonFatalErrors{} -> return []
+    _ -> localState $ do
+      put tcst
+      ws <- getAllWarnings' ErrorWarnings RespectFlags
+      -- We filter out the unsolved(Metas/Constraints) to stay
+      -- true to the previous error messages.
+      return $ filter (not . isUnsolvedWarning . tcWarning) ws
+  _ -> return []
+
+-- | Reconstruct the 'iScope' (not serialized)
+--   from the 'iInsideScope' (serialized).
+
 constructIScope :: Interface -> Interface
-constructIScope i = i{ iScope = billToPure [ Deserialization ] $ publicModules $ iInsideScope i }
+constructIScope i = billToPure [ Deserialization ] $
+  i{ iScope = publicModules $ iInsideScope i }
 
 -- | Builds an interface for the current module, which should already
 -- have been successfully type checked.
@@ -863,14 +900,10 @@ buildInterface
      -- ^ 'TopLevelInfo' for the current module.
   -> HighlightingInfo
      -- ^ Syntax highlighting info for the module.
-  -> Set String
-     -- ^ MAlonzo: Haskell modules imported in imported modules (transitively).
-  -> Set String
-     -- ^ UHC backend: Haskell modules imported in imported modules (transitively).
   -> [OptionsPragma]
      -- ^ Options set in @OPTIONS@ pragmas.
   -> TCM Interface
-buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC pragmas = do
+buildInterface file topLevel syntaxInfo pragmas = do
     reportSLn "import.iface" 5 "Building interface..."
     let m = topLevelModuleName topLevel
     scope'  <- getScope
@@ -888,9 +921,7 @@ buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC p
     builtin <- use stLocalBuiltins
     ms      <- getImports
     mhs     <- mapM (\ m -> (m,) <$> moduleHash m) $ Set.toList ms
-    hsImps  <- getHaskellImports
-    uhcHsImps <- getHaskellImportsUHC
-    hsCode  <- use stHaskellCode
+    foreignCode <- use stForeignCode
     -- Ulf, 2016-04-12:
     -- Non-closed display forms are not applicable outside the module anyway,
     -- and should be dead-code eliminated (#1928).
@@ -903,6 +934,7 @@ buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC p
     patsyns <- killRange <$> getPatternSyns
     h       <- liftIO $ hashFile file
     let builtin' = Map.mapWithKey (\ x b -> (x,) . primFunName <$> b) builtin
+    warnings <- getAllWarnings' AllWarnings RespectFlags
     reportSLn "import.iface" 7 "  instantiating all meta variables"
     i <- instantiateFull $ Interface
       { iSourceHash      = h
@@ -913,12 +945,11 @@ buildInterface file topLevel syntaxInfo previousHsImports previousHsImportsUHC p
       , iSignature       = sig
       , iDisplayForms    = display
       , iBuiltin         = builtin'
-      , iHaskellImports  = hsImps `Set.difference` previousHsImports
-      , iHaskellImportsUHC = uhcHsImps `Set.difference` previousHsImportsUHC
-      , iHaskellCode     = hsCode
+      , iForeignCode     = foreignCode
       , iHighlighting    = syntaxInfo
       , iPragmaOptions   = pragmas
       , iPatternSyns     = patsyns
+      , iWarnings        = warnings
       }
     reportSLn "import.iface" 7 "  interface complete"
     return i
