@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                   #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
+{-# LANGUAGE CPP #-}
 
 {-# OPTIONS_GHC -fno-cse #-}
 
@@ -13,15 +12,19 @@ module Agda.Interaction.InteractionTop
 import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
+import Control.Concurrent
+import Control.Concurrent.Async
+import Control.Concurrent.STM.TChan
 import qualified Control.Exception as E
 import Control.Monad.Identity
 import Control.Monad.Reader
 import Control.Monad.State
+import Control.Monad.STM
 
 import qualified Data.Char as Char
 import Data.Foldable (Foldable)
 import Data.Function
-import Data.List as List hiding (null)
+import qualified Data.List as List
 import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
@@ -36,6 +39,7 @@ import Agda.TypeChecking.Monad as TM
   hiding (initState, setCommandLineOptions)
 import qualified Agda.TypeChecking.Monad as TM
 import qualified Agda.TypeChecking.Pretty as TCP
+import Agda.TypeChecking.Rules.Term (checkExpr, isType_)
 import Agda.TypeChecking.Errors
 
 import Agda.Syntax.Fixity
@@ -64,15 +68,15 @@ import qualified Agda.Interaction.BasicOps as B
 import Agda.Interaction.BasicOps hiding (whyInScope)
 import Agda.Interaction.Highlighting.Precise hiding (Postulate)
 import qualified Agda.Interaction.Imports as Imp
+import Agda.TypeChecking.Warnings
 import Agda.Interaction.Highlighting.Generate
 import qualified Agda.Interaction.Highlighting.LaTeX as LaTeX
 import qualified Agda.Interaction.Highlighting.Range as H
 
 import Agda.Compiler.Common (IsMain (..))
-import qualified Agda.Compiler.MAlonzo.Compiler as MAlonzo
-import qualified Agda.Compiler.JS.Compiler as JS
+import Agda.Compiler.Backend
 
-import qualified Agda.Auto.Auto as Auto
+import Agda.Auto.Auto as Auto
 
 import Agda.Utils.Except
   ( ExceptT
@@ -97,7 +101,8 @@ import Agda.Utils.Time
 #include "undefined.h"
 import Agda.Utils.Impossible
 
-------------------------------------------
+------------------------------------------------------------------------
+-- The CommandM monad
 
 -- | Auxiliary state of an interactive computation.
 
@@ -119,20 +124,30 @@ data CommandState = CommandState
     -- ^ We remember (the scope of) old interaction points to make it
     --   possible to parse and compute highlighting information for the
     --   expression that it got replaced by.
+  , commandQueue :: CommandQueue
+    -- ^ Command queue.
+    --
+    -- The commands in the queue are processed in the order in which
+    -- they are received. Abort commands do not have precedence over
+    -- other commands, they only abort the immediately preceding
+    -- command. (The Emacs mode is expected not to send a new command,
+    -- other than the abort command, before the previous command has
+    -- completed.)
   }
 
 type OldInteractionScopes = Map InteractionId ScopeInfo
 
 -- | Initial auxiliary interaction state
 
-initCommandState :: CommandState
-initCommandState = CommandState
-  { theInteractionPoints = []
-  , theCurrentFile       = Nothing
-  , optionsOnReload      = defaultOptions
-  , oldInteractionScopes = Map.empty
-  }
-
+initCommandState :: CommandQueue -> CommandState
+initCommandState commandQueue =
+  CommandState
+    { theInteractionPoints = []
+    , theCurrentFile       = Nothing
+    , optionsOnReload      = defaultOptions
+    , oldInteractionScopes = Map.empty
+    , commandQueue         = commandQueue
+    }
 
 -- | Monad for computing answers to interactive commands.
 --
@@ -229,7 +244,7 @@ handleCommand_ = handleCommand id (return ())
 
 handleCommand :: (forall a. CommandM a -> CommandM a) -> CommandM () -> CommandM () -> CommandM ()
 handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
-    tcSt <- lift get
+    oldState <- lift get
 
     -- -- Andreas, 2016-11-18 OLD CODE:
     -- -- onFail and handleErr are executed in "new" command state (not TCState).
@@ -246,7 +261,10 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
       handleErr e
       -- Andreas, 2016-11-18, issue #2174
       -- Reset TCState after error is handled, to get rid of metas created during failed command
-      lift $ put tcSt
+      lift $ do
+        newPersistentState <- use lensPersistentState
+        put oldState
+        lensPersistentState .= newPersistentState
 
   where
     -- Preserves state so we can do unsolved meta highlighting
@@ -259,12 +277,24 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
       put s'
       return x
 
-    -- | Handle nasty errors like stack space overflow (issue 637)
-    -- We assume that the input action handles other kind of errors.
+    -- | Handle every possible kind of error (#637), except for
+    -- ThreadKilled, which is used to abort Agda.
     handleNastyErrors :: CommandM () -> CommandM ()
-    handleNastyErrors m = commandMToIO $ \ toIO ->
-        toIO m `E.catch` \ (e :: E.SomeException) ->
-          toIO $ handleErr $ Exception noRange $ text $ show e
+    handleNastyErrors m = commandMToIO $ \ toIO -> do
+      let handle e =
+            Right <$>
+              (toIO $ handleErr $ Exception noRange $ text $ show e)
+
+          asyncHandler e@E.ThreadKilled = return (Left e)
+          asyncHandler e                = handle e
+
+          generalHandler (e :: E.SomeException) = handle e
+
+      r <- ((Right <$> toIO m) `E.catch` asyncHandler)
+             `E.catch` generalHandler
+      case r of
+        Right x -> return x
+        Left e  -> E.throwIO e
 
     -- | Displays an error and instructs Emacs to jump to the site of the
     -- error. Because this function may switch the focus to another file
@@ -275,15 +305,18 @@ handleCommand wrap onFail cmd = handleNastyErrors $ wrap $ do
         constr  <- lift $ computeUnsolvedConstraints
         err     <- lift $ errorHighlighting e
         modFile <- lift $ use stModuleToSource
+        method  <- lift $ view eHighlightingMethod
         let info = compress $ mconcat $
                      -- Errors take precedence over unsolved things.
                      err : if unsolvedNotOK then [meta, constr] else []
-        s <- lift $ prettyError e
+        s1 <- lift $ prettyError e
+        s2 <- lift $ prettyTCWarnings' =<< Imp.errorWarningsOfTCErr e
+        let s = List.intercalate "\n" $ filter (not . null) $ s1 : s2
         x <- lift $ optShowImplicit <$> use stPragmaOptions
-        unless (null s) $ mapM_ putResponse $
+        unless (null s1) $ mapM_ putResponse $
             [ Resp_DisplayInfo $ Info_Error s ] ++
             tellEmacsToJumpToError (getRange e) ++
-            [ Resp_HighlightingInfo info modFile ] ++
+            [ Resp_HighlightingInfo info method modFile ] ++
             [ Resp_Status $ Status { sChecked = False
                                    , sShowImplicitArguments = x
                                    } ]
@@ -320,6 +353,95 @@ runInteraction (IOTCM current highlighting highlightingMethod cmd) =
     onFail | independent cmd = modify $ \ s -> s { theCurrentFile = Nothing }
            | otherwise       = return ()
 
+------------------------------------------------------------------------
+-- Command queues
+
+-- | Commands.
+
+data Command
+  = Command IOTCM
+    -- ^ An 'IOTCM' command.
+  | Done
+    -- ^ Stop processing commands.
+  | Error String
+    -- ^ An error message for a command that could not be parsed.
+  deriving Show
+
+-- | Command queues.
+
+type CommandQueue = TChan Command
+
+-- | The next command.
+
+nextCommand :: CommandM Command
+nextCommand =
+  liftIO . atomically . readTChan =<< gets commandQueue
+
+-- | Runs the given computation, but if an abort command is
+-- encountered (and acted upon), then the computation is interrupted,
+-- the persistent state and all options are restored, and some
+-- commands are sent to the frontend.
+
+-- TODO: It might be nice if some of the changes to the persistent
+-- state inflicted by the interrupted computation were preserved.
+
+maybeAbort :: CommandM () -> CommandM ()
+maybeAbort c = do
+  commandState <- get
+  tcState      <- lift get
+  tcEnv        <- lift ask
+  result       <- liftIO $ race
+                    (runTCM tcEnv tcState $ runStateT c commandState)
+                    (waitForAbort $ commandQueue commandState)
+  case result of
+    Left (((), commandState), tcState) -> do
+      lift $ put tcState
+      put commandState
+    Right () -> do
+      lift $ put $ initState
+        { stPersistentState = stPersistentState tcState
+        , stPreScopeState   =
+            (stPreScopeState initState)
+              { stPrePragmaOptions =
+                  stPrePragmaOptions
+                    (stPreScopeState tcState)
+              }
+        }
+      put $ (initCommandState (commandQueue commandState))
+        { optionsOnReload = optionsOnReload commandState
+        }
+      putResponse Resp_DoneAborting
+      displayStatus
+  where
+
+  -- | Returns if the first element in the queue is an abort command.
+  -- The abort command is removed from the queue.
+
+  waitForAbort :: CommandQueue -> IO ()
+  waitForAbort q = atomically $ do
+    c <- peekTChan q
+    case c of
+      Command (IOTCM _ _ _ Cmd_abort) -> void $ readTChan q
+      _                               -> retry
+
+-- | Creates a command queue, and forks a thread that writes commands
+-- to the queue. The queue is returned.
+
+initialiseCommandQueue
+  :: IO Command
+     -- ^ Returns the next command.
+  -> IO CommandQueue
+initialiseCommandQueue next = do
+  q <- newTChanIO
+  let readCommands = do
+        c <- next
+        atomically $ writeTChan q c
+        case c of
+          Done -> return ()
+          _    -> readCommands
+  _ <- forkIO readCommands
+  return q
+
 ----------------------------------------------------------------------------
 -- | An interactive computation.
 
@@ -332,7 +454,7 @@ data Interaction' range
 
     -- | @cmd_compile b m argv@ compiles the module in file @m@ using
     -- the backend @b@, using @argv@ as the command-line options.
-  | Cmd_compile         Backend FilePath [String]
+  | Cmd_compile         CompilerBackend FilePath [String]
 
   | Cmd_constraints
 
@@ -410,7 +532,7 @@ data Interaction' range
     -- If the range is 'noRange', then the string comes from the
     -- minibuffer rather than the goal.
 
-  | Cmd_give            InteractionId range String
+  | Cmd_give            UseForce InteractionId range String
 
   | Cmd_refine          InteractionId range String
 
@@ -436,6 +558,11 @@ data Interaction' range
   | Cmd_goal_type_context_infer
                         B.Rewrite InteractionId range String
 
+  -- | Grabs the current goal's type and checks the expression in the hole
+  -- against it.
+  | Cmd_goal_type_context_check
+                        B.Rewrite InteractionId range String
+
     -- | Shows all the top-level names in the given module, along with
     -- their types. Uses the scope of the given goal.
   | Cmd_show_module_contents
@@ -450,8 +577,11 @@ data Interaction' range
   | Cmd_why_in_scope_toplevel String
     -- | Displays version of the running Agda
   | Cmd_show_version
-        deriving (Read, Functor, Foldable, Traversable)
-
+  | Cmd_abort
+    -- ^ Abort the current computation.
+    --
+    -- Does nothing if no computation is in progress.
+        deriving (Show, Read, Functor, Foldable, Traversable)
 
 type IOTCM = IOTCM' Range
 data IOTCM' range
@@ -464,7 +594,7 @@ data IOTCM' range
         HighlightingMethod
         (Interaction' range)
          -- -^ What to do
-            deriving (Read, Functor, Foldable, Traversable)
+            deriving (Show, Read, Functor, Foldable, Traversable)
 
 ---------------------------------------------------------
 -- Read instances
@@ -496,7 +626,7 @@ parseToReadsPrec p i s = case runIdentity . flip runStateT s . runExceptT $ pare
 -- | Demand an exact string.
 
 exact :: String -> Parse ()
-exact s = readsToParse (show s) $ fmap (\x -> ((),x)) . stripPrefix s . dropWhile (==' ')
+exact s = readsToParse (show s) $ fmap (\x -> ((),x)) . List.stripPrefix s . dropWhile (==' ')
 
 readParse :: Read a => Parse a
 readParse = readsToParse "read failed" $ listToMaybe . reads
@@ -556,23 +686,30 @@ independent _                               = False
 interpret :: Interaction -> CommandM ()
 
 interpret (Cmd_load m argv) =
-  cmd_load' m argv True $ \_ -> interpret Cmd_metas
+  cmd_load' m argv True Imp.TypeCheck $ \_ -> interpret Cmd_metas
 
 interpret (Cmd_compile b file argv) =
-  cmd_load' file argv (b == LaTeX) $ \(i, mw) -> do
+  cmd_load' file argv (b `elem` [LaTeX, QuickLaTeX])
+            (if b == QuickLaTeX
+             then Imp.ScopeCheck
+             else Imp.TypeCheck) $ \(i, mw) -> do
     mw <- lift $ Imp.applyFlagsToMaybeWarnings RespectFlags mw
     case mw of
       Imp.NoWarnings -> do
         lift $ case b of
-          GHC       -> MAlonzo.compilerMain IsMain i
-          GHCNoMain -> MAlonzo.compilerMain NotMain i
-          JS        -> JS.compilerMain i
-          LaTeX     -> LaTeX.generateLaTeX i
-        display_info $ Info_CompilationOk
-      Imp.SomeWarnings w ->
+          LaTeX                    -> LaTeX.generateLaTeX i
+          QuickLaTeX               -> LaTeX.generateLaTeX i
+          OtherBackend "GHCNoMain" -> callBackend "GHC" NotMain i   -- for backwards compatibility
+          OtherBackend b           -> callBackend b IsMain  i
+        (pwe, pwa) <- interpretWarnings
+        display_info $ Info_CompilationOk pwa pwe
+      Imp.SomeWarnings w -> do
+        pw <- lift $ prettyTCWarnings w
         display_info $ Info_Error $ unlines
-          [ "You can only compile modules without unsolved metavariables"
-          , "or termination checking problems."
+          [ "You need to fix the following errors before you can compile"
+          , "the module:"
+          , ""
+          , pw
           ]
 
 interpret Cmd_constraints =
@@ -654,7 +791,8 @@ interpret (Cmd_load_highlighting_info source) = do
               if sourceH == iSourceHash (miInterface mi)
                then do
                 modFile <- use stModuleToSource
-                return $ Just (iHighlighting $ miInterface mi, modFile)
+                method  <- view eHighlightingMethod
+                return $ Just (iHighlighting $ miInterface mi, method, modFile)
                else
                 return Nothing
     mapM_ putResponse resp
@@ -680,15 +818,15 @@ interpret (Cmd_highlight ii rng s) = do
     try err m = mkExceptT $ do
       (Right <$> m) `catchError` \ _ -> return (Left err)
 
-interpret (Cmd_give   ii rng s) = give_gen ii rng s Give
-interpret (Cmd_refine ii rng s) = give_gen ii rng s Refine
+interpret (Cmd_give   force ii rng s) = give_gen force ii rng s Give
+interpret (Cmd_refine ii rng s) = give_gen WithoutForce ii rng s Refine
 
 interpret (Cmd_intro pmLambda ii rng _) = do
   ss <- lift $ B.introTactic pmLambda ii
   liftCommandMT (B.withInteractionId ii) $ case ss of
     []    -> do
       display_info $ Info_Intro $ text "No introduction forms found."
-    [s]   -> give_gen ii rng s Intro
+    [s]   -> give_gen WithoutForce ii rng s Intro
     _:_:_ -> do
       display_info $ Info_Intro $
         sep [ text "Don't know which constructor to introduce of"
@@ -707,11 +845,11 @@ interpret (Cmd_auto ii rng s) = do
   -- Save the state to have access to even those interaction ids
   -- that Auto solves (since Auto gives the solution right away).
   st <- lift $ get
-  (res, msg) <- lift $ Auto.auto ii rng s
-  case res of
-   Left xs -> do
-    lift $ reportSLn "auto" 10 $ "Auto produced the following solutions " ++ show xs
-    forM_ xs $ \(ii, s) -> do
+  (time , res) <- maybeTimed $ lift $ Auto.auto ii rng s
+  case autoProgress res of
+   Solutions sols -> do
+    lift $ reportSLn "auto" 10 $ "Auto produced the following solutions " ++ show sols
+    forM_ sols $ \(ii, s) -> do
       -- Andreas, 2014-07-05 Issue 1226:
       -- For highlighting, Resp_GiveAction needs to access
       -- the @oldInteractionScope@s of the interaction points solved by Auto.
@@ -723,16 +861,17 @@ interpret (Cmd_auto ii rng s) = do
       -- modifyTheInteractionPoints $ filter (/= ii)
       putResponse $ Resp_GiveAction ii $ Give_String s
     -- Andreas, 2014-07-07: Remove the interaction points in one go.
-    modifyTheInteractionPoints (\\ (map fst xs))
-    case msg of
-     Nothing -> interpret Cmd_metas
+    modifyTheInteractionPoints (List.\\ (map fst sols))
+    case autoMessage res of
+     Nothing  -> interpret Cmd_metas
      Just msg -> display_info $ Info_Auto msg
-   Right (Left cs) -> do
-    case msg of
-     Nothing -> return ()
+   FunClauses cs -> do
+    case autoMessage res of
+     Nothing  -> return ()
      Just msg -> display_info $ Info_Auto msg
     putResponse $ Resp_MakeCase R.Function cs
-   Right (Right s) -> give_gen ii rng s Refine
+   Refinement s -> give_gen WithoutForce ii rng s Refine
+  maybe (return ()) (display_info . Info_Time) time
 
 interpret (Cmd_context norm ii _ _) =
   display_info . Info_Context =<< liftLocalState (prettyContext norm False ii)
@@ -762,6 +901,17 @@ interpret (Cmd_goal_type_context_infer norm ii rng s) = do
     return $ text "Have:" <+> typ
   cmd_goal_type_context_and have norm ii rng s
 
+interpret (Cmd_goal_type_context_check norm ii rng s) = do
+  have <- liftLocalState $ B.withInteractionId ii $ do
+    expr <- B.parseExprIn ii rng s
+    goal <- B.typeOfMeta AsIs ii
+    term <- case goal of
+      OfType _ ty -> checkExpr expr =<< isType_ ty
+      _           -> __IMPOSSIBLE__
+    txt <- TCP.prettyTCM =<< normalForm norm term
+    return $ text "Elaborates to:" <+> txt
+  cmd_goal_type_context_and have norm ii rng s
+
 interpret (Cmd_show_module_contents norm ii rng s) =
   liftCommandMT (B.withInteractionId ii) $ showModuleContents norm rng s
 
@@ -788,6 +938,13 @@ interpret (Cmd_make_case ii rng s) = do
         , TCP.text "pcs' = " TCP.<+> TCP.vcat (map TCP.text pcs')
         ]
       ]
+    lift $ reportSDoc "interaction.case" 90 $ TCP.vcat
+      [ TCP.text "InteractionTop.Cmd_make_case"
+      , TCP.nest 2 $ TCP.vcat
+        [ TCP.text "cs   = " TCP.<+> TCP.text (show cs)
+        , TCP.text "cs'  = " TCP.<+> TCP.text (show cs')
+        ]
+      ]
     putResponse $ Resp_MakeCase (makeCaseVariant casectxt) pcs'
   where
     render = renderStyle (style { mode = OneLineMode })
@@ -811,11 +968,10 @@ interpret (Cmd_make_case ii rng s) = do
     -- Drops pattern added to extended lambda functions when lambda lifting them
     extlam_dropLLifted :: CaseContext -> Bool -> A.Clause -> A.Clause
     extlam_dropLLifted Nothing _ x = x
-    extlam_dropLLifted _ _ (A.Clause (A.LHS _ A.LHSProj{} _) _ _ _ _) = __IMPOSSIBLE__
-    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden (A.Clause (A.LHS info (A.LHSHead name nps) ps) dots rhs decl catchall)
+    extlam_dropLLifted _ _ (A.Clause (A.LHS _ A.LHSProj{} _) _ _ _ _ _) = __IMPOSSIBLE__
+    extlam_dropLLifted (Just (ExtLamInfo h nh)) hidden cl@A.Clause{ A.clauseLHS = A.LHS info (A.LHSHead name nps) ps }
       = let n = if hidden then h + nh else nh
-        in
-         (A.Clause (A.LHS info (A.LHSHead name (drop n nps)) ps) dots rhs decl catchall)
+        in cl{ A.clauseLHS = A.LHS info (A.LHSHead name (drop n nps)) ps }
 
 interpret (Cmd_compute cmode ii rng s) = display_info . Info_NormalForm =<< do
   liftLocalState $ do
@@ -826,13 +982,15 @@ interpret (Cmd_compute cmode ii rng s) = display_info . Info_NormalForm =<< do
 
 interpret Cmd_show_version = display_info Info_Version
 
+interpret Cmd_abort = return ()
+
 -- | Show warnings
 interpretWarnings :: CommandM (String, String)
 interpretWarnings = do
-  mws <- lift $ Imp.getAllWarnings Imp.AllWarnings RespectFlags
+  mws <- lift $ Imp.getAllWarnings AllWarnings RespectFlags
   case filter isNotMeta <$> mws of
     Imp.SomeWarnings ws@(_:_) -> do
-      let (we, wa) = Imp.classifyWarnings ws
+      let (we, wa) = classifyWarnings ws
       pwe <- lift $ prettyTCWarnings we
       pwa <- lift $ prettyTCWarnings wa
       return (pwe, pwa)
@@ -895,10 +1053,12 @@ showOpenMetas = do
 -- result of 'Imp.typeCheckMain'.
 
 cmd_load' :: FilePath -> [String]
-          -> Bool -- ^ Allow unsolved meta-variables?
+          -> Bool      -- ^ Allow unsolved meta-variables?
+          -> Imp.Mode  -- ^ Full type-checking, or only
+                       --   scope-checking?
           -> ((Interface, Imp.MaybeWarnings) -> CommandM ())
           -> CommandM ()
-cmd_load' file argv unsolvedOK cmd = do
+cmd_load' file argv unsolvedOK mode cmd = do
     f <- liftIO $ absolute file
     ex <- liftIO $ doesFileExist $ filePath f
     let relativeTo | ex        = ProjectRoot f
@@ -937,7 +1097,7 @@ cmd_load' file argv unsolvedOK cmd = do
     -- We activate the cache only when agda is used interactively
     lift activateLoadedFileCache
 
-    ok <- lift $ Imp.typeCheckMain f
+    ok <- lift $ Imp.typeCheckMain f mode
 
     -- The module type checked. If the file was not changed while the
     -- type checker was running then the interaction points and the
@@ -959,30 +1119,42 @@ withCurrentFile m = do
 
 -- | Available backends.
 
-data Backend = GHC
-             | GHCNoMain
-             | JS
-             | LaTeX
-    deriving (Show, Read, Eq)
+data CompilerBackend = LaTeX | QuickLaTeX | OtherBackend String
+    deriving (Eq)
+
+instance Show CompilerBackend where
+  show LaTeX            = "LaTeX"
+  show QuickLaTeX       = "QuickLaTeX"
+  show (OtherBackend s) = s
+
+instance Read CompilerBackend where
+  readsPrec _ s = do
+    (t, s) <- lex s
+    let b = case t of
+              "LaTeX"      -> LaTeX
+              "QuickLaTeX" -> QuickLaTeX
+              _            -> OtherBackend t
+    return (b, s)
 
 data GiveRefine = Give | Refine | Intro
   deriving (Eq, Show)
 
 -- | A "give"-like action (give, refine, etc).
 --
---   @give_gen ii rng s give_ref mk_newtxt@
+--   @give_gen force ii rng s give_ref mk_newtxt@
 --   acts on interaction point @ii@
 --   occupying range @rng@,
 --   placing the new content given by string @s@,
 --   and replacing @ii@ by the newly created interaction points
---   in the state.
+--   in the state if safety checks pass (unless @force@ is applied).
 give_gen
-  :: InteractionId
+  :: UseForce       -- ^ Should safety checks be skipped?
+  -> InteractionId
   -> Range
   -> String
   -> GiveRefine
   -> CommandM ()
-give_gen ii rng s0 giveRefine = do
+give_gen force ii rng s0 giveRefine = do
   let s = trim s0
   lift $ reportSLn "interaction.give" 20 $ "give_gen  " ++ s
   -- Andreas, 2015-02-26 if string is empty do nothing rather
@@ -1001,10 +1173,10 @@ give_gen ii rng s0 giveRefine = do
         mis  <- getInteractionPoints
         reportSLn "interaction.give" 30 $ "interaction points before = " ++ show mis
         given <- B.parseExprIn ii rng s
-        ae    <- give_ref ii Nothing given
+        ae    <- give_ref force ii Nothing given
         mis' <- getInteractionPoints
         reportSLn "interaction.give" 30 $ "interaction points after = " ++ show mis'
-        return (ae, given, mis' \\ mis)
+        return (ae, given, mis' List.\\ mis)
     -- favonia: backup the old scope for highlighting
     insertOldInteractionScope ii scope
     -- sort the new interaction points and put them into the state
@@ -1053,7 +1225,7 @@ highlightExpr e =
   local (\e -> e { envModuleNestingLevel = 0
                  , envHighlightingLevel  = NonInteractive
                  , envHighlightingMethod = Direct }) $
-    generateAndPrintSyntaxInfo decl Full
+    generateAndPrintSyntaxInfo decl Full True
   where
     dummy = mkName_ (NameId 0 0) "dummy"
     info  = mkDefInfo (nameConcrete dummy) noFixity' PublicAccess ConcreteDef (getRange e)
@@ -1063,7 +1235,7 @@ highlightExpr e =
 
 sortInteractionPoints :: [InteractionId] -> TCM [InteractionId]
 sortInteractionPoints is =
-  map fst . sortBy (compare `on` snd) <$> do
+  map fst . List.sortBy (compare `on` snd) <$> do
     forM is $ \ i -> do
       (i,) <$> getInteractionRange i
 
@@ -1123,7 +1295,7 @@ showModuleContents norm rng s = display_info . Info_ModuleContents =<< do
     (modules, types) <- B.moduleContents norm rng s
     types' <- forM types $ \ (x, t) -> do
       t <- TCP.prettyTCM t
-      return (show x, text ":" <+> t)
+      return (prettyShow x, text ":" <+> t)
     return $ vcat
       [ text "Modules"
       , nest 2 $ vcat $ map (text . show) modules
@@ -1142,9 +1314,9 @@ searchAbout norm rg nm = do
        hits <- findMentions norm rg tnm
        forM hits $ \ (x, t) -> do
          t <- TCP.prettyTCM t
-         return (show x, text ":" <+> t)
+         return (prettyShow x, text ":" <+> t)
     display_info $ Info_SearchAbout $
-      text "Definitions about" <+> text (intercalate ", " $ words nm) $$
+      text "Definitions about" <+> text (List.intercalate ", " $ words nm) $$
       nest 2 (align 10 fancy)
 
 -- | Explain why something is in scope.
@@ -1181,8 +1353,8 @@ whyInScope s = display_info . Info_WhyInScope =<< do
             asVar = do
               TCP.text "* a variable bound at" TCP.<+> TCP.prettyTCM (nameBindingSite $ localVar x)
             shadowing :: LocalVar -> TCM Doc
-            shadowing LocalVar{}    = TCP.text "shadowing"
-            shadowing ShadowedVar{} = TCP.text "in conflict with"
+            shadowing (LocalVar _ _ [])    = TCP.text "shadowing"
+            shadowing _ = TCP.text "in conflict with"
         names   xs = TCP.vcat $ map pName xs
         modules ms = TCP.vcat $ map pMod ms
 
@@ -1256,7 +1428,7 @@ status = do
           mm <- lookupModuleFromSource f
           case mm of
             Nothing -> return False -- work-around for Issue1007
-            Just m  -> not . miWarnings . fromMaybe __IMPOSSIBLE__ <$> getVisitedModule m
+            Just m  -> maybe False (not . miWarnings) <$> getVisitedModule m
 
   return $ Status { sShowImplicitArguments = showImpl
                   , sChecked               = checked
@@ -1322,10 +1494,10 @@ maybeTimed work = do
 -- info (unless it is @Nothing@).
 
 tellToUpdateHighlighting
-  :: Maybe (HighlightingInfo, ModuleToSource) -> IO [Response]
+  :: Maybe (HighlightingInfo, HighlightingMethod, ModuleToSource) -> IO [Response]
 tellToUpdateHighlighting Nothing                = return []
-tellToUpdateHighlighting (Just (info, modFile)) =
-  return [Resp_HighlightingInfo info modFile]
+tellToUpdateHighlighting (Just (info, method, modFile)) =
+  return [Resp_HighlightingInfo info method modFile]
 
 -- | Tells the Emacs mode to go to the first error position (if any).
 
