@@ -36,6 +36,7 @@ import Agda.TheTypeChecker
 import Agda.Interaction.Options
 import Agda.Interaction.BasicOps
 
+import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
@@ -124,11 +125,14 @@ parseVariables f tel ii rng ss = do
 
         -- If s is a variable name in scope, get its de Bruijn index
         -- via the type checker.
-        VarName x -> do
+        VarName x _ -> do
           (v, _) <- getVarInfo x
           case ignoreSharing v of
             Var i [] -> done $ i - nlocals
-            _        -> failNotVar
+            _        -> typeError . GenericDocError =<< sep
+              [ text $ "Cannot split on variable " ++ s ++ ", because it is bound to"
+              , prettyTCM v
+              ]
 
         -- If s is not a name, compare it to the printed variable representation.
         -- This fallback is to enable splitting on hidden variables.
@@ -162,7 +166,7 @@ getClauseForIP f clauseNo = do
 
 -- | Entry point for case splitting tactic.
 
-makeCase :: InteractionId -> Range -> String -> TCM (QName, CaseContext , [A.Clause])
+makeCase :: InteractionId -> Range -> String -> TCM (QName, CaseContext, [A.Clause])
 makeCase hole rng s = withInteractionId hole $ do
 
   -- Get function clause which contains the interaction point.
@@ -181,8 +185,8 @@ makeCase hole rng s = withInteractionId hole $ do
     [ text "splitting clause:"
     , nest 2 $ vcat
       [ text "f       =" <+> prettyTCM f
-      , text "context =" <+> (prettyTCM =<< getContextTelescope)
-      , text "tel     =" <+> prettyTCM tel
+      , text "context =" <+> ((inTopContext . prettyTCM) =<< getContextTelescope)
+      , text "tel     =" <+> (inTopContext . prettyTCM) tel
       , text "perm    =" <+> text (show perm)
       , text "ps      =" <+> text (show ps)
       ]
@@ -192,9 +196,21 @@ makeCase hole rng s = withInteractionId hole $ do
 
   let vars = words s
 
+  -- If the user just entered ".", do nothing.
+  -- This will expand an ellipsis, if present.
+
+  if concat vars == "." then do
+    cl <- makeAbstractClause f rhs $ clauseToSplitClause clause
+    return (f, casectxt, [cl])
+
   -- If we have no split variables, split on result.
 
-  if null vars then do
+  else if null vars then do
+    -- Andreas, 2017-07-24, issue #2654:
+    -- When we introduce projection patterns in an extended lambda,
+    -- we need to print them postfix.
+    let postProjInExtLam = applyWhen (isJust casectxt) $
+          withPragmaOptions $ \ opt -> opt { optPostfixProjections = True }
     (piTel, sc) <- fixTarget $ clauseToSplitClause clause
     -- Andreas, 2015-05-05 If we introduced new function arguments
     -- do not split on result.  This might be more what the user wants.
@@ -207,7 +223,7 @@ makeCase hole rng s = withInteractionId hole $ do
       -- if any of them is shown by the printer
       imp <- optShowImplicit <$> pragmaOptions
       return $ imp || any visible (telToList piTel)
-    scs <- if newPats then return [sc] else do
+    scs <- if newPats then return [sc] else postProjInExtLam $ do
       res <- splitResult f sc
       case res of
         Nothing  -> typeError $ GenericError $ "Cannot split on result here"
@@ -247,7 +263,7 @@ makeCase hole rng s = withInteractionId hole $ do
   split :: QName -> [Nat] -> SplitClause -> TCM [(SplitClause, Bool)]
   split f [] clause = return [(clause,False)]
   split f (var : vars) clause = do
-    z <- splitClauseWithAbsurd clause var
+    z <- dontAssignMetas $ splitClauseWithAbsurd clause var
     case z of
       Left err          -> typeError $ SplitError err
       Right (Left cl)   -> return [(cl,True)]
@@ -265,7 +281,7 @@ makeCase hole rng s = withInteractionId hole $ do
   -- In this case, we refuse to split, as this might lose the refinements.
   checkClauseIsClean :: IPClause -> TCM ()
   checkClauseIsClean ipCl = do
-    sips <- Map.elems <$> use stSolvedInteractionPoints
+    sips <- filter ipSolved . Map.elems <$> use stInteractionPoints
     when (List.any ((== ipCl) . ipClause) sips) $
       typeError $ GenericError $ "Cannot split as clause rhs has been refined.  Please reload"
 
@@ -282,7 +298,7 @@ makePatternVarsVisible is sc@SClause{ scPats = ps } =
       -- We could introduce extra consistency checks, like
       -- if visible ai then __IMPOSSIBLE__ else
       -- or passing the parsed name along and comparing it with @x@
-      setOrigin UserWritten nx
+      setOrigin CaseSplit nx
     | otherwise = nx
 
 -- | Make clause with no rhs (because of absurd match).
@@ -292,9 +308,9 @@ makeAbsurdClause f (SClause tel ps _ _ t) = do
   reportSDoc "interaction.case" 10 $ vcat
     [ text "Interaction.MakeCase.makeAbsurdClause: split clause:"
     , nest 2 $ vcat
-      [ text "context =" <+> (prettyTCM =<< getContextTelescope)
-      , text "tel =" <+> prettyTCM tel
-      , text "ps =" <+> text (show ps)
+      [ text "context =" <+> do (inTopContext . prettyTCM) =<< getContextTelescope
+      , text "tel     =" <+> do inTopContext $ prettyTCM tel
+      , text "ps      =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList ps -- P.sep <$> prettyTCMPatterns ps
       ]
     ]
   withCurrentModule (qnameModule f) $ do
@@ -302,7 +318,7 @@ makeAbsurdClause f (SClause tel ps _ _ t) = do
     -- Contract implicit record patterns before printing.
     -- c <- translateRecordPatterns $ Clause noRange tel perm ps NoBody t False
     -- Jesper, 2015-09-19 Don't contract, since we do on-demand splitting
-    let c = Clause noRange tel ps Nothing t False
+    let c = Clause noRange noRange tel ps Nothing t False Nothing
     -- Normalise the dot patterns
     ps <- addContext tel $ normalise $ namedClausePats c
     reportSDoc "interaction.case" 60 $ text "normalized patterns: " <+> text (show ps)
@@ -314,9 +330,9 @@ makeAbsurdClause f (SClause tel ps _ _ t) = do
 makeAbstractClause :: QName -> A.RHS -> SplitClause -> TCM A.Clause
 makeAbstractClause f rhs cl = do
 
-  A.Clause lhs _ _ _ _ <- makeAbsurdClause f cl
+  lhs <- A.clauseLHS <$> makeAbsurdClause f cl
   reportSDoc "interaction.case" 60 $ text "reified lhs: " <+> text (show lhs)
-  return $ A.Clause lhs [] rhs [] False
+  return $ A.Clause lhs [] [] rhs [] False
   -- let ii = InteractionId (-1)  -- Dummy interaction point since we never type check this.
   --                              -- Can end up in verbose output though (#1842), hence not __IMPOSSIBLE__.
   -- let info = A.emptyMetaInfo   -- metaNumber = Nothing in order to print as ?, not ?n
