@@ -1,12 +1,17 @@
 {-# LANGUAGE CPP #-}
 
+#if __GLASGOW_HASKELL__ <= 708
+{-# LANGUAGE DeriveDataTypeable #-}
+#endif
+
 module Agda.Interaction.Options
     ( CommandLineOptions(..)
     , IgnoreFlags(..)
     , PragmaOptions(..)
     , OptionsPragma
-    , Flag, OptM, runOptM
+    , Flag, OptM, runOptM, OptDescr(..), ArgDescr(..)
     , Verbosity
+    , WarningMode(..)
     , checkOpts
     , parseStandardOptions, parseStandardOptions'
     , parsePragmaOptions
@@ -32,14 +37,14 @@ import Control.Applicative
 import Control.Monad            ( (>=>), when )
 import Control.Monad.Trans
 
--- base-4.7 defines the Functor instances for OptDescr and ArgDescr
-#if !(MIN_VERSION_base(4,7,0))
-import Data.Orphans             ()
-#endif
-
 import Data.Either
 import Data.Maybe
 import Data.List                ( isSuffixOf , intercalate )
+
+#if __GLASGOW_HASKELL__ <= 708
+import Data.Typeable            ( Typeable )
+#endif
+
 import System.Console.GetOpt    ( getOpt', usageInfo, ArgOrder(ReturnInOrder)
                                 , OptDescr(..), ArgDescr(..)
                                 )
@@ -82,6 +87,26 @@ type Verbosity = Trie String Int
 data IgnoreFlags = IgnoreFlags | RespectFlags
   deriving Eq
 
+-- Potentially turn harmless warnings into nothing, or errors
+-- (does not apply to non-fatal errors)
+data WarningMode = LeaveAlone | TurnIntoErrors | IgnoreAllWarnings
+  deriving (Show, Eq
+#if __GLASGOW_HASKELL__ <= 708
+           , Typeable
+#endif
+           )
+
+warningModes :: [ (String, WarningMode) ]
+warningModes =
+  [ (defaultWarningMode, LeaveAlone)
+  , ("ignore"          , IgnoreAllWarnings)
+  , ("error"           , TurnIntoErrors)
+  ]
+
+-- Don't forget to update
+--   doc/user-manual/tools/command-line-options.rst
+-- if you make changes to the command-line options!
+
 data CommandLineOptions = Options
   { optProgramName      :: String
   , optInputFile        :: Maybe FilePath
@@ -98,18 +123,6 @@ data CommandLineOptions = Options
   , optShowHelp         :: Bool
   , optInteractive      :: Bool
   , optGHCiInteraction  :: Bool
-  , optCompileNoMain    :: Bool
-  , optGhcCompile       :: Bool
-  , optGhcCallGhc       :: Bool
-  , optGhcFlags         :: [String]
-  , optEpicCompile      :: Bool
-  , optJSCompile        :: Bool
-  , optUHCCompile       :: Bool
-  , optUHCBin           :: Maybe FilePath
-  , optUHCTextualCore   :: Bool
-  , optUHCCallUHC       :: Bool
-  , optUHCTraceLevel    :: Int
-  , optUHCFlags         :: [String]
   , optOptimSmashing    :: Bool
   , optCompileDir       :: Maybe FilePath
   -- ^ In the absence of a path the project root is used.
@@ -118,15 +131,19 @@ data CommandLineOptions = Options
   , optGenerateHTML     :: Bool
   , optDependencyGraph  :: Maybe FilePath
   , optLaTeXDir         :: FilePath
+  , optCountClusters    :: Bool
+    -- ^ Count extended grapheme clusters rather than code points when
+    -- generating LaTeX.
   , optHTMLDir          :: FilePath
   , optCSSFile          :: Maybe FilePath
   , optIgnoreInterfaces :: Bool
   , optForcing          :: Bool
   , optPragmaOptions    :: PragmaOptions
-  , optEpicFlags        :: [String]
-  , optSafe             :: Bool
   , optSharing          :: Bool
   , optCaching          :: Bool
+  , optOnlyScopeChecking :: Bool
+    -- ^ Should the top-level module only be scope-checked, and not
+    --   type-checked?
   }
   deriving Show
 
@@ -160,8 +177,16 @@ data PragmaOptions = PragmaOptions
       -- ^ Should system generated projections 'ProjSystem' be printed
       --   postfix (True) or prefix (False).
   , optInstanceSearchDepth       :: Int
+  , optSafe                      :: Bool
+  , optWarningMode               :: WarningMode
+  , optCompileNoMain             :: Bool
   }
-  deriving (Show,Eq)
+  deriving ( Show
+           , Eq
+#if __GLASGOW_HASKELL__ <= 708
+           , Typeable
+#endif
+           )
 
 -- | The options from an @OPTIONS@ pragma.
 --
@@ -195,18 +220,6 @@ defaultOptions = Options
   , optShowHelp         = False
   , optInteractive      = False
   , optGHCiInteraction  = False
-  , optCompileNoMain    = False
-  , optGhcCompile       = False
-  , optGhcCallGhc       = True
-  , optGhcFlags         = []
-  , optEpicCompile      = False
-  , optJSCompile        = False
-  , optUHCCompile       = False
-  , optUHCBin           = Nothing
-  , optUHCTextualCore   = False
-  , optUHCCallUHC       = True
-  , optUHCTraceLevel    = 0
-  , optUHCFlags         = []
   , optOptimSmashing    = True
   , optCompileDir       = Nothing
   , optGenerateVimFile  = False
@@ -214,15 +227,15 @@ defaultOptions = Options
   , optGenerateHTML     = False
   , optDependencyGraph  = Nothing
   , optLaTeXDir         = defaultLaTeXDir
+  , optCountClusters    = False
   , optHTMLDir          = defaultHTMLDir
   , optCSSFile          = Nothing
   , optIgnoreInterfaces = False
   , optForcing          = True
   , optPragmaOptions    = defaultPragmaOptions
-  , optEpicFlags        = []
-  , optSafe             = False
   , optSharing          = False
   , optCaching          = False
+  , optOnlyScopeChecking = False
   }
 
 defaultPragmaOptions :: PragmaOptions
@@ -251,6 +264,9 @@ defaultPragmaOptions = PragmaOptions
   , optRewriting                 = False
   , optPostfixProjections        = False
   , optInstanceSearchDepth       = 500
+  , optSafe                      = False
+  , optWarningMode               = fromJust $ lookup defaultWarningMode warningModes
+  , optCompileNoMain             = False
   }
 
 -- | The default termination depth.
@@ -268,6 +284,12 @@ defaultLaTeXDir = "latex"
 defaultHTMLDir :: String
 defaultHTMLDir = "html"
 
+-- | The default warning mode.
+
+defaultWarningMode :: String
+defaultWarningMode = "warn"
+
+
 type OptM = ExceptT String IO
 
 runOptM :: OptM a -> IO (Either String a)
@@ -283,36 +305,42 @@ type Flag opts = opts -> OptM opts
 
 checkOpts :: Flag CommandLineOptions
 checkOpts opts
-  | not (atMostOne [optAllowUnsolved . p, \x -> optGhcCompile x]) = throwError
-      "Unsolved meta variables are not allowed when compiling.\n"
-  | optCompileNoMain opts && (not (optGhcCompile opts || optUHCCompile opts)) = throwError
-      "--no-main only allowed in combination with --compile.\n"
-  | not (atMostOne [optGHCiInteraction, isJust . optInputFile]) =
+  | not (matches [optGHCiInteraction, isJust . optInputFile] <= 1) =
       throwError "Choose at most one: input file or --interaction.\n"
-  | not (atMostOne $ interactive ++ [\x -> optGhcCompile x, optEpicCompile, optJSCompile]) =
-      throwError "Choose at most one: compilers/--interactive/--interaction.\n"
-  | not (atMostOne $ interactive ++ [optGenerateHTML]) =
-      throwError "Choose at most one: --html/--interactive/--interaction.\n"
-  | not (atMostOne $ interactive ++ [isJust . optDependencyGraph]) =
-      throwError "Choose at most one: --dependency-graph/--interactive/--interaction.\n"
-  | not (atMostOne $ interactive ++ [optGenerateLaTeX]) =
-      throwError "Choose at most one: --latex/--interactive/--interaction.\n"
-  | (not . null . optEpicFlags $ opts)
-      && not (optEpicCompile opts) =
-      throwError "Cannot set Epic flags without using the Epic backend.\n"
-  | (isJust $ optUHCBin opts)
-      && not (optUHCCompile opts) =
-      throwError "Cannot set uhc binary without using UHC backend.\n"
-  | (optUHCTextualCore opts)
-      && not (optUHCCompile opts) =
-      throwError "Cannot set --uhc-textual-core without using UHC backend.\n"
+  | or [ p opts && matches ps > 1 | (p, ps) <- exclusive ] =
+      throwError exclusiveMessage
   | otherwise = return opts
   where
-  atMostOne bs = length (filter ($ opts) bs) <= 1
+  matches = length . filter ($ opts)
 
-  p = optPragmaOptions
+  atMostOne =
+    [ optGenerateHTML
+    , isJust . optDependencyGraph
+    ] ++
+    map fst exclusive
 
-  interactive = [optInteractive, optGHCiInteraction]
+  exclusive =
+    [ ( optOnlyScopeChecking
+      , optSafe . optPragmaOptions :
+        optGenerateVimFile :
+        atMostOne
+      )
+    , ( optInteractive
+      , optGenerateLaTeX : atMostOne
+      )
+    , ( optGHCiInteraction
+      , optGenerateLaTeX : atMostOne
+      )
+    ]
+
+  exclusiveMessage = unlines $
+    [ "The options --interactive, --interaction and"
+    , "--only-scope-checking cannot be combined with each other or"
+    , "with --html or --dependency-graph. Furthermore"
+    , "--interactive and --interaction cannot be combined with"
+    , "--latex, and --only-scope-checking cannot be combined with"
+    , "--safe or --vim."
+    ]
 
 -- | Check for unsafe pramas. Gives a list of used unsafe flags.
 
@@ -341,7 +369,7 @@ versionFlag o = return $ o { optShowVersion = True }
 helpFlag :: Flag CommandLineOptions
 helpFlag o = return $ o { optShowHelp = True }
 
-safeFlag :: Flag CommandLineOptions
+safeFlag :: Flag PragmaOptions
 safeFlag o = return $ o { optSafe = True }
 
 sharingFlag :: Bool -> Flag CommandLineOptions
@@ -380,6 +408,18 @@ vimFlag o = return $ o { optGenerateVimFile = True }
 latexFlag :: Flag CommandLineOptions
 latexFlag o = return $ o { optGenerateLaTeX = True }
 
+onlyScopeCheckingFlag :: Flag CommandLineOptions
+onlyScopeCheckingFlag o = return $ o { optOnlyScopeChecking = True }
+
+countClustersFlag :: Flag CommandLineOptions
+countClustersFlag o =
+#ifdef COUNT_CLUSTERS
+  return $ o { optCountClusters = True }
+#else
+  throwError
+    "Cluster counting has not been enabled in this build of Agda."
+#endif
+
 latexDirFlag :: FilePath -> Flag CommandLineOptions
 latexDirFlag d o = return $ o { optLaTeXDir = d }
 
@@ -462,55 +502,12 @@ interactiveFlag  o = return $ o { optInteractive    = True
                                                       { optAllowUnsolved = True }
                                 }
 
-compileFlagNoMain :: Flag CommandLineOptions
+compileFlagNoMain :: Flag PragmaOptions
 compileFlagNoMain o = return $ o { optCompileNoMain = True }
 
-compileGhcFlag :: Flag CommandLineOptions
-compileGhcFlag o = return $ o { optGhcCompile = True }
-
-ghcDontCallGhcFlag :: Flag CommandLineOptions
-ghcDontCallGhcFlag o = return $ o { optGhcCallGhc = False }
-
--- NOTE: Quadratic in number of flags.
-ghcFlag :: String -> Flag CommandLineOptions
-ghcFlag f o = return $ o { optGhcFlags = optGhcFlags o ++ [f] }
-
--- The Epic backend has been removed. See Issue 1481.
-compileEpicFlag :: Flag CommandLineOptions
--- compileEpicFlag o = return $ o { optEpicCompile = True}
-compileEpicFlag o = throwError "the Epic backend has been disabled"
-
-compileJSFlag :: Flag CommandLineOptions
-compileJSFlag  o = return $ o { optJSCompile = True }
-
-compileUHCFlag :: Flag CommandLineOptions
-compileUHCFlag o = return $ o { optUHCCompile = True}
-
 compileDirFlag :: FilePath -> Flag CommandLineOptions
 compileDirFlag f o = return $ o { optCompileDir = Just f }
 
--- NOTE: Quadratic in number of flags.
--- The Epic backend has been removed. See Issue 1481.
-epicFlagsFlag :: String -> Flag CommandLineOptions
--- epicFlagsFlag s o = return $ o { optEpicFlags = optEpicFlags o ++ [s] }
-epicFlagsFlag s o = throwError "the Epic backend has been disabled"
-
-uhcBinFlag :: String -> Flag CommandLineOptions
-uhcBinFlag s o = return $ o { optUHCBin  = Just s }
-
-uhcTextualCoreFlag :: Flag CommandLineOptions
-uhcTextualCoreFlag o = return $ o { optUHCTextualCore = True }
-
-uhcDontCallUHCFlag :: Flag CommandLineOptions
-uhcDontCallUHCFlag o = return $ o { optUHCCallUHC = False }
-
-uhcTraceLevelFlag :: String -> Flag CommandLineOptions
--- TODO proper parsing and error handling
-uhcTraceLevelFlag i o = return $ o { optUHCTraceLevel = read i }
-
-uhcFlagsFlag :: String -> Flag CommandLineOptions
-uhcFlagsFlag s o = return $ o { optUHCFlags = optUHCFlags o ++ [s] }
-
 htmlFlag :: Flag CommandLineOptions
 htmlFlag o = return $ o { optGenerateHTML = True }
 
@@ -553,6 +550,15 @@ verboseFlag s o =
         return (init ss, n)
     usage = throwError "argument to verbose should be on the form x.y.z:N or N"
 
+warningModeFlag :: String -> Flag PragmaOptions
+warningModeFlag s o =
+    case lookup s warningModes of
+      Just m -> return $ o { optWarningMode = m }
+      Nothing -> usage
+  where
+    usage = throwError $ "unknown warning mode (available: " ++
+                           intercalate ", " (map fst warningModes) ++ ")"
+
 terminationDepthFlag :: String -> Flag PragmaOptions
 terminationDepthFlag s o =
     do k <- readM s `catchError` \_ -> usage
@@ -573,35 +579,10 @@ standardOptions =
                     "start in interactive mode"
     , Option []     ["interaction"] (NoArg ghciInteractionFlag)
                     "for use with the Emacs mode"
-    , Option ['c']  ["compile", "ghc"] (NoArg compileGhcFlag)
-                    "compile program using the GHC backend"
-    , Option []     ["ghc-dont-call-ghc"] (NoArg ghcDontCallGhcFlag) "Don't call ghc, just write the GHC Haskell files."
-    , Option []     ["ghc-flag"] (ReqArg ghcFlag "GHC-FLAG")
-                    "give the flag GHC-FLAG to GHC when compiling using the GHC backend"
-    , Option []     ["no-main"] (NoArg compileFlagNoMain)
-                    "when compiling using the GHC backend or the UHC backend (experimental), do not treat the requested module as the main module of a program"
 
-    -- The Epic backend has been removed. See Issue 1481.
-    , Option []     ["epic"] (NoArg compileEpicFlag)
-    --                "compile program using the Epic backend"
-                    "the Epic backend has been removed"
-
-    , Option []     ["js"] (NoArg compileJSFlag) "compile program using the JS backend"
-    , Option []     ["uhc"] (NoArg compileUHCFlag) "compile program using the UHC backend"
-    , Option []     ["uhc-bin"] (ReqArg uhcBinFlag "UHC") "The uhc binary to use when compiling with the UHC backend."
-    , Option []     ["uhc-textual-core"] (NoArg uhcTextualCoreFlag) "Use textual core as intermediate representation instead of binary core."
-    , Option []     ["uhc-dont-call-uhc"] (NoArg uhcDontCallUHCFlag) "Don't call uhc, just write the UHC Core files."
-    , Option []     ["uhc-gen-trace"] (ReqArg uhcTraceLevelFlag "TRACE") "Add tracing code to generated executable."
-    , Option []     ["uhc-flag"] (ReqArg uhcFlagsFlag "UHC-FLAG")
-                    "give the flag UHC-FLAG to UHC when compiling using the UHC backend"
     , Option []     ["compile-dir"] (ReqArg compileDirFlag "DIR")
                     ("directory for compiler output (default: the project root)")
 
-    -- The Epic backend has been removed. See Issue 1481.
-    , Option []     ["epic-flag"] (ReqArg epicFlagsFlag "EPIC-FLAG")
-    --                "give the flag EPIC-FLAG to Epic when compiling using Epic"
-                    "the Epic backend has been removed"
-
     , Option []     ["vim"] (NoArg vimFlag)
                     "generate Vim highlighting files"
     , Option []     ["latex"] (NoArg latexFlag)
@@ -609,15 +590,24 @@ standardOptions =
     , Option []     ["latex-dir"] (ReqArg latexDirFlag "DIR")
                     ("directory in which LaTeX files are placed (default: " ++
                      defaultLaTeXDir ++ ")")
+    , Option []     ["count-clusters"] (NoArg countClustersFlag)
+                    ("count extended grapheme clusters when " ++
+                     "generating LaTeX (note that this flag " ++
+#ifdef COUNT_CLUSTERS
+                     "is not enabled in all builds of Agda)"
+#else
+                     "has not been enabled in this build of Agda)"
+#endif
+                    )
     , Option []     ["html"] (NoArg htmlFlag)
                     "generate HTML files with highlighted source code"
-    , Option []     ["dependency-graph"] (ReqArg dependencyGraphFlag "FILE")
-                    "generate a Dot file with a module dependency graph"
     , Option []     ["html-dir"] (ReqArg htmlDirFlag "DIR")
                     ("directory in which HTML files are placed (default: " ++
                      defaultHTMLDir ++ ")")
     , Option []     ["css"] (ReqArg cssFlag "URL")
                     "the CSS file used by the HTML files (can be relative)"
+    , Option []     ["dependency-graph"] (ReqArg dependencyGraphFlag "FILE")
+                    "generate a Dot file with a module dependency graph"
     , Option []     ["ignore-interfaces"] (NoArg ignoreInterfacesFlag)
                     "ignore interface files (re-type check everything)"
     , Option ['i']  ["include-path"] (ReqArg includeFlag "DIR")
@@ -632,8 +622,6 @@ standardOptions =
                     "don't use default libraries"
     , Option []     ["no-forcing"] (NoArg noForcingFlag)
                     "disable the forcing optimisation"
-    , Option []     ["safe"] (NoArg safeFlag)
-                    "disable postulates, unsafe OPTION pragmas and primTrustMe"
     , Option []     ["sharing"] (NoArg $ sharingFlag True)
                     "enable sharing and call-by-need evaluation (experimental) (default: OFF)"
     , Option []     ["no-sharing"] (NoArg $ sharingFlag False)
@@ -642,6 +630,8 @@ standardOptions =
                     "enable caching of typechecking (experimental) (default: OFF)"
     , Option []     ["no-caching"] (NoArg $ cachingFlag False)
                     "disable caching of typechecking"
+    , Option []     ["only-scope-checking"] (NoArg onlyScopeCheckingFlag)
+                    "only scope-check the top-level module, do not type-check it"
     ] ++ map (fmap lift) pragmaOptions
   where
   lift :: Flag PragmaOptions -> Flag CommandLineOptions
@@ -660,7 +650,7 @@ pragmaOptions =
     -- , Option []          ["proof-irrelevance"] (NoArg proofIrrelevanceFlag)
     --              "enable proof irrelevance (experimental feature)"
     , Option []     ["allow-unsolved-metas"] (NoArg allowUnsolvedFlag)
-                    "allow unsolved meta variables (only needed in batch mode)"
+                    "succeed and create interface file regardless of unsolved meta variables"
     , Option []     ["no-positivity-check"] (NoArg noPositivityFlag)
                     "do not warn about not strictly positive data types"
     , Option []     ["no-termination-check"] (NoArg dontTerminationCheckFlag)
@@ -698,17 +688,25 @@ pragmaOptions =
     , Option []     ["no-pattern-matching"] (NoArg noPatternMatchingFlag)
                     "disable pattern matching completely"
     , Option []     ["exact-split"] (NoArg exactSplitFlag)
-                    "require all clauses in a definition by pattern matching to hold as definitional equalities (except those marked as CATCHALL)"
+                    "require all clauses in a definition to hold as definitional equalities (unless marked CATCHALL)"
     , Option []     ["no-exact-split"] (NoArg noExactSplitFlag)
-                    "do not require all clauses in a definition by pattern matching to hold as definitional equalities (ignore those marked as CATCHALL)"
+                    "do not require all clauses in a definition to hold as definitional equalities (default)"
     , Option []     ["no-eta-equality"] (NoArg noEtaFlag)
-                    "disable eta rules for records"
+                    "default records to no-eta-equality"
     , Option []     ["rewriting"] (NoArg rewritingFlag)
                     "enable declaration and use of REWRITE rules"
     , Option []     ["postfix-projections"] (NoArg postfixProjectionsFlag)
                     "make postfix projection notation the default"
     , Option []     ["instance-search-depth"] (ReqArg instanceDepthFlag "N")
                     "set instance search depth to N (default: 500)"
+    , Option []     ["safe"] (NoArg safeFlag)
+                    "disable postulates, unsafe OPTION pragmas and primTrustMe"
+    , Option ['W']  ["warning"] (ReqArg warningModeFlag "MODE")
+                    ("set warning mode to MODE (available: "
+                       ++ intercalate ", " (map fst warningModes)
+                       ++ ". Default: " ++  defaultWarningMode ++ ")")
+    , Option []     ["no-main"] (NoArg compileFlagNoMain)
+                    "do not treat the requested module as the main module of a program when compiling"
     ]
 
 -- | Used for printing usage info.
