@@ -33,7 +33,7 @@ import Agda.Utils.Impossible
 --   so the user can have timely response even during long computations.
 
 data Response
-    = Resp_HighlightingInfo HighlightingInfo ModuleToSource
+    = Resp_HighlightingInfo HighlightingInfo HighlightingMethod ModuleToSource
     | Resp_Status Status
     | Resp_JumpToError FilePath Int32
     | Resp_InteractionPoints [InteractionId]
@@ -46,6 +46,9 @@ data Response
       -- ^ The integer is the message's debug level.
     | Resp_ClearRunningInfo
     | Resp_ClearHighlighting
+    | Resp_DoneAborting
+      -- ^ A command sent when an abort command has completed
+      -- successfully.
 
 -- | There are two kinds of \"make case\" commands.
 
@@ -54,7 +57,8 @@ data MakeCaseVariant = Function | ExtendedLambda
 -- | Info to display at the end of an interactive command
 
 data DisplayInfo
-    = Info_CompilationOk
+    = Info_CompilationOk String String
+      -- ^ Strings are the warnings and the (non-fatal) errors
     | Info_Constraints String
     | Info_AllGoalsWarnings String String String
         -- ^ Strings are the goals, the warnings and the (non-fatal) errors
@@ -119,7 +123,7 @@ data GiveResult
 --      closure of the 'InteractionOutputCallback' function.
 --      (suitable for intra-process communication).
 
-type InteractionOutputCallback = Response -> TCM ()
+type InteractionOutputCallback = Response -> IO ()
 
 -- | The default 'InteractionOutputCallback' function prints certain
 -- things to stdout (other things generate internal errors).
@@ -139,3 +143,4 @@ defaultInteractionOutputCallback r = case r of
                                  hFlush stdout
   Resp_ClearRunningInfo {}  -> __IMPOSSIBLE__
   Resp_ClearHighlighting {} -> __IMPOSSIBLE__
+  Resp_DoneAborting {}      -> __IMPOSSIBLE__
