@@ -22,9 +22,11 @@ import Data.Maybe
 import Data.Sequence (Seq, (<|), (><))
 import qualified Data.Sequence as Seq
 import Data.Traversable
-import Data.Typeable (Typeable)
 import Data.Void
 
+import Data.Data (Data)
+import Data.Typeable (Typeable)
+
 import Agda.Syntax.Concrete.Name (NumHoles(..))
 import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA)
 import qualified Agda.Syntax.Concrete as C
@@ -71,9 +73,9 @@ data Expr
   | Dot ExprInfo Expr                  -- ^ @.e@, for postfix projection.
   | App  ExprInfo Expr (NamedArg Expr) -- ^ Ordinary (binary) application.
   | WithApp ExprInfo Expr [Expr]       -- ^ With application.
-  | Lam  ExprInfo LamBinding Expr      -- ^ @λ bs → e@.
-  | AbsurdLam ExprInfo Hiding          -- ^ @λ()@ or @λ{}@.
-  | ExtendedLam ExprInfo DefInfo QName [Clause]
+  | Lam  LamInfo LamBinding Expr      -- ^ @λ bs → e@.
+  | AbsurdLam LamInfo Hiding          -- ^ @λ()@ or @λ{}@.
+  | ExtendedLam LamInfo DefInfo QName [Clause]
   | Pi   ExprInfo Telescope Expr       -- ^ Dependent function space @Γ → A@.
   | Fun  ExprInfo (Arg Expr) Expr      -- ^ Non-dependent function space.
   | Set  ExprInfo Integer              -- ^ @Set@, @Set1@, @Set2@, ...
@@ -91,7 +93,7 @@ data Expr
   | Tactic ExprInfo Expr [NamedArg Expr] [NamedArg Expr]
                                        -- ^ @tactic e x1 .. xn | y1 | .. | yn@
   | DontCare Expr                      -- ^ For printing @DontCare@ from @Syntax.Internal@.
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 -- | Record field assignment @f = e@.
 type Assign  = FieldAssignment' Expr
@@ -104,7 +106,7 @@ data Axiom
   = FunSig    -- ^ A function signature.
   | NoFunSig  -- ^ Not a function signature, i.e., a postulate (in user input)
               --   or another (e.g. data/record) type signature (internally).
-  deriving (Typeable, Eq, Ord, Show)
+  deriving (Typeable, Data, Eq, Ord, Show)
 
 -- | Renaming (generic).
 type Ren a = [(a, a)]
@@ -112,7 +114,7 @@ type Ren a = [(a, a)]
 data ScopeCopyInfo = ScopeCopyInfo
   { renModules :: Ren ModuleName
   , renNames   :: Ren QName }
-  deriving (Eq, Show)
+  deriving (Eq, Show, Typeable, Data)
 
 initCopyInfo :: ScopeCopyInfo
 initCopyInfo = ScopeCopyInfo
@@ -146,19 +148,19 @@ data Declaration
     -- ^ only retained for highlighting purposes
   | FunDef     DefInfo QName Delayed [Clause] -- ^ sequence of function clauses
   | DataSig    DefInfo QName Telescope Expr -- ^ lone data signature
-      -- ^ the 'LamBinding's are 'DomainFree' and binds the parameters of the datatype.
   | DataDef    DefInfo QName [LamBinding] [Constructor]
-      -- ^ the 'LamBinding's are 'DomainFree' and binds the parameters of the datatype.
+      -- ^ the 'LamBinding's are 'DomainFree' and bind the parameters of the datatype.
   | RecSig     DefInfo QName Telescope Expr -- ^ lone record signature
   | RecDef     DefInfo QName (Maybe (Ranged Induction)) (Maybe Bool) (Maybe QName) [LamBinding] Expr [Declaration]
-      -- ^ The 'Expr' gives the constructor type telescope, @(x1 : A1)..(xn : An) -> Prop@,
+      -- ^ The 'LamBinding's are 'DomainFree' and bind the parameters of the datatype.
+      --   The 'Expr' gives the constructor type telescope, @(x1 : A1)..(xn : An) -> Prop@,
       --   and the optional name is the constructor's name.
   | PatternSynDef QName [Arg Name] (Pattern' Void)
       -- ^ Only for highlighting purposes
   | UnquoteDecl MutualInfo [DefInfo] [QName] Expr
   | UnquoteDef  [DefInfo] [QName] Expr
   | ScopedDecl ScopeInfo [Declaration]  -- ^ scope annotation
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 class GetDefInfo a where
   getDefInfo :: a -> Maybe DefInfo
@@ -184,7 +186,7 @@ data ModuleApplication
       -- ^ @tel. M args@:  applies @M@ to @args@ and abstracts @tel@.
     | RecordModuleIFS ModuleName
       -- ^ @M {{...}}@
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 data Pragma
   = OptionsPragma [String]
@@ -193,19 +195,22 @@ data Pragma
     -- ^ Builtins that do not come with a definition,
     --   but declare a name for an Agda concept.
   | RewritePragma QName
+  | CompilePragma String QName String
   | CompiledPragma QName String
   | CompiledExportPragma QName String
-  | CompiledDeclareDataPragma QName String
   | CompiledTypePragma QName String
   | CompiledDataPragma QName String [String]
-  | CompiledEpicPragma QName String
   | CompiledJSPragma QName String
   | CompiledUHCPragma QName String
   | CompiledDataUHCPragma QName String [String]
   | StaticPragma QName
+  | EtaPragma QName
+    -- ^ For coinductive records, use pragma instead of regular
+    --   @eta-equality@ definition (as it is might make Agda loop).
+  | InjectivePragma QName
   | InlinePragma QName
   | DisplayPragma QName [NamedArg Pattern] Expr
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 -- | Bindings that are valid in a @let@.
 data LetBinding
@@ -221,7 +226,7 @@ data LetBinding
   | LetDeclaredVariable Name
     -- ^ Only used for highlighting. Refers to the first occurrence of
     -- @x@ in @let x : A; x = e@.
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 -- | Only 'Axiom's.
 type TypeSignature  = Declaration
@@ -232,12 +237,12 @@ type Field          = TypeSignature
 data LamBinding
   = DomainFree ArgInfo Name   -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@
   | DomainFull TypedBindings  -- ^ . @(xs:e)@ or @{xs:e}@ or @(let Ds)@
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 -- | Typed bindings with hiding information.
 data TypedBindings = TypedBindings Range (Arg TypedBinding)
             -- ^ . @(xs : e)@ or @{xs : e}@
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 -- | A typed binding.  Appears in dependent function spaces, typed lambdas, and
 --   telescopes.  It might be tempting to simplify this to only bind a single
@@ -258,15 +263,19 @@ data TypedBinding
     -- ^ As in telescope @(x y z : A)@ or type @(x y z : A) -> B@.
   | TLet Range [LetBinding]
     -- ^ E.g. @(let x = e)@ or @(let open M)@.
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 type Telescope  = [TypedBindings]
 
 data NamedDotPattern = NamedDot Name I.Term I.Type
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
-instance Eq NamedDotPattern where
-  _ == _ = True -- These are not relevant for caching purposes
+data StrippedDotPattern = StrippedDot Expr I.Term I.Type
+  deriving (Typeable, Data, Show)
+
+-- These are not relevant for caching purposes
+instance Eq NamedDotPattern    where _ == _ = True
+instance Eq StrippedDotPattern where _ == _ = True
 
 -- | We could throw away @where@ clauses at this point and translate them to
 --   @let@. It's not obvious how to remember that the @let@ was really a
@@ -276,10 +285,14 @@ data Clause' lhs = Clause
   , clauseNamedDots  :: [NamedDotPattern]
       -- ^ Only in with-clauses where we inherit some already checked dot patterns from the parent.
       --   These live in the context of the parent clause left-hand side.
+  , clauseStrippedDots :: [StrippedDotPattern]
+      -- ^ In with-clauses where a dot pattern from the parent clause is
+      --   repeated in the with-clause. In this case it's not actually part of
+      --   the clause, but it still needs to be checked (Issue 142).
   , clauseRHS        :: RHS
   , clauseWhereDecls :: [Declaration]
   , clauseCatchall   :: Bool
-  } deriving (Typeable, Show, Functor, Foldable, Traversable, Eq)
+  } deriving (Typeable, Data, Show, Functor, Foldable, Traversable, Eq)
 
 type Clause = Clause' LHS
 type SpineClause = Clause' SpineLHS
@@ -305,7 +318,7 @@ data RHS
       -- ^ The where clauses are attached to the @RewriteRHS@ by
       ---  the scope checker (instead of to the clause).
     }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 instance Eq RHS where
   RHS e _          == RHS e' _            = e == e'
@@ -323,7 +336,7 @@ data SpineLHS = SpineLHS
   , spLhsPats     :: [NamedArg Pattern]  -- ^ Function parameters (patterns).
   , spLhsWithPats :: [Pattern]           -- ^ @with@ patterns (after @|@).
   }
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 
 instance Eq LHS where
@@ -336,7 +349,7 @@ data LHS = LHS
   , lhsCore     :: LHSCore               -- ^ Copatterns.
   , lhsWithPats :: [Pattern]             -- ^ @with@ patterns (after @|@).
   }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 -- | The lhs minus @with@-patterns in projection-application view.
 --   Parameterised over the type @e@ of dot patterns.
@@ -353,7 +366,7 @@ data LHSCore' e
              , lhsPatsRight  :: [NamedArg (Pattern' e)]
                -- ^ Further applied to patterns.
              }
-  deriving (Typeable, Show, Functor, Foldable, Traversable, Eq)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable, Eq)
 
 type LHSCore = LHSCore' Expr
 
@@ -444,11 +457,14 @@ data Pattern' e
     --   Or generated at type checking for implicit arguments.
   | AsP PatInfo Name (Pattern' e)
   | DotP PatInfo Origin e
+    -- ^ Dot pattern @.e@: the Origin keeps track whether this dot pattern was
+    --   written by the user or inserted by the system (e.g. while expanding
+    --   the ellipsis in a with clause).
   | AbsurdP PatInfo
   | LitP Literal
   | PatternSynP PatInfo QName [NamedArg (Pattern' e)]
   | RecP PatInfo [FieldAssignment' (Pattern' e)]
-  deriving (Typeable, Show, Functor, Foldable, Traversable, Eq)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable, Eq)
 
 type Pattern  = Pattern' Expr
 type Patterns = [NamedArg Pattern]
@@ -646,7 +662,7 @@ instance HasRange (LHSCore' e) where
     getRange (LHSProj d lhscore ps) = d `fuseRange` lhscore `fuseRange` ps
 
 instance HasRange a => HasRange (Clause' a) where
-    getRange (Clause lhs _ rhs ds catchall) = getRange (lhs,rhs,ds)
+    getRange (Clause lhs _ _ rhs ds catchall) = getRange (lhs, rhs, ds)
 
 instance HasRange RHS where
     getRange AbsurdRHS                = noRange
@@ -770,11 +786,14 @@ instance KillRange e => KillRange (LHSCore' e) where
   killRange (LHSProj a b c) = killRange3 LHSProj a b c
 
 instance KillRange a => KillRange (Clause' a) where
-  killRange (Clause lhs dots rhs ds catchall) = killRange5 Clause lhs dots rhs ds catchall
+  killRange (Clause lhs dots sdots rhs ds catchall) = killRange6 Clause lhs dots sdots rhs ds catchall
 
 instance KillRange NamedDotPattern where
   killRange (NamedDot a b c) = killRange3 NamedDot a b c
 
+instance KillRange StrippedDotPattern where
+  killRange (StrippedDot a b c) = killRange3 StrippedDot a b c
+
 instance KillRange RHS where
   killRange AbsurdRHS                = AbsurdRHS
   killRange (RHS e c)                = killRange2 RHS e c
@@ -856,7 +875,7 @@ instance AllNames Declaration where
   allNames (ScopedDecl _ decls)       = allNames decls
 
 instance AllNames Clause where
-  allNames (Clause _ _ rhs decls _) = allNames rhs >< allNames decls
+  allNames cl = allNames (clauseRHS cl, clauseWhereDecls cl)
 
 instance AllNames RHS where
   allNames (RHS e _)                 = allNames e
@@ -990,7 +1009,7 @@ type PatternSynDefns = Map QName PatternSynDefn
 
 lambdaLiftExpr :: [Name] -> Expr -> Expr
 lambdaLiftExpr []     e = e
-lambdaLiftExpr (n:ns) e = Lam (ExprRange noRange) (DomainFree defaultArgInfo n) $
+lambdaLiftExpr (n:ns) e = Lam defaultLamInfo_ (DomainFree defaultArgInfo n) $
                                      lambdaLiftExpr ns e
 
 substPattern :: [(Name, Pattern)] -> Pattern -> Pattern
@@ -1090,19 +1109,19 @@ insertImplicitPatSynArgs :: HasRange a => (Range -> a) -> Range -> [Arg Name] ->
 insertImplicitPatSynArgs wild r ns as = matchArgs r ns as
   where
     matchNextArg r n as@(~(a : as'))
-      | matchNext n as           = return (namedArg a, as')
-      | getHiding n == NotHidden = Nothing
-      | otherwise                = return (wild r, as)
+      | matchNext n as = return (namedArg a, as')
+      | visible n      = Nothing
+      | otherwise      = return (wild r, as)
 
     matchNext _ [] = False
-    matchNext n (a:as) = getHiding n == getHiding a && matchName
+    matchNext n (a:as) = sameHiding n a && matchName
       where
         x = unranged $ C.nameToRawName $ nameConcrete $ unArg n
         matchName = maybe True (== x) (nameOf $ unArg a)
 
     matchArgs r [] []     = return ([], [])
     matchArgs r [] as     = Nothing
-    matchArgs r (n:ns) [] | getHiding n == NotHidden = return ([], n : ns)    -- under-applied
+    matchArgs r (n:ns) [] | visible n = return ([], n : ns)    -- under-applied
     matchArgs r (n:ns) as = do
       (p, as) <- matchNextArg r n as
       first ((unArg n, p) :) <$> matchArgs (getRange p) ns as
