@@ -1,5 +1,4 @@
 {-# LANGUAGE CPP #-}
-{-# LANGUAGE ScopedTypeVariables #-}
 
 module Agda.Syntax.Abstract.Copatterns (translateCopatternClauses) where
 
@@ -180,10 +179,10 @@ groupClauses (pc@(Path p c) : pcs) = (c, Path p (rhs c) : grp) : groupClauses re
     rhsExpr _       = __IMPOSSIBLE__
 
 clauseToPath :: Clause -> ScopeM (ProjPath Clause)
-clauseToPath (Clause (LHS i lhs wps) dots (RHS e c) [] catchall) =
-  fmap (\ lhs -> Clause (LHS i lhs wps) dots (RHS e c) [] catchall) <$> lhsToPath [] lhs
-clauseToPath (Clause lhs _ (RHS e _) (_:_) _) = typeError $ NotImplemented $ "copattern clauses with where declarations"
-clauseToPath (Clause lhs _ _ wheredecls _) = typeError $ NotImplemented $ "copattern clauses with absurd, with or rewrite right hand side"
+clauseToPath (Clause (LHS i lhs wps) dots sdots (RHS e c) [] catchall) =
+  fmap (\ lhs -> Clause (LHS i lhs wps) dots sdots (RHS e c) [] catchall) <$> lhsToPath [] lhs
+clauseToPath (Clause lhs _ _ (RHS e _) (_:_) _) = typeError $ NotImplemented $ "copattern clauses with where declarations"
+clauseToPath (Clause lhs _ _ _ wheredecls _) = typeError $ NotImplemented $ "copattern clauses with absurd, with or rewrite right hand side"
 
 lhsToPath :: [ProjEntry] -> LHSCore -> ScopeM (ProjPath LHSCore)
 lhsToPath acc lhs@LHSHead{}         = return $ Path acc lhs
@@ -214,7 +213,7 @@ pathToRecord pps =
 
           abstract :: NamedArg Name -> ScopeM Expr -> ScopeM Expr
           abstract (Arg info (Named Nothing x)) me =
-            Lam (ExprRange noRange) (DomainFree info x) <$> me
+            Lam defaultLamInfo_ (DomainFree info x) <$> me
           abstract (Arg _ (Named Just{} _)) me = typeError $ NotImplemented $
             "named arguments in projection patterns"
 
@@ -322,12 +321,15 @@ instance Rename TypedBinding where
   rename rho (TLet r lbs)   = TLet r (rename rho lbs)
 
 instance Rename Clause where
-  rename rho (Clause lhs dots rhs wheredecls catchall) =
-    Clause (rename rho lhs) (rename rho dots) (rename rho rhs) (rename rho wheredecls) catchall
+  rename rho (Clause lhs dots sdots rhs wheredecls catchall) =
+    Clause (rename rho lhs) (rename rho dots) (rename rho sdots) (rename rho rhs) (rename rho wheredecls) catchall
 
 instance Rename NamedDotPattern where
   rename rho (NamedDot x v t) = NamedDot (rename rho x) v t
 
+instance Rename StrippedDotPattern where
+  rename rho (StrippedDot e v t) = StrippedDot (rename rho e) v t
+
 instance Rename RHS where
   rename rho e = case e of
       RHS e c               -> RHS (rename rho e) c
