@@ -53,6 +53,15 @@ unAppView :: AppView -> Expr
 unAppView (Application h es) =
   foldl (App (ExprRange noRange)) h es
 
+-- | Collects plain lambdas.
+data LamView = LamView [(LamInfo, LamBinding)] Expr
+
+lamView :: Expr -> LamView
+lamView (Lam i b e) = cons (i, b) $ lamView e
+  where cons b (LamView bs e) = LamView (b : bs) e
+lamView (ScopedExpr _ e) = lamView e
+lamView e = LamView [] e
+
 -- | Gather top-level 'AsP'atterns to expose underlying pattern.
 asView :: A.Pattern -> ([Name], A.Pattern)
 asView (A.AsP _ x p) = first (x :) $ asView p
@@ -311,7 +320,7 @@ instance ExprLike LetBinding where
 instance ExprLike a => ExprLike (Pattern' a) where
 
 instance ExprLike a => ExprLike (Clause' a) where
-  recurseExpr f (Clause lhs dots rhs ds ca) = Clause <$> rec lhs <*> pure dots <*> rec rhs <*> rec ds <*> pure ca
+  recurseExpr f (Clause lhs dots sdots rhs ds ca) = Clause <$> rec lhs <*> pure dots <*> pure sdots <*> rec rhs <*> rec ds <*> pure ca
     where rec = recurseExpr f
 
 instance ExprLike RHS where
@@ -337,17 +346,18 @@ instance ExprLike Pragma where
       OptionsPragma{}             -> pure p
       BuiltinNoDefPragma{}        -> pure p
       RewritePragma{}             -> pure p
+      CompilePragma{}             -> pure p
       CompiledPragma{}            -> pure p
       CompiledExportPragma{}      -> pure p
-      CompiledDeclareDataPragma{} -> pure p
       CompiledTypePragma{}        -> pure p
       CompiledDataPragma{}        -> pure p
-      CompiledEpicPragma{}        -> pure p
       CompiledJSPragma{}          -> pure p
       CompiledUHCPragma{}         -> pure p
       CompiledDataUHCPragma{}     -> pure p
       StaticPragma{}              -> pure p
+      InjectivePragma{}           -> pure p
       InlinePragma{}              -> pure p
+      EtaPragma{}                 -> pure p
       DisplayPragma f xs e        -> DisplayPragma f <$> rec xs <*> rec e
     where rec e = recurseExpr f e
 
