@@ -13,10 +13,11 @@ import Control.DeepSeq
 import Data.ByteString.Char8 (ByteString)
 import qualified Data.ByteString.Char8 as ByteString
 import Data.Foldable
-import Data.Hashable
+import Data.Hashable (Hashable(..))
 import qualified Data.Strict.Maybe as Strict
 import Data.Semigroup hiding (Arg)
 import Data.Traversable
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 import Data.Word
 
@@ -36,7 +37,7 @@ import Agda.Utils.Impossible
 
 -- | Used to specify whether something should be delayed.
 data Delayed = Delayed | NotDelayed
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 instance KillRange Delayed where
   killRange = id
@@ -46,7 +47,7 @@ instance KillRange Delayed where
 ---------------------------------------------------------------------------
 
 data Induction = Inductive | CoInductive
-  deriving (Typeable, Eq, Ord)
+  deriving (Typeable, Data, Eq, Ord)
 
 instance Show Induction where
   show Inductive   = "inductive"
@@ -66,17 +67,31 @@ instance NFData Induction where
 -- * Hiding
 ---------------------------------------------------------------------------
 
-data Hiding  = Hidden | Instance | NotHidden
-  deriving (Typeable, Show, Eq, Ord)
+data Overlappable = YesOverlap | NoOverlap
+  deriving (Typeable, Data, Show, Eq, Ord)
+
+data Hiding  = Hidden | Instance Overlappable | NotHidden
+  deriving (Typeable, Data, Show, Eq, Ord)
+
+-- | Just for the 'Hiding' instance. Should never combine different
+--   overlapping.
+instance Semigroup Overlappable where
+  NoOverlap  <> NoOverlap  = NoOverlap
+  YesOverlap <> YesOverlap = YesOverlap
+  _          <> _          = __IMPOSSIBLE__
 
 -- | 'Hiding' is an idempotent partial monoid, with unit 'NotHidden'.
 --   'Instance' and 'NotHidden' are incompatible.
 instance Semigroup Hiding where
-  NotHidden <> h         = h
-  h         <> NotHidden = h
-  Hidden    <> Hidden    = Hidden
-  Instance  <> Instance  = Instance
-  _         <> _         = __IMPOSSIBLE__
+  NotHidden  <> h           = h
+  h          <> NotHidden   = h
+  Hidden     <> Hidden      = Hidden
+  Instance o <> Instance o' = Instance (o <> o')
+  _          <> _           = __IMPOSSIBLE__
+
+instance Monoid Overlappable where
+  mempty  = NoOverlap
+  mappend = (<>)
 
 instance Monoid Hiding where
   mempty = NotHidden
@@ -85,14 +100,21 @@ instance Monoid Hiding where
 instance KillRange Hiding where
   killRange = id
 
+instance NFData Overlappable where
+  rnf NoOverlap  = ()
+  rnf YesOverlap = ()
+
 instance NFData Hiding where
-  rnf Hidden    = ()
-  rnf Instance  = ()
-  rnf NotHidden = ()
+  rnf Hidden       = ()
+  rnf (Instance o) = rnf o
+  rnf NotHidden    = ()
 
 -- | Decorating something with 'Hiding' information.
-data WithHiding a = WithHiding !Hiding a
-  deriving (Typeable, Eq, Ord, Show, Functor, Foldable, Traversable)
+data WithHiding a = WithHiding
+  { whHiding :: !Hiding
+  , whThing  :: a
+  }
+  deriving (Typeable, Data, Eq, Ord, Show, Functor, Foldable, Traversable)
 
 instance Decoration WithHiding where
   traverseF f (WithHiding h a) = WithHiding h <$> f a
@@ -139,10 +161,6 @@ instance LensHiding (WithHiding a) where
 mergeHiding :: LensHiding a => WithHiding a -> a
 mergeHiding (WithHiding h a) = mapHiding (mappend h) a
 
--- | Visible ('NotHidden') arguments are @notHidden@. (DEPRECATED, use 'visible'.)
-notHidden :: LensHiding a => a -> Bool
-notHidden a = getHiding a == NotHidden
-
 -- | 'NotHidden' arguments are @visible@.
 visible :: LensHiding a => a -> Bool
 visible a = getHiding a == NotHidden
@@ -151,18 +169,44 @@ visible a = getHiding a == NotHidden
 notVisible :: LensHiding a => a -> Bool
 notVisible a = getHiding a /= NotHidden
 
+-- | 'Hidden' arguments are @hidden@.
+hidden :: LensHiding a => a -> Bool
+hidden a = getHiding a == Hidden
+
 hide :: LensHiding a => a -> a
 hide = setHiding Hidden
 
 hideOrKeepInstance :: LensHiding a => a -> a
 hideOrKeepInstance x =
   case getHiding x of
-    Hidden -> x
-    Instance -> x
-    NotHidden -> setHiding Hidden x
+    Hidden     -> x
+    Instance{} -> x
+    NotHidden  -> setHiding Hidden x
 
 makeInstance :: LensHiding a => a -> a
-makeInstance = setHiding Instance
+makeInstance = makeInstance' NoOverlap
+
+makeInstance' :: LensHiding a => Overlappable -> a -> a
+makeInstance' o = setHiding (Instance o)
+
+isOverlappable :: LensHiding a => a -> Bool
+isOverlappable x =
+  case getHiding x of
+    Instance YesOverlap -> True
+    _ -> False
+
+isInstance :: LensHiding a => a -> Bool
+isInstance x =
+  case getHiding x of
+    Instance{} -> True
+    _          -> False
+
+-- | Ignores 'Overlappable'.
+sameHiding :: (LensHiding a, LensHiding b) => a -> b -> Bool
+sameHiding x y =
+  case (getHiding x, getHiding y) of
+    (Instance{}, Instance{}) -> True
+    (hx, hy)                 -> hx == hy
 
 ---------------------------------------------------------------------------
 -- * Relevance
@@ -178,7 +222,7 @@ makeInstance = setHiding Instance
 --   @A@ is big in constructor @nil@ as the sort @Set1@ of its type @Set@
 --   is bigger than the sort @Set@ of the data type @List@.
 data Big = Big | Small
-  deriving (Typeable, Show, Eq, Enum, Bounded)
+  deriving (Typeable, Data, Show, Eq, Enum, Bounded)
 
 instance Ord Big where
   Big <= Small = False
@@ -200,11 +244,7 @@ data Relevance
                 --   because its value is already determined by the type.
                 --   If a constructor argument is big, it has to be regarded
                 --   absent, otherwise we get into paradoxes.
-  | UnusedArg   -- ^ The polarity checker has determined that this argument
-                --   is unused in the definition.  It can be skipped during
-                --   equality checking but should be mined for solutions
-                --   of meta-variables with relevance 'UnusedArg'
-    deriving (Typeable, Show, Eq)
+    deriving (Typeable, Data, Show, Eq)
 
 allRelevances :: [Relevance]
 allRelevances =
@@ -213,7 +253,6 @@ allRelevances =
   , Irrelevant
   , Forced Small
   , Forced Big
-  , UnusedArg
   ]
 
 instance KillRange Relevance where
@@ -227,7 +266,6 @@ instance NFData Relevance where
   rnf NonStrict  = ()
   rnf Irrelevant = ()
   rnf (Forced a) = rnf a
-  rnf UnusedArg  = ()
 
 -- | A lens to access the 'Relevance' attribute in data structures.
 --   Minimal implementation: @getRelevance@ and one of @setRelevance@ or @mapRelevance@.
@@ -254,7 +292,6 @@ isIrrelevant a = getRelevance a == Irrelevant
 
 -- | Information ordering.
 -- @Relevant  \`moreRelevant\`
---  UnusedArg \`moreRelevant\`
 --  Forced    \`moreRelevant\`
 --  NonStrict \`moreRelevant\`
 --  Irrelevant@
@@ -268,26 +305,26 @@ moreRelevant r r' =
     (Relevant, _)   -> True
     (_, Relevant)   -> False
     -- second bottom
-    (UnusedArg, _)  -> True
-    (_, UnusedArg)  -> False
-    -- third bottom
     (Forced{}, _)   -> True
     (_, Forced{})   -> False
     -- remaining case
     (NonStrict,NonStrict) -> True
 
-irrelevantOrUnused :: Relevance -> Bool
-irrelevantOrUnused r =
+irrelevant :: Relevance -> Bool
+irrelevant r =
   case r of
     Irrelevant -> True
-    UnusedArg  -> True
     NonStrict  -> False
     Relevant   -> False
     Forced{}   -> False
 
 -- | @unusableRelevance rel == True@ iff we cannot use a variable of @rel@.
-unusableRelevance :: Relevance -> Bool
-unusableRelevance rel = NonStrict `moreRelevant` rel
+unusableRelevance :: LensRelevance a => a -> Bool
+unusableRelevance a = case getRelevance a of
+  Irrelevant -> True
+  NonStrict  -> True
+  Forced{}   -> False -- @Forced@ has no semantic relevance
+  Relevant   -> False
 
 -- | 'Relevance' composition.
 --   'Irrelevant' is dominant, 'Relevant' is neutral.
@@ -301,8 +338,6 @@ composeRelevance r r' =
     (Forced b, Forced b') -> Forced (max b b')  -- prefer Big over Small
     (Forced b, _)     -> Forced b
     (_, Forced b)     -> Forced b
-    (UnusedArg, _)  -> UnusedArg
-    (_, UnusedArg)  -> UnusedArg
     (Relevant, Relevant) -> Relevant
 
 -- | @inverseComposeRelevance r x@ returns the most irrelevant @y@
@@ -316,17 +351,14 @@ inverseComposeRelevance r x =
     (Relevant, x)        -> x          -- going to relevant arg.: nothing changes
     _ | r == x           -> Relevant   -- because Relevant is comp.-neutral
     (Forced{}, Forced{}) -> Relevant   -- same, but (==) does not ignore Big
-    (UnusedArg, x)       -> x
-    (Forced{}, UnusedArg)  -> Relevant
     (Forced{}, x)          -> x
     (Irrelevant, x)      -> Relevant   -- going irrelevant: every thing usable
     (_, Irrelevant)      -> Irrelevant -- otherwise: irrelevant things remain unusable
     (NonStrict, _)       -> Relevant   -- but @NonStrict@s become usable
 
--- | For comparing @Relevance@ ignoring @Forced@ and @UnusedArg@.
+-- | For comparing @Relevance@ ignoring @Forced@.
 ignoreForced :: Relevance -> Relevance
 ignoreForced Forced{}   = Relevant
-ignoreForced UnusedArg  = Relevant
 ignoreForced Relevant   = Relevant
 ignoreForced NonStrict  = NonStrict
 ignoreForced Irrelevant = Irrelevant
@@ -350,8 +382,13 @@ nonStrictToIrr rel       = rel
 -- * Origin of arguments (user-written, inserted or reflected)
 ---------------------------------------------------------------------------
 
-data Origin = UserWritten | Inserted | Reflected
-  deriving (Typeable, Show, Eq, Ord)
+-- | Origin of arguments.
+data Origin
+  = UserWritten  -- ^ From the source file / user input.  (Preserve!)
+  | Inserted     -- ^ E.g. inserted hidden arguments.
+  | Reflected    -- ^ Produced by the reflection machinery.
+  | CaseSplit    -- ^ Produced by an interactive case split.
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 instance KillRange Origin where
   killRange = id
@@ -360,6 +397,32 @@ instance NFData Origin where
   rnf UserWritten = ()
   rnf Inserted = ()
   rnf Reflected = ()
+  rnf CaseSplit = ()
+
+-- | Decorating something with 'Origin' information.
+data WithOrigin a = WithOrigin
+  { woOrigin :: !Origin
+  , woThing  :: a
+  }
+  deriving (Typeable, Data, Eq, Ord, Show, Functor, Foldable, Traversable)
+
+instance Decoration WithOrigin where
+  traverseF f (WithOrigin h a) = WithOrigin h <$> f a
+
+instance HasRange a => HasRange (WithOrigin a) where
+  getRange = getRange . dget
+
+instance SetRange a => SetRange (WithOrigin a) where
+  setRange = fmap . setRange
+
+instance KillRange a => KillRange (WithOrigin a) where
+  killRange = fmap killRange
+
+instance NFData a => NFData (WithOrigin a) where
+  rnf (WithOrigin _ a) = rnf a
+
+-- | A lens to access the 'Origin' attribute in data structures.
+--   Minimal implementation: @getOrigin@ and one of @setOrigin@ or @mapOrigin@.
 
 class LensOrigin a where
 
@@ -376,6 +439,11 @@ instance LensOrigin Origin where
   setOrigin = const
   mapOrigin = id
 
+instance LensOrigin (WithOrigin a) where
+  getOrigin   (WithOrigin h _) = h
+  setOrigin h (WithOrigin _ a) = WithOrigin h a
+  mapOrigin f (WithOrigin h a) = WithOrigin (f h) a
+
 ---------------------------------------------------------------------------
 -- * Argument decoration
 ---------------------------------------------------------------------------
@@ -386,11 +454,10 @@ data ArgInfo = ArgInfo
   { argInfoHiding       :: Hiding
   , argInfoRelevance    :: Relevance
   , argInfoOrigin       :: Origin
-  , argInfoOverlappable :: Bool
-  } deriving (Typeable, Eq, Ord, Show)
+  } deriving (Typeable, Data, Eq, Ord, Show)
 
 instance KillRange ArgInfo where
-  killRange (ArgInfo h r o v) = killRange3 ArgInfo h r o v
+  killRange (ArgInfo h r o) = killRange3 ArgInfo h r o
 
 class LensArgInfo a where
   getArgInfo :: a -> ArgInfo
@@ -405,7 +472,7 @@ instance LensArgInfo ArgInfo where
   mapArgInfo = id
 
 instance NFData ArgInfo where
-  rnf (ArgInfo a b c d) = rnf a `seq` rnf b `seq` rnf c `seq` rnf d
+  rnf (ArgInfo a b c) = rnf a `seq` rnf b `seq` rnf c
 
 instance LensHiding ArgInfo where
   getHiding = argInfoHiding
@@ -425,8 +492,7 @@ instance LensOrigin ArgInfo where
 defaultArgInfo :: ArgInfo
 defaultArgInfo =  ArgInfo { argInfoHiding       = NotHidden
                           , argInfoRelevance    = Relevant
-                          , argInfoOrigin       = UserWritten
-                          , argInfoOverlappable = False }
+                          , argInfoOrigin       = UserWritten }
 
 
 ---------------------------------------------------------------------------
@@ -436,7 +502,7 @@ defaultArgInfo =  ArgInfo { argInfoHiding       = NotHidden
 data Arg e  = Arg
   { argInfo :: ArgInfo
   , unArg :: e
-  } deriving (Typeable, Ord, Functor, Foldable, Traversable)
+  } deriving (Data, Typeable, Ord, Functor, Foldable, Traversable)
 
 instance Decoration Arg where
   traverseF f (Arg ai a) = Arg ai <$> f a
@@ -451,25 +517,27 @@ instance KillRange a => KillRange (Arg a) where
   killRange (Arg info a) = killRange2 Arg info a
 
 instance Eq a => Eq (Arg a) where
-  Arg (ArgInfo h1 _ _ _) x1 == Arg (ArgInfo h2 _ _ _) x2 = (h1, x1) == (h2, x2)
+  Arg (ArgInfo h1 _ _) x1 == Arg (ArgInfo h2 _ _) x2 = (h1, x1) == (h2, x2)
 
 instance Show a => Show (Arg a) where
-    show (Arg (ArgInfo h r o v) x) = showR r $ showO o $ showH h $ show x
+    show (Arg (ArgInfo h r o) a) = showR r $ showO o $ showH h $ show a
       where
-        showH Hidden     s = "{" ++ s ++ "}"
-        showH NotHidden  s = "(" ++ s ++ ")"
-        showH Instance   s = (if v then "overlap " else "") ++ "{{" ++ s ++ "}}"
+        showH Hidden       s = "{" ++ s ++ "}"
+        showH NotHidden    s = "(" ++ s ++ ")"
+        showH (Instance o) s = showOv o ++ "{{" ++ s ++ "}}"
+          where showOv YesOverlap = "overlap "
+                showOv NoOverlap  = ""
         showR r s = case r of
           Irrelevant   -> "." ++ s
           NonStrict    -> "?" ++ s
           Forced Big   -> "!b" ++ s
           Forced Small -> "!" ++ s
-          UnusedArg    -> "k" ++ s -- constant
           Relevant     -> "r" ++ s -- Andreas: I want to see it explicitly
         showO o s = case o of
           UserWritten -> "u" ++ s
           Inserted    -> "i" ++ s
           Reflected   -> "g" ++ s -- generated by reflection
+          CaseSplit   -> "c" ++ s -- generated by case split
 
 instance NFData e => NFData (Arg e) where
   rnf (Arg a b) = rnf a `seq` rnf b
@@ -539,7 +607,7 @@ instance Underscore Doc where
 data Dom e = Dom
   { domInfo   :: ArgInfo
   , unDom     :: e
-  } deriving (Typeable, Ord, Functor, Foldable, Traversable)
+  } deriving (Typeable, Data, Ord, Functor, Foldable, Traversable)
 
 instance Decoration Dom where
   traverseF f (Dom ai a) = Dom ai <$> f a
@@ -551,7 +619,7 @@ instance KillRange a => KillRange (Dom a) where
   killRange (Dom info a) = killRange2 Dom info a
 
 instance Eq a => Eq (Dom a) where
-  Dom (ArgInfo h1 r1 _ _) x1 == Dom (ArgInfo h2 r2 _ _) x2 =
+  Dom (ArgInfo h1 r1 _) x1 == Dom (ArgInfo h2 r2 _) x2 =
     (h1, ignoreForced r1, x1) == (h2, ignoreForced r2, x2)
 
 instance Show a => Show (Dom a) where
@@ -569,6 +637,10 @@ instance LensArgInfo (Dom e) where
   getArgInfo = domInfo
   mapArgInfo f arg = arg { domInfo = f $ domInfo arg }
 
+instance LensOrigin (Dom e) where
+  getOrigin = getOrigin . getArgInfo
+  mapOrigin = mapArgInfo . mapOrigin
+
 argFromDom :: Dom a -> Arg a
 argFromDom (Dom i a) = Arg i a
 
@@ -587,7 +659,7 @@ data Named name a =
     Named { nameOf     :: Maybe name
           , namedThing :: a
           }
-    deriving (Eq, Ord, Typeable, Functor, Foldable, Traversable)
+    deriving (Eq, Ord, Typeable, Data, Functor, Foldable, Traversable)
 
 -- | Standard naming.
 type Named_ = Named RString
@@ -611,8 +683,8 @@ instance (KillRange name, KillRange a) => KillRange (Named name a) where
   killRange (Named n a) = Named (killRange n) (killRange a)
 
 instance Show a => Show (Named_ a) where
-    show (Named Nothing x)  = show x
-    show (Named (Just n) x) = rawNameToString (rangedThing n) ++ " = " ++ show x
+    show (Named Nothing a)  = show a
+    show (Named (Just n) a) = rawNameToString (rangedThing n) ++ " = " ++ show a
 
 instance (NFData name, NFData a) => NFData (Named name a) where
   rnf (Named a b) = rnf a `seq` rnf b
@@ -632,6 +704,10 @@ defaultNamedArg = defaultArg . unnamed
 updateNamedArg :: (a -> b) -> NamedArg a -> NamedArg b
 updateNamedArg = fmap . fmap
 
+-- | @setNamedArg a b = updateNamedArg (const b) a@
+setNamedArg :: NamedArg a -> b -> NamedArg b
+setNamedArg a b = (b <$) <$> a
+
 ---------------------------------------------------------------------------
 -- * Range decoration.
 ---------------------------------------------------------------------------
@@ -641,7 +717,7 @@ data Ranged a = Ranged
   { rangeOf     :: Range
   , rangedThing :: a
   }
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 -- | Thing with no range info.
 unranged :: a -> Ranged a
@@ -695,7 +771,8 @@ data ConOrigin
   = ConOSystem  -- ^ Inserted by system or expanded from an implicit pattern.
   | ConOCon     -- ^ User wrote a constructor (pattern).
   | ConORec     -- ^ User wrote a record (pattern).
-  deriving (Typeable, Show, Eq, Ord, Enum, Bounded)
+  | ConOSplit   -- ^ Generated by interactive case splitting.
+  deriving (Typeable, Data, Show, Eq, Ord, Enum, Bounded)
 
 instance KillRange ConOrigin where
   killRange = id
@@ -710,11 +787,14 @@ data ProjOrigin
   = ProjPrefix    -- ^ User wrote a prefix projection.
   | ProjPostfix   -- ^ User wrote a postfix projection.
   | ProjSystem    -- ^ Projection was generated by the system.
-  deriving (Typeable, Show, Eq, Ord, Enum, Bounded)
+  deriving (Typeable, Data, Show, Eq, Ord, Enum, Bounded)
 
 instance KillRange ProjOrigin where
   killRange = id
 
+data DataOrRecord = IsData | IsRecord
+  deriving (Typeable, Data, Eq, Ord, Show)
+
 ---------------------------------------------------------------------------
 -- * Infixity, access, abstract, etc.
 ---------------------------------------------------------------------------
@@ -722,7 +802,7 @@ instance KillRange ProjOrigin where
 -- | Functions can be defined in both infix and prefix style. See
 --   'Agda.Syntax.Concrete.LHS'.
 data IsInfix = InfixDef | PrefixDef
-    deriving (Typeable, Show, Eq, Ord)
+    deriving (Typeable, Data, Show, Eq, Ord)
 
 -- | Access modifier.
 data Access
@@ -732,7 +812,13 @@ data Access
   | PublicAccess
   | OnlyQualified  -- ^ Visible from outside, but not exported when opening the module
                              --   Used for qualified constructors.
-    deriving (Typeable, Show, Eq, Ord)
+    deriving (Typeable, Data, Show, Eq, Ord)
+
+instance Pretty Access where
+  pretty = text . \case
+    PrivateAccess _ -> "private"
+    PublicAccess    -> "public"
+    OnlyQualified   -> "only-qualified"
 
 instance NFData Access where
   rnf _ = ()
@@ -745,14 +831,14 @@ instance KillRange Access where
 
 -- | Abstract or concrete
 data IsAbstract = AbstractDef | ConcreteDef
-    deriving (Typeable, Show, Eq, Ord)
+    deriving (Typeable, Data, Show, Eq, Ord)
 
 instance KillRange IsAbstract where
   killRange = id
 
 -- | Is this definition eligible for instance search?
 data IsInstance = InstanceDef | NotInstanceDef
-    deriving (Typeable, Show, Eq, Ord)
+    deriving (Typeable, Data, Show, Eq, Ord)
 
 instance KillRange IsInstance where
   killRange = id
@@ -766,7 +852,7 @@ instance NFData IsInstance where
 
 -- | Is this a macro definition?
 data IsMacro = MacroDef | NotMacroDef
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 instance KillRange IsMacro where killRange = id
 instance HasRange  IsMacro where getRange _ = noRange
@@ -781,13 +867,13 @@ type Arity  = Nat
 -- | The unique identifier of a name. Second argument is the top-level module
 --   identifier.
 data NameId = NameId {-# UNPACK #-} !Word64 {-# UNPACK #-} !Word64
-    deriving (Eq, Ord, Typeable, Generic)
+    deriving (Eq, Ord, Typeable, Data, Generic)
 
 instance KillRange NameId where
   killRange = id
 
 instance Show NameId where
-  show (NameId x i) = show x ++ "@" ++ show i
+  show (NameId n m) = show n ++ "@" ++ show m
 
 instance Enum NameId where
   succ (NameId n m)     = NameId (n + 1) m
@@ -809,7 +895,7 @@ instance Hashable NameId where
 -- | A meta variable identifier is just a natural number.
 --
 newtype MetaId = MetaId { metaId :: Nat }
-    deriving (Eq, Ord, Num, Real, Enum, Integral, Typeable)
+    deriving (Eq, Ord, Num, Real, Enum, Integral, Typeable, Data)
 
 instance Pretty MetaId where
   pretty (MetaId n) = text $ "_" ++ show n
@@ -839,7 +925,7 @@ data PositionInName
     -- @foo_bar@.
   | End
     -- ^ The following underscore is at the end of the name: @foo_@.
-  deriving (Show, Eq, Ord)
+  deriving (Show, Eq, Ord, Typeable, Data)
 
 -- | Placeholders are used to represent the underscores in a section.
 
@@ -848,7 +934,7 @@ data MaybePlaceholder e
   | NoPlaceholder !(Strict.Maybe PositionInName) e
     -- ^ The second argument is used only (but not always) for name
     -- parts other than underscores.
-  deriving (Typeable, Eq, Ord, Functor, Foldable, Traversable, Show)
+  deriving (Typeable, Data, Eq, Ord, Functor, Foldable, Traversable, Show)
 
 -- | An abbreviation: @noPlaceholder = 'NoPlaceholder'
 -- 'Strict.Nothing'@.
@@ -873,10 +959,20 @@ instance NFData a => NFData (MaybePlaceholder a) where
 ---------------------------------------------------------------------------
 
 newtype InteractionId = InteractionId { interactionId :: Nat }
-    deriving (Eq,Ord,Num,Integral,Real,Enum)
+    deriving ( Eq
+             , Ord
+             , Num
+             , Integral
+             , Real
+             , Enum
+             , Data
+#if __GLASGOW_HASKELL__ <= 708
+             , Typeable
+#endif
+             )
 
 instance Show InteractionId where
-    show (InteractionId x) = "?" ++ show x
+    show (InteractionId i) = "?" ++ show i
 
 instance KillRange InteractionId where killRange = id
 
@@ -893,18 +989,18 @@ data ImportDirective' a b = ImportDirective
   , impRenaming    :: [Renaming' a b]
   , publicOpen     :: Bool -- ^ Only for @open@. Exports the opened names from the current module.
   }
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq)
 
 data Using' a b = UseEverything | Using [ImportedName' a b]
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq)
 
 instance Semigroup (Using' a b) where
-  UseEverything <> u = u
-  u <> UseEverything = u
-  Using xs <> Using ys = Using (xs ++ ys)
+  UseEverything <> u             = u
+  u             <> UseEverything = u
+  Using xs      <> Using ys      = Using (xs ++ ys)
 
 instance Monoid (Using' a b) where
-  mempty = UseEverything
+  mempty  = UseEverything
   mappend = (<>)
 
 -- | Default is directive is @private@ (use everything, but do not export).
@@ -919,15 +1015,15 @@ isDefaultImportDir _                                             = False
 data ImportedName' a b
   = ImportedModule  b
   | ImportedName    a
-  deriving (Typeable, Eq, Ord)
+  deriving (Typeable, Data, Eq, Ord)
 
 setImportedName :: ImportedName' a a -> a -> ImportedName' a a
 setImportedName (ImportedName   x) y = ImportedName   y
 setImportedName (ImportedModule x) y = ImportedModule y
 
 instance (Show a, Show b) => Show (ImportedName' a b) where
-  show (ImportedModule x) = "module " ++ show x
-  show (ImportedName   x) = show x
+  show (ImportedModule b) = "module " ++ show b
+  show (ImportedName   a) = show a
 
 data Renaming' a b = Renaming
   { renFrom    :: ImportedName' a b
@@ -937,7 +1033,7 @@ data Renaming' a b = Renaming
   , renToRange :: Range
     -- ^ The range of the \"to\" keyword.  Retained for highlighting purposes.
   }
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq)
 
 -- ** HasRange instances
 
@@ -1008,7 +1104,7 @@ data TerminationCheck m
     -- ^ Treat as terminating (unsafe).  Same effect as 'NoTerminationCheck'.
   | TerminationMeasure Range m
     -- ^ Skip termination checking but use measure instead.
-    deriving (Typeable, Show, Eq, Functor)
+    deriving (Typeable, Data, Show, Eq, Functor)
 
 instance KillRange m => KillRange (TerminationCheck m) where
   killRange (TerminationMeasure _ m) = TerminationMeasure noRange (killRange m)
