@@ -52,13 +52,15 @@ module Agda.Syntax.Concrete
 import Prelude hiding (null)
 
 import Control.DeepSeq
-import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
 import Data.List hiding (null)
 import Data.Set (Set)
 import Data.Monoid
 
+import Data.Typeable (Typeable)
+import Data.Data (Data)
+
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Fixity
@@ -81,14 +83,14 @@ data OpApp e
     -- ^ An abstraction inside a special syntax declaration
     --   (see Issue 358 why we introduce this).
   | Ordinary e
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 fromOrdinary :: e -> OpApp e -> e
 fromOrdinary d (Ordinary e) = e
 fromOrdinary d _            = d
 
 data FieldAssignment' a = FieldAssignment { _nameFieldA :: Name, _exprFieldA :: a }
-  deriving (Typeable, Functor, Foldable, Traversable, Show, Eq)
+  deriving (Typeable, Data, Functor, Foldable, Traversable, Show, Eq)
 
 type FieldAssignment = FieldAssignment' Expr
 
@@ -97,7 +99,7 @@ data ModuleAssignment  = ModuleAssignment
                            , _exprModA      :: [Expr]
                            , _importDirModA :: ImportDirective
                            }
-  deriving (Typeable)
+  deriving (Typeable, Data)
 type RecordAssignment  = Either FieldAssignment ModuleAssignment
 type RecordAssignments = [RecordAssignment]
 
@@ -161,7 +163,7 @@ data Expr
   | Unquote Range                              -- ^ ex: @unquote@, should be applied to a term of type @Term@
   | DontCare Expr                              -- ^ to print irrelevant things
   | Equal Range Expr Expr                      -- ^ ex: @a = b@, used internally in the parser
-  deriving (Typeable)
+  deriving (Typeable, Data)
 
 -- | Concrete patterns. No literals in patterns at the moment.
 data Pattern
@@ -181,17 +183,20 @@ data Pattern
   | WildP Range                            -- ^ @_@
   | AbsurdP Range                          -- ^ @()@
   | AsP Range Name Pattern                 -- ^ @x\@p@ unused
-  | DotP Range Origin Expr                 -- ^ @.e@
+  | DotP Range Origin Expr                 -- ^ @.e@ (the Origin keeps track
+                                           -- whether this dot pattern was
+                                           -- written by the user or inserted
+                                           -- by the system)
   | LitP Literal                           -- ^ @0@, @1@, etc.
   | RecP Range [FieldAssignment' Pattern]  -- ^ @record {x = p; y = q}@
-  deriving (Typeable)
+  deriving (Typeable, Data)
 
 -- | A lambda binding is either domain free or typed.
 type LamBinding = LamBinding' TypedBindings
 data LamBinding' a
   = DomainFree ArgInfo BoundName  -- ^ . @x@ or @{x}@ or @.x@ or @.{x}@ or @{.x}@
   | DomainFull a                  -- ^ . @(xs : e)@ or @{xs : e}@
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 
 -- | A sequence of typed bindings with hiding information. Appears in dependent
@@ -204,14 +209,14 @@ type TypedBindings = TypedBindings' TypedBinding
 
 data TypedBindings' a = TypedBindings Range (Arg a)
      -- ^ . @(xs : e)@ or @{xs : e}@ or something like @(x {y} _ : e)@.
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 data BoundName = BName
   { boundName   :: Name
   , boundLabel  :: Name    -- ^ for implicit function types the label matters and can't be alpha-renamed
   , bnameFixity :: Fixity'
   }
-  deriving (Typeable)
+  deriving (Typeable, Data, Eq, Show)
 
 mkBoundName_ :: Name -> BoundName
 mkBoundName_ x = mkBoundName x noFixity'
@@ -226,7 +231,7 @@ type TypedBinding = TypedBinding' Expr
 data TypedBinding' e
   = TBind Range [WithHiding BoundName] e  -- ^ Binding @(x1 ... xn : A)@.
   | TLet  Range [Declaration]  -- ^ Let binding @(let Ds)@ or @(open M args)@.
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 -- | A telescope is a sequence of typed bindings. Bound variables are in scope
 --   in later types.
@@ -255,7 +260,7 @@ data LHS
     -- ^ original pattern, with-patterns, rewrite equations and with-expressions
   | Ellipsis Range [Pattern] [RewriteEqn] [WithExpr]
     -- ^ new with-patterns, rewrite equations and with-expressions
-  deriving (Typeable)
+  deriving (Typeable, Data)
 
 type RewriteEqn = Expr
 type WithExpr   = Expr
@@ -277,7 +282,7 @@ type RHS = RHS' Expr
 data RHS' e
   = AbsurdRHS -- ^ No right hand side because of absurd match.
   | RHS e
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 
 type WhereClause = WhereClause' [Declaration]
@@ -288,7 +293,7 @@ data WhereClause' decls
     -- ^ Named where: @module M where@.
     --   The 'Access' flag applies to the 'Name' (not the module contents!)
     --   and is propagated from the parent function.
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 -- | An expression followed by a where clause.
 --   Currently only used to give better a better error message in interaction.
@@ -309,7 +314,7 @@ data AsName = AsName
   , asRange :: Range
     -- ^ The range of the \"as\" keyword.  Retained for highlighting purposes.
   }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 {--------------------------------------------------------------------------
     Declarations
@@ -355,17 +360,17 @@ data Declaration
   | UnquoteDecl Range [Name] Expr
   | UnquoteDef  Range [Name] Expr
   | Pragma      Pragma
-  deriving (Typeable)
+  deriving (Typeable, Data)
 
 data ModuleApplication
   = SectionApp Range [TypedBindings] Expr
     -- ^ @tel. M args@
   | RecordModuleIFS Range QName
     -- ^ @M {{...}}@
-  deriving (Typeable)
+  deriving (Typeable, Data)
 
 data OpenShortHand = DoOpen | DontOpen
-  deriving (Typeable, Eq, Show)
+  deriving (Typeable, Data, Eq, Show)
 
 -- Pragmas ----------------------------------------------------------------
 
@@ -374,28 +379,32 @@ data Pragma
   | BuiltinPragma             Range String Expr
   | RewritePragma             Range [QName]
   | CompiledDataPragma        Range QName String [String]
-  | CompiledDeclareDataPragma Range QName String
   | CompiledTypePragma        Range QName String
   | CompiledPragma            Range QName String
   | CompiledExportPragma      Range QName String
-  | CompiledEpicPragma        Range QName String
   | CompiledJSPragma          Range QName String
   | CompiledUHCPragma         Range QName String
   | CompiledDataUHCPragma     Range QName String [String]
   | HaskellCodePragma         Range String
+  | ForeignPragma             Range String String       -- ^ first string is backend name
+  | CompilePragma             Range String QName String -- ^ first string is backend name
   | StaticPragma              Range QName
+  | InjectivePragma           Range QName
   | InlinePragma              Range QName
   | ImportPragma              Range String
     -- ^ Invariant: The string must be a valid Haskell module name.
   | ImportUHCPragma           Range String
     -- ^ same as above, but for the UHC backend
   | ImpossiblePragma          Range
+  | EtaPragma                 Range QName
+    -- ^ For coinductive records, use pragma instead of regular
+    --   @eta-equality@ definition (as it is might make Agda loop).
   | TerminationCheckPragma    Range (TerminationCheck Name)
   | CatchallPragma            Range
   | DisplayPragma             Range Pattern Expr
   | NoPositivityCheckPragma   Range
   | PolarityPragma            Range Name [Occurrence]
-  deriving (Typeable)
+  deriving (Typeable, Data)
 
 ---------------------------------------------------------------------------
 
@@ -455,8 +464,8 @@ appView e =
   where
     vApp (AppView e es) arg = AppView e (es ++ [arg])
 
-    arg (HiddenArg   _ e) = setHiding Hidden   $ defaultArg e
-    arg (InstanceArg _ e) = setHiding Instance $ defaultArg e
+    arg (HiddenArg   _ e) = hide         $ defaultArg e
+    arg (InstanceArg _ e) = makeInstance $ defaultArg e
     arg e                 = defaultArg (unnamed e)
 
 {--------------------------------------------------------------------------
@@ -642,20 +651,22 @@ instance HasRange Pragma where
   getRange (BuiltinPragma r _ _)             = r
   getRange (RewritePragma r _)               = r
   getRange (CompiledDataPragma r _ _ _)      = r
-  getRange (CompiledDeclareDataPragma r _ _) = r
   getRange (CompiledTypePragma r _ _)        = r
   getRange (CompiledPragma r _ _)            = r
   getRange (CompiledExportPragma r _ _)      = r
-  getRange (CompiledEpicPragma r _ _)        = r
   getRange (CompiledJSPragma r _ _)          = r
   getRange (CompiledUHCPragma r _ _)         = r
   getRange (CompiledDataUHCPragma r _ _ _)   = r
   getRange (HaskellCodePragma r _)           = r
+  getRange (CompilePragma r _ _ _)           = r
+  getRange (ForeignPragma r _ _)             = r
   getRange (StaticPragma r _)                = r
+  getRange (InjectivePragma r _)             = r
   getRange (InlinePragma r _)                = r
   getRange (ImportPragma r _)                = r
   getRange (ImportUHCPragma r _)             = r
   getRange (ImpossiblePragma r)              = r
+  getRange (EtaPragma r _)                   = r
   getRange (TerminationCheckPragma r _)      = r
   getRange (CatchallPragma r)                = r
   getRange (DisplayPragma r _ _)             = r
@@ -818,23 +829,25 @@ instance KillRange Pragma where
   killRange (BuiltinPragma _ s e)             = killRange1 (BuiltinPragma noRange s) e
   killRange (RewritePragma _ qs)              = killRange1 (RewritePragma noRange) qs
   killRange (CompiledDataPragma _ q s ss)     = killRange1 (\q -> CompiledDataPragma noRange q s ss) q
-  killRange (CompiledDeclareDataPragma _ q s) = killRange1 (\q -> CompiledDeclareDataPragma noRange q s) q
   killRange (CompiledTypePragma _ q s)        = killRange1 (\q -> CompiledTypePragma noRange q s) q
   killRange (CompiledPragma _ q s)            = killRange1 (\q -> CompiledPragma noRange q s) q
   killRange (CompiledExportPragma _ q s)      = killRange1 (\q -> CompiledExportPragma noRange q s) q
-  killRange (CompiledEpicPragma _ q s)        = killRange1 (\q -> CompiledEpicPragma noRange q s) q
   killRange (CompiledJSPragma _ q s)          = killRange1 (\q -> CompiledJSPragma noRange q s) q
   killRange (CompiledUHCPragma _ q s)         = killRange1 (\q -> CompiledUHCPragma noRange q s) q
   killRange (CompiledDataUHCPragma _ q s ss)  = killRange1 (\q -> CompiledDataUHCPragma noRange q s ss) q
   killRange (HaskellCodePragma _ s)           = HaskellCodePragma noRange s
   killRange (StaticPragma _ q)                = killRange1 (StaticPragma noRange) q
+  killRange (InjectivePragma _ q)             = killRange1 (InjectivePragma noRange) q
   killRange (InlinePragma _ q)                = killRange1 (InlinePragma noRange) q
   killRange (ImportPragma _ s)                = ImportPragma noRange s
   killRange (ImportUHCPragma _ s)             = ImportUHCPragma noRange s
+  killRange (CompilePragma _ b q s)           = killRange1 (\ q -> CompilePragma noRange b q s) q
+  killRange (ForeignPragma _ b s)             = ForeignPragma noRange b s
   killRange (ImpossiblePragma _)              = ImpossiblePragma noRange
   killRange (TerminationCheckPragma _ t)      = TerminationCheckPragma noRange (killRange t)
   killRange (CatchallPragma _)                = CatchallPragma noRange
   killRange (DisplayPragma _ lhs rhs)         = killRange2 (DisplayPragma noRange) lhs rhs
+  killRange (EtaPragma _ q)                   = killRange1 (EtaPragma noRange) q
   killRange (NoPositivityCheckPragma _)       = NoPositivityCheckPragma noRange
   killRange (PolarityPragma _ q occs)         = killRange1 (\q -> PolarityPragma noRange q occs) q
 
@@ -949,20 +962,22 @@ instance NFData Pragma where
   rnf (BuiltinPragma _ a b)             = rnf a `seq` rnf b
   rnf (RewritePragma _ a)               = rnf a
   rnf (CompiledDataPragma _ a b c)      = rnf a `seq` rnf b `seq` rnf c
-  rnf (CompiledDeclareDataPragma _ a b) = rnf a `seq` rnf b
   rnf (CompiledTypePragma _ a b)        = rnf a `seq` rnf b
   rnf (CompiledPragma _ a b)            = rnf a `seq` rnf b
   rnf (CompiledExportPragma _ a b)      = rnf a `seq` rnf b
-  rnf (CompiledEpicPragma _ a b)        = rnf a `seq` rnf b
   rnf (CompiledJSPragma _ a b)          = rnf a `seq` rnf b
   rnf (CompiledUHCPragma _ a b)         = rnf a `seq` rnf b
   rnf (CompiledDataUHCPragma _ a b c)   = rnf a `seq` rnf b `seq` rnf c
   rnf (HaskellCodePragma _ s)           = rnf s
+  rnf (CompilePragma _ a b c)           = rnf a `seq` rnf b `seq` rnf c
+  rnf (ForeignPragma _ b s)             = rnf b `seq` rnf s
   rnf (StaticPragma _ a)                = rnf a
+  rnf (InjectivePragma _ a)             = rnf a
   rnf (InlinePragma _ a)                = rnf a
   rnf (ImportPragma _ a)                = rnf a
   rnf (ImportUHCPragma _ a)             = rnf a
   rnf (ImpossiblePragma _)              = ()
+  rnf (EtaPragma _ a)                   = rnf a
   rnf (TerminationCheckPragma _ a)      = rnf a
   rnf (CatchallPragma _)                = ()
   rnf (DisplayPragma _ a b)             = rnf a `seq` rnf b
