@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE DeriveDataTypeable  #-}
-{-# LANGUAGE GADTs               #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP                #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE GADTs              #-}
 
 -- | Preprocess 'Agda.Syntax.Concrete.Declaration's, producing 'NiceDeclaration's.
 --
@@ -31,6 +30,7 @@ module Agda.Syntax.Concrete.Definitions
     , NiceConstructor, NiceTypeSignature
     , Clause(..)
     , DeclarationException(..)
+    , DeclarationWarning(..)
     , Nice, runNice
     , niceDeclarations
     , notSoNiceDeclarations
@@ -40,7 +40,7 @@ module Agda.Syntax.Concrete.Definitions
 
 import Prelude hiding (null)
 
-import Control.Arrow ((***))
+import Control.Arrow ((***), first, second)
 import Control.Applicative hiding (empty)
 import Control.Monad.State
 
@@ -52,9 +52,12 @@ import qualified Data.Map as Map
 import Data.Map (Map)
 import Data.Maybe
 import Data.Semigroup ( Semigroup, Monoid, (<>), mempty, mappend )
-import Data.List as List hiding (null)
+import qualified Data.List as List
 import qualified Data.Set as Set
 import Data.Traversable (Traversable, traverse)
+import qualified Data.Traversable as Trav
+
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Concrete
@@ -67,10 +70,12 @@ import Agda.Syntax.Concrete.Pretty ()
 
 import Agda.TypeChecking.Positivity.Occurrence
 
-import Agda.Utils.Except ( Error(strMsg), MonadError(throwError) )
+import Agda.Utils.Except ( MonadError(throwError,catchError) )
+import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List (caseList, headMaybe, isSublistOf)
+import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import qualified Agda.Utils.Pretty as Pretty
@@ -130,16 +135,18 @@ data NiceDeclaration
     --   without type signature or a pattern lhs (e.g. for irrefutable let).
     --   The 'Declaration' is the actual 'FunClause'.
   | FunSig Range Fixity' Access IsAbstract IsInstance IsMacro ArgInfo TerminationCheck Name Expr
-  | FunDef Range [Declaration] Fixity' IsAbstract TerminationCheck Name [Clause]
+  | FunDef Range [Declaration] Fixity' IsAbstract IsInstance TerminationCheck Name [Clause]
       -- ^ Block of function clauses (we have seen the type signature before).
       --   The 'Declaration's are the original declarations that were processed
       --   into this 'FunDef' and are only used in 'notSoNiceDeclaration'.
+      --   Andreas, 2017-01-01: Because of issue #2372, we add 'IsInstance' here.
+      --   An alias should know that it is an instance.
   | DataDef Range Fixity' IsAbstract PositivityCheck Name [LamBinding] [NiceConstructor]
   | RecDef Range Fixity' IsAbstract PositivityCheck Name (Maybe (Ranged Induction)) (Maybe Bool) (Maybe (ThingWithFixity Name, IsInstance)) [LamBinding] [NiceDeclaration]
   | NicePatternSyn Range Fixity' Name [Arg Name] Pattern
   | NiceUnquoteDecl Range [Fixity'] Access IsAbstract IsInstance TerminationCheck [Name] Expr
   | NiceUnquoteDef Range [Fixity'] Access IsAbstract TerminationCheck [Name] Expr
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 type TerminationCheck = Common.TerminationCheck Measure
 
@@ -157,7 +164,7 @@ type NiceTypeSignature  = NiceDeclaration
 -- | One clause in a function definition. There is no guarantee that the 'LHS'
 --   actually declares the 'Name'. We will have to check that later.
 data Clause = Clause Name Catchall LHS RHS WhereClause [Clause]
-    deriving (Typeable, Show)
+    deriving (Typeable, Data, Show)
 
 -- | The exception type.
 data DeclarationException
@@ -170,16 +177,11 @@ data DeclarationException
         | MissingTypeSignature LHS -- Andreas 2012-06-02: currently unused, remove after a while -- Fredrik 2012-09-20: now used, can we keep it?
         | MissingDataSignature Name
         | WrongDefinition Name DataRecOrFun DataRecOrFun
-        | WrongParameters Name
+        | WrongParameters Name Params Params
+          -- ^ 'Name' of symbol, 'Params' of signature, 'Params' of definition.
         | NotAllowedInMutual NiceDeclaration
-        | UnknownNamesInFixityDecl [Name]
-        | UnknownNamesInPolarityPragmas [Name]
-        | PolarityPragmasButNotPostulates [Name]
         | Codata Range
         | DeclarationPanic String
-        | UselessPrivate Range
-        | UselessAbstract Range
-        | UselessInstance Range
         | WrongContentBlock KindOfBlock Range
         | AmbiguousFunClauses LHS [Name] -- ^ in a mutual block, a clause could belong to any of the @[Name]@ type signatures
         | InvalidTerminationCheckPragma Range
@@ -194,7 +196,17 @@ data DeclarationException
         | BadMacroDef NiceDeclaration
         | InvalidNoPositivityCheckPragma Range
 
-    deriving (Typeable)
+    deriving (Typeable, Data, Show)
+
+-- | Non-fatal errors encountered in the Nicifier
+data DeclarationWarning
+  = UnknownNamesInFixityDecl [Name]
+  | UnknownNamesInPolarityPragmas [Name]
+  | PolarityPragmasButNotPostulates [Name]
+  | UselessPrivate Range
+  | UselessAbstract Range
+  | UselessInstance Range
+  deriving (Typeable, Data, Show)
 
 -- | Several declarations expect only type signatures as sub-declarations.  These are:
 data KindOfBlock
@@ -203,7 +215,7 @@ data KindOfBlock
   | InstanceBlock   -- ^ @instance@.  Actually, here all kinds of sub-declarations are allowed a priori.
   | FieldBlock      -- ^ @field@.  Ensured by parser.
   | DataBlock       -- ^ @data ... where@.  Here we got a bad error message for Agda-2.5 (Issue 1698).
-  deriving (Typeable, Eq, Ord, Show)
+  deriving (Typeable, Data, Eq, Ord, Show)
 
 
 instance HasRange DeclarationException where
@@ -216,17 +228,11 @@ instance HasRange DeclarationException where
   getRange (MissingTypeSignature x)             = getRange x
   getRange (MissingDataSignature x)             = getRange x
   getRange (WrongDefinition x k k')             = getRange x
-  getRange (WrongParameters x)                  = getRange x
+  getRange (WrongParameters x _ _)              = getRange x
   getRange (AmbiguousFunClauses lhs xs)         = getRange lhs
   getRange (NotAllowedInMutual x)               = getRange x
-  getRange (UnknownNamesInFixityDecl xs)        = getRange . head $ xs
-  getRange (UnknownNamesInPolarityPragmas xs)   = getRange . head $ xs
-  getRange (PolarityPragmasButNotPostulates xs) = getRange . head $ xs
   getRange (Codata r)                           = r
   getRange (DeclarationPanic _)                 = noRange
-  getRange (UselessPrivate r)                   = r
-  getRange (UselessAbstract r)                  = r
-  getRange (UselessInstance r)                  = r
   getRange (WrongContentBlock _ r)              = r
   getRange (InvalidTerminationCheckPragma r)    = r
   getRange (InvalidMeasureMutual r)             = r
@@ -236,6 +242,14 @@ instance HasRange DeclarationException where
   getRange (BadMacroDef d)                      = getRange d
   getRange (InvalidNoPositivityCheckPragma r)   = r
 
+instance HasRange DeclarationWarning where
+  getRange (UnknownNamesInFixityDecl xs)        = getRange . head $ xs
+  getRange (UnknownNamesInPolarityPragmas xs)   = getRange . head $ xs
+  getRange (PolarityPragmasButNotPostulates xs) = getRange . head $ xs
+  getRange (UselessPrivate r)                   = r
+  getRange (UselessAbstract r)                  = r
+  getRange (UselessInstance r)                  = r
+
 instance HasRange NiceDeclaration where
   getRange (Axiom r _ _ _ _ _ _ _ _)         = r
   getRange (NiceField r _ _ _ _ _ _)         = r
@@ -247,7 +261,7 @@ instance HasRange NiceDeclaration where
   getRange (NicePragma r _)                  = r
   getRange (PrimitiveFunction r _ _ _ _ _)   = r
   getRange (FunSig r _ _ _ _ _ _ _ _ _)      = r
-  getRange (FunDef r _ _ _ _ _ _)            = r
+  getRange (FunDef r _ _ _ _ _ _ _)          = r
   getRange (DataDef r _ _ _ _ _ _)           = r
   getRange (RecDef r _ _ _ _ _ _ _ _ _)      = r
   getRange (NiceRecSig r _ _ _ _ _ _ _)      = r
@@ -257,9 +271,6 @@ instance HasRange NiceDeclaration where
   getRange (NiceUnquoteDecl r _ _ _ _ _ _ _) = r
   getRange (NiceUnquoteDef r _ _ _ _ _ _)    = r
 
-instance Error DeclarationException where
-  strMsg = DeclarationPanic
-
 -- These error messages can (should) be terminated by a dot ".",
 -- there is no error context printed after them.
 instance Pretty DeclarationException where
@@ -286,31 +297,21 @@ instance Pretty DeclarationException where
   pretty (WrongDefinition x k k') = fsep $ pretty x :
     pwords ("has been declared as a " ++ show k ++
       ", but is being defined as a " ++ show k')
-  pretty (WrongParameters x) = fsep $
-    pwords "List of parameters does not match previous signature for" ++ [pretty x]
+  pretty (WrongParameters x sig def) = fsep $
+    pwords "List of parameters " ++ map pretty def ++
+    pwords " does not match parameters " ++ map pretty sig ++
+    pwords " of previous signature for " ++ [pretty x]
   pretty (AmbiguousFunClauses lhs xs) = sep
     [ fsep $
         pwords "More than one matching type signature for left hand side " ++ [pretty lhs] ++
         pwords "it could belong to any of:"
     , vcat $ map (pretty . PrintRange) xs
     ]
-  pretty (UnknownNamesInFixityDecl xs) = fsep $
-    pwords "The following names are not declared in the same scope as their syntax or fixity declaration (i.e., either not in scope at all, imported from another module, or declared in a super module):" ++ map pretty xs
-  pretty (UnknownNamesInPolarityPragmas xs) = fsep $
-    pwords "The following names are not declared in the same scope as their polarity pragmas (they could for instance be out of scope, imported from another module, or declared in a super module):" ++ map pretty xs
-  pretty (PolarityPragmasButNotPostulates xs) = fsep $
-    pwords "Polarity pragmas have been given for the following identifiers which are not postulates:" ++ map pretty xs
-  pretty (UselessPrivate _)      = fsep $
-    pwords "Using private here has no effect. Private applies only to declarations that introduce new identifiers into the module, like type signatures and data, record, and module declarations."
-  pretty (UselessAbstract _)      = fsep $
-    pwords "Using abstract here has no effect. Abstract applies only definitions like data definitions, record type definitions and function clauses."
-  pretty (UselessInstance _)      = fsep $
-    pwords "Using instance here has no effect. Instance applies only to declarations that introduce new identifiers into the module, like type signatures and axioms."
   pretty (WrongContentBlock b _)      = fsep . pwords $
     case b of
       PostulateBlock -> "A postulate block can only contain type signatures, possibly under keyword instance"
       DataBlock -> "A data definition can only contain type signatures, possibly under keyword instance"
-      _ -> __IMPOSSIBLE__
+      _ -> "Unexpected declaration"
   pretty (PragmaNoTerminationCheck _) = fsep $
     pwords "Pragma {-# NO_TERMINATION_CHECK #-} has been removed.  To skip the termination check, label your definitions either as {-# TERMINATING #-} or {-# NON_TERMINATING #-}."
   pretty (InvalidTerminationCheckPragma _) = fsep $
@@ -332,6 +333,20 @@ instance Pretty DeclarationException where
   pretty (InvalidNoPositivityCheckPragma _) = fsep $
     pwords "No positivity checking pragmas can only precede a mutual block or a data/record definition."
 
+instance Pretty DeclarationWarning where
+  pretty (UnknownNamesInFixityDecl xs) = fsep $
+    pwords "The following names are not declared in the same scope as their syntax or fixity declaration (i.e., either not in scope at all, imported from another module, or declared in a super module):" ++ map pretty xs
+  pretty (UnknownNamesInPolarityPragmas xs) = fsep $
+    pwords "The following names are not declared in the same scope as their polarity pragmas (they could for instance be out of scope, imported from another module, or declared in a super module):" ++ map pretty xs
+  pretty (PolarityPragmasButNotPostulates xs) = fsep $
+    pwords "Polarity pragmas have been given for the following identifiers which are not postulates:" ++ map pretty xs
+  pretty (UselessPrivate _)      = fsep $
+    pwords "Using private here has no effect. Private applies only to declarations that introduce new identifiers into the module, like type signatures and data, record, and module declarations."
+  pretty (UselessAbstract _)      = fsep $
+    pwords "Using abstract here has no effect. Abstract applies to only definitions like data definitions, record type definitions and function clauses."
+  pretty (UselessInstance _)      = fsep $
+    pwords "Using instance here has no effect. Instance applies only to declarations that introduce new identifiers into the module, like type signatures and axioms."
+
 declName :: NiceDeclaration -> String
 declName Axiom{}             = "Postulates"
 declName NiceField{}         = "Fields"
@@ -365,9 +380,13 @@ data InMutual
 -- | The kind of the forward declaration, remembering the parameters.
 
 data DataRecOrFun
-  = DataName PositivityCheck Params  -- ^ name of a data with parameters
-  | RecName  PositivityCheck Params  -- ^ name of a record with parameters
-  | FunName  TerminationCheck        -- ^ name of a function
+  = DataName { kindPosCheck :: PositivityCheck, kindParams :: Params }
+    -- ^ Name of a data type with parameters.
+  | RecName  { kindPosCheck :: PositivityCheck, kindParams :: Params }
+    -- ^ Name of a record type with parameters.
+  | FunName  TerminationCheck
+    -- ^ Name of a function.
+  deriving (Typeable, Data)
 
 -- Ignore pragmas when checking equality
 instance Eq DataRecOrFun where
@@ -376,7 +395,7 @@ instance Eq DataRecOrFun where
   FunName  _   == FunName  _   = True
   _            == _            = False
 
-type Params = [Hiding]
+type Params = [Arg BoundName]
 
 instance Show DataRecOrFun where
   show (DataName _ n) = "data type" --  "with " ++ show n ++ " visible parameters"
@@ -407,6 +426,7 @@ positivityCheck _               = True
 --   NO_TERMINATION_CHECK pragma.
 combineTermChecks :: Range -> [TerminationCheck] -> Nice TerminationCheck
 combineTermChecks r tcs = loop tcs where
+  loop :: [TerminationCheck] -> Nice TerminationCheck
   loop []         = return TerminationCheck
   loop (tc : tcs) = do
     let failure r = throwError $ InvalidMeasureMutual r
@@ -431,10 +451,75 @@ combineTermChecks r tcs = loop tcs where
       (NonTerminating        , NoTerminationCheck    ) -> failure r
       (NonTerminating        , Terminating           ) -> failure r
 
+-- | Check that the parameters of the data/record definition
+--   match the parameters of the corresponding signature.
+--
+--   The definition may omit some hidden parameters.
+--   The names need to match.
+--   The types are not checked here.
+--
+--   Precondition: the signature and definition have the same kind (data/record/fun).
+--
+matchParameters
+  :: Name          -- ^ The data/record name.
+  -> DataRecOrFun  -- ^ The data/record signature.
+  -> DataRecOrFun  -- ^ The parameters as given in the data/record definition.
+  -> Nice ()
+matchParameters _ FunName{} FunName{} = return ()
+matchParameters x sig def = loop (kindParams sig) (kindParams def)
+  where
+  failure = throwError $ WrongParameters x (kindParams sig) (kindParams def)
+  loop hs     []     = unless (all notVisible hs) failure
+  loop []     (_:_)  = failure
+  loop (h:hs) (j:js)
+    | h == j                  = loop hs js
+    | notVisible h, visible j = loop hs (j:js)
+    | otherwise               = failure
 
 -- | Nicifier monad.
+--   Preserve the state when throwing an exception.
 
-type Nice = StateT NiceEnv (Either DeclarationException)
+newtype Nice a = Nice { unNice :: NiceEnv -> (Either DeclarationException a, NiceEnv) }
+
+-- We have to hand-roll the instances ourselves, since the automagic does not
+-- work for @Nice a = State s (Except e a)@, only for the usual
+-- @Nice a = StateT s (Except e) a@.
+
+instance Functor Nice where
+  fmap f m = Nice $ \ s ->
+    let (r, s') = unNice m s in
+    case r of
+      Left  e -> (Left e, s')
+      Right a -> (Right (f a), s')
+
+instance Applicative Nice where
+  pure a = Nice $ \ s -> (Right a, s)
+  (<*>)  = ap
+
+instance Monad Nice where
+  return = pure
+  m >>= k  = Nice $ \ s ->
+    let (r, s') = unNice m s in
+    case r of
+      Left e  -> (Left e, s')
+      Right a -> unNice (k a) s'
+
+instance MonadState NiceEnv Nice where
+  state f  = Nice $ \ s -> first Right $ f s
+  -- get = Nice $ \ s -> (Right s, s)  -- Subsumed by state
+
+instance MonadError DeclarationException Nice where
+  throwError e   = Nice $ \ s -> (Left e, s)
+  catchError m h = Nice $ \ s ->
+    let (r, s') = unNice m s in
+    case r of
+      Left e  -> unNice (h e) s'
+      Right a -> (Right a, s')
+
+-- | Run a Nicifier computation, return result and warnings
+--   (in chronological order).
+runNice :: Nice a -> (Either DeclarationException a, NiceWarnings)
+runNice m = second (reverse . niceWarn) $ unNice m initNiceEnv
 
 -- | Nicifier state.
 
@@ -449,11 +534,15 @@ data NiceEnv = NiceEnv
     -- ^ Catchall pragma waiting for a function clause.
   , fixs     :: Fixities
   , pols     :: Polarities
+  , niceWarn :: NiceWarnings
+    -- ^ Stack of warnings. Head is last warning.
   }
 
 type LoneSigs   = Map Name DataRecOrFun
 type Fixities   = Map Name Fixity'
 type Polarities = Map Name [Occurrence]
+type NiceWarnings = [DeclarationWarning]
+     -- ^ Stack of warnings. Head is last warning.
 
 -- | Initial nicifier state.
 
@@ -465,6 +554,7 @@ initNiceEnv = NiceEnv
   , _catchall = False
   , fixs      = empty
   , pols      = empty
+  , niceWarn  = []
   }
 
 -- * Handling the lone signatures, stored to infer mutual blocks.
@@ -553,6 +643,10 @@ withCatchallPragma ca f = do
   catchallPragma .= ca_old
   return result
 
+-- | Add a new warning.
+niceWarning :: DeclarationWarning -> Nice ()
+niceWarning w = modify $ \ st -> st { niceWarn = w : niceWarn st }
+
 -- | Check whether name is not "_" and return its fixity.
 getFixity :: Name -> Nice Fixity'
 getFixity x = Map.findWithDefault noFixity' x <$> gets fixs -- WAS: defaultFixity'
@@ -565,9 +659,6 @@ getPolarity x = do
   modify (\s -> s { pols = Map.delete x (pols s) })
   return p
 
-runNice :: Nice a -> Either DeclarationException a
-runNice nice = nice `evalStateT` initNiceEnv
-
 data DeclKind
     = LoneSig DataRecOrFun Name
     | LoneDefs DataRecOrFun [Name]
@@ -578,7 +669,7 @@ declKind :: NiceDeclaration -> DeclKind
 declKind (FunSig _ _ _ _ _ _ _ tc x _)    = LoneSig (FunName tc) x
 declKind (NiceRecSig _ _ _ _ pc x pars _) = LoneSig (RecName pc $ parameters pars) x
 declKind (NiceDataSig _ _ _ _ pc x pars _)= LoneSig (DataName pc $ parameters pars) x
-declKind (FunDef _ _ _ _ tc x _)          = LoneDefs (FunName tc) [x]
+declKind (FunDef _ _ _ _ _ tc x _)        = LoneDefs (FunName tc) [x]
 declKind (DataDef _ _ _ pc x pars _)      = LoneDefs (DataName pc $ parameters pars) [x]
 declKind (RecDef _ _ _ pc x _ _ _ pars _) = LoneDefs (RecName pc $ parameters pars) [x]
 declKind (NiceUnquoteDef _ _ _ _ tc xs _) = LoneDefs (FunName tc) xs
@@ -595,43 +686,61 @@ declKind NiceFunClause{}                  = OtherDecl
 declKind NicePatternSyn{}                 = OtherDecl
 declKind NiceUnquoteDecl{}                = OtherDecl
 
--- | Compute visible parameters of a data or record signature or definition.
+-- | Compute parameters of a data or record signature or definition.
 parameters :: [LamBinding] -> Params
-parameters = List.concat . List.map numP where
-  numP (DomainFree i _) = [argInfoHiding i]
-  numP (DomainFull (TypedBindings _ (Arg i (TBind _ xs _)))) = List.replicate (length xs) $ argInfoHiding i
-  numP (DomainFull (TypedBindings _ (Arg _ TLet{})))         = []
+parameters = List.concatMap $ \case
+  DomainFree i x                                      -> [Arg i x]
+  DomainFull (TypedBindings _ (Arg _ TLet{}))         -> []
+  DomainFull (TypedBindings _ (Arg i (TBind _ xs _))) -> for xs $ \ (WithHiding h x) ->
+    mergeHiding $ WithHiding h $ Arg i x
 
 -- | Main.
 niceDeclarations :: [Declaration] -> Nice [NiceDeclaration]
 niceDeclarations ds = do
+
   -- Get fixity and syntax declarations.
   (fixs, polarities) <- fixitiesAndPolarities ds
   let declared    = Set.fromList (concatMap declaredNames ds)
-      unknownFixs = Map.keysSet fixs       Set.\\ declared
-      unknownPols = Map.keysSet polarities Set.\\ declared
-  case (Set.null unknownFixs, Set.null unknownPols) of
-    -- If we have fixity/syntax decls for names not declared
-    -- in the current scope, fail.
-    (False, _)   -> throwError $ UnknownNamesInFixityDecl
-                                   (Set.toList unknownFixs)
-    -- Fail if there are polarity pragmas with undeclared names.
-    (_, False)   -> throwError $ UnknownNamesInPolarityPragmas
-                                   (Set.toList unknownPols)
-    (True, True) -> localState $ do
-      -- Run the nicifier in an initial environment of fixity decls
-      -- and polarities.
-      put $ initNiceEnv { fixs = fixs, pols = polarities }
-      ds <- nice ds
-      -- Check that every polarity pragma was used.
-      unusedPolarities <- gets (Map.keys . pols)
-      unless (null unusedPolarities) $ do
-        throwError $ PolarityPragmasButNotPostulates unusedPolarities
-      -- Check that every signature got its definition.
-      checkLoneSigs . Map.toList =<< use loneSigs
-      -- Note that loneSigs is ensured to be empty.
-      -- (Important, since inferMutualBlocks also uses loneSigs state).
-      inferMutualBlocks ds
+
+  -- If we have names in fixity declarations
+  -- which are not defined in the appropriate scope,
+  -- raise a warning and delete them from fixs.
+  fixs <- ifNull (Map.keysSet fixs Set.\\ declared) (return fixs) $ \ unknownFixs -> do
+    niceWarning $ UnknownNamesInFixityDecl $ Set.toList unknownFixs
+    -- Note: Data.Map.restrictKeys requires containers >= 0.5.8.2
+    -- return $ Map.restrictKeys fixs declared
+    return $ Map.filterWithKey (\ k _ -> Set.member k declared) fixs
+
+  -- Same for undefined names in polarity declarations.
+  polarities <- ifNull (Map.keysSet polarities Set.\\ declared) (return polarities) $
+    \ unknownPols -> do
+      niceWarning $ UnknownNamesInPolarityPragmas $ Set.toList unknownPols
+      -- Note: Data.Map.restrictKeys requires containers >= 0.5.8.2
+      -- return $ Map.restrictKeys polarities declared
+      return $ Map.filterWithKey (\ k _ -> Set.member k declared) polarities
+
+  -- Run the nicifier in an initial environment of fixity decls
+  -- and polarities.  But keep the warnings.
+  st <- get
+  put $ initNiceEnv { fixs = fixs, pols = polarities, niceWarn = niceWarn st }
+  ds <- nice ds
+
+  -- Check that every polarity pragma was used.
+  unlessNullM (Map.keys <$> gets pols) $ \ unusedPolarities -> do
+    niceWarning $ PolarityPragmasButNotPostulates unusedPolarities
+
+  -- Check that every signature got its definition.
+  checkLoneSigs . Map.toList =<< use loneSigs
+
+  -- Note that loneSigs is ensured to be empty.
+  -- (Important, since inferMutualBlocks also uses loneSigs state).
+  res <- inferMutualBlocks ds
+
+  -- Restore the old state, but keep the warnings.
+  warns <- gets niceWarn
+  put $ st { niceWarn = warns }
+  return res
+
   where
     -- Compute the names defined in a declaration.
     -- We stay in the current scope, i.e., do not go into modules.
@@ -767,7 +876,7 @@ niceDeclarations ds = do
                removeLoneSig x
                cs  <- mkClauses x (expandEllipsis fits) False
                fx  <- getFixity x
-               return ([FunDef (getRange fits) fits fx ConcreteDef termCheck x cs] , rest)
+               return ([FunDef (getRange fits) fits fx ConcreteDef NotInstanceDef termCheck x cs] , rest)
 
             -- case: clauses match more than one sigs (ambiguity)
             l -> throwError $ AmbiguousFunClauses lhs $ reverse $ map fst l -- "ambiguous function clause; cannot assign it uniquely to one type signature"
@@ -779,13 +888,13 @@ niceDeclarations ds = do
         (DataSig r Inductive x tel t) -> do
           pc <- use positivityCheckPragma
           addLoneSig x (DataName pc $ parameters tel)
-          (,) <$> dataOrRec pc DataDef NiceDataSig (niceAxioms DataBlock) r x tel (Just t) Nothing
+          (,) <$> dataOrRec pc DataDef NiceDataSig (niceAxioms DataBlock) r x (Just (tel, t)) Nothing
               <*> return ds
 
-        (Data r Inductive x tel t cs) -> do
+        (Data r Inductive x tel mt cs) -> do
           pc <- use positivityCheckPragma
-          t <- defaultTypeSig (DataName pc $ parameters tel) x t
-          (,) <$> dataOrRec pc DataDef NiceDataSig (niceAxioms DataBlock) r x tel t (Just cs)
+          mt <- defaultTypeSig (DataName pc $ parameters tel) x mt
+          (,) <$> dataOrRec pc DataDef NiceDataSig (niceAxioms DataBlock) r x ((tel,) <$> mt) (Just (tel, cs))
               <*> return ds
 
         (RecordSig r x tel t)         -> do
@@ -794,12 +903,12 @@ niceDeclarations ds = do
           fx <- getFixity x
           return ([NiceRecSig r fx PublicAccess ConcreteDef pc x tel t] , ds)
 
-        (Record r x i e c tel t cs)   -> do
+        (Record r x i e c tel mt cs)   -> do
           pc <- use positivityCheckPragma
-          t <- defaultTypeSig (RecName pc $ parameters tel) x t
+          mt <- defaultTypeSig (RecName pc $ parameters tel) x mt
           c <- traverse (\(cname, cinst) -> do fix <- getFixity cname; return (ThingWithFixity cname fix, cinst)) c
           (,) <$> dataOrRec pc (\ r f a pc x tel cs -> RecDef r f a pc x i e c tel cs) NiceRecSig
-                    niceDeclarations r x tel t (Just cs)
+                    niceDeclarations r x ((tel,) <$> mt) (Just (tel, cs))
               <*> return ds
 
         Mutual r ds' ->
@@ -942,34 +1051,36 @@ niceDeclarations ds = do
     defaultTypeSig :: DataRecOrFun -> Name -> Maybe Expr -> Nice (Maybe Expr)
     defaultTypeSig k x t@Just{} = return t
     defaultTypeSig k x Nothing  = do
-      mk <- getSig x
-      case mk of
-        Nothing -> throwError $ MissingDataSignature x
-        Just k' | k == k'       -> Nothing <$ removeLoneSig x
-                | sameKind k k' -> throwError $ WrongParameters x
-                | otherwise     -> throwError $ WrongDefinition x k' k
+      caseMaybeM (getSig x) (throwError $ MissingDataSignature x) $ \ k' -> do
+        unless (sameKind k k') $ throwError $ WrongDefinition x k' k
+        unless (k == k') $ matchParameters x k' k
+        Nothing <$ removeLoneSig x
 
-    dataOrRec :: forall a .
-                 PositivityCheck ->
-                 (Range -> Fixity' -> IsAbstract -> PositivityCheck -> Name -> [LamBinding] ->
-                   [NiceConstructor] -> NiceDeclaration) ->
-                 (Range -> Fixity' -> Access -> IsAbstract -> PositivityCheck -> Name -> [LamBinding] -> Expr ->
-                   NiceDeclaration) ->
-                 ([a] -> Nice [NiceDeclaration]) ->
-                 Range ->
-                 Name ->
-                 [LamBinding] ->
-                 Maybe Expr ->
-                 Maybe [a] ->
-                 Nice [NiceDeclaration]
-    dataOrRec pc mkDef mkSig niceD r x tel mt mcs = do
-      mds <- traverse niceD mcs
+    dataOrRec
+      :: forall a
+      .  PositivityCheck
+      -> (Range -> Fixity' -> IsAbstract -> PositivityCheck -> Name -> [LamBinding] -> [NiceConstructor] -> NiceDeclaration)
+         -- ^ Construct definition.
+      -> (Range -> Fixity' -> Access -> IsAbstract -> PositivityCheck -> Name -> [LamBinding] -> Expr -> NiceDeclaration)
+         -- ^ Construct signature.
+      -> ([a] -> Nice [NiceDeclaration])
+         -- ^ Constructor checking.
+      -> Range
+      -> Name          -- ^ Data/record type name.
+      -> Maybe ([LamBinding], Expr)    -- ^ Parameters and type.  If not @Nothing@ a signature is created.
+      -> Maybe ([LamBinding], [a])     -- ^ Parameters and constructors.  If not @Nothing@, a definition body is created.
+      -> Nice [NiceDeclaration]
+    dataOrRec pc mkDef mkSig niceD r x mt mcs = do
+      mds <- Trav.forM mcs $ \ (tel, cs) -> (tel,) <$> niceD cs
       f   <- getFixity x
       return $ catMaybes $
-        [ mt <&> \ t -> mkSig (fuseRange x t) f PublicAccess ConcreteDef pc x tel t
-        , mkDef r f ConcreteDef pc x (concatMap dropType tel) <$> mds
+        [ mt  <&> \ (tel, t)  -> mkSig (fuseRange x t) f PublicAccess ConcreteDef pc x tel t
+        , mds <&> \ (tel, ds) -> mkDef r f ConcreteDef pc x (caseMaybe mt tel $ const $ concatMap dropType tel) ds
+          -- If a type is given (mt /= Nothing), we have to delete the types in @tel@
+          -- for the data definition, lest we duplicate them.
         ]
       where
+        -- | Drop type annotations and lets from bindings.
         dropType :: LamBinding -> [LamBinding]
         dropType (DomainFull (TypedBindings _r (Arg ai (TBind _ xs _)))) =
           map (mergeHiding . fmap (DomainFree ai)) xs
@@ -985,7 +1096,7 @@ niceDeclarations ds = do
       TypeSig rel x t -> do
         fx <- getFixity x
         return [ Axiom (getRange d) fx PublicAccess ConcreteDef NotInstanceDef rel Nothing x t ]
-      Field i x argt -> do
+      Field i x argt | b == FieldBlock -> do
         fx <- getFixity x
         return [ NiceField (getRange d) fx PublicAccess ConcreteDef i x argt ]
       InstanceB r decls -> do
@@ -1003,7 +1114,7 @@ niceDeclarations ds = do
       cs <- mkClauses x (expandEllipsis ds0) False
       f  <- getFixity x
       return [ FunSig (fuseRange x t) f PublicAccess ConcreteDef NotInstanceDef NotMacroDef info termCheck x t
-             , FunDef (getRange ds0) ds0 f ConcreteDef termCheck x cs ]
+             , FunDef (getRange ds0) ds0 f ConcreteDef NotInstanceDef termCheck x cs ]
         where
           t = case mt of
                 Just t  -> t
@@ -1017,21 +1128,32 @@ niceDeclarations ds = do
     expandEllipsis (d@(FunClause Ellipsis{} _ _ _) : ds) =
       d : expandEllipsis ds
     expandEllipsis (d@(FunClause lhs@(LHS p ps _ _) _ _ _) : ds) =
-      d : expand (setInserted p) (map setInserted ps) ds
+      d : expand (wipe p) (map wipe ps) ds
       where
         expand _ _ [] = []
         expand p ps (d@(Pragma (CatchallPragma r)) : ds) = d : expand p ps ds
-        expand p ps (FunClause (Ellipsis r ps' eqs []) rhs wh ca : ds) =
-          FunClause (LHS (setRange r p) ((setRange r ps) ++ ps') eqs []) rhs wh ca : expand p ps ds
         expand p ps (FunClause (Ellipsis r ps' eqs es) rhs wh ca : ds) =
-          FunClause (LHS (setRange r p) ((setRange r ps) ++ ps') eqs es) rhs wh ca : expand p (ps ++ ps') ds
+          FunClause (LHS (setRange r p) ((setRange r ps) ++ ps') eqs es) rhs wh ca
+            : expand p (applyUnless (null es) (++ (map wipe ps')) ps) ds
+                       -- If we have with-expressions (es /= []) then the following
+                       -- ellipses also get the additional with patterns ps'
+        -- We can have ellipses after a fun clause.
+        -- They refer to the last clause that introduced new with-expressions.
         expand p ps (d@(FunClause (LHS _ _ _ []) _ _ _) : ds) =
           d : expand p ps ds
-        expand _ _ (d@(FunClause (LHS p ps _ (_ : _)) _ _ _) : ds) =
-          d : expand p ps ds
+        -- Same here: If we have new with-expressions, the next ellipses will
+        -- refer to us.
+        expand _ _ (d@(FunClause (LHS p' ps' _ (_ : _)) _ _ _) : ds) =
+          d : expand (wipe p') (map wipe ps') ds
+          -- Andreas, Jesper, 2017-05-13, issue #2578
+          -- Need to update the range also on the next with-patterns.
         expand _ _ (_ : ds) = __IMPOSSIBLE__
     expandEllipsis (_ : ds) = __IMPOSSIBLE__
 
+    -- Before copying a pattern, remove traces to its origin.
+    wipe :: Pattern -> Pattern
+    wipe = killRange . setInserted
+
     setInserted :: Pattern -> Pattern
     setInserted p = case p of
       IdentP{} -> p
@@ -1093,14 +1215,14 @@ niceDeclarations ds = do
       xStrings   = nameStringParts x
       patStrings = concatMap nameStringParts pns
       in
---          trace ("x = " ++ show x) $
+--          trace ("x = " ++ prettyShow x) $
 --          trace ("pns = " ++ show pns) $
 --          trace ("xStrings = " ++ show xStrings) $
 --          trace ("patStrings = " ++ show patStrings) $
 --          trace ("mFixity = " ++ show mFixity) $
       case (headMaybe pns, mFixity) of
         -- first identifier in the patterns is the fun.symbol?
-        (Just y, _) | x == y -> True -- trace ("couldBe since y = " ++ show y) $ True
+        (Just y, _) | x == y -> True -- trace ("couldBe since y = " ++ prettyShow y) $ True
         -- are the parts of x contained in p
         _ | xStrings `isSublistOf` patStrings -> True -- trace ("couldBe since isSublistOf") $ True
         -- looking for a mixfix fun.symb
@@ -1163,7 +1285,7 @@ niceDeclarations ds = do
         notAllowedInMutual Axiom{} = False
         notAllowedInMutual d       = declKind d == OtherDecl
         -- Pull type signatures to the top
-        (sigs, other) = partition isTypeSig ds
+        (sigs, other) = List.partition isTypeSig ds
         isTypeSig Axiom{}                     = True
         isTypeSig d | LoneSig{} <- declKind d = True
         isTypeSig _                           = False
@@ -1177,7 +1299,7 @@ niceDeclarations ds = do
         -- do not termination check a mutual block if any of its
         -- inner declarations comes with a {-# NO_TERMINATION_CHECK #-}
         termCheck (FunSig _ _ _ _ _ _ _ tc _ _)      = tc
-        termCheck (FunDef _ _ _ _ tc _ _)            = tc
+        termCheck (FunDef _ _ _ _ _ tc _ _)          = tc
         -- ASR (28 December 2015): Is this equation necessary?
         termCheck (NiceMutual _ tc _ _)              = __IMPOSSIBLE__
         termCheck (NiceUnquoteDecl _ _ _ _ _ tc _ _) = tc
@@ -1215,19 +1337,24 @@ niceDeclarations ds = do
     abstractBlock r ds = do
       let (ds', anyChange) = runChange $ mkAbstract ds
           inherited        = r == noRange
-          -- hack to avoid failing on inherited abstract blocks in where clauses
-      if anyChange || inherited then return ds' else throwError $ UselessAbstract r
+      if anyChange then return ds' else do
+        -- hack to avoid failing on inherited abstract blocks in where clauses
+        unless inherited $ niceWarning $ UselessAbstract r
+        return ds -- no change!
 
     privateBlock _ _ [] = return []
     privateBlock r o ds = do
       let (ds', anyChange) = runChange $ mkPrivate o ds
-      if anyChange then return ds' else
-        if o == UserWritten then throwError $ UselessPrivate r else return ds -- no change!
+      if anyChange then return ds' else do
+        when (o == UserWritten) $ niceWarning $ UselessPrivate r
+        return ds -- no change!
 
     instanceBlock _ [] = return []
     instanceBlock r ds = do
       let (ds', anyChange) = runChange $ mapM mkInstance ds
-      if anyChange then return ds' else throwError $ UselessInstance r
+      if anyChange then return ds' else do
+        niceWarning $ UselessInstance r
+        return ds -- no change!
 
     -- Make a declaration eligible for instance search.
     mkInstance :: Updater NiceDeclaration
@@ -1238,7 +1365,7 @@ niceDeclarations ds = do
         NiceUnquoteDecl r f p a i tc x e -> (\ i -> NiceUnquoteDecl r f p a i tc x e) <$> setInstance i
         NiceMutual{}                     -> return d
         NiceFunClause{}                  -> return d
-        FunDef{}                         -> return d
+        FunDef r ds f a i tc x cs        -> (\ i -> FunDef r ds f a i tc x cs) <$> setInstance i
         NiceField{}                      -> return d  -- Field instance are handled by the parser
         PrimitiveFunction{}              -> return d
         NiceUnquoteDef{}                 -> return d
@@ -1297,7 +1424,7 @@ instance MakeAbstract NiceDeclaration where
   mkAbstract d =
     case d of
       NiceMutual r termCheck pc ds     -> NiceMutual r termCheck pc <$> mkAbstract ds
-      FunDef r ds f a tc x cs          -> (\ a -> FunDef r ds f a tc x) <$> mkAbstract a <*> mkAbstract cs
+      FunDef r ds f a i tc x cs        -> (\ a -> FunDef r ds f a i tc x) <$> mkAbstract a <*> mkAbstract cs
       DataDef r f a pc x ps cs         -> (\ a -> DataDef r f a pc x ps) <$> mkAbstract a <*> mkAbstract cs
       RecDef r f a pc x i e c ps cs    -> (\ a -> RecDef r f a pc x i e c ps) <$> mkAbstract a <*> mkAbstract cs
       NiceFunClause r p a termCheck catchall d  -> (\ a -> NiceFunClause r p a termCheck catchall d) <$> mkAbstract a
@@ -1379,7 +1506,7 @@ instance MakePrivate NiceDeclaration where
       NiceImport _ _ _ _ _                     -> return $ d
       -- Andreas, 2016-07-08, issue #2089
       -- we need to propagate 'private' to the named where modules
-      FunDef r ds f a tc x cls                 -> FunDef r ds f a tc x <$> mkPrivate o cls
+      FunDef r ds f a i tc x cls               -> FunDef r ds f a i tc x <$> mkPrivate o cls
       DataDef{}                                -> return $ d
       RecDef{}                                 -> return $ d
       NicePatternSyn _ _ _ _ _                 -> return $ d
@@ -1517,7 +1644,7 @@ notSoNiceDeclarations d =
     NiceDataSig r _ _ _ _ x bs e     -> [DataSig r Inductive x bs e]
     NiceFunClause _ _ _ _ _ d        -> [d]
     FunSig _ _ _ _ i _ rel tc x e    -> inst i [TypeSig rel x e]
-    FunDef _r ds _ _ _ _ _           -> ds
+    FunDef _r ds _ _ _ _ _ _         -> ds
     DataDef r _ _ _ x bs cs          -> [Data r Inductive x bs Nothing $ concatMap notSoNiceDeclarations cs]
     RecDef r _ _ _ x i e c bs ds     -> [Record r x i e (unThing <$> c) bs Nothing $ concatMap notSoNiceDeclarations ds]
       where unThing (ThingWithFixity c _, inst) = (c, inst)
@@ -1545,7 +1672,7 @@ niceHasAbstract d =
     NiceDataSig{}                   -> Nothing
     NiceFunClause _ _ a _ _ _       -> Just a
     FunSig{}                        -> Nothing
-    FunDef _ _ _ a _ _ _            -> Just a
+    FunDef _ _ _ a _ _ _ _          -> Just a
     DataDef _ _ a _ _ _ _           -> Just a
     RecDef _ _ a _ _ _ _ _ _ _      -> Just a
     NicePatternSyn{}                -> Nothing
