@@ -1,6 +1,5 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE GADTs               #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP   #-}
+{-# LANGUAGE GADTs #-}
 
 {-| The parser doesn't know about operators and parses everything as normal
     function application. This module contains the functions that parses the
@@ -26,7 +25,7 @@ import Control.Monad
 import Data.Either (partitionEithers)
 import qualified Data.Foldable as Fold
 import Data.Function
-import Data.List
+import qualified Data.List as List
 import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
@@ -49,16 +48,13 @@ import Agda.Syntax.Scope.Monad
 
 import Agda.TypeChecking.Monad.Base (typeError, TypeError(..), LHSOrPatSyn(..))
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
+import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.State (getScope)
 import Agda.TypeChecking.Monad.Options
 
 import Agda.Utils.Either
 import Agda.Utils.Pretty
-#if MIN_VERSION_base(4,8,0)
-import Agda.Utils.List hiding ( uncons )
-#else
 import Agda.Utils.List
-#endif
 import Agda.Utils.Trie (Trie)
 import qualified Agda.Utils.Trie as Trie
 
@@ -71,8 +67,13 @@ import Agda.Utils.Impossible
 
 -- | Bills the operator parser.
 
-billToParser :: ScopeM a -> ScopeM a
-billToParser = Bench.billTo [Bench.Parsing, Bench.Operators]
+billToParser :: ExprKind -> ScopeM a -> ScopeM a
+billToParser k = Bench.billTo
+  [ Bench.Parsing
+  , case k of
+      IsExpr    -> Bench.OperatorsExpr
+      IsPattern -> Bench.OperatorsPattern
+  ]
 
 ---------------------------------------------------------------------------
 -- * Building the parser
@@ -138,9 +139,9 @@ data ExprKind = IsPattern | IsExpr
 -- and function symbols in scope.
 --
 -- When parsing a pattern we do not use bound names. The effect is
--- that operator parts (that are not constructor parts) can be used as
--- atomic names in the pattern (so they can be rebound). See
--- @test/succeed/OpBind.agda@ for an example.
+-- that unqualified operator parts (that are not constructor parts)
+-- can be used as atomic names in the pattern (so they can be
+-- rebound). See @test/succeed/OpBind.agda@ for an example.
 --
 -- When parsing a pattern we also disallow the use of sections, mainly
 -- because there is little need for sections in patterns. Note that
@@ -258,7 +259,7 @@ buildParsers r flat kind exprNames = do
         -- resulting in a confusing error message claiming that "if"
         -- is not in scope.
 
-        (non, fix) = partition nonfix (filter (and . partsPresent) ops)
+        (non, fix) = List.partition nonfix (filter (and . partsPresent) ops)
 
         cons       = getDefinedNames [ConName, PatternSynName] flat
         conNames   = Set.fromList $
@@ -276,9 +277,11 @@ buildParsers r flat kind exprNames = do
                         concatMap notationNames $
                         filter (or . partsPresent) ops)
 
-        isAtom   x = case kind of
-                       IsExpr    -> not (Set.member x allParts) || Set.member x allNames
-                       IsPattern -> not (Set.member x conParts) || Set.member x conNames
+        isAtom x
+          | kind == IsPattern && not (isQualified x) =
+            not (Set.member x conParts) || Set.member x conNames
+          | otherwise =
+            not (Set.member x allParts) || Set.member x allNames
         -- If string is a part of notation, it cannot be used as an identifier,
         -- unless it is also used as an identifier. See issue 307.
 
@@ -324,8 +327,8 @@ buildParsers r flat kind exprNames = do
         relatedOperators :: [(Integer, [NotationSection])]
         relatedOperators =
           map (\((l, ns) : rest) -> (l, ns ++ concat (map snd rest))) .
-          groupBy ((==) `on` fst) .
-          sortBy (compare `on` fst) .
+          List.groupBy ((==) `on` fst) .
+          List.sortBy (compare `on` fst) .
           mapMaybe (\n -> case level n of
                             Unrelated     -> Nothing
                             r@(Related l) ->
@@ -600,11 +603,9 @@ parseLHS' lhsOrPatSyn top p = do
         [(p,lhs)] -> do reportSDoc "scope.operators" 50 $ return $
                           text "Parsed lhs:" <+> pretty lhs
                         return (lhs, ops)
-        []        -> typeError $ OperatorChangeMessage
-                               $ OperatorInformation ops
+        []        -> typeError $ OperatorInformation ops
                                $ NoParseForLHS lhsOrPatSyn p
-        rs        -> typeError $ OperatorChangeMessage
-                               $ OperatorInformation ops
+        rs        -> typeError $ OperatorInformation ops
                                $ AmbiguousParseForLHS lhsOrPatSyn p $
                        map (fullParen . fst) rs
     where
@@ -663,12 +664,11 @@ classifyPattern conf p =
 
 -- | Parses a left-hand side, and makes sure that it defined the expected name.
 parseLHS :: QName -> Pattern -> ScopeM LHSCore
-parseLHS top p = billToParser $ do
+parseLHS top p = billToParser IsPattern $ do
   (res, ops) <- parseLHS' IsLHS (Just top) p
   case res of
     Right (f, lhs) -> return lhs
-    _ -> typeError $ OperatorChangeMessage
-                   $ OperatorInformation ops
+    _ -> typeError $ OperatorInformation ops
                    $ NoParseForLHS IsLHS p
 
 -- | Parses a pattern.
@@ -679,12 +679,11 @@ parsePatternSyn :: Pattern -> ScopeM Pattern
 parsePatternSyn = parsePatternOrSyn IsPatSyn
 
 parsePatternOrSyn :: LHSOrPatSyn -> Pattern -> ScopeM Pattern
-parsePatternOrSyn lhsOrPatSyn p = billToParser $ do
+parsePatternOrSyn lhsOrPatSyn p = billToParser IsPattern $ do
   (res, ops) <- parseLHS' lhsOrPatSyn Nothing p
   case res of
     Left p -> return p
-    _      -> typeError $ OperatorChangeMessage
-                        $ OperatorInformation ops
+    _      -> typeError $ OperatorInformation ops
                         $ NoParseForLHS lhsOrPatSyn p
 
 -- | Helper function for 'parseLHS' and 'parsePattern'.
@@ -715,12 +714,12 @@ appView p = case p of
 --   for @Data.Nat._+_@ we return the list @[Data,Nat]@.
 qualifierModules :: [QName] -> [[Name]]
 qualifierModules qs =
-  nub $ filter (not . null) $ map (init . qnameParts) qs
+  List.nub $ filter (not . null) $ map (init . qnameParts) qs
 
 -- | Parse a list of expressions into an application.
 parseApplication :: [Expr] -> ScopeM Expr
 parseApplication [e] = return e
-parseApplication es  = billToParser $ do
+parseApplication es  = billToParser IsExpr $ do
     -- Build the parser
     let names = [ q | Ident q <- es ]
         ms    = qualifierModules names
@@ -734,11 +733,9 @@ parseApplication es  = billToParser $ do
           reportSDoc "scope.operators" 50 $ return $
             text "Parsed an operator application:" <+> pretty e
           return e
-        []  -> typeError $ OperatorChangeMessage
-                         $ OperatorInformation ops
+        []  -> typeError $ OperatorInformation ops
                          $ NoParseForApplication es
-        es' -> typeError $ OperatorChangeMessage
-                         $ OperatorInformation ops
+        es' -> typeError $ OperatorInformation ops
                          $ AmbiguousParseForApplication es
                          $ map fullParen es'
 
@@ -747,7 +744,7 @@ parseModuleIdentifier (Ident m) = return m
 parseModuleIdentifier e = typeError $ NotAModuleExpr e
 
 parseRawModuleApplication :: [Expr] -> ScopeM (QName, [NamedArg Expr])
-parseRawModuleApplication es = billToParser $ do
+parseRawModuleApplication es = billToParser IsExpr $ do
     let e : es_args = es
     m <- parseModuleIdentifier e
 
@@ -762,13 +759,11 @@ parseRawModuleApplication es = billToParser $ do
     -- TODO: not sure about forcing
     case {-force $-} parse (parseSections, pArgs p) es_args of
         [as] -> return (m, as)
-        []   -> typeError $ OperatorChangeMessage
-                          $ OperatorInformation ops
+        []   -> typeError $ OperatorInformation ops
                           $ NoParseForApplication es
         ass -> do
           let f = fullParen . foldl (App noRange) (Ident m)
-          typeError $ OperatorChangeMessage
-                    $ OperatorInformation ops
+          typeError $ OperatorInformation ops
                     $ AmbiguousParseForApplication es
                     $ map f ass
 
@@ -803,9 +798,9 @@ fullParen' e = case exprView e of
     AppV e1 (Arg info e2) -> par $ unExprView $ AppV (fullParen' e1) (Arg info e2')
         where
             e2' = case argInfoHiding info of
-                Hidden    -> e2
-                Instance  -> e2
-                NotHidden -> fullParen' <$> e2
+                Hidden     -> e2
+                Instance{} -> e2
+                NotHidden  -> fullParen' <$> e2
     OpAppV x ns es -> par $ unExprView $ OpAppV x ns $ (map . fmap . fmap . fmap . fmap) fullParen' es
     LamV bs e -> par $ unExprView $ LamV bs (fullParen e)
     where
