@@ -74,9 +74,9 @@ lambda = text "\x03bb"
 prettyHiding :: LensHiding a => a -> (Doc -> Doc) -> Doc -> Doc
 prettyHiding a parens =
   case getHiding a of
-    Hidden    -> braces'
-    Instance  -> dbraces
-    NotHidden -> parens
+    Hidden     -> braces'
+    Instance{} -> dbraces
+    NotHidden  -> parens
 
 prettyRelevance :: LensRelevance a => a -> Doc -> Doc
 prettyRelevance a d =
@@ -93,7 +93,6 @@ instance Pretty a => Pretty (WithHiding a) where
 
 instance Pretty Relevance where
   pretty Forced{}   = empty
-  pretty UnusedArg  = empty
   pretty Relevant   = empty
   pretty Irrelevant = text "."
   pretty NonStrict  = text ".."
@@ -137,9 +136,9 @@ instance Pretty Expr where
                 sep [ lambda <+> fsep (map pretty bs) <+> arrow
                     , nest 2 $ pretty e
                     ]
-            AbsurdLam _ NotHidden -> lambda <+> text "()"
-            AbsurdLam _ Instance -> lambda <+> text "{{}}"
-            AbsurdLam _ Hidden -> lambda <+> text "{}"
+            AbsurdLam _ NotHidden  -> lambda <+> text "()"
+            AbsurdLam _ Instance{} -> lambda <+> text "{{}}"
+            AbsurdLam _ Hidden     -> lambda <+> text "{}"
             ExtendedLam _ pes ->
               lambda <+> bracesAndSemicolons (map (\(x,y,z,_) -> prettyClause x y z) pes)
                    where prettyClause lhs rhs wh = sep [ pretty lhs
@@ -286,15 +285,13 @@ instance Pretty LHSCore where
     pretty d : map (parens . pretty) ps ++
     parens (pretty lhscore) : map (parens . pretty) ps'
 
-instance Pretty [Declaration] where
-  pretty = vcat . map pretty
-
 instance Show ModuleApplication where show = show . pretty
 instance Pretty ModuleApplication where
   pretty (SectionApp _ bs e) = fsep (map pretty bs) <+> text "=" <+> pretty e
   pretty (RecordModuleIFS _ rec) = text "=" <+> pretty rec <+> text "{{...}}"
 
 instance Pretty Declaration where
+    prettyList = vcat . map pretty
     pretty d =
         case d of
             TypeSig i x e ->
@@ -311,8 +308,8 @@ instance Pretty Declaration where
                   mkInst InstanceDef    d = sep [ text "instance", nest 2 d ]
                   mkInst NotInstanceDef d = d
 
-                  mkOverlap i d | argInfoOverlappable i = text "overlap" <+> d
-                                | otherwise             = d
+                  mkOverlap i d | isOverlappable i = text "overlap" <+> d
+                                | otherwise        = d
             FunClause lhs rhs wh _ ->
                 sep [ pretty lhs
                     , nest 2 $ pretty rhs
@@ -438,12 +435,8 @@ instance Pretty Pragma where
       hsep [ text "COMPILED_EXPORT", pretty x, text hs ]
     pretty (CompiledTypePragma _ x hs) =
       hsep [ text "COMPILED_TYPE", pretty x, text hs ]
-    pretty (CompiledDeclareDataPragma _ x hs) =
-      hsep [ text "COMPILED_DECLARE_DATA", pretty x, text hs ]
     pretty (CompiledDataPragma _ x hs hcs) =
       hsep $ [text "COMPILED_DATA", pretty x] ++ map text (hs : hcs)
-    pretty (CompiledEpicPragma _ x e) =
-      hsep [ text "COMPILED_EPIC", pretty x, text e ]
     pretty (CompiledJSPragma _ x e) =
       hsep [ text "COMPILED_JS", pretty x, text e ]
     pretty (CompiledUHCPragma _ x e) =
@@ -452,8 +445,14 @@ instance Pretty Pragma where
       hsep $ [ text "COMPILED_DATA_UHC", pretty x] ++ map text (crd : crcs)
     pretty (HaskellCodePragma _ s) =
       vcat (text "HASKELL" : map text (lines s))
+    pretty (CompilePragma _ b x e) =
+      hsep [ text "COMPILE", text b, pretty x, text e ]
+    pretty (ForeignPragma _ b s) =
+      vcat $ text ("FOREIGN " ++ b) : map text (lines s)
     pretty (StaticPragma _ i) =
       hsep $ [text "STATIC", pretty i]
+    pretty (InjectivePragma _ i) =
+      hsep $ [text "INJECTIVE", pretty i]
     pretty (InlinePragma _ i) =
       hsep $ [text "INLINE", pretty i]
     pretty (ImportPragma _ i) =
@@ -462,6 +461,8 @@ instance Pretty Pragma where
       hsep $ [text "IMPORT_UHC", text i]
     pretty (ImpossiblePragma _) =
       hsep $ [text "IMPOSSIBLE"]
+    pretty (EtaPragma _ x) =
+      hsep $ [text "ETA", pretty x]
     pretty (TerminationCheckPragma _ tc) =
       case tc of
         TerminationCheck       -> __IMPOSSIBLE__
@@ -473,7 +474,7 @@ instance Pretty Pragma where
     pretty (DisplayPragma _ lhs rhs) = text "DISPLAY" <+> sep [ pretty lhs <+> text "=", nest 2 $ pretty rhs ]
     pretty (NoPositivityCheckPragma _) = text "NO_POSITIVITY_CHECK"
     pretty (PolarityPragma _ q occs) =
-      hsep (text "NO_POSITIVITY_CHECK" : pretty q : map pretty occs)
+      hsep (text "POLARITY" : pretty q : map pretty occs)
 
 instance Pretty Fixity where
     pretty (Fixity _ Unrelated   _)   = __IMPOSSIBLE__
@@ -484,25 +485,13 @@ instance Pretty Fixity where
             RightAssoc -> "infixr"
             NonAssoc   -> "infix"
 
-instance Pretty Occurrence where
-  pretty Unused    = text "_"
-  pretty Mixed     = text "*"
-  pretty JustNeg   = text "-"
-  pretty JustPos   = text "+"
-  pretty StrictPos = text "++"
-
-  -- No syntax has been assigned to GuardPos.
-
-  pretty GuardPos  = __IMPOSSIBLE__
-
 instance Pretty GenPart where
     pretty (IdPart x)   = text x
     pretty BindHole{}   = underscore
     pretty NormalHole{} = underscore
     pretty WildHole{}   = underscore
 
-instance Pretty Notation where
-    pretty = hcat . map pretty
+    prettyList = hcat . map pretty
 
 instance Pretty Fixity' where
     pretty (Fixity' fix nota _)
@@ -513,17 +502,15 @@ instance Pretty Fixity' where
  -- Andreas 2010-09-24: and in record fields
 instance Pretty a => Pretty (Arg a) where
   prettyPrec p (Arg ai e) = prettyHiding ai id $ prettyPrec p' e
-      where p' | getHiding ai == NotHidden = p
-               | otherwise                 = 0
+      where p' | visible ai = p
+               | otherwise  = 0
 
 instance Pretty e => Pretty (Named_ e) where
     prettyPrec p (Named Nothing e) = prettyPrec p e
     prettyPrec p (Named (Just s) e) = mparens (p > 0) $ sep [ text (rawNameToString $ rangedThing s) <+> text "=", pretty e ]
 
-instance Pretty [Pattern] where
-    pretty = fsep . map pretty
-
 instance Pretty Pattern where
+    prettyList = fsep . map pretty
     pretty p =
         case p of
             IdentP x        -> pretty x
@@ -541,15 +528,18 @@ instance Pretty Pattern where
             QuoteP _        -> text "quote"
             RecP _ fs       -> sep [ text "record", bracesAndSemicolons (map pretty fs) ]
 
-prettyOpApp ::
+prettyOpApp :: forall a .
   Pretty a => QName -> [NamedArg (MaybePlaceholder a)] -> [Doc]
 prettyOpApp q es = merge [] $ prOp ms xs es
   where
+    -- ms: the module part of the name.
     ms = init (qnameParts q)
+    -- xs: the concrete name (alternation of @Id@ and @Hole@)
     xs = case unqualify q of
            Name _ xs -> xs
            NoName{}  -> __IMPOSSIBLE__
 
+    prOp :: [Name] -> [NamePart] -> [NamedArg (MaybePlaceholder a)] -> [(Doc, Maybe PositionInName)]
     prOp ms (Hole : xs) (e : es) = (pretty e, case namedArg e of
                                                 Placeholder p -> Just p
                                                 _             -> Nothing) :
@@ -558,7 +548,9 @@ prettyOpApp q es = merge [] $ prOp ms xs es
     prOp ms (Id x : xs) es       = ( pretty (foldr Qual (QName (Name noRange $ [Id x])) ms)
                                    , Nothing
                                    ) : prOp [] xs es
-    prOp _  []       []          = []
+      -- Qualify the name part with the module.
+      -- We then clear @ms@ such that the following name parts will not be qualified.
+
     prOp _  []       es          = map (\e -> (pretty e, Nothing)) es
 
     -- Section underscores should be printed without surrounding
