@@ -16,6 +16,8 @@ import Data.Maybe
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Traversable
+
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
@@ -26,6 +28,7 @@ import Agda.Syntax.Notation
 
 import Agda.Utils.Lens
 import Agda.Utils.List
+import Agda.Utils.Pretty
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -41,14 +44,14 @@ data Fixity' = Fixity'
       -- ^ Range of the name in the fixity declaration
       --   (used for correct highlighting, see issue #2140).
     }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 instance Eq Fixity' where
   Fixity' f n _ == Fixity' f' n' _ = f == f' && n == n'
 
 -- | Decorating something with @Fixity'@.
 data ThingWithFixity x = ThingWithFixity x Fixity'
-  deriving (Functor, Foldable, Traversable, Typeable, Show)
+  deriving (Functor, Foldable, Traversable, Typeable, Data, Show)
 
 -- | All the notation information related to a name.
 data NewNotation = NewNotation
@@ -215,12 +218,12 @@ data PrecedenceLevel
     -- ^ No fixity declared.
   | Related !Integer
     -- ^ Fixity level declared as the @Integer@.
-  deriving (Eq, Ord, Show, Typeable)
+  deriving (Eq, Ord, Show, Typeable, Data)
 
 -- | Associativity.
 
 data Associativity = NonAssoc | LeftAssoc | RightAssoc
-   deriving (Eq, Ord, Show, Typeable)
+   deriving (Eq, Ord, Show, Typeable, Data)
 
 -- | Fixity of operators.
 
@@ -230,7 +233,7 @@ data Fixity = Fixity
   , fixityLevel :: !PrecedenceLevel
   , fixityAssoc :: !Associativity
   }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 instance Eq Fixity where
   f1 == f2 = compare f1 f2 == EQ
@@ -253,70 +256,114 @@ data Precedence = TopCtx | FunctionSpaceDomainCtx
                 | LeftOperandCtx Fixity | RightOperandCtx Fixity
                 | FunctionCtx | ArgumentCtx | InsideOperandCtx
                 | WithFunCtx | WithArgCtx | DotPatternCtx
-    deriving (Show, Typeable, Eq)
+    deriving (Show, Typeable, Data, Eq)
+
+instance Pretty Precedence where
+  pretty = text . show
+
+-- | When printing we keep track of a stack of precedences in order to be able
+--   to decide whether it's safe to leave out parens around lambdas. An empty
+--   stack is equivalent to `TopCtx`. Invariant: `notElem TopCtx`.
+type PrecedenceStack = [Precedence]
+
+pushPrecedence :: Precedence -> PrecedenceStack -> PrecedenceStack
+pushPrecedence TopCtx _  = []
+pushPrecedence p      ps = p : ps
+
+headPrecedence :: PrecedenceStack -> Precedence
+headPrecedence []      = TopCtx
+headPrecedence (p : _) = p
 
 -- | The precedence corresponding to a possibly hidden argument.
 hiddenArgumentCtx :: Hiding -> Precedence
-hiddenArgumentCtx NotHidden = ArgumentCtx
-hiddenArgumentCtx Hidden    = TopCtx
-hiddenArgumentCtx Instance  = TopCtx
+hiddenArgumentCtx NotHidden  = ArgumentCtx
+hiddenArgumentCtx Hidden     = TopCtx
+hiddenArgumentCtx Instance{} = TopCtx
 
 -- | Do we need to bracket an operator application of the given fixity
 --   in a context with the given precedence.
-opBrackets :: Fixity -> Precedence -> Bool
-opBrackets                   (Fixity _ (Related n1) LeftAssoc)
-           (LeftOperandCtx   (Fixity _ (Related n2) LeftAssoc))  | n1 >= n2 = False
-opBrackets                   (Fixity _ (Related n1) RightAssoc)
-           (RightOperandCtx  (Fixity _ (Related n2) RightAssoc)) | n1 >= n2 = False
-opBrackets f1
-           (LeftOperandCtx  f2) | Related f1 <- fixityLevel f1
-                                , Related f2 <- fixityLevel f2
-                                , f1 > f2 = False
-opBrackets f1
-           (RightOperandCtx f2) | Related f1 <- fixityLevel f1
-                                , Related f2 <- fixityLevel f2
-                                , f1 > f2 = False
-opBrackets _ TopCtx = False
-opBrackets _ FunctionSpaceDomainCtx = False
-opBrackets _ InsideOperandCtx       = False
-opBrackets _ WithArgCtx             = False
-opBrackets _ WithFunCtx             = False
-opBrackets _ _                      = True
+opBrackets :: Fixity -> PrecedenceStack -> Bool
+opBrackets = opBrackets' False
+
+-- | Do we need to bracket an operator application of the given fixity
+--   in a context with the given precedence.
+opBrackets' :: Bool ->   -- Is the last argument a parenless lambda?
+               Fixity -> PrecedenceStack -> Bool
+opBrackets' isLam f ps = brack f (headPrecedence ps)
+  where
+    false = isLam && lamBrackets' False ps -- require more parens for `e >>= λ x → e₁` than `e >>= e₁`
+    brack                        (Fixity _ (Related n1) LeftAssoc)
+               (LeftOperandCtx   (Fixity _ (Related n2) LeftAssoc))  | n1 >= n2 = false
+    brack                        (Fixity _ (Related n1) RightAssoc)
+               (RightOperandCtx  (Fixity _ (Related n2) RightAssoc)) | n1 >= n2 = false
+    brack f1   (LeftOperandCtx  f2) | Related f1 <- fixityLevel f1
+                                    , Related f2 <- fixityLevel f2
+                                    , f1 > f2 = false
+    brack f1   (RightOperandCtx f2) | Related f1 <- fixityLevel f1
+                                    , Related f2 <- fixityLevel f2
+                                    , f1 > f2 = false
+    brack _ TopCtx                 = false
+    brack _ FunctionSpaceDomainCtx = false
+    brack _ InsideOperandCtx       = false
+    brack _ WithArgCtx             = false
+    brack _ WithFunCtx             = false
+    brack _ _                      = True
 
 -- | Does a lambda-like thing (lambda, let or pi) need brackets in the
 -- given context? A peculiar thing with lambdas is that they don't
--- need brackets in certain right operand contexts. However, we insert
--- brackets anyway, for the following reasons:
---
--- * Clarity.
---
--- * Sometimes brackets are needed. Example: @m₁ >>= (λ x → x) >>= m₂@
---   (here @_>>=_@ is left associative).
-lamBrackets :: Precedence -> Bool
-lamBrackets TopCtx = False
-lamBrackets _      = True
+-- need brackets in certain right operand contexts. To decide we need to look
+-- at the stack of precedences and not just the current precedence.
+-- Example: @m₁ >>= (λ x → x) >>= m₂@ (for @_>>=_@ left associative).
+-- The first argument is the parenthesis preference. If True, we add
+-- parentheses also in right operand contexts where they aren't strictly
+-- needed.
+lamBrackets' :: Bool -> PrecedenceStack -> Bool
+lamBrackets' _      []       = False
+lamBrackets' strict (p : ps) = case p of
+  TopCtx                 -> __IMPOSSIBLE__
+  ArgumentCtx            -> strict || lamBrackets' strict ps
+  RightOperandCtx{}      -> strict || lamBrackets' strict ps
+  FunctionSpaceDomainCtx -> True
+  LeftOperandCtx{}       -> True
+  FunctionCtx            -> True
+  InsideOperandCtx       -> True
+  WithFunCtx             -> True
+  WithArgCtx             -> True
+  DotPatternCtx          -> True
+
+-- | Preference for adding parentheses.
+lamBrackets :: PrecedenceStack -> Bool
+lamBrackets = lamBrackets' True
 
 -- | Does a function application need brackets?
-appBrackets :: Precedence -> Bool
-appBrackets ArgumentCtx   = True
-appBrackets DotPatternCtx = True
-appBrackets _             = False
+appBrackets :: PrecedenceStack -> Bool
+appBrackets = appBrackets' False
+
+-- | Does a function application need brackets?
+appBrackets' :: Bool ->   -- Is the argument of the application a parenless lambda?
+                PrecedenceStack -> Bool
+appBrackets' isLam ps = brack (headPrecedence ps)
+  where
+    brack ArgumentCtx   = True
+    brack DotPatternCtx = True
+    brack _             = isLam && lamBrackets' False ps -- allow e + e₁ λ x → e₂
 
 -- | Does a with application need brackets?
-withAppBrackets :: Precedence -> Bool
-withAppBrackets TopCtx                 = False
-withAppBrackets FunctionSpaceDomainCtx = False
-withAppBrackets WithFunCtx             = False
-withAppBrackets _                      = True
+withAppBrackets :: PrecedenceStack -> Bool
+withAppBrackets = brack . headPrecedence
+  where
+    brack TopCtx                 = False
+    brack FunctionSpaceDomainCtx = False
+    brack WithFunCtx             = False
+    brack _                      = True
 
 -- | Does a function space need brackets?
-piBrackets :: Precedence -> Bool
-piBrackets TopCtx   = False
-piBrackets _        = True
+piBrackets :: PrecedenceStack -> Bool
+piBrackets [] = False
+piBrackets _  = True
 
-roundFixBrackets :: Precedence -> Bool
-roundFixBrackets DotPatternCtx = True
-roundFixBrackets _ = False
+roundFixBrackets :: PrecedenceStack -> Bool
+roundFixBrackets ps = DotPatternCtx == headPrecedence ps
 
 instance HasRange Fixity where
   getRange = fixityRange
