@@ -1,6 +1,5 @@
-{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE UndecidableInstances #-}
 
 {-| An info object contains additional information about a piece of abstract
     syntax that isn't part of the actual syntax. For instance, it might contain
@@ -12,6 +11,7 @@ module Agda.Syntax.Info where
 
 import Prelude hiding (null)
 
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import qualified Agda.Syntax.Concrete.Name as C
@@ -34,7 +34,7 @@ data MetaInfo = MetaInfo
   , metaNumber         :: Maybe MetaId
   , metaNameSuggestion :: String
   }
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 emptyMetaInfo :: MetaInfo
 emptyMetaInfo = MetaInfo
@@ -55,7 +55,7 @@ instance KillRange MetaInfo where
  --------------------------------------------------------------------------}
 
 newtype ExprInfo = ExprRange Range
-  deriving (Typeable, Show, Eq, Null)
+  deriving (Typeable, Data, Show, Eq, Null)
 
 exprNoRange :: ExprInfo
 exprNoRange = ExprRange noRange
@@ -66,6 +66,35 @@ instance HasRange ExprInfo where
 instance KillRange ExprInfo where
   killRange (ExprRange r) = exprNoRange
 
+{--------------------------------------------------------------------------
+    Lambda information
+ --------------------------------------------------------------------------}
+
+-- | Information about lambdas.
+data LamInfo = LamInfo { lamRange  :: Range
+                       , lamOrigin :: Origin
+                       , lamParens :: Bool    -- ^ Do we prefer the lambda with or without parens?
+                       }
+  deriving (Typeable, Data, Show, Eq, Ord)
+
+-- | Default is system inserted and prefer parens.
+defaultLamInfo :: Range -> LamInfo
+defaultLamInfo r = LamInfo{ lamRange = r, lamOrigin = Inserted, lamParens = True }
+
+-- | `LamInfo` with no range information.
+defaultLamInfo_ :: LamInfo
+defaultLamInfo_ = defaultLamInfo noRange
+
+instance HasRange LamInfo where
+  getRange = lamRange
+
+instance KillRange LamInfo where
+  killRange (LamInfo r o p) = LamInfo (killRange r) o p
+
+instance LensOrigin LamInfo where
+  getOrigin = lamOrigin
+  setOrigin o i = i { lamOrigin = o }
+
 {--------------------------------------------------------------------------
     Module information
  --------------------------------------------------------------------------}
@@ -81,9 +110,9 @@ data ModuleInfo = ModuleInfo
   , minfoDirective :: Maybe ImportDirective
     -- ^ Retained for @abstractToConcrete@ of 'ModuleMacro'.
   }
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq)
 
-deriving instance (Show OpenShortHand, Show ImportDirective) => Show ModuleInfo
+deriving instance Show ModuleInfo
 
 instance HasRange ModuleInfo where
   getRange = minfoRange
@@ -99,7 +128,7 @@ instance KillRange ModuleInfo where
 ---------------------------------------------------------------------------
 
 newtype LetInfo = LetRange Range
-  deriving (Typeable, Show, Eq, Null)
+  deriving (Typeable, Data, Show, Eq, Null)
 
 instance HasRange LetInfo where
   getRange (LetRange r)   = r
@@ -119,7 +148,7 @@ data DefInfo = DefInfo
   , defMacro    :: IsMacro
   , defInfo     :: DeclInfo
   }
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 mkDefInfo :: Name -> Fixity' -> Access -> IsAbstract -> Range -> DefInfo
 mkDefInfo x f a ab r = DefInfo f a ab NotInstanceDef NotMacroDef (DeclInfo x r)
@@ -145,7 +174,7 @@ data DeclInfo = DeclInfo
   { declName  :: Name
   , declRange :: Range
   }
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 instance HasRange DeclInfo where
   getRange = declRange
@@ -165,7 +194,7 @@ data MutualInfo = MutualInfo
   , mutualPositivityCheck :: PositivityCheck
   , mutualRange           :: Range
   }
-  deriving (Typeable, Show, Eq)
+  deriving (Typeable, Data, Show, Eq)
 
 -- | Default value for 'MutualInfo'.
 instance Null MutualInfo where
@@ -182,7 +211,7 @@ instance KillRange MutualInfo where
  --------------------------------------------------------------------------}
 
 newtype LHSInfo = LHSRange Range
-  deriving (Typeable, Show, Eq, Null)
+  deriving (Typeable, Data, Show, Eq, Null)
 
 instance HasRange LHSInfo where
   getRange (LHSRange r) = r
@@ -197,7 +226,7 @@ instance KillRange LHSInfo where
 -- | For a general pattern we remember the source code position.
 newtype PatInfo
   = PatRange Range
-  deriving (Typeable, Eq, Null, Show, HasRange, KillRange)
+  deriving (Typeable, Data, Eq, Null, Show, HasRange, KillRange)
 
 -- | Empty range for patterns.
 patNoRange :: PatInfo
@@ -210,7 +239,7 @@ data ConPatInfo = ConPatInfo
     ---  Or from a user written constructor or record pattern?
   , patInfo     :: PatInfo
   }
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq)
 
 instance Show ConPatInfo where
   show (ConPatInfo po i) = applyWhen (po == ConOSystem) ("implicit " ++) $ show i
