@@ -3,6 +3,8 @@
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE TemplateHaskell            #-}
+{-# LANGUAGE UndecidableInstances       #-}  -- because of shortcomings of FunctionalDependencies
+
 #if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
 #endif
@@ -26,12 +28,8 @@ import qualified Data.List as List
 import Data.Maybe
 import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, Sum(..))
 
--- base-4.7 defines the Num instance for Sum
-#if !(MIN_VERSION_base(4,7,0))
-import Data.Orphans             ()
-#endif
-
 import Data.Traversable
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
@@ -42,11 +40,6 @@ import Agda.Syntax.Abstract.Name
 
 import Agda.Utils.Empty
 
--- See Issue 1593.
-#if !MIN_VERSION_transformers(0,4,1)
-import Agda.Utils.Except ( Error(noMsg) )
-#endif
-
 import Agda.Utils.Functor
 import Agda.Utils.Geniplate
 import Agda.Utils.Lens
@@ -78,7 +71,7 @@ data ConHead = ConHead
                               --   Empty list for data constructors.
                               --   'Arg' is not needed here since it
                               --   is stored in the constructor args.
-  } deriving (Typeable)
+  } deriving (Typeable, Data, Show)
 
 instance Eq ConHead where
   (==) = (==) `on` conName
@@ -86,8 +79,8 @@ instance Eq ConHead where
 instance Ord ConHead where
   (<=) = (<=) `on` conName
 
-instance Show ConHead where
-  show (ConHead c i fs) = show c ++ "(" ++ show i ++ ")" ++ show fs
+instance Pretty ConHead where
+  pretty = pretty . conName
 
 instance HasRange ConHead where
   getRange = getRange . conName
@@ -129,7 +122,7 @@ data Term = Var {-# UNPACK #-} !Int Elims -- ^ @x es@ neutral
             --   version of the irrelevance axiom @.irrAx : .A -> A@.
           | Shared !(Ptr Term)
             -- ^ Explicit sharing
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 type ConInfo = ConOrigin
 
@@ -138,11 +131,19 @@ type ConInfo = ConOrigin
 data Elim' a
   = Apply (Arg a)         -- ^ Application.
   | Proj ProjOrigin QName -- ^ Projection.  'QName' is name of a record projection.
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
 
 type Elim = Elim' Term
 type Elims = [Elim]  -- ^ eliminations ordered left-to-right.
 
+-- | This instance cheats on 'Proj', use with care.
+--   'Proj's are always assumed to be 'UserWritten', since they have no 'ArgInfo'.
+instance LensOrigin (Elim' a) where
+  getOrigin (Apply a)   = getOrigin a
+  getOrigin Proj{}      = UserWritten
+  mapOrigin f (Apply a) = Apply $ mapOrigin f a
+  mapOrigin f e@Proj{}  = e
+
 -- | Names in binders and arguments.
 type ArgName = String
 
@@ -166,7 +167,7 @@ data Abs a = Abs   { absName :: ArgName, unAbs :: a }
                -- ^ The body has (at least) one free variable.
                --   Danger: 'unAbs' doesn't shift variables properly
            | NoAbs { absName :: ArgName, unAbs :: a }
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable)
 
 instance Decoration Abs where
   traverseF f (Abs   x a) = Abs   x <$> f a
@@ -175,7 +176,7 @@ instance Decoration Abs where
 -- | Types are terms with a sort annotation.
 --
 data Type' a = El { _getSort :: Sort, unEl :: a }
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
 
 type Type = Type' Term
 
@@ -202,32 +203,10 @@ instance LensSort a => LensSort (Abs a) where
 --   and so on.
 data Tele a = EmptyTel
             | ExtendTel a (Abs (Tele a))  -- ^ 'Abs' is never 'NoAbs'.
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
 
 type Telescope = Tele (Dom Type)
 
--- | A traversal for the names in a telescope.
-mapAbsNamesM :: Applicative m => (ArgName -> m ArgName) -> Tele a -> m (Tele a)
-mapAbsNamesM f EmptyTel                  = pure EmptyTel
-mapAbsNamesM f (ExtendTel a (  Abs x b)) = ExtendTel a <$> (  Abs <$> f x <*> mapAbsNamesM f b)
-mapAbsNamesM f (ExtendTel a (NoAbs x b)) = ExtendTel a <$> (NoAbs <$> f x <*> mapAbsNamesM f b)
-  -- Ulf, 2013-11-06: Last case is really impossible but I'd rather find out we
-  --                  violated that invariant somewhere other than here.
-
-mapAbsNames :: (ArgName -> ArgName) -> Tele a -> Tele a
-mapAbsNames f = runIdentity . mapAbsNamesM (Identity . f)
-
--- Ulf, 2013-11-06
--- The record parameter is named "" inside the record module so we can avoid
--- printing it (issue 208), but we don't want that to show up in the type of
--- the functions in the module (issue 892). This function is used on the record
--- module telescope before adding it to a type in
--- TypeChecking.Monad.Signature.addConstant (to handle functions defined in
--- record modules) and TypeChecking.Rules.Record.checkProjection (to handle
--- record projections).
-replaceEmptyName :: ArgName -> Tele a -> Tele a
-replaceEmptyName x = mapAbsNames $ \ y -> if null y then x else y
-
 -- | Sorts.
 --
 data Sort
@@ -240,19 +219,19 @@ data Sort
     --   If the free variable occurs in the second sort,
     --   the whole thing should reduce to Inf,
     --   otherwise it's the normal lub.
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 -- | A level is a maximum expression of 0..n 'PlusLevel' expressions
 --   each of which is a number or an atom plus a number.
 --
 --   The empty maximum is the canonical representation for level 0.
 newtype Level = Max [PlusLevel]
-  deriving (Show, Typeable)
+  deriving (Show, Typeable, Data)
 
 data PlusLevel
   = ClosedLevel Integer     -- ^ @n@, to represent @Setₙ@.
   | Plus Integer LevelAtom  -- ^ @n + ℓ@.
-  deriving (Show, Typeable)
+  deriving (Show, Typeable, Data)
 
 -- | An atomic term of type @Level@.
 data LevelAtom
@@ -264,7 +243,7 @@ data LevelAtom
     -- ^ A neutral term of type @Level@.
   | UnreducedLevel Term
     -- ^ Introduced by 'instantiate', removed by 'reduce'.
-  deriving (Show, Typeable)
+  deriving (Show, Typeable, Data)
 
 ---------------------------------------------------------------------------
 -- * Blocked Terms
@@ -289,7 +268,7 @@ data NotBlocked
   | ReallyNotBlocked
     -- ^ Reduction was not blocked, we reached a whnf
     --   which can be anything but a stuck @'Def'@.
-  deriving (Show, Typeable)
+  deriving (Show, Typeable, Data)
 
 -- | 'ReallyNotBlocked' is the unit.
 --   'MissingClauses' is dominant.
@@ -306,7 +285,7 @@ instance Semigroup NotBlocked where
 
 instance Monoid NotBlocked where
   -- ReallyNotBlocked is neutral
-  mempty = ReallyNotBlocked
+  mempty  = ReallyNotBlocked
   mappend = (<>)
 
 -- | Something where a meta variable may block reduction.
@@ -340,12 +319,6 @@ instance Monoid Blocked_ where
   mempty = notBlocked ()
   mappend = (<>)
 
--- See issues 1573 and 1674.
-#if !MIN_VERSION_transformers(0,4,1)
-instance Error Blocked_ where
-  noMsg = mempty
-#endif
-
 -- | When trying to reduce @f es@, on match failed on one
 --   elimination @e ∈ es@ that came with info @r :: NotBlocked@.
 --   @stuckOn e r@ produces the new @NotBlocked@ info.
@@ -398,7 +371,8 @@ stuckOn e r =
 --  For the purpose of the permutation and the body dot patterns count
 --  as variables. TODO: Change this!
 data Clause = Clause
-    { clauseRange     :: Range
+    { clauseLHSRange  :: Range
+    , clauseFullRange :: Range
     , clauseTel       :: Telescope
       -- ^ @Δ@: The types of the pattern variables in dependency order.
     , namedClausePats :: [NamedArg DeBruijnPattern]
@@ -413,14 +387,19 @@ data Clause = Clause
       --   pattern on the lhs.
     , clauseCatchall  :: Bool
       -- ^ Clause has been labelled as CATCHALL.
+    , clauseUnreachable :: Maybe Bool
+      -- ^ Clause has been labelled as unreachable by the coverage checker.
+      --   @Nothing@ means coverage checker has not run yet (clause may be unreachable).
+      --   @Just False@ means clause is not unreachable.
+      --   @Just True@ means clause is unreachable.
     }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 clausePats :: Clause -> [Arg DeBruijnPattern]
 clausePats = map (fmap namedThing) . namedClausePats
 
 instance HasRange Clause where
-  getRange = clauseRange
+  getRange = clauseLHSRange
 
 -- | Pattern variables.
 type PatVarName = ArgName
@@ -445,11 +424,15 @@ data Pattern' x
   | ConP ConHead ConPatternInfo [NamedArg (Pattern' x)]
     -- ^ @c ps@
     --   The subpatterns do not contain any projection copatterns.
+  | AbsurdP (Pattern' x)
+    -- ^ @()@
+    --   The argument is to keep track of the original pattern
+    --   (before the absurd match).
   | LitP Literal
     -- ^ E.g. @5@, @"hello"@.
   | ProjP ProjOrigin QName
     -- ^ Projection copattern.  Can only appear by itself.
-  deriving (Typeable, Show, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Show, Functor, Foldable, Traversable)
 
 type Pattern = Pattern' PatVarName
     -- ^ The @PatVarName@ is a name suggestion.
@@ -461,7 +444,7 @@ varP = VarP
 data DBPatVar = DBPatVar
   { dbPatVarName  :: PatVarName
   , dbPatVarIndex :: Int
-  } deriving (Typeable, Show)
+  } deriving (Typeable, Data, Show)
 
 type DeBruijnPattern = Pattern' DBPatVar
 
@@ -491,7 +474,7 @@ data ConPatternInfo = ConPatternInfo
     --   plugin (like Agsy).
     --   Needed e.g. for with-clause stripping.
   }
-  deriving (Typeable, Show)
+  deriving (Typeable, Data, Show)
 
 noConPatternInfo :: ConPatternInfo
 noConPatternInfo = ConPatternInfo Nothing Nothing
@@ -507,17 +490,29 @@ fromConPatternInfo = fromMaybe ConOSystem . conPRecord
 
 -- | Extract pattern variables in left-to-right order.
 --   A 'DotP' is also treated as variable (see docu for 'Clause').
-patternVars :: Arg (Pattern' a) -> [Arg (Either a Term)]
-patternVars (Arg i (VarP x)     ) = [Arg i $ Left x]
-patternVars (Arg i (DotP t)     ) = [Arg i $ Right t]
-patternVars (Arg i (ConP _ _ ps)) = List.concat $ map (patternVars . fmap namedThing) ps
-patternVars (Arg i (LitP l)     ) = []
-patternVars (Arg i ProjP{}      ) = []
+class PatternVars a b | b -> a where
+  patternVars :: b -> [Arg (Either a Term)]
+
+instance PatternVars a (Arg (Pattern' a)) where
+  -- patternVars :: Arg (Pattern' a) -> [Arg (Either a Term)]
+  patternVars (Arg i (VarP x)     ) = [Arg i $ Left x]
+  patternVars (Arg i (DotP t)     ) = [Arg i $ Right t]
+  patternVars (Arg i (AbsurdP p)  ) = patternVars (Arg i p)
+  patternVars (Arg _ (ConP _ _ ps)) = patternVars ps
+  patternVars (Arg _ (LitP _)     ) = []
+  patternVars (Arg _ ProjP{}      ) = []
+
+instance PatternVars a (NamedArg (Pattern' a)) where
+  patternVars = patternVars . fmap namedThing
+
+instance PatternVars a b => PatternVars a [b] where
+  patternVars = concatMap patternVars
 
 -- | Does the pattern perform a match that could fail?
 properlyMatching :: DeBruijnPattern -> Bool
-properlyMatching (VarP x) = isAbsurdPatternName $ dbPatVarName x
+properlyMatching (VarP x) = False
 properlyMatching DotP{} = False
+properlyMatching AbsurdP{} = True
 properlyMatching LitP{} = True
 properlyMatching (ConP _ ci ps) = isNothing (conPRecord ci) || -- not a record cons
   List.any (properlyMatching . namedArg) ps  -- or one of subpatterns is a proper m
@@ -539,8 +534,8 @@ data Substitution' a
     -- ^ Identity substitution.
     --   @Γ ⊢ IdS : Γ@
 
-  | EmptyS
-    -- ^ Empty substitution, lifts from the empty context.
+  | EmptyS Empty
+    -- ^ Empty substitution, lifts from the empty context. First argument is @__IMPOSSIBLE__@.
     --   Apply this to closed terms you want to use in a non-empty context.
     --   @Γ ⊢ EmptyS : ()@
 
@@ -580,7 +575,15 @@ data Substitution' a
     --     Γ, Ψρ ⊢ Lift |Ψ| ρ : Δ, Ψ
     --   @
 
-  deriving (Show, Functor, Foldable, Traversable)
+  deriving ( Show
+           , Functor
+           , Foldable
+           , Traversable
+           , Data
+#if __GLASGOW_HASKELL__ <= 708
+           , Typeable
+#endif
+           )
 
 type Substitution = Substitution' Term
 type PatternSubstitution = Substitution' DeBruijnPattern
@@ -603,7 +606,7 @@ data EqualityView
   = EqualityType
     { eqtSort  :: Sort     -- ^ Sort of this type.
     , eqtName  :: QName    -- ^ Builtin EQUALITY.
-    , eqtLevel :: Arg Term -- ^ Hidden
+    , eqtParams :: [Arg Term] -- ^ Hidden.  Empty or @Level@.
     , eqtType  :: Arg Term -- ^ Hidden
     , eqtLhs   :: Arg Term -- ^ NotHidden
     , eqtRhs   :: Arg Term -- ^ NotHidden
@@ -749,12 +752,79 @@ levelSuc (Max as) = Max $ map inc as
 mkType :: Integer -> Sort
 mkType n = Type $ Max [ClosedLevel n | n > 0]
 
+isSort :: Term -> Maybe Sort
+isSort v = case ignoreSharing v of
+  Sort s -> Just s
+  _      -> Nothing
+
 impossibleTerm :: String -> Int -> Term
 impossibleTerm file line = Lit $ LitString noRange $ unlines
   [ "An internal error has occurred. Please report this as a bug."
   , "Location of the error: " ++ file ++ ":" ++ show line
   ]
 
+hackReifyToMeta :: Term
+hackReifyToMeta = DontCare $ Lit $ LitNat noRange (-42)
+
+isHackReifyToMeta :: Term -> Bool
+isHackReifyToMeta (DontCare (Lit (LitNat r (-42)))) = r == noRange
+isHackReifyToMeta _ = False
+
+---------------------------------------------------------------------------
+-- * Telescopes.
+---------------------------------------------------------------------------
+
+-- | A traversal for the names in a telescope.
+mapAbsNamesM :: Applicative m => (ArgName -> m ArgName) -> Tele a -> m (Tele a)
+mapAbsNamesM f EmptyTel                  = pure EmptyTel
+mapAbsNamesM f (ExtendTel a (  Abs x b)) = ExtendTel a <$> (  Abs <$> f x <*> mapAbsNamesM f b)
+mapAbsNamesM f (ExtendTel a (NoAbs x b)) = ExtendTel a <$> (NoAbs <$> f x <*> mapAbsNamesM f b)
+  -- Ulf, 2013-11-06: Last case is really impossible but I'd rather find out we
+  --                  violated that invariant somewhere other than here.
+
+mapAbsNames :: (ArgName -> ArgName) -> Tele a -> Tele a
+mapAbsNames f = runIdentity . mapAbsNamesM (Identity . f)
+
+-- Ulf, 2013-11-06
+-- The record parameter is named "" inside the record module so we can avoid
+-- printing it (issue 208), but we don't want that to show up in the type of
+-- the functions in the module (issue 892). This function is used on the record
+-- module telescope before adding it to a type in
+-- TypeChecking.Monad.Signature.addConstant (to handle functions defined in
+-- record modules) and TypeChecking.Rules.Record.checkProjection (to handle
+-- record projections).
+replaceEmptyName :: ArgName -> Tele a -> Tele a
+replaceEmptyName x = mapAbsNames $ \ y -> if null y then x else y
+
+-- | Telescope as list.
+type ListTel' a = [Dom (a, Type)]
+type ListTel = ListTel' ArgName
+
+telFromList' :: (a -> ArgName) -> ListTel' a -> Telescope
+telFromList' f = List.foldr extTel EmptyTel
+  where
+    extTel (Dom info (x, a)) = ExtendTel (Dom info a) . Abs (f x)
+
+-- | Convert a list telescope to a telescope.
+telFromList :: ListTel -> Telescope
+telFromList = telFromList' id
+
+-- | Convert a telescope to its list form.
+telToList :: Telescope -> ListTel
+telToList EmptyTel                    = []
+telToList (ExtendTel arg (Abs x tel)) = fmap (x,) arg : telToList tel
+telToList (ExtendTel _    NoAbs{}   ) = __IMPOSSIBLE__
+
+-- | Drop the types from a telescope.
+class TelToArgs a where
+  telToArgs :: a -> [Arg ArgName]
+
+instance TelToArgs ListTel where
+  telToArgs = map $ \ dom -> Arg (domInfo dom) (fst $ unDom dom)
+
+instance TelToArgs Telescope where
+  telToArgs = telToArgs . telToList
+
 -- | Constructing a singleton telescope.
 class SgTel a where
   sgTel :: a -> Telescope
@@ -768,13 +838,6 @@ instance SgTel (Dom (ArgName, Type)) where
 instance SgTel (Dom Type) where
   sgTel dom = sgTel (stringToArgName "_", dom)
 
-hackReifyToMeta :: Term
-hackReifyToMeta = DontCare $ Lit $ LitNat noRange (-42)
-
-isHackReifyToMeta :: Term -> Bool
-isHackReifyToMeta (DontCare (Lit (LitNat r (-42)))) = r == noRange
-isHackReifyToMeta _ = False
-
 ---------------------------------------------------------------------------
 -- * Handling blocked terms.
 ---------------------------------------------------------------------------
@@ -872,13 +935,6 @@ hasElims v =
     DontCare{} -> Nothing
     Shared{}   -> __IMPOSSIBLE__
 
-{- PROBABLY USELESS
-getElims :: Term -> (Elims -> Term, Elims)
-getElims v = maybe default id $ hasElims v
-  where
-    default = (\ [] -> v, [])
--}
-
 -- | Drop 'Apply' constructor. (Unsafe!)
 argFromElim :: Elim' a -> Arg a
 argFromElim (Apply u) = u
@@ -929,8 +985,8 @@ instance Null (Tele a) where
 -- | A 'null' clause is one with no patterns and no rhs.
 --   Should not exist in practice.
 instance Null Clause where
-  empty = Clause empty empty empty empty empty False
-  null (Clause r tel pats body t catchall)
+  empty = Clause empty empty empty empty empty empty False Nothing
+  null (Clause _ _ tel pats body _ _ _)
     =  null tel
     && null pats
     && null body
@@ -1023,7 +1079,7 @@ instance TermSize LevelAtom where
 
 instance TermSize a => TermSize (Substitution' a) where
   tsize IdS                = 1
-  tsize EmptyS             = 1
+  tsize (EmptyS _)         = 1
   tsize (Wk _ rho)         = 1 + tsize rho
   tsize (t :# rho)         = 1 + tsize t + tsize rho
   tsize (Strengthen _ rho) = 1 + tsize rho
@@ -1076,7 +1132,7 @@ instance KillRange Sort where
 
 instance KillRange Substitution where
   killRange IdS                  = IdS
-  killRange EmptyS               = EmptyS
+  killRange (EmptyS err)         = EmptyS err
   killRange (Wk n rho)           = killRange1 (Wk n) rho
   killRange (t :# rho)           = killRange2 (:#) t rho
   killRange (Strengthen err rho) = killRange1 (Strengthen err) rho
@@ -1093,12 +1149,14 @@ instance KillRange a => KillRange (Pattern' a) where
     case p of
       VarP x           -> killRange1 VarP x
       DotP v           -> killRange1 DotP v
+      AbsurdP p        -> killRange1 AbsurdP p
       ConP con info ps -> killRange3 ConP con info ps
       LitP l           -> killRange1 LitP l
       ProjP o q        -> killRange1 (ProjP o) q
 
 instance KillRange Clause where
-  killRange (Clause r tel ps body t catchall) = killRange6 Clause r tel ps body t catchall
+  killRange (Clause rl rf tel ps body t catchall unreachable) =
+    killRange8 Clause rl rf tel ps body t catchall unreachable
 
 instance KillRange a => KillRange (Tele a) where
   killRange = fmap killRange
@@ -1133,7 +1191,7 @@ instance Pretty a => Pretty (Substitution' a) where
     where
     pr p rho = case rho of
       IdS              -> text "idS"
-      EmptyS           -> text "emptyS"
+      EmptyS err       -> text "emptyS"
       t :# rho         -> mparens (p > 2) $ sep [ pr 2 rho P.<> text ",", prettyPrec 3 t ]
       Strengthen _ rho -> mparens (p > 9) $ text "strS" <+> pr 10 rho
       Wk n rho         -> mparens (p > 9) $ text ("wkS " ++ show n) <+> pr 10 rho
@@ -1145,11 +1203,11 @@ instance Pretty Term where
       Var x els -> text ("@" ++ show x) `pApp` els
       Lam ai b   ->
         mparens (p > 0) $
-        sep [ text "λ" <+> prettyHiding ai id (text . show . absName $ b) <+> text "->"
+        sep [ text "λ" <+> prettyHiding ai id (text . absName $ b) <+> text "->"
             , nest 2 $ pretty (unAbs b) ]
       Lit l                -> pretty l
-      Def q els            -> text (show q) `pApp` els
-      Con c ci vs          -> text (show $ conName c) `pApp` map Apply vs
+      Def q els            -> pretty q `pApp` els
+      Con c ci vs          -> pretty (conName c) `pApp` map Apply vs
       Pi a (NoAbs _ b)     -> mparens (p > 0) $
         sep [ prettyPrec 1 (unDom a) <+> text "->"
             , nest 2 $ pretty b ]
@@ -1168,9 +1226,9 @@ instance Pretty Term where
 pDom :: LensHiding a => a -> Doc -> Doc
 pDom i =
   case getHiding i of
-    NotHidden -> parens
-    Hidden    -> braces
-    Instance  -> braces . braces
+    NotHidden  -> parens
+    Hidden     -> braces
+    Instance{} -> braces . braces
 
 instance Pretty Clause where
   pretty Clause{clauseTel = tel, namedClausePats = ps, clauseBody = b, clauseType = t} =
@@ -1223,15 +1281,15 @@ instance Pretty Sort where
       SizeUniv -> text "SizeUniv"
       DLub s b -> mparens (p > 9) $
         text "dlub" <+> prettyPrec 10 s
-                    <+> parens (sep [ text ("λ " ++ show (absName b) ++ " ->")
+                    <+> parens (sep [ text ("λ " ++ absName b ++ " ->")
                                     , nest 2 $ pretty (unAbs b) ])
 
 instance Pretty Type where
   prettyPrec p (El _ a) = prettyPrec p a
 
-instance Pretty Elim where
+instance Pretty tm => Pretty (Elim' tm) where
   prettyPrec p (Apply v)    = prettyPrec p v
-  prettyPrec _ (Proj _o x)  = text ("." ++ show x)
+  prettyPrec _ (Proj _o x)  = text ("." ++ prettyShow x)
 
 instance Pretty DBPatVar where
   prettyPrec _ x = text $ patVarNameToString (dbPatVarName x) ++ "@" ++ show (dbPatVarIndex x)
@@ -1239,8 +1297,9 @@ instance Pretty DBPatVar where
 instance Pretty a => Pretty (Pattern' a) where
   prettyPrec n (VarP x)      = prettyPrec n x
   prettyPrec _ (DotP t)      = text "." P.<> prettyPrec 10 t
+  prettyPrec _ (AbsurdP _)   = text "()"
   prettyPrec n (ConP c i nps)= mparens (n > 0) $
-    text (show $ conName c) <+> fsep (map pretty ps)
+    pretty (conName c) <+> fsep (map pretty ps)
     where ps = map (fmap namedThing) nps
   -- -- Version with printing record type:
   -- prettyPrec _ (ConP c i ps) = (if b then braces else parens) $ prTy $
@@ -1248,8 +1307,8 @@ instance Pretty a => Pretty (Pattern' a) where
   --   where
   --     b = maybe False (== ConOSystem) $ conPRecord i
   --     prTy d = caseMaybe (conPType i) d $ \ t -> d  <+> text ":" <+> pretty t
-  prettyPrec _ (LitP l)      = text (show l)
-  prettyPrec _ (ProjP _o q)  = text ("." ++ show q)
+  prettyPrec _ (LitP l)      = pretty l
+  prettyPrec _ (ProjP _o q)  = text ("." ++ prettyShow q)
 
 -----------------------------------------------------------------------------
 -- * NFData instances
