@@ -1,4 +1,7 @@
-{-# LANGUAGE CPP               #-}
+{-# LANGUAGE CPP          #-}
+{-# LANGUAGE TypeFamilies #-}
+
+-- | Tree traversal for internal syntax.
 
 module Agda.Syntax.Internal.Generic where
 
@@ -9,116 +12,85 @@ import Data.Foldable
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
+-- | Generic term traversal.
+--
+--   Note: ignores sorts in terms!
+--   (Does not traverse into or collect from them.)
+
 class TermLike a where
-  traverseTerm :: (Term -> Term) -> a -> a
-  traverseTermM
+
+  -- | Generic traversal with post-traversal action.
+  --   Ignores sorts.
+  traverseTermM :: (Monad m
 #if __GLASGOW_HASKELL__ <= 708
-    :: (Applicative m, Monad m)
-#else
-    :: Monad m
+    , Applicative m
 #endif
-    => (Term -> m Term) -> a -> m a
+    ) => (Term -> m Term) -> a -> m a
+
+  default traverseTermM :: (Monad m, Traversable f, TermLike b, f b ~ a
+#if __GLASGOW_HASKELL__ <= 708
+    , Applicative m
+#endif
+    ) => (Term -> m Term) -> a -> m a
+  traverseTermM = traverse . traverseTermM
+
+  -- | Generic fold, ignoring sorts.
   foldTerm :: Monoid m => (Term -> m) -> a -> m
 
--- * Constants
+  default foldTerm
+    :: (Monoid m, Foldable f, TermLike b, f b ~ a) => (Term -> m) -> a -> m
+  foldTerm = foldMap . foldTerm
+
+-- Constants
 
 instance TermLike Bool where
-  traverseTerm  _ = id
   traverseTermM _ = pure
   foldTerm _      = mempty
 
 instance TermLike Int where
-  traverseTerm  _ = id
   traverseTermM _ = pure
   foldTerm _      = mempty
 
 instance TermLike Integer where
-  traverseTerm  _ = id
   traverseTermM _ = pure
   foldTerm _      = mempty
 
 instance TermLike Char where
-  traverseTerm  _ = id
   traverseTermM _ = pure
   foldTerm _      = mempty
 
 instance TermLike QName where
-  traverseTerm  _ = id
   traverseTermM _ = pure
   foldTerm _      = mempty
 
--- * Functors
+-- Functors
 
-instance TermLike a => TermLike (Elim' a) where
-  traverseTerm  f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
+instance TermLike a => TermLike (Elim' a)   where
+instance TermLike a => TermLike (Arg a)     where
+instance TermLike a => TermLike (Dom a)     where
+instance TermLike a => TermLike [a]         where
+instance TermLike a => TermLike (Maybe a)   where
+instance TermLike a => TermLike (Abs a)     where
+instance TermLike a => TermLike (Ptr a)     where
+instance TermLike a => TermLike (Blocked a) where
 
-instance TermLike a => TermLike (Arg a) where
-  traverseTerm  f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
-
-instance TermLike a => TermLike (Dom a) where
-  traverseTerm  f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
-
-instance TermLike a => TermLike [a] where
-  traverseTerm f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
-
-instance TermLike a => TermLike (Maybe a) where
-  traverseTerm f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
+-- Tuples
 
 instance (TermLike a, TermLike b) => TermLike (a, b) where
-  traverseTerm f (x, y) = (traverseTerm f x, traverseTerm f y)
   traverseTermM f (x, y) = (,) <$> traverseTermM f x <*> traverseTermM f y
   foldTerm f (x, y) = foldTerm f x `mappend` foldTerm f y
 
 instance (TermLike a, TermLike b, TermLike c) => TermLike (a, b, c) where
-  traverseTerm f (x, y, z) = (traverseTerm f x, traverseTerm f y, traverseTerm f z)
   traverseTermM f (x, y, z) = (,,) <$> traverseTermM f x <*> traverseTermM f y <*> traverseTermM f z
   foldTerm f (x, y, z) = mconcat [foldTerm f x, foldTerm f y, foldTerm f z]
 
 instance (TermLike a, TermLike b, TermLike c, TermLike d) => TermLike (a, b, c, d) where
-  traverseTerm f (x, y, z, u) = (traverseTerm f x, traverseTerm f y, traverseTerm f z, traverseTerm f u)
   traverseTermM f (x, y, z, u) = (,,,) <$> traverseTermM f x <*> traverseTermM f y <*> traverseTermM f z <*> traverseTermM f u
   foldTerm f (x, y, z, u) = mconcat [foldTerm f x, foldTerm f y, foldTerm f z, foldTerm f u]
 
-instance TermLike a => TermLike (Abs a) where
-  traverseTerm f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
-
-instance TermLike a => TermLike (Ptr a) where
-  traverseTerm f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
-
-instance TermLike a => TermLike (Blocked a) where
-  traverseTerm f = fmap (traverseTerm f)
-  traverseTermM f = traverse (traverseTermM f)
-  foldTerm f = foldMap (foldTerm f)
-
--- * Real terms
+-- Real terms
 
 instance TermLike Term where
-  traverseTerm f t = case t of
-    Var i xs    -> f $ Var i $ traverseTerm f xs
-    Def c xs    -> f $ Def c $ traverseTerm f xs
-    Con c ci xs -> f $ Con c ci $ traverseTerm f xs
-    Lam h b     -> f $ Lam h $ traverseTerm f b
-    Pi a b      -> f $ uncurry Pi $ traverseTerm f (a, b)
-    MetaV m xs  -> f $ MetaV m $ traverseTerm f xs
-    Level l     -> f $ Level $ traverseTerm f l
-    Lit _       -> f t
-    Sort _      -> f t
-    DontCare mv -> f $ DontCare $ traverseTerm f mv
-    Shared p    -> f $ Shared $ traverseTerm f p
 
   traverseTermM f t = case t of
     Var i xs    -> f =<< Var i <$> traverseTermM f xs
@@ -147,14 +119,10 @@ instance TermLike Term where
     Shared p    -> foldTerm f p
 
 instance TermLike Level where
-  traverseTerm f  (Max as) = Max $ traverseTerm f as
   traverseTermM f (Max as) = Max <$> traverseTermM f as
   foldTerm f      (Max as) = foldTerm f as
 
 instance TermLike PlusLevel where
-  traverseTerm f l = case l of
-    ClosedLevel{} -> l
-    Plus n l      -> Plus n $ traverseTerm f l
   traverseTermM f l = case l of
     ClosedLevel{} -> return l
     Plus n l      -> Plus n <$> traverseTermM f l
@@ -162,11 +130,6 @@ instance TermLike PlusLevel where
   foldTerm f (Plus _ l)    = foldTerm f l
 
 instance TermLike LevelAtom where
-  traverseTerm f l = case l of
-    MetaLevel m vs   -> MetaLevel m $ traverseTerm f vs
-    NeutralLevel r v -> NeutralLevel r $ traverseTerm f v
-    BlockedLevel m v -> BlockedLevel m $ traverseTerm f v
-    UnreducedLevel v -> UnreducedLevel $ traverseTerm f v
   traverseTermM f l = case l of
     MetaLevel m vs   -> MetaLevel m <$> traverseTermM f vs
     NeutralLevel r v -> NeutralLevel r <$> traverseTermM f v
@@ -179,40 +142,28 @@ instance TermLike LevelAtom where
     UnreducedLevel v -> foldTerm f v
 
 instance TermLike Type where
-  traverseTerm f (El s t) = El s $ traverseTerm f t
   traverseTermM f (El s t) = El s <$> traverseTermM f t
   foldTerm f (El s t) = foldTerm f t
 
 instance TermLike EqualityView where
 
-  traverseTerm f v = case v of
-    OtherType t -> OtherType
-      (traverseTerm f t)
-    EqualityType s eq l t a b -> EqualityType s eq
-      (traverseTerm f l)
-      (traverseTerm f t)
-      (traverseTerm f a)
-      (traverseTerm f b)
-
   traverseTermM f v = case v of
     OtherType t -> OtherType
       <$> traverseTermM f t
     EqualityType s eq l t a b -> EqualityType s eq
-      <$> traverseTermM f l
+      <$> traverse (traverseTermM f) l
       <*> traverseTermM f t
       <*> traverseTermM f a
       <*> traverseTermM f b
 
   foldTerm f v = case v of
     OtherType t -> foldTerm f t
-    EqualityType s eq l t a b -> foldTerm f [l, t, a, b]
+    EqualityType s eq l t a b -> foldTerm f (l ++ [t, a, b])
 
 -- | Put it in a monad to make it possible to do strictly.
-copyTerm
+copyTerm :: (TermLike a, Monad m
 #if __GLASGOW_HASKELL__ <= 708
-  :: (TermLike a, Applicative m, Monad m)
-#else
-  :: (TermLike a, Monad m)
+  , Applicative m
 #endif
-  => a -> m a
+  ) => a -> m a
 copyTerm = traverseTermM return
