@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP                    #-}
+{-# LANGUAGE TypeFamilies           #-}  -- because of type equality ~
 {-# LANGUAGE UndecidableInstances   #-}  -- because of func. deps.
 
 #if __GLASGOW_HASKELL__ <= 708
@@ -11,8 +12,10 @@ import Control.Applicative
 import Control.Monad.State
 
 import Data.Maybe
-import Data.List
-import Data.Traversable (traverse)
+import Data.Monoid
+import qualified Data.List as List
+import Data.Foldable (Foldable, foldMap)
+import Data.Traversable (Traversable, traverse)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Abstract (IsProjP(..))
@@ -79,17 +82,19 @@ class LabelPatVars a b i | b -> i where
   -- ^ Intended, but unpractical due to the absence of type-level lambda, is:
   --   @labelPatVars :: f (Pattern' x) -> State [i] (f (Pattern' (i,x)))@
 
-instance LabelPatVars a b i => LabelPatVars (Arg a) (Arg b) i where
+  default labelPatVars
+    :: (Traversable f, LabelPatVars a' b' i, f a' ~ a, f b' ~ b)
+    => a -> State [i] b
   labelPatVars = traverse labelPatVars
+
+  default unlabelPatVars
+    :: (Traversable f, LabelPatVars a' b' i, f a' ~ a, f b' ~ b)
+    => b -> a
   unlabelPatVars = fmap unlabelPatVars
 
+instance LabelPatVars a b i => LabelPatVars (Arg a) (Arg b) i         where
 instance LabelPatVars a b i => LabelPatVars (Named x a) (Named x b) i where
-  labelPatVars = traverse labelPatVars
-  unlabelPatVars = fmap unlabelPatVars
-
-instance LabelPatVars a b i => LabelPatVars [a] [b] i where
-  labelPatVars = traverse labelPatVars
-  unlabelPatVars = fmap unlabelPatVars
+instance LabelPatVars a b i => LabelPatVars [a] [b] i                 where
 
 instance LabelPatVars Pattern DeBruijnPattern Int where
   labelPatVars p =
@@ -97,6 +102,7 @@ instance LabelPatVars Pattern DeBruijnPattern Int where
       VarP x       -> do i <- next
                          return $ VarP (DBPatVar x i)
       DotP t       -> DotP t <$ next
+      AbsurdP p    -> AbsurdP <$> labelPatVars p
       ConP c mt ps -> ConP c mt <$> labelPatVars ps
       LitP l       -> return $ LitP l
       ProjP o q    -> return $ ProjP o q
@@ -142,12 +148,13 @@ dbPatPerm' countDots ps = Perm (size ixs) <$> picks
   where
     ixs   = concatMap (getIndices . namedThing . unArg) ps
     n     = size $ catMaybes ixs
-    picks = forM (downFrom n) $ \ i -> findIndex (Just i ==) ixs
+    picks = forM (downFrom n) $ \ i -> List.findIndex (Just i ==) ixs
 
     getIndices :: DeBruijnPattern -> [Maybe Int]
     getIndices (VarP x)      = [Just $ dbPatVarIndex x]
     getIndices (ConP c _ ps) = concatMap (getIndices . namedThing . unArg) ps
     getIndices (DotP _)      = [Nothing | countDots]
+    getIndices (AbsurdP p)   = getIndices p
     getIndices (LitP _)      = []
     getIndices ProjP{}       = []
 
@@ -169,6 +176,7 @@ patternToElim (Arg ai (ConP c cpi ps)) = Apply $ Arg ai $ Con c ci $
       map (argFromElim . patternToElim . fmap namedThing) ps
   where ci = fromConPatternInfo cpi
 patternToElim (Arg ai (DotP t)     ) = Apply $ Arg ai t
+patternToElim (Arg ai (AbsurdP p))   = patternToElim $ Arg ai p
 patternToElim (Arg ai (LitP l)     ) = Apply $ Arg ai $ Lit l
 patternToElim (Arg ai (ProjP o dest)) = Proj o dest
 
@@ -183,17 +191,112 @@ patternToTerm p = case patternToElim (defaultArg p) of
   Apply x -> unArg x
   Proj{}  -> __IMPOSSIBLE__
 
+
 class MapNamedArg f where
   mapNamedArg :: (NamedArg a -> NamedArg b) -> NamedArg (f a) -> NamedArg (f b)
 
+-- | Modify the content of @VarP@, and the closest surrounding @NamedArg@.
+--
+--   Note: the @mapNamedArg@ for @Pattern'@ is not expressible simply
+--   by @fmap@ or @traverse@ etc., since @ConP@ has @NamedArg@ subpatterns,
+--   which are taken into account by @mapNamedArg@.
+
 instance MapNamedArg Pattern' where
   mapNamedArg f np =
     case namedArg np of
-      VarP  x     -> map2 VarP $ f $ map2 (const x) np
-      DotP  t     -> map2 (const $ DotP t) np  -- just Haskell type conversion
-      LitP  l     -> map2 (const $ LitP l) np  -- ditto
-      ProjP o q   -> map2 (const $ ProjP o q) np -- ditto
-      ConP c i ps -> map2 (const $ ConP c i $ map (mapNamedArg f) ps) np
+      VarP  x     -> updateNamedArg VarP $ f $ setNamedArg np x
+      AbsurdP p   -> updateNamedArg AbsurdP $ mapNamedArg f $ setNamedArg np p
+      DotP  t     -> setNamedArg np $ DotP t     -- just Haskell type conversion
+      LitP  l     -> setNamedArg np $ LitP l     -- ditto
+      ProjP o q   -> setNamedArg np $ ProjP o q  -- ditto
+      ConP c i ps -> setNamedArg np $ ConP c i $ map (mapNamedArg f) ps
+
+
+-- | Generic pattern traversal.
+--
+--   Pre-applies a pattern modification, recurses, and post-applies another one.
+
+class PatternLike a b where
+
+  -- | Fold pattern.
+  foldrPattern
+    :: Monoid m
+    => (Pattern' a -> m -> m)
+         -- ^ Combine a pattern and the value computed from its subpatterns.
+    -> b -> m
+
+  default foldrPattern
+    :: (Monoid m, Foldable f, PatternLike a p, f p ~ b)
+    => (Pattern' a -> m -> m) -> b -> m
+  foldrPattern = foldMap . foldrPattern
+
+  -- | Traverse pattern.
+  traversePatternM :: (Monad m
+#if __GLASGOW_HASKELL__ <= 708
+    , Applicative m, Functor m
+#endif
+    ) => (Pattern' a -> m (Pattern' a))  -- ^ @pre@: Modification before recursion.
+      -> (Pattern' a -> m (Pattern' a))  -- ^ @post@: Modification after recursion.
+      -> b -> m b
+
+  default traversePatternM :: (Traversable f, PatternLike a p, f p ~ b, Monad m
+#if __GLASGOW_HASKELL__ <= 708
+    , Applicative m, Functor m
+#endif
+    ) => (Pattern' a -> m (Pattern' a))
+      -> (Pattern' a -> m (Pattern' a))
+      -> b -> m b
+  traversePatternM pre post = traverse $ traversePatternM pre post
+
+-- | Compute from each subpattern a value and collect them all in a monoid.
+
+foldPattern :: (PatternLike a b, Monoid m) => (Pattern' a -> m) -> b -> m
+foldPattern f = foldrPattern $ \ p m -> f p `mappend` m
+
+-- | Traverse pattern(s) with a modification before the recursive descent.
+
+preTraversePatternM :: (PatternLike a b, Monad m
+#if __GLASGOW_HASKELL__ <= 708
+  , Applicative m, Functor m
+#endif
+  ) => (Pattern' a -> m (Pattern' a))  -- ^ @pre@: Modification before recursion.
+    -> b -> m b
+preTraversePatternM pre = traversePatternM pre return
+
+-- | Traverse pattern(s) with a modification after the recursive descent.
+
+postTraversePatternM :: (PatternLike a b, Monad m
+#if __GLASGOW_HASKELL__ <= 708
+  , Applicative m, Functor m
+#endif
+  ) => (Pattern' a -> m (Pattern' a))  -- ^ @post@: Modification after recursion.
+    -> b -> m b
+postTraversePatternM = traversePatternM return
+
+-- This is where the action is:
+
+instance PatternLike a (Pattern' a) where
+
+  foldrPattern f p = f p $ case p of
+    ConP _ _ ps -> foldrPattern f ps
+    VarP _      -> mempty
+    AbsurdP _   -> mempty
+    LitP _      -> mempty
+    DotP _      -> mempty
+    ProjP _ _   -> mempty
+
+  traversePatternM pre post = pre >=> recurse >=> post
     where
-    map2 :: (a -> b) -> NamedArg a -> NamedArg b
-    map2 = fmap . fmap
+    recurse p = case p of
+      ConP c ci ps -> ConP c ci <$> traversePatternM pre post ps
+      VarP  _      -> return p
+      LitP  _      -> return p
+      DotP  _      -> return p
+      AbsurdP _    -> return p
+      ProjP _ _    -> return p
+
+-- Boilerplate instances:
+
+instance PatternLike a b => PatternLike a [b]         where
+instance PatternLike a b => PatternLike a (Arg b)     where
+instance PatternLike a b => PatternLike a (Named x b) where
