@@ -1,6 +1,5 @@
 {-# LANGUAGE CPP                 #-}
 {-# LANGUAGE DeriveDataTypeable  #-}
-{-# LANGUAGE ScopedTypeVariables #-}
 
 {-| As a concrete name, a notation is a non-empty list of alternating 'IdPart's and holes.
     In contrast to concrete names, holes can be binders.
@@ -19,8 +18,10 @@ import Control.Applicative
 import Control.DeepSeq
 import Control.Monad
 
-import Data.List
+import qualified Data.List as List
 import Data.Maybe
+
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Common
@@ -58,7 +59,7 @@ data GenPart
   | WildHole !Int
     -- ^ An underscore in binding position.
   | IdPart RawName
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 instance KillRange GenPart where
   killRange p = case p of
@@ -187,8 +188,8 @@ mkNotation holes ids = do
       -- The hole names are the keys of the @holeMap@.
       uniqueHoleNames = distinct [ x | (x, _) <- holeMap, x /= "_" ]
 
-      isExprLinear   xs = sort [ i | x <- xs, isNormalHole x, let Just i = holeTarget x ] == holeNumbers
-      isLambdaLinear xs = sort [ x | BindHole x <- xs ] ==
+      isExprLinear   xs = List.sort [ i | x <- xs, isNormalHole x, let Just i = holeTarget x ] == holeNumbers
+      isLambdaLinear xs = List.sort [ x | BindHole x <- xs ] ==
                           [ i | (i, h) <- numberedHoles,
                                 LambdaHole x _ <- [namedArg h], x /= "_" ]
 
