@@ -28,7 +28,8 @@ import Control.Monad ((>=>), forM_)
 import Control.Monad.State
 import Control.Monad.Reader
 import Control.Monad.Writer hiding ((<>))
-import Data.List
+
+import qualified Data.List as List
 import Data.Typeable ( Typeable )
 
 import Agda.Syntax.Position
@@ -42,12 +43,12 @@ import Agda.Syntax.Concrete.Definitions
 import Agda.Syntax.Parser.Tokens
 
 import Agda.Utils.Except
-  ( Error(strMsg)
-  , ExceptT
+  ( ExceptT
   , MonadError(catchError, throwError)
   , runExceptT
   )
 import Agda.Utils.FileName
+import Agda.Utils.IO.UTF8 (readTextFile)
 import qualified Agda.Utils.Maybe.Strict as Strict
 import Agda.Utils.Pretty
 
@@ -120,7 +121,7 @@ parseLiterateWithoutComments p layers = parseStringFromFile (literateSrcFile lay
 parseLiterateWithComments :: LiterateParser [Token]
 parseLiterateWithComments p layers = do
   code <- map Left <$> parseLiterateWithoutComments p layers
-  let literate = Right <$> filter (not . isCode) layers
+  let literate = Right <$> filter (not . isCodeLayer) layers
   let (terms, overlaps) = interleaveRanges code literate
   forM_ (map fst overlaps) $ \c ->
     warning$ OverlappingTokensWarning { warnRange = getRange c }
@@ -133,7 +134,7 @@ parseLiterateWithComments p layers = do
                    | m <- terms ]
 
 readFilePM :: AbsolutePath -> PM String
-readFilePM path = wrapIOM (ReadFileError path) (readFile (filePath path))
+readFilePM path = wrapIOM (ReadFileError path) (readTextFile (filePath path))
 
 parseLiterateFile :: Processor -> Parser a -> AbsolutePath -> PM a
 parseLiterateFile po p path = readFilePM path >>= parseLiterate p p . po (startPos (Just path))
@@ -147,7 +148,7 @@ parseFileExts = ".agda":literateExts
 
 parseFile' :: (Show a) => Parser a -> AbsolutePath -> PM a
 parseFile' p file =
-  if ".agda" `isSuffixOf` filePath file then
+  if ".agda" `List.isSuffixOf` filePath file then
     Agda.Syntax.Parser.parseFile p file
   else
     go literateProcessors
@@ -156,7 +157,7 @@ parseFile' p file =
                      errPath = file
                    , errValidExts = parseFileExts
                    }
-    go ((ext, po):pos) | ext `isSuffixOf` filePath file = parseLiterateFile po p file
+    go ((ext, po):pos) | ext `List.isSuffixOf` filePath file = parseLiterateFile po p file
     go (_:pos) = go pos
 
 ------------------------------------------------------------------------
