@@ -33,7 +33,9 @@ module Agda.Syntax.Parser.Monad
 
 import Control.Exception (catch)
 import Data.Int
+
 import Data.Typeable ( Typeable )
+import Data.Data (Data)
 
 import Control.Monad.State
 import Control.Applicative
@@ -41,10 +43,6 @@ import Control.Applicative
 import Agda.Syntax.Position
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
-#if !(MIN_VERSION_mtl(2,2,1))
-import Agda.Utils.Except ( Error(noMsg) )
-#endif
-
 import Agda.Utils.FileName
 import qualified Agda.Utils.IO.UTF8 as UTF8
 import qualified Agda.Utils.Maybe.Strict as Strict
@@ -137,7 +135,7 @@ data ParseWarning =
   OverlappingTokensWarning
   { warnRange    :: !(Range' SrcFile)
                     -- ^ The range of the bigger overlapping token
-  }
+  } deriving (Typeable, Data)
 
 -- | The result of parsing something.
 data ParseResult a  = ParseOk ParseState a
@@ -197,7 +195,7 @@ instance Pretty ParseError where
   pretty InvalidExtensionError{errPath,errValidExts} = vcat
       [ pretty errPath <> colon <+>
         text "Unsupported extension."
-      , text "Supported extensions are:" <+> prettyList errValidExts
+      , text "Supported extensions are:" <+> prettyList_ errValidExts
       ]
   pretty ReadFileError{errPath,errIOError} = vcat
       [ text "Cannot read file" <+> pretty errPath
@@ -214,12 +212,6 @@ instance HasRange ParseError where
   getRange ReadFileError{errPath} = posToRange p p
     where p = startPos (Just errPath)
 
-#if !(MIN_VERSION_mtl(2,2,1))
--- Stupid ErrorT!
-instance Error ParseError where
-  noMsg = __IMPOSSIBLE__
-#endif
-
 instance Show ParseWarning where
   show = prettyShow
 
