@@ -1,28 +1,29 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE DeriveDataTypeable  #-}
-{-# LANGUAGE GADTs               #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP                #-}
+{-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE GADTs              #-}
 
 {-| This module defines the notion of a scope and operations on scopes.
 -}
 module Agda.Syntax.Scope.Base where
 
+import Prelude hiding (null)
+
 import Control.Arrow (first, second, (***))
-import Control.Applicative
+import Control.Applicative hiding (empty)
 import Control.DeepSeq
 import Control.Monad
 
 import Data.Either (partitionEithers)
 import Data.Function
-import Data.List as List
+import qualified Data.List as List
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Maybe
-import Data.Typeable (Typeable)
 
--- import Debug.Trace (trace)
+import Data.Typeable (Typeable)
+import Data.Data (Data)
 
 import Agda.Benchmarking
 
@@ -38,6 +39,7 @@ import qualified Agda.Utils.AssocList as AssocList
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
+import Agda.Utils.Null
 import Agda.Utils.Pretty
 import qualified Agda.Utils.Map as Map
 
@@ -53,9 +55,9 @@ data Scope = Scope
       , scopeParents        :: [A.ModuleName]
       , scopeNameSpaces     :: ScopeNameSpaces
       , scopeImports        :: Map C.QName A.ModuleName
-      , scopeDatatypeModule :: Bool
+      , scopeDatatypeModule :: Maybe DataOrRecord
       }
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq, Show)
 
 -- | See 'Agda.Syntax.Common.Access'.
 data NameSpaceId
@@ -65,7 +67,7 @@ data NameSpaceId
   | OnlyQualifiedNS  -- ^ Visible (as qualified) from outside,
                      --   but not exported when opening the module.
                      --   Used for qualified constructors.
-  deriving (Typeable, Eq, Bounded, Enum)
+  deriving (Typeable, Data, Eq, Bounded, Enum, Show)
 
 type ScopeNameSpaces = [(NameSpaceId, NameSpace)]
 
@@ -98,12 +100,12 @@ data ScopeInfo = ScopeInfo
       { scopeCurrent       :: A.ModuleName
       , scopeModules       :: Map A.ModuleName Scope
       , scopeLocals        :: LocalVars
-      , scopePrecedence    :: Precedence
+      , scopePrecedence    :: PrecedenceStack
       , scopeInverseName   :: Map A.QName [C.QName]
       , scopeInverseModule :: Map A.ModuleName [C.QName]
       , scopeInScope       :: InScopeSet
       }
-  deriving Typeable
+  deriving (Typeable, Data, Show)
 
 instance Eq ScopeInfo where
   ScopeInfo c1 m1 l1 p1 _ _ _ == ScopeInfo c2 m2 l2 p2 _ _ _ =
@@ -115,13 +117,16 @@ type LocalVars = AssocList C.Name LocalVar
 -- | A local variable can be shadowed by an import.
 --   In case of reference to a shadowed variable, we want to report
 --   a scope error.
-data LocalVar
-  = LocalVar    { localVar :: A.Name }
+data LocalVar = LocalVar
+  { localVar        :: A.Name
     -- ^ Unique ID of local variable.
-  | ShadowedVar { localVar :: A.Name, localShadowedBy :: [AbstractName] }
-    -- ^ This local variable is shadowed by one or more imports.
-    --   (List not empty).
-  deriving (Typeable)
+  , localLetBound   :: Bool
+    -- ^ Flag whether the variable is introduced by a @let@.
+  , localShadowedBy :: [AbstractName]
+     -- ^ If this list is not empty, the local variable is
+     --   shadowed by one or more imports.
+  }
+  deriving (Typeable, Data, Show)
 
 instance Eq LocalVar where
   (==) = (==) `on` localVar
@@ -130,20 +135,19 @@ instance Ord LocalVar where
   compare = compare `on` localVar
 
 -- | We show shadowed variables as prefixed by a ".", as not in scope.
-instance Show LocalVar where
-  show (LocalVar    x)    = show x
-  show (ShadowedVar x xs) = "." ++ show x
+instance Pretty LocalVar where
+  pretty (LocalVar x _ []) = pretty x
+  pretty (LocalVar x _ xs) = text "." <> pretty x
 
 -- | Shadow a local name by a non-empty list of imports.
 shadowLocal :: [AbstractName] -> LocalVar -> LocalVar
 shadowLocal [] _ = __IMPOSSIBLE__
-shadowLocal ys (LocalVar    x   ) = ShadowedVar x ys
-shadowLocal ys (ShadowedVar x zs) = ShadowedVar x (ys ++ zs)
+shadowLocal ys (LocalVar x b zs) = LocalVar x b (ys ++ zs)
 
 -- | Project name of unshadowed local variable.
 notShadowedLocal :: LocalVar -> Maybe A.Name
-notShadowedLocal (LocalVar x) = Just x
-notShadowedLocal ShadowedVar{} = Nothing
+notShadowedLocal (LocalVar x _ []) = Just x
+notShadowedLocal _ = Nothing
 
 -- | Get all locals that are not shadowed __by imports__.
 notShadowedLocals :: LocalVars -> AssocList C.Name A.Name
@@ -175,7 +179,7 @@ data NameSpace = NameSpace
         -- ^ Maps concrete module names to a list of abstract module names.
       , nsInScope :: InScopeSet
       }
-  deriving (Typeable, Eq)
+  deriving (Typeable, Data, Eq, Show)
 
 type ThingsInScope a = Map C.Name [a]
 type NamesInScope    = ThingsInScope AbstractName
@@ -223,7 +227,7 @@ data KindOfName
   | PatternSynName -- ^ Name of a pattern synonym.
   | MacroName      -- ^ Name of a macro
   | QuotableName   -- ^ A name that can only be quoted.
-  deriving (Eq, Show, Typeable, Enum, Bounded)
+  deriving (Eq, Show, Typeable, Data, Enum, Bounded)
 
 -- | A list containing all name kinds.
 allKindsOfNames :: [KindOfName]
@@ -240,7 +244,7 @@ data WhyInScope
     -- ^ Imported from another module.
   | Applied C.QName WhyInScope
     -- ^ Imported by a module application.
-  deriving (Typeable)
+  deriving (Typeable, Data, Show)
 
 -- | A decoration of 'Agda.Syntax.Abstract.Name.QName'.
 data AbstractName = AbsName
@@ -251,7 +255,7 @@ data AbstractName = AbsName
   , anameLineage :: WhyInScope
     -- ^ Explanation where this name came from.
   }
-  deriving (Typeable)
+  deriving (Typeable, Data, Show)
 
 -- | A decoration of abstract syntax module names.
 data AbstractModule = AbsModule
@@ -260,7 +264,7 @@ data AbstractModule = AbsModule
   , amodLineage :: WhyInScope
     -- ^ Explanation where this name came from.
   }
-  deriving (Typeable)
+  deriving (Typeable, Data, Show)
 
 instance Eq AbstractName where
   (==) = (==) `on` anameName
@@ -285,7 +289,7 @@ lensAmodName f am = f (amodName am) <&> \ m -> am { amodName = m }
 -- * Operations on name and module maps.
 
 mergeNames :: Eq a => ThingsInScope a -> ThingsInScope a -> ThingsInScope a
-mergeNames = Map.unionWith union
+mergeNames = Map.unionWith List.union
 
 ------------------------------------------------------------------------
 -- * Operations on name spaces
@@ -339,7 +343,7 @@ emptyScope = Scope
   , scopeParents        = []
   , scopeNameSpaces     = [ (nsid, emptyNameSpace) | nsid <- [minBound..maxBound] ]
   , scopeImports        = Map.empty
-  , scopeDatatypeModule = False
+  , scopeDatatypeModule = Nothing
   }
 
 -- | The empty scope info.
@@ -348,7 +352,7 @@ emptyScopeInfo = ScopeInfo
   { scopeCurrent       = noModuleName
   , scopeModules       = Map.singleton noModuleName emptyScope
   , scopeLocals        = []
-  , scopePrecedence    = TopCtx
+  , scopePrecedence    = []
   , scopeInverseName   = Map.empty
   , scopeInverseModule = Map.empty
   , scopeInScope       = Set.empty
@@ -405,11 +409,14 @@ zipScope :: (NameSpaceId -> NamesInScope   -> NamesInScope   -> NamesInScope  )
             (NameSpaceId -> InScopeSet     -> InScopeSet     -> InScopeSet    ) ->
             Scope -> Scope -> Scope
 zipScope fd fm fs s1 s2 =
-  s1 { scopeNameSpaces = [ (nsid, zipNS nsid ns1 ns2)
-                         | ((nsid, ns1), (nsid', ns2)) <- zipWith' (,) (scopeNameSpaces s1) (scopeNameSpaces s2)
-                         , assert (nsid == nsid')
-                         ]
-     , scopeImports  = Map.union (scopeImports s1) (scopeImports s2)
+  s1 { scopeNameSpaces =
+         [ (nsid, zipNS nsid ns1 ns2)
+         | ((nsid, ns1), (nsid', ns2)) <-
+             fromMaybe __IMPOSSIBLE__ $
+               zipWith' (,) (scopeNameSpaces s1) (scopeNameSpaces s2)
+         , assert (nsid == nsid')
+         ]
+     , scopeImports = Map.union (scopeImports s1) (scopeImports s2)
      }
   where
     assert True  = True
@@ -714,7 +721,9 @@ scopeLookup :: InScope a => C.QName -> ScopeInfo -> [a]
 scopeLookup q scope = map fst $ scopeLookup' q scope
 
 scopeLookup' :: forall a. InScope a => C.QName -> ScopeInfo -> [(a, Access)]
-scopeLookup' q scope = nubBy ((==) `on` fst) $ findName q root ++ maybeToList topImports ++ imports
+scopeLookup' q scope =
+  List.nubBy ((==) `on` fst) $
+    findName q root ++ maybeToList topImports ++ imports
   where
 
     -- 1. Finding a name in the current scope and its parents.
@@ -824,7 +833,7 @@ inverseScopeLookup' amb name scope = billToPure [ Scoping , InverseScopeLookup ]
     len (C.Qual _ x) = 1 + len x
 
     best :: [C.QName] -> [C.QName]
-    best = sortBy (compare `on` len)
+    best = List.sortBy (compare `on` len)
 
     unique :: forall a . [a] -> Bool
     unique []      = __IMPOSSIBLE__
@@ -922,58 +931,56 @@ inverseScopeLookupModule x = inverseScopeLookup (Left x)
 -- * (Debug) printing
 ------------------------------------------------------------------------
 
-instance Show AbstractName where
-  show = show . anameName
+instance Pretty AbstractName where
+  pretty = pretty . anameName
 
-instance Show AbstractModule where
-  show = show . amodName
+instance Pretty AbstractModule where
+  pretty = pretty . amodName
 
-instance Show NameSpaceId where
-  show nsid = case nsid of
+instance Pretty NameSpaceId where
+  pretty = text . \case
     PublicNS        -> "public"
     PrivateNS       -> "private"
     ImportedNS      -> "imported"
     OnlyQualifiedNS -> "only-qualified"
 
-instance Show NameSpace where
-  show (NameSpace names mods _) =
-    unlines $
-      blockOfLines "names"   (map pr $ Map.toList names) ++
-      blockOfLines "modules" (map pr $ Map.toList mods)
-    where
-      pr :: (Show a, Show b) => (a,b) -> String
-      pr (x, y) = show x ++ " --> " ++ show y
+instance Pretty NameSpace where
+  pretty = vcat . prettyNameSpace
 
-instance Show Scope where
-  show (scope@Scope{ scopeName = name, scopeParents = parents, scopeImports = imps }) =
-    unlines $
-      [ "scope " ++ show name ] ++ ind (
-        concat [ blockOfLines (show nsid) (lines $ show $ scopeNameSpace nsid scope)
+prettyNameSpace :: NameSpace -> [Doc]
+prettyNameSpace (NameSpace names mods _) =
+    blockOfLines (text "names")   (map pr $ Map.toList names) ++
+    blockOfLines (text "modules") (map pr $ Map.toList mods)
+  where
+    pr :: (Pretty a, Pretty b) => (a,b) -> Doc
+    pr (x, y) = pretty x <+> text "-->" <+> pretty y
+
+instance Pretty Scope where
+  pretty (scope@Scope{ scopeName = name, scopeParents = parents, scopeImports = imps }) =
+    vcat $
+      [ text "scope" <+> pretty name ] ++ ind (
+        concat [ blockOfLines (pretty nsid) $ prettyNameSpace $ scopeNameSpace nsid scope
                | nsid <- [minBound..maxBound] ]
-      ++ blockOfLines "imports"  (case Map.keys imps of
-                                    [] -> []
-                                    ks -> [ show ks ]
-                                 )
+      ++ blockOfLines (text "imports")
+           (case Map.keys imps of [] -> []; ks -> [ prettyList ks ])
       )
-    where ind = map ("  " ++)
+    where ind = map $ nest 2
 
 -- | Add first string only if list is non-empty.
-blockOfLines :: String -> [String] -> [String]
+blockOfLines :: Doc -> [Doc] -> [Doc]
 blockOfLines _  [] = []
-blockOfLines hd ss = hd : map ("  " ++) ss
+blockOfLines hd ss = hd : map (nest 2) ss
 
-instance Show ScopeInfo where
-  show (ScopeInfo this mods locals ctx _ _ _) =
-    unlines $
-      [ "ScopeInfo"
-      , "  current = " ++ show this
-      ] ++
-      (if null locals then [] else [ "  locals  = " ++ show locals ]) ++
-      [ "  context = " ++ show ctx
-      , "  modules"
-      ] ++ map ("    "++) (relines . map show $ Map.elems mods)
-    where
-      relines = filter (not . null) . lines . unlines
+instance Pretty ScopeInfo where
+  pretty (ScopeInfo this mods locals ctx _ _ _) = vcat $
+    [ text "ScopeInfo"
+    , text "  current = " <> pretty this
+    ] ++
+    (if null locals then [] else [ text "  locals  = " <> pretty locals ]) ++
+    [ text "  context = " <> pretty ctx
+    , text "  modules"
+    ] ++
+    map (nest 4) (List.filter (not . null) $ map pretty $ Map.elems mods)
 
 ------------------------------------------------------------------------
 -- * Boring instances
