@@ -13,7 +13,7 @@ import Control.Monad hiding (mapM, forM)
 import Control.Monad.Writer hiding (mapM, forM)
 import Control.Monad.State hiding (mapM, forM)
 
-import Data.List as List
+import qualified Data.List as List
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Maybe
@@ -31,6 +31,7 @@ import Agda.Syntax.Concrete as C
 import Agda.Syntax.Scope.Base
 
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Options
 
@@ -55,10 +56,9 @@ type ScopeM = TCM
 
 -- * Errors
 
-isDatatypeModule :: A.ModuleName -> ScopeM Bool
+isDatatypeModule :: A.ModuleName -> ScopeM (Maybe DataOrRecord)
 isDatatypeModule m = do
-   sc <- getScope
-   return $ maybe __IMPOSSIBLE__ scopeDatatypeModule (Map.lookup m (scopeModules sc))
+   scopeDatatypeModule . Map.findWithDefault __IMPOSSIBLE__ m . scopeModules <$> getScope
 
 -- * General operations
 
@@ -90,14 +90,15 @@ getNamedScope m = do
   case Map.lookup m (scopeModules scope) of
     Just s  -> return s
     Nothing -> do
-      reportSLn "" 0 $ "ERROR: In scope\n" ++ show scope ++ "\nNO SUCH SCOPE " ++ show m
+      reportSLn "" 0 $ "ERROR: In scope\n" ++ prettyShow scope ++ "\nNO SUCH SCOPE " ++ prettyShow m
       __IMPOSSIBLE__
 
 getCurrentScope :: ScopeM Scope
 getCurrentScope = getNamedScope =<< getCurrentModule
 
--- | Create a new module with an empty scope (Bool is True if it is a datatype module)
-createModule :: Bool -> A.ModuleName -> ScopeM ()
+-- | Create a new module with an empty scope.
+--   (@Just@ if it is a datatype or record module.)
+createModule :: Maybe DataOrRecord -> A.ModuleName -> ScopeM ()
 createModule b m = do
   reportSLn "scope.createModule" 10 $ "createModule " ++ prettyShow m
   s <- getCurrentScope
@@ -139,18 +140,17 @@ modifyCurrentNameSpace :: NameSpaceId -> (NameSpace -> NameSpace) -> ScopeM ()
 modifyCurrentNameSpace acc f = modifyCurrentScope $ updateScopeNameSpaces $
   AssocList.updateAt acc f
 
-setContextPrecedence :: Precedence -> ScopeM ()
-setContextPrecedence p = modifyScope_ $ \s -> s { scopePrecedence = p }
+pushContextPrecedence :: Precedence -> ScopeM ()
+pushContextPrecedence p = modifyScope_ $ \ s -> s { scopePrecedence = pushPrecedence p $ scopePrecedence s }
 
-getContextPrecedence :: ScopeM Precedence
-getContextPrecedence = scopePrecedence <$> getScope
+popContextPrecedence :: ScopeM ()
+popContextPrecedence = modifyScope_ $ \ s -> s { scopePrecedence = drop 1 $ scopePrecedence s }
 
 withContextPrecedence :: Precedence -> ScopeM a -> ScopeM a
 withContextPrecedence p m = do
-  p' <- getContextPrecedence
-  setContextPrecedence p
+  pushContextPrecedence p
   x <- m
-  setContextPrecedence p'
+  popContextPrecedence
   return x
 
 getLocalVars :: ScopeM LocalVars
@@ -201,15 +201,37 @@ freshAbstractQName fx x = do
 -- * Resolving names
 
 data ResolvedName
-  = VarName A.Name
-  | DefinedName Access AbstractName
-  | FieldName [AbstractName]
-      -- ^ record fields names need to be distinguished to parse copatterns
-  | ConstructorName [AbstractName]
-  | PatternSynResName AbstractName
-  | UnknownName
+  =   -- | Local variable bound by λ, Π, module telescope, pattern, @let@.
+    VarName
+    { resolvedVar      :: A.Name
+    , resolvedLetBound :: Bool    -- ^ Variable bound by @let@?
+    }
+
+  |   -- | Function, data/record type, postulate.
+    DefinedName Access AbstractName
+
+  |   -- | Record field name.  Needs to be distinguished to parse copatterns.
+    FieldName [AbstractName]
+
+  |   -- | Data or record constructor name.
+    ConstructorName [AbstractName]
+
+  |   -- | Name of pattern synonym.
+    PatternSynResName AbstractName
+
+  |   -- | Unbound name.
+    UnknownName
   deriving (Show, Eq)
 
+instance Pretty ResolvedName where
+  pretty = \case
+    VarName x _         -> text "variable"    <+> pretty x
+    DefinedName a x     -> pretty a           <+> pretty x
+    FieldName xs        -> text "field"       <+> pretty xs
+    ConstructorName xs  -> text "constructor" <+> pretty xs
+    PatternSynResName x -> text "pattern"     <+> pretty x
+    UnknownName         -> text "<unknown name>"
+
 -- | Look up the abstract name referred to by a given concrete name.
 resolveName :: C.QName -> ScopeM ResolvedName
 resolveName = resolveName' allKindsOfNames Nothing
@@ -225,16 +247,16 @@ resolveName' ::
 resolveName' kinds names x = do
   scope <- getScope
   let vars     = AssocList.mapKeysMonotonic C.QName $ scopeLocals scope
-      retVar y = return $ VarName $ y { nameConcrete = unqualify x }
+      retVar y = return . VarName y{ nameConcrete = unqualify x }
       aName    = A.qnameName . anameName
   case lookup x vars of
     -- Case: we have a local variable x.
-    Just (LocalVar y)  -> retVar y
+    Just (LocalVar y b []) -> retVar y b
     -- Case: ... but is (perhaps) shadowed by some imports.
-    Just (ShadowedVar y ys) -> case names of
+    Just (LocalVar y b ys) -> case names of
       Nothing -> shadowed ys
-      Just ns -> case filter (\y -> aName y `Set.member` ns) ys of
-        [] -> retVar y
+      Just ns -> case filter (\ y -> aName y `Set.member` ns) ys of
+        [] -> retVar y b
         ys -> shadowed ys
       where
       shadowed ys =
@@ -272,7 +294,7 @@ resolveModule :: C.QName -> ScopeM AbstractModule
 resolveModule x = do
   ms <- scopeLookup x <$> getScope
   case ms of
-    [AbsModule m why] -> return $ AbsModule (m `withRangesOfQ` x) why
+    [AbsModule m why] -> return $ AbsModule (m `withRangeOf` x) why
     []                -> typeError $ NoSuchModule x
     ms                -> typeError $ AmbiguousModule x (map amodName ms)
 
@@ -285,7 +307,7 @@ getNotation
 getNotation x ns = do
   r <- resolveName' allKindsOfNames (Just ns) x
   case r of
-    VarName y           -> return $ namesToNotation x y
+    VarName y _         -> return $ namesToNotation x y
     DefinedName _ d     -> return $ notation d
     FieldName ds        -> return $ oneNotation ds
     ConstructorName ds  -> return $ oneNotation ds
@@ -300,9 +322,13 @@ getNotation x ns = do
 
 -- * Binding names
 
--- | Bind a variable. The abstract name is supplied as the second argument.
-bindVariable :: C.Name -> A.Name -> ScopeM ()
-bindVariable x y = modifyScope_ $ updateScopeLocals $ AssocList.insert x $ LocalVar y
+-- | Bind a variable.
+bindVariable
+  :: Bool    -- ^ Let-bound variable?
+  -> C.Name  -- ^ Concrete name.
+  -> A.Name  -- ^ Abstract name.
+  -> ScopeM ()
+bindVariable b x y = modifyScope_ $ updateScopeLocals $ AssocList.insert x $ LocalVar y b []
 
 -- | Bind a defined name. Must not shadow anything.
 bindName :: Access -> KindOfName -> C.Name -> A.QName -> ScopeM ()
@@ -314,7 +340,7 @@ bindName acc kind x y = do
     UnknownName   | isNoName x -> success
     DefinedName{} | isNoName x -> success <* modifyCurrentScope (removeNameFromScope PrivateNS x)
     DefinedName _ d     -> clash $ anameName d
-    VarName z           -> clash $ A.qualify (mnameFromList []) z
+    VarName z _         -> clash $ A.qualify (mnameFromList []) z
     FieldName       ds  -> ambiguous FldName ds
     ConstructorName ds  -> ambiguous ConName ds
     PatternSynResName n -> clash $ anameName n
@@ -380,8 +406,8 @@ copyScope oldc new0 s = (inScopeBecause (Applied oldc) *** memoToScopeInfo) <$>
   where
     copy :: A.ModuleName -> Scope -> WSM Scope
     copy new s = do
-      lift $ reportSLn "scope.copy" 20 $ "Copying scope " ++ show old ++ " to " ++ show new
-      lift $ reportSLn "scope.copy" 50 $ show s
+      lift $ reportSLn "scope.copy" 20 $ "Copying scope " ++ prettyShow old ++ " to " ++ prettyShow new
+      lift $ reportSLn "scope.copy" 50 $ prettyShow s
       s0 <- lift $ getNamedScope new
       -- Delete private names, then copy names and modules. Recompute inScope
       -- set rather than trying to copy it.
@@ -454,7 +480,7 @@ copyScope oldc new0 s = (inScopeBecause (Applied oldc) *** memoToScopeInfo) <$>
           -- For now, we just set their range
           -- to the new module name's one, which fixes issue 1619.
           y <- setRange rnew . A.qualify m <$> refresh (qnameName x)
-          lift $ reportSLn "scope.copy" 50 $ "  Copying " ++ show x ++ " to " ++ show y
+          lift $ reportSLn "scope.copy" 50 $ "  Copying " ++ prettyShow x ++ " to " ++ prettyShow y
           addName x y
           return y
 
@@ -491,9 +517,9 @@ copyScope oldc new0 s = (inScopeBecause (Applied oldc) *** memoToScopeInfo) <$>
             -- Andreas, Jesper, 2015-07-02: Issue 1597
             -- Don't copy a module over itself, it will just be emptied of its contents.
             if (x == y) then return x else do
-            lift $ reportSLn "scope.copy" 50 $ "  Copying module " ++ show x ++ " to " ++ show y
+            lift $ reportSLn "scope.copy" 50 $ "  Copying module " ++ prettyShow x ++ " to " ++ prettyShow y
             addMod x y rec
-            lift $ createModule False y
+            lift $ createModule Nothing y
             -- We need to copy the contents of included modules recursively (only when 'rec')
             when rec $ copyRec x y
             return y
@@ -522,7 +548,7 @@ applyImportDirectiveM m dir@ImportDirective{ impRenaming = ren, using = u, hidin
 
     -- Check for duplicate imports in a single import directive.
     -- @dup@ : To be imported names that are mentioned more than once.
-    let dup = targetNames \\ nub targetNames
+    let dup = targetNames List.\\ List.nub targetNames
     unless (null dup) $ typeError $ DuplicateImports m dup
 
     -- Apply the import directive.
@@ -565,7 +591,7 @@ applyImportDirectiveM m dir@ImportDirective{ impRenaming = ren, using = u, hidin
         (Using xs, ys) -> do
           let uselessHiding = [ x | x@ImportedName{} <- ys ] ++
                               [ x | x@(ImportedModule y) <- ys, ImportedName y `notElem` names ]
-          unless (null uselessHiding) $ typeError $ GenericError $ "Hiding " ++ intercalate ", " (map show uselessHiding)
+          unless (null uselessHiding) $ typeError $ GenericError $ "Hiding " ++ List.intercalate ", " (map prettyShow uselessHiding)
                                                                 ++ " has no effect"
           return dir{ hiding = [] }
         _ -> return dir
@@ -605,7 +631,7 @@ applyImportDirectiveM m dir@ImportDirective{ impRenaming = ren, using = u, hidin
     doesntExport = do
       -- Names @xs@ mentioned in the import directive @dir@ but not in the @scope@.
       let xs = filter doesntExist names
-      reportSLn "scope.import.apply" 20 $ "non existing names: " ++ show xs
+      reportSLn "scope.import.apply" 20 $ "non existing names: " ++ prettyShow xs
       typeError $ ModuleDoesntExport m xs
 
     doesntExist (ImportedName   x) = isNothing $ Map.lookup x namesInScope
@@ -679,7 +705,7 @@ openModule_ cm dir = do
     locals <- mapMaybe (\ (c,x) -> c <$ notShadowedLocal x) <$> getLocalVars
     let newdefs = Map.keys $ nsNames ns
         shadowed = List.intersect locals newdefs
-    reportSLn "scope.locals" 10 $ "opening module shadows the following locals vars: " ++ show shadowed
+    reportSLn "scope.locals" 10 $ "opening module shadows the following locals vars: " ++ prettyShow shadowed
   -- Andreas, 2014-09-03, issue 1266: shadow local variables by imported defs.
   modifyLocalVars $ AssocList.mapWithKey $ \ c x ->
     case Map.lookup c $ nsNames ns of
