@@ -27,6 +27,7 @@ import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
 import Control.Monad.Reader
+import Control.Monad.State
 
 import qualified Data.Map as Map
 import Data.Maybe
@@ -41,6 +42,7 @@ import Agda.Syntax.Position
 import Agda.Syntax.Literal
 import Agda.Syntax.Info
 import Agda.Syntax.Internal (MetaId(..))
+import qualified Agda.Syntax.Internal as I
 import Agda.Syntax.Fixity
 import Agda.Syntax.Concrete as C
 import Agda.Syntax.Abstract as A
@@ -48,11 +50,13 @@ import Agda.Syntax.Abstract.Views as AV
 import Agda.Syntax.Scope.Base
 
 import Agda.TypeChecking.Monad.State (getScope)
-import Agda.TypeChecking.Monad.Base  (TCM, NamedMeta(..))
+import Agda.TypeChecking.Monad.Base  (TCM, NamedMeta(..), stBuiltinThings, BuiltinThings, Builtin(..))
+import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Options
 
 import qualified Agda.Utils.AssocList as AssocList
 import Agda.Utils.Either
+import Agda.Utils.Function
 import Agda.Utils.Functor
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
@@ -70,10 +74,11 @@ data Env = Env { takenNames   :: Set C.Name
                , currentScope :: ScopeInfo
                }
 
-defaultEnv :: Env
-defaultEnv = Env { takenNames   = Set.empty
-                 , currentScope = emptyScopeInfo
-                 }
+-- -- UNUSED
+-- defaultEnv :: Env
+-- defaultEnv = Env { takenNames   = Set.empty
+--                  , currentScope = emptyScopeInfo
+--                  }
 
 makeEnv :: ScopeInfo -> Env
 makeEnv scope = Env { takenNames   = Set.union vars defs
@@ -83,12 +88,17 @@ makeEnv scope = Env { takenNames   = Set.union vars defs
     vars  = Set.fromList $ map fst $ scopeLocals scope
     defs  = Map.keysSet $ nsNames $ everythingInScope scope
 
-currentPrecedence :: AbsToCon Precedence
+currentPrecedence :: AbsToCon PrecedenceStack
 currentPrecedence = asks $ scopePrecedence . currentScope
 
+withPrecedence' :: PrecedenceStack -> AbsToCon a -> AbsToCon a
+withPrecedence' ps = local $ \e ->
+  e { currentScope = (currentScope e) { scopePrecedence = ps } }
+
 withPrecedence :: Precedence -> AbsToCon a -> AbsToCon a
-withPrecedence p = local $ \e ->
-  e { currentScope = (currentScope e) { scopePrecedence = p } }
+withPrecedence p ret = do
+  ps <- currentPrecedence
+  withPrecedence' (pushPrecedence p ps) ret
 
 withScope :: ScopeInfo -> AbsToCon a -> AbsToCon a
 withScope scope = local $ \e -> e { currentScope = scope }
@@ -96,6 +106,14 @@ withScope scope = local $ \e -> e { currentScope = scope }
 noTakenNames :: AbsToCon a -> AbsToCon a
 noTakenNames = local $ \e -> e { takenNames = Set.empty }
 
+-- | Bind a concrete name to an abstract in the translation environment.
+addBinding :: C.Name -> A.Name -> Env -> Env
+addBinding y x e =
+  e { takenNames   = Set.insert y $ takenNames e
+    , currentScope = (`updateScopeLocals` currentScope e) $
+        AssocList.insert y (LocalVar x __IMPOSSIBLE__ [])
+    }
+
 -- The Monad --------------------------------------------------------------
 
 -- | We put the translation into TCM in order to print debug messages.
@@ -110,10 +128,7 @@ abstractToConcreteEnv :: ToConcrete a c => Env -> a -> TCM c
 abstractToConcreteEnv flags a = runReaderT (toConcrete a) flags
 
 abstractToConcreteCtx :: ToConcrete a c => Precedence -> a -> TCM c
-abstractToConcreteCtx ctx x = do
-  scope <- getScope
-  let scope' = scope { scopePrecedence = ctx }
-  abstractToConcreteEnv (makeEnv scope') x
+abstractToConcreteCtx ctx x = runAbsToCon $ withPrecedence ctx (toConcrete x)
 
 abstractToConcrete_ :: ToConcrete a c => a -> TCM c
 abstractToConcrete_ = runAbsToCon . toConcrete
@@ -140,7 +155,7 @@ lookupQName :: AllowAmbiguousNames -> A.QName -> AbsToCon C.QName
 lookupQName ambCon x = do
   ys <- inverseScopeLookupName' ambCon x <$> asks currentScope
   lift $ reportSLn "scope.inverse" 100 $
-    "inverse looking up abstract name " ++ show x ++ " yields " ++ show ys
+    "inverse looking up abstract name " ++ prettyShow x ++ " yields " ++ prettyShow ys
   case ys of
     (y : _) -> return y
     [] -> do
@@ -165,25 +180,30 @@ lookupModule x =
             []      -> return $ mnameToConcrete x
                 -- this is what happens for names that are not in scope (private names)
 
+-- | Add a abstract name to the scope and produce an available concrete version of it.
 bindName :: A.Name -> (C.Name -> AbsToCon a) -> AbsToCon a
 bindName x ret = do
-  names <- asks takenNames
   let y = nameConcrete x
-  case (Set.member y names) of
-    _ | isNoName y -> ret y
-    True           -> bindName (nextName x) ret
-    False          ->
-        local (\e -> e { takenNames   = Set.insert y $ takenNames e
-                       , currentScope = (`updateScopeLocals` currentScope e) $
-                           AssocList.insert y (LocalVar x)
-                       }
-              ) $ ret y
+  if isNoName y then ret y else do
+    names <- asks takenNames
+    let loop x = do
+          let y = nameConcrete x
+          if y `Set.member` names then loop $ nextName x
+          else local (addBinding y x) $ ret y
+    loop x
+
+
+-- | Like 'bindName', but do not care whether name is already taken.
+bindName' :: A.Name -> AbsToCon a -> AbsToCon a
+bindName' x = applyUnless (isNoName y) $ local $ addBinding y x
+  where y = nameConcrete x
+
 
 -- Dealing with precedences -----------------------------------------------
 
 -- | General bracketing function.
 bracket' ::    (e -> e)             -- ^ the bracketing function
-            -> (Precedence -> Bool) -- ^ Should we bracket things
+            -> (PrecedenceStack -> Bool) -- ^ Should we bracket things
                                     --   which have the given
                                     --   precedence?
             -> e -> AbsToCon e
@@ -192,20 +212,20 @@ bracket' paren needParen e =
         return $ if needParen p then paren e else e
 
 -- | Expression bracketing
-bracket :: (Precedence -> Bool) -> AbsToCon C.Expr -> AbsToCon C.Expr
+bracket :: (PrecedenceStack -> Bool) -> AbsToCon C.Expr -> AbsToCon C.Expr
 bracket par m =
     do  e <- m
         bracket' (Paren (getRange e)) par e
 
 -- | Pattern bracketing
-bracketP_ :: (Precedence -> Bool) -> AbsToCon C.Pattern -> AbsToCon C.Pattern
+bracketP_ :: (PrecedenceStack -> Bool) -> AbsToCon C.Pattern -> AbsToCon C.Pattern
 bracketP_ par m =
     do  e <- m
         bracket' (ParenP (getRange e)) par e
 
 {- UNUSED
 -- | Pattern bracketing
-bracketP :: (Precedence -> Bool) -> (C.Pattern -> AbsToCon a)
+bracketP :: (PrecedenceStack -> Bool) -> (C.Pattern -> AbsToCon a)
                                  -> ((C.Pattern -> AbsToCon a) -> AbsToCon a)
                                  -> AbsToCon a
 bracketP par ret m = m $ \p -> do
@@ -213,6 +233,17 @@ bracketP par ret m = m $ \p -> do
     ret p
 -}
 
+-- | Applications where the argument is a lambda without parentheses need
+--   parens more often than other applications.
+isParenlessLambda :: NamedArg A.Expr -> Bool
+isParenlessLambda e | notVisible e = False
+isParenlessLambda e =
+  case unScope $ namedArg e of
+    A.Lam i _ _           -> not $ lamParens i
+    A.AbsurdLam i _       -> not $ lamParens i
+    A.ExtendedLam i _ _ _ -> not $ lamParens i
+    _                     -> False
+
 -- Dealing with infix declarations ----------------------------------------
 
 -- | If a name is defined with a fixity that differs from the default, we have
@@ -255,6 +286,9 @@ class ToConcrete a c | a -> c where
     toConcrete :: a -> AbsToCon c
     bindToConcrete :: a -> (c -> AbsToCon b) -> AbsToCon b
 
+    -- Christian Sattler, 2017-08-05:
+    -- These default implementations are not valid semantically (at least
+    -- the second one). Perhaps they (it) should be removed.
     toConcrete     x     = bindToConcrete x return
     bindToConcrete x ret = ret =<< toConcrete x
 
@@ -278,23 +312,33 @@ bindToConcreteTop = bindToConcreteCtx TopCtx
 toConcreteHiding :: (LensHiding h, ToConcrete a c) => h -> a -> AbsToCon c
 toConcreteHiding h =
   case getHiding h of
-    NotHidden -> toConcrete
-    Hidden    -> toConcreteTop
-    Instance  -> toConcreteTop
+    NotHidden  -> toConcrete
+    Hidden     -> toConcreteTop
+    Instance{} -> toConcreteTop
 
 -- | Translate something in a context indicated by 'Hiding' info.
 bindToConcreteHiding :: (LensHiding h, ToConcrete a c) => h -> a -> (c -> AbsToCon b) -> AbsToCon b
 bindToConcreteHiding h =
   case getHiding h of
-    NotHidden -> bindToConcrete
-    Hidden    -> bindToConcreteTop
-    Instance  -> bindToConcreteTop
+    NotHidden  -> bindToConcrete
+    Hidden     -> bindToConcreteTop
+    Instance{} -> bindToConcreteTop
 
 -- General instances ------------------------------------------------------
 
 instance ToConcrete a c => ToConcrete [a] [c] where
     toConcrete     = mapM toConcrete
-    bindToConcrete = thread bindToConcrete
+    -- Andreas, 2017-04-11, Issue #2543
+    -- The naive `thread'ing does not work as we have to undo
+    -- changes to the Precedence.
+    -- bindToConcrete = thread bindToConcrete
+    bindToConcrete []     ret = ret []
+    bindToConcrete (a:as) ret = do
+      p <- currentPrecedence  -- save precedence
+      bindToConcrete a $ \ c ->
+        withPrecedence' p $ -- reset precedence
+          bindToConcrete as $ \ cs ->
+            ret (c : cs)
 
 instance (ToConcrete a1 c1, ToConcrete a2 c2) => ToConcrete (Either a1 a2) (Either c1 c2) where
     toConcrete = traverseEither toConcrete toConcrete
@@ -391,8 +435,9 @@ instance ToConcrete A.Expr C.Expr where
 
     toConcrete e@(A.App i e1 e2)    =
         tryToRecoverOpApp e
+        $ tryToRecoverNatural e
         -- or fallback to App
-        $ bracket appBrackets
+        $ bracket (appBrackets' $ isParenlessLambda e2)
         $ do e1' <- toConcreteCtx FunctionCtx e1
              e2' <- toConcreteCtx ArgumentCtx e2
              return $ C.App (getRange i) e1' e2'
@@ -404,9 +449,10 @@ instance ToConcrete A.Expr C.Expr where
         return $ C.WithApp (getRange i) e es
 
     toConcrete (A.AbsurdLam i h) =
-      bracket lamBrackets $ return $ C.AbsurdLam (getRange i) h
+      bracket (lamBrackets' $ lamParens i) $ return $ C.AbsurdLam (getRange i) h
     toConcrete e@(A.Lam i _ _)      =
-        bracket lamBrackets
+        tryToRecoverOpApp e   -- recover sections
+        $ bracket (lamBrackets' $ lamParens i)
         $ case lamView e of
             (bs, e) ->
                 bindToConcrete (map makeDomainFree bs) $ \bs -> do
@@ -425,12 +471,12 @@ instance ToConcrete A.Expr C.Expr where
                     _                              -> ([b], e)
             lamView e = ([], e)
     toConcrete (A.ExtendedLam i di qname cs) =
-        bracket lamBrackets $ do
+        bracket (lamBrackets' $ lamParens i) $ do
           decls <- concat <$> toConcrete cs
           let namedPat np = case getHiding np of
-                 NotHidden -> namedArg np
-                 Hidden    -> C.HiddenP noRange (unArg np)
-                 Instance  -> C.InstanceP noRange (unArg np)
+                 NotHidden  -> namedArg np
+                 Hidden     -> C.HiddenP noRange (unArg np)
+                 Instance{} -> C.InstanceP noRange (unArg np)
               -- we know all lhs are of the form `.extlam p1 p2 ... pn`,
               -- with the name .extlam leftmost. It is our mission to remove it.
           let removeApp (C.RawAppP r (_:es)) = return $ C.RawAppP r es
@@ -473,9 +519,9 @@ instance ToConcrete A.Expr C.Expr where
                            _          -> e
             addDot a e = C.Dot (getRange a) e
             mkArg (Arg info e) = case getHiding info of
-                                          Hidden    -> HiddenArg   (getRange e) (unnamed e)
-                                          Instance  -> InstanceArg (getRange e) (unnamed e)
-                                          NotHidden -> e
+                                          Hidden     -> HiddenArg   (getRange e) (unnamed e)
+                                          Instance{} -> InstanceArg (getRange e) (unnamed e)
+                                          NotHidden  -> e
 
     toConcrete (A.Set i 0)  = return $ C.Set (getRange i)
     toConcrete (A.Set i n)  = return $ C.SetN (getRange i) n
@@ -531,9 +577,18 @@ makeDomainFree b@(A.DomainFull (A.TypedBindings r (Arg info (A.TBind _ [WithHidi
     _ -> b
 makeDomainFree b = b
 
+-- Christian Sattler, 2017-08-05, fixing #2669
+-- Both methods of ToConcrete (FieldAssignment' a) (FieldAssignment' c) need
+-- to be implemented, each in terms of the corresponding one of ToConcrete a c.
+-- This mirrors the instance ToConcrete (Arg a) (Arg c).
+-- The default implementations of ToConcrete are not valid semantically.
 instance ToConcrete a c => ToConcrete (FieldAssignment' a) (FieldAssignment' c) where
     toConcrete = traverse toConcrete
 
+    bindToConcrete (FieldAssignment name a) ret =
+      bindToConcrete a $ ret . FieldAssignment name
+
+
 -- Binder instances -------------------------------------------------------
 
 instance ToConcrete A.LamBinding [C.LamBinding] where
@@ -547,7 +602,7 @@ instance ToConcrete A.TypedBindings [C.TypedBindings] where
     where
       recoverLabels :: Arg A.TypedBinding -> Arg C.TypedBinding -> [Arg C.TypedBinding]
       recoverLabels b cb
-        | getHiding b == NotHidden = [cb]   -- We don't care about labels for explicit args
+        | visible b = [cb]   -- We don't care about labels for explicit args
         | otherwise = traverse (recover (unArg b)) cb
 
       recover (A.TBind _ xs _) (C.TBind r ys e) = tbind r e (zipWith label xs ys)
@@ -577,7 +632,7 @@ instance ToConcrete LetBinding [C.Declaration] where
     bindToConcrete (LetBind i info x t e) ret =
         bindToConcrete x $ \x ->
         do (t,(e, [], [], [])) <- toConcrete (t, A.RHS e Nothing)
-           ret $ addInstanceB (getHiding info == Instance) $
+           ret $ addInstanceB (isInstance info) $
                [ C.TypeSig info x t
                , C.FunClause (C.LHS (C.IdentP $ C.QName x) [] [] [])
                              e C.NoWhere False
@@ -591,14 +646,14 @@ instance ToConcrete LetBinding [C.Declaration] where
       x' <- unqualify <$> toConcrete x
       modapp <- toConcrete modapp
       let r = getRange modapp
-          open = maybe DontOpen id $ minfoOpenShort i
-          dir  = maybe defaultImportDir{ importDirRange = r } id $ minfoDirective i
+          open = fromMaybe DontOpen $ minfoOpenShort i
+          dir  = fromMaybe defaultImportDir{ importDirRange = r } $ minfoDirective i
       -- This is no use since toAbstract LetDefs is in localToAbstract.
       local (openModule' x dir id) $
         ret [ C.ModuleMacro (getRange i) x' modapp open dir ]
     bindToConcrete (LetOpen i x _) ret = do
       x' <- toConcrete x
-      let dir = maybe defaultImportDir id $ minfoDirective i
+      let dir = fromMaybe defaultImportDir $ minfoDirective i
       local (openModule' x dir restrictPrivate) $
             ret [ C.Open (getRange i) x' dir ]
     bindToConcrete (LetDeclaredVariable _) ret =
@@ -652,7 +707,7 @@ instance ToConcrete A.RHS (C.RHS, [C.Expr], [C.Expr], [C.Declaration]) where
     toConcrete A.AbsurdRHS = return (C.AbsurdRHS, [], [], [])
     toConcrete (A.WithRHS _ es cs) = do
       es <- toConcrete es
-      cs <- concat <$> toConcrete cs
+      cs <- noTakenNames $ concat <$> toConcrete cs
       return (C.AbsurdRHS, [], es, cs)
     toConcrete (A.RewriteRHS xeqs rhs wh) = do
       wh <- declsToConcrete wh
@@ -679,7 +734,7 @@ instance ToConcrete (Constr A.Constructor) C.Declaration where
   toConcrete (Constr d) = head <$> toConcrete d
 
 instance ToConcrete a C.LHS => ToConcrete (A.Clause' a) [C.Declaration] where
-  toConcrete (A.Clause lhs _ rhs wh catchall) =
+  toConcrete (A.Clause lhs _ _ rhs wh catchall) =
       bindToConcrete lhs $ \lhs ->
         case lhs of
           C.LHS p wps _ _ -> do
@@ -780,8 +835,8 @@ instance ToConcrete A.Declaration [C.Declaration] where
 
   toConcrete (A.Import i x _) = do
     x <- toConcrete x
-    let open = maybe DontOpen id $ minfoOpenShort i
-        dir  = maybe defaultImportDir id $ minfoDirective i
+    let open = fromMaybe DontOpen $ minfoOpenShort i
+        dir  = fromMaybe defaultImportDir $ minfoDirective i
     return [ C.Import (getRange i) x Nothing open dir]
 
   toConcrete (A.Pragma i p)     = do
@@ -824,18 +879,15 @@ instance ToConcrete RangeAndPragma C.Pragma where
     A.CompiledDataPragma x hs hcs -> do
       x <- toConcrete x
       return $ C.CompiledDataPragma r x hs hcs
-    A.CompiledDeclareDataPragma x hs -> do
-      x <- toConcrete x
-      return $ C.CompiledDeclareDataPragma r x hs
     A.CompiledPragma x hs -> do
       x <- toConcrete x
       return $ C.CompiledPragma r x hs
+    A.CompilePragma b x s -> do
+      x <- toConcrete x
+      return $ C.CompilePragma r b x s
     A.CompiledExportPragma x hs -> do
       x <- toConcrete x
       return $ C.CompiledExportPragma r x hs
-    A.CompiledEpicPragma x e -> do
-      x <- toConcrete x
-      return $ C.CompiledEpicPragma r x e
     A.CompiledJSPragma x e -> do
       x <- toConcrete x
       return $ C.CompiledJSPragma r x e
@@ -846,7 +898,9 @@ instance ToConcrete RangeAndPragma C.Pragma where
       x <- toConcrete x
       return $ C.CompiledDataUHCPragma r x crd crcs
     A.StaticPragma x -> C.StaticPragma r <$> toConcrete x
+    A.InjectivePragma x -> C.InjectivePragma r <$> toConcrete x
     A.InlinePragma x -> C.InlinePragma r <$> toConcrete x
+    A.EtaPragma x    -> C.EtaPragma    r <$> toConcrete x
     A.DisplayPragma f ps rhs ->
       C.DisplayPragma r <$> toConcrete (A.DefP (PatRange noRange) (AmbQ [f]) ps) <*> toConcrete rhs
 
@@ -864,18 +918,82 @@ instance ToConcrete A.LHS C.LHS where
 instance ToConcrete A.LHSCore C.Pattern where
   bindToConcrete = bindToConcrete . lhsCoreToPattern
 
-appBrackets' :: [arg] -> Precedence -> Bool
-appBrackets' []    _   = False
-appBrackets' (_:_) ctx = appBrackets ctx
+appBracketsArgs :: [arg] -> PrecedenceStack -> Bool
+appBracketsArgs []    _   = False
+appBracketsArgs (_:_) ctx = appBrackets ctx
 
+-- Auxiliary wrappers for processing the bindings in patterns in the right order.
+newtype UserPattern a  = UserPattern a
+newtype SplitPattern a = SplitPattern a
 newtype BindingPattern = BindingPat A.Pattern
 newtype FreshName = FreshenName A.Name
 
 instance ToConcrete FreshName A.Name where
   bindToConcrete (FreshenName x) ret = bindToConcrete x $ \ y -> ret x{ nameConcrete = y }
 
--- Takes care of freshening and binding pattern variables, but doesn't actually
+-- Pass 1: (Issue #2729)
+-- Takes care of binding the originally user-written pattern variables, but doesn't actually
 -- translate anything to Concrete.
+instance ToConcrete (UserPattern A.Pattern) A.Pattern where
+  bindToConcrete (UserPattern p) ret =
+    case p of
+      A.VarP x               -> bindName' x $ ret $ A.VarP x
+      A.WildP{}              -> ret p
+      A.ProjP{}              -> ret p
+      A.AbsurdP{}            -> ret p
+      A.LitP{}               -> ret p
+      A.DotP{}               -> ret p
+      -- Andreas, 2017-09-03, issue #2729:
+      -- Do not go into patterns generated by case-split here!
+      -- They are treated in a second pass.
+      A.ConP i c args
+        | patOrigin i == ConOSplit -> ret p
+        | otherwise          -> bindToConcrete (map UserPattern args) $ ret . A.ConP i c
+      A.DefP i f args        -> bindToConcrete (map UserPattern args) $ ret . A.DefP i f
+      A.PatternSynP i f args -> bindToConcrete (map UserPattern args) $ ret . A.PatternSynP i f
+      A.RecP i args          -> bindToConcrete ((map . fmap) UserPattern args) $ ret . A.RecP i
+      A.AsP i x p            -> bindName' x $
+                                bindToConcrete (UserPattern p) $ \ p ->
+                                ret (A.AsP i x p)
+
+instance ToConcrete (UserPattern (NamedArg A.Pattern)) (NamedArg A.Pattern) where
+  bindToConcrete (UserPattern np) ret =
+    case getOrigin np of
+      CaseSplit -> ret np
+      _         -> bindToConcrete (fmap (fmap UserPattern) np) ret
+
+-- Pass 2a: locate case-split pattern.  Don't bind anything!
+instance ToConcrete (SplitPattern A.Pattern) A.Pattern where
+  bindToConcrete (SplitPattern p) ret =
+    case p of
+      A.VarP x               -> ret p
+      A.WildP{}              -> ret p
+      A.ProjP{}              -> ret p
+      A.AbsurdP{}            -> ret p
+      A.LitP{}               -> ret p
+      A.DotP{}               -> ret p
+      -- Andreas, 2017-09-03, issue #2729:
+      -- For patterns generated by case-split here, switch to freshening & binding.
+      A.ConP i c args
+        | patOrigin i == ConOSplit
+                             -> bindToConcrete ((map . fmap . fmap) BindingPat args) $ ret . A.ConP i c
+        | otherwise          -> bindToConcrete (map SplitPattern args) $ ret . A.ConP i c
+      A.DefP i f args        -> bindToConcrete (map SplitPattern args) $ ret . A.DefP i f
+      A.PatternSynP i f args -> bindToConcrete (map SplitPattern args) $ ret . A.PatternSynP i f
+      A.RecP i args          -> bindToConcrete ((map . fmap) SplitPattern args) $ ret . A.RecP i
+      A.AsP i x p            -> bindToConcrete (SplitPattern p)  $ \ p ->
+                                ret (A.AsP i x p)
+
+instance ToConcrete (SplitPattern (NamedArg A.Pattern)) (NamedArg A.Pattern) where
+  bindToConcrete (SplitPattern np) ret =
+    case getOrigin np of
+      CaseSplit -> bindToConcrete (fmap (fmap BindingPat  ) np) ret
+      _         -> bindToConcrete (fmap (fmap SplitPattern) np) ret
+
+
+-- Pass 2b:
+-- Takes care of freshening and binding pattern variables introduced by case split.
+-- Still does not translate anything to Concrete.
 instance ToConcrete BindingPattern A.Pattern where
   bindToConcrete (BindingPat p) ret =
     case p of
@@ -896,7 +1014,9 @@ instance ToConcrete BindingPattern A.Pattern where
 instance ToConcrete A.Pattern C.Pattern where
   bindToConcrete p ret = do
     prec <- currentPrecedence
-    bindToConcrete (BindingPat p) (ret <=< withPrecedence prec . toConcrete)
+    bindToConcrete (UserPattern p) $ \ p -> do
+      bindToConcrete (SplitPattern p) $ \ p -> do
+        ret =<< do withPrecedence' prec $ toConcrete p
   toConcrete p =
     case p of
       A.VarP x ->
@@ -957,7 +1077,7 @@ instance ToConcrete A.Pattern C.Pattern where
         Just c  -> applyTo args2 c
         Nothing -> applyTo args . C.IdentP =<< toConcrete x
     -- Note: applyTo [] c = return c
-    applyTo args c = bracketP_ (appBrackets' args) $ do
+    applyTo args c = bracketP_ (appBracketsArgs args) $ do
       foldl C.AppP c <$> toConcreteCtx ArgumentCtx args
 
 
@@ -965,46 +1085,112 @@ instance ToConcrete A.Pattern C.Pattern where
 
 data Hd = HdVar A.Name | HdCon A.QName | HdDef A.QName
 
-cOpApp :: Range -> C.QName -> A.Name -> [C.Expr] -> C.Expr
+cOpApp :: Range -> C.QName -> A.Name -> [Maybe C.Expr] -> C.Expr
 cOpApp r x n es =
   C.OpApp r x (Set.singleton n)
-          (map (defaultNamedArg . noPlaceholder . Ordinary) es)
+          (map (defaultNamedArg . placeholder) eps)
+  where
+    x0 = C.unqualify x
+    positions | isPrefix  x0 =                [ Middle | _ <- drop 1 es ] ++ [End]
+              | isPostfix x0 = [Beginning] ++ [ Middle | _ <- drop 1 es ]
+              | isInfix x0   = [Beginning] ++ [ Middle | _ <- drop 2 es ] ++ [End]
+              | otherwise    =                [ Middle | _ <- es ]
+    eps = zip es positions
+    placeholder (Nothing, pos) = Placeholder pos
+    placeholder (Just e,  _)   = noPlaceholder (Ordinary e)
+
+tryToRecoverNatural :: A.Expr -> AbsToCon C.Expr -> AbsToCon C.Expr
+tryToRecoverNatural e def = do
+  builtins <- stBuiltinThings <$> lift get
+  let reified = do
+        zero <- getAQName "ZERO" builtins
+        suc  <- getAQName "SUC"  builtins
+        explore zero suc 0 e
+  case reified of
+    Just n  -> return $ C.Lit $ LitNat noRange n
+    Nothing -> def
+  where
+
+    getAQName :: String -> BuiltinThings a -> Maybe A.QName
+    getAQName str bs = do
+      Builtin (I.Con hd _ _) <- Map.lookup str bs
+      return $ I.conName hd
+
+    explore :: A.QName -> A.QName -> Integer -> A.Expr -> Maybe Integer
+    explore z s k (A.App _ (A.Con (AmbQ [f])) t) | f == s =
+      let v = 1+k in v `seq` explore z s v $ namedArg t
+    explore z s k (A.Con (AmbQ [x]))             | x == z = Just k
+    explore z s k (A.Lit (LitNat _ l))                    = Just (k + l)
+    explore _ _ _ _ = Nothing
 
 tryToRecoverOpApp :: A.Expr -> AbsToCon C.Expr -> AbsToCon C.Expr
-tryToRecoverOpApp e def = caseMaybeM (recoverOpApp bracket cOpApp view e) def return
+tryToRecoverOpApp e def = caseMaybeM (recoverOpApp bracket (isParenlessLambda . defaultNamedArg) cOpApp view e) def return
   where
-    view e = do
-      let Application hd args = AV.appView e
-      case hd of
-        Var x            -> Just (HdVar x, args)
-        Def f            -> Just (HdDef f, args)
-        Con (AmbQ (c:_)) -> Just (HdCon c, args)
-        Con (AmbQ [])    -> __IMPOSSIBLE__
-        _                -> Nothing
+    view e
+        -- Do we have a series of inserted lambdas?
+      | Just xs@(_:_) <- traverse insertedName bs =
+        (,) <$> getHead hd <*> sectionArgs xs args
+      where
+        LamView     bs body = AV.lamView e
+        Application hd args = AV.appView body
+
+        -- Only inserted domain-free visible lambdas come from sections.
+        insertedName (i, A.DomainFree ai x)
+          | getOrigin i == Inserted && visible ai = Just x
+        insertedName _ = Nothing
+
+        -- Build section arguments. Need to check that:
+        -- lambda bound variables appear in the right order and only as
+        -- top-level arguments.
+        sectionArgs :: [A.Name] -> [NamedArg A.Expr] -> Maybe [NamedArg (Maybe A.Expr)]
+        sectionArgs xs = go xs
+          where
+            noXs = getAll . foldExpr (\ case A.Var x -> All (notElem x xs)
+                                             _       -> All True)
+            go [] [] = return []
+            go (y : ys) (arg : args)
+              | visible arg
+              , A.Var y' <- namedArg arg
+              , y == y'       = (fmap (Nothing <$) arg :) <$> go ys args
+            go ys (arg : args)
+              | visible arg, noXs arg = ((fmap . fmap) Just arg :) <$> go ys args
+            go _ _ = Nothing
+
+    view e = (, (map . fmap . fmap) Just args) <$> getHead hd
+      where Application hd args = AV.appView e
+
+    getHead (Var x)              = Just (HdVar x)
+    getHead (Def f)              = Just (HdDef f)
+    getHead (Con (AmbQ (c : _))) = Just (HdCon c)
+    getHead (Con (AmbQ []))      = __IMPOSSIBLE__
+    getHead _                    = Nothing
 
 tryToRecoverOpAppP :: A.Pattern -> AbsToCon (Maybe C.Pattern)
-tryToRecoverOpAppP = recoverOpApp bracketP_ opApp view
+tryToRecoverOpAppP = recoverOpApp bracketP_ (const False) opApp view
   where
     opApp r x n ps =
-      C.OpAppP r x (Set.singleton n) (map defaultNamedArg ps)
+      C.OpAppP r x (Set.singleton n) (map (defaultNamedArg . fromjust) ps)
+    fromjust (Just x) = x
+    fromjust Nothing  = __IMPOSSIBLE__  -- `view` does not generate any `Nothing`s
 
     view p = case p of
-      ConP _ (AmbQ (c:_)) ps -> Just (HdCon c, ps)
-      DefP _ (AmbQ (f:_)) ps -> Just (HdDef f, ps)
+      ConP _ (AmbQ (c:_)) ps -> Just (HdCon c, (map . fmap . fmap) Just ps)
+      DefP _ (AmbQ (f:_)) ps -> Just (HdDef f, (map . fmap . fmap) Just ps)
       _ -> __IMPOSSIBLE__
       -- ProjP _ _ (AmbQ (d:_))   -> Just (HdDef d, [])   -- ? Andreas, 2016-04-21
       -- _                      -> Nothing
 
 recoverOpApp :: (ToConcrete a c, HasRange c)
-  => ((Precedence -> Bool) -> AbsToCon c -> AbsToCon c)
-  -> (Range -> C.QName -> A.Name -> [c] -> c)
-  -> (a -> Maybe (Hd, [NamedArg a]))
+  => ((PrecedenceStack -> Bool) -> AbsToCon c -> AbsToCon c)
+  -> (a -> Bool)  -- ^ Check for parenless lambdas
+  -> (Range -> C.QName -> A.Name -> [Maybe c] -> c)
+  -> (a -> Maybe (Hd, [NamedArg (Maybe a)]))  -- ^ `Nothing` for sections
   -> a
   -> AbsToCon (Maybe c)
-recoverOpApp bracket opApp view e = case view e of
+recoverOpApp bracket isParenlessLam opApp view e = case view e of
   Nothing -> mDefault
   Just (hd, args)
-    | all notHidden args  -> do
+    | all visible args    -> do
       let  args' = map namedArg args
       case hd of
         HdVar  n
@@ -1036,11 +1222,11 @@ recoverOpApp bracket opApp view e = case view e of
             as' = case as of
                     as@(_ : _ : _) -> init $ tail as
                     _              -> __IMPOSSIBLE__
-        e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
-        es <- mapM (toConcreteCtx InsideOperandCtx) as'
-        en <- toConcreteCtx (RightOperandCtx fixity) an
         Just <$> do
-          bracket (opBrackets fixity) $
+          bracket (opBrackets' (maybe False isParenlessLam an) fixity) $ do
+            e1 <- traverse (toConcreteCtx $ LeftOperandCtx fixity) a1
+            es <- (mapM . traverse) (toConcreteCtx InsideOperandCtx) as'
+            en <- traverse (toConcreteCtx $ RightOperandCtx fixity) an
             return $ opApp (getRange (e1, en)) x n ([e1] ++ es ++ [en])
 
   -- prefix
@@ -1050,10 +1236,10 @@ recoverOpApp bracket opApp view e = case view e of
             as' = case as of
                     as@(_ : _) -> init as
                     _          -> __IMPOSSIBLE__
-        es <- mapM (toConcreteCtx InsideOperandCtx) as'
-        en <- toConcreteCtx (RightOperandCtx fixity) an
         Just <$> do
-          bracket (opBrackets fixity) $
+          bracket (opBrackets' (maybe False isParenlessLam an) fixity) $ do
+            es <- (mapM . traverse) (toConcreteCtx InsideOperandCtx) as'
+            en <- traverse (toConcreteCtx $ RightOperandCtx fixity) an
             return $ opApp (getRange (n, en)) x n (es ++ [en])
 
   -- postfix
@@ -1061,15 +1247,15 @@ recoverOpApp bracket opApp view e = case view e of
     | Hole <- head xs = do
         let a1  = head as
             as' = tail as
-        e1 <- toConcreteCtx (LeftOperandCtx fixity) a1
-        es <- mapM (toConcreteCtx InsideOperandCtx) as'
+        e1 <- traverse (toConcreteCtx $ LeftOperandCtx fixity) a1
+        es <- (mapM . traverse) (toConcreteCtx InsideOperandCtx) as'
         Just <$> do
           bracket (opBrackets fixity) $
             return $ opApp (getRange (e1, n)) x n ([e1] ++ es)
 
   -- roundfix
   doQName _ x n as xs = do
-    es <- mapM (toConcreteCtx InsideOperandCtx) as
+    es <- (mapM . traverse) (toConcreteCtx InsideOperandCtx) as
     Just <$> do
       bracket roundFixBrackets $
         return $ opApp (getRange x) x n es
@@ -1082,3 +1268,4 @@ instance ToConcrete InteractionId C.Expr where
 instance ToConcrete NamedMeta C.Expr where
     toConcrete i = do
       return $ C.Underscore noRange (Just $ prettyShow i)
+
