@@ -1,8 +1,8 @@
-{-# LANGUAGE CPP                    #-}
+{-# LANGUAGE CPP                        #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE NondecreasingIndentation #-}
-{-# LANGUAGE ScopedTypeVariables    #-}
-{-# LANGUAGE UndecidableInstances   #-}
+{-# LANGUAGE NondecreasingIndentation   #-}
+{-# LANGUAGE TypeFamilies               #-}  -- for type equality ~
+{-# LANGUAGE UndecidableInstances       #-}
 
 {-|
     Translating from internal syntax to abstract syntax. Enables nice
@@ -26,23 +26,25 @@ import Control.Applicative hiding (empty)
 import Control.Monad.State hiding (mapM_, mapM)
 import Control.Monad.Reader hiding (mapM_, mapM)
 
-import Data.Foldable (foldMap)
-import Data.List hiding (null, sort)
+import Data.Foldable (Foldable, foldMap)
+import qualified Data.List as List
 import qualified Data.Map as Map
 import Data.Maybe
 import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend)
 import Data.Set (Set)
 import qualified Data.Set as Set
-import Data.Traversable (traverse, mapM)
+import Data.Traversable (Traversable, traverse, mapM)
 import qualified Data.Traversable as Trav
 
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 import Agda.Syntax.Common
 import Agda.Syntax.Fixity
+import qualified Agda.Syntax.Concrete.Name as C
 import Agda.Syntax.Concrete (FieldAssignment'(..), exprFieldA)
 import Agda.Syntax.Info as Info
 import Agda.Syntax.Abstract as A
+import Agda.Syntax.Abstract.Pattern ( foldAPattern )
 import Agda.Syntax.Abstract.Pretty
 import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern as I
@@ -52,7 +54,7 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Reduce
 import {-# SOURCE #-} Agda.TypeChecking.Records
-import Agda.TypeChecking.CompiledClause (CompiledClauses(Fail))
+import Agda.TypeChecking.CompiledClause (CompiledClauses'(Fail))
 import Agda.TypeChecking.DisplayForm
 import Agda.TypeChecking.Level
 import {-# SOURCE #-} Agda.TypeChecking.Datatypes
@@ -63,6 +65,7 @@ import Agda.TypeChecking.DropArgs
 
 import Agda.Interaction.Options ( optPostfixProjections )
 
+import Agda.Utils.Either
 import Agda.Utils.Except ( MonadError(catchError) )
 import Agda.Utils.Function
 import Agda.Utils.Functor
@@ -195,14 +198,14 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     md <- liftTCM $ -- addContext (replicate (length ps) "x") $
       displayForm f $ zipWith (\ p i -> I.Apply $ p $> I.var i) ps [0..]
     reportSLn "reify.display" 60 $
-      "display form of " ++ show f ++ " " ++ show ps ++ " " ++ show wps ++ ":\n  " ++ show md
+      "display form of " ++ prettyShow f ++ " " ++ show ps ++ " " ++ show wps ++ ":\n  " ++ show md
     case md of
       Just d  | okDisplayForm d -> do
         -- In the display term @d@, @var i@ should be a placeholder
         -- for the @i@th pattern of @ps@.
         -- Andreas, 2014-06-11:
         -- Are we sure that @d@ did not use @var i@ otherwise?
-        lhs' <- displayLHS (map namedArg ps) wps d
+        lhs' <- displayLHS ps wps d
         reportSDoc "reify.display" 70 $ do
           doc <- prettyA lhs'
           return $ vcat
@@ -220,6 +223,7 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     -- can serve as a valid left-hand side. That means checking that it is a
     -- defined name applied to valid lhs eliminators (projections or
     -- applications to constructor patterns).
+    okDisplayForm :: DisplayTerm -> Bool
     okDisplayForm (DWithApp d ds es) =
       okDisplayForm d && all okDisplayTerm ds  && all okToDropE es
       -- Andreas, 2016-05-03, issue #1950.
@@ -230,29 +234,36 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     okDisplayForm DCon{}               = False
     okDisplayForm DTerm{}              = False
 
+    okDisplayTerm :: DisplayTerm -> Bool
     okDisplayTerm (DTerm v) = okTerm v
     okDisplayTerm DDot{}    = True
     okDisplayTerm DCon{}    = True
     okDisplayTerm DDef{}    = False
     okDisplayTerm _         = False
 
+    okDElim :: Elim' DisplayTerm -> Bool
     okDElim (I.Apply v) = okDisplayTerm $ unArg v
     okDElim I.Proj{}    = True
 
+    okToDropE :: Elim' Term -> Bool
     okToDropE (I.Apply v) = okToDrop v
     okToDropE I.Proj{}    = False
 
+    okToDrop :: Arg I.Term -> Bool
     okToDrop arg = notVisible arg && case ignoreSharing $ unArg arg of
       I.Var _ []   -> True
       I.DontCare{} -> True  -- no matching on irrelevant things.  __IMPOSSIBLE__ anyway?
       I.Level{}    -> True  -- no matching on levels. __IMPOSSIBLE__ anyway?
       _ -> False
 
+    okArg :: Arg I.Term -> Bool
     okArg = okTerm . unArg
 
+    okElim :: Elim' I.Term -> Bool
     okElim (I.Apply a) = okArg a
     okElim (I.Proj{})  = True
 
+    okTerm :: I.Term -> Bool
     okTerm (I.Var _ []) = True
     okTerm (I.Con c ci vs) = all okArg vs
     okTerm (I.Def x []) = isNoName $ qnameToConcrete x -- Handling wildcards in display forms
@@ -267,37 +278,45 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
     flattenWith (DTerm (I.Def f es)) = (f, map (fmap DTerm) es, [])
     flattenWith _ = __IMPOSSIBLE__
 
-    displayLHS :: [A.Pattern] -> [A.Pattern] -> DisplayTerm -> TCM A.SpineLHS
+    displayLHS :: [NamedArg A.Pattern] -> [A.Pattern] -> DisplayTerm -> TCM A.SpineLHS
     displayLHS ps wps d = do
         let (f, vs, es) = flattenWith d
         ds <- mapM (namedArg <.> elimToPat) es
         vs <- mapM elimToPat vs
         return $ SpineLHS i f vs (ds ++ wps)
       where
-        argToPat arg = fmap unnamed <$> traverse termToPat arg
+        argToPat :: Arg DisplayTerm -> TCM (NamedArg A.Pattern)
+        argToPat arg = traverse termToPat arg
+
+        elimToPat :: I.Elim' DisplayTerm -> TCM (NamedArg A.Pattern)
         elimToPat (I.Apply arg) = argToPat arg
         elimToPat (I.Proj o d)  = return $ defaultNamedArg $ A.ProjP patNoRange o $ AmbQ [d]
 
-        termToPat :: DisplayTerm -> TCM A.Pattern
+        termToPat :: DisplayTerm -> TCM (Named_ A.Pattern)
 
-        termToPat (DTerm (I.Var n [])) = return $ case ps !!! n of
-                                           Nothing -> __IMPOSSIBLE__
-                                           Just p  -> p
+        termToPat (DTerm (I.Var n [])) = return $ unArg $ fromMaybe __IMPOSSIBLE__ $ ps !!! n
 
-        termToPat (DCon c ci vs)          = tryRecPFromConP =<< do
+        termToPat (DCon c ci vs)          = fmap unnamed <$> tryRecPFromConP =<< do
            A.ConP (ConPatInfo ci patNoRange) (AmbQ [conName c]) <$> mapM argToPat vs
 
-        termToPat (DTerm (I.Con c ci vs)) = tryRecPFromConP =<< do
+        termToPat (DTerm (I.Con c ci vs)) = fmap unnamed <$> tryRecPFromConP =<< do
            A.ConP (ConPatInfo ci patNoRange) (AmbQ [conName c]) <$> mapM (argToPat . fmap DTerm) vs
 
-        termToPat (DTerm (I.Def _ [])) = return $ A.WildP patNoRange
-        termToPat (DDef _ [])          = return $ A.WildP patNoRange
+        termToPat (DTerm (I.Def _ [])) = return $ unnamed $ A.WildP patNoRange
+        termToPat (DDef _ [])          = return $ unnamed $ A.WildP patNoRange
 
-        termToPat (DDot v)             = A.DotP patNoRange Inserted <$> termToExpr v
-        termToPat v                    = A.DotP patNoRange Inserted <$> reify v -- __IMPOSSIBLE__
+        -- Currently we don't keep track of the origin of a dot pattern in the internal syntax,
+        -- so here we give __IMPOSSIBLE__. This is only used for printing purposes, the origin
+        -- should not be used anyway after this point.
+        -- Andreas, 2017-02-14: This crashes with -v 100.
+        -- termToPat (DDot v)             = A.DotP patNoRange __IMPOSSIBLE__ <$> termToExpr v
+        -- termToPat v                    = A.DotP patNoRange __IMPOSSIBLE__ <$> reify v -- __IMPOSSIBLE__
+        termToPat (DDot v)             = unnamed . A.DotP patNoRange Inserted <$> termToExpr v
+        termToPat v                    = unnamed . A.DotP patNoRange Inserted <$> reify v
 
-        len = genericLength ps
+        len = length ps
 
+        argsToExpr :: I.Args -> TCM [Arg A.Expr]
         argsToExpr = mapM (traverse termToExpr)
 
         -- TODO: restructure this to avoid having to repeat the code for reify
@@ -318,7 +337,7 @@ reifyDisplayFormP lhs@(A.SpineLHS i f ps wps) =
               -- even the pattern variables @n < len@ can be
               -- applied to some args @vs@.
               e <- if n < len
-                   then return $ A.patternToExpr $ ps !! n
+                   then return $ A.patternToExpr $ namedArg $ ps !! n
                    else reify (I.var (n - len))
               apps e =<< argsToExpr vs
             _ -> return underscore
@@ -354,7 +373,7 @@ reifyTerm expandAnonDefs0 v = do
       case isR || ci == ConORec of
         True -> do
           showImp <- showImplicitArguments
-          let keep (a, v) = showImp || notHidden a
+          let keep (a, v) = showImp || visible a
           r  <- getConstructorData x
           xs <- getRecordFieldNames r
           vs <- map unArg <$> reify vs
@@ -403,13 +422,13 @@ reifyTerm expandAnonDefs0 v = do
 --    I.Lam info b | isAbsurdBody b -> return $ A. AbsurdLam noExprInfo $ getHiding info
     I.Lam info b    -> do
       (x,e) <- reify b
-      return $ A.Lam noExprInfo (DomainFree info x) e
+      return $ A.Lam (setOrigin (getOrigin info) defaultLamInfo_) (DomainFree info x) e
       -- Andreas, 2011-04-07 we do not need relevance information at internal Lambda
     I.Lit l        -> reify l
     I.Level l      -> reify l
     I.Pi a b       -> case b of
         NoAbs _ b'
-          | notHidden a -> uncurry (A.Fun $ noExprInfo) <$> reify (a, b')
+          | visible a   -> uncurry (A.Fun $ noExprInfo) <$> reify (a, b')
             -- Andreas, 2013-11-11 Hidden/Instance I.Pi must be A.Pi
             -- since (a) the syntax {A} -> B or {{A}} -> B is not legal
             -- and (b) the name of the binder might matter.
@@ -449,14 +468,14 @@ reifyTerm expandAnonDefs0 v = do
         YesReduction _ v -> do
           reportSLn "reify.anon" 60 $ unlines
             [ "reduction on defined ident. in anonymous module"
-            , "x = " ++ show x
+            , "x = " ++ prettyShow x
             , "v = " ++ show v
             ]
           reify v
         NoReduction () -> do
           reportSLn "reify.anon" 60 $ unlines
             [ "no reduction on defined ident. in anonymous module"
-            , "x  = " ++ show x
+            , "x  = " ++ prettyShow x
             , "es = " ++ show es
             ]
           reifyDef' x es
@@ -464,13 +483,14 @@ reifyTerm expandAnonDefs0 v = do
 
     reifyDef' :: QName -> I.Elims -> TCM Expr
     reifyDef' x es = do
-      reportSLn "reify.def" 60 $ "reifying call to " ++ show x
+      reportSLn "reify.def" 60 $ "reifying call to " ++ prettyShow x
       -- We should drop this many arguments from the local context.
       n <- getDefFreeVars x
+      reportSLn "reify.def" 70 $ "freeVars for " ++ prettyShow x ++ " = " ++ show n
       -- If the definition is not (yet) in the signature,
       -- we just do the obvious.
-      let fallback = elims (A.Def x) =<< reify (drop n es)
-      caseMaybeM (tryMaybe $ getConstInfo x) fallback $ \ defn -> do
+      let fallback _ = elims (A.Def x) =<< reify (drop n es)
+      caseEitherM (getConstInfo' x) fallback $ \ defn -> do
       let def = theDef defn
 
       -- Check if we have an absurd lambda.
@@ -479,7 +499,7 @@ reifyTerm expandAnonDefs0 v = do
                 | isAbsurdLambdaName x -> do
                   -- get hiding info from last pattern, which should be ()
                   let h = getHiding $ last $ namedClausePats cl
-                  elims (A.AbsurdLam noExprInfo h) =<< reify (drop n es)
+                  elims (A.AbsurdLam defaultLamInfo_ h) =<< reify (drop n es)
 
       -- Otherwise (no absurd lambda):
        _ -> do
@@ -504,6 +524,8 @@ reifyTerm expandAnonDefs0 v = do
            (pad, nes :: [Elim' (Named_ Term)]) <- case def of
 
             Function{ funProjection = Just Projection{ projIndex = np } } | np > 0 -> do
+              reportSLn "reify.def" 70 $ "  def. is a projection with projIndex = " ++ show np
+
               -- This is tricky:
               --  * getDefFreeVars x tells us how many arguments
               --    are part of the local context
@@ -547,7 +569,7 @@ reifyTerm expandAnonDefs0 v = do
               let padVis  = map (fmap (unnamed . namedThing)) padVisNamed
 
               -- Keep only the rest with the same visibility of @dom@...
-              let padTail = filter ((getHiding dom ==) . getHiding) padRest
+              let padTail = filter (sameHiding dom) padRest
 
               -- ... and even the same name.
               let padSame = filter ((Just (fst (unDom dom)) ==) . fmap rangedThing . nameOf . unArg) padTail
@@ -558,7 +580,12 @@ reifyTerm expandAnonDefs0 v = do
 
             -- If it is not a projection(-like) function, we need no padding.
             _ -> return ([], map (fmap unnamed) $ drop n es)
-           let hd = foldl' (A.App noExprInfo) (A.Def x) pad
+
+           reportSLn "reify.def" 70 $ unlines
+             [ "  pad = " ++ show pad
+             , "  nes = " ++ show nes
+             ]
+           let hd = List.foldl' (A.App noExprInfo) (A.Def x) pad
            nelims hd =<< reify nes
 
     -- Andreas, 2016-07-06 Issue #2047
@@ -582,8 +609,8 @@ reifyTerm expandAnonDefs0 v = do
 
     reifyExtLam :: QName -> Int -> [I.Clause] -> I.Elims -> TCM Expr
     reifyExtLam x npars cls es = do
-      reportSLn "reify.def" 10 $ "reifying extended lambda " ++ show x
-      reportSLn "reify.def" 50 $ show $ nest 2 $ vcat
+      reportSLn "reify.def" 10 $ "reifying extended lambda " ++ prettyShow x
+      reportSLn "reify.def" 50 $ render $ nest 2 $ vcat
         [ text "npars =" <+> pretty npars
         , text "es    =" <+> fsep (map (prettyPrec 10) es)
         , text "def   =" <+> vcat (map pretty cls) ]
@@ -596,7 +623,7 @@ reifyTerm expandAnonDefs0 v = do
       cls <- mapM (reify . NamedClause x False . (`applyE` pars)) cls
       let cx    = nameConcrete $ qnameName x
           dInfo = mkDefInfo cx noFixity' PublicAccess ConcreteDef (getRange x)
-      elims (A.ExtendedLam noExprInfo dInfo x cls) =<< reify rest
+      elims (A.ExtendedLam defaultLamInfo_ dInfo x cls) =<< reify rest
 
 -- | @nameFirstIfHidden (x:a) ({e} es) = {x = e} es@
 nameFirstIfHidden :: Dom (ArgName, t) -> [Elim' a] -> [Elim' (Named_ a)]
@@ -660,20 +687,28 @@ stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the n
       strip ps = stripArgs True ps
         where
           stripArgs _ [] = []
-          stripArgs fixedPos (a : as) =
-            if canStrip a
-            then if   all canStrip $ takeWhile isUnnamedHidden as
+          stripArgs fixedPos (a : as)
+            -- Andreas, 2017-01-18, issue #819: preserves _ when splitting:
+            -- An Inserted visible variable comes form a WildP and is restored as such.
+            | visible a, getOrigin a == Inserted, varOrDot (namedArg a) = goWild
+            -- A hidden non-UserWritten variable is removed if not needed for
+            -- correct position of the following hidden arguments.
+            | canStrip a =
+                 if   all canStrip $ takeWhile isUnnamedHidden as
                  then stripArgs False as
-                 else let a' = fmap ($> A.WildP (Info.PatRange $ getRange a)) a
-                      in  stripName fixedPos a' : stripArgs True as
-            else stripName fixedPos (stripArg a) : stripArgs True as
+                 else goWild
+            -- Other arguments are kept.
+            | otherwise = stripName fixedPos (stripArg a) : stripArgs True as
+            where
+              a'     = setNamedArg a $ A.WildP $ Info.PatRange $ getRange a
+              goWild = stripName fixedPos a' : stripArgs True as
 
           stripName True  = fmap (unnamed . namedThing)
           stripName False = id
 
           canStrip a = and
             [ notVisible a
-            , getOrigin a /= UserWritten
+            , getOrigin a `notElem` [ UserWritten , CaseSplit ]
             , varOrDot (namedArg a)
             ]
 
@@ -701,21 +736,22 @@ stripImplicits (ps, wps) = do          -- v if show-implicit we don't need the n
                                  = all varOrDot $ map namedArg ps
           varOrDot _             = False
 
--- | @blank bound x@ replaces all variables in @x@ that are not in @bound@ by
+-- | @blank bound e@ replaces all variables in expression @e@ that are not in @bound@ by
 --   an underscore @_@. It is used for printing dot patterns: we don't want to
 --   make implicit variables explicit, so we blank them out in the dot patterns
 --   instead (this is fine since dot patterns can be inferred anyway).
+
 class BlankVars a where
   blank :: Set Name -> a -> a
 
-instance BlankVars a => BlankVars (Arg a) where
-  blank bound = fmap $ blank bound
+  default blank :: (Functor f, BlankVars b, f b ~ a) => Set Name -> a -> a
+  blank = fmap . blank
 
-instance BlankVars a => BlankVars (Named s a) where
-  blank bound = fmap $ blank bound
-
-instance BlankVars a => BlankVars [a] where
-  blank bound = fmap $ blank bound
+instance BlankVars a => BlankVars (Arg a)              where
+instance BlankVars a => BlankVars (Named s a)          where
+instance BlankVars a => BlankVars [a]                  where
+instance BlankVars a => BlankVars (A.Pattern' a)       where
+instance BlankVars a => BlankVars (FieldAssignment' a) where
 
 instance (BlankVars a, BlankVars b) => BlankVars (a, b) where
   blank bound (x, y) = (blank bound x, blank bound y)
@@ -727,13 +763,17 @@ instance (BlankVars a, BlankVars b) => BlankVars (Either a b) where
 instance BlankVars A.NamedDotPattern where
   blank bound = id
 
+instance BlankVars A.StrippedDotPattern where
+  blank bound = id
+
 instance BlankVars A.Clause where
-  blank bound (A.Clause lhs namedDots rhs [] ca) =
+  blank bound (A.Clause lhs namedDots strippedDots rhs [] ca) =
     let bound' = varsBoundIn lhs `Set.union` bound
     in  A.Clause (blank bound' lhs)
                  (blank bound' namedDots)
+                 (blank bound' strippedDots)
                  (blank bound' rhs) [] ca
-  blank bound (A.Clause lhs namedDots rhs (_:_) ca) = __IMPOSSIBLE__
+  blank bound (A.Clause lhs namedDots strippedDots rhs (_:_) ca) = __IMPOSSIBLE__
 
 instance BlankVars A.LHS where
   blank bound (A.LHS i core wps) = uncurry (A.LHS i) $ blank bound (core, wps)
@@ -742,25 +782,11 @@ instance BlankVars A.LHSCore where
   blank bound (A.LHSHead f ps) = A.LHSHead f $ blank bound ps
   blank bound (A.LHSProj p b ps) = uncurry (A.LHSProj p) $ blank bound (b, ps)
 
-instance BlankVars A.Pattern where
-  blank bound p = case p of
-    A.VarP _      -> p   -- do not blank pattern vars
-    A.ConP c i ps -> A.ConP c i $ blank bound ps
-    A.ProjP{}     -> p
-    A.DefP i f ps -> A.DefP i f $ blank bound ps
-    A.DotP i o e  -> A.DotP i o $ blank bound e
-    A.WildP _     -> p
-    A.AbsurdP _   -> p
-    A.LitP _      -> p
-    A.AsP i n p   -> A.AsP i n $ blank bound p
-    A.PatternSynP _ _ _ -> __IMPOSSIBLE__
-    A.RecP i fs   -> A.RecP i $ blank bound fs
-
 instance BlankVars A.Expr where
   blank bound e = case e of
     A.ScopedExpr i e       -> A.ScopedExpr i $ blank bound e
     A.Var x                -> if x `Set.member` bound then e
-                              else A.Underscore emptyMetaInfo
+                              else A.Underscore emptyMetaInfo  -- Here is the action!
     A.Def _                -> e
     A.Proj{}               -> e
     A.Con _                -> e
@@ -793,9 +819,6 @@ instance BlankVars A.Expr where
     A.PatternSyn {}        -> e
     A.Macro {}             -> e
 
-instance BlankVars a => BlankVars (FieldAssignment' a) where
-  blank bound = over exprFieldA (blank bound)
-
 instance BlankVars A.ModuleName where
   blank bound = id
 
@@ -816,9 +839,15 @@ instance BlankVars TypedBinding where
   blank bound (TBind r n e) = TBind r n $ blank bound e
   blank bound (TLet _ _)    = __IMPOSSIBLE__ -- Since the internal syntax has no let bindings left
 
+
+-- | Collect the binders in some abstract syntax lhs.
+
 class Binder a where
   varsBoundIn :: a -> Set Name
 
+  default varsBoundIn :: (Foldable f, Binder b, f b ~ a) => a -> Set Name
+  varsBoundIn = foldMap varsBoundIn
+
 instance Binder A.LHS where
   varsBoundIn (A.LHS _ core ps) = varsBoundIn (core, ps)
 
@@ -827,18 +856,18 @@ instance Binder A.LHSCore where
   varsBoundIn (A.LHSProj _ b ps) = varsBoundIn (b, ps)
 
 instance Binder A.Pattern where
-  varsBoundIn p = case p of
-    A.VarP x             -> if show x == "()" then empty else singleton x
-    A.ConP _ _ ps        -> varsBoundIn ps
-    A.ProjP{}            -> empty
-    A.DefP _ _ ps        -> varsBoundIn ps
-    A.WildP{}            -> empty
-    A.AsP _ _ p          -> varsBoundIn p
-    A.DotP{}             -> empty
-    A.AbsurdP{}          -> empty
-    A.LitP{}             -> empty
-    A.PatternSynP _ _ ps -> varsBoundIn ps
-    A.RecP _ fs          -> varsBoundIn fs
+  varsBoundIn = foldAPattern $ \case
+    A.VarP x            -> if prettyShow x == "()" then empty else singleton x -- TODO: get rid of this hack?
+    A.AsP _ x _         -> empty
+    A.ConP _ _ _        -> empty
+    A.ProjP{}           -> empty
+    A.DefP _ _ _        -> empty
+    A.WildP{}           -> empty
+    A.DotP{}            -> empty
+    A.AbsurdP{}         -> empty
+    A.LitP{}            -> empty
+    A.PatternSynP _ _ _ -> empty
+    A.RecP _ _          -> empty
 
 instance Binder A.LamBinding where
   varsBoundIn (A.DomainFree _ x) = singleton x
@@ -858,20 +887,13 @@ instance Binder LetBinding where
   varsBoundIn LetOpen{}           = empty
   varsBoundIn LetDeclaredVariable{} = empty
 
-instance Binder a => Binder (FieldAssignment' a) where
-  varsBoundIn = varsBoundIn . (^. exprFieldA)
-
-instance Binder a => Binder (Arg a) where
-  varsBoundIn = varsBoundIn . unArg
-
-instance Binder a => Binder (Named x a) where
-  varsBoundIn = varsBoundIn . namedThing
-
 instance Binder (WithHiding Name) where
   varsBoundIn (WithHiding _ x) = singleton x
 
-instance Binder a => Binder [a] where
-  varsBoundIn xs = Set.unions $ map varsBoundIn xs
+instance Binder a => Binder (FieldAssignment' a) where
+instance Binder a => Binder (Arg a)              where
+instance Binder a => Binder (Named x a)          where
+instance Binder a => Binder [a]                  where
 
 instance (Binder a, Binder b) => Binder (a, b) where
   varsBoundIn (x, y) = varsBoundIn x `Set.union` varsBoundIn y
@@ -884,19 +906,32 @@ reifyPatterns = mapM $ stripNameFromExplicit <.> traverse (traverse reifyPat)
   where
     stripNameFromExplicit :: NamedArg p -> NamedArg p
     stripNameFromExplicit a
-      | getHiding a == NotHidden = fmap (unnamed . namedThing) a
-      | otherwise                = a
+      | visible a = fmap (unnamed . namedThing) a
+      | otherwise = a
 
     reifyPat :: MonadTCM tcm => I.DeBruijnPattern -> tcm A.Pattern
     reifyPat p = do
      liftTCM $ reportSLn "reify.pat" 80 $ "reifying pattern " ++ show p
      case p of
-      I.VarP x | isAbsurdPatternName (dbPatVarName x)
-               -> return $ A.AbsurdP patNoRange    -- HACK
-      I.VarP x -> liftTCM $ A.VarP <$> nameOfBV (dbPatVarIndex x)
+      I.VarP x -> do
+        n <- liftTCM $ nameOfBV $ dbPatVarIndex x
+        case dbPatVarName x of
+          "_"  -> return $ A.VarP n
+          -- Andreas, 2017-09-03: TODO for #2580
+          -- Patterns @VarP "()"@ should have been replaced by @AbsurdP@, but the
+          -- case splitter still produces them.
+          y    -> if prettyShow (nameConcrete n) == "()" then return $ A.VarP n else
+            -- Andreas, 2017-09-03, issue #2729
+            -- Restore original pattern name.  AbstractToConcrete picks unique names.
+            return $ A.VarP n { nameConcrete = C.Name noRange [ C.Id y ] }
       I.DotP v -> do
         t <- liftTCM $ reify v
+        -- This is only used for printing purposes, so the Origin shouldn't be
+        -- used after this point anyway.
         return $ A.DotP patNoRange Inserted t
+        -- WAS: return $ A.DotP patNoRange __IMPOSSIBLE__ t
+        -- Crashes on -v 100.
+      I.AbsurdP p -> return $ A.AbsurdP patNoRange
       I.LitP l  -> return $ A.LitP l
       I.ProjP o d     -> return $ A.ProjP patNoRange o $ AmbQ [d]
       I.ConP c cpi ps -> do
@@ -932,7 +967,7 @@ instance Reify (QNamed I.Clause) A.Clause where
 instance Reify NamedClause A.Clause where
   reify (NamedClause f toDrop cl) = addContext (clauseTel cl) $ do
     reportSLn "reify.clause" 60 $ "reifying NamedClause"
-      ++ "\n  f      = " ++ show f
+      ++ "\n  f      = " ++ prettyShow f
       ++ "\n  toDrop = " ++ show toDrop
       ++ "\n  cl     = " ++ show cl
     ps  <- reifyPatterns $ namedClausePats cl
@@ -947,14 +982,14 @@ instance Reify NamedClause A.Clause where
     rhs <- caseMaybe (clauseBody cl) (return AbsurdRHS) $ \ e -> do
        RHS <$> reify e <*> pure Nothing
     reportSLn "reify.clause" 60 $ "reifying NamedClause, rhs = " ++ show rhs
-    let result = A.Clause (spineToLhs lhs) [] rhs [] (I.clauseCatchall cl)
+    let result = A.Clause (spineToLhs lhs) [] [] rhs [] (I.clauseCatchall cl)
     reportSLn "reify.clause" 60 $ "reified NamedClause, result = " ++ show result
     return result
     where
       perm = fromMaybe __IMPOSSIBLE__ $ clausePerm cl
       info = LHSRange noRange
 
-      dropParams n (SpineLHS i f ps wps) = SpineLHS i f (genericDrop n ps) wps
+      dropParams n (SpineLHS i f ps wps) = SpineLHS i f (drop n ps) wps
       stripImps (SpineLHS i f ps wps) = do
         (ps, wps) <- stripImplicits (ps, wps)
         return $ SpineLHS i f ps wps
