@@ -1,5 +1,4 @@
-{-# LANGUAGE ScopedTypeVariables    #-}
-{-# LANGUAGE UndecidableInstances   #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 {-# OPTIONS_GHC -fwarn-missing-signatures #-}
 
@@ -75,7 +74,7 @@ instance ToAbstract (Expr, Elim) Expr where
   toAbstract (f, Apply arg) = do
     arg     <- toAbstract arg
     showImp <- lift showImplicitArguments
-    return $ if showImp || getHiding arg == NotHidden
+    return $ if showImp || visible arg
              then App (ExprRange noRange) f arg
              else f
 
@@ -106,7 +105,7 @@ instance ToAbstract Term Expr where
     R.Lam h t  -> do
       (e, name) <- toAbstract t
       let info  = setHiding h $ setOrigin Reflected defaultArgInfo
-      return $ A.Lam exprNoRange (DomainFree info name) e
+      return $ A.Lam (setOrigin Reflected defaultLamInfo_) (DomainFree info name) e
     R.ExtLam cs es -> do
       name <- freshName_ extendedLambdaName
       m    <- lift $ getCurrentModule
@@ -114,7 +113,7 @@ instance ToAbstract Term Expr where
           cname   = nameConcrete name
           defInfo = mkDefInfo cname noFixity' PublicAccess ConcreteDef noRange
       cs <- toAbstract $ map (QNamed qname) cs
-      toAbstract (A.ExtendedLam exprNoRange defInfo qname cs, es)
+      toAbstract (A.ExtendedLam (setOrigin Reflected defaultLamInfo_) defInfo qname cs, es)
     R.Pi a b   -> do
       (b, name) <- toAbstract b
       a         <- toAbstract (a, name)
@@ -167,11 +166,11 @@ instance ToAbstract (QNamed R.Clause) A.Clause where
     (names, pats) <- toAbstractPats pats
     rhs           <- local (names++) $ toAbstract rhs
     let lhs = spineToLhs $ SpineLHS (LHSRange noRange) name pats []
-    return $ A.Clause lhs [] (RHS rhs Nothing) [] False
+    return $ A.Clause lhs [] [] (RHS rhs Nothing) [] False
   toAbstract (QNamed name (R.AbsurdClause pats)) = do
     (_, pats) <- toAbstractPats pats
     let lhs = spineToLhs $ SpineLHS (LHSRange noRange) name pats []
-    return $ A.Clause lhs [] AbsurdRHS [] False
+    return $ A.Clause lhs [] [] AbsurdRHS [] False
 
 instance ToAbstract [QNamed R.Clause] [A.Clause] where
   toAbstract = traverse toAbstract
