@@ -14,6 +14,7 @@ module Agda.Syntax.Treeless
 import Control.Arrow (first, second)
 
 import Data.Map (Map)
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import Agda.Syntax.Position
@@ -23,7 +24,7 @@ import Agda.Syntax.Abstract.Name
 data Compiled = Compiled
   { cTreeless :: TTerm
   , cArgUsage :: [Bool] }
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 type Args = [TTerm]
 
@@ -53,7 +54,7 @@ data TTerm = TVar Int
            | TErased
            | TError TError
            -- ^ A runtime error, something bad has happened.
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 -- | Compiler-related primitives. This are NOT the same thing as primitives
 -- in Agda's surface or internal syntax!
@@ -80,7 +81,7 @@ data TPrim
   | PEqQ
   | PIf
   | PSeq
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 isPrimEq :: TPrim -> Bool
 isPrimEq p = p `elem` [PEqI, PEqF, PEqS, PEqC, PEqQ]
@@ -155,7 +156,7 @@ data CaseType
   | CTString
   | CTFloat
   | CTQName
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 data TAlt
   = TACon    { aCon  :: QName, aArity :: Int, aBody :: TTerm }
@@ -165,7 +166,7 @@ data TAlt
   | TAGuard  { aGuard :: TTerm, aBody :: TTerm }
   -- ^ Binds no variables
   | TALit    { aLit :: Literal,   aBody:: TTerm }
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 data TError
   = TUnreachable
@@ -173,7 +174,7 @@ data TError
   -- Runtime behaviour of unreachable code is undefined, but preferably
   -- the program will exit with an error message. The compiler is free
   -- to assume that this code is unreachable and to remove it.
-  deriving (Typeable, Show, Eq, Ord)
+  deriving (Typeable, Data, Show, Eq, Ord)
 
 
 class Unreachable a where
