@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP                        #-}
+{-# LANGUAGE ScopedTypeVariables        #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 
 -- | The monad for the termination checker.
@@ -13,18 +14,18 @@ import Prelude hiding (null)
 
 import Control.Applicative hiding (empty)
 import Control.Monad.Reader
-import Control.Monad.Writer
 import Control.Monad.State
 
 import Data.Foldable (Foldable)
 import Data.Traversable (Traversable)
-import Data.Semigroup (Semigroup)
+import Data.Semigroup (Semigroup(..), Monoid(..))
 
 import Agda.Interaction.Options
 
 import Agda.Syntax.Abstract (IsProjP(..), AllNames)
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
+import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Literal
 import Agda.Syntax.Position (noRange)
 
@@ -35,7 +36,7 @@ import Agda.Termination.RecCheck (anyDefs)
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Benchmark
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Pretty as TCP
+import Agda.TypeChecking.Pretty hiding ((<>))
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
@@ -46,6 +47,7 @@ import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
+import Agda.Utils.Monoid
 import Agda.Utils.Null
 import Agda.Utils.Pretty (Pretty, prettyShow)
 import qualified Agda.Utils.Pretty as P
@@ -101,6 +103,9 @@ data TerEnv = TerEnv
   , terUserNames :: [QName]
     -- ^ The list of name actually appearing in the file (abstract syntax).
     --   Excludes the internally generated functions.
+  , terHaveInlinedWith :: Bool
+    -- ^ Does the actual clause result from with-inlining?
+    --   (If yes, it may be ill-typed.)
   , terTarget  :: Maybe Target
     -- ^ Target type of the function we are currently termination checking.
     --   Only the constructors of 'Target' are considered guarding.
@@ -115,7 +120,7 @@ data TerEnv = TerEnv
     -- ^ How many @SIZELT@ relations do we have in the context
     --   (= clause telescope).  Used to approximate termination
     --   for metas in call args.
-  , terPatterns :: MaskedDeBruijnPats
+  , terPatterns :: MaskedDeBruijnPatterns
     -- ^ The patterns of the clause we are checking.
   , terPatternsRaise :: !Int
     -- ^ Number of additional binders we have gone under
@@ -158,6 +163,7 @@ defaultTerEnv = TerEnv
   , terUserNames                = __IMPOSSIBLE__ -- needs to be set!
   , terMutual                   = __IMPOSSIBLE__ -- needs to be set!
   , terCurrent                  = __IMPOSSIBLE__ -- needs to be set!
+  , terHaveInlinedWith          = False
   , terTarget                   = Nothing
   , terDelayed                  = NotDelayed
   , terMaskArgs                 = repeat False   -- use all arguments (mask none)
@@ -182,7 +188,7 @@ class (Functor m, Monad m) => MonadTer m where
 -- | Termination monad.
 
 newtype TerM a = TerM { terM :: ReaderT TerEnv TCM a }
-  deriving (Functor, Applicative, Monad, MonadBench Phase)
+  deriving (Functor, Applicative, Monad, MonadBench Phase, HasOptions, MonadDebug)
 
 instance MonadTer TerM where
   terAsk     = TerM $ ask
@@ -249,6 +255,14 @@ instance HasConstInfo TerM where
   getConstInfo       = liftTCM . getConstInfo
   getRewriteRulesFor = liftTCM . getRewriteRulesFor
 
+instance Semigroup m => Semigroup (TerM m) where
+  (<>) = liftA2 (<>)
+
+instance (Semigroup m, Monoid m) => Monoid (TerM m) where
+  mempty  = pure mempty
+  mappend = (<>)
+  mconcat = mconcat <.> sequence
+
 -- * Modifiers and accessors for the termination environment in the monad.
 
 terGetGuardingTypeConstructors :: TerM Bool
@@ -290,6 +304,12 @@ terGetTarget = terAsks terTarget
 terSetTarget :: Maybe Target -> TerM a -> TerM a
 terSetTarget t = terLocal $ \ e -> e { terTarget = t }
 
+terGetHaveInlinedWith :: TerM Bool
+terGetHaveInlinedWith = terAsks terHaveInlinedWith
+
+terSetHaveInlinedWith :: TerM a -> TerM a
+terSetHaveInlinedWith = terLocal $ \ e -> e { terHaveInlinedWith = True }
+
 terGetDelayed :: TerM Delayed
 terGetDelayed = terAsks terDelayed
 
@@ -308,13 +328,13 @@ terGetMaskResult = terAsks terMaskResult
 terSetMaskResult :: Bool -> TerM a -> TerM a
 terSetMaskResult b = terLocal $ \ e -> e { terMaskResult = b }
 
-terGetPatterns :: TerM (MaskedDeBruijnPats)
+terGetPatterns :: TerM (MaskedDeBruijnPatterns)
 terGetPatterns = do
   n   <- terAsks terPatternsRaise
   mps <- terAsks terPatterns
-  return $ if n == 0 then mps else map (fmap (fmap (n +))) mps
+  return $ if n == 0 then mps else map (fmap (raise n)) mps
 
-terSetPatterns :: MaskedDeBruijnPats -> TerM a -> TerM a
+terSetPatterns :: MaskedDeBruijnPatterns -> TerM a -> TerM a
 terSetPatterns ps = terLocal $ \ e -> e { terPatterns = ps }
 
 terRaise :: TerM a -> TerM a
@@ -367,7 +387,10 @@ terSetUseSizeLt = terModifyUseSizeLt . const
 withUsableVars :: UsableSizeVars a => a -> TerM b -> TerM b
 withUsableVars pats m = do
   vars <- usableSizeVars pats
-  reportSLn "term.size" 20 $ "usableSizeVars = " ++ show vars
+  reportSLn "term.size" 70 $ "usableSizeVars = " ++ show vars
+  reportSDoc "term.size" 20 $ if null vars then text "no usuable size vars" else
+    text "the size variables amoung these variables are usable: " <+>
+      sep (map (prettyTCM . var) $ VarSet.toList vars)
   terSetUsableVars vars $ m
 
 -- | Set 'terUseSizeLt' when going under constructor @c@.
@@ -408,7 +431,7 @@ isProjectionButNotCoinductive qn = liftTCM $ do
         else do
           mp <- isProjection qn
           case mp of
-            Just Projection{ projProper = True, projFromType = t }
+            Just Projection{ projProper = Just{}, projFromType = t }
               -> isInductiveRecord (unArg t)
             _ -> return False
 
@@ -428,11 +451,11 @@ isCoinductiveProjection mustBeRecursive q = liftTCM $ do
   if Just q == flat then return True else do
     pdef <- getConstInfo q
     case isProjection_ (theDef pdef) of
-      Just Projection{ projProper = True, projFromType = Arg _ r, projIndex = n } ->
+      Just Projection{ projProper = Just{}, projFromType = Arg _ r, projIndex = n } ->
         caseMaybeM (isRecord r) __IMPOSSIBLE__ $ \ rdef -> do
           -- no for inductive or non-recursive record
           if recInduction rdef /= Just CoInductive then return False else do
-            reportSLn "term.guardedness" 40 $ prettyShow q ++ " is coinductive"
+            reportSLn "term.guardedness" 40 $ prettyShow q ++ " is coinductive; record type is " ++ prettyShow r
             if not mustBeRecursive then return True else do
               reportSLn "term.guardedness" 40 $ prettyShow q ++ " must be recursive"
               if not (recRecursive rdef) then return False else do
@@ -442,81 +465,45 @@ isCoinductiveProjection mustBeRecursive q = liftTCM $ do
                 -- Now check if type of field mentions mutually recursive symbol.
                 -- Get the type of the field by dropping record parameters and record argument.
                 let TelV tel core = telView' (defType pdef)
-                    tel' = drop n $ telToList tel
+                    (pars, tel') = splitAt n $ telToList tel
+                    mut = fromMaybe __IMPOSSIBLE__ $ recMutual rdef
                 -- Check if any recursive symbols appear in the record type.
                 -- Q (2014-07-01): Should we normalize the type?
-                reportSDoc "term.guardedness" 40 $ sep
-                  [ text "looking for recursive occurrences in"
-                  , prettyTCM (telFromList tel')
+                -- A (2017-01-13): Yes, since we also normalize during positivity check?
+                -- See issue #1899.
+                reportSDoc "term.guardedness" 40 $ inTopContext $ sep
+                  [ text "looking for recursive occurrences of"
+                  , sep (map prettyTCM mut)
+                  , text "in"
+                  , addContext pars $ prettyTCM (telFromList tel')
                   , text "and"
-                  , prettyTCM core
+                  , addContext tel $ prettyTCM core
                   ]
-                names <- anyDefs (r : recMutual rdef) (map (snd . unDom) tel', core)
+                when (null mut) __IMPOSSIBLE__
+                names <- anyDefs mut =<< normalise (map (snd . unDom) tel', core)
                 reportSDoc "term.guardedness" 40 $
-                  text "found" <+> sep (map prettyTCM names)
+                  text "found" <+> if null names then text "none" else sep (map prettyTCM names)
                 return $ not $ null names
       _ -> do
         reportSLn "term.guardedness" 40 $ prettyShow q ++ " is not a proper projection"
         return False
 
+-- * De Bruijn pattern stuff
 
--- * De Bruijn patterns.
-
-type DeBruijnPats = [DeBruijnPat]
-
--- | Patterns with variables as de Bruijn indices.
-type DeBruijnPat = DeBruijnPat' Int
-
-data DeBruijnPat' a
-  = VarDBP a
-    -- ^ De Bruijn Index.
-  | ConDBP QName [DeBruijnPat' a]
-    -- ^ The name refers to either an ordinary
-    --   constructor or the successor function on sized types.
-  | LitDBP Literal
-    -- ^ Literal.  Also abused to censor part of a pattern.
-  | TermDBP Term
-    -- ^ Part of dot pattern that cannot be converted into a pattern.
-  | ProjDBP ProjOrigin QName
-    -- ^ Projection pattern.
-  deriving (Functor, Show)
-
-instance IsProjP (DeBruijnPat' a) where
-  isProjP (ProjDBP o d) = Just (o, AmbQ [d])
-  isProjP _ = Nothing
-
-instance PrettyTCM DeBruijnPat where
-  prettyTCM (VarDBP i)    = prettyTCM $ var i
-  prettyTCM (ConDBP c ps) = parens $ do prettyTCM c <+> hsep (map prettyTCM ps)
-  prettyTCM (LitDBP l)    = prettyTCM l
-  prettyTCM (TermDBP v)   = parens $ prettyTCM v
-  prettyTCM (ProjDBP o d) = text "." TCP.<> prettyTCM d
-
--- | How long is the path to the deepest variable?
-patternDepth :: DeBruijnPat' a -> Int
-patternDepth p =
-  case p of
-    ConDBP _ ps -> succ $ maximum $ 0 : map patternDepth ps
-    VarDBP{}    -> 0
-    LitDBP{}    -> 0
-    TermDBP{}   -> 0
-    ProjDBP{}   -> 0
-
+-- | How long is the path to the deepest atomic pattern?
+patternDepth :: forall a. Pattern' a -> Int
+patternDepth = getMaxNat . foldrPattern depth where
+  depth :: Pattern' a -> MaxNat -> MaxNat
+  depth ConP{} = succ      -- add 1 to the maximum of the depth of the subpatterns
+  depth _      = id        -- atomic pattern (leaf) has depth 0
 
 -- | A dummy pattern used to mask a pattern that cannot be used
 --   for structural descent.
 
-unusedVar :: DeBruijnPat
-unusedVar = LitDBP (LitString noRange "term.unused.pat.var")
+unusedVar :: DeBruijnPattern
+unusedVar = LitP (LitString noRange "term.unused.pat.var")
 
--- | @raiseDBP n ps@ increases each de Bruijn index in @ps@ by @n@.
---   Needed when going under a binder during analysis of a term.
-
-raiseDBP :: Int -> DeBruijnPats -> DeBruijnPats
-raiseDBP 0 = id
-raiseDBP n = map $ fmap (n +)
-
--- | Extract variables from 'DeBruijnPat's that could witness a decrease
+-- | Extract variables from 'DeBruijnPattern's that could witness a decrease
 --   via a SIZELT constraint.
 --
 --   These variables must be under an inductive constructor (with no record
@@ -526,44 +513,46 @@ raiseDBP n = map $ fmap (n +)
 class UsableSizeVars a where
   usableSizeVars :: a -> TerM VarSet
 
-instance UsableSizeVars DeBruijnPat where
-  usableSizeVars p = do
-    let none = return mempty
-    case p of
-      VarDBP i    -> ifM terGetUseSizeLt (return $ VarSet.singleton i) {- else -} none
-      ConDBP c ps -> conUseSizeLt c $ usableSizeVars ps
-      LitDBP{}    -> none
-      TermDBP{}   -> none
-      ProjDBP{}   -> none
+instance UsableSizeVars DeBruijnPattern where
+  usableSizeVars = foldrPattern $ \case
+    VarP x    -> const $ ifM terGetUseSizeLt (return $ VarSet.singleton $ dbPatVarIndex x) $
+                   {-else-} return mempty
+    ConP c _ _ -> conUseSizeLt $ conName c
+    LitP{}     -> none
+    DotP{}     -> none
+    AbsurdP{}  -> none
+    ProjP{}    -> none
+    where none _ = return mempty
 
-instance UsableSizeVars DeBruijnPats where
+instance UsableSizeVars [DeBruijnPattern] where
   usableSizeVars ps =
     case ps of
-      []                 -> return mempty
-      (ProjDBP _ q : ps) -> projUseSizeLt q $ usableSizeVars ps
-      (p           : ps) -> mappend <$> usableSizeVars p <*> usableSizeVars ps
+      []               -> return mempty
+      (ProjP _ q : ps) -> projUseSizeLt q $ usableSizeVars ps
+      (p         : ps) -> mappend <$> usableSizeVars p <*> usableSizeVars ps
 
-instance UsableSizeVars (Masked DeBruijnPat) where
-  usableSizeVars (Masked m p) = do
-    let none = return mempty
-    case p of
-      VarDBP i    -> ifM terGetUseSizeLt (return $ VarSet.singleton i) {- else -} none
-      ConDBP c ps -> if m then none else conUseSizeLt c $ usableSizeVars ps
-      LitDBP{}    -> none
-      TermDBP{}   -> none
-      ProjDBP{}   -> none
+instance UsableSizeVars (Masked DeBruijnPattern) where
+  usableSizeVars (Masked m p) = (`foldrPattern` p) $ \case
+    VarP x    -> const $ ifM terGetUseSizeLt (return $ VarSet.singleton $ dbPatVarIndex x) $
+                   {-else-} return mempty
+    ConP c _ _ -> if m then none else conUseSizeLt $ conName c
+    LitP{}     -> none
+    DotP{}     -> none
+    AbsurdP{}  -> none
+    ProjP{}    -> none
+    where none _ = return mempty
 
-instance UsableSizeVars MaskedDeBruijnPats where
+instance UsableSizeVars MaskedDeBruijnPatterns where
   usableSizeVars ps =
     case ps of
-      []                            -> return mempty
-      (Masked _ (ProjDBP _ q) : ps) -> projUseSizeLt q $ usableSizeVars ps
-      (p                      : ps) -> mappend <$> usableSizeVars p <*> usableSizeVars ps
+      []                          -> return mempty
+      (Masked _ (ProjP _ q) : ps) -> projUseSizeLt q $ usableSizeVars ps
+      (p                    : ps) -> mappend <$> usableSizeVars p <*> usableSizeVars ps
 
 -- * Masked patterns (which are not eligible for structural descent, only for size descent)
 --   See issue #1023.
 
-type MaskedDeBruijnPats = [Masked DeBruijnPat]
+type MaskedDeBruijnPatterns = [Masked DeBruijnPattern]
 
 data Masked a = Masked
   { getMask   :: Bool  -- ^ True if thing not eligible for structural descent.
