@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP                        #-}
+{-# LANGUAGE GADTs                      #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE ImplicitParams             #-}
 {-# LANGUAGE NondecreasingIndentation   #-}
@@ -13,7 +14,7 @@
 module Agda.Termination.TermCheck
     ( termDecl
     , termMutual
-    , Result, DeBruijnPat
+    , Result
     ) where
 
 import Prelude hiding (null)
@@ -23,11 +24,10 @@ import Control.Monad.Reader
 import Control.Monad.State
 
 import Data.Foldable (toList)
-import Data.List hiding (null)
 import qualified Data.List as List
-import Data.Monoid
+import Data.Monoid hiding ((<>))
 import qualified Data.Set as Set
-import Data.Traversable (traverse)
+import Data.Traversable (Traversable, traverse)
 
 import Agda.Syntax.Abstract (IsProjP(..), AllNames(..))
 import qualified Agda.Syntax.Abstract as A
@@ -51,18 +51,18 @@ import qualified Agda.Termination.Termination  as Term
 import Agda.Termination.RecCheck
 import Agda.Termination.Inlining
 
-import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Pretty
-import Agda.TypeChecking.Reduce (reduce, normalise, instantiate, instantiateFull)
-import Agda.TypeChecking.Records -- (isRecordConstructor, isInductiveRecord)
-import Agda.TypeChecking.Telescope
-import Agda.TypeChecking.EtaContract
-import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Records
-import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Datatypes
+import Agda.TypeChecking.EtaContract
+import Agda.TypeChecking.Functions
+import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Positivity.Occurrence
+import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Records -- (isRecordConstructor, isInductiveRecord)
+import Agda.TypeChecking.Reduce (reduce, normalise, instantiate, instantiateFull)
+import Agda.TypeChecking.SizedTypes
+import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Telescope
 
 import qualified Agda.Benchmarking as Benchmark
 import Agda.TypeChecking.Monad.Benchmark (billTo, billPureTo)
@@ -162,7 +162,7 @@ termMutual
      --   (For error-reporting only.)
   -> TCM Result
 termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempty) $ {-else-}
- disableDestructiveUpdate $ do
+ inTopContext $ disableDestructiveUpdate $ do
 
   -- Get set of mutually defined names from the TCM.
   -- This includes local and auxiliary functions introduced
@@ -182,25 +182,29 @@ termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempt
         -- Recursivity checker has to see through abstract definitions!
         ignoreAbstractMode $ do
         billTo [Benchmark.Termination, Benchmark.RecCheck] $ recursive allNames
+      -- -- Andreas, 2017-03-24, use positivity info to skip non-recursive functions
+      -- skip = ignoreAbstractMode $ allM allNames $ \ x -> do
+      --   null <$> getMutual x
+      -- PROBLEMS with test/Succeed/AbstractCoinduction.agda
 
   -- We set the range to avoid panics when printing error messages.
   setCurrentRange i $ do
 
-  reportSLn "term.mutual" 10 $ "Termination checking " ++ show allNames
+  reportSLn "term.mutual" 10 $ "Termination checking " ++ prettyShow allNames
 
   -- NO_TERMINATION_CHECK
   if (Info.mutualTermCheck i `elem` [ NoTerminationCheck, Terminating ]) then do
-      reportSLn "term.warn.yes" 2 $ "Skipping termination check for " ++ show names
+      reportSLn "term.warn.yes" 2 $ "Skipping termination check for " ++ prettyShow names
       forM_ allNames $ \ q -> setTerminates q True -- considered terminating!
       return mempty
   -- NON_TERMINATING
     else if (Info.mutualTermCheck i == NonTerminating) then do
-      reportSLn "term.warn.yes" 2 $ "Considering as non-terminating: " ++ show names
+      reportSLn "term.warn.yes" 2 $ "Considering as non-terminating: " ++ prettyShow names
       forM_ allNames $ \ q -> setTerminates q False
       return mempty
   -- Trivially terminating (non-recursive)
     else ifM skip (do
-      reportSLn "term.warn.yes" 2 $ "Trivially terminating: " ++ show names
+      reportSLn "term.warn.yes" 2 $ "Trivially terminating: " ++ prettyShow names
       forM_ allNames $ \ q -> setTerminates q True
       return mempty)
    $ {- else -} do
@@ -212,7 +216,7 @@ termMutual names0 = ifNotM (optTerminationCheck <$> pragmaOptions) (return mempt
            }
          runTerm cont = runTerDefault $ do
            cutoff <- terGetCutOff
-           reportSLn "term.top" 10 $ "Termination checking " ++ show names ++
+           reportSLn "term.top" 10 $ "Termination checking " ++ prettyShow names ++
              " with cutoff=" ++ show cutoff ++ "..."
            terLocal setNames cont
 
@@ -268,18 +272,15 @@ termMutual' = do
     Left calls -> return $ singleton $ terminationError names $ callInfos calls
     Right{} -> do
       liftTCM $ reportSLn "term.warn.yes" 2 $
-        show (names) ++ " does termination check"
+        prettyShow (names) ++ " does termination check"
       return mempty
 
 -- | Smart constructor for 'TerminationError'.
 --   Removes 'termErrFunctions' that are not mentioned in 'termErrCalls'.
 terminationError :: [QName] -> [CallInfo] -> TerminationError
 terminationError names calls = TerminationError names' calls
-  where names' = names `intersect` toList (allNames calls)
+  where names' = names `List.intersect` toList (allNames calls)
 
--- ASR (08 November 2014). The type of the function could be
---
--- @Either a b -> TerM (Either a b)@.
 billToTerGraph :: a -> TerM a
 billToTerGraph a = liftTCM $ billPureTo [Benchmark.Termination, Benchmark.Graph] a
 
@@ -296,7 +297,7 @@ reportCalls no calls = do
   liftTCM $ do
 
     reportS "term.lex" 20 $ unlines
-      [ "Calls (" ++ no ++ "dot patterns): " ++ show calls
+      [ "Calls (" ++ no ++ "dot patterns): " ++ prettyShow calls
       ]
 
     -- Print the whole completion phase.
@@ -392,16 +393,16 @@ termFunction name = do
 
     names <- terGetUserNames
     case r of
-      Left calls -> return $ singleton $ terminationError ([name] `intersect` names) calls
+      Left calls -> return $ singleton $ terminationError ([name] `List.intersect` names) calls
       Right () -> do
         liftTCM $ reportSLn "term.warn.yes" 2 $
-          show name ++ " does termination check"
+          prettyShow name ++ " does termination check"
         return mempty
    where
      reportTarget r = liftTCM $
        reportSLn "term.target" 20 $ "  target type " ++
          caseMaybe r "not recognized" (\ q ->
-           "ends in " ++ show q)
+           "ends in " ++ prettyShow q)
 
 -- | To process the target type.
 typeEndsInDef :: MonadTCM tcm => Type -> tcm (Maybe QName)
@@ -471,26 +472,6 @@ setMasks t cont = do
     return (ds, d)
   terSetMaskArgs (ds ++ repeat True) $ terSetMaskResult d $ cont
 
-{- Termination check clauses:
-
-   For instance
-
-   f x (cons y nil) = g x y
-
-   Clause
-     [VarP "x", ConP "List.cons" [VarP "y", ConP "List.nil" []]]
-     Bind (Abs { absName = "x"
-               , absBody = Bind (Abs { absName = "y"
-                                     , absBody = Def "g" [ Var 1 []
-                                                         , Var 0 []]})})
-
-   Outline:
-   - create "De Bruijn pattern"
-   - collect recursive calls
-   - going under a binder, lift de Bruijn pattern
-   - compare arguments of recursive call to pattern
-
--}
 
 -- | Is the current target type among the given ones?
 
@@ -517,97 +498,107 @@ matchingTarget conf t = maybe (return True) (match t) (currentTarget conf)
 --
 --   The term is first normalized and stripped of all non-coinductive projections.
 
-termToDBP :: Term -> TerM DeBruijnPat
+termToDBP :: Term -> TerM DeBruijnPattern
 termToDBP t = ifNotM terGetUseDotPatterns (return unusedVar) $ {- else -} do
-    suc <- terGetSizeSuc
-    let
-      loop :: Term -> TCM DeBruijnPat
-      loop t = do
-        t <- constructorForm t
-        case ignoreSharing t of
-          -- Constructors.
-          Con c ci args -> ConDBP (conName c) <$> mapM (loop . unArg) args
-          Def s [Apply arg] | Just s == suc
-                      -> ConDBP s . (:[]) <$> loop (unArg arg)
-          DontCare t  -> __IMPOSSIBLE__  -- removed by stripAllProjections
-          -- Leaves.
-          Var i []    -> return $ VarDBP i
-          Lit l       -> return $ LitDBP l
-          t           -> return $ TermDBP t
-    liftTCM $ loop =<< stripAllProjections =<< normalise t
+  termToPattern =<< do liftTCM $ stripAllProjections =<< normalise t
 
+-- | Convert a term (from a dot pattern) to a pattern for the purposes of the termination checker.
+--
+--   @SIZESUC@ is treated as a constructor.
+
+class TermToPattern a b where
+  termToPattern :: a -> TerM b
+
+  default termToPattern :: (TermToPattern a' b', Traversable f, a ~ f a', b ~ f b') => a -> TerM b
+  termToPattern = traverse termToPattern
+
+instance TermToPattern a b => TermToPattern [a] [b] where
+instance TermToPattern a b => TermToPattern (Arg a) (Arg b) where
+instance TermToPattern a b => TermToPattern (Named c a) (Named c b) where
+
+-- OVERLAPPING
+-- instance TermToPattern a b => TermToPattern a (Named c b) where
+--   termToPattern t = unnamed <$> termToPattern t
+
+instance TermToPattern Term DeBruijnPattern where
+  termToPattern t = (liftTCM $ ignoreSharing <$> constructorForm t) >>= \case
+    -- Constructors.
+    Con c _ args -> ConP c noConPatternInfo . map (fmap unnamed) <$> termToPattern args
+    Def s [Apply arg] -> do
+      suc <- terGetSizeSuc
+      if Just s == suc then ConP (ConHead s Inductive []) noConPatternInfo . map (fmap unnamed) <$> termToPattern [arg]
+       else return $ DotP t
+    DontCare t  -> termToPattern t -- OR: __IMPOSSIBLE__  -- removed by stripAllProjections
+    -- Leaves.
+    Var i []    -> VarP . (`DBPatVar` i) . prettyShow <$> nameOfBV i
+    Lit l       -> return $ LitP l
+    t           -> return $ DotP t
 
--- | Masks coconstructor patterns in a deBruijn pattern.
-stripCoConstructors :: DeBruijnPat -> TerM DeBruijnPat
-stripCoConstructors p = do
-  case p of
-    ConDBP c args -> do
-      ind <- ifM ((Just c ==) <$> terGetSizeSuc) (return Inductive) {- else -}
-               (liftTCM $ whatInduction c)
-      case ind of
-        Inductive   -> ConDBP c <$> mapM stripCoConstructors args
-        CoInductive -> return unusedVar
-    -- The remaining (atomic) patterns cannot contain coconstructors, obviously.
-    VarDBP{}  -> return p
-    LitDBP{}  -> return p
-    TermDBP{} -> return p  -- Can contain coconstructors, but they do not count here.
-    ProjDBP{} -> return p
 
 -- | Masks all non-data/record type patterns if --without-K.
 --   See issue #1023.
-maskNonDataArgs :: [DeBruijnPat] -> TerM [Masked DeBruijnPat]
+maskNonDataArgs :: [DeBruijnPattern] -> TerM [Masked DeBruijnPattern]
 maskNonDataArgs ps = zipWith mask ps <$> terGetMaskArgs
   where
-    mask p@ProjDBP{} _ = Masked False p
-    mask p           d = Masked d     p
+    mask p@ProjP{} _ = Masked False p
+    mask p         d = Masked d     p
 
--- | Convert patterns to the format used by the termination checker
---   TODO: refactor termination checking to use regular patterns
-convertPatterns :: [DeBruijnPattern] -> TerM [DeBruijnPat]
-convertPatterns ps = mapM build ps
-  where
-    build :: DeBruijnPattern -> TerM DeBruijnPat
-    build (VarP x)        = return $ VarDBP $ dbPatVarIndex x
-    build (ConP con _ ps) = ConDBP (conName con) <$> mapM (build . namedArg) ps
-    build (DotP t)        = termToDBP t
-    build (LitP l)        = return $ LitDBP l
-    build (ProjP o d)     = return $ ProjDBP o d
 
 -- | Extract recursive calls from one clause.
+
 termClause :: Clause -> TerM Calls
 termClause clause = do
-  ifNotM (terGetInlineWithFunctions) (termClause' clause) $ {- else -} do
+
+  -- If with-function inlining is disallowed (e.g. --without-K),
+  -- we check the original clause.
+
+  let fallback = termClause' clause
+  ifNotM (terGetInlineWithFunctions) fallback $ {- else -} do
+
+    -- Otherwise, we will do inlining, hence, can skip with-generated functions.
+
     name <- terGetCurrent
-    ifM (isJust <$> do isWithFunction name) (return mempty) $
-      mapM' termClause' =<< do liftTCM $ inlineWithClauses name clause
+    ifM (isJust <$> isWithFunction name) (return mempty) $ {- else -} do
+
+      -- With inlining, the termination check for all subordinated
+      -- with-functions is included in the parent function.
+
+      (liftTCM $ inlineWithClauses name clause) >>= \case
+        Nothing  -> fallback
+        Just cls -> terSetHaveInlinedWith $ mapM' termClause' cls
 
 termClause' :: Clause -> TerM Calls
 termClause' clause = do
-  cl <- introHiddenLambdas clause
-  let tel      = clauseTel cl
-      argPats' = namedClausePats cl
-      body     = clauseBody cl
+  Clause{ clauseTel = tel, namedClausePats = ps, clauseBody = body } <- etaExpandClause clause
   liftTCM $ reportSDoc "term.check.clause" 25 $ vcat
     [ text "termClause"
-    , nest 2 $ text "tel      =" <+> prettyTCM tel
-    , nest 2 $ text "argPats' =" <+> do
-      addContext tel $ do
-       aps <- reifyPatterns argPats'
-       fsep $ map prettyA aps
+    , nest 2 $ text "tel =" <+> prettyTCM tel
+    , nest 2 $ text "ps  =" <+> do addContext tel $ prettyTCMPatternList ps
     ]
-  addContext tel $ do
-    dbpats <- convertPatterns $ map namedArg argPats'
-    case body of
-      Nothing -> return empty
-      Just v -> do
-        dbpats <- mapM stripCoConstructors dbpats
-        mdbpats <- maskNonDataArgs dbpats
-        terSetPatterns mdbpats $ do
-          terSetSizeDepth tel $ do
-            reportBody v
-            extract v
+  forM' body $ \ v -> addContext tel $ do
+    -- TODO: combine the following two traversals, avoid full normalisation.
+    -- Parse dot patterns as patterns as far as possible.
+    ps <- postTraversePatternM parseDotP ps
+    -- Blank out coconstructors.
+    ps <- preTraversePatternM stripCoCon ps
+    -- Mask non-data arguments.
+    mdbpats <- maskNonDataArgs $ map namedArg ps
+    terSetPatterns mdbpats $ do
+      terSetSizeDepth tel $ do
+        reportBody v
+        extract v
 
   where
+    parseDotP = \case
+      DotP t -> termToDBP t
+      p      -> return p
+    stripCoCon p = case p of
+      ConP (ConHead c _ _) _ _ -> do
+        ifM ((Just c ==) <$> terGetSizeSuc) (return p) $ {- else -} do
+        whatInduction c >>= \case
+          Inductive   -> return p
+          CoInductive -> return unusedVar
+      _ -> return p
     reportBody :: Term -> TerM ()
     reportBody v = verboseS "term.check.clause" 6 $ do
       f       <- terGetCurrent
@@ -618,45 +609,10 @@ termClause' clause = do
                     (if delayed == Delayed then "delayed " else "") ++
                     "clause of")
                 <+> prettyTCM f
-            , nest 2 $ text "lhs:" <+> hsep (map prettyTCM pats)
+            , nest 2 $ text "lhs:" <+> sep (map prettyTCM pats)
             , nest 2 $ text "rhs:" <+> prettyTCM v
             ]
 
--- | Rewrite a clause @f ps =tel= \ {xs} -> v@ to @f ps {xs} =(tel {xs})= v@.
---   The pupose is to move hidden size quantifications
---   to the lhs such that the termination checker can make use of them.
---   See, e.g., test/succeed/SizedTypesExtendedLambda.agda.
-introHiddenLambdas :: MonadTCM tcm => Clause -> tcm Clause
-introHiddenLambdas clause = liftTCM $ do
-  case clause of
-    Clause range ctel ps body        Nothing  catchall -> return clause
-
-    Clause range ctel ps Nothing     (Just t) catchall -> return clause
-    Clause range ctel ps (Just body) (Just t) catchall -> do
-      case removeHiddenLambdas body of
-        -- nobody or no hidden lambdas
-        ([], _) -> return clause
-        -- hidden lambdas
-        (axs, body') -> do
-          -- n = number of hidden lambdas
-          let n = length axs
-          -- take n abstractions from rhs type
-          TelV ttel t' <- telViewUpTo n $ unArg t
-          when (size ttel < n) __IMPOSSIBLE__
-          -- join with lhs telescope
-          let ctel' = telFromList $ telToList ctel ++ telToList ttel
-              ps'   = raise n ps ++ zipWith toPat (downFrom $ size axs) axs
-          return $ Clause range ctel' ps' (Just body') (Just (t $> t')) catchall
-  where
-    toPat i (Arg info x) = Arg info $ namedDBVarP i x
-
-    removeHiddenLambdas :: Term -> ([Arg ArgName], Term)
-    removeHiddenLambdas v =
-      case ignoreSharing v of
-        Lam info b | getHiding info == Hidden ->
-          let (xs, b') = removeHiddenLambdas $ unAbs b
-          in  (Arg info (absName b) : xs, b')
-        _ -> ([], v)
 
 -- | Extract recursive calls from expressions.
 class ExtractCalls a where
@@ -721,14 +677,12 @@ constructor
 constructor c ind args = do
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
-  mapM' loopArg args
-  where
-    loopArg (arg, preserves) = terModifyGuarded g' $ extract arg where
-      g' = case (preserves, ind) of
+  forM' args $ \ (arg, preserves) -> do
+    let g' = case (preserves, ind) of
              (True,  Inductive)   -> id
              (True,  CoInductive) -> (Order.lt .*.)
              (False, _)           -> const Order.unknown
-
+    terModifyGuarded g' $ extract arg
 
 -- | Handle guardedness preserving type constructor.
 
@@ -763,16 +717,16 @@ withFunction g es = do
 -- | Handles function applications @g es@.
 
 function :: QName -> Elims -> TerM Calls
-function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunction g) (withFunction g es)
+function g es0 = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunction g) (withFunction g es0)
   $ {-else, no with function-} do
 
     f       <- terGetCurrent
     names   <- terGetMutual
     guarded <- terGetGuarded
 
-    let gArgs = Def g es
+    -- let gArgs = Def g es0
     liftTCM $ reportSDoc "term.function" 30 $
-      text "termination checking function call " <+> prettyTCM gArgs
+      text "termination checking function call " <+> prettyTCM (Def g es0)
 
     -- First, look for calls in the arguments of the call gArgs.
 
@@ -780,7 +734,7 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
     -- Andreas, Issue 1530: constructors have to be reduced deep inside terms,
     -- thus, we need to use traverseTermM.  Sharing is handled by traverseTermM,
     -- so no ignoreSharing needed here.
-    let reduceCon = traverseTermM $ \ t -> case t of
+    let (reduceCon :: Term -> TCM Term) = traverseTermM $ \ t -> case t of
            Con c ci vs -> (`apply` vs) <$> reduce (Con c ci [])  -- make sure we don't reduce the arguments
            _ -> return t
 
@@ -794,7 +748,7 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
     let unguards = repeat Order.unknown
     let guards = applyWhen isProj (guarded :) unguards
     -- Collect calls in the arguments of this call.
-    let args = map unArg $ argsFromElims es
+    let args = map unArg $ argsFromElims es0
     calls <- forM' (zip guards args) $ \ (guard, a) -> do
       terSetGuarded guard $ extract a
 
@@ -815,10 +769,26 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
        Just gInd -> do
          delayed <- terGetDelayed
          pats    <- terGetPatterns
-         -- 2014-03-25 Andreas, the costs seem small, benchmark turned off.
-         es <- liftTCM $ -- billTo [Benchmark.Termination, Benchmark.Reduce] $
-           forM es $
-              etaContract <=< traverse reduceCon <=< instantiateFull
+         -- Andreas, 2017-02-14, issue #2458:
+         -- If we have inlined with-functions, we could be illtyped,
+         -- hence, do not reduce anything.
+         -- Andreas, 2017-06-20 issue #2613:
+         -- We still need to reduce constructors, even when with-inlining happened.
+         es <- -- ifM terGetHaveInlinedWith (return es0) {-else-} $
+           liftTCM $ forM es0 $
+             etaContract <=< traverse reduceCon <=< instantiateFull
+           -- 2017-05-16, issue #2403: Argument normalization is too expensive,
+           -- even if we only expand non-recursive functions.
+           -- Argument normalization TURNED OFF.
+           -- liftTCM $ billTo [Benchmark.Termination, Benchmark.Reduce] $ do
+           --  -- Andreas, 2017-01-13, issue #2403, normalize arguments for the structural ordering.
+           --  -- Andreas, 2017-03-25, issue #2495, restrict this to non-recursive functions
+           --  -- otherwise, the termination checking may run forever.
+           --  reportSLn "term.reduce" 90 $ "normalizing call arguments"
+           --  modifyAllowedReductions (List.\\ [UnconfirmedReductions,RecursiveReductions]) $
+           --    forM es0 $ \ e -> do
+           --      reportSDoc "term.reduce" 95 $ text "normalizing " <+> prettyTCM e
+           --      etaContract =<< normalise e
 
          -- Compute the call matrix.
 
@@ -849,7 +819,10 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
          -- Andreas, 2015-01-21 Issue 1410: Go to the module where g is defined
          -- otherwise its free variables with be prepended to the call
          -- in the error message.
-         doc <- liftTCM $ withCurrentModule (qnameModule g) $ buildClosure gArgs
+         doc <- liftTCM $ withCurrentModule (qnameModule g) $ buildClosure $
+           Def g $ filter ((/= Inserted) . getOrigin) es0
+           -- Andreas, 2017-01-05, issue #2376
+           -- Remove arguments inserted by etaExpandClause.
 
          let src  = fromMaybe __IMPOSSIBLE__ $ List.elemIndex f names
              tgt  = gInd
@@ -860,8 +833,8 @@ function g es = ifM (terGetInlineWithFunctions `and2M` do isJust <$> isWithFunct
                       , callInfoCall   = doc
                       }]
          liftTCM $ reportSDoc "term.kept.call" 5 $ vcat
-           [ text "kept call from" <+> text (show f) <+> hsep (map prettyTCM pats)
-           , nest 2 $ text "to" <+> text (show g) <+>
+           [ text "kept call from" <+> text (prettyShow f) <+> hsep (map prettyTCM pats)
+           , nest 2 $ text "to" <+> text (prettyShow g) <+>
                        hsep (map (parens . prettyTCM) args)
            , nest 2 $ text "call matrix (with guardedness): "
            , nest 2 $ pretty cm
@@ -889,13 +862,11 @@ instance ExtractCalls Term where
         -- in a type mutual with the current target
         -- then we count it as guarding.
         ind <- ifM ((Just c ==) <$> terGetSharp) (return CoInductive) $ do
-          r <- liftTCM $ isRecordConstructor c
-          case r of
-            Nothing       -> return Inductive
-            Just (q, def) -> (\ b -> if b then CoInductive else Inductive) <$>
-              andM [ return $ recRecursive def
-                   , return $ recInduction def == Just CoInductive
-                   , targetElem (q : recMutual def)
+          caseMaybeM (liftTCM $ isRecordConstructor c) (return Inductive) $ \ (q, def) -> do
+            reportSLn "term.check.term" 50 $ "constructor " ++ prettyShow c ++ " has record type " ++ prettyShow q
+            (\ b -> if b then CoInductive else Inductive) <$>
+              andM [ return $ recInduction def == Just CoInductive
+                   , targetElem . fromMaybe __IMPOSSIBLE__ $ recMutual def
                    ]
         constructor c ind argsg
 
@@ -975,8 +946,11 @@ maskSizeLt dom@(Dom info a) = liftTCM $ do
      The guardedness is the number of projection patterns in @pats@
      minus the number of projections in @es@.
  -}
-compareArgs :: (Integral n) => [Elim] -> TerM (n, n, [[Order]])
+compareArgs :: [Elim] -> TerM (Int, Int, [[Order]])
 compareArgs es = do
+  liftTCM $ reportSDoc "term.compareArgs" 90 $ vcat
+    [ text $ "comparing " ++ show (length es) ++ " args"
+    ]
   pats <- terGetPatterns
   -- apats <- annotatePatsWithUseSizeLt pats
   -- reportSDoc "term.compare" 20 $
@@ -986,18 +960,18 @@ compareArgs es = do
 
   -- Count the number of coinductive projection(pattern)s in caller and callee.
   -- Only recursive coinductive projections are eligible (Issue 1209).
-  projsCaller <- genericLength <$> do
+  projsCaller <- length <$> do
     filterM (isCoinductiveProjection True) $ mapMaybe (fmap (head . unAmbQ . snd) . isProjP . getMasked) pats
-  projsCallee <- genericLength <$> do
+  projsCallee <- length <$> do
     filterM (isCoinductiveProjection True) $ mapMaybe (fmap snd . isProjElim) es
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
-  let guardedness = decr $ projsCaller - projsCallee
+  let guardedness = decr True $ projsCaller - projsCallee
   liftTCM $ reportSDoc "term.guardedness" 30 $ sep
     [ text "compareArgs:"
-    , nest 2 $ text $ "projsCaller = " ++ show projsCaller
-    , nest 2 $ text $ "projsCallee = " ++ show projsCallee
-    , nest 2 $ text $ "guardedness of call: " ++ show guardedness
+    , nest 2 $ text $ "projsCaller = " ++ prettyShow projsCaller
+    , nest 2 $ text $ "projsCallee = " ++ prettyShow projsCallee
+    , nest 2 $ text $ "guardedness of call: " ++ prettyShow guardedness
     ]
   return $ addGuardedness guardedness (size es) (size pats) matrix
 
@@ -1008,35 +982,41 @@ compareArgs es = do
 --   off, if inductive.
 --
 --   UNUSED
-annotatePatsWithUseSizeLt :: [DeBruijnPat] -> TerM [(Bool,DeBruijnPat)]
+annotatePatsWithUseSizeLt :: [DeBruijnPattern] -> TerM [(Bool,DeBruijnPattern)]
 annotatePatsWithUseSizeLt = loop where
   loop [] = return []
-  loop (p@(ProjDBP _ q) : pats) = ((False,p) :) <$> do projUseSizeLt q $ loop pats
+  loop (p@(ProjP _ q) : pats) = ((False,p) :) <$> do projUseSizeLt q $ loop pats
   loop (p : pats) = (\ b ps -> (b,p) : ps) <$> terGetUseSizeLt <*> loop pats
 
 
 -- | @compareElim e dbpat@
 
-compareElim :: Elim -> Masked DeBruijnPat -> TerM Order
+compareElim :: Elim -> Masked DeBruijnPattern -> TerM Order
 compareElim e p = do
   liftTCM $ do
     reportSDoc "term.compare" 30 $ sep
       [ text "compareElim"
-      , nest 2 $ text "e = " <+> prettyTCM e
-      , nest 2 $ text "p = " <+> prettyTCM p
+      , nest 2 $ text "e = " <> prettyTCM e
+      , nest 2 $ text "p = " <> prettyTCM p
       ]
     reportSDoc "term.compare" 50 $ sep
       [ nest 2 $ text $ "e = " ++ show e
       , nest 2 $ text $ "p = " ++ show p
       ]
   case (e, getMasked p) of
-    (Proj _ d, ProjDBP _ d')       -> do
-      d  <- liftTCM $ getOriginalProjection d
-      d' <- liftTCM $ getOriginalProjection d'
-      compareProj d d'
-    (Proj{}, _         )           -> return Order.unknown
-    (Apply{}, ProjDBP{})           -> return Order.unknown
-    (Apply arg, _)                 -> compareTerm (unArg arg) p
+    (Proj _ d, ProjP _ d') -> do
+      d  <- getOriginalProjection d
+      d' <- getOriginalProjection d'
+      o  <- compareProj d d'
+      reportSDoc "term.compare" 30 $ sep
+        [ text $ "comparing callee projection " ++ prettyShow d
+        , text $ "against caller projection " ++ prettyShow d'
+        , text $ "yields order " ++ prettyShow o
+        ]
+      return o
+    (Proj{}, _)            -> return Order.unknown
+    (Apply{}, ProjP{})     -> return Order.unknown
+    (Apply arg, _)         -> compareTerm (unArg arg) p
 
 -- | In dependent records, the types of later fields may depend on the
 --   values of earlier fields.  Thus when defining an inhabitant of a
@@ -1061,7 +1041,7 @@ compareProj d d'
           case def of
             Record{ recFields = fs } -> do
               fs <- return $ map unArg fs
-              case (find (d==) fs, find (d'==) fs) of
+              case (List.find (d==) fs, List.find (d'==) fs) of
                 (Just i, Just i')
                   -- earlier field is smaller
                   | i < i'    -> return Order.lt
@@ -1084,10 +1064,10 @@ addGuardedness g nrows ncols m = (nrows, ncols, m)
 -}
 
 -- | 'addGuardedness' adds guardedness flag in the upper left corner (0,0).
-addGuardedness :: Integral n => Order -> n -> n -> [[Order]] -> (n, n, [[Order]])
+addGuardedness :: Order -> Int -> Int -> [[Order]] -> (Int, Int, [[Order]])
 addGuardedness o nrows ncols m =
   (nrows + 1, ncols + 1,
-   (o : genericReplicate ncols Order.unknown) : map (Order.unknown :) m)
+   (o : replicate ncols Order.unknown) : map (Order.unknown :) m)
 
 -- | Compose something with the upper-left corner of a call matrix
 composeGuardedness :: (?cutoff :: CutOff) => Order -> [[Order]] -> [[Order]]
@@ -1101,16 +1081,18 @@ offsetFromConstructor :: MonadTCM tcm => QName -> tcm Int
 offsetFromConstructor c = maybe 1 (const 0) <$> do
   liftTCM $ isRecordConstructor c
 
--- | Compute the proper subpatterns of a 'DeBruijnPat'.
-subPatterns :: DeBruijnPat -> [DeBruijnPat]
-subPatterns p = case p of
-  ConDBP c ps -> ps ++ concatMap subPatterns ps
-  VarDBP _    -> []
-  LitDBP _    -> []
-  TermDBP _   -> []
-  ProjDBP{}   -> []
+-- | Compute the proper subpatterns of a 'DeBruijnPattern'.
+subPatterns :: DeBruijnPattern -> [DeBruijnPattern]
+subPatterns = foldPattern $ \case
+  ConP _ _ ps -> map namedArg ps
+  VarP _      -> mempty
+  LitP _      -> mempty
+  DotP _      -> mempty
+  AbsurdP _   -> mempty
+  ProjP _ _   -> mempty
 
-compareTerm :: Term -> Masked DeBruijnPat -> TerM Order
+
+compareTerm :: Term -> Masked DeBruijnPattern -> TerM Order
 compareTerm t p = do
 --   reportSDoc "term.compare" 25 $
 --     text " comparing term " <+> prettyTCM t <+>
@@ -1120,13 +1102,8 @@ compareTerm t p = do
   liftTCM $ reportSDoc "term.compare" 25 $
     text " comparing term " <+> prettyTCM t <+>
     text " to pattern " <+> prettyTCM p <+>
-    text (" results in " ++ show o)
+    text (" results in " ++ prettyShow o)
   return o
-{-
-compareTerm t p = Order.supremum $ compareTerm' t p : map cmp (subPatterns p)
-  where
-    cmp p' = (Order..*.) Order.lt (compareTerm' t p')
--}
 
 
 -- | Remove all non-coinductive projections from an algebraic term
@@ -1137,14 +1114,6 @@ class StripAllProjections a where
 
 instance StripAllProjections a => StripAllProjections (Arg a) where
   stripAllProjections = traverse stripAllProjections
-  -- stripAllProjections (Arg info a) = Arg info <$> stripAllProjections a
-
-{- DOES NOT WORK, since s.th. special is needed for Elims
-instance StripAllProjections a => StripAllProjections [a] where
-  stripAllProjections = traverse stripAllProjections
-
-instance StripAllProjections a => StripAllProjections (Elim' a) where
--}
 
 instance StripAllProjections Elims where
   stripAllProjections es =
@@ -1172,7 +1141,7 @@ instance StripAllProjections Term where
 --
 --   Precondition: top meta variable resolved
 
-compareTerm' :: Term -> Masked DeBruijnPat -> TerM Order
+compareTerm' :: Term -> Masked DeBruijnPattern -> TerM Order
 compareTerm' v mp@(Masked m p) = do
   suc  <- terGetSizeSuc
   cutoff <- terGetCutOff
@@ -1196,13 +1165,13 @@ compareTerm' v mp@(Masked m p) = do
     -- deepest variable in @p@.
     -- For sized types, the depth is maximally
     -- the number of SIZELT hypotheses one can have in a context.
-    (MetaV{}, p) -> Order.decr . max (if m then 0 else patternDepth p) . pred <$>
+    (MetaV{}, p) -> Order.decr True . max (if m then 0 else patternDepth p) . pred <$>
        terAsks _terSizeDepth
 
     -- Successor on both sides cancel each other.
     -- We ignore the mask for sizes.
-    (Def s [Apply t], ConDBP s' [p]) | s == s' && Just s == suc ->
-      compareTerm' (unArg t) (notMasked p)
+    (Def s [Apply t], ConP s' _ [p]) | s == conName s' && Just s == suc ->
+      compareTerm' (unArg t) (notMasked $ namedArg p)
 
     -- Register also size increase.
     (Def s [Apply t], p) | Just s == suc ->
@@ -1214,7 +1183,7 @@ compareTerm' v mp@(Masked m p) = do
 
     _ | m -> return Order.unknown
 
-    (Lit l, LitDBP l')
+    (Lit l, LitP l')
       | l == l'     -> return Order.le
       | otherwise   -> return Order.unknown
 
@@ -1226,47 +1195,48 @@ compareTerm' v mp@(Masked m p) = do
 
     -- Andreas, 2011-04-19 give subterm priority over matrix order
 
-    (Con{}, ConDBP c ps) | any (isSubTerm v) ps ->
-      decrease <$> offsetFromConstructor c <*> return Order.le
+    (Con{}, ConP c _ ps) | any (isSubTerm v . namedArg) ps ->
+      decr True <$> offsetFromConstructor (conName c)
 
-    (Con c ci ts, ConDBP c' ps) | conName c == c'->
+    (Con c _ ts, ConP c' _ ps) | conName c == conName c'->
       compareConArgs ts ps
 
-    (Con c ci [], _) -> return Order.le
+    (Con _ _ [], _) -> return Order.le
 
     -- new case for counting constructors / projections
     -- register also increase
-    (Con c ci ts, _) -> do
+    (Con c _ ts, _) -> do
       increase <$> offsetFromConstructor (conName c)
                <*> (infimum <$> mapM (\ t -> compareTerm' (unArg t) mp) ts)
 
     (t, p) -> return $ subTerm t p
 
 -- | @subTerm@ computes a size difference (Order)
-subTerm :: (?cutoff :: CutOff) => Term -> DeBruijnPat -> Order
+subTerm :: (?cutoff :: CutOff) => Term -> DeBruijnPattern -> Order
 subTerm t p = if equal t p then Order.le else properSubTerm t p
   where
     equal (Shared p) dbp = equal (derefPtr p) dbp
-    equal (Con c ci ts) (ConDBP c' ps) =
-      and $ (conName c == c')
+    equal (Con c _ ts) (ConP c' _ ps) =
+      and $ (conName c == conName c')
           : (length ts == length ps)
-          : zipWith equal (map unArg ts) ps
-    equal (Var i []) (VarDBP i') = i == i'
-    equal (Lit l)    (LitDBP l') = l == l'
+          : zipWith (\ t p -> equal (unArg t) (namedArg p)) ts ps
+    equal (Var i []) (VarP x) = i == dbPatVarIndex x
+    equal (Lit l)    (LitP l') = l == l'
     -- Terms.
     -- Checking for identity here is very fragile.
     -- However, we cannot do much more, as we are not allowed to normalize t.
     -- (It might diverge, and we are just in the process of termination checking.)
-    equal t         (TermDBP t') = t == t'
+    equal t         (DotP t') = t == t'
     equal _ _ = False
 
-    properSubTerm t (ConDBP _ ps) = decrease 1 $ supremum $ map (subTerm t) ps
+    properSubTerm t (ConP _ _ ps) =
+      setUsability True $ decrease 1 $ supremum $ map (subTerm t . namedArg) ps
     properSubTerm _ _ = Order.unknown
 
-isSubTerm :: (?cutoff :: CutOff) => Term -> DeBruijnPat -> Bool
+isSubTerm :: (?cutoff :: CutOff) => Term -> DeBruijnPattern -> Bool
 isSubTerm t p = nonIncreasing $ subTerm t p
 
-compareConArgs :: Args -> [DeBruijnPat] -> TerM Order
+compareConArgs :: Args -> [NamedArg DeBruijnPattern] -> TerM Order
 compareConArgs ts ps = do
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
@@ -1276,9 +1246,9 @@ compareConArgs ts ps = do
     (0,0) -> return Order.le        -- c <= c
     (0,1) -> return Order.unknown   -- c not<= c x
     (1,0) -> __IMPOSSIBLE__
-    (1,1) -> compareTerm' (unArg (head ts)) (notMasked (head ps))
+    (1,1) -> compareTerm' (unArg (head ts)) (notMasked $ namedArg $ head ps)
     (_,_) -> foldl (Order..*.) Order.le <$>
-               zipWithM compareTerm' (map unArg ts) (map notMasked ps)
+               zipWithM compareTerm' (map unArg ts) (map (notMasked . namedArg) ps)
        -- corresponds to taking the size, not the height
        -- allows examples like (x, y) < (Succ x, y)
 {- version which does an "order matrix"
@@ -1287,7 +1257,7 @@ compareConArgs ts ps = do
    -- Trigges issue 787.
         (_,_) -> do -- build "call matrix"
           m <- mapM (\t -> mapM (compareTerm' suc (unArg t)) ps) ts
-          let m2 = makeCM (genericLength ps) (genericLength ts) m
+          let m2 = makeCM (length ps) (length ts) m
           return $ Order.orderMat (Order.mat m2)
 -}
 {- version which takes height
@@ -1295,34 +1265,45 @@ compareConArgs ts ps = do
 --               else Order.infimum (zipWith compareTerm' (map unArg ts) ps)
 -}
 
-compareVar :: Nat -> Masked DeBruijnPat -> TerM Order
+compareVar :: Nat -> Masked DeBruijnPattern -> TerM Order
 compareVar i (Masked m p) = do
   suc    <- terGetSizeSuc
   cutoff <- terGetCutOff
   let ?cutoff = cutoff
   let no = return Order.unknown
   case p of
-    ProjDBP{}   -> no
-    LitDBP{}    -> no
-    TermDBP{}   -> no
-    VarDBP j    -> compareVarVar i (Masked m j)
-    ConDBP s [p] | Just s == suc -> decrease 1 <$> compareVar i (notMasked p)
-    ConDBP c ps -> if m then no else do
-      decrease <$> offsetFromConstructor c
-               <*> (Order.supremum <$> mapM (compareVar i . notMasked) ps)
+    ProjP{}   -> no
+    LitP{}    -> no
+    DotP{}   -> no
+    AbsurdP{} -> no
+    VarP x    -> compareVarVar i (Masked m x)
+
+    ConP s _ [p] | Just (conName s) == suc ->
+      setUsability True . decrease 1 <$> compareVar i (notMasked $ namedArg p)
+
+    ConP c _ ps -> if m then no else setUsability True <$> do
+      decrease <$> offsetFromConstructor (conName c)
+               <*> (Order.supremum <$> mapM (compareVar i . notMasked . namedArg) ps)
 
 -- | Compare two variables.
 --
 --   The first variable comes from a term, the second from a pattern.
-compareVarVar :: Nat -> Masked Nat -> TerM Order
-compareVarVar i (Masked m j)
-  | i == j    = if not m then return Order.le else liftTCM $
+compareVarVar :: Nat -> Masked DBPatVar -> TerM Order
+compareVarVar i (Masked m x@(DBPatVar _ j))
+  | i == j = if not m then return Order.le else liftTCM $
       -- If j is a size, we ignore the mask.
       ifM (isJust <$> do isSizeType =<< reduce =<< typeOfBV j)
         {- then -} (return Order.le)
         {- else -} (return Order.unknown)
-  | otherwise = ifNotM ((i `VarSet.member`) <$> terGetUsableVars) (return Order.unknown) $ {- else -} do
+  | otherwise = do
+      -- record usability of variable
+      u <- (i `VarSet.member`) <$> terGetUsableVars
+      -- Andreas, 2017-07-26, issue #2331.
+      -- The usability logic is refuted by bounded size quantification in terms.
+      -- Thus, it is switched off (the infrastructure remains in place for now).
+      if not u then return Order.unknown else do
+      -- Only if usable:
       res <- isBounded i
       case res of
         BoundedNo  -> return Order.unknown
-        BoundedLt v -> decrease 1 <$> compareTerm' v (Masked  m (VarDBP j))
+        BoundedLt v -> setUsability u . decrease 1 <$> compareTerm' v (Masked m $ VarP x)
