@@ -55,9 +55,10 @@ piAbstractTerm v a b = do
 
 piAbstract :: (Term, EqualityView) -> Type -> TCM Type
 piAbstract (v, OtherType a) b = piAbstractTerm v a b
-piAbstract (prf, eqt@(EqualityType s _ _ a v _)) b = do
+piAbstract (prf, eqt@(EqualityType _ _ _ (Arg _ a) v _)) b = do
+  s <- inferSort a
   let prfTy = equalityUnview eqt
-      vTy   = El s (unArg a)
+      vTy   = El s a
   b <- abstractType prfTy prf b
   b <- addContext ("w", defaultDom prfTy) $ abstractType (raise 1 vTy) (unArg $ raise 1 v) b
   return . funType vTy . funType eqTy' . swap01 $ b
@@ -108,7 +109,7 @@ abstractTerm a u@Con{} b v = do
   let n = length args
 
   let abstr b v = do
-        m <- size <$> getContext
+        m <- getContextSize
         let (a', u') = raise (m - n) (a, u)
         case isPrefixOf u' v of
           Nothing -> return v
@@ -129,6 +130,11 @@ abstractTerm a u@Con{} b v = do
                , nest 2 $ (prettyTCM =<< getContextTelescope) <+> text "⊢"
                , nest 2 $ sep [ prettyTCM v <+> text ":", nest 2 $ prettyTCM b ]
                , nest 2 $ prettyTCM err ]
+        reportSDoc "impossible" 60 $
+          vcat [ text "Type error in term to abstract (raw)"
+               , nest 2 $ ((text . show) =<< getContextTelescope) <+> text "⊢"
+               , nest 2 $ sep [ (text . show) v <+> text ":", nest 2 $ (text . show) b ]
+               ]
         __IMPOSSIBLE__
   reportSDoc "tc.abstract" 50 $ sep [ text "Resulting abstraction", nest 2 $ prettyTCM v ]
   modifySignature $ updateDefinitions $ HMap.delete hole
