@@ -12,8 +12,9 @@ module Agda.TypeChecking.CheckInternal
   ( checkType
   , checkInternal
   , checkInternal'
-  , defaultAction, Action(..)
+  , Action(..), defaultAction, eraseUnusedAction
   , infer
+  , inferSort
   ) where
 
 import Control.Arrow ((&&&), (***), first, second)
@@ -24,7 +25,7 @@ import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Conversion
-import Agda.TypeChecking.Datatypes (getConType)
+import Agda.TypeChecking.Datatypes -- (getConType, getFullyAppliedConType)
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
@@ -42,6 +43,8 @@ import Agda.Utils.Size
 #include "undefined.h"
 import Agda.Utils.Impossible
 
+-- * Bidirectional rechecker
+
 -- -- | Entry point for e.g. checking WithFunctionType.
 -- checkType :: Type -> TCM ()
 -- checkType t = -- dontAssignMetas $ ignoreSorts $
@@ -96,11 +99,43 @@ checkType' t = do
 checkTypeSpine :: Type -> Term -> Elims -> TCM Sort
 checkTypeSpine a self es = shouldBeSort =<< do snd <$> inferSpine a self es
 
-data Action = Action { preAction  :: Type -> Term -> TCM Term
-                     , postAction :: Type -> Term -> TCM Term }
 
+-- | 'checkInternal' traverses the whole 'Term', and we can use this
+--   traversal to modify the term.
+data Action = Action
+  { preAction  :: Type -> Term -> TCM Term
+    -- ^ Called on each subterm before the checker runs.
+  , postAction :: Type -> Term -> TCM Term
+    -- ^ Called on each subterm after the type checking.
+  , relevanceAction :: Relevance -> Relevance -> Relevance
+    -- ^ Called for each @ArgInfo@.
+    --   The first 'Relevance' is from the type,
+    --   the second from the term.
+  }
+
+-- | The default action is to not change the 'Term' at all.
 defaultAction :: Action
-defaultAction = Action (\ _ v -> return v) (\ _ v -> return v)
+defaultAction = Action
+  { preAction       = \ _ -> return
+  , postAction      = \ _ -> return
+  , relevanceAction = \ _ -> id
+  }
+
+eraseUnusedAction :: Action
+eraseUnusedAction = defaultAction { postAction = eraseUnused }
+  where
+    eraseUnused :: Type -> Term -> TCM Term
+    eraseUnused t v = case ignoreSharing v of
+      Def f es -> do
+        pols <- getPolarity f
+        return $ Def f $ eraseIfNonvariant pols es
+      _        -> return v
+
+    eraseIfNonvariant :: [Polarity] -> Elims -> Elims
+    eraseIfNonvariant []                  es             = es
+    eraseIfNonvariant pols                []             = []
+    eraseIfNonvariant (Nonvariant : pols) (e : es) = (fmap DontCare e) : eraseIfNonvariant pols es
+    eraseIfNonvariant (_          : pols) (e : es) = e : eraseIfNonvariant pols es
 
 -- | Entry point for term checking.
 checkInternal :: Term -> Type -> TCM ()
@@ -126,20 +161,16 @@ checkInternal' action v t = do
       a <- metaType x
       checkSpine action a (MetaV x []) es t
     Con c ci vs -> do
-      -- we need to fully apply the constructor to make getConType work
-      TelV tel t <- telView t
-      addContext tel $ do
-        let failure = typeError $ DoesNotConstructAnElementOf (conName c) t
-            vs'     = raise (size tel) vs ++ teleArgs tel
-        a <- maybe failure return =<< getConType c t
+      -- We need to fully apply the constructor to make getConType work!
+      fullyApplyCon c vs t $ \ _d _dt _pars a vs' tel t -> do
         Con c ci vs2 <- checkArgs action a (Con c ci []) vs' t
-                 -- Strip away the extra arguments
+        -- Strip away the extra arguments
         return $ applySubst (strengthenS __IMPOSSIBLE__ (size tel))
-               $ Con c ci (take (length vs) vs2)
+          $ Con c ci $ take (length vs) vs2
     Lit l      -> Lit l <$ ((`subtype` t) =<< litType l)
     Lam ai vb  -> do
       (a, b) <- shouldBePi t
-      checkArgInfo ai $ domInfo a
+      ai <- checkArgInfo action ai $ domInfo a
       addContext (suggest vb b, a) $ do
         Lam ai . Abs (absName vb) <$> checkInternal' action (absBody vb) (absBody b)
     Pi a b     -> do
@@ -165,7 +196,41 @@ checkInternal' action v t = do
     DontCare v -> DontCare <$> checkInternal' action v t
     Shared{}   -> __IMPOSSIBLE__
 
-checkSpine :: Action -> Type -> Term -> Elims -> Type -> TCM Term
+-- | Make sure a constructor is fully applied
+--   and infer the type of the constructor.
+--   Raises a type error if the constructor does not belong to the given type.
+fullyApplyCon
+  :: ConHead -- ^ Constructor.
+  -> Args    -- ^ Constructor arguments.
+  -> Type    -- ^ Type of the constructor application.
+  -> (QName -> Type -> Args -> Type -> Args -> Telescope -> Type -> TCM a)
+       -- ^ Name of the data/record type,
+       --   type of the data/record type,
+       --   reconstructed parameters,
+       --   type of the constructor (applied to parameters),
+       --   full application arguments,
+       --   types of missing arguments (already added to context),
+       --   type of the full application.
+  -> TCM a
+fullyApplyCon c vs t0 ret = do
+  TelV tel t <- telView t0
+  -- The type of the constructor application may still be a function
+  -- type.  In this case, we introduce the domains @tel@ into the context
+  -- and apply the constructor to these fresh variables.
+  addContext tel $ do
+    getFullyAppliedConType c t >>= \case
+      Nothing ->
+        typeError $ DoesNotConstructAnElementOf (conName c) t
+      Just ((d, dt, pars), a) ->
+        ret d dt pars a (raise (size tel) vs ++ teleArgs tel) tel t
+
+checkSpine
+  :: Action
+  -> Type      -- ^ Type of the head @self@.
+  -> Term      -- ^ The head @self@.
+  -> Elims     -- ^ The eliminations @es@.
+  -> Type      -- ^ Expected type of the application @self es@.
+  -> TCM Term  -- ^ The application after modification by the @Action@.
 checkSpine action a self es t = do
   reportSDoc "tc.check.internal" 20 $ sep
     [ text "checking spine "
@@ -174,27 +239,45 @@ checkSpine action a self es t = do
                    , nest 4 $ prettyTCM es <+> text ":"
                    , nest 2 $ prettyTCM t ] ]
   ((v, v'), t') <- inferSpine' action a self self es
+  t' <- reduce t'
   v' <$ coerceSize subtype v t' t
 
-checkArgs :: Action -> Type -> Term -> Args -> Type -> TCM Term
+checkArgs
+  :: Action
+  -> Type      -- ^ Type of the head.
+  -> Term      -- ^ The head.
+  -> Args      -- ^ The arguments.
+  -> Type      -- ^ Expected type of the application.
+  -> TCM Term  -- ^ The application after modification by the @Action@.
 checkArgs action a self vs t = checkSpine action a self (map Apply vs) t
 
 -- | @checkArgInfo actual expected@.
-checkArgInfo :: ArgInfo -> ArgInfo -> TCM ()
-checkArgInfo ai ai' = do
+--
+--   The @expected@ 'ArgInfo' comes from the type.
+--   The @actual@ 'ArgInfo' comes from the term and can be updated
+--   by an action.
+checkArgInfo :: Action -> ArgInfo -> ArgInfo -> TCM ArgInfo
+checkArgInfo action ai ai' = do
   checkHiding    (getHiding ai)     (getHiding ai')
-  checkRelevance (getRelevance ai)  (getRelevance ai')
+  r <- checkRelevance action (getRelevance ai)  (getRelevance ai')
+  return $ setRelevance r ai
 
 checkHiding    :: Hiding -> Hiding -> TCM ()
-checkHiding    h h' = unless (h == h') $ typeError $ HidingMismatch h h'
+checkHiding    h h' = unless (sameHiding h h') $ typeError $ HidingMismatch h h'
 
-checkRelevance :: Relevance -> Relevance -> TCM ()
-checkRelevance r0 r0' = unless (r == r') $ typeError $ RelevanceMismatch r r'
+-- | @checkRelevance action term type@.
+--
+--   The @term@ 'Relevance' can be updated by the @action@.
+--   Note that the relevances might not match precisedly,
+--   because of the non-semantic 'Forced' relevance.
+checkRelevance :: Action -> Relevance -> Relevance -> TCM Relevance
+checkRelevance action r0 r0' = do
+  unless (r == r') $ typeError $ RelevanceMismatch r r'
+  return $ relevanceAction action r0' r0  -- Argument order for actions: @type@ @term@
   where
     r  = canon r0
     r' = canon r0'
     canon Forced{}  = Relevant
-    canon UnusedArg = Relevant
     canon r         = r
 
 -- | Infer type of a neutral term.
@@ -252,7 +335,7 @@ inferSpine' action t self self' (e : es) = do
   case e of
     Apply (Arg ai v) -> do
       (a, b) <- shouldBePi t
-      checkArgInfo ai $ domInfo a
+      ai <- checkArgInfo action ai $ domInfo a
       v' <- checkInternal' action v $ unDom a
       inferSpine' action (b `absApp` v) (self `applyE` [e]) (self' `applyE` [Apply (Arg ai v')]) es
     -- case: projection or projection-like
@@ -321,5 +404,47 @@ metaType x = jMetaType . mvJudgement <$> lookupMeta x
 -- | Universe subsumption and type equality (subtyping for sizes, resp.).
 subtype :: Type -> Type -> TCM ()
 subtype t1 t2 = do
-  ifIsSort t1 (\ s1 -> (s1 `leqSort`) =<< shouldBeSort t2) $
-    leqType t1 t2
+  ifIsSort t1 (\ s1 -> (s1 `leqSort`) =<< shouldBeSort t2) $ do
+    -- Andreas, 2017-03-09, issue #2493
+    -- Only check subtyping, do not solve any metas!
+    -- TODO: NEED? disableDestructiveUpdate
+    dontAssignMetas $ leqType t1 t2
+
+-- | Compute the sort of a type.
+
+inferSort :: Term -> TCM Sort
+inferSort t = case ignoreSharing t of
+    Var i es   -> do
+      a <- typeOfBV i
+      (_, s) <- eliminate (Var i []) a es
+      shouldBeSort s
+    Def f es   -> do  -- f is not projection(-like)!
+      a <- defType <$> getConstInfo f
+      (_, s) <- eliminate (Def f []) a es
+      shouldBeSort s
+    MetaV x es -> do
+      a <- metaType x
+      (_, s) <- eliminate (MetaV x []) a es
+      shouldBeSort s
+    Pi a b     -> return $ dLub (getSort a) (getSort <$> b)
+    Sort s     -> return $ sSuc s
+    Con{}      -> __IMPOSSIBLE__
+    Lit{}      -> __IMPOSSIBLE__
+    Lam{}      -> __IMPOSSIBLE__
+    Level{}    -> __IMPOSSIBLE__
+    DontCare{} -> __IMPOSSIBLE__
+    Shared{}   -> __IMPOSSIBLE__
+
+-- | @eliminate t self es@ eliminates value @self@ of type @t@ by spine @es@
+--   and returns the remaining value and its type.
+eliminate :: Term -> Type -> Elims -> TCM (Term, Type)
+eliminate self t [] = return (self, t)
+eliminate self t (e : es) = case e of
+    Apply (Arg _ v) -> do
+      (_, b) <- shouldBePi t
+      eliminate (self `apply1` v) (b `absApp` v) es
+    -- case: projection or projection-like
+    Proj o f -> do
+      (Dom ai _, b) <- shouldBePi =<< shouldBeProjectible t f
+      u  <- applyDef o f $ Arg ai self
+      eliminate u (b `absApp` self) es
