@@ -14,12 +14,15 @@ import Prelude hiding (null)
 import qualified Data.Map as Map
 import Data.Map (Map)
 import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, Any(..))
-import Data.Typeable (Typeable)
 import Data.Foldable (Foldable, foldMap)
-import Data.Traversable (Traversable)
+import Data.Traversable (Traversable, traverse)
+
+import Data.Data (Data)
+import Data.Typeable (Typeable)
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
+import Agda.Syntax.Internal.Generic
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 
@@ -30,7 +33,7 @@ import Agda.Utils.Pretty hiding ((<>))
 import Agda.Utils.Impossible
 
 data WithArity c = WithArity { arity :: Int, content :: c }
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable, Show)
 
 -- | Branches in a case tree.
 
@@ -46,24 +49,26 @@ data Case c = Branches
   , catchAllBranch :: Maybe c
     -- ^ (Possibly additional) catch-all clause.
   }
-  deriving (Typeable, Functor, Foldable, Traversable)
+  deriving (Typeable, Data, Functor, Foldable, Traversable, Show)
 
 -- | Case tree with bodies.
 
-data CompiledClauses
-  = Case (Arg Int) (Case CompiledClauses)
+data CompiledClauses' a
+  = Case (Arg Int) (Case (CompiledClauses' a))
     -- ^ @Case n bs@ stands for a match on the @n@-th argument
     -- (counting from zero) with @bs@ as the case branches.
     -- If the @n@-th argument is a projection, we have only 'conBranches'
     -- with arity 0.
-  | Done [Arg ArgName] Term
+  | Done [Arg ArgName] a
     -- ^ @Done xs b@ stands for the body @b@ where the @xs@ contains hiding
     --   and name suggestions for the free variables. This is needed to build
     --   lambdas on the right hand side for partial applications which can
     --   still reduce.
   | Fail
     -- ^ Absurd case.
-  deriving (Typeable)
+  deriving (Typeable, Data, Functor, Traversable, Foldable, Show)
+
+type CompiledClauses = CompiledClauses' Term
 
 litCase :: Literal -> c -> Case c
 litCase l x = Branches False Map.empty (Map.singleton l x) Nothing
@@ -86,24 +91,24 @@ hasCatchAll = getAny . loop
     Done{}    -> mempty
     Case _ br -> maybe (foldMap loop br) (const $ Any True) $ catchAllBranch br
 
-instance Monoid c => Semigroup (WithArity c) where
+instance Semigroup c => Semigroup (WithArity c) where
   WithArity n1 c1 <> WithArity n2 c2
-    | n1 == n2  = WithArity n1 $ mappend c1 c2
+    | n1 == n2  = WithArity n1 (c1 <> c2)
     | otherwise = __IMPOSSIBLE__   -- arity must match!
 
-instance Monoid c => Monoid (WithArity c) where
-  mempty = WithArity __IMPOSSIBLE__ mempty
+instance (Semigroup c, Monoid c) => Monoid (WithArity c) where
+  mempty  = WithArity __IMPOSSIBLE__ mempty
   mappend = (<>)
 
-instance Monoid m => Semigroup (Case m) where
+instance Semigroup m => Semigroup (Case m) where
   Branches cop  cs  ls  m <> Branches cop' cs' ls' m' =
     Branches (cop || cop') -- for @projCase <> mempty@
-             (Map.unionWith mappend cs cs')
-             (Map.unionWith mappend ls ls')
-             (mappend m m')
+             (Map.unionWith (<>) cs cs')
+             (Map.unionWith (<>) ls ls')
+             (m <> m')
 
-instance Monoid m => Monoid (Case m) where
-  mempty = empty
+instance (Semigroup m, Monoid m) => Monoid (Case m) where
+  mempty  = empty
   mappend = (<>)
 
 instance Null (Case m) where
@@ -112,12 +117,6 @@ instance Null (Case m) where
 
 -- * Pretty instances.
 
-instance Pretty a => Show (Case a) where
-  show = show . pretty
-
-instance Show CompiledClauses where
-  show = show . pretty
-
 instance Pretty a => Pretty (WithArity a) where
   pretty = pretty . content
 
@@ -129,10 +128,10 @@ instance Pretty a => Pretty (Case a) where
       prC Nothing = []
       prC (Just x) = [text "_ ->" <+> pretty x]
 
-prettyMap :: (Show k, Pretty v) => Map k v -> [Doc]
-prettyMap m = [ sep [ text (show x ++ " ->")
+prettyMap :: (Pretty k, Pretty v) => Map k v -> [Doc]
+prettyMap m = [ sep [ pretty k <+> text "->"
                     , nest 2 $ pretty v ]
-              | (x, v) <- Map.toList m ]
+              | (k, v) <- Map.toList m ]
 
 instance Pretty CompiledClauses where
   pretty (Done hs t) = text ("done" ++ show hs) <+> pretty t
@@ -142,7 +141,7 @@ instance Pretty CompiledClauses where
         , nest 2 $ pretty bs
         ]
   pretty (Case n bs) =
-    sep [ text ("case " ++ show n ++ " of")
+    sep [ text ("case " ++ prettyShow n ++ " of")
         , nest 2 $ pretty bs
         ]
 
@@ -161,3 +160,17 @@ instance KillRange CompiledClauses where
   killRange (Case i br) = killRange2 Case i br
   killRange (Done xs v) = killRange2 Done xs v
   killRange Fail        = Fail
+
+-- * TermLike instances
+
+instance TermLike a => TermLike (WithArity a) where
+  traverseTermM = traverse . traverseTermM
+  foldTerm      = foldMap . foldTerm
+
+instance TermLike a => TermLike (Case a) where
+  traverseTermM = traverse . traverseTermM
+  foldTerm      = foldMap . foldTerm
+
+instance TermLike a => TermLike (CompiledClauses' a) where
+  traverseTermM = traverse . traverseTermM
+  foldTerm      = foldMap . foldTerm
