@@ -4,6 +4,8 @@ module Agda.TypeChecking.CompiledClause.Compile where
 
 import Prelude hiding (null)
 
+import Control.Monad
+
 import Data.Maybe
 import Data.Monoid
 import qualified Data.Map as Map
@@ -46,17 +48,18 @@ compileClauses ::
 compileClauses mt cs = do
   -- Construct clauses with pattern variables bound in left-to-right order.
   -- Discard de Bruijn indices in patterns.
-  cs <- normaliseProjP cs
   let unBruijn cs = [ Cl (map (fmap (fmap dbPatVarName . namedThing)) $ namedClausePats c)
                          (compiledClauseBody c) | c <- cs ]
   shared <- sharedFun
   case mt of
-    Nothing -> return $ compile shared $ unBruijn cs
+    Nothing -> compile shared . unBruijn <$> normaliseProjP cs
     Just (q, t)  -> do
       splitTree <- coverageCheck q t cs
 
       -- The coverage checker might have added some clauses (#2288)!
-      cs <- normaliseProjP =<< defClauses <$> getConstInfo q
+      -- Throw away the unreachable clauses (#2723).
+      let notUnreachable = (Just True /=) . clauseUnreachable
+      cs <- normaliseProjP =<< filter notUnreachable . defClauses <$> getConstInfo q
       let cls = unBruijn cs
 
       reportSDoc "tc.cc" 30 $ sep $ do
@@ -69,7 +72,7 @@ compileClauses mt cs = do
       let cc = compileWithSplitTree shared splitTree cls
       reportSDoc "tc.cc" 12 $ sep
         [ text "compiled clauses (still containing record splits)"
-        , nest 2 $ text (show cc)
+        , nest 2 $ return $ P.pretty cc
         ]
       cc <- translateCompiledClauses cc
       return cc
@@ -126,6 +129,7 @@ compile shared cs = case nextSplit cs of
     c = headWithDefault __IMPOSSIBLE__ cs
     name (VarP x) = x
     name (DotP _) = underscore
+    name AbsurdP{} = absurdPatternName
     name ConP{}  = __IMPOSSIBLE__
     name LitP{}  = __IMPOSSIBLE__
     name ProjP{} = __IMPOSSIBLE__
@@ -144,6 +148,7 @@ properSplit (ConP _ cpi _) = Just $ isJust $ conPRecord cpi
 properSplit LitP{}  = Just False
 properSplit ProjP{} = Just False
 properSplit VarP{}  = Nothing
+properSplit AbsurdP{} = Nothing -- for purposes of compilation
 properSplit DotP{}  = Nothing
 
 -- | Is this a variable pattern?
@@ -152,6 +157,7 @@ properSplit DotP{}  = Nothing
 isVar :: Pattern' a -> Bool
 isVar VarP{}  = True
 isVar DotP{}  = True
+isVar AbsurdP{} = True
 isVar ConP{}  = False
 isVar LitP{}  = False
 isVar ProjP{} = False
@@ -171,6 +177,7 @@ splitC n (Cl ps b) = caseMaybe mp fallback $ \case
   LitP l      -> litCase l $ Cl (ps0 ++ ps1) b
   VarP{}      -> fallback
   DotP{}      -> fallback
+  AbsurdP{}   -> fallback
   where
     (ps0, rest) = splitAt n ps
     mp          = unArg <$> headMaybe rest
@@ -264,9 +271,15 @@ expandCatchAlls single n cs =
     expansions = nubBy ((==) `on` (classify . unArg . snd))
                . mapMaybe (notVarNth . clPats)
                $ cs
-    notVarNth ps = caseMaybe (headMaybe ps2) Nothing $ \ p ->
-      if isVar (unArg p) then Nothing else Just (ps1, p)
-      where (ps1, ps2) = splitAt n ps
+    notVarNth
+      :: [Arg Pattern]
+      -> Maybe ([Arg Pattern]  -- First @n@ patterns.
+               , Arg Pattern)  -- @n+1@st pattern, not a variable
+    notVarNth ps = do
+      let (ps1, ps2) = splitAt n ps
+      p <- headMaybe ps2
+      guard $ not $ isVar $ unArg p
+      return (ps1, p)
 
     expand cl (qs, q) =
       case unArg q of
@@ -293,6 +306,7 @@ expandCatchAlls single n cs =
         count VarP{}        = 1
         count (ConP _ _ ps) = countVars $ map (fmap namedThing) ps
         count DotP{}        = 1   -- dot patterns are treated as variables in the clauses
+        count (AbsurdP p)   = count p
         count _             = 0
 
 -- | Make sure (by eta-expansion) that clause has arity at least @n@
