@@ -12,7 +12,8 @@ import Control.Monad.Reader
 import Control.Monad.State
 import Control.Monad.Trans.Maybe
 
-import Data.List as List hiding (null)
+import qualified Data.List as List
+import qualified Data.Set as Set
 
 import Agda.Syntax.Internal
 
@@ -24,6 +25,8 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.LevelConstraints
 import Agda.TypeChecking.SizedTypes
+import Agda.TypeChecking.MetaVars.Mention
+import Agda.TypeChecking.Warnings
 
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Term
 import {-# SOURCE #-} Agda.TypeChecking.Conversion
@@ -61,7 +64,7 @@ addConstraint c = do
     pids <- asks envActiveProblems
     reportSDoc "tc.constr.add" 20 $ hsep
       [ text "adding constraint"
-      , text (show pids)
+      , text (show $ Set.toList pids)
       , prettyTCM c ]
     -- Need to reduce to reveal possibly blocking metas
     c <- reduce =<< instantiateFull c
@@ -90,7 +93,6 @@ addConstraint c = do
     -- Try to simplify a level constraint
     simpl :: Constraint -> TCM Constraint
     simpl c = if not $ isLvl c then return c else do
-      n <- genericLength <$> getContext
       cs <- map theConstraint <$> getAllConstraints
       lvls <- instantiateFull $ List.filter (isLvl . clValue) cs
       when (not $ null lvls) $ do
@@ -149,7 +151,7 @@ whenConstraints action handler =
 -- | Wake up the constraints depending on the given meta.
 wakeupConstraints :: MetaId -> TCM ()
 wakeupConstraints x = do
-  wakeConstraints (return . const True) -- (mentionsMeta x) -- TODO: needs fixing to cope with shared updates
+  wakeConstraints (return . mentionsMeta x)
   solveAwakeConstraints
 
 -- | Wake up all constraints.
@@ -163,12 +165,12 @@ solveAwakeConstraints = solveAwakeConstraints' False
 
 solveAwakeConstraints' :: Bool -> TCM ()
 solveAwakeConstraints' force = do
-    verboseS "profile.constraints" 10 $ liftTCM $ tickMax "max-open-constraints" . genericLength =<< getAllConstraints
+    verboseS "profile.constraints" 10 $ liftTCM $ tickMax "max-open-constraints" . List.genericLength =<< getAllConstraints
     whenM ((force ||) . not <$> isSolvingConstraints) $ nowSolvingConstraints $ do
      -- solveSizeConstraints -- Andreas, 2012-09-27 attacks size constrs too early
      -- Ulf, 2016-12-06: Don't inherit problems here! Stored constraints
      -- already contain all their dependencies.
-     locally eActiveProblems (const []) solve
+     locally eActiveProblems (const Set.empty) solve
   where
     solve = do
       reportSDoc "tc.constr.solve" 10 $ hsep [ text "Solving awake constraints."
@@ -183,7 +185,7 @@ solveConstraint c = do
     verboseS "profile.constraints" 10 $ liftTCM $ tick "attempted-constraints"
     verboseBracket "tc.constr.solve" 20 "solving constraint" $ do
       pids <- asks envActiveProblems
-      reportSDoc "tc.constr.solve" 20 $ text (show pids) <+> prettyTCM c
+      reportSDoc "tc.constr.solve" 20 $ text (show $ Set.toList pids) <+> prettyTCM c
       solveConstraint_ c
 
 solveConstraint_ :: Constraint -> TCM ()
@@ -194,8 +196,13 @@ solveConstraint_ (TelCmp a b cmp tela telb) = compareTel a b cmp tela telb
 solveConstraint_ (SortCmp cmp s1 s2)        = compareSort cmp s1 s2
 solveConstraint_ (LevelCmp cmp a b)         = compareLevel cmp a b
 solveConstraint_ c0@(Guarded c pid)         = do
-  ifM (isProblemSolved pid) (solveConstraint_ c)
-                            (addConstraint c0)
+  ifM (isProblemSolved pid)
+    {-then-} (do
+      reportSLn "tc.constr.solve" 50 $ "Guarding problem " ++ show pid ++ " is solved, moving on..."
+      solveConstraint_ c)
+    {-else-} $ do
+      reportSLn "tc.constr.solve" 50 $ "Guarding problem " ++ show pid ++ " is still unsolved."
+      addConstraint c0
 solveConstraint_ (IsEmpty r t)              = isEmptyType r t
 solveConstraint_ (CheckSizeLtSat t)         = checkSizeLtSat t
 solveConstraint_ (UnBlock m)                =
@@ -217,6 +224,10 @@ solveConstraint_ (UnBlock m)                =
       --
       -- already solved metavariables: should only happen for size
       -- metas (not sure why it does, Andreas?)
+      -- Andreas, 2017-07-11:
+      -- I think this is because the size solver instantiates
+      -- some metas with infinity but does not clean up the UnBlock constraints.
+      -- See also issue #2637.
       InstV{} -> return ()
       -- Open (whatever that means)
       Open -> __IMPOSSIBLE__
