@@ -1,4 +1,5 @@
-{-# LANGUAGE CPP               #-}
+{-# LANGUAGE CPP                      #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 module Agda.TypeChecking.Conversion where
 
@@ -7,7 +8,6 @@ import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State
 
-import Data.List hiding (sort)
 import qualified Data.List as List
 import Data.Traversable hiding (mapM, sequence)
 
@@ -18,7 +18,7 @@ import Agda.Syntax.Translation.InternalToAbstract (reify)
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.CompiledClause (CompiledClauses(Fail))
+import Agda.TypeChecking.CompiledClause (CompiledClauses'(Fail))
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.MetaVars.Occurs (killArgs,PruneResult(..))
 import Agda.TypeChecking.Reduce
@@ -29,7 +29,7 @@ import Agda.TypeChecking.Constraints
 import {-# SOURCE #-} Agda.TypeChecking.CheckInternal (infer)
 import Agda.TypeChecking.Errors
 import Agda.TypeChecking.Free
-import Agda.TypeChecking.Datatypes (getConType)
+import Agda.TypeChecking.Datatypes (getConType, getFullyAppliedConType)
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Injectivity
@@ -169,10 +169,8 @@ compareTerm cmp a u v = do
         ]
       ifM (isInstantiatedMeta x) patternViolation {-else-} $ do
         assignE dir x es v $ compareTermDir dir a
-      _ <- instantiate u
-      -- () <- seq u' $ return ()
-      reportSLn "tc.conv.term.shortcut" 50 $
-        "shortcut successful\n  result: " ++ show u
+      reportSDoc "tc.conv.term.shortcut" 50 $
+        text "shortcut successful" $$ nest 2 (text "result:" <+> (pretty =<< instantiate (MetaV x es)))
     -- Should be ok with catchError_ but catchError is much safer since we don't
     -- rethrow errors.
     orelse m h = catchError m (\_ -> h)
@@ -237,8 +235,8 @@ compareTerm' cmp a m n =
     s        <- reduce $ getSort a'
     mlvl     <- tryMaybe primLevel
     reportSDoc "tc.conv.level" 60 $ nest 2 $ sep
-      [ text $ "a'   = " ++ show a'
-      , text $ "mlvl = " ++ show mlvl
+      [ text "a'   =" <+> pretty a'
+      , text "mlvl =" <+> pretty mlvl
       , text $ "(Just (ignoreSharing $ unEl a') == mlvl) = " ++ show (Just (ignoreSharing $ unEl a') == mlvl)
       ]
     case s of
@@ -307,7 +305,7 @@ compareTerm' cmp a m n =
     equalFun (Shared p) m n = equalFun (derefPtr p) m n
     equalFun (Pi dom@(Dom info _) b) m n = do
         name <- freshName_ $ suggest (absName b) "x"
-        addContext (name, dom) $ compareTerm cmp (absBody b) m' n'
+        addContext' (name, dom) $ compareTerm cmp (absBody b) m' n'
       where
         (m',n') = raise 1 (m,n) `apply` [Arg info $ var 0]
     equalFun _ _ _ = __IMPOSSIBLE__
@@ -376,6 +374,31 @@ etaInequal cmp t m n = do
 compareAtomDir :: CompareDirection -> Type -> Term -> Term -> TCM ()
 compareAtomDir dir a = dirToCmp (`compareAtom` a) dir
 
+-- | Compute the head type of an elimination. For projection-like functions
+--   this requires inferring the type of the principal argument.
+computeElimHeadType :: QName -> Elims -> Elims -> TCM Type
+computeElimHeadType f es es' = do
+  def <- getConstInfo f
+  -- To compute the type @a@ of a projection-like @f@,
+  -- we have to infer the type of its first argument.
+  if projectionArgs (theDef def) <= 0 then return $ defType def else do
+    -- Find an first argument to @f@.
+    let arg = case (es, es') of
+              (Apply arg : _, _) -> arg
+              (_, Apply arg : _) -> arg
+              _ -> __IMPOSSIBLE__
+    -- Infer its type.
+    reportSDoc "tc.conv.infer" 30 $
+      text "inferring type of internal arg: " <+> prettyTCM arg
+    targ <- infer $ unArg arg
+    reportSDoc "tc.conv.infer" 30 $
+      text "inferred type: " <+> prettyTCM targ
+    -- getDefType wants the argument type reduced.
+    -- Andreas, 2016-02-09, Issue 1825: The type of arg might be
+    -- a meta-variable, e.g. in interactive development.
+    -- In this case, we postpone.
+    fromMaybeM patternViolation $ getDefType f =<< reduce targ
+
 -- | Syntax directed equality on atomic values
 --
 compareAtom :: Comparison -> Type -> Term -> Term -> TCM ()
@@ -492,26 +515,7 @@ compareAtom cmp t m n =
                 compareElims [] a (var i) es es'
             (Def f [], Def f' []) | f == f' -> return ()
             (Def f es, Def f' es') | f == f' -> do
-                def <- getConstInfo f
-                -- To compute the type @a@ of a projection-like @f@,
-                -- we have to infer the type of its first argument.
-                a <- if projectionArgs (theDef def) <= 0 then return $ defType def else do
-                  -- Find an first argument to @f@.
-                  let arg = case (es, es') of
-                            (Apply arg : _, _) -> arg
-                            (_, Apply arg : _) -> arg
-                            _ -> __IMPOSSIBLE__
-                  -- Infer its type.
-                  reportSDoc "tc.conv.infer" 30 $
-                    text "inferring type of internal arg: " <+> prettyTCM arg
-                  targ <- infer $ unArg arg
-                  reportSDoc "tc.conv.infer" 30 $
-                    text "inferred type: " <+> prettyTCM targ
-                  -- getDefType wants the argument type reduced.
-                  -- Andreas, 2016-02-09, Issue 1825: The type of arg might be
-                  -- a meta-variable, e.g. in interactive development.
-                  -- In this case, we postpone.
-                  fromMaybeM patternViolation $ getDefType f =<< reduce targ
+                a <- computeElimHeadType f es es'
                 -- The polarity vector of projection-like functions
                 -- does not include the parameters.
                 pol <- getPolarity' cmp f
@@ -519,6 +523,7 @@ compareAtom cmp t m n =
             (Def f es, Def f' es') ->
               unlessM (bothAbsurd f f') $ do
                 trySizeUniv cmp t m n f es f' es'
+            -- Due to eta-expansion, these constructors are fully applied.
             (Con x ci xArgs, Con y _ yArgs)
                 | x == y -> do
                     -- Get the type of the constructor instantiated to the datatype parameters.
@@ -533,7 +538,7 @@ compareAtom cmp t m n =
           let impossible = do
                 reportSDoc "impossible" 10 $
                   text "expected data/record type, found " <+> prettyTCM t
-                reportSLn "impossible" 70 $ "  raw = " ++ show t
+                reportSDoc "impossible" 70 $ nest 2 $ text "raw =" <+> pretty t
                 -- __IMPOSSIBLE__
                 -- Andreas, 2013-10-20:  in case termination checking fails
                 -- we might get some unreduced types here.
@@ -541,7 +546,7 @@ compareAtom cmp t m n =
                 -- to solve left-over constraints.
                 -- Thus, instead of crashing, just give up gracefully.
                 patternViolation
-          maybe impossible return =<< getConType c t
+          maybe impossible (return . snd) =<< getFullyAppliedConType c t
         equalFun t1 t2 = case (ignoreSharing t1, ignoreSharing t2) of
           (Pi dom1 b1, Pi dom2 b2) -> do
             verboseBracket "tc.conv.fun" 15 "compare function types" $ do
@@ -579,7 +584,7 @@ compareDom :: Free c
   -> TCM ()     -- ^ Continuation if comparison is successful.
   -> TCM ()
 compareDom cmp dom1@(Dom i1 a1) dom2@(Dom i2 a2) b1 b2 errH errR cont
-  | getHiding dom1 /= getHiding dom2 = errH
+  | not (sameHiding dom1 dom2) = errH
   -- Andreas 2010-09-21 compare r1 and r2, but ignore forcing annotations!
   | not $ compareRelevance cmp (ignoreForced $ getRelevance dom1)
                                (ignoreForced $ getRelevance dom2) = errR
@@ -594,7 +599,7 @@ compareDom cmp dom1@(Dom i1 a1) dom2@(Dom i2 a2) b1 b2 errH errR cont
         -- We only need to require a1 == a2 if b2 is dependent
         -- If it's non-dependent it doesn't matter what we add to the context.
       name <- freshName_ $ suggest b1 b2
-      addContext (name, dom) $ cont
+      addContext' (name, dom) $ cont
       stealConstraints pid
         -- Andreas, 2013-05-15 Now, comparison of codomains is not
         -- blocked any more by getting stuck on domains.
@@ -605,6 +610,79 @@ compareRelevance :: Comparison -> Relevance -> Relevance -> Bool
 compareRelevance CmpEq  = (==)
 compareRelevance CmpLeq = (<=)
 
+-- | When comparing argument spines (in compareElims) where the first arguments
+--   don't match, we keep going, substituting the anti-unification of the two
+--   terms in the telescope. More precisely:
+--
+--  @@
+--    (u = v : A)[pid]   w = antiUnify pid A u v   us = vs : Δ[w/x]
+--    -------------------------------------------------------------
+--                    u us = v vs : (x : A) Δ
+--  @@
+--
+--   The simplest case of anti-unification is to return a fresh metavariable
+--   (created by blockTermOnProblem), but if there's shared structure between
+--   the two terms we can expose that.
+--
+--   This is really a crutch that lets us get away with things that otherwise
+--   would require heterogenous conversion checking. See for instance issue
+--   #2384.
+antiUnify :: ProblemId -> Type -> Term -> Term -> TCM Term
+antiUnify pid a u v = do
+  ((u, v), eq) <- runReduceM (SynEq.checkSyntacticEquality u v)
+  if eq then return u else do
+  (u, v) <- reduce (u, v)
+  case (ignoreSharing u, ignoreSharing v) of
+    (Pi ua ub, Pi va vb) -> do
+      wa0 <- antiUnifyType pid (unDom ua) (unDom va)
+      let wa = wa0 <$ ua
+      wb <- addContext wa $ antiUnifyType pid (unAbs ub) (unAbs vb)
+      return $ Pi wa (wb <$ ub)
+    (Lam i u, Lam _ v) ->
+      case ignoreSharing $ unEl a of
+        Pi a b -> Lam i . (<$ u) <$> addContext a (antiUnify pid (unAbs b) (unAbs u) (unAbs v))
+        _      -> fallback
+    (Var i us, Var j vs) | i == j -> maybeGiveUp $ do
+      a <- typeOfBV i
+      antiUnifyElims pid a (var i) us vs
+    -- Andreas, 2017-07-27:
+    -- It seems that nothing guarantees here that the constructors are fully
+    -- applied!?  Thus, @a@ could be a function type and we need the robust
+    -- @getConType@ here.
+    -- (Note that @patternViolation@ swallows exceptions coming from @getConType@
+    -- thus, we would not see clearly if we used @getFullyAppliedConType@ instead.)
+    (Con x ci us, Con y _ vs) | x == y -> maybeGiveUp $ do
+      a <- maybe patternViolation (return . snd) =<< getConType x a
+      antiUnifyElims pid a (Con x ci []) (map Apply us) (map Apply vs)
+    (Def f us, Def g vs) | f == g, length us == length vs -> maybeGiveUp $ do
+      a <- computeElimHeadType f us vs
+      antiUnifyElims pid a (Def f []) us vs
+    _ -> fallback
+  where
+    fallback = blockTermOnProblem a u pid
+    maybeGiveUp m = m `catchError_` \ err ->
+      case err of
+        PatternErr{} -> fallback
+        _            -> throwError err
+
+antiUnifyType :: ProblemId -> Type -> Type -> TCM Type
+antiUnifyType pid (El s a) (El _ b) = El s <$> antiUnify pid (sort s) a b
+
+antiUnifyElims :: ProblemId -> Type -> Term -> Elims -> Elims -> TCM Term
+antiUnifyElims pid a self [] [] = return self
+antiUnifyElims pid a self (Proj o f : es1) (Proj _ g : es2) | f == g = do
+  res <- projectTyped self a o f
+  case res of
+    Just (_, self, a) -> antiUnifyElims pid a self es1 es2
+    Nothing -> patternViolation -- can fail for projection like
+antiUnifyElims pid a self (Apply u : es1) (Apply v : es2) = do
+  case ignoreSharing $ unEl a of
+    Pi a b -> do
+      w <- antiUnify pid (unDom a) (unArg u) (unArg v)
+      antiUnifyElims pid (b `lazyAbsApp` w) (apply self [w <$ u]) es1 es2
+    _ -> patternViolation
+antiUnifyElims _ _ _ _ _ = patternViolation -- trigger maybeGiveUp in antiUnify
+
 -- | @compareElims pols a v els1 els2@ performs type-directed equality on eliminator spines.
 --   @t@ is the type of the head @v@.
 compareElims :: [Polarity] -> Type -> Term -> [Elim] -> [Elim] -> TCM ()
@@ -640,9 +718,9 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
         , text "arg2 =" <+> prettyTCM arg2
         ]
       reportSDoc "tc.conv.elim" 50 $ nest 2 $ vcat
-        [ text "v    =" <+> text (show v)
-        , text "arg1 =" <+> text (show arg1)
-        , text "arg2 =" <+> text (show arg2)
+        [ text "v    =" <+> pretty v
+        , text "arg1 =" <+> pretty arg1
+        , text "arg2 =" <+> pretty arg2
         , text ""
         ]
       let (pol, pols) = nextPolarity pols0
@@ -667,13 +745,18 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
             pid <- newProblem_ $ applyRelevanceToContext r $
                 case r of
                   Forced{}   -> return ()
-                  r | irrelevantOrUnused r ->
+                  r | irrelevant r ->
                                 compareIrrelevant b (unArg arg1) (unArg arg2)
                   _          -> compareWithPol pol (flip compareTerm b)
                                   (unArg arg1) (unArg arg2)
             -- if comparison got stuck and function type is dependent, block arg
-            arg <- if dependent
-                   then (arg1 $>) <$> blockTermOnProblem b (unArg arg1) pid
+            solved <- isProblemSolved pid
+            arg <- if dependent && not solved
+                   then do
+                    arg <- (arg1 $>) <$> antiUnify pid b (unArg arg1) (unArg arg2)
+                    reportSDoc "tc.conv.elims" 30 $ hang (text "Anti-unification:") 2 (prettyTCM arg)
+                    reportSDoc "tc.conv.elims" 70 $ nest 2 $ text "raw:" <+> pretty arg
+                    return arg
                    else return arg1
             -- continue, possibly with blocked instantiation
             compareElims pols (codom `lazyAbsApp` unArg arg) (apply v [arg]) els1 els2
@@ -693,7 +776,7 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
             let checkArg = applyRelevanceToContext r $
                                case r of
                   Forced     -> return ()
-                  r | irrelevantOrUnused r ->
+                  r | irrelevant r ->
                                 compareIrrelevant b (unArg arg1) (unArg arg2)
                   _          -> compareWithPol pol (flip compareTerm b)
                                   (unArg arg1) (unArg arg2)
@@ -708,7 +791,7 @@ compareElims pols0 a v els01 els02 = catchConstraint (ElimCmp pols0 a v els01 el
           a -> do
             reportSDoc "impossible" 10 $
               text "unexpected type when comparing apply eliminations " <+> prettyTCM a
-            reportSLn "impossible" 50 $ "raw type: " ++ show a
+            reportSDoc "impossible" 50 $ text "raw type:" <+> pretty a
             patternViolation
             -- Andreas, 2013-10-22
             -- in case of disabled reductions (due to failing termination check)
@@ -752,8 +835,8 @@ compareIrrelevant t v w = do
     , nest 2 $ text "w =" <+> prettyTCM w
     ]
   reportSDoc "tc.conv.irr" 50 $ vcat
-    [ nest 2 $ text $ "v = " ++ show v
-    , nest 2 $ text $ "w = " ++ show w
+    [ nest 2 $ text "v =" <+> pretty v
+    , nest 2 $ text "w =" <+> pretty w
     ]
   try v w $ try w v $ return ()
   where
@@ -766,9 +849,9 @@ compareIrrelevant t v w = do
                    _       -> False
       reportSDoc "tc.conv.irr" 20 $ vcat
         [ nest 2 $ text $ "rel  = " ++ show rel
-        , nest 2 $ text $ "inst = " ++ show inst
+        , nest 2 $ text "inst =" <+> pretty inst
         ]
-      if not (irrelevantOrUnused rel) || inst
+      if not (irrelevant rel) || inst
         then fallback
         -- Andreas, 2016-08-08, issue #2131:
         -- Mining for solutions for irrelevant metas is not definite.
@@ -850,6 +933,8 @@ leqType = compareType CmpLeq
 --
 --   In principle, this function can host coercive subtyping, but
 --   currently it only tries to fix problems with hidden function types.
+--
+--   Precondition: @a@ and @b@ are reduced.
 coerce :: Term -> Type -> Type -> TCM Term
 coerce v t1 t2 = blockTerm t2 $ do
   verboseS "tc.conv.coerce" 10 $ do
@@ -861,10 +946,16 @@ coerce v t1 t2 = blockTerm t2 $ do
         , text "from type t1 =" <+> prettyTCM a1
         , text "to type   t2 =" <+> prettyTCM a2
         ]
+    reportSDoc "tc.conv.coerce" 70 $
+      text "coerce" <+> vcat
+        [ text "term      v  =" <+> pretty v
+        , text "from type t1 =" <+> pretty t1
+        , text "to type   t2 =" <+> pretty t2
+        ]
   -- v <$ do workOnTypes $ leqType t1 t2
   -- take off hidden/instance domains from t1 and t2
-  TelV tel1 b1 <- telViewUpTo' (-1) ((NotHidden /=) . getHiding) t1
-  TelV tel2 b2 <- telViewUpTo' (-1) ((NotHidden /=) . getHiding) t2
+  TelV tel1 b1 <- telViewUpTo' (-1) notVisible t1
+  TelV tel2 b2 <- telViewUpTo' (-1) notVisible t2
   let n = size tel1 - size tel2
   -- the crude solution would be
   --   v' = λ {tel2} → v {tel1}
@@ -873,7 +964,7 @@ coerce v t1 t2 = blockTerm t2 $ do
   -- insert n many hidden args
   if n <= 0 then fallback else do
     ifBlockedType b2 (\ _ _ -> fallback) $ \ _ -> do
-      (args, t1') <- implicitArgs n (NotHidden /=) t1
+      (args, t1') <- implicitArgs n notVisible t1
       coerceSize leqType (v `apply` args) t1' t2
   where
     fallback = coerceSize leqType v t1 t2
@@ -886,18 +977,32 @@ coerce v t1 t2 = blockTerm t2 $ do
 --   TODO.
 --
 --   For now, we do a cheap heuristics.
+--
+--   Precondition: types are reduced.
 coerceSize :: (Type -> Type -> TCM ()) -> Term -> Type -> Type -> TCM Term
 coerceSize leqType v t1 t2 = workOnTypes $ do
+    reportSDoc "tc.conv.coerce" 70 $
+      text "coerceSize" <+> vcat
+        [ text "term      v  =" <+> pretty v
+        , text "from type t1 =" <+> pretty t1
+        , text "to type   t2 =" <+> pretty t2
+        ]
     let fallback = v <$ leqType t1 t2
         done = caseMaybeM (isSizeType t1) fallback $ \ b1 -> return v
     -- Andreas, 2015-07-22, Issue 1615:
     -- If t1 is a meta and t2 a type like Size< v2, we need to make sure we do not miss
     -- the constraint v < v2!
     caseMaybeM (isSizeType t2) fallback $ \ b2 -> do
+      -- Andreas, 2017-01-20, issue #2329:
+      -- If v is not a size suitable for the solver, like a neutral term,
+      -- we can only rely on the type.
+      mv <- sizeMaxView v
+      if any (\case{ DOtherSize{} -> True; _ -> False }) mv then fallback else do
       -- Andreas, 2015-02-11 do not instantiate metas here (triggers issue 1203).
       ifM (tryConversion $ dontAssignMetas $ leqType t1 t2) (return v) $ {- else -} do
         -- A (most probably weaker) alternative is to just check syn.eq.
         -- ifM (snd <$> checkSyntacticEquality t1 t2) (return v) $ {- else -} do
+        reportSDoc "tc.conv.coerce" 20 $ text "coercing to a size type"
         case b2 of
           -- @t2 = Size@.  We are done!
           BoundedNo -> done
@@ -985,8 +1090,8 @@ leqLevel a b = liftTCM $ do
     leqView a@(Max as) b@(Max bs) = catchConstraint (LevelCmp CmpLeq a b) $ do
       reportSDoc "tc.conv.nat" 30 $
         text "compareLevelView" <+>
-          sep [ text (show a) <+> text "=<"
-              , text (show b) ]
+          sep [ pretty a <+> text "=<"
+              , pretty b ]
       wrap $ case (as, bs) of
 
         -- same term
@@ -1015,7 +1120,7 @@ leqLevel a b = liftTCM $ do
         -- remove subsumed
         -- Andreas, 2014-04-07: This is ok if we do not go back to equalLevel
         (as, bs)
-          | not $ null subsumed -> leqView (Max $ as \\ subsumed) (Max bs)
+          | not $ null subsumed -> leqView (Max $ as List.\\ subsumed) (Max bs)
           where
             subsumed = [ a | a@(Plus m l) <- as, n <- findN l, m <= n ]
             -- @findN a@ finds the unique(?) term @Plus n a@ in @bs@, if any.
@@ -1100,7 +1205,7 @@ equalLevel a b = do
 -- | Precondition: levels are 'normalise'd.
 equalLevel' :: Level -> Level -> TCM ()
 equalLevel' a b = do
-  reportSLn "tc.conv.level" 50 $ "equalLevel (" ++ show a ++ ") (" ++ show b ++ ")"
+  reportSDoc "tc.conv.level" 50 $ sep [ text "equalLevel", nest 2 $ parens $ pretty a, nest 2 $ parens $ pretty b ]
   liftTCM $ catchConstraint (LevelCmp CmpEq a b) $
     check a b
   where
@@ -1132,8 +1237,8 @@ equalLevel' a b = do
             ]
       reportSDoc "tc.conv.level" 50 $
         sep [ text "equalLevel"
-            , vcat [ nest 2 $ sep [ text (show (Max as)) <+> text "=="
-                                  , text (show (Max bs))
+            , vcat [ nest 2 $ sep [ pretty (Max as) <+> text "=="
+                                  , pretty (Max bs)
                                   ]
                    ]
             ]
@@ -1198,7 +1303,7 @@ equalLevel' a b = do
         notok    = unlessM typeInType notOk
         notOk    = typeError $ UnequalSorts (Type a) (Type b)
         postpone = do
-          reportSLn "tc.conv.level" 30 $ "postponing: " ++ show a ++ " == " ++ show b
+          reportSDoc "tc.conv.level" 30 $ hang (text "postponing:") 2 $ hang (pretty a <+> text "==") 0 (pretty b)
           patternViolation
 
         closed0 [] = [ClosedLevel 0]
@@ -1207,7 +1312,7 @@ equalLevel' a b = do
         -- perform assignment (Plus n (MetaLevel x as)) := bs
         meta n x as bs = do
           reportSLn "tc.meta.level" 30 $ "Assigning meta level"
-          reportSLn "tc.meta.level" 50 $ "meta " ++ show as ++ " " ++ show bs
+          reportSDoc "tc.meta.level" 50 $ text "meta" <+> sep [prettyList $ map pretty as, prettyList $ map pretty bs]
           bs' <- mapM (subtr n) bs
           assignE DirEq x as (levelTm (Max bs')) (===) -- fallback: check equality as atoms
 
@@ -1291,8 +1396,8 @@ equalSort s1 s2 = do
           [ text "equalSort"
           , vcat [ nest 2 $ fsep [ prettyTCM s1 <+> text "=="
                                  , prettyTCM s2 ]
-                 , nest 2 $ fsep [ text (show s1) <+> text "=="
-                                 , text (show s2) ]
+                 , nest 2 $ fsep [ pretty s1 <+> text "=="
+                                 , pretty s2 ]
                  ]
           ]
 
