@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP              #-}
+{-# LANGUAGE CPP                      #-}
 {-# LANGUAGE NondecreasingIndentation #-}
 
 {-| Coverage checking, case splitting, and splitting for refine tactics.
@@ -25,7 +25,8 @@ import Control.Monad.Trans ( lift )
 import Control.Applicative hiding (empty)
 #endif
 
-import Data.List hiding (null)
+import Data.Either (lefts)
+import qualified Data.List as List
 import Data.Monoid (Any(..))
 import Data.Map (Map)
 import qualified Data.Map as Map
@@ -40,7 +41,6 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad
-import Agda.TypeChecking.Monad.Exception
 
 import Agda.TypeChecking.Rules.LHS.Problem (allFlexVars)
 import Agda.TypeChecking.Rules.LHS.Unify
@@ -55,17 +55,25 @@ import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.MetaVars
+import Agda.TypeChecking.Warnings
 
 import Agda.Interaction.Options
 
 import Agda.Utils.Either
+import Agda.Utils.Except
+  ( ExceptT
+  , MonadError(catchError, throwError)
+  , runExceptT
+  )
 import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Size
+import Agda.Utils.Suffix (nameVariant)
 import Agda.Utils.Tuple
 import Agda.Utils.Lens
 
@@ -124,9 +132,16 @@ clauseToSplitClause cl = SClause
   , scTarget = clauseType cl
   }
 
-type CoverM = ExceptionT SplitError TCM
+type CoverM = ExceptT SplitError TCM
 
 -- | Top-level function for checking pattern coverage.
+--
+--   Effects:
+--
+--   - Marks unreachable clauses as such in the signature.
+--
+--   - Adds missing instances clauses to the signature.
+--
 coverageCheck :: QName -> Type -> [Clause] -> TCM SplitTree
 coverageCheck f t cs = do
   TelV gamma a <- telView t
@@ -145,54 +160,93 @@ coverageCheck f t cs = do
     let prCl cl = addContext (clauseTel cl) $
                   prettyTCMPatternList $ namedClausePats cl
     vcat
-      [ text $ "Coverage checking " ++ show f ++ " with patterns:"
+      [ text $ "Coverage checking " ++ prettyShow f ++ " with patterns:"
       , nest 2 $ vcat $ map prCl cs
       ]
 
   -- used = actually used clauses for cover
   -- pss  = uncovered cases
-  (splitTree, used, pss) <- cover f cs sc
+  CoverResult splitTree used pss noex <- cover f cs sc
+  reportSDoc "tc.cover.top" 10 $ vcat
+    [ text "cover computed!"
+    , text $ "used clauses: " ++ show used
+    , text $ "non-exact clauses: " ++ show noex
+    ]
   reportSDoc "tc.cover.splittree" 10 $ vcat
     [ text "generated split tree for" <+> prettyTCM f
-    , text $ show splitTree
+    , text $ prettyShow splitTree
     ]
-  -- report an error if there are uncovered cases
+  -- report a warning if there are uncovered cases,
+  -- generate a catch-all clause with a metavariable as its body to avoid
+  -- internal errors in the reduction machinery.
   unless (null pss) $
       setCurrentRange cs $
-        typeError $ CoverageFailure f pss
+        warning $ CoverageIssue f pss
+
+  -- Andreas, 2017-08-28, issue #2723:
+  -- Mark clauses as reachable or unreachable in the signature.
+  let (is0, cs1) = unzip $ for (zip [0..] cs) $ \ (i, cl) ->
+        let unreachable = i `Set.notMember` used in
+        (boolToMaybe unreachable i, cl { clauseUnreachable = Just unreachable  })
   -- is = indices of unreachable clauses
-  let is = Set.toList $ Set.difference (Set.fromList [0..genericLength cs - 1]) used
-  -- report an error if there are unreachable clauses
+  let is = catMaybes is0
+  reportSDoc "tc.cover.top" 10 $ vcat
+    [ text $ "unreachable clauses: " ++ if null is then "(none)" else show is
+    ]
+  -- Replace the first clauses by @cs1@.  There might be more
+  -- added by @inferMissingClause@.
+  modifyFunClauses f $ \ cs0 -> cs1 ++ drop (length cs1) cs0
+
+  -- Warn if there are unreachable clauses and mark them as unreachable.
   unless (null is) $ do
-      let unreached = map (cs !!) is
-      setCurrentRange unreached $
-        typeError $ UnreachableClauses f $ map namedClausePats unreached
+    -- Warn about unreachable clauses.
+    let unreached = filter ((Just True ==) . clauseUnreachable) cs1
+    setCurrentRange (map clauseFullRange unreached) $
+      warning $ UnreachableClauses f $ map namedClausePats unreached
+
+  -- report a warning if there are clauses that are not preserved as
+  -- definitional equalities and --exact-split is enabled
+  unless (null noex) $ do
+      let noexclauses = map (cs1 !!) $ Set.toList noex
+      setCurrentRange (map clauseLHSRange noexclauses) $
+        warning $ CoverageNoExactSplit f $ noexclauses
   return splitTree
 
 -- | Top-level function for eliminating redundant clauses in the interactive
 --   case splitter
 isCovered :: QName -> [Clause] -> SplitClause -> TCM Bool
 isCovered f cs sc = do
-  (_, _, missing) <- cover f cs sc
+  CoverResult { coverMissingClauses = missing } <- cover f cs sc
   return $ null missing
 
+data CoverResult = CoverResult
+  { coverSplitTree       :: SplitTree
+  , coverUsedClauses     :: Set Nat
+  , coverMissingClauses  :: [(Telescope, [NamedArg DeBruijnPattern])]
+  , coverNoExactClauses  :: Set Nat
+  }
+
 -- | @cover f cs (SClause _ _ ps _) = return (splitTree, used, pss)@.
 --   checks that the list of clauses @cs@ covers the given split clause.
 --   Returns the @splitTree@, the @used@ clauses, and missing cases @pss@.
+--
+--   Effect: adds missing instance clauses for @f@ to signature.
+--
 cover :: QName -> [Clause] -> SplitClause ->
-         TCM (SplitTree, Set Nat, [[NamedArg DeBruijnPattern]])
+         TCM CoverResult
 cover f cs sc@(SClause tel ps _ _ target) = do
-  reportSDoc "tc.cover.cover" 10 $ vcat
+  reportSDoc "tc.cover.cover" 10 $ inTopContext $ vcat
     [ text "checking coverage of pattern:"
     , nest 2 $ text "tel  =" <+> prettyTCM tel
     , nest 2 $ text "ps   =" <+> do addContext tel $ prettyTCMPatternList ps
     ]
-  exactSplitEnabled <- optExactSplit <$> pragmaOptions
   cs' <- normaliseProjP cs
   case match cs' ps of
-    Yes (i,(mps,ls0))
-     | not exactSplitEnabled || (clauseCatchall (cs !! i) || all isTrivialPattern mps)
-     -> do
+    Yes (i,(mps,ls0)) -> do
+      exact <- allM mps isTrivialPattern
+      let noExactClause = if exact || clauseCatchall (cs !! i)
+                          then Set.empty
+                          else Set.singleton i
       reportSLn "tc.cover.cover" 10 $ "pattern covered by clause " ++ show i
       -- Check if any earlier clauses could match with appropriate literals
       let lsis = mapMaybe (\(j,c) -> (,j) <$> matchLits c ps) $ zip [0..i-1] cs
@@ -201,57 +255,45 @@ cover f cs sc@(SClause tel ps _ _ target) = do
       -- If we have several literal matches with the same literals
       -- only take the first matching clause of these.
       let is = Map.elems $ Map.fromListWith min $ (ls0,i) : lsis
-      return (SplittingDone (size tel), Set.fromList is, [])
-
-     | otherwise -> do
-         reportSDoc "tc.cover.cover" 10 $ vcat
-           [ text $ "pattern covered by clause " ++ show i ++ " but case splitting was not exact. remaining mpats: "
-           , nest 2 $ vcat $ map (text . show) mps
-           ]
-         setCurrentRange (cs !! i) $
-           typeError $ CoverageNoExactSplit f (cs !! i)
+      return $ CoverResult (SplittingDone (size tel)) (Set.fromList is) [] noExactClause
 
     No        ->  do
       reportSLn "tc.cover" 20 $ "pattern is not covered"
       case fmap getHiding target of
-        Just h | h == Instance -> do
+        Just h | isInstance h -> do
           -- Ulf, 2016-10-31: For now we only infer instance clauses. It would
           -- make sense to do it also for hidden, but since the value of a
           -- hidden clause is expected to be forced by later clauses, it's too
           -- late to add it now. If it was inferrable we would have gotten a
           -- type error before getting to this point.
           inferMissingClause f sc
-          return (SplittingDone (size tel), Set.empty, [])
-        _ -> return (SplittingDone (size tel), Set.empty, [ps])
+          return $ CoverResult (SplittingDone (size tel)) Set.empty [] Set.empty
+        _ -> return $ CoverResult (SplittingDone (size tel)) Set.empty [(tel, ps)] Set.empty
 
     -- We need to split!
     -- If all clauses have an unsplit copattern, we try that first.
     Block res bs -> tryIf (getAny res) splitRes $ do
-      let done = return (SplittingDone (size tel), Set.empty, [ps])
+      let done = return $ CoverResult (SplittingDone (size tel)) Set.empty [(tel, ps)] Set.empty
       if null bs then done else do
       -- Otherwise, if there are variables to split, we try them
       -- in the order determined by a split strategy.
-      reportSLn "tc.cover.strategy" 20 $ "blocking vars = " ++ show bs
+      reportSLn "tc.cover.strategy" 20 $ "blocking vars = " ++ prettyShow bs
       -- xs is a non-empty lists of blocking variables
       -- try splitting on one of them
       xs <- splitStrategy bs tel
       r <- altM1 (split Inductive sc) xs
       case r of
-        Left err  -> case err of
-          CantSplit c tel us vs   -> typeError $ CoverageCantSplitOn c tel us vs
-          NotADatatype a          -> enterClosure a $ typeError . CoverageCantSplitType
-          IrrelevantDatatype a    -> enterClosure a $ typeError . CoverageCantSplitIrrelevantType
-          CoinductiveDatatype a   -> enterClosure a $ typeError . CoverageCantSplitType
-{- UNUSED
-          NoRecordConstructor a   -> typeError $ CoverageCantSplitType a
- -}
-          GenericSplitError s     -> fail $ "failed to split: " ++ s
+        Left err -> typeError $ SplitError err
         -- If we get the empty covering, we have reached an impossible case
         -- and are done.
         Right (Covering n []) ->
-          return (SplittingDone (size tel), Set.empty, [])
+          return $ CoverResult (SplittingDone (size tel)) Set.empty [] Set.empty
         Right (Covering n scs) -> do
-          (trees, useds, psss) <- unzip3 <$> mapM (cover f cs) (map snd scs)
+          results <- mapM (cover f cs) (map snd scs)
+          let trees = map coverSplitTree results
+              useds = map coverUsedClauses results
+              psss  = map coverMissingClauses results
+              noex  = map coverNoExactClauses results
           -- Jesper, 2016-03-10  We need to remember which variables were
           -- eta-expanded by the unifier in order to generate a correct split
           -- tree (see Issue 1872).
@@ -265,7 +307,7 @@ cover f cs sc@(SClause tel ps _ _ target) = do
             ]
           let trees' = zipWith (etaRecordSplits (unArg n) ps) scs trees
               tree   = SplitAt n trees'
-          return (tree, Set.unions useds, concat psss)
+          return $ CoverResult tree (Set.unions useds) (concat psss) (Set.unions noex)
 
   where
     tryIf :: Monad m => Bool -> m (Maybe a) -> m a -> m a
@@ -273,21 +315,25 @@ cover f cs sc@(SClause tel ps _ _ target) = do
     tryIf False me m = m
 
     -- Try to split result
-    splitRes :: TCM (Maybe (SplitTree, Set Nat, [[NamedArg DeBruijnPattern]]))
+    splitRes :: TCM (Maybe CoverResult)
     splitRes = do
       reportSLn "tc.cover" 20 $ "blocked by projection pattern"
       -- forM is a monadic map over a Maybe here
       mcov <- splitResult f sc
       Trav.forM mcov $ \ (Covering n scs) -> do
         -- If result splitting was successful, continue coverage checking.
-        (projs, (trees, useds, psss)) <- mapSnd unzip3 . unzip <$> do
+        (projs, results) <- unzip <$> do
           mapM (traverseF $ cover f cs <=< (snd <.> fixTarget)) scs
           -- OR:
           -- forM scs $ \ (proj, sc') -> (proj,) <$> do
           --   cover f cs =<< do
           --     snd <$> fixTarget sc'
-        let tree = SplitAt n $ zip projs trees
-        return (tree, Set.unions useds, concat psss)
+        let trees = map coverSplitTree results
+            useds = map coverUsedClauses results
+            psss  = map coverMissingClauses results
+            noex  = map coverNoExactClauses results
+            tree  = SplitAt n $ zip projs trees
+        return $ CoverResult tree (Set.unions useds) (concat psss) (Set.unions noex)
 
     gatherEtaSplits :: Int -> SplitClause
                     -> [NamedArg DeBruijnPattern] -> [NamedArg DeBruijnPattern]
@@ -300,6 +346,7 @@ cover f cs sc@(SClause tel ps _ _ target) = do
        | n == 0    -> case p' of -- this is the main split
            VarP  _      -> __IMPOSSIBLE__
            DotP  _      -> __IMPOSSIBLE__
+           AbsurdP _    -> __IMPOSSIBLE__
            ConP  _ _ qs -> qs ++ gatherEtaSplits (-1) sc ps
            LitP  _      -> __IMPOSSIBLE__
            ProjP{}      -> __IMPOSSIBLE__
@@ -307,6 +354,7 @@ cover f cs sc@(SClause tel ps _ _ target) = do
            updateNamedArg (\ _ -> p') p : gatherEtaSplits (n-1) sc ps
         where p' = lookupS (scSubst sc) $ dbPatVarIndex x
       DotP  _      -> p : gatherEtaSplits (n-1) sc ps -- count dot patterns
+      AbsurdP _    -> p : gatherEtaSplits (n-1) sc ps
       ConP  _ _ qs -> gatherEtaSplits n sc (qs ++ ps)
       LitP  _      -> gatherEtaSplits n sc ps
       ProjP{}      -> gatherEtaSplits n sc ps
@@ -316,6 +364,7 @@ cover f cs sc@(SClause tel ps _ _ target) = do
     addEtaSplits k (p:ps) t = case namedArg p of
       VarP  _       -> addEtaSplits (k+1) ps t
       DotP  _       -> addEtaSplits (k+1) ps t
+      AbsurdP _     -> addEtaSplits (k+1) ps t
       ConP c cpi qs -> SplitAt (p $> k) [(conName c , addEtaSplits k (qs ++ ps) t)]
       LitP  _       -> __IMPOSSIBLE__
       ProjP{}       -> __IMPOSSIBLE__
@@ -325,21 +374,30 @@ cover f cs sc@(SClause tel ps _ _ target) = do
     etaRecordSplits n ps (q , sc) t =
       (q , addEtaSplits 0 (gatherEtaSplits n sc ps) t)
 
-inferMissingClause :: QName -> SplitClause -> TCM ()
+-- | Append a instance clause to the clauses of a function.
+inferMissingClause
+  :: QName
+       -- ^ Function name.
+  -> SplitClause
+       -- ^ Clause to add.  Clause hiding (in 'clauseType') must be 'Instance'.
+   -> TCM ()
 inferMissingClause f (SClause tel ps _ mpsub (Just t)) = setCurrentRange f $ do
   reportSDoc "tc.cover.infer" 20 $ addContext tel $ text "Trying to infer right-hand side of type" <+> prettyTCM t
   cl <- addContext tel $ withModuleParameters mpsub $ do
-    (x, rhs) <- case getHiding t of
-                  Instance  -> newIFSMeta "" (unArg t)
-                  Hidden    -> __IMPOSSIBLE__
-                  NotHidden -> __IMPOSSIBLE__
-    return $ Clause { clauseRange     = noRange
+    (_x, rhs) <- case getHiding t of
+                  Instance{} -> newIFSMeta "" (unArg t)
+                  Hidden     -> __IMPOSSIBLE__
+                  NotHidden  -> __IMPOSSIBLE__
+    return $ Clause { clauseLHSRange  = noRange
+                    , clauseFullRange = noRange
                     , clauseTel       = tel
                     , namedClausePats = ps
                     , clauseBody      = Just rhs
                     , clauseType      = Just t
-                    , clauseCatchall  = False }
-  addClauses f [cl]
+                    , clauseCatchall  = False
+                    , clauseUnreachable = Just False  -- missing, thus, not unreachable
+                    }
+  addClauses f [cl]  -- Important: add at the end.
 inferMissingClause _ (SClause _ _ _ _ Nothing) = __IMPOSSIBLE__
 
 splitStrategy :: BlockingVars -> Telescope -> TCM BlockingVars
@@ -362,12 +420,12 @@ splitStrategy bs tel = return $ updateLast clearBlockingVarCons xs
 -- | Check that a type is a non-irrelevant datatype or a record with
 -- named constructor. Unless the 'Induction' argument is 'CoInductive'
 -- the data type must be inductive.
-isDatatype :: (MonadTCM tcm, MonadException SplitError tcm) =>
+isDatatype :: (MonadTCM tcm, MonadError SplitError tcm) =>
               Induction -> Dom Type ->
               tcm (QName, [Arg Term], [Arg Term], [QName])
 isDatatype ind at = do
   let t       = unDom at
-      throw f = throwException . f =<< do liftTCM $ buildClosure t
+      throw f = throwError . f =<< do liftTCM $ buildClosure t
   t' <- liftTCM $ reduce t
   case ignoreSharing $ unEl t' of
     Def d es -> do
@@ -382,7 +440,7 @@ isDatatype ind at = do
           | isIrrelevant at && not splitOnIrrelevantDataAllowed ->
               throw IrrelevantDatatype
           | otherwise -> do
-              let (ps, is) = genericSplitAt np args
+              let (ps, is) = splitAt np args
               return (d, ps, is, cs)
         Record{recPars = np, recConHead = con, recInduction = i}
           | i == Just CoInductive && ind /= CoInductive ->
@@ -420,7 +478,7 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParam
     let n         = size tel
         -- Andreas, 2016-10-04 issue #2236
         -- Need to set origin to "Inserted" to avoid printing of hidden patterns.
-        xs        = map (setOrigin Inserted) $ teleNamedArgs tel
+        xs        = map (mapArgInfo hiddenInserted) $ teleNamedArgs tel
         -- Compute new split clause
         sctel'    = telFromList $ telToList (raise n sctel) ++ telToList tel
         -- Dot patterns in @ps@ need to be raised!  (Issue 1298)
@@ -455,6 +513,13 @@ fixTarget sc@SClause{ scTel = sctel, scPats = ps, scSubst = sigma, scModuleParam
       ]
     return $ if n == 0 then (empty, sc { scTarget = newTarget }) else (tel, sc')
 
+-- Andreas, 2017-01-18, issue #819, set visible arguments to UserWritten.
+-- Otherwise, they will be printed as _.
+hiddenInserted :: ArgInfo -> ArgInfo
+hiddenInserted ai
+  | visible ai = setOrigin UserWritten ai
+  | otherwise  = setOrigin Inserted ai
+
 -- | @computeNeighbourhood delta1 delta2 d pars ixs hix tel ps mpsub con@
 --
 --   @
@@ -490,13 +555,9 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
   dtype <- liftTCM $ (`piApply` pars) . defType <$> getConstInfo d
 
   -- Get the real constructor name
-  con <- liftTCM $ getConForm c
+  con <- liftTCM $ fromRight __IMPOSSIBLE__ <$> getConForm c
   con <- return $ con { conName = c }  -- What if we restore the current name?
                                        -- Andreas, 2013-11-29 changes nothing!
-{-
-  con <- conSrcCon . theDef <$> getConstInfo con
-  Con con ci [] <- liftTCM $ ignoreSharing <$> (constructorForm =<< normalise (Con con ci []))
--}
 
   -- Get the type of the constructor
   ctype <- liftTCM $ defType <$> getConInfo con
@@ -513,7 +574,7 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
 
   let preserve (x, t@(El _ (Def d' _))) | d == d' = (n, t)
       preserve (x, (El s (Shared p))) = preserve (x, El s $ derefPtr p)
-      preserve p = p
+      preserve (x, t) = (x, t)
       gammal = map (fmap preserve) . telToList $ gamma0
       gamma  = telFromList gammal
       delta1Gamma = delta1 `abstract` gamma
@@ -537,20 +598,21 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
   case r of
     NoUnify {} -> debugNoUnify $> Nothing
 
-    DontKnow{} -> do
+    DontKnow errs -> do
       debugCantSplit
-      throwException $ CantSplit (conName con) (delta1 `abstract` gamma) conIxs givenIxs
+      throwError $ UnificationStuck (conName con) (delta1 `abstract` gamma) conIxs givenIxs errs
     Unifies (delta1',rho0,_) -> do
       debugSubst "rho0" rho0
 
       -- We have Δ₁' ⊢ ρ₀ : Δ₁Γ, so split it into the part for Δ₁ and the part for Γ
       let (rho1,rho2) = splitS (size gamma) rho0
 
-      -- Andreas, 2015-05-01  I guess it is fine to use @noConPatternInfo@
+      -- Andreas, 2015-05-01  I guess it is fine to use no @conPType@
       -- as the result of splitting is never used further down the pipeline.
       -- After splitting, Agda reloads the file.
-      let conp    = ConP con noConPatternInfo $ applySubst rho2 $
-                      map (setOrigin Inserted) $ tele2NamedArgs gamma0 gamma
+      -- Andreas, 2017-09-03, issue #2729: remember that pattern was generated by case split.
+      let conp    = ConP con (ConPatternInfo (Just ConOSplit) Nothing) $ applySubst rho2 $
+                      map (mapArgInfo hiddenInserted) $ tele2NamedArgs gamma0 gamma
           -- Andreas, 2016-09-08, issue #2166: use gamma0 for correct argument names
 
       -- Compute final context and substitution
@@ -561,11 +623,11 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
 
       debugTel "delta'" delta'
       debugSubst "rho" rho
-      addContext tel $ debugPs ps
+      debugPs tel ps
 
       -- Apply the substitution
       let ps' = applySubst rho ps
-      addContext delta' $ debugPlugged ps'
+      debugPlugged delta' ps'
 
       let mpsub' = applySubst (fromPatternSubstitution rho) mpsub
 
@@ -579,14 +641,14 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
           [ text "context=" <+> (inTopContext . prettyTCM =<< getContextTelescope)
           , text "con    =" <+> prettyTCM con
           , text "ctype  =" <+> prettyTCM ctype
-          , text "ps     =" <+> do addContext tel $ prettyTCMPatternList ps
+          , text "ps     =" <+> do inTopContext $ addContext tel $ prettyTCMPatternList ps
           , text "d      =" <+> prettyTCM d
-          , text "pars   =" <+> prettyList (map prettyTCM pars)
-          , text "ixs    =" <+> addContext delta1 (prettyList (map prettyTCM ixs))
-          , text "cixs   =" <+> do addContext gamma $ prettyList (map prettyTCM cixs)
-          , text "delta1 =" <+> prettyTCM delta1
-          , text "delta2 =" <+> prettyTCM delta2
-          , text "gamma  =" <+> prettyTCM gamma
+          , text "pars   =" <+> do prettyList $ map prettyTCM pars
+          , text "ixs    =" <+> do addContext delta1 $ prettyList $ map prettyTCM ixs
+          , text "cixs   =" <+> do addContext gamma  $ prettyList $ map prettyTCM cixs
+          , text "delta1 =" <+> do inTopContext $ prettyTCM delta1
+          , text "delta2 =" <+> do inTopContext $ addContext delta1 $ addContext gamma $ prettyTCM delta2
+          , text "gamma  =" <+> do inTopContext $ addContext delta1 $ prettyTCM gamma
           , text "hix    =" <+> text (show hix)
           ]
         ]
@@ -607,15 +669,17 @@ computeNeighbourhood delta1 n delta2 d pars ixs hix tel ps mpsub c = do
         [ text (s ++ " =") <+> prettyTCM tel
         ]
 
-    debugPs ps =
-      liftTCM $ reportSDoc "tc.cover.split.con" 20 $ nest 2 $ vcat
-        [ text "ps     =" <+> prettyTCMPatternList ps
-        ]
+    debugPs tel ps =
+      liftTCM $ reportSDoc "tc.cover.split.con" 20 $
+        inTopContext $ addContext tel $ nest 2 $ vcat
+          [ text "ps     =" <+> prettyTCMPatternList ps
+          ]
 
-    debugPlugged ps' =
-      liftTCM $ reportSDoc "tc.cover.split.con" 20 $ nest 2 $ vcat
-        [ text "ps'    =" <+> do prettyTCMPatternList ps'
-        ]
+    debugPlugged delta' ps' =
+      liftTCM $ reportSDoc "tc.cover.split.con" 20 $
+        inTopContext $ addContext delta' $ nest 2 $ vcat
+          [ text "ps'    =" <+> do prettyTCMPatternList ps'
+          ]
 
 -- | Entry point from @Interaction.MakeCase@.
 splitClauseWithAbsurd :: SplitClause -> Nat -> TCM (Either SplitError (Either SplitClause Covering))
@@ -691,14 +755,14 @@ split' :: Induction
        -> SplitClause
        -> BlockingVar
        -> TCM (Either SplitError (Either SplitClause Covering))
-split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) = liftTCM $ runExceptionT $ do
+split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) = liftTCM $ runExceptT $ do
 
   debugInit tel x ps mpsub
 
   -- Split the telescope at the variable
   -- t = type of the variable,  Δ₁ ⊢ t
   (n, t, delta1, delta2) <- do
-    let (tel1, Dom info (n, t) : tel2) = genericSplitAt (size tel - x - 1) $ telToList tel
+    let (tel1, Dom info (n, t) : tel2) = splitAt (size tel - x - 1) $ telToList tel
     return (n, Dom info t, telFromList tel1, telFromList tel2)
 
   -- Check that t is a datatype or a record
@@ -735,7 +799,7 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
     -- if more than one constructor matches, we cannot be irrelevant
     -- (this piece of code is unreachable if --experimental-irrelevance is off)
     (_ : _ : _) | unusableRelevance (getRelevance t) ->
-      throwException . IrrelevantDatatype =<< do liftTCM $ buildClosure (unDom t)
+      throwError . IrrelevantDatatype =<< do liftTCM $ buildClosure (unDom t)
 
   -- Andreas, 2012-10-10 fail if precomputed constructor set does not cover
   -- all the data type constructors
@@ -749,7 +813,7 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
             [ hsep $ text "pcons =" : map prettyTCM pcons
             , hsep $ text "cons  =" : map prettyTCM cons
             ]
-          throwException (GenericSplitError "precomputed set of constructors does not cover all cases")
+          throwError (GenericSplitError "precomputed set of constructors does not cover all cases")
 
     _  -> return $ Right $ Covering (lookupPatternVar sc x) ns
 
@@ -761,12 +825,12 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
     inContextOfDelta2 = addContext tel . escapeContext x
 
     -- Debug printing
-    debugInit tel x ps mpsub = liftTCM $ do
+    debugInit tel x ps mpsub = liftTCM $ inTopContext $ do
       reportSDoc "tc.cover.top" 10 $ vcat
         [ text "TypeChecking.Coverage.split': split"
         , nest 2 $ vcat
           [ text "tel     =" <+> prettyTCM tel
-          , text "x       =" <+> text (show x)
+          , text "x       =" <+> prettyTCM x
           , text "ps      =" <+> do addContext tel $ prettyTCMPatternList ps
           , text "mpsub   =" <+> prettyTCM mpsub
           ]
@@ -775,7 +839,7 @@ split' ind fixtarget sc@(SClause tel ps _ mpsub target) (BlockingVar x mcons) =
     debugHoleAndType delta1 delta2 s ps t =
       liftTCM $ reportSDoc "tc.cover.top" 10 $ nest 2 $ vcat $
         [ text "p      =" <+> text (patVarNameToString s)
-        , text "ps    =" <+> text (show ps)
+        , text "ps     =" <+> text (show ps)
         , text "delta1 =" <+> prettyTCM delta1
         , text "delta2 =" <+> inContextOfDelta2 (prettyTCM delta2)
         , text "t      =" <+> inContextOfT (prettyTCM t)
@@ -791,7 +855,7 @@ splitResult :: QName -> SplitClause -> TCM (Maybe Covering)
 splitResult f sc@(SClause tel ps _ _ target) = do
   reportSDoc "tc.cover.split" 10 $ vcat
     [ text "splitting result:"
-    , nest 2 $ text "f      =" <+> text (show f)
+    , nest 2 $ text "f      =" <+> prettyTCM f
     , nest 2 $ text "target =" <+> (addContext tel $ maybe empty prettyTCM target)
     ]
   -- if we want to split projections, but have no target type, we give up
@@ -801,9 +865,9 @@ splitResult f sc@(SClause tel ps _ _ target) = do
     case isR of
       Just (_r, vs, Record{ recFields = fs }) -> do
         reportSDoc "tc.cover" 20 $ sep
-          [ text $ "we are of record type _r = " ++ show _r
+          [ text $ "we are of record type _r = " ++ prettyShow _r
           , text   "applied to parameters vs = " <+> (addContext tel $ prettyTCM vs)
-          , text $ "and have fields       fs = " ++ show fs
+          , text $ "and have fields       fs = " ++ prettyShow fs
           ]
         let es = patternsToElims ps
         -- Note: module parameters are part of ps
