@@ -16,11 +16,15 @@ import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern ()
 import Agda.Syntax.Literal
 
+import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Records
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty ( Pretty(pretty), text, (<+>) )
 import Agda.Utils.Size
 import Agda.Utils.List
+import Agda.Utils.Monad
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -68,13 +72,15 @@ buildPattern (Shared p)     = buildPattern (derefPtr p)
 buildPattern _              = Nothing
 
 -- | A pattern that matches anything (modulo eta).
-isTrivialPattern :: Pattern' a -> Bool
+isTrivialPattern :: (HasConstInfo m) => Pattern' a -> m Bool
 isTrivialPattern p = case p of
-  VarP{}      -> True
-  DotP{}      -> True
-  ConP c i ps -> isJust (conPRecord i) && all (isTrivialPattern . namedArg) ps
-  LitP{}      -> False
-  ProjP{}     -> False
+  VarP{}      -> return True
+  DotP{}      -> return True
+  AbsurdP{}   -> return True
+  ConP c i ps -> andM $ (isEtaCon $ conName c)
+                      : (map (isTrivialPattern . namedArg) ps)
+  LitP{}      -> return False
+  ProjP{}     -> return False
 
 -- | If matching succeeds, we return the instantiation of the clause pattern vector
 --   to obtain the split clause pattern vector, plus the literals of the clause patterns
@@ -105,6 +111,13 @@ data BlockingVar  = BlockingVar
     --   @Just cons@ means that it is an non-overlapping match and
     --   @cons@ are the encountered constructors.
   } deriving (Show)
+
+instance Pretty BlockingVar where
+  pretty (BlockingVar i mcs) = case mcs of
+    Nothing -> text $ "blocking var " ++ show i ++ " (overlapping)"
+    Just cs -> text ("blocking var " ++ show i ++ " (non-overlapping); constructors:")
+      <+> pretty cs
+
 type BlockingVars = [BlockingVar]
 
 -- | Lens for 'blockingVarCons'.
@@ -156,6 +169,7 @@ yesMatchLit l q@VarP{} = Yes ([q], [l])
 yesMatchLit l (DotP t) = maybe No (yesMatchLit l) $ buildPattern t
 yesMatchLit _ ConP{}   = No
 yesMatchLit _ ProjP{}  = No
+yesMatchLit _ AbsurdP{} = __IMPOSSIBLE__
 yesMatchLit _ LitP{}   = __IMPOSSIBLE__
 
 -- | Check if a clause could match given generously chosen literals
@@ -246,15 +260,15 @@ matchPats mlit ps qs = mconcat $ [ projPatternsLeftInSplitClause ] ++
 --   that have to be instantiated (an projection names of copattern matches)
 --   to make the split clause an instance of the function clause.
 
-instance Monoid a => Semigroup (Match a) where
-  Yes a      <> Yes b      = Yes $ mappend a b
+instance Semigroup a => Semigroup (Match a) where
+  Yes a      <> Yes b      = Yes (a <> b)
   Yes _      <> m          = m
   No         <> _          = No
   Block{}    <> No         = No
-  Block r xs <> Block s ys = Block (mappend r s) $ mappend xs ys
+  Block r xs <> Block s ys = Block (r <> s) (xs <> ys)
   m@Block{}  <> Yes{}      = m
 
-instance Monoid a => Monoid (Match a) where
+instance (Semigroup a, Monoid a) => Monoid (Match a) where
   mempty  = Yes mempty
   mappend = (<>)
 
@@ -281,6 +295,7 @@ matchPat
 
 matchPat _    VarP{}   q = Yes ([q],[])
 matchPat _    DotP{}   q = mempty
+matchPat _    AbsurdP{} q = mempty
 -- Jesper, 2014-11-04: putting 'Yes [q]' here triggers issue 1333.
 -- Not checking for trivial patterns should be safe here, as dot patterns are
 -- guaranteed to match if the rest of the pattern does, so some extra splitting
@@ -294,5 +309,6 @@ matchPat mlit p@(ConP c _ ps) q = case q of
     | c == c'   -> matchPats mlit ps qs
     | otherwise -> No
   DotP t  -> maybe No (matchPat mlit p) $ buildPattern t
+  AbsurdP{} -> __IMPOSSIBLE__  -- excluded by typing
   LitP _  -> __IMPOSSIBLE__  -- split clause has no literal patterns
   ProjP{} -> __IMPOSSIBLE__  -- excluded by typing
