@@ -3,6 +3,7 @@
 module Agda.TypeChecking.Datatypes where
 
 import Data.Maybe (fromMaybe)
+import qualified Data.List as List
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal
@@ -11,9 +12,12 @@ import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Pretty
 
-import Agda.Utils.Size
+import Agda.Utils.Either
 import Agda.Utils.Functor
+import Agda.Utils.Pretty ( prettyShow )
+import Agda.Utils.Size
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -23,21 +27,20 @@ import Agda.Utils.Impossible
 ---------------------------------------------------------------------------
 
 -- | Get true constructor with record fields.
-getConHead :: QName -> TCM ConHead
-getConHead c = conSrcCon . theDef <$> getConstInfo c
+getConHead :: QName -> TCM (Either SigError ConHead)
+getConHead c = mapRight (conSrcCon . theDef) <$> getConstInfo' c
 
 -- | Get true constructor with fields, expanding literals to constructors
 --   if possible.
-getConForm :: QName -> TCM ConHead
-getConForm c = do
-  ch <- getConHead c
+getConForm :: QName -> TCM (Either SigError ConHead)
+getConForm c = caseEitherM (getConHead c) (return . Left) $ \ ch -> do
   Con con _ [] <- ignoreSharing <$> constructorForm (Con ch ConOCon [])
-  return con
+  return $ Right con
 
 -- | Augment constructor with record fields (preserve constructor name).
 --   The true constructor might only surface via 'reduce'.
-getOrigConHead :: QName -> TCM ConHead
-getOrigConHead c = setConName c <$> getConHead c
+getOrigConHead :: QName -> TCM (Either SigError ConHead)
+getOrigConHead c = mapRight (setConName c) <$> getConHead c
 
 -- | Get the name of the datatype constructed by a given constructor.
 --   Precondition: The argument must refer to a constructor
@@ -50,31 +53,96 @@ getConstructorData c = do
     _                        -> __IMPOSSIBLE__
 
 -- | @getConType c t@ computes the constructor parameters from type @t@
---   and returns the instantiated type of constructor @c@.
+--   and returns them plus the instantiated type of constructor @c@.
+--   This works also if @t@ is a function type ending in a data/record type;
+--   the term from which @c@ comes need not be fully applied
+--
 --   @Nothing@ if @t@ is not a data/record type or does not have
 --   a constructor @c@.
---   Precondition: @t@ is reduced.
-getConType :: ConHead -> Type -> TCM (Maybe Type)
+getConType
+  :: ConHead  -- ^ Constructor.
+  -> Type     -- ^ Ending in data/record type.
+  -> TCM (Maybe ((QName, Type, Args), Type))
+       -- ^ @Nothing@ if not ends in data or record type.
+       --
+       --   @Just ((d, dt, pars), ct)@ otherwise, where
+       --     @d@    is the data or record type name,
+       --     @dt@   is the type of the data or record name,
+       --     @pars@ are the reconstructed parameters,
+       --     @ct@   is the type of the constructor instantiated to the parameters.
 getConType c t = do
-  c <- getConHead $ conName c
+  reportSDoc "tc.getConType" 30 $ sep $
+    [ text "getConType: constructor "
+    , prettyTCM c
+    , text " at type "
+    , prettyTCM t
+    ]
+  TelV tel t <- telView t
+  -- Now @t@ lives under @tel@, we need to remove the dependency on @tel@.
+  -- This will succeed if @t@ is indeed a data/record type that is the
+  -- type of a constructor coming from a term
+  -- (applied to at least the parameters).
+  -- Note: @t@ will have some unbound deBruijn indices if view outside of @tel@.
+  reportSLn "tc.getConType" 35 $ "  target type: " ++ prettyShow t
+  applySubst (strengthenS __IMPOSSIBLE__ (size tel)) <$> do
+    addContext tel $ getFullyAppliedConType c t
+  -- Andreas, 2017-08-18, issue #2703:
+  -- The original code
+  --    getFullyAppliedConType c $ applySubst (strengthenS __IMPOSSIBLE__ (size tel)) t
+  -- crashes because substitution into @Def@s is slightly too strict
+  -- (see @defApp@ and @canProject@).
+  -- Strengthening the parameters after the call to @getFullyAppliedConType@
+  -- does not produce intermediate terms with __IMPOSSIBLE__s and this thus
+  -- robust wrt. strictness/laziness of substitution.
+
+-- | @getFullyAppliedConType c t@ computes the constructor parameters
+--   from data type @t@ and returns them
+--   plus the instantiated type of constructor @c@.
+--
+--   @Nothing@ if @t@ is not a data/record type or does not have
+--   a constructor @c@.
+--
+--   Precondition: @t@ is reduced.
+getFullyAppliedConType
+  :: ConHead  -- ^ Constructor.
+  -> Type     -- ^ Reduced type of the fully applied constructor.
+  -> TCM (Maybe ((QName, Type, Args), Type))
+       -- ^ @Nothing@ if not data or record type.
+       --
+       --   @Just ((d, dt, pars), ct)@ otherwise, where
+       --     @d@    is the data or record type name,
+       --     @dt@   is the type of the data or record name,
+       --     @pars@ are the reconstructed parameters,
+       --     @ct@   is the type of the constructor instantiated to the parameters.
+getFullyAppliedConType c t = do
+  reportSLn "tc.getConType" 35 $ List.intercalate " " $
+    [ "getFullyAppliedConType", prettyShow c, prettyShow t ]
+  c <- fromRight __IMPOSSIBLE__ <$> do getConHead $ conName c
   case ignoreSharing $ unEl t of
+    -- Note that if we come e.g. from getConType,
+    -- then the non-parameter arguments of @es@ might contain __IMPOSSIBLE__
+    -- coming from strengthening.  (Thus, printing them is not safe.)
     Def d es -> do
-      def <- theDef <$> getConstInfo d
-      case def of
+      reportSLn "tc.getConType" 35 $ List.intercalate " " $
+        [ "getFullyAppliedConType: case Def", prettyShow d, prettyShow es ]
+      def <- getConstInfo d
+      let cont n = do
+            -- At this point we can be sure that the parameters are well-scoped.
+            let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims $ take n es
+            Just . ((d, defType def, pars),) <$> do
+              (`piApplyM` pars) . defType =<< getConInfo c
+      case theDef def of
         Datatype { dataPars = n, dataCons   = cs  } | conName c `elem` cs -> cont n
         Record   { recPars  = n, recConHead = con } | c == con            -> cont n
         _ ->  return Nothing
-      where
-        cont n = do
-          let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims $ take n es
-          Just <$> do (`piApplyM` pars) . defType =<< getConInfo c
     _ -> return Nothing
 
 data HasEta = NoEta | YesEta
   deriving (Eq)
 
-data ConstructorInfo = DataCon Nat  -- arity
-                     | RecordCon HasEta [Arg QName]
+data ConstructorInfo
+  = DataCon Nat                  -- ^ Arity.
+  | RecordCon HasEta [Arg QName] -- ^ List of field names.
 
 -- | Return the number of non-parameter arguments to a data constructor,
 --   or the field names of a record constructor.
@@ -83,28 +151,16 @@ data ConstructorInfo = DataCon Nat  -- arity
 --   use @either id size <$> getConstructorArity c@.
 getConstructorInfo :: QName -> TCM ConstructorInfo
 getConstructorInfo c = do
-  Defn{ defType = t, theDef = def } <- getConstInfo c
-  case def of
-    Constructor{ conData = d, conPars = n } -> do
-      def <- theDef <$> getConstInfo d
-      case def of
+  (theDef <$> getConstInfo c) >>= \case
+    Constructor{ conData = d, conArity = n } -> do
+      (theDef <$> getConstInfo d) >>= \case
         r@Record{ recFields = fs } ->
            return $ RecordCon (if recEtaEquality r then YesEta else NoEta) fs
-        Datatype{} -> do
-          -- TODO: I do not want to take the type of constructor apart
-          -- to see its arity!
-          TelV tel _ <- telView t
-          return $ DataCon $ size tel - n
+        Datatype{} ->
+           return $ DataCon n
         _ -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
-getConstructorArity :: QName -> TCM Nat
-getConstructorArity c =
-  for (getConstructorInfo c) $ \i ->
-  case i of
-    DataCon n      -> n
-    RecordCon _ fs -> size fs
-
 ---------------------------------------------------------------------------
 -- * Data types
 ---------------------------------------------------------------------------
@@ -118,9 +174,6 @@ isDatatype d = do
     Record{recNamedCon = namedC} -> return namedC
     _                            -> return False
 
-data DataOrRecord = IsData | IsRecord
-  deriving (Eq, Ord, Show)
-
 -- | Check if a name refers to a datatype or a record.
 isDataOrRecordType :: QName -> TCM (Maybe DataOrRecord)
 isDataOrRecordType d = do
