@@ -8,7 +8,7 @@ import Control.Monad.Reader
 import Control.Monad.State
 import qualified Data.Map as Map
 import qualified Data.Set as Set
-import Data.List as List
+import qualified Data.List as List
 
 import Agda.Syntax.Common
 import Agda.Syntax.Internal as I
@@ -45,39 +45,46 @@ import Agda.Utils.Impossible
 --   for instance search.
 initialIFSCandidates :: Type -> TCM (Maybe [Candidate])
 initialIFSCandidates t = do
-  cands1 <- getContextVars
   otn <- getOutputTypeName t
   case otn of
-    NoOutputTypeName -> typeError $ GenericError $ "Instance search can only be used to find elements in a named type"
+    NoOutputTypeName -> typeError $ GenericError $
+      "Instance search can only be used to find elements in a named type"
     OutputTypeNameNotYetKnown -> return Nothing
-    OutputTypeVar -> return $ Just cands1
-    OutputTypeName n -> do
-      cands2 <- getScopeDefs n
-      return $ Just $ cands1 ++ cands2
+    OutputTypeVar    -> Just <$> getContextVars
+    OutputTypeName n -> Just <$> do (++) <$> getContextVars <*> getScopeDefs n
   where
     -- get a list of variables with their type, relative to current context
     getContextVars :: TCM [Candidate]
     getContextVars = do
       ctx <- getContext
-      let vars = [ Candidate (var i) (raise (i + 1) t) ExplicitStayExplicit (argInfoOverlappable info)
-                 | (Dom info (x, t), i) <- zip ctx [0..]
-                 , getHiding info == Instance
+      reportSDoc "tc.instance.cands" 40 $ hang (text "Getting candidates from context") 2 (inTopContext $ prettyTCM ctx)
+          -- Context variables with their types lifted to live in the full context
+      let varsAndRaisedTypes = [ (var i, raise (i + 1) t) | (i, t) <- zip [0..] ctx ]
+          vars = [ Candidate x t ExplicitStayExplicit (isOverlappable info)
+                 | (x, Dom info (_, t)) <- varsAndRaisedTypes
+                 , isInstance info
                  , not (unusableRelevance $ argInfoRelevance info)
                  ]
 
       -- {{}}-fields of variables are also candidates
-      let cxtAndTypes = [ (var i, snd $ unDom $ raise (i + 1) t) | (i, t) <- zip [0..] ctx ]
+      let cxtAndTypes = [ (x, t) | (x, Dom _ (_, t)) <- varsAndRaisedTypes ]
       fields <- concat <$> mapM instanceFields (reverse cxtAndTypes)
-      reportSDoc "tc.instance.fields" 30 $ text "instance field candidates" $$ nest 2 (vcat
-        [ sep [ (if overlap then text "overlap" else empty) <+> prettyTCM v <+> text ":"
-              , nest 2 $ prettyTCM t ] | Candidate v t _ overlap <- fields ])
+      reportSDoc "tc.instance.fields" 30 $
+        if null fields then text "no instance field candidates" else
+          text "instance field candidates" $$ do
+            nest 2 $ vcat
+              [ sep [ (if overlap then text "overlap" else empty) <+> prettyTCM v <+> text ":"
+                    , nest 2 $ prettyTCM t
+                    ]
+              | Candidate v t _ overlap <- fields
+              ]
 
       -- get let bindings
       env <- asks envLetBindings
       env <- mapM (getOpen . snd) $ Map.toList env
       let lets = [ Candidate v t ExplicitStayExplicit False
                  | (v, Dom info t) <- env
-                 , getHiding info == Instance
+                 , isInstance info
                  , not (unusableRelevance $ argInfoRelevance info)
                  ]
       return $ vars ++ fields ++ lets
@@ -91,7 +98,7 @@ initialIFSCandidates t = do
               m <- currentModule
               -- Are we inside the record module? If so it's safe and desirable
               -- to eta-expand once (issue #2320).
-              if qnameToList r `isPrefixOf` mnameToList m
+              if qnameToList r `List.isPrefixOf` mnameToList m
                 then return (Just (r, vs))
                 else return Nothing
         r -> return r
@@ -102,8 +109,8 @@ initialIFSCandidates t = do
         (tel, args) <- forceEtaExpandRecord r pars v
         let types = map unDom $ applySubst (parallelS $ reverse $ map unArg args) (flattenTel tel)
         fmap concat $ forM (zip args types) $ \ (arg, t) ->
-          ([ Candidate (unArg arg) t ExplicitStayExplicit (argInfoOverlappable $ argInfo arg)
-           | getHiding arg == Instance ] ++) <$>
+          ([ Candidate (unArg arg) t ExplicitStayExplicit (isOverlappable arg)
+           | isInstance arg ] ++) <$>
           instanceFields' False (unArg arg, t)
 
     getScopeDefs :: QName -> TCM [Candidate]
@@ -114,17 +121,19 @@ initialIFSCandidates t = do
       catMaybes <$> mapM (candidate rel) qs
 
     candidate :: Relevance -> QName -> TCM (Maybe Candidate)
-    candidate rel q =
+    candidate rel q = ifNotM (isNameInScope q <$> getScope) (return Nothing) $ do
       -- Andreas, 2012-07-07:
       -- we try to get the info for q
       -- while opening a module, q may be in scope but not in the signature
       -- in this case, we just ignore q (issue 674)
       flip catchError handle $ do
         def <- getConstInfo q
-        let r = defRelevance def
-        if not (r `moreRelevant` rel) then return Nothing else do
-          t   <- defType <$> instantiateDef def
+        if not (defRelevance def `moreRelevant` rel) then return Nothing else do
+          -- Andreas, 2017-01-14: instantiateDef is a bit of an overkill
+          -- if we anyway get the freeVarsToApply
+          -- WAS: t <- defType <$> instantiateDef def
           args <- freeVarsToApply q
+          let t = defType def `piApply` args
           let v = case theDef def of
                -- drop parameters if it's a projection function...
                Function{ funProjection = Just p } -> projDropParsApply p ProjSystem args
@@ -134,8 +143,7 @@ initialIFSCandidates t = do
                -- Ulf, 2014-08-20: constructors are always instances.
                Constructor{ conSrcCon = c }       -> Con c ConOSystem []
                _                                  -> Def q $ map Apply args
-          inScope <- isNameInScope q <$> getScope
-          return $ Candidate v t ExplicitToInstance False <$ guard inScope
+          return $ Just $ Candidate v t ExplicitToInstance False
       where
         -- unbound constant throws an internal error
         handle (TypeError _ (Closure {clValue = InternalError _})) = return Nothing
@@ -156,33 +164,33 @@ findInScope m Nothing = do
   mv <- lookupMeta m
   setCurrentRange mv $ do
     reportSLn "tc.instance" 20 $ "The type of the FindInScope constraint isn't known, trying to find it again."
-    t <- getMetaType m
+    t <- instantiate =<< getMetaTypeInContext m
+    reportSLn "tc.instance" 70 $ "findInScope 1: t: " ++ prettyShow t
 
---     -- We create a new meta (which can have additional leading lambdas, if the
---     -- type @t@ now happens to be a function type) and the associated constraint
---     newM <- initializeIFSMeta (miNameSuggestion $ mvInfo mv) t
-
---     -- ... and we assign it to the previous one
---     ctxElims <- map Apply <$> getContextArgs
---     solveConstraint $ ValueCmp CmpEq t (MetaV m ctxElims) newM
-
--- {-
-    cands <- initialIFSCandidates t
+    -- Issue #2577: If the target is a function type the arguments are
+    -- potential candidates, so we add them to the context to make
+    -- initialIFSCandidates pick them up.
+    TelV tel t <- telView t
+    cands <- addContext' tel $ initialIFSCandidates t
     case cands of
-      Nothing -> addConstraint $ FindInScope m Nothing Nothing
+      Nothing -> do
+        reportSLn "tc.instance" 20 "Can't figure out target of instance goal. Postponing constraint."
+        addConstraint $ FindInScope m Nothing Nothing
       Just {} -> findInScope m cands
--- -}
+
 findInScope m (Just cands) =
   whenJustM (findInScope' m cands) $ (\ (cands, b) -> addConstraint $ FindInScope m b $ Just cands)
 
 -- | Result says whether we need to add constraint, and if so, the set of
 --   remaining candidates and an eventual blocking metavariable.
 findInScope' :: MetaId -> [Candidate] -> TCM (Maybe ([Candidate], Maybe MetaId))
-findInScope' m cands = ifM (isFrozen m) (return (Just (cands, Nothing))) $ do
+findInScope' m cands = ifM (isFrozen m) (do
+    reportSLn "tc.instance" 20 "Refusing to solve frozen instance meta."
+    return (Just (cands, Nothing))) $ do
   -- Andreas, 2013-12-28 issue 1003:
   -- If instance meta is already solved, simply discard the constraint.
   -- Ulf, 2016-12-06 issue 2325: But only if *fully* instantiated.
-  ifM (isFullyInstantiatedMeta m) (return Nothing) $ do
+  ifM (isFullyInstantiatedMeta m) (Nothing <$ reportSLn "tc.instance" 20 "Instance meta already solved.") $ do
     -- Andreas, 2015-02-07: New metas should be created with range of the
     -- current instance meta, thus, we set the range.
     mv <- lookupMeta m
@@ -192,16 +200,23 @@ findInScope' m cands = ifM (isFrozen m) (return (Just (cands, Nothing))) $ do
       reportSDoc "tc.instance" 60 $ nest 2 $ vcat
         [ sep [ (if overlap then text "overlap" else empty) <+> prettyTCM v <+> text ":"
               , nest 2 $ prettyTCM t ] | Candidate v t _ overlap <- cands ]
+      reportSDoc "tc.instance" 70 $ text "raw" $$ do
+       nest 2 $ vcat
+        [ sep [ (if overlap then text "overlap" else empty) <+> pretty v <+> text ":"
+              , nest 2 $ pretty t ] | Candidate v t _ overlap <- cands ]
       t <- normalise =<< getMetaTypeInContext m
+      reportSLn "tc.instance" 70 $ "findInScope 2: t: " ++ prettyShow t
       insidePi t $ \ t -> do
       reportSDoc "tc.instance" 15 $ text "findInScope 3: t =" <+> prettyTCM t
-      reportSLn "tc.instance" 70 $ "findInScope 3: t: " ++ show t
+      reportSLn "tc.instance" 70 $ "findInScope 3: t: " ++ prettyShow t
 
       -- If one of the arguments of the typeclass is a meta which is not rigidly
       -- constrained, then don’t do anything because it may loop.
-      ifJustM (areThereNonRigidMetaArguments (unEl t)) (\ m -> do
-        reportSLn "tc.instance" 15 "aborting due to non-rigidly constrained metas"
-        return (Just (cands, Just m))) $ do
+      let abortNonRigid m = do
+            reportSLn "tc.instance" 15 $
+              "aborting due to non-rigidly constrained meta " ++ show m
+            return $ Just (cands, Just m)
+      ifJustM (areThereNonRigidMetaArguments (unEl t)) abortNonRigid $ {-else-} do
 
         mcands <- checkCandidates m t cands
         debugConstraints
@@ -237,7 +252,7 @@ findInScope' m cands = ifM (isFrozen m) (return (Just (cands, Nothing))) $ do
 insidePi :: Type -> (Type -> TCM a) -> TCM a
 insidePi t ret =
   case ignoreSharing $ unEl t of
-    Pi a b     -> addContext' (absName b, a) $ insidePi (unAbs b) ret
+    Pi a b     -> addContext' (absName b, a) $ insidePi (absBody b) ret
     Def{}      -> ret t
     Var{}      -> ret t
     Sort{}     -> __IMPOSSIBLE__
@@ -340,16 +355,16 @@ areThereNonRigidMetaArguments t = case ignoreSharing t of
         Var _ es  -> areThereNonRigidMetaArgs es
         Con _ _ vs-> areThereNonRigidMetaArgs (map Apply vs)
         MetaV i _ -> ifM (isRigid i) (return Nothing) $ do
-                      -- Ignore unconstrained level and size metas (#1865)
-                      Def lvl [] <- ignoreSharing <$> primLevel
-                      sz <- for (fmap ignoreSharing <$> getBuiltin' builtinSize) $ \case
-                              Just (Def sz []) -> [sz]
-                              Nothing          -> []
-                              _                -> __IMPOSSIBLE__
-                      o          <- getOutputTypeName . jMetaType . mvJudgement =<< lookupMeta i
-                      case o of
-                        OutputTypeName l | elem l (lvl : sz) -> return Nothing
-                        _                                   -> return (Just i)
+          -- Ignore unconstrained level and size metas (#1865)
+          mlvl <- getBuiltinDefName builtinLevel
+          (msz, mszlt) <- getBuiltinSize
+          let ok = catMaybes [ mlvl, msz ]  -- , mszlt ]  -- ?! Andreas, 2016-12-22
+            -- @Ulf: are SIZELT metas excluded on purpose?
+            -- How to you know the level/size meta is unconstrained?
+          o <- getOutputTypeName . jMetaType . mvJudgement =<< lookupMeta i
+          case o of
+            OutputTypeName l | elem l ok -> return Nothing
+            _ -> return $ Just i
         Lam _ t   -> isNonRigidMeta (unAbs t)
         _         -> return Nothing
 
@@ -387,13 +402,13 @@ filterResetingState m cands f = disableDestructiveUpdate $ do
 -- Drop all candidates which are judgmentally equal to the first one.
 -- This is sufficient to reduce the list to a singleton should all be equal.
 dropSameCandidates :: MetaId -> [(Candidate, Term, Type, a)] -> TCM [(Candidate, Term, Type, a)]
-dropSameCandidates m cands0 = do
+dropSameCandidates m cands0 = verboseBracket "tc.instance" 30 "dropSameCandidates" $ do
   metas <- Set.fromList . Map.keys <$> getMetaStore
   let freshMetas x = not $ Set.null $ Set.difference (Set.fromList $ allMetas x) metas
 
   -- Take overlappable candidates into account
   let cands =
-        case partition (\ (c, _, _, _) -> candidateOverlappable c) cands0 of
+        case List.partition (\ (c, _, _, _) -> candidateOverlappable c) cands0 of
           (cand : _, []) -> [cand]  -- only overlappable candidates: pick the first one
           _              -> cands0  -- otherwise require equality
 
@@ -405,16 +420,18 @@ dropSameCandidates m cands0 = do
   rel <- getMetaRelevance <$> lookupMeta m
   case cands of
     []            -> return cands
+    cvd : _ | isIrrelevant rel -> do
+      reportSLn "tc.instance" 30 "Meta is irrelevant so any candidate will do."
+      return [cvd]
     cvd@(_, v, a, _) : vas -> do
         if freshMetas (v, a)
           then return (cvd : vas)
           else (cvd :) <$> dropWhileM equal vas
       where
-        equal _ | isIrrelevant rel = return True
         equal (_, v', a', _)
             | freshMetas (v', a') = return False  -- If there are fresh metas we can't compare
             | otherwise           =
-          verboseBracket "tc.instance" 30 "checkEqualCandidates" $ do
+          verboseBracket "tc.instance" 30 "comparingCandidates" $ do
           reportSDoc "tc.instance" 30 $ sep [ prettyTCM v <+> text "==", nest 2 $ prettyTCM v' ]
           localTCState $ dontAssignMetas $ ifNoConstraints_ (equalType a a' >> equalTerm a v v')
                              {- then -} (return True)
@@ -471,7 +488,7 @@ checkCandidates m t cands = disableDestructiveUpdate $
         debugConstraints
         verboseBracket "tc.instance" 20 ("checkCandidateForMeta " ++ prettyShow m) $
           liftTCM $ runCandidateCheck $ do
-            reportSLn "tc.instance" 70 $ "  t: " ++ show t ++ "\n  t':" ++ show t' ++ "\n  term: " ++ show term ++ "."
+            reportSLn "tc.instance" 70 $ "  t: " ++ prettyShow t ++ "\n  t':" ++ prettyShow t' ++ "\n  term: " ++ prettyShow term ++ "."
             reportSDoc "tc.instance" 20 $ vcat
               [ text "checkCandidateForMeta"
               , text "t    =" <+> prettyTCM t
@@ -480,20 +497,28 @@ checkCandidates m t cands = disableDestructiveUpdate $
               ]
 
             -- Apply hidden and instance arguments (recursive inst. search!).
-            (args, t'') <- implicitArgs (-1) (\h -> h /= NotHidden || eti == ExplicitToInstance) t'
+            (args, t'') <- implicitArgs (-1) (\h -> notVisible h || eti == ExplicitToInstance) t'
 
             reportSDoc "tc.instance" 20 $
               text "instance search: checking" <+> prettyTCM t''
               <+> text "<=" <+> prettyTCM t
-            ctxElims <- map Apply <$> getContextArgs
+            reportSDoc "tc.instance" 70 $ vcat
+              [ text "instance search: checking (raw)"
+              , nest 4 $ pretty t''
+              , nest 2 $ text "<="
+              , nest 4 $ pretty t
+              ]
             v <- (`applyDroppingParameters` args) =<< reduce term
             reportSDoc "tc.instance" 15 $ vcat
               [ text "instance search: attempting"
               , nest 2 $ prettyTCM m <+> text ":=" <+> prettyTCM v
               ]
+            reportSDoc "tc.instance" 70 $ nest 2 $
+              text "candidate v = " <+> pretty v
             -- if constraints remain, we abort, but keep the candidate
             -- Jesper, 05-12-2014: When we abort, we should add a constraint to
             -- instantiate the meta at a later time (see issue 1377).
+            ctxElims <- map Apply <$> getContextArgs
             guardConstraint (ValueCmp CmpEq t'' (MetaV m ctxElims) v) $ leqType t'' t
             -- make a pass over constraints, to detect cases where some are made
             -- unsolvable by the assignment, but don't do this for FindInScope's
@@ -553,7 +578,7 @@ applyDroppingParameters t vs = do
     Con c ci [] -> do
       def <- theDef <$> getConInfo c
       case def of
-        Constructor {conPars = n} -> return $ Con c ci (genericDrop n vs)
+        Constructor {conPars = n} -> return $ Con c ci (drop n vs)
         _ -> __IMPOSSIBLE__
     Def f [] -> do
       mp <- isProjection f
