@@ -4,6 +4,7 @@
 -}
 module Agda.TypeChecking.Irrelevance where
 
+import Control.Arrow (first, second)
 import Control.Applicative
 import Control.Monad.Reader
 
@@ -50,7 +51,7 @@ workOnTypes cont = do
 workOnTypes' :: Bool -> TCM a -> TCM a
 workOnTypes' experimental cont = modifyContext (modifyContextEntries $ mapRelevance f) cont
   where
-    f | experimental = irrToNonStrict
+    f | experimental = irrToNonStrict . nonStrictToRel
       | otherwise    = nonStrictToRel
 
 -- | (Conditionally) wake up irrelevant variables and make them relevant.
@@ -63,17 +64,24 @@ applyRelevanceToContext rel =
     Relevant -> id
     Forced{} -> id
     _        -> local $ \ e -> e
-      { envContext   = modifyContextEntries (inverseApplyRelevance rel) (envContext e)
-      , envLetBindings = Map.map
-          (fmap $ \ (t, a) -> (t, inverseApplyRelevance rel a))
-          (envLetBindings e)
+      { envContext     = modifyContextEntries      (inverseApplyRelevance rel) (envContext e)
+      , envLetBindings = (Map.map . fmap . second) (inverseApplyRelevance rel) (envLetBindings e)
                                                   -- enable local  irr. defs
-      , envRelevance = composeRelevance rel (envRelevance e)
+      , envRelevance   = composeRelevance rel (envRelevance e)
                                                   -- enable global irr. defs
       }
 
--- | Wake up irrelevant variables and make them relevant.  For instance,
---   in an irrelevant function argument otherwise irrelevant variables
---   may be used, so they are awoken before type checking the argument.
+-- | Wake up irrelevant variables and make them relevant. This is used
+--   when type checking terms in a hole, in which case you want to be able to
+--   (for instance) infer the type of an irrelevant variable. In the course
+--   of type checking an irrelevant function argument 'applyRelevanceToContext'
+--   is used instead, which also sets the context relevance to 'Irrelevant'.
+--   This is not the right thing to do when type checking interactively in a
+--   hole since it also marks all metas created during type checking as
+--   irrelevant (issue #2568).
 wakeIrrelevantVars :: TCM a -> TCM a
-wakeIrrelevantVars = applyRelevanceToContext Irrelevant
+wakeIrrelevantVars = local $ \ e -> e
+  { envContext     = modifyContextEntries      (inverseApplyRelevance Irrelevant) (envContext e)
+  , envLetBindings = (Map.map . fmap . second) (inverseApplyRelevance Irrelevant) (envLetBindings e)
+  }
+
