@@ -8,7 +8,6 @@ import Prelude hiding (null)
 import Control.Monad.Reader
 
 import Data.Function
-import Data.List hiding (sort, null)
 import qualified Data.List as List
 import qualified Data.Map as Map
 import qualified Data.Foldable as Fold
@@ -40,9 +39,6 @@ import Agda.TypeChecking.MetaVars.Occurs
 
 import Agda.Utils.Except
   ( ExceptT
-#if !MIN_VERSION_transformers(0,4,1)
-  , Error(noMsg)
-#endif
   , MonadError(throwError)
   , runExceptT
   )
@@ -67,7 +63,7 @@ import Agda.Utils.Impossible
 --   @reverse@ is necessary because we are directly abstracting over the list.
 --
 findIdx :: Eq a => [a] -> a -> Maybe Int
-findIdx vs v = findIndex (==v) (reverse vs)
+findIdx vs v = List.findIndex (==v) (reverse vs)
 
 -- | Check whether a meta variable is a place holder for a blocked term.
 isBlockedTerm :: MetaId -> TCM Bool
@@ -128,19 +124,13 @@ assignTerm' x tel v = do
     -- dontAssignMetas $ do
     --   checkInternal t . jMetaType . mvJudgement =<< lookupMeta x
 
-    let i = metaInstance tel $ killRange v
-    verboseS "profile.metas" 10 $ liftTCM $ tickMax "max-open-metas" . size =<< getOpenMetas
-    modifyMetaStore $ ins x i
+    verboseS "profile.metas" 10 $ liftTCM $ tickMax "max-open-metas" . (fromIntegral . size) =<< getOpenMetas
+    modifyMetaStore $ ins x $ InstV tel $ killRange v
     etaExpandListeners x
     wakeupConstraints x
     reportSLn "tc.meta.assign" 20 $ "completed assignment of " ++ prettyShow x
   where
-    metaInstance tel v = InstV tel v
-    ins x i store = Map.adjust (inst i) x store
-    inst i mv = mv { mvInstantiation = i }
-
-    -- Andreas, 2013-10-25 hack to fool the unused-imports-checking-Nazi
-    -- phantomUseToOverruleStrictImportsChecking = checkInternal
+    ins x i = Map.adjust (\ mv -> mv { mvInstantiation = i }) x
 
 -- * Creating meta variables.
 
@@ -190,15 +180,11 @@ newTypeMeta_  = newTypeMeta =<< (workOnTypes $ newSortMeta)
 
 -- | @newIFSMeta s t cands@ creates a new "implicit from scope" metavariable
 --   of type the output type of @t@ with name suggestion @s@.
---   If @t@ is a function type, then insert enough
---   lambdas in front of it.
 newIFSMeta :: MetaNameSuggestion -> Type -> TCM (MetaId, Term)
 newIFSMeta s t = do
-  TelV tel t' <- telView t
-  addContext tel $ do
-    vs  <- getContextArgs
-    ctx <- getContextTelescope
-    mapSnd (teleLam tel) <$> newIFSMetaCtx s (telePi_ ctx t') vs
+  vs  <- getContextArgs
+  ctx <- getContextTelescope
+  newIFSMetaCtx s (telePi_ ctx t) vs
 
 newIFSMetaCtx :: MetaNameSuggestion -> Type -> Args -> TCM (MetaId, Term)
 newIFSMetaCtx s t vs = do
@@ -214,7 +200,11 @@ newIFSMetaCtx s t vs = do
   reportSDoc "tc.meta.new" 50 $ fsep
     [ nest 2 $ pretty x <+> text ":" <+> prettyTCM t
     ]
-  addConstraint $ FindInScope x Nothing Nothing
+  let c = FindInScope x Nothing Nothing
+  -- If we're not already solving instance constraints we should add this
+  -- to the awake constraints to make sure we don't forget about it. If we
+  -- are solving constraints it will get woken up later (see #2690)
+  ifM isSolvingConstraints (addConstraint c) (addAwakeConstraint' c)
   etaExpandMetaSafe x
   return (x, MetaV x $ map Apply vs)
 
@@ -319,7 +309,7 @@ newRecordMetaCtx r pars tel perm ctx = do
   return $ Con con ConOSystem fields
 
 newQuestionMark :: InteractionId -> Type -> TCM (MetaId, Term)
-newQuestionMark = newQuestionMark' $ newValueMeta' DontRunMetaOccursCheck
+newQuestionMark = newQuestionMark' $ newValueMeta' RunMetaOccursCheck
 
 newQuestionMark' :: (Type -> TCM (MetaId, Term)) -> InteractionId -> Type -> TCM (MetaId, Term)
 newQuestionMark' new ii t = do
@@ -469,7 +459,7 @@ allMetaKinds = [minBound .. maxBound]
 -- | Eta expand a metavariable, if it is of the specified kind.
 --   Don't do anything if the metavariable is a blocked term.
 etaExpandMeta :: [MetaKind] -> MetaId -> TCM ()
-etaExpandMeta kinds m = whenM (isEtaExpandable kinds m) $ do
+etaExpandMeta kinds m = whenM (asks envAssignMetas `and2M` isEtaExpandable kinds m) $ do
   verboseBracket "tc.meta.eta" 20 ("etaExpandMeta " ++ prettyShow m) $ do
     let waitFor x = do
           reportSDoc "tc.meta.eta" 20 $ do
@@ -595,7 +585,11 @@ assign dir x args v = do
   -- arguments to definitions as flexible), if that fails it tries again
   -- with full unfolding.
   v <- instantiate v
-  reportSLn "tc.meta.assign" 50 $ "MetaVars.assign: assigning to " ++ show v
+  reportSDoc "tc.meta.assign" 45 $
+    text "MetaVars.assign: assigning to " <+> prettyTCM v
+
+  reportSLn "tc.meta.assign" 75 $
+    "MetaVars.assign: assigning to " ++ show v
 
   case (ignoreSharing v, mvJudgement mvar) of
       (Sort Inf, HasType{}) -> typeError SetOmegaNotValidType
@@ -629,8 +623,22 @@ assign dir x args v = do
     -- args <- etaContract =<< normalise args
 
     -- Also, try to expand away projected vars in meta args.
+    reportSDoc "tc.meta.assign.proj" 45 $ do
+      cxt <- getContextTelescope
+      vcat
+        [ text "context before projection expansion"
+        , nest 2 $ inTopContext $ prettyTCM cxt
+        ]
+
     expandProjectedVars args v $ \ args v -> do
 
+      reportSDoc "tc.meta.assign.proj" 45 $ do
+        cxt <- getContextTelescope
+        vcat
+          [ text "context after projection expansion"
+          , nest 2 $ inTopContext $ prettyTCM cxt
+          ]
+
       -- If we had the type here we could save the work we put
       -- into expanding projected variables.
       -- catchConstraint (ValueCmp CmpEq ? (MetaV m $ map Apply args) v) $ do
@@ -641,16 +649,12 @@ assign dir x args v = do
 
       (relVL, irrVL) <- do
         -- Andreas, 2016-11-03 #2211 attempt to do s.th. for unused
-        if False -- irrelevantOrUnused $ getMetaRelevance mvar
+        if False -- irrelevant $ getMetaRelevance mvar
           then do
             reportSDoc "tc.meta.assign" 25 $ text "meta is irrelevant or unused"
             return (Set.toList $ allFreeVars args, empty)
           else do
-            -- Andreas, 2016-11-03, issue #2211
-            -- treating UnusedArg as Irrelevant bears trouble
-            -- since the UnusedArg info is not consistently present
-            -- Thus, make sure we include the "unused" variables.
-            let relVL = Set.toList $ allRelevantOrUnusedVars args
+            let relVL = Set.toList $ allRelevantVars args
             -- Andreas, 2011-10-06 only irrelevant vars that are direct
             -- arguments to the meta, hence, can be abstracted over, may
             -- appear on the rhs.  (test/fail/Issue483b)
@@ -663,7 +667,7 @@ assign dir x args v = do
                 fromIrrVar _ = return []
             irrVL <- concat <$> mapM fromIrrVar
                        [ v | Arg info v <- args, isIrrelevant info ]
-                          -- irrelevantOrUnused (getRelevance info) ]
+                          -- irrelevant (getRelevance info) ]
             return (relVL, irrVL)
       reportSDoc "tc.meta.assign" 20 $
           let pr (Var n []) = text (show n)
@@ -907,8 +911,22 @@ assignMeta' m x t n ids v = do
        patternViolation -- WAS: __IMPOSSIBLE__
 
     -- Perform the assignment (and wake constraints).
+
+    let vsol = abstract tel' v'
+    -- -- Andreas, 2013-10-25 double check solution before assigning
+    -- -- Andreas, 2017-07-28
+    -- m <- lookupMeta x
+    -- case mvJudgement m of
+    --   IsSort{}    -> return ()  -- skip double check since type of meta is not accurate
+    --   HasType _ a -> do
+    --     reportSDoc "tc.meta.check" 30 $ vcat
+    --       [ text "double checking solution"
+    --       , nest 2 $ prettyTCM vsol <+> text " : " <+> prettyTCM a
+    --       ]
+    --     dontAssignMetas $ checkInternal vsol a  -- This can crash at assignTerm'!
+
     reportSDoc "tc.meta.assign" 10 $
-      text "solving" <+> prettyTCM x <+> text ":=" <+> prettyTCM (abstract tel' v')
+      text "solving" <+> prettyTCM x <+> text ":=" <+> prettyTCM vsol
     assignTerm x (telToArgs tel') v'
 
 
@@ -991,16 +1009,12 @@ class NoProjectedVar a where
 
 data ProjVarExc = ProjVarExc Int [(ProjOrigin, QName)]
 
--- ASR (17 June 2015). Unused Error instance.
--- instance Error ProjVarExc where
---   noMsg = __IMPOSSIBLE__
-
 instance NoProjectedVar Term where
   noProjectedVar t =
     case ignoreSharing t of
       Var i es
         | qs@(_:_) <- takeWhileJust id $ map isProjElim es -> Left $ ProjVarExc i qs
-      -- Andreas, 2015-09-12 Issue 1316:
+      -- Andreas, 2015-09-12 Issue #1316:
       -- Also look in inductive record constructors
       Con (ConHead _ Inductive (_:_)) _ vs -> noProjectedVar vs
       _ -> return ()
@@ -1071,7 +1085,7 @@ type SubstCand = [(Int,Term)] -- ^ a possibly non-deterministic substitution
 --   Otherwise, raise the error.
 checkLinearity :: SubstCand -> ExceptT () TCM SubstCand
 checkLinearity ids0 = do
-  let ids = sortBy (compare `on` fst) ids0  -- see issue 920
+  let ids = List.sortBy (compare `on` fst) ids0  -- see issue 920
   let grps = groupOn fst ids
   concat <$> mapM makeLinear grps
   where
@@ -1094,11 +1108,6 @@ data InvertExcept
   | NeutralArg                -- ^ A potentially neutral arg: can't invert, but can try pruning.
   | ProjectedVar Int [(ProjOrigin, QName)]  -- ^ Try to eta-expand var to remove projs.
 
-#if !MIN_VERSION_transformers(0,4,1)
-instance Error InvertExcept where
-  noMsg = CantInvert
-#endif
-
 -- | Check that arguments @args@ to a metavar are in pattern fragment.
 --   Assumes all arguments already in whnf and eta-reduced.
 --   Parameters are represented as @Var@s so @checkArgs@ really
@@ -1137,10 +1146,6 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
         Arg info (Con c ci vs) -> do
           let fallback
                | isIrrelevant info = return vars
-                 -- Andreas, 2016-11-03, issue #2211
-                 -- treating UnusedArg as Irrelevant bears trouble
-                 -- since the UnusedArg info is not consistently present
-                 -- irrelevantOrUnused (getRelevance info) = return vars
                | otherwise                              = failure
           isRC <- lift $ isRecordConstructor $ conName c
           case isRC of
@@ -1151,7 +1156,6 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
                        { argInfoHiding       = min (getHiding info) (getHiding info')
                        , argInfoRelevance    = max (getRelevance info) (getRelevance info')
                        , argInfoOrigin       = min (getOrigin info) (getOrigin info')
-                       , argInfoOverlappable = False
                        }
                 res <- loop $ zipWith aux vs fs
                 return $ res `append` vars
@@ -1161,10 +1165,6 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
 
         -- An irrelevant argument which is not an irrefutable pattern is dropped
         Arg info _ | isIrrelevant info -> return vars
-          -- Andreas, 2016-11-03, issue #2211
-          -- treating UnusedArg as Irrelevant bears trouble
-          -- since the UnusedArg info is not consistently present
-          -- irrelevantOrUnused (getRelevance info) -> return vars
         -- Andreas, 2013-10-29
         -- An irrelevant part can also be marked by a DontCare
         -- (coming from an irrelevant projection), see Issue 927:
@@ -1189,7 +1189,7 @@ inverseSubst args = map (mapFst unArg) <$> loop (zip args terms)
 
     -- adding an irrelevant entry only if not present
     cons :: (Arg Nat, Term) -> Res -> Res
-    cons a@(Arg (ArgInfo _ Irrelevant _ _) i, t) vars    -- TODO? UnusedArg?!
+    cons a@(Arg (ArgInfo _ Irrelevant _) i, t) vars
       | any ((i==) . unArg . fst) vars  = vars
       | otherwise                       = a : vars
     -- adding a relevant entry:
