@@ -256,7 +256,11 @@ instance Occurs Term where
             if (i `allowedVar` xs) then Var i <$> occ (weakly ctx) es else do
               -- if the offending variable is of singleton type,
               -- eta-expand it away
-              isST <- isSingletonType =<< typeOfBV i
+              reportSDoc "tc.meta.occurs" 35 $ text "offending variable: " <+> prettyTCM (var i)
+              t <-  typeOfBV i
+              reportSDoc "tc.meta.occurs" 35 $ nest 2 $ text "of type " <+> prettyTCM t
+              isST <- isSingletonType t
+              reportSDoc "tc.meta.occurs" 35 $ nest 2 $ text "(after singleton test)"
               case isST of
                 -- cannot decide, blocked by meta-var
                 Left mid -> patternViolation' 70 $ "Disallowed var " ++ show i ++ " not obviously singleton"
@@ -486,7 +490,6 @@ prune m' vs xs = do
       [ text "attempting kills"
       , nest 2 $ vcat
         [ text "m'    =" <+> pretty m'
-        -- , text "xs    =" <+> text (show xs)
         , text "xs    =" <+> prettyList (map (prettyTCM . var) xs)
         , text "vs    =" <+> prettyList (map prettyTCM vs)
         , text "kills =" <+> text (show kills)
@@ -546,7 +549,7 @@ isNeutral b f es = liftTCM $ do
   def <- getConstInfo f
   if defMatchable def then no else do
   case theDef def of
-    AbstractDefn -> yes
+    AbstractDefn{} -> yes
     Axiom{}    -> yes
     Datatype{} -> yes
     Record{}   -> yes
@@ -643,7 +646,6 @@ instance FoldRigid a => FoldRigid (Arg a) where
   foldRigid f a =
     case getRelevance a of
       Irrelevant -> mempty
-      UnusedArg  -> mempty
       _          -> foldRigid f $ unArg a
 
 instance FoldRigid a => FoldRigid (Dom a) where
