@@ -13,6 +13,7 @@ import Agda.Syntax.Position
 import Agda.Syntax.Literal
 import Agda.Syntax.Internal as I
 import Agda.TypeChecking.Monad.Base
+-- import Agda.TypeChecking.Functions  -- LEADS TO IMPORT CYCLE
 import Agda.TypeChecking.Substitute
 
 import Agda.Utils.Except ( MonadError(catchError) )
@@ -34,7 +35,7 @@ litType :: Literal -> TCM Type
 litType l = case l of
   LitNat _ n    -> do
     _ <- primZero
-    when_ (n > 0) $ primSuc
+    when (n > 0) $ void $ primSuc
     el <$> primNat
   LitFloat _ _  -> el <$> primFloat
   LitChar _ _   -> el <$> primChar
@@ -140,7 +141,7 @@ primInteger, primIntegerPos, primIntegerNegSuc,
     primAgdaTCMGetType, primAgdaTCMGetDefinition,
     primAgdaTCMQuoteTerm, primAgdaTCMUnquoteTerm,
     primAgdaTCMBlockOnMeta, primAgdaTCMCommit, primAgdaTCMIsMacro,
-    primAgdaTCMWithNormalisation
+    primAgdaTCMWithNormalisation, primAgdaTCMDebugPrint
     :: TCM Term
 
 primInteger      = getBuiltin builtinInteger
@@ -278,6 +279,7 @@ primAgdaTCMBlockOnMeta        = getBuiltin builtinAgdaTCMBlockOnMeta
 primAgdaTCMCommit             = getBuiltin builtinAgdaTCMCommit
 primAgdaTCMIsMacro            = getBuiltin builtinAgdaTCMIsMacro
 primAgdaTCMWithNormalisation  = getBuiltin builtinAgdaTCMWithNormalisation
+primAgdaTCMDebugPrint         = getBuiltin builtinAgdaTCMDebugPrint
 
 builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinNatTimes, builtinNatDivSucAux, builtinNatModSucAux, builtinNatEquals,
@@ -323,7 +325,7 @@ builtinNat, builtinSuc, builtinZero, builtinNatPlus, builtinNatMinus,
   builtinAgdaTCMGetType, builtinAgdaTCMGetDefinition,
   builtinAgdaTCMQuoteTerm, builtinAgdaTCMUnquoteTerm,
   builtinAgdaTCMBlockOnMeta, builtinAgdaTCMCommit, builtinAgdaTCMIsMacro,
-  builtinAgdaTCMWithNormalisation
+  builtinAgdaTCMWithNormalisation, builtinAgdaTCMDebugPrint
   :: String
 
 builtinNat                           = "NATURAL"
@@ -461,6 +463,7 @@ builtinAgdaTCMQuoteTerm     = "AGDATCMQUOTETERM"
 builtinAgdaTCMUnquoteTerm   = "AGDATCMUNQUOTETERM"
 builtinAgdaTCMIsMacro       = "AGDATCMISMACRO"
 builtinAgdaTCMWithNormalisation = "AGDATCMWITHNORMALISATION"
+builtinAgdaTCMDebugPrint = "AGDATCMDEBUGPRINT"
 
 -- | Builtins that come without a definition in Agda syntax.
 --   These are giving names to Agda internal concepts which
@@ -513,6 +516,7 @@ coinductionKit = tryMaybe coinductionKit'
 
 -- | Get the name of the equality type.
 primEqualityName :: TCM QName
+-- primEqualityName = getDef =<< primEquality  -- LEADS TO IMPORT CYCLE
 primEqualityName = do
   eq <- primEquality
   -- Andreas, 2014-05-17 moved this here from TC.Rules.Def
@@ -527,9 +531,11 @@ primEqualityName = do
       lamV (Shared p) = lamV (derefPtr p)
       lamV v          = ([], v)
   return $ case lamV eq of
-    ([Hidden, Hidden], Def equality _) -> equality
-    ([Hidden],         Def equality _) -> equality
-    ([],               Def equality _) -> equality
+    (_, Def equality _) -> equality
+  -- OLD:
+  --   ([Hidden, Hidden], Def equality _) -> equality
+  --   ([Hidden],         Def equality _) -> equality
+  --   ([],               Def equality _) -> equality
     _                                  -> __IMPOSSIBLE__
 
 -- | Check whether the type is actually an equality (lhs ≡ rhs)
@@ -541,8 +547,12 @@ equalityView :: Type -> TCM EqualityView
 equalityView t0@(El s t) = do
   equality <- primEqualityName
   case ignoreSharing t of
-    Def equality' [ Apply level , Apply typ , Apply lhs , Apply rhs ]
-      | equality' == equality -> return $ EqualityType s equality level typ lhs rhs
+    Def equality' es | equality' == equality -> do
+      let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+      let n = length vs
+      unless (n >= 3) __IMPOSSIBLE__
+      let (pars, [ typ , lhs, rhs ]) = splitAt (n-3) vs
+      return $ EqualityType s equality pars typ lhs rhs
     _ -> return $ OtherType t0
 
 -- | Revert the 'EqualityView'.
@@ -552,4 +562,4 @@ equalityView t0@(El s t) = do
 equalityUnview :: EqualityView -> Type
 equalityUnview (OtherType t) = t
 equalityUnview (EqualityType s equality l t lhs rhs) =
-  El s $ Def equality $ map Apply [l, t, lhs, rhs]
+  El s $ Def equality $ map Apply (l ++ [t, lhs, rhs])
