@@ -25,7 +25,7 @@ import Agda.Syntax.Scope.Base
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.Trace
 import Agda.TypeChecking.Monad.Closure
-import Agda.TypeChecking.Monad.Options (reportSLn)
+import Agda.TypeChecking.Monad.Debug (reportSLn)
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Substitute
 import {-# SOURCE #-} Agda.TypeChecking.Telescope
@@ -47,7 +47,9 @@ import Agda.Utils.Impossible
 
 -- | Switch off assignment of metas.
 dontAssignMetas :: TCM a -> TCM a
-dontAssignMetas = local $ \ env -> env { envAssignMetas = False }
+dontAssignMetas cont = do
+  reportSLn "tc.meta" 45 $ "don't assign metas"
+  local (\ env -> env { envAssignMetas = False }) cont
 
 -- | Get the meta store.
 getMetaStore :: TCM MetaStore
@@ -209,6 +211,10 @@ setMetaNameSuggestion mi s = unless (null s || isUnderscore s) $ do
 updateMetaVarRange :: MetaId -> Range -> TCM ()
 updateMetaVarRange mi r = updateMetaVar mi (setRange r)
 
+setMetaOccursCheck :: MetaId -> RunMetaOccursCheck -> TCM ()
+setMetaOccursCheck mi b = updateMetaVar mi $ \ mvar ->
+  mvar { mvInfo = (mvInfo mvar) { miMetaOccursCheck = b } }
+
 -- * Query and manipulate interaction points.
 
 modifyInteractionPoints :: (InteractionPoints -> InteractionPoints) -> TCM ()
@@ -232,7 +238,7 @@ registerInteractionPoint preciseRange r maybeId = do
   ii <- case maybeId of
     Just i  -> return $ InteractionId i
     Nothing -> fresh
-  let ip = InteractionPoint { ipRange = r, ipMeta = Nothing, ipClause = IPNoClause }
+  let ip = InteractionPoint { ipRange = r, ipMeta = Nothing, ipSolved = False, ipClause = IPNoClause }
   case Map.insertLookupWithKey (\ key new old -> old) ii ip m of
     -- If the interaction point is already present, we keep the old ip.
     -- However, it needs to be at the same range as the new one.
@@ -253,7 +259,7 @@ findInteractionPoint_ r m = do
   headMaybe $ mapMaybe sameRange $ Map.toList m
   where
     sameRange :: (InteractionId, InteractionPoint) -> Maybe InteractionId
-    sameRange (ii, InteractionPoint r' _ _) | r == r' = Just ii
+    sameRange (ii, InteractionPoint r' _ _ _) | r == r' = Just ii
     sameRange _ = Nothing
 
 -- | Hook up meta variable to interaction point.
@@ -261,32 +267,30 @@ connectInteractionPoint :: InteractionId -> MetaId -> TCM ()
 connectInteractionPoint ii mi = do
   ipCl <- asks envClause
   m <- use stInteractionPoints
-  let ip = InteractionPoint { ipRange = __IMPOSSIBLE__, ipMeta = Just mi, ipClause = ipCl }
+  let ip = InteractionPoint { ipRange = __IMPOSSIBLE__, ipMeta = Just mi, ipSolved = False, ipClause = ipCl }
   -- The interaction point needs to be present already, we just set the meta.
   case Map.insertLookupWithKey (\ key new old -> new { ipRange = ipRange old }) ii ip m of
     (Nothing, _) -> __IMPOSSIBLE__
     (Just _, m') -> modifyInteractionPoints $ const m'
 
--- | Move an interaction point from the current ones to the old ones.
+-- | Mark an interaction point as solved.
 removeInteractionPoint :: InteractionId -> TCM ()
 removeInteractionPoint ii = do
-  ip <- stInteractionPoints %%= \ m -> do
-    let (mip, m') = Map.updateLookupWithKey (\ _ _ -> Nothing) ii m
-    return (m',
-      fromMaybe __IMPOSSIBLE__ mip)
-  stSolvedInteractionPoints %= Map.insert ii ip
+  stInteractionPoints %= Map.update (\ ip -> Just ip{ ipSolved = True }) ii
 
 -- | Get a list of interaction ids.
 getInteractionPoints :: TCM [InteractionId]
 getInteractionPoints = Map.keys <$> use stInteractionPoints
 
--- | Get all metas that correspond to interaction ids.
+-- | Get all metas that correspond to unsolved interaction ids.
 getInteractionMetas :: TCM [MetaId]
-getInteractionMetas = mapMaybe ipMeta . Map.elems <$> use stInteractionPoints
+getInteractionMetas =
+  mapMaybe ipMeta . filter (not . ipSolved) . Map.elems <$> use stInteractionPoints
 
--- | Get all metas that correspond to interaction ids.
+-- | Get all metas that correspond to unsolved interaction ids.
 getInteractionIdsAndMetas :: TCM [(InteractionId,MetaId)]
-getInteractionIdsAndMetas = mapMaybe f . Map.toList <$> use stInteractionPoints
+getInteractionIdsAndMetas =
+  mapMaybe f . filter (not . ipSolved . snd) . Map.toList <$> use stInteractionPoints
   where f (ii, ip) = (ii,) <$> ipMeta ip
 
 -- | Does the meta variable correspond to an interaction point?
