@@ -11,14 +11,15 @@ import Control.Monad.State  hiding (mapM)
 import Data.Maybe
 import Data.Traversable
 
-import Text.PrettyPrint
 import System.Directory
 import System.FilePath
 
 import Agda.Syntax.Internal
 import Agda.Syntax.Common
 import Agda.Syntax.Concrete
+import {-# SOURCE #-} Agda.TypeChecking.Monad.Debug
 import {-# SOURCE #-} Agda.TypeChecking.Errors
+import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Monad.Base
 import Agda.TypeChecking.Monad.State
 import Agda.TypeChecking.Monad.Benchmark
@@ -34,6 +35,7 @@ import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Lens
 import Agda.Utils.List
+import Agda.Utils.Pretty
 import Agda.Utils.Trie (Trie)
 import qualified Agda.Utils.Trie as Trie
 import Agda.Utils.Except
@@ -46,9 +48,10 @@ import Agda.Utils.Impossible
 
 setPragmaOptions :: PragmaOptions -> TCM ()
 setPragmaOptions opts = do
+  stPragmaOptions %= Lens.mapSafeMode (Lens.getSafeMode opts ||)
   clo <- commandLineOptions
   let unsafe = unsafePragmaOptions opts
-  when (optSafe clo && not (null unsafe)) $ typeError (SafeFlagPragma unsafe)
+  when (Lens.getSafeMode clo && not (null unsafe)) $ warning $ SafeFlagPragma unsafe
   ok <- liftIO $ runOptM $ checkOpts (clo { optPragmaOptions = opts })
   case ok of
     Left err   -> __IMPOSSIBLE__
@@ -113,20 +116,6 @@ addDefaultLibraries rel o
   (libs, incs) <- libToTCM $ getDefaultLibraries (filePath root) (optDefaultLibs o)
   return o{ optIncludePaths = incs ++ optIncludePaths o, optLibraries = libs }
 
-class (Functor m, Applicative m, Monad m) => HasOptions m where
-  -- | Returns the pragma options which are currently in effect.
-  pragmaOptions      :: m PragmaOptions
-  -- | Returns the command line options which are currently in effect.
-  commandLineOptions :: m CommandLineOptions
-
-instance MonadIO m => HasOptions (TCMT m) where
-  pragmaOptions = use stPragmaOptions
-
-  commandLineOptions = do
-    p  <- use stPragmaOptions
-    cl <- stPersistentOptions . stPersistentState <$> get
-    return $ cl { optPragmaOptions = p }
-
 setOptionsFromPragma :: OptionsPragma -> TCM ()
 setOptionsFromPragma ps = do
     opts <- commandLineOptions
@@ -174,7 +163,7 @@ data RelativeTo
 getProjectRoot :: RelativeTo -> TCM AbsolutePath
 getProjectRoot CurrentDir = liftIO (absolute =<< getCurrentDirectory)
 getProjectRoot (ProjectRoot f) = do
-  Ranged _ m <- moduleName' f
+  m <- moduleName' f
   return (projectRoot f m)
 
 -- | Makes the given directories absolute and stores them as include
@@ -209,12 +198,26 @@ setIncludeDirs incs relativeTo = do
       -- Might also be useful to overwrite default imports...
   incs <- return $ incs ++ [primdir]
 
+  reportSDoc "setIncludeDirs" 10 $ return $ vcat
+    [ text "Old include directories:"
+    , nest 2 $ vcat $ map pretty oldIncs
+    , text "New include directories:"
+    , nest 2 $ vcat $ map pretty incs
+    ]
+
   -- Check whether the include dirs have changed.  If yes, reset state.
   -- Andreas, 2013-10-30 comments:
   -- The logic, namely using the include-dirs variable as a driver
   -- for the interaction, qualifies for a code-obfuscation contest.
   -- I guess one Boolean more in the state cost 10.000 EUR at the time
   -- of this implementation...
+  --
+  -- Andreas, perhaps you have misunderstood something: If the include
+  -- directories have changed and we do not reset the decoded modules,
+  -- then we might (depending on how the rest of the code works) end
+  -- up in a situation in which we use the contents of the file
+  -- "old-path/M.agda", when the user actually meant
+  -- "new-path/M.agda".
   when (oldIncs /= incs) $ do
     ho <- getInteractionOutputCallback
     resetAllState
@@ -236,7 +239,7 @@ setIncludeDirs incs relativeTo = do
     ProjectRoot f -> void $ moduleName f
      -- Andreas, 2016-07-12 WAS:
      -- do
-     --  Ranged _ m <- moduleName' f
+     --  m <- moduleName' f
      --  checkModuleName m f Nothing
 
 
@@ -291,31 +294,28 @@ showIrrelevantArguments = optShowIrrelevant <$> pragmaOptions
 
 -- | Switch on printing of implicit and irrelevant arguments.
 --   E.g. for reification in with-function generation.
+--
+--   Restores all 'PragmaOptions' after completion.
+--   Thus, do not attempt to make persistent 'PragmaOptions'
+--   changes in a `withShowAllArguments` bracket.
+
 withShowAllArguments :: TCM a -> TCM a
 withShowAllArguments = withShowAllArguments' True
 
 withShowAllArguments' :: Bool -> TCM a -> TCM a
-withShowAllArguments' yes ret = do
+withShowAllArguments' yes = withPragmaOptions $ \ opts ->
+  opts { optShowImplicit = yes, optShowIrrelevant = yes }
+
+-- | Change 'PragmaOptions' for a computation and restore afterwards.
+
+withPragmaOptions :: (PragmaOptions -> PragmaOptions) -> TCM a -> TCM a
+withPragmaOptions f cont = do
   opts <- pragmaOptions
-  let imp = optShowImplicit opts
-      irr = optShowIrrelevant opts
-  setPragmaOptions $ opts { optShowImplicit = yes, optShowIrrelevant = yes }
-  x <- ret
-  opts <- pragmaOptions
-  setPragmaOptions $ opts { optShowImplicit = imp, optShowIrrelevant = irr }
+  setPragmaOptions $ f opts
+  x <- cont
+  setPragmaOptions opts
   return x
 
-{- RETIRED, Andreas, 2012-04-30
-setShowImplicitArguments :: Bool -> TCM a -> TCM a
-setShowImplicitArguments showImp ret = do
-  opts <- pragmaOptions
-  let imp = optShowImplicit opts
-  setPragmaOptions $ opts { optShowImplicit = showImp }
-  x <- ret
-  opts <- pragmaOptions
-  setPragmaOptions $ opts { optShowImplicit = imp }
-  return x
--}
 
 ignoreInterfaces :: TCM Bool
 ignoreInterfaces = optIgnoreInterfaces <$> commandLineOptions
@@ -362,48 +362,24 @@ hasVerbosity k n | n < 0     = __IMPOSSIBLE__
         m  = last $ 0 : Trie.lookupPath ks t
     return (n <= m)
 
--- | Displays a debug message in a suitable way.
-{-# SPECIALIZE displayDebugMessage :: Int -> String -> TCM () #-}
-displayDebugMessage :: MonadTCM tcm
-  => Int     -- ^ The message's debug level.
-  -> String  -- ^ Message.
-  -> tcm ()
-displayDebugMessage n s = liftTCM $
-  appInteractionOutputCallback (Resp_RunningInfo n s)
+-- | Check whether a certain verbosity level is activated (exact match).
+
+{-# SPECIALIZE hasExactVerbosity :: VerboseKey -> Int -> TCM Bool #-}
+hasExactVerbosity :: HasOptions m => VerboseKey -> Int -> m Bool
+hasExactVerbosity k n =
+  (Just n ==) . Trie.lookup (wordsBy (`elem` ".:") k) <$> getVerbosity
+
+-- | Run a computation if a certain verbosity level is activated (exact match).
+
+{-# SPECIALIZE whenExactVerbosity :: VerboseKey -> Int -> TCM () -> TCM () #-}
+whenExactVerbosity :: MonadTCM tcm => VerboseKey -> Int -> tcm () -> tcm ()
+whenExactVerbosity k n = whenM $ liftTCM $ hasExactVerbosity k n
 
 -- | Run a computation if a certain verbosity level is activated.
 --
 --   Precondition: The level must be non-negative.
 {-# SPECIALIZE verboseS :: VerboseKey -> Int -> TCM () -> TCM () #-}
-verboseS :: MonadTCM tcm => VerboseKey -> Int -> tcm () -> tcm ()
-verboseS k n action = whenM (liftTCM $ hasVerbosity k n) action
-
--- | Conditionally print debug string.
-{-# SPECIALIZE reportS :: VerboseKey -> Int -> String -> TCM () #-}
-reportS :: MonadTCM tcm => VerboseKey -> Int -> String -> tcm ()
-reportS k n s = liftTCM $ verboseS k n $ displayDebugMessage n s
-
--- | Conditionally println debug string.
-{-# SPECIALIZE reportSLn :: VerboseKey -> Int -> String -> TCM () #-}
-reportSLn :: MonadTCM tcm => VerboseKey -> Int -> String -> tcm ()
-reportSLn k n s = verboseS k n $
-  displayDebugMessage n (s ++ "\n")
-
--- | Conditionally render debug 'Doc' and print it.
-{-# SPECIALIZE reportSDoc :: VerboseKey -> Int -> TCM Doc -> TCM () #-}
-reportSDoc :: MonadTCM tcm => VerboseKey -> Int -> TCM Doc -> tcm ()
-reportSDoc k n d = liftTCM $ verboseS k n $ do
-  displayDebugMessage n . (++ "\n") . show =<< do
-    d `catchError` \ err ->
-      (\ s -> (sep $ map text
-                 [ "Printing debug message"
-                 , k  ++ ":" ++show n
-                 , "failed due to error:" ]) $$
-              (nest 2 $ text s)) <$> prettyError err
-
--- | Print brackets around debug messages issued by a computation.
-{-# SPECIALIZE verboseBracket :: VerboseKey -> Int -> String -> TCM a -> TCM a #-}
-verboseBracket :: MonadTCM tcm => VerboseKey -> Int -> String -> TCM a -> tcm a
-verboseBracket k n s m = liftTCM $ ifNotM (hasVerbosity k n) m $ {- else -} do
-  displayDebugMessage n $ "{ " ++ s ++ "\n"
-  m `finally` displayDebugMessage n "}\n"
+-- {-# SPECIALIZE verboseS :: MonadIO m => VerboseKey -> Int -> TCMT m () -> TCMT m () #-} -- RULE left-hand side too complicated to desugar
+{-# SPECIALIZE verboseS :: MonadTCM tcm => VerboseKey -> Int -> tcm () -> tcm () #-}
+verboseS :: HasOptions m => VerboseKey -> Int -> m () -> m ()
+verboseS k n action = whenM (hasVerbosity k n) action
