@@ -11,7 +11,7 @@ import Control.Monad.State
 import Control.Monad.Reader
 import Control.Monad.Trans.Maybe
 
-import Data.List hiding (null)
+import qualified Data.List as List
 import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Map (Map)
@@ -27,17 +27,17 @@ import Agda.Syntax.Internal.Names
 import Agda.Syntax.Position
 import Agda.Syntax.Treeless (Compiled(..), TTerm)
 
-import qualified Agda.Compiler.UHC.Pragmas.Base as CR
-
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Context
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Env
-import Agda.TypeChecking.Monad.Exception ( ExceptionT )
 import Agda.TypeChecking.Monad.Mutual
 import Agda.TypeChecking.Monad.Open
 import Agda.TypeChecking.Monad.Local
 import Agda.TypeChecking.Monad.State
+import Agda.TypeChecking.Monad.Trace
+import Agda.TypeChecking.Warnings
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Substitute
 import {-# SOURCE #-} Agda.TypeChecking.Telescope
@@ -45,7 +45,7 @@ import {-# SOURCE #-} Agda.TypeChecking.CompiledClause.Compile
 import {-# SOURCE #-} Agda.TypeChecking.Polarity
 import {-# SOURCE #-} Agda.TypeChecking.ProjectionLike
 
-import Agda.Utils.Except ( Error )
+import Agda.Utils.Except ( ExceptT )
 import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.List
@@ -64,11 +64,11 @@ import Agda.Utils.Impossible
 -- | Add a constant to the signature. Lifts the definition to top level.
 addConstant :: QName -> Definition -> TCM ()
 addConstant q d = do
-  reportSLn "tc.signature" 20 $ "adding constant " ++ show q ++ " to signature"
+  reportSLn "tc.signature" 20 $ "adding constant " ++ prettyShow q ++ " to signature"
   tel <- getContextTelescope
   let tel' = replaceEmptyName "r" $ killRange $ case theDef d of
               Constructor{} -> fmap hideOrKeepInstance tel
-              Function{ funProjection = Just Projection{ projProper = True, projIndex = n } } ->
+              Function{ funProjection = Just Projection{ projProper = Just{}, projIndex = n } } ->
                 let fallback = fmap hideOrKeepInstance tel in
                 if n > 0 then fallback else
                 -- if the record value is part of the telescope, its hiding should left unchanged
@@ -77,7 +77,7 @@ addConstant q d = do
                     Just (doms, dom) -> telFromList $ fmap hideOrKeepInstance doms ++ [dom]
               _ -> tel
   let d' = abstract tel' $ d { defName = q }
-  reportSLn "tc.signature" 30 $ "lambda-lifted definition = " ++ show d'
+  reportSDoc "tc.signature" 60 $ return $ text "lambda-lifted definition =" <?> pretty d'
   modifySignature $ updateDefinitions $ HMap.insertWith (+++) q d'
   i <- currentOrFreshMutualBlock
   setMutualBlock i q
@@ -87,10 +87,9 @@ addConstant q d = do
 
 -- | Set termination info of a defined function symbol.
 setTerminates :: QName -> Bool -> TCM ()
-setTerminates q b = modifySignature $ updateDefinition q $ updateTheDef $ setT
-  where
-    setT def@Function{} = def { funTerminates = Just b }
-    setT def            = def
+setTerminates q b = modifySignature $ updateDefinition q $ updateTheDef $ \case
+    def@Function{} -> def { funTerminates = Just b }
+    def -> def
 
 -- | Modify the clauses of a function.
 modifyFunClauses :: QName -> ([Clause] -> [Clause]) -> TCM ()
@@ -98,65 +97,73 @@ modifyFunClauses q f =
   modifySignature $ updateDefinition q $ updateTheDef $ updateFunClauses f
 
 -- | Lifts clauses to the top-level and adds them to definition.
+--   Also adjusts the 'funCopatternLHS' field if necessary.
 addClauses :: QName -> [Clause] -> TCM ()
 addClauses q cls = do
   tel <- getContextTelescope
-  modifyFunClauses q (++ abstract tel cls)
+  modifySignature $ updateDefinition q $ updateTheDef
+    $ updateFunClauses      (++ abstract tel cls)
+    . updateFunCopatternLHS (|| isCopatternLHS cls)
 
-ensureNoCompiledHaskell :: QName -> TCM ()
-ensureNoCompiledHaskell q =
-  whenM (isJust . compiledHaskell . defCompiledRep <$> getConstInfo q) $
-    typeError $ GenericError $ "Multiple Haskell bindings for " ++ show q ++ ". " ++
-                               "Note that builtin numbers, booleans, chars and strings don't need " ++
-                               "COMPILED pragmas."
+mkPragma :: String -> TCM CompilerPragma
+mkPragma s = CompilerPragma <$> getCurrentRange <*> pure s
 
-addHaskellCode :: QName -> HaskellType -> HaskellCode -> TCM ()
-addHaskellCode q hsTy hsDef = do
-  ensureNoCompiledHaskell q
-  modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
-  where
-    addHs crep = crep { compiledHaskell = Just $ HsDefn hsTy hsDef }
+-- | Add a compiler pragma `{-# COMPILE <backend> <name> <text> #-}`
+addPragma :: BackendName -> QName -> String -> TCM ()
+addPragma b q s = modifySignature . updateDefinition q . addCompilerPragma b =<< mkPragma s
 
-addHaskellExport :: QName -> HaskellType -> String -> TCM ()
-addHaskellExport q hsTy hsName = do
-  ensureNoCompiledHaskell q
-  modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
-  where
-    addHs crep = crep { exportHaskell = Just (HsExport hsTy hsName)}
+-- ** Temporary **
+--  The functions below are only needed while we still parse the old COMPILED
+--  pragmas.
+
+type HaskellCode = String
+type HaskellType = String
+type JSCode = String
+type CoreCode = String
+
+addDeprecatedPragma :: String -> BackendName -> QName -> String -> TCM ()
+addDeprecatedPragma old b q s = do
+  let pq = prettyShow $ nameConcrete $ qnameName q
+  warning $ DeprecationWarning (unwords ["The", old, "pragma"])
+                               (unwords ["{-# COMPILE", b, pq, s, "#-}"]) "2.6"
+  addPragma b q s
+
+dataFormat :: String -> [String] -> String
+dataFormat ty cons = "= data " ++ ty ++ " (" ++ List.intercalate " | " cons ++ ")"
+
+addHaskellCode :: QName -> HaskellCode -> TCM ()
+addHaskellCode q hsCode = addDeprecatedPragma "COMPILED" ghcBackendName q $ "= " ++ hsCode
+
+addHaskellExport :: QName -> String -> TCM ()
+addHaskellExport q hsName = addDeprecatedPragma "COMPILED_EXPORT" ghcBackendName q $ "as " ++ hsName
 
 addHaskellType :: QName -> HaskellType -> TCM ()
-addHaskellType q hsTy = do
-  ensureNoCompiledHaskell q
-  modifySignature $ updateDefinition q $ updateDefCompiledRep $ addHs
-  where
-    addHs crep = crep { compiledHaskell = Just $ HsType hsTy }
+addHaskellType q hsTy = addDeprecatedPragma "COMPILED_TYPE" ghcBackendName q $ "= type " ++ hsTy
 
-addEpicCode :: QName -> EpicCode -> TCM ()
-addEpicCode q epDef = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addEp
-  -- TODO: sanity checking
-  where
-    addEp crep = crep { compiledEpic = Just epDef }
+addHaskellData :: QName -> HaskellType -> [HaskellCode] -> TCM ()
+addHaskellData q hsTy hsCons = addDeprecatedPragma "COMPILED_DATA" ghcBackendName q $ dataFormat hsTy hsCons
 
-addJSCode :: QName -> String -> TCM ()
-addJSCode q jsDef = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addJS
-  where
-    addJS crep = crep { compiledJS = Just jsDef }
+addJSCode :: QName -> JSCode -> TCM ()
+addJSCode q jsDef = addDeprecatedPragma "COMPILED_JS" jsBackendName q ("= " ++ jsDef)
 
-addCoreCode :: QName -> CR.CoreExpr -> TCM ()
-addCoreCode q crDef =  modifySignature $ updateDefinition q $ updateDefCompiledRep $ addCore crDef
-  where
-    addCore e crep = crep { compiledCore = Just $ CrDefn e }
+addCoreCode :: QName -> CoreCode -> TCM ()
+addCoreCode q crDef = addDeprecatedPragma "COMPILED_UHC" uhcBackendName q $ "= " ++ crDef
 
-addCoreConstr :: QName -> CR.CoreConstr -> TCM ()
-addCoreConstr q con = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addCore
-  where
-    addCore crep = crep {compiledCore = Just $ CrConstr con }
+addCoreType :: QName -> CoreCode -> [CoreCode] -> TCM ()
+addCoreType q crTy crCons = addDeprecatedPragma "COMPILED_DATA_UHC" uhcBackendName q $ dataFormat crTy crCons
 
-addCoreType :: QName -> CR.CoreType -> TCM ()
-addCoreType q crTy = modifySignature $ updateDefinition q $ updateDefCompiledRep $ addCr
-  -- TODO: sanity checking
-  where
-    addCr crep = crep { compiledCore = Just $ CrType crTy }
+-- ** End of temporary functions **
+
+getUniqueCompilerPragma :: BackendName -> QName -> TCM (Maybe CompilerPragma)
+getUniqueCompilerPragma backend q = do
+  ps <- defCompilerPragmas backend <$> getConstInfo q
+  case ps of
+    []  -> return Nothing
+    [p] -> return $ Just p
+    _   -> setCurrentRange (ps !! 1) $
+            genericDocError $
+                  hang (text ("Conflicting " ++ backend ++ " pragmas for") <+> pretty q <+> text "at") 2 $
+                       vcat [ text "-" <+> pretty (getRange p) | p <- ps ]
 
 setFunctionFlag :: FunctionFlag -> Bool -> QName -> TCM ()
 setFunctionFlag flag val q = modifyGlobalDefinition q $ set (theDefLens . funFlag flag) val
@@ -167,6 +174,9 @@ markStatic = setFunctionFlag FunStatic True
 markInline :: QName -> TCM ()
 markInline = setFunctionFlag FunInline True
 
+markInjective :: QName -> TCM ()
+markInjective q = modifyGlobalDefinition q $ \def -> def { defInjective = True }
+
 unionSignatures :: [Signature] -> Signature
 unionSignatures ss = foldr unionSignature emptySignature ss
   where
@@ -189,12 +199,12 @@ addSection m = do
     if (sec == sec') then do
       -- Andreas, 2015-12-02: test/Succeed/Issue1701II.agda
       -- reports a "redundantly adding existing section".
-      reportSLn "tc.section" 10 $ "warning: redundantly adding existing section " ++ show m
-      reportSLn "tc.section" 60 $ "with content " ++ show sec
+      reportSLn "tc.section" 10 $ "warning: redundantly adding existing section " ++ prettyShow m
+      reportSLn "tc.section" 60 $ "with content " ++ prettyShow sec
     else do
-      reportSLn "impossible" 10 $ "overwriting existing section " ++ show m
-      reportSLn "impossible" 60 $ "of content   " ++ show sec'
-      reportSLn "impossible" 60 $ "with content " ++ show sec
+      reportSLn "impossible" 10 $ "overwriting existing section " ++ prettyShow m
+      reportSLn "impossible" 60 $ "of content   " ++ prettyShow sec'
+      reportSLn "impossible" 60 $ "with content " ++ prettyShow sec
       __IMPOSSIBLE__
   -- Add the new section.
   setDefaultModuleParameters m
@@ -258,8 +268,10 @@ addDisplayForms x = do
           case unSpine <$> body of
             Just (Def y es) -> do
               let df = Display m es $ DTerm $ Def top $ map Apply args
-              reportSLn "tc.display.section" 20 $ "adding display form " ++ show y ++ " --> " ++ show top
-                                                ++ "\n  " ++ show df
+              reportSLn "tc.display.section" 20 $ unlines
+                [ "adding display form " ++ prettyShow y ++ " --> " ++ prettyShow top
+                , show df
+                ]
               addDisplayForm y df
               add args top y es
             Just v          -> noDispForm x $ "not a def body, but " ++ show v
@@ -267,14 +279,16 @@ addDisplayForms x = do
         [] | Constructor{ conSrcCon = h } <- theDef def -> do
               let y  = conName h
                   df = Display 0 [] $ DTerm $ Con (h {conName = top }) ConOSystem []
-              reportSLn "tc.display.section" 20 $ "adding display form " ++ show y ++ " --> " ++ show top
-                                                ++ "\n  " ++ show df
+              reportSLn "tc.display.section" 20 $ unlines
+                [ "adding display form " ++ prettyShow y ++ " --> " ++ prettyShow top
+                , show df
+                ]
               addDisplayForm y df
         [] -> noDispForm x "no clauses"
         (_:_:_) -> noDispForm x "many clauses"
 
     noDispForm x reason = reportSLn "tc.display.section" 30 $
-      "no display form from " ++ show x ++ " because " ++ reason
+      "no display form from " ++ prettyShow x ++ " because " ++ reason
 
     isVar VarP{} = True
     isVar _      = False
@@ -293,47 +307,49 @@ applySection new ptel old ts ScopeCopyInfo{ renModules = rm, renNames = rd } = d
   where
     -- If a datatype is being copied, all its constructors need to be copied,
     -- and if a constructor is copied its datatype needs to be.
+    closeConstructors :: Ren QName -> TCM (Ren QName)
     closeConstructors rd = do
-        ds <- nub . concat <$> mapM (constructorData . fst) rd
-        cs <- nub . concat <$> mapM (dataConstructors . fst) rd
+        ds <- List.nub . concat <$> mapM (constructorData . fst) rd
+        cs <- List.nub . concat <$> mapM (dataConstructors . fst) rd
         new <- concat <$> mapM rename (ds ++ cs)
         reportSLn "tc.mod.apply.complete" 30 $
-          "also copying: " ++ show new
+          "also copying: " ++ prettyShow new
         return $ new ++ rd
       where
+        rename :: QName -> TCM (Ren QName)
         rename x =
           case lookup x rd of
-            Nothing -> do y <- freshName_ (show x)
+            Nothing -> do y <- freshName_ (prettyShow x)
                           return [(x, qnameFromList [y])]
             Just{}  -> return []
 
+        constructorData :: QName -> TCM [QName]
         constructorData x = do
           def <- theDef <$> getConstInfo x
           return $ case def of
             Constructor{ conData = d } -> [d]
             _                          -> []
+
+        dataConstructors :: QName -> TCM [QName]
         dataConstructors x = do
           def <- theDef <$> getConstInfo x
           return $ case def of
             Datatype{ dataCons = cs } -> cs
-            Record{ recConHead = h }      -> [conName h]
+            Record{ recConHead = h }  -> [conName h]
             _                         -> []
 
 applySection' :: ModuleName -> Telescope -> ModuleName -> Args -> ScopeCopyInfo -> TCM ()
 applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } = do
   reportSLn "tc.mod.apply" 10 $ render $ vcat
     [ text "applySection"
-    , text "new  =" <+> text (show new)
-    , text "ptel =" <+> text (show ptel)
-    , text "old  =" <+> text (show old)
-    , text "ts   =" <+> text (show ts)
-    ]
-  reportSLn "tc.mod.apply" 80 $ render $ vcat
-    [ text "arguments:  " <+> text (show ts)
+    , text "new  =" <+> pretty new
+    , text "ptel =" <+> pretty ptel
+    , text "old  =" <+> pretty old
+    , text "ts   =" <+> pretty ts
     ]
   mapM_ (copyDef ts) rd
   mapM_ (copySec ts) rm
-  mapM_ computePolarity (map snd rd)
+  computePolarity (map snd rd)
   where
     -- Andreas, 2013-10-29
     -- Here, if the name x is not imported, it persists as
@@ -347,7 +363,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
 
     argsToUse x = do
       let m = commonParentModule old x
-      reportSLn "tc.mod.apply" 80 $ "Common prefix: " ++ show m
+      reportSLn "tc.mod.apply" 80 $ "Common prefix: " ++ prettyShow m
       size <$> lookupSection m
 
     copyDef :: Args -> (QName, QName) -> TCM ()
@@ -357,7 +373,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
       copyDef' np def
       where
         copyDef' np d = do
-          reportSLn "tc.mod.apply" 60 $ "making new def for " ++ show y ++ " from " ++ show x ++ " with " ++ show np ++ " args " ++ show (defAbstract d)
+          reportSLn "tc.mod.apply" 60 $ "making new def for " ++ prettyShow y ++ " from " ++ prettyShow x ++ " with " ++ show np ++ " args " ++ show (defAbstract d)
           reportSLn "tc.mod.apply" 80 $
             "args = " ++ show ts' ++ "\n" ++
             "old type = " ++ prettyShow (defType d)
@@ -400,10 +416,11 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                     , defInstance       = inst
                     , defCopy           = True
                     , defMatchable      = False
+                    , defInjective      = False
                     , theDef            = df }
             oldDef = theDef d
             isCon  = case oldDef of { Constructor{} -> True ; _ -> False }
-            mutual = case oldDef of { Function{funMutual = m} -> m              ; _ -> [] }
+            mutual = case oldDef of { Function{funMutual = m} -> m              ; _ -> Nothing }
             extlam = case oldDef of { Function{funExtLam = e} -> e              ; _ -> Nothing }
             with   = case oldDef of { Function{funWith = w}   -> copyName <$> w ; _ -> Nothing }
             -- Andreas, 2015-05-11, to fix issue 1413:
@@ -416,6 +433,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                 | size ts' < n || (size ts' == n && maybe True isVar0 (lastMaybe ts'))
                 -> Just $ p { projIndex = n - size ts'
                             , projLams  = projLams p `apply` ts'
+                            , projProper= fmap copyName $ projProper p
                             }
               _ -> Nothing
             def =
@@ -450,10 +468,11 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                         , funWith           = with
                         , funCopatternLHS   = isCopatternLHS [cl]
                         }
-                  reportSLn "tc.mod.apply" 80 $ "new def for " ++ show x ++ "\n  " ++ show newDef
+                  reportSDoc "tc.mod.apply" 80 $ return $ (text "new def for" <+> pretty x) <?> pretty newDef
                   return newDef
 
-            cl = Clause { clauseRange     = getRange $ defClauses d
+            cl = Clause { clauseLHSRange  = getRange $ defClauses d
+                        , clauseFullRange = getRange $ defClauses d
                         , clauseTel       = EmptyTel
                         , namedClausePats = []
                         , clauseBody      = Just $ case oldDef of
@@ -461,6 +480,7 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
                             _ -> Def x $ map Apply ts'
                         , clauseType      = Just $ defaultArg t
                         , clauseCatchall  = False
+                        , clauseUnreachable = Just False -- definitely not unreachable
                         }
 
     {- Example
@@ -498,11 +518,11 @@ applySection' new ptel old ts ScopeCopyInfo{ renNames = rd, renModules = rm } =
       totalArgs <- argsToUse x
       tel       <- lookupSection x
       let sectionTel =  apply tel $ take totalArgs ts
-      reportSLn "tc.mod.apply" 80 $ "Copying section " ++ show x ++ " to " ++ show y
-      reportSLn "tc.mod.apply" 80 $ "  ts           = " ++ intercalate "; " (map prettyShow ts)
+      reportSLn "tc.mod.apply" 80 $ "Copying section " ++ prettyShow x ++ " to " ++ prettyShow y
+      reportSLn "tc.mod.apply" 80 $ "  ts           = " ++ List.intercalate "; " (map prettyShow ts)
       reportSLn "tc.mod.apply" 80 $ "  totalArgs    = " ++ show totalArgs
-      reportSLn "tc.mod.apply" 80 $ "  tel          = " ++ intercalate " " (map (fst . unDom) $ telToList tel)  -- only names
-      reportSLn "tc.mod.apply" 80 $ "  sectionTel   = " ++ intercalate " " (map (fst . unDom) $ telToList ptel) -- only names
+      reportSLn "tc.mod.apply" 80 $ "  tel          = " ++ List.intercalate " " (map (fst . unDom) $ telToList tel)  -- only names
+      reportSLn "tc.mod.apply" 80 $ "  sectionTel   = " ++ List.intercalate " " (map (fst . unDom) $ telToList ptel) -- only names
       addContext sectionTel $ addSection y
 
 -- | Add a display form to a definition (could be in this or imported signature).
@@ -517,13 +537,13 @@ addDisplayForm x df = do
     typeError . GenericDocError $ text "Cannot add recursive display form for" <+> pretty x
 
 isLocal :: QName -> TCM Bool
-isLocal x = isJust . HMap.lookup x <$> use (stSignature . sigDefinitions)
+isLocal x = HMap.member x <$> use (stSignature . sigDefinitions)
 
 getDisplayForms :: QName -> TCM [LocalDisplayForm]
 getDisplayForms q = do
-  ds  <- defDisplay <$> getConstInfo q
-  ds1 <- maybe [] id . HMap.lookup q <$> use stImportsDisplayForms
-  ds2 <- maybe [] id . HMap.lookup q <$> use stImportedDisplayForms
+  ds  <- either (const []) defDisplay <$> getConstInfo' q
+  ds1 <- HMap.lookupDefault [] q <$> use stImportsDisplayForms
+  ds2 <- HMap.lookupDefault [] q <$> use stImportedDisplayForms
   ifM (isLocal q) (return $ ds ++ ds1 ++ ds2)
                   (return $ ds1 ++ ds ++ ds2)
 
@@ -564,12 +584,12 @@ sameDef d1 d2 = do
 
 -- | Can be called on either a (co)datatype, a record type or a
 --   (co)constructor.
-whatInduction :: QName -> TCM Induction
-whatInduction c = do
+whatInduction :: MonadTCM tcm => QName -> tcm Induction
+whatInduction c = liftTCM $ do
   def <- theDef <$> getConstInfo c
   case def of
     Datatype{ dataInduction = i } -> return i
-    Record{ recRecursive = False} -> return Inductive
+    Record{} | not (recRecursive def) -> return Inductive
     Record{ recInduction = i    } -> return $ fromMaybe Inductive i
     Constructor{ conInd = i }     -> return i
     _                             -> __IMPOSSIBLE__
@@ -590,10 +610,31 @@ singleConstructorType q = do
         _                          -> __IMPOSSIBLE__
     _ -> __IMPOSSIBLE__
 
-class (Functor m, Applicative m, Monad m) => HasConstInfo m where
+-- | Signature lookup errors.
+data SigError
+  = SigUnknown String -- ^ The name is not in the signature; default error message.
+  | SigAbstract       -- ^ The name is not available, since it is abstract.
+
+-- | Standard eliminator for 'SigError'.
+sigError :: (String -> a) -> a -> SigError -> a
+sigError f a = \case
+  SigUnknown s -> f s
+  SigAbstract  -> a
+
+class (Functor m, Applicative m, Monad m, HasOptions m, MonadDebug m) => HasConstInfo m where
   -- | Lookup the definition of a name. The result is a closed thing, all free
   --   variables have been abstracted over.
   getConstInfo :: QName -> m Definition
+  getConstInfo q = getConstInfo' q >>= \case
+      Right d -> return d
+      Left (SigUnknown err) -> __IMPOSSIBLE_VERBOSE__ err
+      Left SigAbstract      -> __IMPOSSIBLE_VERBOSE__ $
+        "Abstract, thus, not in scope: " ++ prettyShow q
+
+  -- | Version that reports exceptions:
+  getConstInfo' :: QName -> m (Either SigError Definition)
+  getConstInfo' q = Right <$> getConstInfo q
+
   -- | Lookup the rewrite rules with the given head symbol.
   getRewriteRulesFor :: QName -> m RewriteRules
 
@@ -614,21 +655,33 @@ getOriginalProjection q = projOrig . fromMaybe __IMPOSSIBLE__ <$> isProjection q
 
 instance HasConstInfo (TCMT IO) where
   getRewriteRulesFor = defaultGetRewriteRulesFor get
-  getConstInfo q = join $ pureTCM $ \st env ->
+  getConstInfo' q = do
+    st  <- get
+    env <- ask
+    defaultGetConstInfo st env q
+  getConstInfo q = getConstInfo' q >>= \case
+      Right d -> return d
+      Left (SigUnknown err) -> fail err
+      Left SigAbstract      -> notInScope $ qnameToConcrete q
+
+defaultGetConstInfo
+  :: (HasOptions m, MonadDebug m)
+  => TCState -> TCEnv -> QName -> m (Either SigError Definition)
+defaultGetConstInfo st env q = do
     let defs  = st^.(stSignature . sigDefinitions)
         idefs = st^.(stImports . sigDefinitions)
-    in case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
-        []  -> fail $ "Unbound name: " ++ show q ++ " " ++ showQNameId q
+    case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
+        []  -> return $ Left $ SigUnknown $ "Unbound name: " ++ prettyShow q ++ " " ++ showQNameId q
         [d] -> mkAbs env d
-        ds  -> fail $ "Ambiguous name: " ++ show q
+        ds  -> __IMPOSSIBLE_VERBOSE__ $ "Ambiguous name: " ++ prettyShow q
     where
       mkAbs env d
         | treatAbstractly' q' env =
           case makeAbstract d of
-            Just d      -> return d
-            Nothing     -> notInScope $ qnameToConcrete q
+            Just d      -> return $ Right d
+            Nothing     -> return $ Left SigAbstract
               -- the above can happen since the scope checker is a bit sloppy with 'abstract'
-        | otherwise = return d
+        | otherwise = return $ Right d
         where
           q' = case theDef d of
             -- Hack to make abstract constructors work properly. The constructors
@@ -640,12 +693,12 @@ instance HasConstInfo (TCMT IO) where
           dropLastModule q@QName{ qnameModule = m } =
             q{ qnameModule = mnameFromList $ ifNull (mnameToList m) __IMPOSSIBLE__ init }
 
-instance (HasConstInfo m) => HasConstInfo (MaybeT m) where
-  getConstInfo = lift . getConstInfo
+instance HasConstInfo m => HasConstInfo (MaybeT m) where
+  getConstInfo' = lift . getConstInfo'
   getRewriteRulesFor = lift . getRewriteRulesFor
 
-instance (HasConstInfo m, Error err) => HasConstInfo (ExceptionT err m) where
-  getConstInfo = lift . getConstInfo
+instance HasConstInfo m => HasConstInfo (ExceptT err m) where
+  getConstInfo' = lift . getConstInfo'
   getRewriteRulesFor = lift . getRewriteRulesFor
 
 {-# INLINE getConInfo #-}
@@ -666,7 +719,7 @@ getPolarity' CmpLeq q = getPolarity q -- composition with Covariant is identity
 setPolarity :: QName -> [Polarity] -> TCM ()
 setPolarity q pol = do
   reportSLn "tc.polarity.set" 20 $
-    "Setting polarity of " ++ show q ++ " to " ++ show pol ++ "."
+    "Setting polarity of " ++ prettyShow q ++ " to " ++ prettyShow pol ++ "."
   modifySignature $ updateDefinition q $ updateDefPolarity $ const pol
 
 -- | Get argument occurrence info for argument @i@ of definition @d@ (never fails).
@@ -687,23 +740,21 @@ modifyArgOccurrences d f =
   modifySignature $ updateDefinition d $ updateDefArgOccurrences f
 
 setTreeless :: QName -> TTerm -> TCM ()
-setTreeless q t = modifyGlobalDefinition q $ setTT
-  where
-    setTT def@Defn{theDef = fun@Function{}} =
-      def{theDef = fun{funTreeless = Just (Compiled t [])}}
-    setTT def = __IMPOSSIBLE__
+setTreeless q t =
+  modifyGlobalDefinition q $ updateTheDef $ \case
+    fun@Function{} -> fun{ funTreeless = Just $ Compiled t [] }
+    _ -> __IMPOSSIBLE__
 
 setCompiledArgUse :: QName -> [Bool] -> TCM ()
-setCompiledArgUse q use = modifyGlobalDefinition q $ setTT
-  where
-    setTT def@Defn{theDef = fun@Function{}} =
-      def{theDef = fun{funTreeless = for (funTreeless fun) $ \ c -> c { cArgUsage = use }}}
-    setTT def = __IMPOSSIBLE__
+setCompiledArgUse q use =
+  modifyGlobalDefinition q $ updateTheDef $ \case
+    fun@Function{} ->
+      fun{ funTreeless = for (funTreeless fun) $ \ c -> c { cArgUsage = use } }
+    _ -> __IMPOSSIBLE__
 
 getCompiled :: QName -> TCM (Maybe Compiled)
 getCompiled q = do
-  def <- theDef <$> getConstInfo q
-  return $ case def of
+  (theDef <$> getConstInfo q) <&> \case
     Function{ funTreeless = t } -> t
     _                           -> Nothing
 
@@ -711,22 +762,13 @@ getErasedConArgs :: QName -> TCM [Bool]
 getErasedConArgs q = do
   def <- getConstInfo q
   case theDef def of
-    Constructor{ conData = d, conPars = np, conErased = es } -> do
-      ddef <- getConstInfo d
-      case compiledHaskell $ defCompiledRep ddef of
-        Nothing -> return es
-        Just _  -> do
-          -- Can't erase arguments of COMPILED_DATA constructors yet
-          TelV tel _ <- telView $ defType def
-          return $ replicate (size tel - np) False
+    Constructor{ conData = d, conPars = np, conErased = es } -> return es
     _ -> __IMPOSSIBLE__
 
 setErasedConArgs :: QName -> [Bool] -> TCM ()
-setErasedConArgs q args = modifyGlobalDefinition q setArgs
-  where
-    setArgs def@Defn{theDef = con@Constructor{}} =
-      def{ theDef = con{ conErased = args } }
-    setArgs def = def   -- no-op for non-constructors
+setErasedConArgs q args = modifyGlobalDefinition q $ updateTheDef $ \case
+    def@Constructor{} -> def{ conErased = args }
+    def -> def   -- no-op for non-constructors
 
 getTreeless :: QName -> TCM (Maybe TTerm)
 getTreeless q = fmap cTreeless <$> getCompiled q
@@ -734,28 +776,35 @@ getTreeless q = fmap cTreeless <$> getCompiled q
 getCompiledArgUse :: QName -> TCM [Bool]
 getCompiledArgUse q = maybe [] cArgUsage <$> getCompiled q
 
--- | Get the mutually recursive identifiers.
-getMutual :: QName -> TCM [QName]
-getMutual d = do
-  def <- theDef <$> getConstInfo d
-  return $ case def of
+-- | Get the mutually recursive identifiers of a symbol from the signature.
+getMutual :: QName -> TCM (Maybe [QName])
+getMutual d = getMutual_ . theDef <$> getConstInfo d
+
+-- | Get the mutually recursive identifiers from a `Definition`.
+getMutual_ :: Defn -> Maybe [QName]
+getMutual_ = \case
     Function {  funMutual = m } -> m
     Datatype { dataMutual = m } -> m
     Record   {  recMutual = m } -> m
-    _ -> []
+    _ -> Nothing
 
 -- | Set the mutually recursive identifiers.
 setMutual :: QName -> [QName] -> TCM ()
 setMutual d m = modifySignature $ updateDefinition d $ updateTheDef $ \ def ->
   case def of
-    Function{} -> def { funMutual = m }
-    Datatype{} -> def {dataMutual = m }
-    Record{}   -> def { recMutual = m }
-    _          -> __IMPOSSIBLE__
+    Function{} -> def { funMutual = Just m }
+    Datatype{} -> def {dataMutual = Just m }
+    Record{}   -> def { recMutual = Just m }
+    _          -> if null m then def else __IMPOSSIBLE__ -- nothing to do
 
 -- | Check whether two definitions are mutually recursive.
 mutuallyRecursive :: QName -> QName -> TCM Bool
-mutuallyRecursive d d' = (d `elem`) <$> getMutual d'
+mutuallyRecursive d d1 = (d `elem`) . fromMaybe __IMPOSSIBLE__ <$> getMutual d1
+
+-- | A function/data/record definition is nonRecursive if it is not even mutually
+--   recursive with itself.
+definitelyNonRecursive_ :: Defn -> Bool
+definitelyNonRecursive_ = maybe False null . getMutual_
 
 -- | Get the number of parameters to the current module.
 getCurrentModuleFreeVars :: TCM Nat
@@ -767,12 +816,20 @@ getCurrentModuleFreeVars = size <$> (lookupSection =<< currentModule)
 getDefFreeVars :: (Functor m, Applicative m, ReadTCState m, MonadReader TCEnv m) => QName -> m Nat
 getDefFreeVars = getModuleFreeVars . qnameModule
 
-freeVarsToApply :: QName -> TCM Args
-freeVarsToApply = moduleParamsToApply . qnameModule
+freeVarsToApply :: (Functor m, HasConstInfo m, HasOptions m,
+                    ReadTCState m, MonadReader TCEnv m, MonadDebug m)
+                => QName -> m Args
+freeVarsToApply q = do
+  vs <- moduleParamsToApply $ qnameModule q
+  t <- defType <$> getConstInfo q
+  let TelV tel _ = telView'UpTo (size vs) t
+  unless (size tel == size vs) __IMPOSSIBLE__
+  return $ zipWith (\ (Arg _ v) (Dom ai _) -> Arg ai v) vs $ telToList tel
 
 {-# SPECIALIZE getModuleFreeVars :: ModuleName -> TCM Nat #-}
 {-# SPECIALIZE getModuleFreeVars :: ModuleName -> ReduceM Nat #-}
-getModuleFreeVars :: (Functor m, Applicative m, ReadTCState m, MonadReader TCEnv m) => ModuleName -> m Nat
+getModuleFreeVars :: (Functor m, Applicative m, MonadReader TCEnv m, ReadTCState m)
+                  => ModuleName -> m Nat
 getModuleFreeVars m = do
   m0   <- commonParentModule m <$> currentModule
   (+) <$> getAnonymousVariables m <*> (size <$> lookupSection m0)
@@ -782,7 +839,7 @@ getModuleFreeVars m = do
 --   We have to insert the module telescope of the common prefix
 --   of the current module and the module where the definition comes from.
 --   (Properly raised to the current context.)
---
+--y
 --   Example:
 --   @
 --      module M₁ Γ where
@@ -791,20 +848,23 @@ getModuleFreeVars m = do
 --        module M₃ Θ where
 --          ... M₁.M₂.f [insert Γ raised by Θ]
 --   @
-moduleParamsToApply :: ModuleName -> TCM Args
+moduleParamsToApply :: (Functor m, Applicative m, HasOptions m,
+                        MonadReader TCEnv m, ReadTCState m, MonadDebug m)
+                    => ModuleName -> m Args
 moduleParamsToApply m = do
   -- Get the correct number of free variables (correctly raised) of @m@.
 
-  reportSLn "tc.sig.param" 90 $ "computing module parameters of " ++ show m
-  cxt <- getContext
+  reportSLn "tc.sig.param" 90 $ "computing module parameters of " ++ prettyShow m
   n   <- getModuleFreeVars m
   tel <- take n . telToList <$> lookupSection m
   sub <- getModuleParameterSub m
-  reportSLn "tc.sig.param" 60 $ unlines $
-    [ "  n    = " ++ show n
-    , "  cxt  = " ++ show (map (fmap fst) cxt)
-    , "  sub  = " ++ show sub
-    ]
+  verboseS "tc.sig.param" 60 $ do
+    cxt <- getContext
+    reportSLn "tc.sig.param" 60 $ unlines $
+      [ "  n    = " ++ show n
+      , "  cxt  = " ++ show (map (fmap fst) cxt)
+      , "  sub  = " ++ show sub
+      ]
   unless (size tel == n) __IMPOSSIBLE__
   let args = applySubst sub $ zipWith (\ i a -> var i <$ argFromDom a) (downFrom n) tel
   reportSLn "tc.sig.param" 60 $ "  args = " ++ show args
@@ -850,10 +910,26 @@ instantiateDef d = do
     ctx <- getContext
     m   <- currentModule
     reportSLn "tc.sig.inst" 30 $
-      "instDef in " ++ show m ++ ": " ++ show (defName d) ++ " " ++
+      "instDef in " ++ prettyShow m ++ ": " ++ prettyShow (defName d) ++ " " ++
       unwords (map show $ zipWith (<$) (reverse $ map (fst . unDom) ctx) vs)
   return $ d `apply` vs
 
+instantiateRewriteRule :: (Functor m, HasConstInfo m, HasOptions m,
+                           ReadTCState m, MonadReader TCEnv m, MonadDebug m)
+                       => RewriteRule -> m RewriteRule
+instantiateRewriteRule rew = do
+  traceSLn "rewriting" 60 ("instantiating rewrite rule " ++ show (rewName rew) ++ " to the local context.") $ do
+  vs  <- freeVarsToApply $ rewName rew
+  let rew' = rew `apply` vs
+  traceSLn "rewriting" 60 ("instantiated rewrite rule: ") $ do
+  traceSLn "rewriting" 60 (show rew') $ do
+  return rew'
+
+instantiateRewriteRules :: (Functor m, HasConstInfo m, HasOptions m,
+                            ReadTCState m, MonadReader TCEnv m, MonadDebug m)
+                        => RewriteRules -> m RewriteRules
+instantiateRewriteRules = mapM instantiateRewriteRule
+
 -- | Give the abstract view of a definition.
 makeAbstract :: Definition -> Maybe Definition
 makeAbstract d =
@@ -867,13 +943,16 @@ makeAbstract d =
                }
   where
     makeAbs Axiom         = Just Axiom
-    makeAbs Datatype   {} = Just AbstractDefn
-    makeAbs Function   {} = Just AbstractDefn
+    makeAbs d@Datatype {} = Just $ AbstractDefn d
+    makeAbs d@Function {} = Just $ AbstractDefn d
     makeAbs Constructor{} = Nothing
     -- Andreas, 2012-11-18:  Make record constructor and projections abstract.
-    makeAbs d@Record{}    = Just AbstractDefn
+    -- Andreas, 2017-08-14:  Projections are actually not abstract (issue #2682).
+    -- Return the Defn under a wrapper to allow e.g. eligibleForProjectionLike
+    -- to see whether the abstract thing is a record type or not.
+    makeAbs d@Record{}    = Just $ AbstractDefn d
     makeAbs Primitive{}   = __IMPOSSIBLE__
-    makeAbs AbstractDefn  = __IMPOSSIBLE__
+    makeAbs AbstractDefn{}= __IMPOSSIBLE__
 
 -- | Enter abstract mode. Abstract definition in the current module are transparent.
 {-# SPECIALIZE inAbstractMode :: TCM a -> TCM a #-}
@@ -942,14 +1021,6 @@ typeOfConst q = defType <$> (instantiateDef =<< getConstInfo q)
 relOfConst :: QName -> TCM Relevance
 relOfConst q = defRelevance <$> getConstInfo q
 
--- | The name must be a datatype.
-sortOfConst :: QName -> TCM Sort
-sortOfConst q =
-    do  d <- theDef <$> getConstInfo q
-        case d of
-            Datatype{dataSort = s} -> return s
-            _                      -> fail $ "Expected " ++ show q ++ " to be a datatype."
-
 -- | The number of dropped parameters for a definition.
 --   0 except for projection(-like) functions and constructors.
 droppedPars :: Definition -> Int
@@ -960,7 +1031,7 @@ droppedPars d = case theDef d of
     Record     {recPars = _} -> 0  -- not dropped
     Constructor{conPars = n} -> n
     Primitive{}              -> 0
-    AbstractDefn             -> __IMPOSSIBLE__
+    AbstractDefn{}           -> __IMPOSSIBLE__
 
 -- | Is it the name of a record projection?
 {-# SPECIALIZE isProjection :: QName -> TCM (Maybe Projection) #-}
@@ -986,7 +1057,7 @@ isInlineFun = (^. funInline)
 --   (projection applied to argument).
 isProperProjection :: Defn -> Bool
 isProperProjection d = caseMaybe (isProjection_ d) False $ \ isP ->
-  if projIndex isP <= 0 then False else projProper isP
+  if projIndex isP <= 0 then False else isJust $ projProper isP
 
 -- | Number of dropped initial arguments of a projection(-like) function.
 projectionArgs :: Defn -> Int
@@ -1008,5 +1079,5 @@ applyDef o f a = do
   caseMaybeM (isProjection f) fallback $ \ isP -> do
     if projIndex isP <= 0 then fallback else do
       -- Get the original projection, if existing.
-      if not (projProper isP) then fallback else do
+      if isNothing (projProper isP) then fallback else do
         return $ unArg a `applyE` [Proj o $ projOrig isP]
