@@ -14,6 +14,7 @@ import Agda.Syntax.Common
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Monad.Debug
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.State
@@ -42,6 +43,12 @@ data BoundedSize
 class IsSizeType a where
   isSizeType :: a -> TCM (Maybe BoundedSize)
 
+instance IsSizeType a => IsSizeType (Dom a) where
+  isSizeType = isSizeType . unDom
+
+instance IsSizeType a => IsSizeType (b,a) where
+  isSizeType = isSizeType . snd
+
 instance IsSizeType a => IsSizeType (Type' a) where
   isSizeType = isSizeType . unEl
 
@@ -163,6 +170,7 @@ sizeMax vs = case vs of
 -- | A useful view on sizes.
 data SizeView = SizeInf | SizeSuc Term | OtherSize Term
 
+-- | Expects argument to be 'reduce'd.
 sizeView :: Term -> TCM SizeView
 sizeView v = do
   Def inf [] <- ignoreSharing <$> primSizeInf
