@@ -8,9 +8,11 @@ module Agda.TypeChecking.Monad.State where
 import Control.Arrow (first)
 import Control.Applicative
 import qualified Control.Exception as E
-import Control.Monad.State (put, get, gets, modify)
+import Control.Monad.Reader (asks)
+import Control.Monad.State (put, get, gets, modify, modify')
 import Control.Monad.Trans (liftIO)
 
+import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Monoid
@@ -31,6 +33,8 @@ import Agda.Syntax.Abstract.Name
 import Agda.Syntax.Internal
 
 import Agda.TypeChecking.Monad.Base
+import Agda.TypeChecking.Warnings
+import {-# SOURCE #-} Agda.TypeChecking.Monad.Debug
 import {-# SOURCE #-} Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.CompiledClause
@@ -38,7 +42,7 @@ import Agda.TypeChecking.CompiledClause
 import Agda.Utils.Hash
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.Lens
-import Agda.Utils.Monad (bracket_, modify')
+import Agda.Utils.Monad (bracket_)
 import Agda.Utils.Pretty
 import Agda.Utils.Tuple
 
@@ -54,12 +58,14 @@ resetState = do
 
 -- | Resets all of the type checking state.
 --
---   Keep only 'Benchmark' information.
+--   Keep only 'Benchmark' and backend information.
 
 resetAllState :: TCM ()
 resetAllState = do
     b <- getBenchmark
+    backends <- use stBackends
     put $ updatePersistentState (\ s -> s { stBenchmark = b }) initState
+    stBackends .= backends
 -- resetAllState = put initState
 
 -- | Restore 'TCState' after performing subcomputation.
@@ -85,6 +91,31 @@ localTCStateSaving compute = do
     modifyBenchmark $ const b
   return (result, newState)
 
+-- | A fresh TCM instance.
+--
+-- The computation is run in a fresh state, with the exception that
+-- the persistent state is preserved. If the computation changes the
+-- state, then these changes are ignored, except for changes to the
+-- persistent state. (Changes to the persistent state are also ignored
+-- if errors other than type errors or IO exceptions are encountered.)
+
+freshTCM :: TCM a -> TCM (Either TCErr a)
+freshTCM m = do
+  ps <- use lensPersistentState
+  let s = set lensPersistentState ps initState
+  r <- liftIO $ (Right <$> runTCM initEnv s m) `E.catch` (return . Left)
+  case r of
+    Right (a, s) -> do
+      lensPersistentState .= s ^. lensPersistentState
+      return $ Right a
+    Left err -> do
+      case err of
+        TypeError { tcErrState = s } ->
+          lensPersistentState .= s ^. lensPersistentState
+        IOException s _ _ ->
+          lensPersistentState .= s ^. lensPersistentState
+        _ -> return ()
+      return $ Left err
 
 ---------------------------------------------------------------------------
 -- * Lens for persistent states and its fields
@@ -162,7 +193,7 @@ notInScope x = do
 printScope :: String -> Int -> String -> TCM ()
 printScope tag v s = verboseS ("scope." ++ tag) v $ do
   scope <- getScope
-  reportSDoc ("scope." ++ tag) v $ return $ vcat [ text s, text $ show scope ]
+  reportSDoc ("scope." ++ tag) v $ return $ vcat [ text s, pretty scope ]
 
 ---------------------------------------------------------------------------
 -- * Signature
@@ -184,8 +215,9 @@ getSignature = use stSignature
 --   module. This function is currently used to update the compiled
 --   representation of a function during compilation.
 modifyGlobalDefinition :: QName -> (Definition -> Definition) -> TCM ()
-modifyGlobalDefinition q f = modifySignature (updateDefinition q f) >>
-                             modifyImportedSignature (updateDefinition q f)
+modifyGlobalDefinition q f = do
+  modifySignature         $ updateDefinition q f
+  modifyImportedSignature $ updateDefinition q f
 
 setSignature :: Signature -> TCM ()
 setSignature sig = modifySignature $ const sig
@@ -237,6 +269,9 @@ updateDefPolarity f def = def { defPolarity = f (defPolarity def) }
 updateDefCompiledRep :: (CompiledRepresentation -> CompiledRepresentation) -> (Definition -> Definition)
 updateDefCompiledRep f def = def { defCompiledRep = f (defCompiledRep def) }
 
+addCompilerPragma :: BackendName -> CompilerPragma -> Definition -> Definition
+addCompilerPragma backend pragma = updateDefCompiledRep $ Map.insertWith (++) backend [pragma]
+
 updateFunClauses :: ([Clause] -> [Clause]) -> (Defn -> Defn)
 updateFunClauses f def@Function{ funClauses = cs} = def { funClauses = f cs }
 updateFunClauses f _                              = __IMPOSSIBLE__
@@ -245,6 +280,10 @@ updateCompiledClauses :: (Maybe CompiledClauses -> Maybe CompiledClauses) -> (De
 updateCompiledClauses f def@Function{ funCompiled = cc} = def { funCompiled = f cc }
 updateCompiledClauses f _                              = __IMPOSSIBLE__
 
+updateFunCopatternLHS :: (Bool -> Bool) -> Defn -> Defn
+updateFunCopatternLHS f def@Function{ funCopatternLHS = b } = def { funCopatternLHS = f b }
+updateFunCopatternLHS f _ = __IMPOSSIBLE__
+
 ---------------------------------------------------------------------------
 -- * Top level module
 ---------------------------------------------------------------------------
@@ -256,7 +295,7 @@ updateCompiledClauses f _                              = __IMPOSSIBLE__
 -- implementation of 'setTopLevelModule' should be changed.
 
 setTopLevelModule :: C.QName -> TCM ()
-setTopLevelModule x = stFreshNameId .= NameId 0 (hashString (show x))
+setTopLevelModule x = stFreshNameId .= NameId 0 (hashString $ prettyShow x)
 
 -- | Use a different top-level module for a computation. Used when generating
 --   names for imported modules.
@@ -269,27 +308,41 @@ withTopLevelModule x m = do
   return y
 
 ---------------------------------------------------------------------------
--- * Haskell imports
+-- * Foreign code
 ---------------------------------------------------------------------------
 
+addForeignCode :: BackendName -> String -> TCM ()
+addForeignCode backend code = do
+  r <- asks envRange  -- can't use TypeChecking.Monad.Trace.getCurrentRange without cycle
+  stForeignCode . key backend %= Just . (ForeignCode r code :) . fromMaybe []
+
+---------------------------------------------------------------------------
+-- * Temporary: Haskell imports
+--   These will go away when we remove the IMPORT and HASKELL pragmas in
+--   favour of the FOREIGN pragma.
+---------------------------------------------------------------------------
+
+addDeprecatedForeignCode :: String -> BackendName -> String -> TCM ()
+addDeprecatedForeignCode old backend code = do
+  warning $ DeprecationWarning (unwords ["The", old, "pragma"])
+                               foreignPragma "2.6"
+  addForeignCode backend code
+  where
+    spc | length (lines code) > 1 = "\n"
+        | otherwise               = " "
+    foreignPragma =
+      "{-# FOREIGN " ++ backend ++ spc ++ code ++ spc ++ "#-}"
+
 -- | Tell the compiler to import the given Haskell module.
 addHaskellImport :: String -> TCM ()
-addHaskellImport i = stHaskellImports %= Set.insert i
-
--- | Get the Haskell imports.
-getHaskellImports :: TCM (Set String)
-getHaskellImports = use stHaskellImports
+addHaskellImport i = addDeprecatedForeignCode "IMPORT" ghcBackendName $ "import qualified " ++ i
 
 -- | Tell the compiler to import the given Haskell module.
 addHaskellImportUHC :: String -> TCM ()
-addHaskellImportUHC i = stHaskellImportsUHC %= Set.insert i
-
--- | Get the Haskell imports.
-getHaskellImportsUHC :: TCM (Set String)
-getHaskellImportsUHC = use stHaskellImportsUHC
+addHaskellImportUHC i = addDeprecatedForeignCode "IMPORT_UHC" ghcBackendName $ "__IMPORT__ " ++ i
 
 addInlineHaskell :: String -> TCM ()
-addInlineHaskell s = stHaskellCode %= (s :)
+addInlineHaskell s = addDeprecatedForeignCode "HASKELL" ghcBackendName s
 
 ---------------------------------------------------------------------------
 -- * Interaction output callback
@@ -301,7 +354,7 @@ getInteractionOutputCallback
 
 appInteractionOutputCallback :: Response -> TCM ()
 appInteractionOutputCallback r
-  = getInteractionOutputCallback >>= \ cb -> cb r
+  = getInteractionOutputCallback >>= \ cb -> liftIO $ cb r
 
 setInteractionOutputCallback :: InteractionOutputCallback -> TCM ()
 setInteractionOutputCallback cb
@@ -355,30 +408,6 @@ getBenchmark = gets $ theBenchmark
 modifyBenchmark :: (Benchmark -> Benchmark) -> TCM ()
 modifyBenchmark = modify' . updateBenchmark
 
--- | Run a fresh instance of the TCM (with initial state).
---   'Benchmark' info is preserved.
-freshTCM :: TCM a -> TCM (Either TCErr a)
-freshTCM m = do
-  -- Prepare an initial state with current benchmark info.
-  b <- getBenchmark
-  a <- use lensAccumStatistics
-  let s = updateBenchmark (const b)
-        . set lensAccumStatistics a
-        $ initState
-  -- Run subcomputation in initial state.
-  -- If we encounter an exception, we lose the state and the
-  -- benchmark info.
-  -- We could retrieve i from a type error, which carries the state,
-  -- but we do not care for benchmarking in the presence of errors.
-  r <- liftIO $ (Right <$> runTCM initEnv s m) `E.catch` (return . Left)
-  case r of
-    Left err     -> return $ Left err
-    Right (a, s) -> do
-      -- Keep only the benchmark info from the final state of the subcomp.
-      modifyBenchmark $ const $ theBenchmark s
-      lensAccumStatistics .= (s^.lensAccumStatistics)
-      return $ Right a
-
 ---------------------------------------------------------------------------
 -- * Instance definitions
 ---------------------------------------------------------------------------
@@ -415,7 +444,9 @@ clearAnonInstanceDefs = modifyInstanceDefs $ mapSnd $ const Set.empty
 -- | Add an instance whose type is still unresolved.
 addUnknownInstance :: QName -> TCM ()
 addUnknownInstance x = do
-  reportSLn "tc.decl.instance" 10 $ "adding definition " ++ show x ++ " to the instance table (the type is not yet known)"
+  reportSLn "tc.decl.instance" 10 $
+    "adding definition " ++ prettyShow x ++
+    " to the instance table (the type is not yet known)"
   modifyInstanceDefs $ mapSnd $ Set.insert x
 
 -- | Add instance to some ``class''.
@@ -424,7 +455,8 @@ addNamedInstance
   -> QName  -- ^ Name of the class.
   -> TCM ()
 addNamedInstance x n = do
-  reportSLn "tc.decl.instance" 10 $ ("adding definition " ++ show x ++ " to instance table for " ++ show n)
+  reportSLn "tc.decl.instance" 10 $
+    "adding definition " ++ prettyShow x ++ " to instance table for " ++ prettyShow n
   -- Mark x as instance for n.
   modifySignature $ updateDefinition x $ \ d -> d { defInstance = Just n }
   -- Add x to n's instances.
