@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP                      #-}
 {-# LANGUAGE NondecreasingIndentation #-}
 
 -- | Pattern matcher used in the reducer for clauses that
@@ -20,7 +19,7 @@ import Agda.Syntax.Internal
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
 import Agda.TypeChecking.Substitute
-import Agda.TypeChecking.Monad hiding (reportSDoc)
+import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Records
 import Agda.TypeChecking.Datatypes
@@ -142,9 +141,9 @@ matchCopatterns ps vs = do
           , nest 2 $ text "ps =" <+> fsep (punctuate comma $ map (prettyTCM . namedArg) ps)
           , nest 2 $ text "vs =" <+> fsep (punctuate comma $ map prettyTCM vs)
           ]) $ do
-     -- Buggy, see issue 1124:
-     -- mapFst mconcat . unzip <$> zipWithM' (matchCopattern . namedArg) ps vs
-     foldMatch (matchCopattern . namedArg) ps vs
+  -- Buggy, see issue 1124:
+  -- mapFst mconcat . unzip <$> zipWithM' (matchCopattern . namedArg) ps vs
+  foldMatch (matchCopattern . namedArg) ps vs
 
 -- | Match a single copattern.
 matchCopattern :: DeBruijnPattern
@@ -168,10 +167,10 @@ matchPatterns ps vs = do
           , nest 2 $ text "ps =" <+> fsep (punctuate comma $ map (text . show) ps)
           , nest 2 $ text "vs =" <+> fsep (punctuate comma $ map prettyTCM vs)
           ]) $ do
-     -- Buggy, see issue 1124:
-     -- (ms,vs) <- unzip <$> zipWithM' (matchPattern . namedArg) ps vs
-     -- return (mconcat ms, vs)
-     foldMatch (matchPattern . namedArg) ps vs
+  -- Buggy, see issue 1124:
+  -- (ms,vs) <- unzip <$> zipWithM' (matchPattern . namedArg) ps vs
+  -- return (mconcat ms, vs)
+  foldMatch (matchPattern . namedArg) ps vs
 
 -- | Match a single pattern.
 matchPattern :: DeBruijnPattern
@@ -182,6 +181,7 @@ matchPattern p u = case (p, u) of
   (VarP x , arg          ) -> return (Yes NoSimplification entry, arg)
     where entry = singleton (dbPatVarIndex x, arg)
   (DotP _ , arg@(Arg _ v)) -> return (Yes NoSimplification empty, arg)
+  (AbsurdP _ , arg@(Arg _ v)) -> return (Yes NoSimplification empty, arg)
   (LitP l , arg@(Arg _ v)) -> do
     w <- reduceB' v
     let arg' = arg $> ignoreBlocking w
