@@ -5,7 +5,6 @@ module Agda.TypeChecking.Polarity where
 import Control.Applicative
 import Control.Monad.State
 
-import Data.List
 import Data.Maybe
 import Data.Traversable (traverse)
 
@@ -16,6 +15,7 @@ import Agda.Syntax.Internal.Pattern
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.SizedTypes
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Reduce
@@ -28,6 +28,7 @@ import Agda.Utils.List
 import Agda.Utils.Maybe ( whenNothingM )
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty ( prettyShow )
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -84,31 +85,53 @@ nextPolarity (p : ps) = (p, ps)
 purgeNonvariant :: [Polarity] -> [Polarity]
 purgeNonvariant = map (\ p -> if p == Nonvariant then Covariant else p)
 
+
+-- | A quick transliterations of occurrences to polarities.
+polarityFromPositivity :: QName -> TCM ()
+polarityFromPositivity x = inConcreteOrAbstractMode x $ \ def -> do
+
+  -- Get basic polarity from positivity analysis.
+  let npars = droppedPars def
+  let pol0 = replicate npars Nonvariant ++ map polFromOcc (defArgOccurrences def)
+  reportSLn "tc.polarity.set" 15 $
+    "Polarity of " ++ prettyShow x ++ " from positivity: " ++ prettyShow pol0
+
+  -- set the polarity in the signature (not the final polarity, though)
+  setPolarity x $ drop npars pol0
+
 ------------------------------------------------------------------------
 -- * Computing the polarity of a symbol.
 ------------------------------------------------------------------------
 
 -- | Main function of this module.
-computePolarity :: QName -> TCM ()
-computePolarity x = inConcreteOrAbstractMode x $ \ def -> do
-  reportSLn "tc.polarity.set" 25 $ "Computing polarity of " ++ show x
+computePolarity :: [QName] -> TCM ()
+computePolarity xs = do
 
-  -- get basic polarity from positivity analysis
+ -- Andreas, 2017-04-26, issue #2554
+ -- First, for mutual definitions, obtain a crude polarity from positivity.
+ when (length xs >= 2) $ mapM_ polarityFromPositivity xs
+
+ -- Then, refine it.
+ forM_ xs $ \ x -> inConcreteOrAbstractMode x $ \ def -> do
+  reportSLn "tc.polarity.set" 25 $ "Refining polarity of " ++ prettyShow x
+
+  -- Again: get basic polarity from positivity analysis.
   let npars = droppedPars def
   let pol0 = replicate npars Nonvariant ++ map polFromOcc (defArgOccurrences def)
-  reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ show x ++ " from positivity: " ++ show pol0
+  reportSLn "tc.polarity.set" 15 $
+    "Polarity of " ++ prettyShow x ++ " from positivity: " ++ prettyShow pol0
 
 {-
   -- get basic polarity from shape of def (arguments matched on or not?)
   def      <- getConstInfo x
   let usagePol = usagePolarity $ theDef def
-  reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ show x ++ " from definition form: " ++ show usagePol
+  reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ prettyShow x ++ " from definition form: " ++ prettyShow usagePol
   let n = genericLength usagePol  -- n <- getArity x
   reportSLn "tc.polarity.set" 20 $ "  arity = " ++ show n
 
   -- refine polarity by positivity information
   pol0 <- zipWith (/\) usagePol <$> mapM getPol [0..n - 1]
-  reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ show x ++ " from positivity: " ++ show pol0
+  reportSLn "tc.polarity.set" 15 $ "Polarity of " ++ prettyShow x ++ " from positivity: " ++ prettyShow pol0
 -}
 
   -- compute polarity of sized types
@@ -119,29 +142,17 @@ computePolarity x = inConcreteOrAbstractMode x $ \ def -> do
   -- Instantiation takes place in Rules.Decl.instantiateDefinitionType
   -- t <- instantiateFull t -- Andreas, 2014-04-11 Issue 1099: needed for
   --                        -- variable occurrence test in  dependentPolarity.
-  reportSDoc "tc.polarity.set" 15 $ text "Refining polarity with type " <+> prettyTCM t
+  reportSDoc "tc.polarity.set" 15 $
+    text "Refining polarity with type " <+> prettyTCM t
+  reportSDoc "tc.polarity.set" 60 $
+    text "Refining polarity with type (raw): " <+> (text .show) t
+
   pol <- dependentPolarity t (enablePhantomTypes (theDef def) pol1) pol1
-  reportSLn "tc.polarity.set" 10 $ "Polarity of " ++ show x ++ ": " ++ show pol
+  reportSLn "tc.polarity.set" 10 $ "Polarity of " ++ prettyShow x ++ ": " ++ prettyShow pol
 
   -- set the polarity in the signature
   setPolarity x $ drop npars pol -- purgeNonvariant pol -- temporarily disable non-variance
 
-  -- make 'Nonvariant' args 'UnusedArg' in type and clause telescope
-  -- Andreas 2012-11-18: skip this for abstract definitions (fixing issue 755).
-  -- This means that the most precise type for abstract definitions
-  -- is not available, even to other abstract definitions.
-  -- A proper fix would be to introduce a second type for use within abstract.
-  --
-  -- Andreas, 2015-07-01: I thought one should do this for
-  -- abstract local definitions in @where@ blocks to fix Issue 1366b,
-  --but it is not necessary.
-  -- t <- if (defAbstract def == AbstractDef) && not (isAnonymousModuleName $ qnameModule x)
-  t <- if (defAbstract def == AbstractDef)
-         then return t
-         else nonvariantToUnusedArg pol t
-  modifySignature $ updateDefinition x $
-   updateTheDef (nonvariantToUnusedArgInDef pol) . updateDefType (const t)
-
 -- | Data and record parameters are used as phantom arguments all over
 --   the test suite (and possibly in user developments).
 --   @enablePhantomTypes@ turns 'Nonvariant' parameters to 'Covariant'
@@ -151,7 +162,7 @@ enablePhantomTypes def pol = case def of
   Datatype{ dataPars = np } -> enable np
   Record  { recPars  = np } -> enable np
   _                         -> pol
-  where enable np = let (pars, rest) = genericSplitAt np pol
+  where enable np = let (pars, rest) = splitAt np pol
                     in  purgeNonvariant pars ++ rest
 
 {- UNUSED
@@ -186,20 +197,21 @@ dependentPolarity t _      []          = return []  -- all remaining are 'Invari
 dependentPolarity t []     (_ : _)     = __IMPOSSIBLE__
 dependentPolarity t (q:qs) pols@(p:ps) = do
   t <- reduce $ unEl t
+  reportSDoc "tc.polarity.dep" 20 $ text "dependentPolarity t = " <+> prettyTCM t
+  reportSDoc "tc.polarity.dep" 70 $ text "dependentPolarity t = " <+> (text . show) t
   case ignoreSharing t of
     Pi dom b -> do
-      let c = absBody b
-      ps <- dependentPolarity c qs ps
-      let mp = ifM (isJust <$> isSizeType (unDom dom)) (return p) (return q)
-      p  <- case b of
-              Abs{} | p /= Invariant  ->
-                -- Andreas, 2014-04-11 see Issue 1099
-                -- Free variable analysis is not in the monad,
-                -- hence metas must have been instantiated before!
-                ifM (relevantInIgnoringNonvariant 0 c ps)
-                  (return Invariant)
-                  mp
-              _ -> mp
+      ps <- underAbstraction dom b $ \ c -> dependentPolarity c qs ps
+      let fallback = ifM (isJust <$> isSizeType (unDom dom)) (return p) (return q)
+      p <- case b of
+        Abs{} | p /= Invariant  ->
+          -- Andreas, 2014-04-11 see Issue 1099
+          -- Free variable analysis is not in the monad,
+          -- hence metas must have been instantiated before!
+          ifM (relevantInIgnoringNonvariant 0 (absBody b) ps)
+            {- then -} (return Invariant)
+            {- else -} fallback
+        _ -> fallback
       return $ p : ps
     _ -> return pols
 
@@ -214,55 +226,6 @@ relevantInIgnoringNonvariant i t (p:ps) = do
               else relevantInIgnoringNonvariant (i + 1) (absBody b) ps
     _ -> return $ i `relevantInIgnoringSortAnn` t
 
--- * Turn polarity 'Nonvariant' into relevance 'UnusedArg'
-
--- | Record information that an argument is unused in 'Relevance'.
-mkUnused :: Relevance -> Relevance
-mkUnused Relevant = UnusedArg   -- commenting out this line switches off 'UnusedArg' polarity machinery
-mkUnused r        = r  -- 'Irrelevant' is more informative than 'UnusedArg'.
-
--- | Improve 'Relevance' information in a type by polarity information.
---   'Nonvariant' becomes 'UnusedArg'.
-nonvariantToUnusedArg :: [Polarity] -> Type -> TCM Type
-nonvariantToUnusedArg []     t = return t
-nonvariantToUnusedArg (p:ps) t = do
-  t <- reduce t
-  case ignoreSharingType t of
-    El s (Pi a b) -> do
-      let a' = if p == Nonvariant then mapRelevance mkUnused a else a
-      El s . Pi a' <$> traverse (nonvariantToUnusedArg ps) b
-        -- we do not lift properly but bound variables do not matter for reduce
-        -- also, we do not maintain the context
-    _ -> return t
-
--- | Propagate 'Nonvariant' 'Polarity' to 'Relevance' information in
---   'Arg's of a defined symbol.
-nonvariantToUnusedArgInDef :: [Polarity] -> Defn -> Defn
-nonvariantToUnusedArgInDef pol def = case def of
-  Function { funClauses = cl } ->
-       def { funClauses = map (nonvariantToUnusedArgInClause pol) cl }
-  _ -> def
-
-nonvariantToUnusedArgInClause :: [Polarity] -> Clause -> Clause
-nonvariantToUnusedArgInClause pol cl@Clause{clauseTel = tel, namedClausePats = ps} =
-  let adjPat p Nonvariant
-        | properlyMatching (namedArg p) = __IMPOSSIBLE__ -- if we match, we cannot be Nonvariant (sanity check)
-        | otherwise                     = mapRelevance mkUnused p
-      adjPat p _    = p
-      -- change relevance of 'Nonvariant' arguments to 'UnusedArg'
-      -- note that the associated patterns cannot be 'ConP' or 'LitP'
-      ps'   = zipWith adjPat ps (pol ++ repeat Invariant)
-      -- get a list of 'Relevance's for the variables bound in the pattern
-      rels0 = getRelevance <$> (concatMap (patternVars . fmap namedThing) ps')
-      -- this is the order the variables appear in the telescope
-      perm  = fromMaybe __IMPOSSIBLE__ $ clausePerm cl
-      rels  = permute perm rels0
-      -- now improve 'Relevance' in 'Telescope' by pattern relevance
-      updateDom UnusedArg = mapRelevance mkUnused
-      updateDom r          = id
-      tel' = telFromList $ zipWith updateDom rels $ telToList tel
-   in cl { namedClausePats = ps', clauseTel = tel'}
-
 ------------------------------------------------------------------------
 -- * Sized types
 ------------------------------------------------------------------------
@@ -278,12 +241,12 @@ sizePolarity d pol0 = do
     case theDef def of
       Datatype{ dataPars = np, dataCons = cons } -> do
         let TelV tel _      = telView' $ defType def
-            (parTel, ixTel) = genericSplitAt np $ telToList tel
+            (parTel, ixTel) = splitAt np $ telToList tel
         case ixTel of
           []                 -> exit  -- No size index
           Dom _ (_, a) : _ -> ifM ((/= Just BoundedNo) <$> isSizeType a) exit $ do
             -- we assume the size index to be 'Covariant' ...
-            let pol   = genericTake np pol0
+            let pol   = take np pol0
                 polCo = pol ++ [Covariant]
                 polIn = pol ++ [Invariant]
             setPolarity d $ polCo
@@ -303,7 +266,7 @@ sizePolarity d pol0 = do
                           let isPos = underAbstraction arg tel $ \ tel -> do
                                 pols <- zipWithM polarity [0..] $ map (snd . unDom) $ telToList tel
                                 reportSDoc "tc.polarity.size" 25 $
-                                  text $ "to pass size polarity check, the following polarities need all to be covariant: " ++ show pols
+                                  text $ "to pass size polarity check, the following polarities need all to be covariant: " ++ prettyShow pols
                                 return $ all (`elem` [Nonvariant, Covariant]) pols
 
                           -- check that the size argument appears in the
@@ -329,25 +292,26 @@ sizePolarity d pol0 = do
       _ -> exit
 
 -- | @checkSizeIndex d np i a@ checks that constructor target type @a@
---   has form @d ps (↑ i) idxs@ where @|ps| = np@.
+--   has form @d ps (↑ⁿ i) idxs@ where @|ps| = np@.
 --
 --   Precondition: @a@ is reduced and of form @d ps idxs0@.
 checkSizeIndex :: QName -> Nat -> Nat -> Type -> TCM Bool
 checkSizeIndex d np i a = do
-  reportSDoc "tc.polarity.size" 15 $ withShowAllArguments $
-    text "checking that constructor target type " <+> prettyTCM a <+>
-    text "is data type " <+> prettyTCM d <+>
-    text "has size index successor of " <+> prettyTCM (var i)
+  reportSDoc "tc.polarity.size" 15 $ withShowAllArguments $ vcat
+    [ text "checking that constructor target type " <+> prettyTCM a
+    , text "  is data type " <+> prettyTCM d
+    , text "  and has size index (successor(s) of) " <+> prettyTCM (var i)
+    ]
   case ignoreSharing $ unEl a of
     Def d0 es -> do
       whenNothingM (sameDef d d0) __IMPOSSIBLE__
-      s <- sizeView $ unArg ix
+      s <- deepSizeView $ unArg ix
       case s of
-        SizeSuc v | Var j [] <- ignoreSharing v, i == j
+        DSizeVar j _ | i == j
           -> return $ not $ freeIn i (pars ++ ixs)
         _ -> return False
       where
-        (pars, Apply ix : ixs) = genericSplitAt np es
+        (pars, Apply ix : ixs) = splitAt np es
     _ -> __IMPOSSIBLE__
 
 -- | @polarities i a@ computes the list of polarities of de Bruijn index @i@
