@@ -18,7 +18,7 @@ import Data.Function
 import Data.Graph (SCC(..), flattenSCC)
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
-import Data.List as List hiding (null)
+import qualified Data.List as List
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Monoid (mconcat)
@@ -30,26 +30,31 @@ import Debug.Trace
 
 import Agda.Syntax.Common
 import qualified Agda.Syntax.Info as Info
-import Agda.Syntax.Position (fuseRange, Range, HasRange(..), noRange)
 import Agda.Syntax.Internal
 import Agda.Syntax.Internal.Pattern
-import Agda.TypeChecking.Datatypes (isDataOrRecordType, DataOrRecord(..))
-import Agda.TypeChecking.Records
+import Agda.Syntax.Position (fuseRange, Range, HasRange(..), noRange)
+import Agda.TypeChecking.Datatypes ( isDataOrRecordType )
+import Agda.TypeChecking.Functions
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin (primInf, CoinductionKit(..), coinductionKit)
-import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Positivity.Occurrence
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Records
+import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
+import Agda.TypeChecking.Warnings
 
 import qualified Agda.Utils.Graph.AdjacencyMap.Unidirectional as Graph
+import Agda.Utils.Function (applyUnless)
 import Agda.Utils.Functor
 import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import qualified Agda.Utils.Permutation as Perm
+import qualified Agda.Utils.Pretty as P
+import Agda.Utils.Pretty (Pretty, prettyShow)
 import Agda.Utils.SemiRing
 import Agda.Utils.Singleton
 import Agda.Utils.Size
@@ -70,7 +75,7 @@ checkStrictlyPositive mi qset = disableDestructiveUpdate $ do
   let qs = Set.toList qset
   reportSDoc "tc.pos.tick" 100 $ text "positivity of" <+> prettyTCM qs
   g <- buildOccurrenceGraph qset
-  let (gstar, sccs') =
+  let (gstar, sccs) =
         Graph.gaussJordanFloydWarshallMcNaughtonYamada $ fmap occ g
   reportSDoc "tc.pos.tick" 100 $ text "constructed graph"
   reportSLn "tc.pos.graph" 5 $ "Positivity graph: N=" ++ show (size $ Graph.nodes g) ++
@@ -92,18 +97,27 @@ checkStrictlyPositive mi qset = disableDestructiveUpdate $ do
   reportSDoc "tc.pos.tick" 100 $ text "set args"
 
   -- check positivity for all strongly connected components of the graph for qs
-  let sccs = map flattenSCC sccs'
   reportSDoc "tc.pos.graph.sccs" 10 $ do
-    let (triv, others) = partitionEithers $ for sccs' $ \ scc -> case scc of
+    let (triv, others) = partitionEithers $ for sccs $ \case
           AcyclicSCC v -> Left v
           CyclicSCC vs -> Right vs
     sep [ text $ show (length triv) ++ " trivial sccs"
         , text $ show (length others) ++ " non-trivial sccs with lengths " ++
             show (map length others)
         ]
-  reportSDoc "tc.pos.graph.sccs" 15 $ text $ "  sccs = " ++ show sccs
-  forM_ sccs $ \ scc -> setMut [ q | DefNode q <- scc ]
-  mapM_ (checkPos g gstar) $ qs
+  reportSLn "tc.pos.graph.sccs" 15 $
+    "  sccs = " ++ prettyShow [ scc | CyclicSCC scc <- sccs ]
+  forM_ sccs $ \case
+    -- If the mutuality information has never been set, we set it to []
+    AcyclicSCC (DefNode q) -> whenM (isNothing <$> getMutual q) $ do
+      reportSLn "tc.pos.mutual" 10 $ "setting " ++ prettyShow q ++ " to non-recursive"
+      -- Andreas, 2017-04-26, issue #2555
+      -- We should not have @DefNode@s pointing outside our formal mutual block.
+      unless (Set.member q qset) __IMPOSSIBLE__
+      setMutual q []
+    AcyclicSCC (ArgNode{}) -> return ()
+    CyclicSCC scc          -> setMut [ q | DefNode q <- scc ]
+  mapM_ (checkPos g gstar) qs
   reportSDoc "tc.pos.tick" 100 $ text "checked positivity"
 
   where
@@ -195,12 +209,12 @@ checkStrictlyPositive mi qset = disableDestructiveUpdate $ do
 
     -- Set the mutually recursive identifiers for a SCC.
     setMut :: [QName] -> TCM ()
-    setMut []  = return ()  -- nothing to do
-    setMut [q] = return ()  -- no mutual recursion
-    setMut qs  = forM_ qs $ \ q -> setMutual q (delete q qs)
-      -- TODO: The previous line is at least quadratic in the length
-      -- of qs (assuming that the expression "delete q qs" is always
-      -- forced, for instance due to serialisation). Presumably qs is
+    setMut [] = return ()  -- nothing to do
+    setMut qs = forM_ qs $ \ q -> do
+      reportSLn "tc.pos.mutual" 10 $ "setting " ++ prettyShow q ++ " to (mutually) recursive"
+      setMutual q qs
+      -- TODO: The previous line produces data of quadratic size
+      -- (which has to be processed upon serialization).  Presumably qs is
       -- usually short, but in some cases (for instance for generated
       -- code) it may be long. Wouldn't it be better to assign a
       -- unique identifier to each SCC, and avoid storing lists?
@@ -221,7 +235,7 @@ checkStrictlyPositive mi qset = disableDestructiveUpdate $ do
               map findOcc [0 .. max m (n - 1)]
         reportSDoc "tc.pos.args" 10 $ sep
           [ text "args of" <+> prettyTCM q <+> text "="
-          , nest 2 $ prettyList $ map (text . show) args
+          , nest 2 $ prettyList $ map prettyTCM args
           ]
         -- The list args can take a long time to compute, but contains
         -- small elements, and is stored in the interface (right?), so
@@ -258,7 +272,7 @@ instance PrettyTCM OccursWhere where
       nth n = pwords $ show (n + 1) ++ "th"
 
       -- remove consecutive duplicates
-      uniq = map head . group
+      uniq = map head . List.group
 
       prettyOs [] = __IMPOSSIBLE__
       prettyOs [o] = prettyO o <> text "."
@@ -564,63 +578,44 @@ computeOccurrences' q = inConcreteOrAbstractMode q $ \ def -> do
       <+> prettyTCM cur
   OccursAs (InDefOf q) <$> case theDef def of
     Function{funClauses = cs} -> do
-      n  <- getDefArity def
-      cs <- map (etaExpandClause n) <$> instantiateFull cs
+      cs <- mapM etaExpandClause =<< instantiateFull cs
       Concat . zipWith (OccursAs . InClause) [0..] <$>
         mapM (getOccurrences []) cs
     Datatype{dataClause = Just c} -> getOccurrences [] =<< instantiateFull c
-    Datatype{dataPars = np, dataCons = cs}       -> do
+    Datatype{dataPars = np0, dataCons = cs}       -> do
       -- Andreas, 2013-02-27: first, each data index occurs as matched on.
       TelV tel t <- telView $ defType def
-      let xs  = [np .. size tel - 1] -- argument positions corresponding to indices
-          ioccs = Concat $ map (OccursAs Matched . OccursHere . AnArg) xs
+      -- Andreas, 2017-04-26, issue #2554: count first index as parameter if it has type Size.
+      -- We compute sizeIndex=1 if first first index has type Size, otherwise sizeIndex==0
+      sizeIndex <- caseMaybe (headMaybe $ drop np0 $ telToList tel) (return 0) $ \ dom -> do
+        caseMaybeM (isSizeType dom) (return 0) $ \ _ -> return 1
+      let np = np0 + sizeIndex
+      let xs = [np .. size tel - 1] -- argument positions corresponding to indices
+          ioccs = Concat $ map (OccursHere . AnArg) [np0 .. np - 1]
+                        ++ map (OccursAs Matched . OccursHere . AnArg) xs
       -- Then, we compute the occurrences in the constructor types.
       let conOcc c = do
             a <- defType <$> getConstInfo c
             TelV tel t <- telView' <$> normalise a -- normalization needed e.g. for test/succeed/Bush.agda
             let indices = case unEl t of
-                            Def _ vs -> genericDrop np vs
+                            Def _ vs -> drop np vs
                             _        -> __IMPOSSIBLE__
-            let tel'    = telFromList $ genericDrop np $ telToList tel
+            let tel'    = telFromList $ drop np $ telToList tel
                 vars np = map (Just . AnArg) $ downFrom np
             (>+<) <$> (OccursAs (ConArgType c) <$> getOccurrences (vars np) tel')
                   <*> (OccursAs (IndArgType c) . OnlyVarsUpTo np <$> getOccurrences (vars $ size tel) indices)
       (>+<) ioccs <$> (Concat <$> mapM conOcc cs)
     Record{recClause = Just c} -> getOccurrences [] =<< instantiateFull c
     Record{recPars = np, recTel = tel} -> do
-      let tel' = telFromList $ genericDrop np $ telToList tel
+      let tel' = telFromList $ drop np $ telToList tel
           vars = map (Just . AnArg) $ downFrom np
-      getOccurrences vars =<< instantiateFull tel'
+      getOccurrences vars =<< normalise tel' -- Andreas, 2017-01-01, issue #1899, treat like data types
 
     -- Arguments to other kinds of definitions are hard-wired.
     Constructor{} -> return emptyOB
     Axiom{}       -> return emptyOB
     Primitive{}   -> return emptyOB
-    AbstractDefn  -> __IMPOSSIBLE__
-
--- | Eta expand a clause to have the given number of variables.
---   Warning: doesn't put correct types in telescope!
---   This is used instead of special treatment of lambdas
---   (which was unsound: issue 121)
-etaExpandClause :: Nat -> Clause -> Clause
-etaExpandClause n c
-  | m <= 0    = c
-  | otherwise = c
-      { namedClausePats = raise m (namedClausePats c) ++
-                          map (\i -> defaultArg $ namedDBVarP i underscore) (downFrom m)
-      , clauseBody      = liftBody m $ clauseBody c
-      , clauseTel       = telFromList $
-          telToList (clauseTel c) ++ (replicate m $ (underscore,) <$> dummyDom)
-          -- dummyDom, not __IMPOSSIBLE__, because of debug printing.
-      }
-  where
-    m = n - genericLength (namedClausePats c)
-
-    vars = map (defaultArg . var) $ downFrom m
---    vars = reverse [ defaultArg $ var i | i <- [0..m - 1] ]
-
-    liftBody m (Just v) = Just $ raise m v `apply` vars
-    liftBody m Nothing  = Nothing
+    AbstractDefn{}-> __IMPOSSIBLE__
 
 -- Building the occurrence graph ------------------------------------------
 
@@ -628,17 +623,19 @@ data Node = DefNode !QName
           | ArgNode !QName !Nat
   deriving (Eq, Ord)
 
-instance Show Node where
-  show (DefNode q)   = show q
-  show (ArgNode q i) = show q ++ "." ++ show i
+instance Pretty Node where
+  pretty = \case
+    DefNode q   -> P.pretty q
+    ArgNode q i -> P.pretty q P.<> P.text ("." ++ show i)
 
 instance PrettyTCM Node where
-  prettyTCM (DefNode q)   = prettyTCM q
-  prettyTCM (ArgNode q i) = prettyTCM q <> text ("." ++ show i)
+  prettyTCM = return . P.pretty
 
 instance PrettyTCM n => PrettyTCM (WithNode n Edge) where
-  prettyTCM (WithNode n (Edge o w)) =
-    prettyTCM o <+> prettyTCM n <+> fsep (pwords $ show w)
+  prettyTCM (WithNode n (Edge o w)) = vcat
+    [ prettyTCM o <+> prettyTCM n
+    , nest 2 $ return $ P.pretty w
+    ]
 
 -- | Edge labels for the positivity graph.
 data Edge = Edge !Occurrence OccursWhere
@@ -698,7 +695,7 @@ buildOccurrenceGraph qs =
                    text (show i) <> text ":" <+> text (show n) <+>
                    text "occurrences, of total size" <+>
                    text (show s)) $
-           sortBy (compare `on` fst . snd) $
+           List.sortBy (compare `on` fst . snd) $
            map (\(i, os) -> (i, (length os, sum $ map size os))) $
            Map.toList (flatten occs))
       reportSDoc "tc.pos.occs" 50 $
@@ -706,7 +703,7 @@ buildOccurrenceGraph qs =
            map (\(i, os) ->
                    (text (show i) <> text ":")
                      $+$
-                   (nest 2 $ vcat $ map (text . show) os))
+                   (nest 2 $ vcat $ map (return . P.pretty) os))
                (Map.toList (flatten occs)))
 
       -- Placing this line before the reportSDoc lines above creates a
@@ -720,8 +717,8 @@ buildOccurrenceGraph qs =
            map (\e ->
                    let Edge o w = Graph.label e in
                    prettyTCM (Graph.source e) <+>
-                   text "-[" <+> text (show o) <> text "," <+>
-                                 text (show w) <+> text "]->" <+>
+                   text "-[" <+> return (P.pretty o) <> text "," <+>
+                                 return (P.pretty w) <+> text "]->" <+>
                    prettyTCM (Graph.target e))
                es)
 
@@ -762,16 +759,20 @@ computeEdges muts q ob =
                          [ mkEdge to pol ob | ob <- obs ]
     OccursAs' w ob  -> do (to, pol) <- mkEdge' to pol w
                           mkEdge to pol ob
-    OccursHere' i o -> return $
-                         if null pol
-                         then id
-                         else (Graph.Edge
-                                 { Graph.source = case i of
-                                                    AnArg i -> ArgNode q i
-                                                    ADef q  -> DefNode q
-                                 , Graph.target = to
-                                 , Graph.label  = Edge pol o
-                                 } :)
+    OccursHere' (AnArg i) o ->
+      return $ applyUnless (null pol) (Graph.Edge
+        { Graph.source = ArgNode q i
+        , Graph.target = to
+        , Graph.label  = Edge pol o
+        } :)
+    OccursHere' (ADef q') o ->
+      -- Andreas, 2017-04-26, issue #2555
+      -- Skip nodes pointing outside the mutual block.
+      return $ applyUnless (null pol || Set.notMember q' muts) (Graph.Edge
+        { Graph.source = DefNode q'
+        , Graph.target = to
+        , Graph.label  = Edge pol o
+        } :)
 
   mkEdge' to !pol w = case w of
     VarArg         -> mixed
