@@ -3,6 +3,8 @@
 module Agda.TypeChecking.Quote where
 
 import Control.Applicative
+import Control.Arrow ((&&&))
+import Control.Monad
 import Control.Monad.State (runState, get, put)
 import Control.Monad.Reader (asks)
 import Control.Monad.Writer (execWriterT, tell)
@@ -20,13 +22,11 @@ import Agda.Syntax.Position
 import Agda.Syntax.Translation.InternalToAbstract
 
 import Agda.TypeChecking.CompiledClause
-import Agda.TypeChecking.Datatypes ( getConHead )
 import Agda.TypeChecking.DropArgs
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Exception
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
@@ -116,19 +116,18 @@ quotingKit = do
       t !@! u = pure t @@ pure u
 
       quoteHiding :: Hiding -> ReduceM Term
-      quoteHiding Hidden    = pure hidden
-      quoteHiding Instance  = pure instanceH
-      quoteHiding NotHidden = pure visible
+      quoteHiding Hidden     = pure hidden
+      quoteHiding Instance{} = pure instanceH
+      quoteHiding NotHidden  = pure visible
 
       quoteRelevance :: Relevance -> ReduceM Term
       quoteRelevance Relevant   = pure relevant
       quoteRelevance Irrelevant = pure irrelevant
       quoteRelevance NonStrict  = pure relevant
       quoteRelevance Forced{}   = pure relevant
-      quoteRelevance UnusedArg  = pure relevant
 
       quoteArgInfo :: ArgInfo -> ReduceM Term
-      quoteArgInfo (ArgInfo h r _ _) = arginfo !@ quoteHiding h @@ quoteRelevance r
+      quoteArgInfo (ArgInfo h r _) = arginfo !@ quoteHiding h @@ quoteRelevance r
 
       quoteLit :: Literal -> ReduceM Term
       quoteLit l@LitNat{}    = litNat    !@! Lit l
@@ -162,7 +161,7 @@ quotingKit = do
       quotePats ps = list $ map (quoteArg quotePat . fmap namedThing) ps
 
       quotePat :: DeBruijnPattern -> ReduceM Term
-      quotePat (VarP x) | isAbsurdPatternName (dbPatVarName x) = pure absurdP
+      quotePat (AbsurdP x)       = pure absurdP
       quotePat (VarP x)          = varP !@! quoteString (dbPatVarName x)
       quotePat (DotP _)          = pure dotP
       quotePat (ConP c _ ps)     = conP !@ quoteQName (conName c) @@ quotePats ps
@@ -206,18 +205,26 @@ quotingKit = do
              in  var !@! Lit (LitNat noRange $ fromIntegral n) @@ quoteArgs ts
           Lam info t -> lam !@ quoteHiding (getHiding info) @@ quoteAbs quoteTerm t
           Def x es   -> do
-            def <- getConstInfo x
-            let d = theDef def
+            defn <- getConstInfo x
             n <- getDefFreeVars x
             -- #2220: remember to restore dropped parameters
-            qx d @@ list (drop n $ defParameters def ++ map (quoteArg quoteTerm) ts)
-            where
+            let
+              conOrProjPars = defParameters defn
               ts = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-              qx Function{ funExtLam = Just (ExtLamInfo h nh), funClauses = cs } =
-                    extlam !@ list (map (quoteClause . dropArgs (h + nh)) cs)
+              qx Function{ funExtLam = Just (ExtLamInfo h nh), funClauses = cs } = do
+                    -- An extended lambda should not have any extra parameters!
+                    unless (null conOrProjPars) __IMPOSSIBLE__
+                    -- Andreas, 2017-01-23 quoting Ulf
+                    -- "One would hope that @n@ includes the @h + nh@ parameters of the ext.lam."
+                    -- Let's see!
+                    -- unless (n >= h + nh) __IMPOSSIBLE__
+                    -- Actually, no, it does not!  ExtLam is not touched by module application.
+                    -- TODO: fixe me!  See #2404.
+                    extlam !@ list (map (quoteClause . (`apply` (take (h + nh) ts))) cs)
               qx Function{ funCompiled = Just Fail, funClauses = [cl] } =
                     extlam !@ list [quoteClause $ dropArgs (length (namedClausePats cl) - 1) cl]
               qx _ = def !@! quoteName x
+            qx (theDef defn) @@ list (drop n $ conOrProjPars ++ map (quoteArg quoteTerm) ts)
           Con x ci ts -> do
             cDef <- getConstInfo (conName x)
             n    <- getDefFreeVars (conName x)
@@ -241,8 +248,8 @@ quotingKit = do
                  Function{ funProjection = Just p } -> projIndex p - 1
                  _                                  -> 0
           TelV tel _ = telView' (defType def)
-          hiding     = map getHiding $ take np $ telToList tel
-          par h      = arg !@ (arginfo !@ quoteHiding h @@ pure relevant) @@ pure unsupported
+          hiding     = map (getHiding &&& getRelevance) $ take np $ telToList tel
+          par (h, r) = arg !@ (arginfo !@ quoteHiding h @@ quoteRelevance r) @@ pure unsupported
 
       quoteDefn :: Definition -> ReduceM Term
       quoteDefn def =
@@ -251,8 +258,10 @@ quotingKit = do
             agdaDefinitionFunDef !@ quoteList quoteClause cs
           Datatype{dataPars = np, dataCons = cs} ->
             agdaDefinitionDataDef !@! quoteNat (fromIntegral np) @@ quoteList (pure . quoteName) cs
-          Record{recConHead = c} ->
-            agdaDefinitionRecordDef !@! quoteName (conName c)
+          Record{recConHead = c, recFields = fs} ->
+            agdaDefinitionRecordDef
+              !@! quoteName (conName c)
+              @@ quoteList (quoteArg (pure . quoteName)) fs
           Axiom{}       -> pure agdaDefinitionPostulate
           AbstractDefn{}-> pure agdaDefinitionPostulate
           Primitive{primClauses = cs} | not $ null cs ->
