@@ -17,7 +17,7 @@ import Control.Monad.Fix
 import Control.Monad.Reader
 import Control.Monad.State
 
-import Data.List
+import qualified Data.List as List
 import Data.Maybe
 import qualified Data.Map as Map
 import qualified Data.Traversable as Trav
@@ -27,6 +27,7 @@ import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern as I
 import Agda.TypeChecking.CompiledClause
 import Agda.TypeChecking.Coverage.SplitTree
+import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Datatypes
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Pretty hiding (pretty)
@@ -64,6 +65,7 @@ recordPatternToProjections p =
     VarP{}       -> return [ \ x -> x ]
     LitP{}       -> typeError $ ShouldBeRecordPattern p
     DotP{}       -> typeError $ ShouldBeRecordPattern p
+    AbsurdP{}    -> typeError $ ShouldBeRecordPattern p
     ConP c ci ps -> do
       whenNothing (conPRecord ci) $
         typeError $ ShouldBeRecordPattern p
@@ -114,13 +116,22 @@ cutSublist i n row =
 
 getEtaAndArity :: QName -> TCM (Bool, Nat)
 getEtaAndArity c =
-  for (getConstructorInfo c) $ \i ->
-  case i of
+  for (getConstructorInfo c) $ \case
     DataCon n        -> (False, n)
     RecordCon eta fs -> (eta == YesEta, size fs)
 
 translateCompiledClauses :: CompiledClauses -> TCM CompiledClauses
-translateCompiledClauses cc = snd <$> loop cc
+translateCompiledClauses cc = do
+  reportSDoc "tc.cc.record" 20 $ vcat
+    [ text "translate record patterns in compiled clauses"
+    , nest 2 $ return $ pretty cc
+    ]
+  cc <- snd <$> loop cc
+  reportSDoc "tc.cc.record" 20 $ vcat
+    [ text "translated compiled clauses (no eta record patterns):"
+    , nest 2 $ return $ pretty cc
+    ]
+  return cc
   where
 
     loop :: CompiledClauses -> TCM ([Bool], CompiledClauses)
@@ -139,7 +150,7 @@ translateCompiledClauses cc = snd <$> loop cc
 
       -- recurse on and compute variable status of catch-all clause
       (xssa, catchAll) <- unzipMaybe <$> Trav.mapM loop catchAll
-      let xsa = maybe (repeat True) id xssa
+      let xsa = fromMaybe (repeat True) xssa
 
       -- recurse on compute variable status of literal clauses
       (xssl, litMap)   <- Map.unzip <$> Trav.mapM loop litMap
@@ -160,8 +171,8 @@ translateCompiledClauses cc = snd <$> loop cc
                   RecordCon YesEta fs -> return $ Right fs
               Just{}  -> return $ Left 0
           let (isRC, n)   = either (False,) ((True,) . size) dataOrRecCon
-              (xs0, rest) = genericSplitAt (unArg i) xs
-              (xs1, xs2 ) = genericSplitAt n rest
+              (xs0, rest) = splitAt (unArg i) xs
+              (xs1, xs2 ) = splitAt n rest
               -- if all dropped variables (xs1) are virgins and we are record cons.
               -- then new variable x is also virgin
               -- and we can translate away the split
@@ -171,7 +182,8 @@ translateCompiledClauses cc = snd <$> loop cc
               -- get the record fields
               fs          = either __IMPOSSIBLE__ id dataOrRecCon
               -- if x we can translate
-              mcc = if x then [replaceByProjections i (map unArg fs) cc] else []
+          mcc <- if x then etaContract [replaceByProjections i (map unArg fs) cc]
+                      else return []
 
           when (n /= ar) __IMPOSSIBLE__
           return (mcc, xs', WithArity ar cc)
@@ -317,8 +329,8 @@ recordSplitTree t = snd <$> loop t
         forM ts $ \ (c, t) -> do
           (xs, t) <- loop t
           (isRC, n) <- getEtaAndArity c
-          let (xs0, rest) = genericSplitAt i xs
-              (xs1, xs2)  = genericSplitAt n rest
+          let (xs0, rest) = splitAt i xs
+              (xs1, xs2)  = splitAt n rest
               x           = isRC && and xs1
               xs'         = xs0 ++ x : xs2
           return (xs, (RecordSplitNode c n x, t))
@@ -360,8 +372,8 @@ translateSplitTree t = snd <$> loop t
           (xs, t) <- loop t
           (isRC, n) <- getEtaAndArity c
           -- now drop variables from i to i+n-1
-          let (xs0, rest) = genericSplitAt i xs
-              (xs1, xs2)  = genericSplitAt n rest
+          let (xs0, rest) = splitAt i xs
+              (xs1, xs2)  = splitAt n rest
               -- if all dropped variables are virgins and we are record cons.
               -- then new variable x is also virgin
               -- and we can translate away the split
@@ -484,7 +496,7 @@ translateRecordPatterns clause = do
         -- It is important that dummyDom does not mention any variable
         -- (see the definition of reorderTel).
         where
-        isDotP n = case genericIndex cs n of
+        isDotP n = case List.genericIndex cs n of
                      Left DotP{} -> True
                      _           -> False
 
@@ -523,8 +535,8 @@ translateRecordPatterns clause = do
       , text "Intermediate results:"
       , nest 2 $ vcat
         [ text "ps        =" <+> text (show ps)
-        , text "s         =" <+> prettyList (map prettyTCM s)
-        , text "cs        =" <+> prettyList (map prettyTCM cs)
+        , text "s         =" <+> prettyTCM s
+        , text "cs        =" <+> prettyTCM cs
         , text "flattenedOldTel =" <+> (text . show) flattenedOldTel
         , text "newTel'   =" <+> (text . show) newTel'
         , text "newPerm   =" <+> prettyTCM newPerm
@@ -562,7 +574,7 @@ translateRecordPatterns clause = do
 
 newtype RecPatM a = RecPatM (TCMT (ReaderT Nat (StateT Nat IO)) a)
   deriving (Functor, Applicative, Monad,
-            MonadIO, MonadTCM,
+            MonadIO, MonadTCM, HasOptions, MonadDebug,
             MonadReader TCEnv, MonadState TCState)
 
 -- | Runs a computation in the 'RecPatM' monad.
@@ -652,7 +664,7 @@ removeTree tree = do
   let ps = projections tree
       s  = map (\(p, _) -> p x) ps
 
-      count k = genericLength $ filter ((== k) . snd) ps
+      count k = length $ filter ((== k) . snd) ps
 
   return $ case tree of
     Leaf p     -> (p,   s, [Left p])
@@ -697,6 +709,7 @@ translatePattern p@(ConP c ci ps)
       return (ConP c ci ps, s, cs)
 translatePattern p@VarP{} = removeTree (Leaf p)
 translatePattern p@DotP{} = removeTree (Leaf p)
+translatePattern p@AbsurdP{} = removeTree (Leaf p)
 translatePattern p@LitP{} = return (p, [], [])
 translatePattern p@ProjP{}= return (p, [], [])
 
@@ -741,6 +754,7 @@ recordTree (ConP c ci ps) | Just ConOSystem <- conPRecord ci = do
 recordTree p@(ConP _ ci _) = return $ Left $ translatePattern p
 recordTree p@VarP{} = return (Right (Leaf p))
 recordTree p@DotP{} = return (Right (Leaf p))
+recordTree p@AbsurdP{} = return (Right (Leaf p))
 recordTree p@LitP{} = return $ Left $ translatePattern p
 recordTree p@ProjP{}= return $ Left $ translatePattern p
 
@@ -762,7 +776,7 @@ translateTel
 translateTel (Left (DotP{}) : rest)   tel = Nothing : translateTel rest tel
 translateTel (Right (n, x, t) : rest) tel = Just (x, t) :
                                               translateTel rest
-                                                (genericDrop (n VarPat) tel)
+                                                (drop (n VarPat) tel)
 translateTel (Left _ : rest) (t : tel)    = Just t : translateTel rest tel
 translateTel []              []           = []
 translateTel (Left _ : _)    []           = __IMPOSSIBLE__
