@@ -7,7 +7,7 @@ module Agda.TypeChecking.Records where
 import Control.Monad
 
 import Data.Function
-import Data.List
+import qualified Data.List as List
 import Data.Maybe
 import qualified Data.Set as Set
 import Data.Traversable (traverse)
@@ -46,9 +46,9 @@ import Agda.Utils.Impossible
 --   Use the second argument for missing fields.
 orderFields :: QName -> a -> [C.Name] -> [(C.Name, a)] -> TCM [a]
 orderFields r def xs fs = do
-  unlessNull (ys \\ nub ys) $ typeError . DuplicateFields . nub
-  unlessNull (ys \\ xs)     $ typeError . TooManyFields r
-  -- shouldBeNull (xs \\ ys)     $ TooFewFields r
+  unlessNull (ys List.\\ List.nub ys) $ typeError . DuplicateFields . List.nub
+  unlessNull (ys List.\\ xs)          $ typeError . TooManyFields r
+  -- shouldBeNull (xs List.\\ ys)     $ TooFewFields r
   return $ order xs fs
   where
     ys = map fst fs
@@ -73,18 +73,16 @@ insertMissingFields
   -> [Arg C.Name]         -- ^ All record field names with 'ArgInfo'.
   -> TCM [NamedArg a]     -- ^ Given fields enriched by placeholders for missing explicit fields.
 insertMissingFields r placeholder fs axs = do
-  let -- Just field names.
-      xs   = map unArg axs
   -- Compute the list of given fields, decorated with the ArgInfo from the record def.
   let arg x e =
         case [ a | a <- axs, unArg a == x ] of
-          [a] -> unnamed e <$ a
+          [a] -> nameIfHidden a e <$ a
           _   -> defaultNamedArg e -- we only end up here if the field names are bad
       givenFields = [ (x, Just $ arg x e) | FieldAssignment x e <- fs ]
   -- Compute a list of p[aceholders for the missing visible fields.
   let missingExplicits =
-       [ (x, Just $ setOrigin Inserted $ unnamed . placeholder <$> a)
-       | a <- filter notHidden axs
+       [ (x, Just $ setOrigin Inserted $ nameIfHidden a . placeholder <$> a)
+       | a <- filter visible axs
        , let x = unArg a
        , x `notElem` map (view nameFieldA) fs
        ]
@@ -94,8 +92,15 @@ insertMissingFields r placeholder fs axs = do
   catMaybes <$> do
     -- Default value @Nothing@ will only be used for missing hidden fields.
     -- These can be ignored as they will be inserted by @checkArguments_@.
-    orderFields r Nothing xs $ givenFields ++ missingExplicits
-
+    orderFields r Nothing (map unArg axs) $ givenFields ++ missingExplicits
+  where
+    -- Andreas, 2017-04-13, issue #2494
+    -- We need to put the field names as argument names for hidden arguments.
+    -- Otherwise, insertImplicit does not do the right thing.
+    nameIfHidden :: Arg C.Name -> c -> Named_ c
+    nameIfHidden ax
+      | visible ax = unnamed
+      | otherwise  = named (Ranged (getRange ax) $ prettyShow $ unArg ax)
 
 -- | The name of the module corresponding to a record.
 recordModule :: QName -> ModuleName
@@ -111,7 +116,7 @@ getRecordDef r = maybe err return =<< isRecord r
 getRecordOfField :: QName -> TCM (Maybe QName)
 getRecordOfField d = caseMaybeM (isProjection d) (return Nothing) $
   \ Projection{ projProper = proper, projFromType = r} ->
-    return $ if proper then Just (unArg r) else Nothing
+    return $ unArg r <$ proper -- if proper then Just (unArg r) else Nothing
 
 -- | Get the field names of a record.
 getRecordFieldNames :: QName -> TCM [Arg C.Name]
@@ -196,7 +201,7 @@ origProjection f = do
       fallback = return (f, def, proj)
   caseMaybe proj fallback $
     \ p@Projection{ projProper = proper, projOrig = f' } ->
-      if not proper || f == f' then fallback else do
+      if isNothing proper || f == f' then fallback else do
         def <- getConstInfo f'
         return (f', def, isProjection_ $ theDef def)
 
@@ -222,7 +227,7 @@ getDefType f t = do
   -- if @f@ is not a projection (like) function, @a@ is the correct type
       fallback = return $ Just a
   reportSDoc "tc.deftype" 20 $ vcat
-    [ text "definition f = " <> prettyTCM f <+> text ("raw: " ++ show f)
+    [ text "definition f = " <> prettyTCM f <+> text ("  -- raw: " ++ prettyShow f)
     , text "has type   a = " <> prettyTCM a
     , text "principal  t = " <> prettyTCM t
     ]
@@ -244,7 +249,7 @@ getDefType f t = do
             -- now we know it is reduced, we can safely take the parameters
             let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims $ take npars es
             reportSDoc "tc.deftype" 20 $ vcat
-              [ text $ "head d     = " ++ show d
+              [ text $ "head d     = " ++ prettyShow d
               , text "parameters =" <+> sep (map prettyTCM pars)
               ]
             reportSLn "tc.deftype" 60 $ "parameters = " ++ show pars
@@ -271,12 +276,46 @@ getDefType f t = do
 --   In this case, the first result is not a record type.
 --
 --   Precondition: @t@ is reduced.
-projectTyped :: Term -> Type -> ProjOrigin -> QName -> TCM (Maybe (Dom Type, Term, Type))
+--
+projectTyped
+  :: Term        -- ^ Head (record value).
+  -> Type        -- ^ Its type.
+  -> ProjOrigin
+  -> QName       -- ^ Projection.
+  -> TCM (Maybe (Dom Type, Term, Type))
 projectTyped v t o f = caseMaybeM (getDefType f t) (return Nothing) $ \ tf -> do
   ifNotPiType tf (const $ return Nothing) {- else -} $ \ dom b -> do
   u <- applyDef o f (argFromDom dom $> v)
   return $ Just (dom, u, b `absApp` v)
 
+-- | Typing of an elimination.
+
+data ElimType
+  = ArgT (Dom Type)           -- ^ Type of the argument.
+  | ProjT
+    { projTRec   :: Dom Type  -- ^ The type of the record which is eliminated.
+    , projTField :: Type      -- ^ The type of the field.
+    }
+
+instance PrettyTCM ElimType where
+  prettyTCM (ArgT a)    = prettyTCM a
+  prettyTCM (ProjT a b) =
+    text "." <> parens (prettyTCM a <+> text "->" <+> prettyTCM b)
+
+-- | Given a head and its type, compute the types of the eliminations.
+
+typeElims :: Type -> Term -> Elims -> TCM [ElimType]
+typeElims a _ [] = return []
+typeElims a self (e : es) = do
+  case e of
+    Apply v -> ifNotPiType a __IMPOSSIBLE__ {- else -} $ \ a b -> do
+      (ArgT a :) <$> typeElims (absApp b $ unArg v) (self `applyE` [e]) es
+    Proj o f -> do
+      a <- reduce a
+      (dom, self, a) <- fromMaybe __IMPOSSIBLE__ <$> projectTyped self a o f
+      (ProjT dom a :) <$> typeElims a self es
+
+
 -- | Check if a name refers to an eta expandable record.
 {-# SPECIALIZE isEtaRecord :: QName -> TCM Bool #-}
 {-# SPECIALIZE isEtaRecord :: QName -> ReduceM Bool #-}
@@ -284,9 +323,10 @@ isEtaRecord :: HasConstInfo m => QName -> m Bool
 isEtaRecord r = maybe False recEtaEquality <$> isRecord r
 
 isEtaCon :: HasConstInfo m => QName -> m Bool
-isEtaCon c = do
-  cdef <- theDef <$> getConstInfo c
-  case cdef of
+isEtaCon c = getConstInfo' c >>= \case
+  Left (SigUnknown err) -> __IMPOSSIBLE__
+  Left SigAbstract -> return False
+  Right def -> case theDef def of
     Constructor {conData = r} -> isEtaRecord r
     _ -> return False
 
@@ -305,39 +345,39 @@ isEtaRecordType a = case ignoreSharing $ unEl a of
 -- | Check if a name refers to a record constructor.
 --   If yes, return record definition.
 isRecordConstructor :: HasConstInfo m => QName -> m (Maybe (QName, Defn))
-isRecordConstructor c = do
-  def <- theDef <$> getConstInfo c
-  case def of
+isRecordConstructor c = getConstInfo' c >>= \case
+  Left (SigUnknown err)        -> __IMPOSSIBLE__
+  Left SigAbstract             -> return Nothing
+  Right def -> case theDef $ def of
     Constructor{ conData = r } -> fmap (r,) <$> isRecord r
     _                          -> return Nothing
 
 -- | Check if a constructor name is the internally generated record constructor.
+--
+--   Works also for abstract constructors.
 isGeneratedRecordConstructor :: QName -> TCM Bool
-isGeneratedRecordConstructor c = do
-  def <- theDef <$> getConstInfo c
-  case def of
-    Constructor{ conData = r } -> do
-      def <- theDef <$> getConstInfo r
-      case def of
-        Record{ recNamedCon = False } -> return True
-        _                             -> return False
-    _ -> return False
+isGeneratedRecordConstructor c = ignoreAbstractMode $ do
+  caseMaybeM (isRecordConstructor c) (return False) $ \ (_, def) ->
+    case def of
+      Record{ recNamedCon = False } -> return True
+      _ -> return False
 
--- | Mark record type as unguarded.
---   No eta-expansion.  Projections do not preserve guardedness.
+
+-- | Turn off eta for unguarded recursive records.
+--   Projections do not preserve guardedness.
 unguardedRecord :: QName -> TCM ()
 unguardedRecord q = modifySignature $ updateDefinition q $ updateTheDef $ \case
-  r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) False, recRecursive = True }
+  r@Record{} -> r { recEtaEquality' = setEtaEquality (recEtaEquality' r) False }
   _ -> __IMPOSSIBLE__
 
--- | Mark record type as recursive.
+-- | Turn on eta for inductive guarded recursive records.
 --   Projections do not preserve guardedness.
 recursiveRecord :: QName -> TCM ()
 recursiveRecord q = do
   ok <- etaEnabled
   modifySignature $ updateDefinition q $ updateTheDef $ \case
     r@Record{ recInduction = ind, recEtaEquality' = eta } ->
-      r { recRecursive = True, recEtaEquality' = eta' }
+      r { recEtaEquality' = eta' }
       where
       eta' | ok, eta == Inferred False, ind /= Just CoInductive = Inferred True
            | otherwise = eta
@@ -359,12 +399,7 @@ nonRecursiveRecord q = whenM etaEnabled $ do
 --
 --   Precondition: record type identifier exists in signature.
 isRecursiveRecord :: QName -> TCM Bool
-isRecursiveRecord q = recRecursive_ . theDef . fromMaybe __IMPOSSIBLE__ . lookupDefinition q <$> getSignature
-
--- | Version of @recRecursive@ with proper internal error.
-recRecursive_ :: Defn -> Bool
-recRecursive_ (Record { recRecursive = b }) = b
-recRecursive_ _ = __IMPOSSIBLE__
+isRecursiveRecord q = recRecursive . theDef . fromMaybe __IMPOSSIBLE__ . lookupDefinition q <$> getSignature
 
 {- | @etaExpandBoundVar i = (Δ, σ, τ)@
 
@@ -431,7 +466,11 @@ expandRecordVar i gamma0 = do
           -- Use "f(x)" as variable name for the projection f(x).
           s     = prettyShow x
           tel'  = mapAbsNames (\ f -> stringToArgName $ argNameToString f ++ "(" ++ s ++ ")") tel
-          delta = telFromList $ gamma1 ++ telToList tel' ++ applySubst tau0 gamma2
+          delta = telFromList $ gamma1 ++ telToList tel' ++
+                    telToList (applySubst tau0 $ telFromList gamma2)
+                    -- Andreas, 2017-07-29, issue #2644
+                    -- We cannot substitute directly into a ListTel like gamma2,
+                    -- we have to convert it to a telescope first, otherwise we get garbage.
 
       return (delta, sigma, tau, tel)
 
@@ -533,8 +572,8 @@ etaExpandRecord'_ forceEta r pars def u = do
       when (con /= con_) $ do
         reportSDoc "impossible" 10 $ vcat
           [ text "etaExpandRecord_: the following two constructors should be identical"
-          , nest 2 $ text $ "con  = " ++ show con
-          , nest 2 $ text $ "con_ = " ++ show con_
+          , nest 2 $ text $ "con  = " ++ prettyShow con
+          , nest 2 $ text $ "con_ = " ++ prettyShow con_
           ]
         __IMPOSSIBLE__
       return (tel', con, ci, args)
@@ -614,7 +653,7 @@ isSingletonRecordModuloRelevance r ps = mapRight isJust <$> isSingletonRecord' T
 --   contains garbage.
 isSingletonRecord' :: Bool -> QName -> Args -> TCM (Either MetaId (Maybe Term))
 isSingletonRecord' regardIrrelevance r ps = do
-  reportSLn "tc.meta.eta" 30 $ "Is " ++ show r ++ " a singleton record type?"
+  reportSLn "tc.meta.eta" 30 $ "Is " ++ prettyShow r ++ " a singleton record type?"
   def <- getRecordDef r
   emap (Con (recConHead def) ConOSystem) <$> check (recTel def `apply` ps)
   where
@@ -663,6 +702,8 @@ isSingletonType' regardIrrelevance t = do
 emap :: (a -> b) -> Either c (Maybe a) -> Either c (Maybe b)
 emap = mapRight . fmap
 
+-- | Replace projection patterns by the original projections.
+--
 class NormaliseProjP a where
   normaliseProjP :: HasConstInfo m => a -> m a
 
@@ -683,6 +724,7 @@ instance NormaliseProjP a => NormaliseProjP (Named_ a) where
 instance NormaliseProjP (Pattern' x) where
   normaliseProjP p@VarP{}        = return p
   normaliseProjP p@DotP{}        = return p
+  normaliseProjP p@AbsurdP{}     = return p
   normaliseProjP (ConP c cpi ps) = ConP c cpi <$> normaliseProjP ps
   normaliseProjP p@LitP{}        = return p
   normaliseProjP (ProjP o d0)    = ProjP o <$> getOriginalProjection d0
