@@ -8,7 +8,7 @@ import Prelude hiding (mapM)
 import Control.Monad.Reader hiding (mapM)
 import Control.Applicative
 
-import Data.List as List hiding (sort)
+import qualified Data.List as List
 import Data.Maybe
 import Data.Map (Map)
 import Data.Traversable
@@ -73,7 +73,7 @@ isFullyInstantiatedMeta :: MetaId -> TCM Bool
 isFullyInstantiatedMeta m = do
   mv <- TCM.lookupMeta m
   case mvInstantiation mv of
-    InstV sub v -> null . allMetas <$> instantiateFull (sub, v)
+    InstV _tel v -> null . allMetas <$> instantiateFull v
     _ -> return False
 
 -- | Instantiate something.
@@ -207,7 +207,7 @@ instance Instantiate EqualityView where
   instantiate' (EqualityType s eq l t a b) = EqualityType
     <$> instantiate' s
     <*> return eq
-    <*> instantiate' l
+    <*> mapM instantiate' l
     <*> instantiate' t
     <*> instantiate' a
     <*> instantiate' b
@@ -327,7 +327,7 @@ maybeFastReduceTerm v = do
                  else do
     s <- optSharing   <$> commandLineOptions
     allowed <- asks envAllowedReductions
-    let notAll = delete NonTerminatingReductions allowed /= allReductions
+    let notAll = List.delete NonTerminatingReductions allowed /= allReductions
     if s || notAll then slowReduceTerm v else fastReduce (elem NonTerminatingReductions allowed) v
 
 slowReduceTerm :: Term -> ReduceM (Blocked Term)
@@ -363,7 +363,7 @@ slowReduceTerm v = do
         case v of
           _ | Just v == mz  -> return $ Lit $ LitNat (getRange c) 0
           _                 -> return v
-      reduceNat v@(Con c ci [a]) | notHidden a && isRelevant a = do
+      reduceNat v@(Con c ci [a]) | visible a && isRelevant a = do
         ms  <- fmap ignoreSharing <$> getBuiltin' builtinSuc
         case v of
           _ | Just (Con c ci []) == ms -> inc <$> reduce' (unArg a)
@@ -422,7 +422,7 @@ unfoldDefinitionStep :: Bool -> Term -> QName -> Elims -> ReduceM (Reduced (Bloc
 unfoldDefinitionStep unfoldDelayed v0 f es =
   {-# SCC "reduceDef" #-} do
   info <- getConstInfo f
-  rewr <- getRewriteRulesFor f
+  rewr <- instantiateRewriteRules =<< getRewriteRulesFor f
   allowed <- asks envAllowedReductions
   let def = theDef info
       v   = v0 `applyE` es
@@ -432,6 +432,7 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
       -- are not unfolded unless explicitely permitted.
       dontUnfold =
         (defNonterminating info && notElem NonTerminatingReductions allowed)
+        || (defTerminationUnconfirmed info && notElem UnconfirmedReductions allowed)
         || (defDelayed info == Delayed && not unfoldDelayed)
       copatterns =
         case def of
@@ -447,10 +448,11 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
                              cls (defCompiled info) rewr
         else noReduction $ notBlocked v
     _  -> do
-      if FunctionReductions `elem` allowed ||
+      if (RecursiveReductions `elem` allowed) ||
          (isJust (isProjection_ def) && ProjectionReductions `elem` allowed) || -- includes projection-like
          (isInlineFun def && InlineReductions `elem` allowed) ||
-         (copatterns && CopatternReductions `elem` allowed)
+         (definitelyNonRecursive_ def && copatterns && CopatternReductions `elem` allowed) ||
+         (definitelyNonRecursive_ def && FunctionReductions `elem` allowed)
         then
           reduceNormalE v0 f (map notReduced es) dontUnfold
                        (defClauses info) (defCompiled info) rewr
@@ -460,10 +462,10 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
     noReduction    = return . NoReduction
     yesReduction s = return . YesReduction s
     reducePrimitive x v0 f es pf dontUnfold cls mcc rewr
-      | genericLength es < ar
+      | length es < ar
                   = noReduction $ NotBlocked Underapplied $ v0 `applyE` es -- not fully applied
       | otherwise = {-# SCC "reducePrimitive" #-} do
-          let (es1,es2) = genericSplitAt ar es
+          let (es1,es2) = splitAt ar es
               args1     = fromMaybe __IMPOSSIBLE__ $ mapM isApplyElim es1
           r <- primFunImplementation pf args1
           case r of
@@ -487,15 +489,32 @@ unfoldDefinitionStep unfoldDelayed v0 f es =
       case (def,rewr) of
         _ | dontUnfold -> defaultResult -- non-terminating or delayed
         ([],[])        -> defaultResult -- no definition for head
-        (cls,rewr)     -> appDefE_ f v0 cls mcc rewr es
-      where defaultResult = noReduction $ NotBlocked AbsurdMatch vfull
-            vfull         = v0 `applyE` map ignoreReduced es
+        (cls,rewr)     -> do
+          ev <- appDefE_ f v0 cls mcc rewr es
+          debugReduce ev
+          return ev
+      where
+      defaultResult = noReduction $ NotBlocked AbsurdMatch vfull
+      vfull         = v0 `applyE` map ignoreReduced es
+      debugReduce ev = verboseS "tc.reduce" 90 $ do
+        case ev of
+          NoReduction v -> do
+            reportSDoc "tc.reduce" 90 $ vcat
+              [ text "*** tried to reduce " <+> prettyTCM f
+              , text "    es =  " <+> sep (map (prettyTCM . ignoreReduced) es)
+              -- , text "*** tried to reduce " <+> prettyTCM vfull
+              , text "    stuck on" <+> prettyTCM (ignoreBlocking v)
+              ]
+          YesReduction _simpl v -> do
+            reportSDoc "tc.reduce"  90 $ text "*** reduced definition: " <+> prettyTCM f
+            reportSDoc "tc.reduce"  95 $ text "    result" <+> prettyTCM v
+            reportSDoc "tc.reduce" 100 $ text "    raw   " <+> text (show v)
 
 -- | Reduce a non-primitive definition if it is a copy linking to another def.
 reduceDefCopy :: QName -> Elims -> TCM (Reduced () Term)
 reduceDefCopy f es = do
   info <- TCM.getConstInfo f
-  rewr <- TCM.getRewriteRulesFor f
+  rewr <- instantiateRewriteRules =<< TCM.getRewriteRulesFor f
   if (defCopy info) then reduceDef_ info rewr f es else return $ NoReduction ()
   where
     reduceDef_ :: Definition -> RewriteRules -> QName -> Elims -> TCM (Reduced () Term)
@@ -649,7 +668,7 @@ instance Reduce EqualityView where
   reduce' (EqualityType s eq l t a b) = EqualityType
     <$> reduce' s
     <*> return eq
-    <*> reduce' l
+    <*> mapM reduce' l
     <*> reduce' t
     <*> reduce' a
     <*> reduce' b
@@ -809,7 +828,7 @@ instance Simplify EqualityView where
   simplify' (EqualityType s eq l t a b) = EqualityType
     <$> simplify' s
     <*> return eq
-    <*> simplify' l
+    <*> mapM simplify' l
     <*> simplify' t
     <*> simplify' a
     <*> simplify' b
@@ -945,6 +964,7 @@ instance Normalise a => Normalise (Pattern' a) where
     LitP _       -> return p
     ConP c mt ps -> ConP c <$> normalise' mt <*> normalise' ps
     DotP v       -> DotP <$> normalise' v
+    AbsurdP x    -> AbsurdP <$> normalise' x
     ProjP{}      -> return p
 
 instance Normalise DisplayForm where
@@ -965,7 +985,7 @@ instance Normalise EqualityView where
   normalise' (EqualityType s eq l t a b) = EqualityType
     <$> normalise' s
     <*> return eq
-    <*> normalise' l
+    <*> mapM normalise' l
     <*> normalise' t
     <*> normalise' a
     <*> normalise' b
@@ -974,9 +994,6 @@ instance Normalise EqualityView where
 -- * Full instantiation
 ---------------------------------------------------------------------------
 
--- STALE: Full instantiatiation = normalisation [ instantiate' / reduce' ]
--- How can we express this? We need higher order classes!
-
 -- | @instantiateFull'@ 'instantiate's metas everywhere (and recursively)
 --   but does not 'reduce'.
 class InstantiateFull t where
@@ -1041,7 +1058,7 @@ instance InstantiateFull Substitution where
   instantiateFull' sigma =
     case sigma of
       IdS                  -> return IdS
-      EmptyS               -> return EmptyS
+      EmptyS err           -> return $ EmptyS err
       Wk   n sigma         -> Wk   n         <$> instantiateFull' sigma
       Lift n sigma         -> Lift n         <$> instantiateFull' sigma
       Strengthen bot sigma -> Strengthen bot <$> instantiateFull' sigma
@@ -1063,6 +1080,7 @@ instance InstantiateFull DBPatVar where
 instance InstantiateFull a => InstantiateFull (Pattern' a) where
     instantiateFull' (VarP x)       = VarP <$> instantiateFull' x
     instantiateFull' (DotP t)       = DotP <$> instantiateFull' t
+    instantiateFull' (AbsurdP p)    = AbsurdP <$> instantiateFull' p
     instantiateFull' (ConP n mt ps) = ConP n <$> instantiateFull' mt <*> instantiateFull' ps
     instantiateFull' l@LitP{}       = return l
     instantiateFull' p@ProjP{}      = return p
@@ -1146,12 +1164,12 @@ instance InstantiateFull Char where
     instantiateFull' = return
 
 instance InstantiateFull Definition where
-    instantiateFull' (Defn rel x t pol occ df i c inst copy ma d) = do
+    instantiateFull' (Defn rel x t pol occ df i c inst copy ma inj d) = do
       (t, df, d) <- instantiateFull' (t, df, d)
-      return $ Defn rel x t pol occ df i c inst copy ma d
+      return $ Defn rel x t pol occ df i c inst copy ma inj d
 
 instance InstantiateFull NLPat where
-  instantiateFull' (PVar x y z) = return $ PVar x y z
+  instantiateFull' (PVar x y) = return $ PVar x y
   instantiateFull' (PWild)    = return PWild
   instantiateFull' (PDef x y) = PDef <$> instantiateFull' x <*> instantiateFull' y
   instantiateFull' (PLam x y) = PLam x <$> instantiateFull' y
@@ -1192,7 +1210,7 @@ instance InstantiateFull DisplayTerm where
 instance InstantiateFull Defn where
     instantiateFull' d = case d of
       Axiom{} -> return d
-      AbstractDefn -> return d
+      AbstractDefn d -> AbstractDefn <$> instantiateFull' d
       Function{ funClauses = cs, funCompiled = cc, funInv = inv } -> do
         (cs, cc, inv) <- instantiateFull' (cs, cc, inv)
         return $ d { funClauses = cs, funCompiled = cc, funInv = inv }
@@ -1229,26 +1247,27 @@ instance InstantiateFull CompiledClauses where
   instantiateFull' (Case n bs) = Case n <$> instantiateFull' bs
 
 instance InstantiateFull Clause where
-    instantiateFull' (Clause r tel ps b t catchall) =
-       Clause r <$> instantiateFull' tel
+    instantiateFull' (Clause rl rf tel ps b t catchall unreachable) =
+       Clause rl rf <$> instantiateFull' tel
        <*> instantiateFull' ps
        <*> instantiateFull' b
        <*> instantiateFull' t
        <*> return catchall
+       <*> return unreachable
 
 instance InstantiateFull Interface where
     instantiateFull' (Interface h ms mod scope inside
-                               sig display b hsImports hsImportsUHC hsCode highlighting pragmas patsyns) =
+                               sig display b foreignCode
+                               highlighting pragmas patsyns warnings) =
         Interface h ms mod scope inside
             <$> instantiateFull' sig
             <*> instantiateFull' display
             <*> instantiateFull' b
-            <*> return hsImports
-            <*> return hsImportsUHC
-            <*> return hsCode
+            <*> return foreignCode
             <*> return highlighting
             <*> return pragmas
             <*> return patsyns
+            <*> return warnings
 
 instance InstantiateFull a => InstantiateFull (Builtin a) where
     instantiateFull' (Builtin t) = Builtin <$> instantiateFull' t
@@ -1270,7 +1289,7 @@ instance InstantiateFull EqualityView where
   instantiateFull' (EqualityType s eq l t a b) = EqualityType
     <$> instantiateFull' s
     <*> return eq
-    <*> instantiateFull' l
+    <*> mapM instantiateFull' l
     <*> instantiateFull' t
     <*> instantiateFull' a
     <*> instantiateFull' b
