@@ -9,7 +9,6 @@ module Agda.TypeChecking.Reduce.Monad
   , getConstInfo
   , isInstantiatedMeta
   , lookupMeta
-  , traceSLn, traceSDoc
   , askR, applyWhenVerboseS
   ) where
 
@@ -31,7 +30,7 @@ import Agda.Syntax.Position
 import Agda.Syntax.Internal
 import Agda.TypeChecking.Monad hiding
   ( enterClosure, underAbstraction_, underAbstraction, addCtx, mkContextEntry,
-    isInstantiatedMeta, verboseS, reportSDoc, reportSLn, typeOfConst, lookupMeta )
+    isInstantiatedMeta, verboseS, typeOfConst, lookupMeta )
 import Agda.TypeChecking.Monad.Builtin hiding ( constructorForm )
 import Agda.TypeChecking.Substitute
 import Agda.Interaction.Options
@@ -127,12 +126,6 @@ isInstantiatedMeta i = do
     InstV{} -> True
     _       -> False
 
--- | Run a computation if a certain verbosity level is activated.
---
---   Precondition: The level must be non-negative.
-verboseS :: VerboseKey -> Int -> ReduceM () -> ReduceM ()
-verboseS k n action = whenM (hasVerbosity k n) action
-
 -- | Apply a function if a certain verbosity level is activated.
 --
 --   Precondition: The level must be non-negative.
@@ -140,43 +133,22 @@ verboseS k n action = whenM (hasVerbosity k n) action
 applyWhenVerboseS :: HasOptions m => VerboseKey -> Int -> (m a -> m a) -> m a -> m a
 applyWhenVerboseS k n f a = ifM (hasVerbosity k n) (f a) a
 
-traceSDoc :: VerboseKey -> Int -> TCM Doc -> ReduceM a -> ReduceM a
-traceSDoc k n doc = applyWhenVerboseS k n $ \ cont -> do
-  ReduceEnv env st <- askR
-  -- return $! unsafePerformIO $ do print . fst =<< runTCM env st doc
-  trace (show $ fst $ unsafePerformIO $ runTCM env st doc) cont
+instance MonadDebug ReduceM where
 
--- traceSDoc :: VerboseKey -> Int -> TCM Doc -> ReduceM a -> ReduceM a
--- traceSDoc k n doc = verboseS k n $ ReduceM $ do
---   ReduceEnv env st <- ask
---   -- return $! unsafePerformIO $ do print . fst =<< runTCM env st doc
---   trace (show $ fst $ unsafePerformIO $ runTCM env st doc) $ return ()
+  traceDebugMessage n s cont = do
+    ReduceEnv env st <- askR
+    unsafePerformIO $ do
+      _ <- runTCM env st $ displayDebugMessage n s
+      return $ cont
 
-{-# SPECIALIZE traceSLn :: VerboseKey -> Int -> String -> ReduceM a -> ReduceM a #-}
-traceSLn :: HasOptions m => VerboseKey -> Int -> String -> m a -> m a
-traceSLn k n s = applyWhenVerboseS k n (trace s)
+  formatDebugMessage k n d = do
+    ReduceEnv env st <- askR
+    unsafePerformIO $ do
+      (s , _) <- runTCM env st $ formatDebugMessage k n d
+      return $ return s
 
 instance HasConstInfo ReduceM where
   getRewriteRulesFor = defaultGetRewriteRulesFor (gets id)
-  getConstInfo q = ReduceM $ \(ReduceEnv env st) ->
-    let defs  = st^.(stSignature . sigDefinitions)
-        idefs = st^.(stImports . sigDefinitions)
-    in case catMaybes [HMap.lookup q defs, HMap.lookup q idefs] of
-        []  -> trace ("Unbound name: " ++ show q ++ " " ++ showQNameId q) __IMPOSSIBLE__
-        [d] -> mkAbs env d
-        ds  -> trace ("Ambiguous name: " ++ show q) __IMPOSSIBLE__
-    where
-      mkAbs env d
-        | treatAbstractly' q' env = fromMaybe err $ makeAbstract d
-        | otherwise               = d
-        where
-          err = trace ("Not in scope: " ++ show q) __IMPOSSIBLE__
-          q' = case theDef d of
-            -- Hack to make abstract constructors work properly. The constructors
-            -- live in a module with the same name as the datatype, but for 'abstract'
-            -- purposes they're considered to be in the same module as the datatype.
-            Constructor{} -> dropLastModule q
-            _                 -> q
-
-          dropLastModule q@QName{ qnameModule = m } =
-            q{ qnameModule = mnameFromList $ ifNull (mnameToList m) __IMPOSSIBLE__ init }
+  getConstInfo' q = do
+    ReduceEnv env st <- askR
+    defaultGetConstInfo st env q
