@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP               #-}
+{-# LANGUAGE NondecreasingIndentation #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
@@ -62,12 +63,15 @@ import Agda.Syntax.Internal as I
 
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
+import Agda.TypeChecking.Monad.Env
 import Agda.TypeChecking.EtaContract
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Free.Lazy
 import Agda.TypeChecking.MetaVars
 import Agda.TypeChecking.Conversion
+import qualified Agda.TypeChecking.Positivity.Occurrence as Pos
 import Agda.TypeChecking.Pretty
+import Agda.TypeChecking.Primitive ( getBuiltinName )
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
@@ -148,7 +152,9 @@ relView t = do
 addRewriteRule :: QName -> TCM ()
 addRewriteRule q = do
   requireOptionRewriting
-  Def rel _ <- primRewrite
+  let failNoBuiltin = typeError $ GenericError $
+        "Cannot add rewrite rule without prior BUILTIN REWRITE"
+  rel <- fromMaybeM failNoBuiltin $ getBuiltinName builtinRewrite
   def <- instantiateDef =<< getConstInfo q
   -- Issue 1651: Check that we are not adding a rewrite rule
   -- for a type signature whose body has not been type-checked yet.
@@ -202,6 +208,11 @@ addRewriteRule q = do
         reportSDoc "rewriting" 30 $ text "metas in gamma1: " <+> text (show $ allMetas $ telToList gamma1)
         failureMetas
 
+      -- 2017-06-18, Jesper: Unfold inlined definitions on the LHS.
+      -- This is necessary to replace copies created by imports by their
+      -- original definition.
+      lhs <- modifyAllowedReductions (const [InlineReductions]) $ normalise lhs
+
       -- Find head symbol f of the lhs and its arguments.
       (f , hd , es) <- case ignoreSharing lhs of
         Def f es -> return (f , Def f , es)
@@ -227,7 +238,7 @@ addRewriteRule q = do
           text "Pattern generated from lhs: " <+> prettyTCM (PDef f ps)
 
         -- check that FV(rhs) ⊆ nlPatVars(lhs)
-        let freeVars  = usedArgs gamma1 `IntSet.union` allFreeVars (ps,rhs)
+        let freeVars  = usedArgs (defArgOccurrences def) `IntSet.union` allFreeVars (ps,rhs)
             boundVars = nlPatVars ps
         reportSDoc "rewriting" 40 $
           text "variables bound by the pattern: " <+> text (show boundVars)
@@ -277,14 +288,13 @@ addRewriteRule q = do
         typeError . GenericDocError =<< do
           text "Rewrite rule " <+> prettyTCM q <+> text " has already been added"
 
-    usedArgs :: Telescope -> IntSet
-    usedArgs tel = IntSet.fromList $ map unDom $ usedIxs
+    usedArgs :: [Pos.Occurrence] -> IntSet
+    usedArgs occs = IntSet.fromList $ map snd $ usedIxs
       where
-        n = size tel
-        allIxs = zipWith ($>) (flattenTel tel) (downFrom n)
-        usedIxs = filter (not . unused . getRelevance) allIxs
-        unused UnusedArg{} = True
-        unused _           = False
+        allIxs = zip occs $ downFrom $ size occs
+        usedIxs = filter (used . fst) allIxs
+        used Pos.Unused = False
+        used _          = True
 
 -- | Append rewrite rules to a definition.
 addRewriteRules :: QName -> RewriteRules -> TCM ()
@@ -299,19 +309,6 @@ addRewriteRules f rews = do
   --  , vcat (map prettyTCM rules)
   --  ]
 
--- | Sledgehammer approach to local rewrite rules. Rebind them after each
---   left-hand side (which scrambles the context).
-rebindLocalRewriteRules :: TCM ()
-rebindLocalRewriteRules = do
-  current <- currentModule
-  ruleMap <- use $ stSignature . sigRewriteRules
-  let isLocal r = m == current || m `isSubModuleOf` current
-        where m = qnameModule $ rewName r
-      ruleMap' = HMap.map (filter (not . isLocal)) ruleMap
-      locals = map rewName $ filter isLocal $ concat $ map reverse $ HMap.elems ruleMap
-  stSignature . sigRewriteRules .= ruleMap'
-  mapM_ addRewriteRule locals
-
 -- | @rewriteWith t f es rew@
 --   tries to rewrite @f es : t@ with @rew@, returning the reduct if successful.
 rewriteWith :: Maybe Type
@@ -320,19 +317,20 @@ rewriteWith :: Maybe Type
             -> Elims
             -> ReduceM (Either (Blocked Term) Term)
 rewriteWith mt v rew@(RewriteRule q gamma _ ps rhs b) es = do
-  Red.traceSDoc "rewriting" 75 (sep
+  traceSDoc "rewriting" 75 (sep
     [ text "attempting to rewrite term " <+> prettyTCM (v `applyE` es)
     , text " with rule " <+> prettyTCM rew
     ]) $ do
-    result <- nonLinMatch gamma ps es
-    case result of
-      Left block -> return $ Left $ block $> v `applyE` es -- TODO: remember reductions
-      Right sub  -> do
-        let v' = applySubst sub rhs
-        Red.traceSDoc "rewriting" 70 (sep
-          [ text "rewrote " <+> prettyTCM (v `applyE` es)
-          , text " to " <+> prettyTCM v'
-          ]) $ return $ Right v'
+  result <- nonLinMatch gamma ps es
+  case result of
+    Left block -> return $ Left $ block $> v `applyE` es -- TODO: remember reductions
+    Right sub  -> do
+      let v' = applySubst sub rhs
+      traceSDoc "rewriting" 70 (sep
+        [ text "rewrote " <+> prettyTCM (v `applyE` es)
+        , text " to " <+> prettyTCM v'
+        ]) $ do
+      return $ Right v'
 
     {- OLD CODE:
     -- Freeze all metas, remember which one where not frozen before.
@@ -388,51 +386,31 @@ rewrite block v rules es = do
 ------------------------------------------------------------------------
 
 class NLPatVars a where
+  nlPatVarsUnder :: Int -> a -> IntSet
+
   nlPatVars :: a -> IntSet
+  nlPatVars = nlPatVarsUnder 0
 
 instance (Foldable f, NLPatVars a) => NLPatVars (f a) where
-  nlPatVars = foldMap nlPatVars
+  nlPatVarsUnder k = foldMap $ nlPatVarsUnder k
 
 instance NLPatVars NLPType where
-  nlPatVars (NLPType l a) = nlPatVars l `IntSet.union` nlPatVars a
+  nlPatVarsUnder k (NLPType l a) = nlPatVarsUnder k l `IntSet.union` nlPatVarsUnder k a
 
 instance NLPatVars NLPat where
-  nlPatVars p =
+  nlPatVarsUnder k p =
     case p of
-      PVar _ i _ -> singleton i
-      PDef _ es -> nlPatVars es
+      PVar i _  -> singleton $ i - k
+      PDef _ es -> nlPatVarsUnder k es
       PWild     -> empty
-      PLam _ p' -> nlPatVars $ unAbs p'
-      PPi a b   -> nlPatVars a `IntSet.union` nlPatVars (unAbs b)
-      PBoundVar _ es -> nlPatVars es
+      PLam _ p' -> nlPatVarsUnder (k+1) $ unAbs p'
+      PPi a b   -> nlPatVarsUnder k a `IntSet.union` nlPatVarsUnder (k+1) (unAbs b)
+      PBoundVar _ es -> nlPatVarsUnder k es
       PTerm{}   -> empty
 
 rewArity :: RewriteRule -> Int
 rewArity = length . rewPats
 
--- | Erase the CtxId's of rewrite rules
-class KillCtxId a where
-  killCtxId :: a -> a
-
-instance (Functor f, KillCtxId a) => KillCtxId (f a) where
-  killCtxId = fmap killCtxId
-
-instance KillCtxId RewriteRule where
-  killCtxId rule@RewriteRule{ rewPats = ps } = rule{ rewPats = killCtxId ps }
-
-instance KillCtxId NLPType where
-  killCtxId (NLPType l a) = NLPType (killCtxId l) (killCtxId a)
-
-instance KillCtxId NLPat where
-  killCtxId p = case p of
-    PVar _ i bvs   -> PVar Nothing i bvs
-    PWild          -> p
-    PDef f es      -> PDef f $ killCtxId es
-    PLam i x       -> PLam i $ killCtxId x
-    PPi a b        -> PPi (killCtxId a) (killCtxId b)
-    PBoundVar i es -> PBoundVar i $ killCtxId es
-    PTerm _        -> p
-
 -- | Get all symbols that a rewrite rule matches against
 class GetMatchables a where
   getMatchables :: a -> [QName]
@@ -443,7 +421,7 @@ instance (Foldable f, GetMatchables a) => GetMatchables (f a) where
 instance GetMatchables NLPat where
   getMatchables p =
     case p of
-      PVar _ _ _     -> empty
+      PVar _ _       -> empty
       PWild          -> empty
       PDef f _       -> singleton f
       PLam _ x       -> empty
@@ -455,9 +433,9 @@ instance GetMatchables RewriteRule where
   getMatchables = getMatchables . rewPats
 
 -- Only computes free variables that are not bound (i.e. those in a PTerm)
-instance Free' NLPat c where
+instance Free NLPat where
   freeVars' p = case p of
-    PVar _ _ _ -> mempty
+    PVar _ _ -> mempty
     PWild -> mempty
     PDef _ es -> freeVars' es
     PLam _ u -> freeVars' u
@@ -465,7 +443,7 @@ instance Free' NLPat c where
     PBoundVar _ es -> freeVars' es
     PTerm t -> freeVars' t
 
-instance Free' NLPType c where
+instance Free NLPType where
   freeVars' (NLPType l a) =
     ifM ((IgnoreNot ==) <$> asks feIgnoreSorts)
       {- then -} (freeVars' (l, a))
