@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Rules.Builtin
   ( bindBuiltin
@@ -29,16 +28,20 @@ import qualified Agda.TypeChecking.CompiledClause as CC
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.EtaContract
+import Agda.TypeChecking.Functions
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.Primitive
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Rules.Term ( checkExpr , inferExpr )
+import Agda.TypeChecking.Warnings
 
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Builtin.Coinduction
 import {-# SOURCE #-} Agda.TypeChecking.Rewriting
 
 import Agda.Utils.Except ( MonadError(catchError) )
+import Agda.Utils.List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
@@ -55,7 +58,7 @@ builtinPostulate :: TCM Type -> BuiltinDescriptor
 builtinPostulate = BuiltinPostulate Relevant
 
 coreBuiltins :: [BuiltinInfo]
-coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
+coreBuiltins =
   [ (builtinList               |-> BuiltinData (tset --> tset) [builtinNil, builtinCons])
   , (builtinArg                |-> BuiltinData (tset --> tset) [builtinArgArg])
   , (builtinAbs                |-> BuiltinData (tset --> tset) [builtinAbsAbs])
@@ -95,12 +98,13 @@ coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
   , builtinAgdaErrorPartString |-> BuiltinDataCons (tstring --> terrorpart)
   , builtinAgdaErrorPartTerm   |-> BuiltinDataCons (tterm --> terrorpart)
   , builtinAgdaErrorPartName   |-> BuiltinDataCons (tqname --> terrorpart)
-  , (builtinEquality           |-> BuiltinData (hPi "a" (el primLevel) $
-                                                hPi "A" (return $ sort $ varSort 0) $
-                                                (El (varSort 1) <$> varM 0) -->
-                                                (El (varSort 1) <$> varM 0) -->
-                                                return (sort $ varSort 1))
-                                               [builtinRefl])
+  -- Andreas, 2017-01-12, issue #2386: special handling of builtinEquality
+  -- , (builtinEquality           |-> BuiltinData (hPi "a" (el primLevel) $
+  --                                               hPi "A" (return $ sort $ varSort 0) $
+  --                                               (El (varSort 1) <$> varM 0) -->
+  --                                               (El (varSort 1) <$> varM 0) -->
+  --                                               return (sort $ varSort 1))
+  --                                              [builtinRefl])
   , (builtinHiding             |-> BuiltinData tset [builtinHidden, builtinInstance, builtinVisible])
     -- Relevance
   , (builtinRelevance          |-> BuiltinData tset [builtinRelevant, builtinIrrelevant])
@@ -118,10 +122,11 @@ coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
     -- Fixity
   , builtinFixity              |-> BuiltinData tset [builtinFixityFixity]
   , builtinFixityFixity        |-> BuiltinDataCons (tassoc --> tprec --> tfixity)
-  , (builtinRefl               |-> BuiltinDataCons (hPi "a" (el primLevel) $
-                                                    hPi "A" (return $ sort $ varSort 0) $
-                                                    hPi "x" (El (varSort 1) <$> varM 0) $
-                                                    El (varSort 2) <$> primEquality <#> varM 2 <#> varM 1 <@> varM 0 <@> varM 0))
+  -- Andreas, 2017-01-12, issue #2386: special handling of builtinEquality
+  -- , (builtinRefl               |-> BuiltinDataCons (hPi "a" (el primLevel) $
+  --                                                   hPi "A" (return $ sort $ varSort 0) $
+  --                                                   hPi "x" (El (varSort 1) <$> varM 0) $
+  --                                                   El (varSort 2) <$> primEquality <#> varM 2 <#> varM 1 <@> varM 0 <@> varM 0))
   , (builtinRewrite           |-> BuiltinUnknown Nothing verifyBuiltinRewrite)
   , (builtinNil                |-> BuiltinDataCons (hPi "A" tset (el (list v0))))
   , (builtinCons               |-> BuiltinDataCons (hPi "A" tset (tv0 --> el (list v0) --> el (list v0))))
@@ -186,7 +191,7 @@ coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
   , (builtinAgdaDefinitionFunDef          |-> BuiltinDataCons (tlist tclause --> tdefn))
   , (builtinAgdaDefinitionDataDef         |-> BuiltinDataCons (tnat --> tlist tqname --> tdefn))
   , (builtinAgdaDefinitionDataConstructor |-> BuiltinDataCons (tqname --> tdefn))
-  , (builtinAgdaDefinitionRecordDef       |-> BuiltinDataCons (tqname --> tdefn))
+  , (builtinAgdaDefinitionRecordDef       |-> BuiltinDataCons (tqname --> tlist (targ tqname) --> tdefn))
   , (builtinAgdaDefinitionPostulate       |-> BuiltinDataCons tdefn)
   , (builtinAgdaDefinitionPrimitive       |-> BuiltinDataCons tdefn)
   , builtinAgdaTCM       |-> builtinPostulate (hPi "a" tlevel $ tsetL 0 --> tsetL 0)
@@ -217,9 +222,10 @@ coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
   , builtinAgdaTCMCommit             |-> builtinPostulate (tTCM_ primUnit)
   , builtinAgdaTCMIsMacro            |-> builtinPostulate (tqname --> tTCM_ primBool)
   , builtinAgdaTCMWithNormalisation  |-> builtinPostulate (hPi "a" tlevel $ hPi "A" (tsetL 0) $ tbool --> tTCM 1 (varM 0) --> tTCM 1 (varM 0))
+  , builtinAgdaTCMDebugPrint         |-> builtinPostulate (tstring --> tnat --> tlist terrorpart --> tTCM_ primUnit)
   ]
   where
-        (|->) = (,)
+        (|->) = BuiltinInfo
 
         v0 = varM 0
         v1 = varM 1
@@ -387,27 +393,35 @@ coreBuiltins = map (\ (x, z) -> BuiltinInfo x z)
             xs <- mapM freshName_ xs
             addContext (xs, domFromArg $ defaultArg nat) $ f apply1 zero suc (==) (===) choice
 
-
-inductiveCheck :: String -> Int -> Term -> TCM ()
+-- | Checks that builtin with name @b : String@ of type @t : Term@
+--   is a data type or inductive record with @n : Int@ constructors.
+--   Returns the name of the data/record type.
+inductiveCheck :: String -> Int -> Term -> TCM (QName, Definition)
 inductiveCheck b n t = do
-    t <- etaContract =<< normalise t
-    let err = typeError (NotInductive t)
-    case ignoreSharing t of
-      Def t _ -> do
-        t <- theDef <$> getConstInfo t
-        case t of
-          Datatype { dataInduction = Inductive
-                   , dataCons      = cs
-                   }
-            | length cs == n -> return ()
-            | otherwise ->
-              typeError $ GenericError $ unwords
-                          [ "The builtin", b
-                          , "must be a datatype with", show n
-                          , "constructors" ]
-          Record { recInduction = ind } | n == 1 && ind /= Just CoInductive -> return ()
-          _ -> err
-      _ -> err
+  t <- etaContract =<< normalise t
+  case ignoreSharing t of
+    Def q _ -> do
+      def <- getConstInfo q
+      let yes = return (q, def)
+      case theDef def of
+        Datatype { dataInduction = Inductive, dataCons = cs }
+          | length cs == n -> yes
+          | otherwise      -> no
+        Record { recInduction = ind } | n == 1 && ind /= Just CoInductive -> yes
+        _ -> no
+    _ -> no
+  where
+  no
+    | n == 1 = typeError $ GenericError $ unwords
+        [ "The builtin", b
+        , "must be a datatype with a single constructor"
+        , "or an (inductive) record type"
+        ]
+    | otherwise = typeError $ GenericError $ unwords
+        [ "The builtin", b
+        , "must be a datatype with", show n
+        , "constructors"
+        ]
 
 -- | @bindPostulatedName builtin e m@ checks that @e@ is a postulated
 -- name @q@, and binds the builtin @builtin@ to the term @m q def@,
@@ -430,28 +444,32 @@ bindPostulatedName builtin e m = do
   getName (A.ScopedExpr _ e) = getName e
   getName _                  = err
 
-getDef :: Term -> TCM QName
-getDef t = do
-  t <- etaContract =<< normalise t
-  case ignoreSharing t of
-    Def d _ -> return d
-    _ -> __IMPOSSIBLE__
+-- REPLACED by TC.Functions.getDef
+-- getDef :: Term -> TCM QName
+-- getDef t = do
+--   t <- etaContract =<< normalise t
+--   case ignoreSharing t of
+--     Def d _ -> return d
+--     _ -> __IMPOSSIBLE__
 
-bindAndSetHaskellType :: String -> String -> Term -> TCM ()
-bindAndSetHaskellType b hs t = do
-  d <- getDef t
-  addHaskellType d hs
+addHaskellPragma :: QName -> String -> TCM ()
+addHaskellPragma = addPragma ghcBackendName
+
+bindAndSetHaskellCode :: String -> String -> Term -> TCM ()
+bindAndSetHaskellCode b hs t = do
+  d <- fromMaybe __IMPOSSIBLE__ <$> getDef t
+  addHaskellPragma d hs
   bindBuiltinName b t
 
 bindBuiltinBool :: Term -> TCM ()
-bindBuiltinBool = bindAndSetHaskellType builtinBool "Bool"
+bindBuiltinBool = bindAndSetHaskellCode builtinBool "= type Bool"
 
 bindBuiltinInt :: Term -> TCM ()
-bindBuiltinInt = bindAndSetHaskellType builtinInteger "Integer"
+bindBuiltinInt = bindAndSetHaskellCode builtinInteger "= type Integer"
 
 bindBuiltinNat :: Term -> TCM ()
 bindBuiltinNat t = do
-  nat <- getDef t
+  nat <- fromMaybe __IMPOSSIBLE__ <$> getDef t
   def <- theDef <$> getConstInfo nat
   case def of
     Datatype { dataCons = [c1, c2] } -> do
@@ -462,7 +480,7 @@ bindBuiltinNat t = do
                       | otherwise = (c2, c1)
           tnat = el primNat
           rerange = setRange (getRange nat)
-      addHaskellType nat "Integer"
+      addHaskellPragma nat "= type Integer"
       bindBuiltinInfo (BuiltinInfo builtinZero $ BuiltinDataCons tnat)
                       (A.Con $ AmbQ [rerange zero])
       bindBuiltinInfo (BuiltinInfo builtinSuc  $ BuiltinDataCons (tnat --> tnat))
@@ -471,7 +489,7 @@ bindBuiltinNat t = do
 
 bindBuiltinUnit :: Term -> TCM ()
 bindBuiltinUnit t = do
-  unit <- getDef t
+  unit <- fromMaybe __IMPOSSIBLE__ <$> getDef t
   def <- theDef <$> getConstInfo unit
   case def of
     Record { recFields = [], recConHead = con } -> do
@@ -479,20 +497,75 @@ bindBuiltinUnit t = do
       bindBuiltinName builtinUnitUnit (Con con ConOSystem [])
     _ -> genericError "Builtin UNIT must be a singleton record type"
 
+-- | Bind BUILTIN EQUALITY and BUILTIN REFL.
+bindBuiltinEquality :: A.Expr -> TCM ()
+bindBuiltinEquality (A.ScopedExpr scope e) = do
+  setScope scope
+  bindBuiltinEquality e
+bindBuiltinEquality e = do
+  (v, _t) <- inferExpr e
+
+  -- Equality needs to be a data type with 1 constructor
+  (eq, def) <- inductiveCheck builtinEquality 1 v
+
+  -- Check that the type is the type of a polymorphic relation, i.e.,
+  -- Γ → (A : Set _) → A → A → Set _
+  TelV eqTel eqCore <- telView $ defType def
+  let no = genericError "The type of BUILTIN EQUALITY must be a polymorphic relation"
+
+  -- The target is a sort since eq is a data type.
+  unless (isJust $ isSort $ unEl eqCore) __IMPOSSIBLE__
+
+  -- The types of the last two arguments must be the third-last argument
+  unless (size eqTel >= 3) no
+  let (a, b) = fromMaybe __IMPOSSIBLE__ $ last2 $ telToList eqTel
+  [a,b] <- reduce $ map (unEl . snd . unDom) [a,b]
+  unless (deBruijnView a == Just 0) no
+  unless (deBruijnView b == Just 1) no
+
+  -- Get the single constructor.
+  case theDef def of
+    Datatype { dataCons = [c] } -> do
+      bindBuiltinName builtinEquality v
+
+      -- Check type of REFL.  It has to be of the form
+      -- pars → (x : A) → Eq ... x x
+
+      -- Check the arguments
+      cdef <- getConstInfo c
+      TelV conTel conCore <- telView $ defType cdef
+      ts <- reduce $ map (unEl . snd . unDom) $ drop (conPars $ theDef cdef) $ telToList conTel
+      -- After dropping the parameters, there should be maximally one argument.
+      unless (length ts <= 1) wrongRefl
+      unless (all ((Just 0 ==) . deBruijnView) ts) wrongRefl
+
+      -- Check the target
+      case ignoreSharing $ unEl conCore of
+        Def _ es -> do
+          let vs = map unArg $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+          (a,b) <- reduce $ fromMaybe __IMPOSSIBLE__ $ last2 vs
+          unless (deBruijnView a == Just 0) wrongRefl
+          unless (deBruijnView b == Just 0) wrongRefl
+          bindBuiltinName builtinRefl (Con (ConHead c Inductive []) ConOSystem [])
+        _ -> __IMPOSSIBLE__
+    _ -> genericError "Builtin EQUALITY must be a data type with a single constructor"
+  where
+  wrongRefl = genericError "Wrong type of constructor of BUILTIN EQUALITY"
+
 bindBuiltinInfo :: BuiltinInfo -> A.Expr -> TCM ()
 bindBuiltinInfo i (A.ScopedExpr scope e) = setScope scope >> bindBuiltinInfo i e
 bindBuiltinInfo (BuiltinInfo s d) e = do
     case d of
       BuiltinData t cs -> do
-        e' <- checkExpr e =<< t
-        let n = length cs
-        inductiveCheck s n e'
-        case () of
-          _ | s == builtinBool    -> bindBuiltinBool e'
-            | s == builtinNat     -> bindBuiltinNat e'
-            | s == builtinInteger -> bindBuiltinInt e'
-            | s == builtinUnit    -> bindBuiltinUnit e'
-            | otherwise           -> bindBuiltinName s e'
+        v <- checkExpr e =<< t
+        unless (s == builtinUnit) $ do
+          void $ inductiveCheck s (length cs) v
+        if | s == builtinEquality -> __IMPOSSIBLE__ -- bindBuiltinEquality v
+           | s == builtinBool     -> bindBuiltinBool     v
+           | s == builtinNat      -> bindBuiltinNat      v
+           | s == builtinInteger  -> bindBuiltinInt      v
+           | s == builtinUnit     -> bindBuiltinUnit     v
+           | otherwise            -> bindBuiltinName s   v
 
       BuiltinDataCons t -> do
 
@@ -501,16 +574,14 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
             name (Shared p) = name $ ignoreSharing (derefPtr p)
             name _          = __IMPOSSIBLE__
 
-        e' <- checkExpr e =<< t
+        v0 <- checkExpr e =<< t
 
         case e of
           A.Con{} -> return ()
           _       -> typeError $ BuiltinMustBeConstructor s e
 
-        let v@(Con h _ []) = name e'
+        let v@(Con h _ []) = name v0
             c = conName h
-        when (s == builtinTrue)  $ addHaskellCode c "Bool" "True"
-        when (s == builtinFalse) $ addHaskellCode c "Bool" "False"
         bindBuiltinName s v
 
       BuiltinPrim pfname axioms -> do
@@ -536,7 +607,7 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
 
       BuiltinPostulate rel t -> do
         t' <- t
-        e' <- applyRelevanceToContext rel $ checkExpr e t'
+        v <- applyRelevanceToContext rel $ checkExpr e t'
         let err = typeError $ GenericError $
                     "The argument to BUILTIN " ++ s ++ " must be a postulated name"
         case e of
@@ -545,10 +616,10 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
             case theDef def of
               Axiom {} -> do
                 builtinSizeHook s q t'
-                when (s == builtinChar)   $ addHaskellType q "Char"
-                when (s == builtinString) $ addHaskellType q "Data.Text.Text"
-                when (s == builtinFloat)  $ addHaskellType q "Double"
-                bindBuiltinName s e'
+                when (s == builtinChar)   $ addHaskellPragma q "= type Char"
+                when (s == builtinString) $ addHaskellPragma q "= type Data.Text.Text"
+                when (s == builtinFloat)  $ addHaskellPragma q "= type Double"
+                bindBuiltinName s v
               _        -> err
           _ -> err
 
@@ -563,14 +634,15 @@ bindBuiltinInfo (BuiltinInfo s d) e = do
 bindBuiltin :: String -> A.Expr -> TCM ()
 bindBuiltin b e = do
   unlessM ((0 ==) <$> getContextSize) $ typeError $ BuiltinInParameterisedModule b
-  case b of
-    _ | b == builtinZero  -> nowNat b
-    _ | b == builtinSuc   -> nowNat b
-    _ | b == builtinInf   -> bindBuiltinInf e
-    _ | b == builtinSharp -> bindBuiltinSharp e
-    _ | b == builtinFlat  -> bindBuiltinFlat e
-    _ | Just i <- find ((b ==) . builtinName) coreBuiltins -> bindBuiltinInfo i e
-    _ -> typeError $ NoSuchBuiltinName b
+  if | b == builtinRefl  -> warning $ OldBuiltin b builtinEquality
+     | b == builtinZero  -> nowNat b
+     | b == builtinSuc   -> nowNat b
+     | b == builtinInf   -> bindBuiltinInf e
+     | b == builtinSharp -> bindBuiltinSharp e
+     | b == builtinFlat  -> bindBuiltinFlat e
+     | b == builtinEquality -> bindBuiltinEquality e
+     | Just i <- find ((b ==) . builtinName) coreBuiltins -> bindBuiltinInfo i e
+     | otherwise -> typeError $ NoSuchBuiltinName b
   where
     nowNat b = warning $ OldBuiltin b builtinNat
 
