@@ -86,15 +86,15 @@ bindBuiltinSharp e =
                   , recNamedCon       = True
                   , recFields         = []  -- flat is added later
                   , recTel            = fieldTel
-                  , recRecursive      = False
                   , recEtaEquality'   = Inferred False
-                  , recMutual         = []
+                  , recMutual         = Just []
                   , recAbstr          = ConcreteDef
                   }
               }
     addConstant sharp $
       sharpDefn { theDef = Constructor
                     { conPars   = 2
+                    , conArity  = 1
                     , conSrcCon = ConHead sharp CoInductive [] -- flat is added as field later
                     , conData   = defName infDefn
                     , conAbstr  = ConcreteDef
@@ -128,13 +128,15 @@ bindBuiltinFlat e =
         infA     = El (varSort 2) $ Def inf [ Apply $ defaultArg $ var 1 ]
         cpi      = ConPatternInfo Nothing $ Just $ defaultArg infA
     let clause   = Clause
-          { clauseRange     = noRange
+          { clauseLHSRange  = noRange
+          , clauseFullRange = noRange
           , clauseTel       = tel
           , namedClausePats = [ argN $ Named Nothing $
               ConP sharpCon cpi [ argN $ Named Nothing $ debruijnNamedVar "x" 0 ] ]
           , clauseBody      = Just $ var 0
           , clauseType      = Just $ defaultArg $ El (varSort 2) $ var 1
           , clauseCatchall  = False
+          , clauseUnreachable = Just False
           }
         cc = Case (defaultArg 0) $ Branches False
                                (Map.singleton sharp
@@ -142,7 +144,7 @@ bindBuiltinFlat e =
                                Map.empty
                                Nothing
         projection = Projection
-          { projProper   = True
+          { projProper   = Just inf
           , projOrig     = flat
           , projFromType = defaultArg inf
           , projIndex    = 3
