@@ -12,15 +12,13 @@ import Control.Monad.Writer (tell)
 
 import Data.Either (partitionEithers)
 import qualified Data.Foldable as Fold
-import Data.List (genericLength)
+import qualified Data.List as List
 import Data.Maybe
 import Data.Map (Map)
 import qualified Data.Sequence as Seq
 import qualified Data.Set as Set
 import Data.Set (Set)
 
-import Agda.Compiler.HaskellTypes
-import Agda.Compiler.UHC.Pragmas.Parse
 import Agda.Interaction.Options
 import Agda.Interaction.Highlighting.Generate
 
@@ -60,6 +58,7 @@ import Agda.TypeChecking.SizedTypes.Solve
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Unquote
+import Agda.TypeChecking.Warnings
 
 import Agda.TypeChecking.Rules.Term
 import Agda.TypeChecking.Rules.Data    ( checkDataDef )
@@ -186,7 +185,7 @@ checkDecl d = setCurrentRange d $ do
                                       current <- asks envMutualBlock
                                       unless (Just blockId == current) $ do
                                         reportSLn "" 0 $ unlines
-                                          [ "mutual block id discrepancy for " ++ show x
+                                          [ "mutual block id discrepancy for " ++ prettyShow x
                                           , "  current    mut. bl. = " ++ show current
                                           , "  calculated mut. bl. = " ++ show blockId
                                           ]
@@ -216,9 +215,12 @@ checkDecl d = setCurrentRange d $ do
 
       -- Post-typing checks.
       whenJust finalChecks $ \ theMutualChecks -> do
+        reportSLn "tc.decl" 20 $ "Attempting to solve constraints before freezing."
+        wakeupConstraints_   -- solve emptiness and instance constraints
         checkingWhere <- asks envCheckingWhere
         solveSizeConstraints $ if checkingWhere then DontDefaultToInfty else DefaultToInfty
-        wakeupConstraints_   -- solve emptiness constraints
+        wakeupConstraints_   -- Size solver might have unblocked some constraints
+        reportSLn "tc.decl" 20 $ "Freezing all metas."
         _ <- freezeMetas
         theMutualChecks
 
@@ -229,7 +231,7 @@ checkDecl d = setCurrentRange d $ do
       A.Axiom A.NoFunSig i defaultArgInfo Nothing x
               (A.Pi (Info.ExprRange (fuseRange ps t)) ps t)
 
-    check x i m = do
+    check x i m = Bench.billTo [Bench.Definition x] $ do
       reportSDoc "tc.decl" 5 $ text "Checking" <+> prettyTCM x <> text "."
       reportSLn "tc.decl.abstract" 25 $ show (Info.defAbstract i)
       r <- abstract (Info.defAbstract i) m
@@ -250,10 +252,14 @@ mutualChecks mi d ds mid names = do
   -- Andreas, 2014-04-11: instantiate metas in definition types
   let nameList = Set.toList names
   mapM_ instantiateDefinitionType nameList
+  -- Andreas, 2017-03-23: check positivity before termination.
+  -- This allows us to reuse the information about SCCs
+  -- to skip termination of non-recursive functions.
+  modifyAllowedReductions (List.delete UnconfirmedReductions) $
+    checkPositivity_ mi names
   -- Andreas, 2013-02-27: check termination before injectivity,
   -- to avoid making the injectivity checker loop.
   local (\ e -> e { envMutualBlock = Just mid }) $ checkTermination_ d
-  checkPositivity_         mi names
   revisitRecordPatternTranslation nameList -- Andreas, 2016-11-19 issue #2308
   -- Andreas, 2015-03-26 Issue 1470:
   -- Restricting coinduction to recursive does not solve the
@@ -292,7 +298,12 @@ revisitRecordPatternTranslation qs = do
   classify q = inConcreteOrAbstractMode q $ \ def -> do
     case theDef def of
       Record{ recEtaEquality' = Inferred True } -> return $ Just $ Left q
-      Function{ funCompiled = Just cc } -> return $ Just $ Right (q, cc)
+      Function
+        { funProjection = Nothing
+            -- Andreas, 2017-08-10, issue #2664:
+            -- Do not record pattern translate record projection definitions!
+        , funCompiled   = Just cc
+        } -> return $ Just $ Right (q, cc)
       _ -> return Nothing
 
 type FinalChecks = Maybe (TCM ())
@@ -329,11 +340,14 @@ unquoteTop xs e = do
 --   Precondition: name has been added to signature already.
 instantiateDefinitionType :: QName -> TCM ()
 instantiateDefinitionType q = do
-  reportSLn "tc.decl.inst" 20 $ "instantiating type of " ++ show q
-  sig <- getSignature
-  let t = defType $ fromMaybe __IMPOSSIBLE__ $ lookupDefinition q sig
-  t <- instantiateFull t
-  modifySignature $ updateDefinition q $ \ def -> def { defType = t }
+  reportSLn "tc.decl.inst" 20 $ "instantiating type of " ++ prettyShow q
+  t  <- defType . fromMaybe __IMPOSSIBLE__ . lookupDefinition q <$> getSignature
+  t' <- instantiateFull t
+  modifySignature $ updateDefinition q $ updateDefType $ const t'
+  reportSDoc "tc.decl.inst" 30 $ vcat
+    [ text "  t  = " <+> prettyTCM t
+    , text "  t' = " <+> prettyTCM t'
+    ]
 
 -- Andreas, 2014-04-11
 -- UNUSED, costs a couple of sec on the std-lib
@@ -345,7 +359,7 @@ instantiateDefinitionType q = do
 -- --   Precondition: name has been added to signature already.
 -- instantiateDefinition :: QName -> TCM ()
 -- instantiateDefinition q = do
---   reportSLn "tc.decl.inst" 20 $ "instantiating " ++ show q
+--   reportSLn "tc.decl.inst" 20 $ "instantiating " ++ prettyShow q
 --   sig <- getSignature
 --   let def = fromMaybe __IMPOSSIBLE__ $ lookupDefinition q sig
 --   def <- instantiateFull def
@@ -354,7 +368,7 @@ instantiateDefinitionType q = do
 -- | Highlight a declaration.
 highlight_ :: A.Declaration -> TCM ()
 highlight_ d = do
-  let highlight d = generateAndPrintSyntaxInfo d Full
+  let highlight d = generateAndPrintSyntaxInfo d Full True
   Bench.billTo [Bench.Highlighting] $ case d of
     A.Axiom{}                -> highlight d
     A.Field{}                -> __IMPOSSIBLE__
@@ -421,7 +435,7 @@ checkPositivity_ mi names = Bench.billTo [Bench.Positivity] $ do
   -- Andreas, 2012-02-13: Polarity computation uses info from
   -- positivity check, so it needs happen after positivity
   -- check.
-  mapM_ computePolarity $ Set.toList names
+  computePolarity $ Set.toList names
 
 -- | Check that all coinductive records are actually recursive.
 --   (Otherwise, one can implement invalid recursion schemes just like
@@ -457,15 +471,15 @@ checkInjectivity_ names = Bench.billTo [Bench.Injectivity] $ do
         case term of
           Just True -> do
             inv <- checkInjectivity q cs
-            modifySignature $ updateDefinition q $ const $
-              def { theDef = d { funInv = inv }}
+            modifySignature $ updateDefinition q $ updateTheDef $ const $
+              d { funInv = inv }
           _ -> reportSLn "tc.inj.check" 20 $
-             show q ++ " is not verified as terminating, thus, not considered for injectivity"
+             prettyShow q ++ " is not verified as terminating, thus, not considered for injectivity"
       _ -> do
         abstr <- asks envAbstractMode
         reportSLn "tc.inj.check" 20 $
           "we are in " ++ show abstr ++ " and " ++
-             show q ++ " is abstract or not a function, thus, not considered for injectivity"
+             prettyShow q ++ " is abstract or not a function, thus, not considered for injectivity"
 
 -- | Check a set of mutual names for projection likeness.
 --
@@ -479,7 +493,7 @@ checkProjectionLikeness_ names = Bench.billTo [Bench.ProjectionLikeness] $ do
       -- Non-mutual definitions can be considered for
       -- projection likeness
       let ds = Set.toList names
-      reportSLn "tc.proj.like" 20 $ "checkDecl: checking projection-likeness of " ++ show ds
+      reportSLn "tc.proj.like" 20 $ "checkDecl: checking projection-likeness of " ++ prettyShow ds
       case ds of
         [d] -> do
           def <- getConstInfo d
@@ -488,7 +502,7 @@ checkProjectionLikeness_ names = Bench.billTo [Bench.ProjectionLikeness] $ do
           case theDef def of
             Function{} -> makeProjection (defName def)
             _          -> reportSLn "tc.proj.like" 25 $
-              show d ++ " is abstract or not a function, thus, not considered for projection-likeness"
+              prettyShow d ++ " is abstract or not a function, thus, not considered for projection-likeness"
         _ -> reportSLn "tc.proj.like" 25 $
                "mutual definitions are not considered for projection-likeness"
 
@@ -537,13 +551,13 @@ checkAxiom funSig i info0 mp x e = whenAbstractFreezeMetasAfter i $ do
     Nothing   -> return ([], [])
     Just occs -> do
       TelV tel _ <- telView t
-      let n = genericLength (telToList tel)
-      when (n < genericLength occs) $
+      let n = length (telToList tel)
+      when (n < length occs) $
         typeError $ TooManyPolarities x n
       let pols = map polFromOcc occs
       reportSLn "tc.polarity.pragma" 10 $
-        "Setting occurrences and polarity for " ++ show x ++ ":\n  " ++
-        show occs ++ "\n  " ++ show pols
+        "Setting occurrences and polarity for " ++ prettyShow x ++ ":\n  " ++
+        prettyShow occs ++ "\n  " ++ prettyShow pols
       return (occs, pols)
 
   -- Not safe. See Issue 330
@@ -595,7 +609,7 @@ assertCurrentModule x err =
   do def <- getConstInfo x
      m <- currentModule
      let m' = qnameModule $ defName def
-     unless (m == m') $ typeError $ GenericError err
+     unless (m == m' || isSubModuleOf m' m) $ typeError $ GenericError err
 
 -- | Check a pragma.
 checkPragma :: Range -> A.Pragma -> TCM ()
@@ -604,17 +618,6 @@ checkPragma r p =
         A.BuiltinPragma x e -> bindBuiltin x e
         A.BuiltinNoDefPragma b x -> bindBuiltinNoDef b x
         A.RewritePragma q   -> addRewriteRule q
-        A.CompiledDeclareDataPragma x hs -> do
-          def <- getConstInfo x
-          assertCurrentModule x $
-              "COMPILED_DECLARE_DATA directives must appear in the same module " ++
-              "as their corresponding datatype definition,"
-          case theDef def of
-            Datatype{} -> addHaskellType x hs
-            Axiom{}    -> -- possible when the data type has only been declared yet
-              addHaskellType x hs
-            _          -> typeError $ GenericError
-                          "COMPILED_DECLARE_DATA directive only works on data types"
         A.CompiledTypePragma x hs -> do
           def <- getConstInfo x
           case theDef def of
@@ -628,83 +631,38 @@ checkPragma r p =
           assertCurrentModule x $
               "COMPILED_DATA directives must appear in the same module " ++
               "as their corresponding datatype definition,"
-          let addCompiledData cs = do
-                addHaskellType x hs
-                let computeHaskellType c = do
-                      def <- getConstInfo c
-                      let Constructor{ conPars = np } = theDef def
-                          underPars 0 a = haskellType a
-                          underPars n a = do
-                            a <- reduce a
-                            case unEl a of
-                              Pi a (NoAbs _ b) -> underPars (n - 1) b
-                              Pi a b  -> underAbstraction a b $ \b -> hsForall <$> getHsVar 0 <*> underPars (n - 1) b
-                              _       -> __IMPOSSIBLE__
-                      ty <- underPars np $ defType def
-                      reportSLn "tc.pragma.compile" 10 $ "Haskell type for " ++ show c ++ ": " ++ ty
-                      return ty
-                hts <- mapM computeHaskellType cs
-                sequence_ $ zipWith3 addHaskellCode cs hts hcs
           case theDef def of
-            Datatype{dataCons = cs}
-              | length cs /= length hcs -> do
-                  let n_forms_are = case length hcs of
-                        1 -> "1 compiled form is"
-                        n -> show n ++ " compiled forms are"
-                      only | null hcs               = ""
-                           | length hcs < length cs = "only "
-                           | otherwise              = ""
-
-                  err <- fsep $ [prettyTCM x] ++ pwords ("has " ++ show (length cs) ++
-                                " constructors, but " ++ only ++ n_forms_are ++ " given [" ++ unwords hcs ++ "]")
-                  typeError $ GenericError $ show err
-              | otherwise -> addCompiledData cs
-            Record{recConHead = ch}
-              | length hcs == 1 -> addCompiledData [conName ch]
-              | otherwise -> do
-                  err <- fsep $ [prettyTCM x] ++ pwords ("has 1 constructor, but " ++
-                                show (length hcs) ++ " Haskell constructors are given [" ++ unwords hcs ++ "]")
-                  typeError $ GenericError $ show err
+            Datatype{dataCons = cs} -> addHaskellData x hs hcs
+            Record{recConHead = ch} -> addHaskellData x hs hcs
             _ -> typeError $ GenericError "COMPILED_DATA on non datatype"
+        A.CompilePragma b x s -> do
+          assertCurrentModule x $
+              "COMPILE pragmas must appear in the same module " ++
+              "as their corresponding definitions,"
+          addPragma b x s
         A.CompiledPragma x hs -> do
           def <- getConstInfo x
+          let addCompiled = addHaskellCode x hs
           case theDef def of
-            Axiom{} -> do
-              ty <- haskellType $ defType def
-              reportSLn "tc.pragma.compile" 10 $ "Haskell type for " ++ show x ++ ": " ++ ty
-              addHaskellCode x ty hs
-            _   -> typeError $ GenericError "COMPILED directive only works on postulates"
+            Axiom{} -> addCompiled
+            Function{} -> addCompiled
+            _   -> typeError $ GenericError "COMPILED directive only works on postulates and functions"
+
         A.CompiledExportPragma x hs -> do
           def <- getConstInfo x
           let correct = case theDef def of
-                            -- Axiom{} -> do
-                            --   ty <- haskellType $ defType def
-                            --   reportSLn "tc.pragma.compile" 10 $ "Haskell type for " ++ show x ++ ": " ++ ty
-                            --   addHaskellCode x ty hs
                             Function{} -> True
                             Constructor{} -> False
                             _   -> False
           if not correct
             then typeError $ GenericError "COMPILED_EXPORT directive only works on functions"
-            else do
-              ty <- haskellType $ defType def
-              addHaskellExport x ty hs
-        A.CompiledEpicPragma x ep -> do
-          def <- getConstInfo x
-          case theDef def of
-            Axiom{} -> do
-              --ty <- haskellType $ defType def
-              --reportSLn "tc.pragma.compile" 10 $ "Haskell type for " ++ show x ++ ": " ++ ty
-              addEpicCode x ep
-            _   -> typeError $ GenericError "COMPILED_EPIC directive only works on postulates"
+            else addHaskellExport x hs
         A.CompiledJSPragma x ep ->
           addJSCode x ep
         A.CompiledUHCPragma x cr -> do
           def <- getConstInfo x
           case theDef def of
-            Axiom{} -> case parseCoreExpr cr of
-                    Left msg -> typeError $ GenericError $ "Could not parse COMPILED_UHC pragma: " ++ msg
-                    Right cr -> addCoreCode x cr
+            Axiom{} -> addCoreCode x cr
             _ -> typeError $ GenericError "COMPILED_UHC directive only works on postulates" -- only allow postulates for the time being
         A.CompiledDataUHCPragma x crd crcs -> do
           -- TODO mostly copy-paste from the CompiledDataPragma, should be refactored into a seperate function
@@ -717,30 +675,14 @@ checkPragma r p =
               "COMPILED_DATA_UHC directives must appear in the same module " ++
               "as their corresponding datatype definition,"
           case theDef def of
-            Datatype{dataCons = cs}
-              | length cs /= length crcs -> do
-                  let n_forms_are = case length crcs of
-                        1 -> "1 compiled form is"
-                        n -> show n ++ " compiled forms are"
-                      only | null crcs               = ""
-                           | length crcs < length cs = "only "
-                           | otherwise               = ""
-
-                  err <- fsep $ [prettyTCM x] ++ pwords ("has " ++ show (length cs) ++
-                                " constructors, but " ++ only ++ n_forms_are ++ " given [" ++ unwords crcs ++ "]")
-                  typeError $ GenericError $ show err
-              | otherwise -> do
-                -- Remark: core pragmas are not type-checked
-                dt' <- parseCoreData crd
-                cons' <- parseCoreConstrs dt' crcs
-                addCoreType x dt'
-                sequence_ $ zipWith addCoreConstr cs cons'
+            Datatype{dataCons = cs} -> addCoreType x crd crcs
             _ -> typeError $ GenericError "COMPILED_DATA_UHC on non datatype"
         A.StaticPragma x -> do
           def <- getConstInfo x
           case theDef def of
             Function{} -> markStatic x
             _          -> typeError $ GenericError "STATIC directive only works on functions"
+        A.InjectivePragma x -> markInjective x
         A.InlinePragma x -> do
           def <- getConstInfo x
           case theDef def of
@@ -748,6 +690,18 @@ checkPragma r p =
             _          -> typeError $ GenericError "INLINE directive only works on functions"
         A.OptionsPragma{} -> typeError $ GenericError $ "OPTIONS pragma only allowed at beginning of file, before top module declaration"
         A.DisplayPragma f ps e -> checkDisplayPragma f ps e
+        A.EtaPragma r -> do
+          let noRecord = typeError $ GenericError $
+                "ETA pragma is only applicable to coinductive records"
+          caseMaybeM (isRecord r) noRecord $ \case
+            Record{ recInduction = ind, recEtaEquality' = eta } -> do
+              unless (ind == Just CoInductive) $ noRecord
+              when (eta == Specified False) $ typeError $ GenericError $
+                "ETA pragram conflicts with no-eta-equality declaration"
+            _ -> __IMPOSSIBLE__
+          modifySignature $ updateDefinition r $ updateTheDef $ \case
+            def@Record{} -> def { recEtaEquality' = Specified True }
+            _ -> __IMPOSSIBLE__
 
 -- | Type check a bunch of mutual inductive recursive definitions.
 --
@@ -773,6 +727,7 @@ checkTypeSignature (A.ScopedDecl scope ds) = do
   setScope scope
   mapM_ checkTypeSignature ds
 checkTypeSignature (A.Axiom funSig i info mp x e) =
+  Bench.billTo [Bench.Definition x] $
   Bench.billTo [Bench.Typing, Bench.TypeSig] $
     let abstr = case Info.defAccess i of
           PrivateAccess{}
@@ -815,21 +770,21 @@ checkModuleArity m tel args = check tel args
           y    = absName btel
           tel  = absBody btel in
       case (argInfoHiding info, argInfoHiding info', name) of
-        (Instance, NotHidden, _) -> check tel args0
-        (Instance, Hidden, _)    -> check tel args0
-        (Instance, Instance, Nothing) -> check tel args
-        (Instance, Instance, Just x)
-          | x == y                -> check tel args
-          | otherwise             -> check tel args0
-        (Hidden, NotHidden, _)    -> check tel args0
-        (Hidden, Instance, _)     -> check tel args0
-        (Hidden, Hidden, Nothing) -> check tel args
+        (Instance{}, NotHidden, _)        -> check tel args0
+        (Instance{}, Hidden, _)           -> check tel args0
+        (Instance{}, Instance{}, Nothing) -> check tel args
+        (Instance{}, Instance{}, Just x)
+          | x == y                        -> check tel args
+          | otherwise                     -> check tel args0
+        (Hidden, NotHidden, _)            -> check tel args0
+        (Hidden, Instance{}, _)           -> check tel args0
+        (Hidden, Hidden, Nothing)         -> check tel args
         (Hidden, Hidden, Just x)
-          | x == y                -> check tel args
-          | otherwise             -> check tel args0
-        (NotHidden, NotHidden, _) -> check tel args
-        (NotHidden, Hidden, _)    -> bad
-        (NotHidden, Instance, _)    -> bad
+          | x == y                        -> check tel args
+          | otherwise                     -> check tel args0
+        (NotHidden, NotHidden, _)         -> check tel args
+        (NotHidden, Hidden, _)            -> bad
+        (NotHidden, Instance{}, _)        -> bad
 
 -- | Check an application of a section (top-level function, includes @'traceCall'@).
 checkSectionApplication
@@ -854,15 +809,25 @@ checkSectionApplication' i m1 (A.SectionApp ptel m2 args) copyInfo = do
   -- lambda-bound variables as additional parameters to the module.
   extraParams <- do
     mfv <- getCurrentModuleFreeVars
-    fv  <- size <$> getContextTelescope
+    fv  <- getContextSize
     return (fv - mfv)
-  when (extraParams > 0) $ reportSLn "tc.mod.apply" 30 $ "Extra parameters to " ++ show m1 ++ ": " ++ show extraParams
+  when (extraParams > 0) $ reportSLn "tc.mod.apply" 30 $ "Extra parameters to " ++ prettyShow m1 ++ ": " ++ show extraParams
   -- Type-check the LHS (ptel) of the module macro.
   checkTelescope ptel $ \ ptel -> do
     -- We are now in the context @ptel@.
-    -- Get the correct parameter telescope of @m2@.
+    -- Get the correct parameter telescope of @m2@. This is the fully lifted
+    -- telescope obtained by `lookupSection` instantiated with the module
+    -- parameters of `m2` currently in scope. For instance
+    -- ```
+    --  module _ (A : Set) where
+    --    module M (B : Set) where ...
+    --    module M' = M B
+    -- ```
+    -- In the application `M' = M B`, `tel = (A B : Set)` and
+    -- `moduleParamsToApply M = [A]`, so the resulting parameter telescope is
+    -- `tel' = (B : Set)`.
     tel <- lookupSection m2
-    vs  <- freeVarsToApply $ mnameToQName m2
+    vs  <- moduleParamsToApply m2
     let tel'  = apply tel vs
     -- Compute the remaining parameter telescope after stripping of
     -- the initial parameters that are determined by the @args@.
@@ -891,7 +856,7 @@ checkSectionApplication' i m1 (A.SectionApp ptel m2 args) copyInfo = do
       ]
     -- Andreas, 2014-04-06, Issue 1094:
     -- Add the section with well-formed telescope.
-    addContext aTel $ do
+    addContext (KeepNames aTel) $ do
       reportSDoc "tc.mod.apply" 80 $
         text "addSection" <+> prettyTCM m1 <+> (getContextTelescope >>= \ tel -> inTopContext (prettyTCM tel))
       addSection m1
@@ -903,13 +868,13 @@ checkSectionApplication' i m1 (A.SectionApp ptel m2 args) copyInfo = do
     args <- instantiateFull $ vs ++ ts
     let n = size aTel
     etaArgs <- inTopContext $ addContext aTel getContextArgs
-    addContext' aTel $
+    addContext' (KeepNames aTel) $
       applySection m1 (ptel `abstract` aTel) m2 (raise n args ++ etaArgs) copyInfo
 
 checkSectionApplication' i m1 (A.RecordModuleIFS x) copyInfo = do
   let name = mnameToQName x
   tel' <- lookupSection x
-  vs   <- freeVarsToApply name
+  vs   <- moduleParamsToApply x
   let tel = tel' `apply` vs
       args = teleArgs tel
 
@@ -923,7 +888,7 @@ checkSectionApplication' i m1 (A.RecordModuleIFS x) copyInfo = do
       -- Found last parameter: switch it to @Instance@.
       instFinal (ExtendTel (Dom info t) (Abs n EmptyTel)) =
                  ExtendTel (Dom ifo' t) (Abs n EmptyTel)
-        where ifo' = setHiding Instance info
+        where ifo' = makeInstance info
       -- Otherwise, keep searching for last parameter:
       instFinal (ExtendTel arg (Abs n tel)) =
                  ExtendTel arg (Abs n (instFinal tel))
@@ -944,10 +909,10 @@ checkSectionApplication' i m1 (A.RecordModuleIFS x) copyInfo = do
     -- , nest 2 $ text "args    =" <+> text (show args)
     ]
   when (tel == EmptyTel) $
-    typeError $ GenericError $ show (qnameToConcrete name) ++ " is not a parameterised section"
+    typeError $ GenericError $ prettyShow (qnameToConcrete name) ++ " is not a parameterised section"
 
   addContext' telInst $ do
-    vs <- freeVarsToApply name
+    vs <- moduleParamsToApply $ qnameModule name
     reportSDoc "tc.mod.apply" 20 $ vcat
       [ nest 2 $ text "vs      =" <+> sep (map prettyTCM vs)
       , nest 2 $ text "args    =" <+> sep (map (parens . prettyTCM) args)
@@ -1000,7 +965,7 @@ debugPrintDecl d = do
       case d of
         A.Section info mname tel ds -> do
           reportSLn "tc.decl" 45 $
-            "section " ++ show mname ++ " has "
+            "section " ++ prettyShow mname ++ " has "
               ++ show (length tel) ++ " parameters and "
               ++ show (length ds) ++ " declarations"
           reportSDoc "tc.decl" 45 $ prettyA $ A.Section info mname tel []
