@@ -10,7 +10,6 @@ import Control.Monad.State hiding (forM, mapM)
 import Control.Monad.Reader hiding (forM, mapM)
 
 import Data.Function
-import Data.List hiding (sort)
 import Data.Maybe
 import Data.Traversable
 import qualified Data.Set as Set
@@ -19,6 +18,7 @@ import Agda.Syntax.Common
 import qualified Agda.Syntax.Concrete as C
 import Agda.Syntax.Concrete (exprFieldA)
 import Agda.Syntax.Position
+import Agda.Syntax.Abstract.Pattern ( containsAbsurdPattern )
 import qualified Agda.Syntax.Abstract as A
 import qualified Agda.Syntax.Abstract.Views as A
 import Agda.Syntax.Internal as I
@@ -40,14 +40,14 @@ import Agda.TypeChecking.Patterns.Abstract (expandPatternSynonyms)
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Free
-import Agda.TypeChecking.CheckInternal (checkType)
+import Agda.TypeChecking.CheckInternal (checkType, inferSort)
 import Agda.TypeChecking.With
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Injectivity
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.SizedTypes.Solve
 import Agda.TypeChecking.RecordPatterns
-import Agda.TypeChecking.CompiledClause (CompiledClauses(..))
+import Agda.TypeChecking.CompiledClause (CompiledClauses'(..))
 import Agda.TypeChecking.CompiledClause.Compile
 import Agda.TypeChecking.Primitive hiding (Nat)
 
@@ -57,12 +57,14 @@ import Agda.TypeChecking.Rules.LHS.Problem         ( AsBinding(..) )
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl ( checkDecls )
 
 import Agda.Utils.Except ( MonadError(catchError, throwError) )
+import Agda.Utils.Functor
 import Agda.Utils.Lens
 import Agda.Utils.Maybe ( whenNothing )
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty ( prettyShow )
+import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
-import Agda.Utils.Functor
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -114,15 +116,15 @@ isAlias cs t =
   where
     isMeta (MetaV x _) = Just x
     isMeta _           = Nothing
-    trivialClause [A.Clause (A.LHS i (A.LHSHead f []) []) _ (A.RHS e mc) [] _] = Just (e, mc)
+    trivialClause [A.Clause (A.LHS i (A.LHSHead f []) []) _ _ (A.RHS e mc) [] _] = Just (e, mc)
     trivialClause _ = Nothing
 
 -- | Check a trivial definition of the form @f = e@
 checkAlias :: Type -> ArgInfo -> Delayed -> Info.DefInfo -> QName -> A.Expr -> Maybe C.Expr -> TCM ()
 checkAlias t' ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
   reportSDoc "tc.def.alias" 10 $ text "checkAlias" <+> vcat
-    [ text (show name) <+> colon  <+> prettyTCM t'
-    , text (show name) <+> equals <+> prettyTCM e
+    [ text (prettyShow name) <+> colon  <+> prettyTCM t'
+    , text (prettyShow name) <+> equals <+> prettyTCM e
     ]
 
 {-
@@ -155,17 +157,25 @@ checkAlias t' ai delayed i name e mc = atClause name 0 (A.RHS e mc) $ do
                    $ set funMacro (Info.defMacro i == MacroDef) $
                      emptyFunction
                       { funClauses = [ Clause  -- trivial clause @name = v@
-                          { clauseRange     = getRange i
+                          { clauseLHSRange  = getRange i
+                          , clauseFullRange = getRange i
                           , clauseTel       = EmptyTel
                           , namedClausePats = []
                           , clauseBody      = Just $ bodyMod v
                           , clauseType      = Just $ Arg ai t
                           , clauseCatchall  = False
+                          , clauseUnreachable = Just False
                           } ]
                       , funCompiled = Just $ Done [] $ bodyMod v
                       , funDelayed  = delayed
                       , funAbstr    = Info.defAbstract i
                       }
+
+  -- Andreas, 2017-01-01, issue #2372:
+  -- Add the definition to the instance table, if needed, to update its type.
+  when (Info.defInstance i == InstanceDef) $ do
+    addTypedInstance name t
+
   reportSDoc "tc.def.alias" 20 $ text "checkAlias: leaving"
 
 
@@ -195,7 +205,7 @@ checkFunDefS :: Type             -- ^ the type we expect the function to have
              -> Maybe Substitution -- ^ substitution (from with abstraction) that needs to be applied to module parameters
              -> [A.Clause]       -- ^ the clauses to check
              -> TCM ()
-checkFunDefS t ai delayed extlam with i name withSub cs =
+checkFunDefS t ai delayed extlam with i name withSub cs = do
 
     traceCall (CheckFunDef (getRange i) (qnameName name) cs) $ do   -- TODO!! (qnameName)
         reportSDoc "tc.def.fun" 10 $
@@ -236,18 +246,18 @@ checkFunDefS t ai delayed extlam with i name withSub cs =
               inTopContext $ addClauses name [c]
               return c
 
-        reportSDoc "tc.def.fun" 70 $
+        reportSDoc "tc.def.fun" 70 $ inTopContext $ do
           sep $ [ text "checked clauses:" ] ++ map (nest 2 . text . show) cs
 
         -- After checking, remove the clauses again.
         -- (Otherwise, @checkInjectivity@ loops for issue 801).
         modifyFunClauses name (const [])
 
-        reportSDoc "tc.cc" 25 $ do
+        reportSDoc "tc.cc" 25 $ inTopContext $ do
           sep [ text "clauses before injectivity test"
               , nest 2 $ prettyTCM $ map (QNamed name) cs  -- broken, reify (QNamed n cl) expect cl to live at top level
               ]
-        reportSDoc "tc.cc" 60 $ do
+        reportSDoc "tc.cc" 60 $ inTopContext $ do
           sep [ text "raw clauses: "
               , nest 2 $ sep $ map (text . show . QNamed name) cs
               ]
@@ -268,7 +278,7 @@ checkFunDefS t ai delayed extlam with i name withSub cs =
         inv <- Bench.billTo [Bench.Injectivity] $
           checkInjectivity name cs
 
-        reportSDoc "tc.cc" 15 $ do
+        reportSDoc "tc.cc" 15 $ inTopContext $ do
           sep [ text "clauses before compilation"
               , nest 2 $ sep $ map (prettyTCM . QNamed name) cs
               ]
@@ -282,7 +292,7 @@ checkFunDefS t ai delayed extlam with i name withSub cs =
         cc <- Bench.billTo [Bench.Coverage] $
           inTopContext $ compileClauses (Just (name, fullType)) cs
 
-        reportSDoc "tc.cc" 10 $ do
+        reportSDoc "tc.cc" 60 $ inTopContext $ do
           sep [ text "compiled clauses of" <+> prettyTCM name
               , nest 2 $ text (show cc)
               ]
@@ -307,9 +317,6 @@ checkFunDefS t ai delayed extlam with i name withSub cs =
              , funCopatternLHS   = isCopatternLHS cs
              }
 
-        -- Andreas 2012-02-13: postpone polarity computation until after positivity check
-        -- computePolarity name
-
         reportSDoc "tc.def.fun" 10 $ do
           sep [ text "added " <+> prettyTCM name <+> text ":"
               , nest 2 $ prettyTCM . defType =<< getConstInfo name
@@ -325,7 +332,7 @@ useTerPragma def@Defn{ defName = name, theDef = fun@Function{}} = do
         Terminating    -> Just True
         _              -> Nothing
   reportSLn "tc.fundef" 30 $ unlines $
-    [ "funTerminates of " ++ show name ++ " set to " ++ show terminates
+    [ "funTerminates of " ++ prettyShow name ++ " set to " ++ show terminates
     , "  tc = " ++ show tc
     ]
   return $ def { theDef = fun { funTerminates = terminates }}
@@ -335,8 +342,8 @@ useTerPragma def = return def
 -- | Insert some patterns in the in with-clauses LHS of the given RHS
 insertPatterns :: [A.Pattern] -> A.RHS -> A.RHS
 insertPatterns pats (A.WithRHS aux es cs) = A.WithRHS aux es (map insertToClause cs)
-    where insertToClause (A.Clause (A.LHS i lhscore ps) dots rhs ds catchall)
-              = A.Clause (A.LHS i lhscore (pats ++ ps)) dots (insertPatterns pats rhs) ds catchall
+    where insertToClause (A.Clause (A.LHS i lhscore ps) dots sdots rhs ds catchall)
+              = A.Clause (A.LHS i lhscore (pats ++ ps)) dots sdots (insertPatterns pats rhs) ds catchall
 insertPatterns pats (A.RewriteRHS qes rhs wh) = A.RewriteRHS qes (insertPatterns pats rhs) wh
 insertPatterns pats rhs = rhs
 
@@ -347,6 +354,7 @@ data WithFunctionProblem
     { wfParentName :: QName                -- ^ Parent function name.
     , wfName       :: QName                -- ^ With function name.
     , wfParentType :: Type                 -- ^ Type of the parent function.
+    , wfParentTel  :: Telescope            -- ^ Context of the parent patterns.
     , wfBeforeTel  :: Telescope            -- ^ Types of arguments to the with function before the with expressions (needed vars).
     , wfAfterTel   :: Telescope            -- ^ Types of arguments to the with function after the with expressions (unneeded vars).
     , wfExprs      :: [Term]               -- ^ With and rewrite expressions.
@@ -368,7 +376,7 @@ checkClause
   -> A.SpineClause -- ^ Clause.
   -> TCM Clause    -- ^ Type-checked clause.
 
-checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots rhs0 wh catchall) = do
+checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots strippedDots rhs0 wh catchall) = do
     reportSDoc "tc.lhs.top" 30 $ text "Checking clause" $$ prettyA c
     unless (null withPats) $
       typeError $ UnexpectedWithPatterns withPats
@@ -379,7 +387,7 @@ checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots rhs0 w
         text "namedDots:" <+> vcat [ prettyTCM x <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.NamedDot x v a <- namedDots ]
       -- Not really an as-pattern, but this does the right thing.
       bindAsPatterns [ AsB x v a | A.NamedDot x v a <- namedDots ] $
-        checkLeftHandSide (CheckPatternShadowing c) (Just x) aps t withSub $ \ lhsResult@(LHSResult npars delta ps trhs patSubst asb) -> do
+        checkLeftHandSide (CheckPatternShadowing c) (Just x) aps t withSub strippedDots $ \ lhsResult@(LHSResult npars delta ps trhs patSubst asb) -> do
         -- Note that we might now be in irrelevant context,
         -- in case checkLeftHandSide walked over an irrelevant projection pattern.
 
@@ -404,7 +412,8 @@ checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots rhs0 w
             updateRHS (A.WithRHS q es cs)       = A.WithRHS q es (map updateClause cs)
             updateRHS (A.RewriteRHS qes rhs wh) = A.RewriteRHS qes (updateRHS rhs) wh
 
-            updateClause (A.Clause f dots rhs wh ca) = A.Clause f (applySubst patSubst dots) (updateRHS rhs) wh ca
+            updateClause (A.Clause f dots sdots rhs wh ca) =
+              A.Clause f (applySubst patSubst dots) (applySubst patSubst sdots) (updateRHS rhs) wh ca
 
         (body, with) <- bindAsPatterns asb $ checkWhere wh $ checkRHS i x aps t' lhsResult rhs
 
@@ -413,29 +422,42 @@ checkClause t withSub c@(A.Clause (A.SpineLHS i x aps withPats) namedDots rhs0 w
         -- to child).
         inTopContext $ Bench.billTo [Bench.Typing, Bench.With] $ checkWithFunction cxtNames with
 
-        reportSDoc "tc.lhs.top" 10 $ escapeContext (size delta) $ vcat
+        reportSDoc "tc.lhs.top" 10 $ vcat
           [ text "Clause before translation:"
           , nest 2 $ vcat
-            [ text "delta =" <+> prettyTCM delta
-            , text "ps    =" <+> text (show ps)
-            , text "body  =" <+> text (show body)
+            [ text "delta =" <+> do escapeContext (size delta) $ prettyTCM delta
+            , text "ps    =" <+> do P.fsep <$> prettyTCMPatterns ps
             , text "body  =" <+> maybe (text "_|_") prettyTCM body
             ]
           ]
 
+        reportSDoc "tc.lhs.top" 60 $ escapeContext (size delta) $ vcat
+          [ text "Clause before translation (raw):"
+          , nest 2 $ vcat
+            [ text "ps    =" <+> text (show ps)
+            , text "body  =" <+> text (show body)
+            ]
+          ]
+
         -- compute body modification for irrelevant definitions, see issue 610
         rel <- asks envRelevance
         let bodyMod body = case rel of
               Irrelevant -> dontCare <$> body
               _          -> body
 
+        -- absurd clauses don't define computational behaviour, so it's fine to
+        -- treat them as catchalls.
+        let catchall' = catchall || isNothing body
+
         return $
-          Clause { clauseRange     = getRange i
+          Clause { clauseLHSRange  = getRange i
+                 , clauseFullRange = getRange c
                  , clauseTel       = killRange delta
                  , namedClausePats = ps
                  , clauseBody      = bodyMod body
                  , clauseType      = Just trhs
-                 , clauseCatchall  = catchall
+                 , clauseCatchall  = catchall'
+                 , clauseUnreachable = Nothing -- we don't know yet
                  }
 
 -- | Type check the @with@ and @rewrite@ lhss and/or the rhs.
@@ -451,7 +473,7 @@ checkRHS
                                               -- Note: the as-bindings are already bound (in checkClause)
 checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS rhs0
   where
-  absurdPat = any (containsAbsurdPattern . namedArg) aps
+  absurdPat = containsAbsurdPattern aps
   handleRHS rhs =
     case rhs of
 
@@ -504,31 +526,37 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
 
         t' <- reduce =<< instantiateFull eqt
         (eqt,rewriteType,rewriteFrom,rewriteTo) <- equalityView t' >>= \case
-          eqt@(EqualityType s _eq _level dom a b) -> return (eqt, El s (unArg dom), unArg a, unArg b)
+          eqt@(EqualityType _s _eq _params (Arg _ dom) a b) -> do
+            s <- inferSort dom
+            return (eqt, El s dom, unArg a, unArg b)
+            -- Note: the sort _s of the equality need not be the sort of the type @dom@!
           OtherType{} -> typeError . GenericDocError =<< do
             text "Cannot rewrite by equation of type" <+> prettyTCM t'
 
         -- Get the name of builtin REFL.
 
         Con reflCon _ [] <- ignoreSharing <$> primRefl
+        reflInfo <- fmap (setOrigin Inserted) <$> getReflArgInfo reflCon
 
-        -- Andreas, 2014-05-17  Issue 1110:
-        -- Rewriting with @refl@ has no effect, but gives an
-        -- incomprehensible error message about the generated
-        -- with clause. Thus, we rather do simply nothing if
-        -- rewriting with @refl@ is attempted.
-
-        let isReflProof = do
-             v <- reduce proof
-             case ignoreSharing v of
-               Con c _ [] | c == reflCon -> return True
-               _ -> return False
-
-        ifM isReflProof recurse $ {- else -} do
+        -- Andreas, 2017-01-11:
+        -- The test for refl is obsolete after fixes of #520 and #1740.
+        -- -- Andreas, 2014-05-17  Issue 1110:
+        -- -- Rewriting with @refl@ has no effect, but gives an
+        -- -- incomprehensible error message about the generated
+        -- -- with clause. Thus, we rather do simply nothing if
+        -- -- rewriting with @refl@ is attempted.
+        -- let isReflProof = do
+        --      v <- reduce proof
+        --      case ignoreSharing v of
+        --        Con c _ [] | c == reflCon -> return True
+        --        _ -> return False
+        -- ifM isReflProof recurse $ {- else -} do
 
         -- Process 'rewrite' clause like a suitable 'with' clause.
 
-        let reflPat  = A.ConP (ConPatInfo ConOCon patNoRange) (AmbQ [conName reflCon]) []
+        -- The REFL constructor might have an argument
+        let reflPat  = A.ConP (ConPatInfo ConOCon patNoRange) (AmbQ [conName reflCon]) $
+              maybeToList $ fmap (\ ai -> Arg ai $ unnamed $ A.WildP patNoRange) reflInfo
 
         -- Andreas, 2015-12-25  Issue #1740:
         -- After the fix of #520, rewriting with a reflexive equation
@@ -547,7 +575,7 @@ checkRHS i x aps t lhsResult@(LHSResult _ delta ps trhs _ _asb) rhs0 = handleRHS
               | otherwise = (A.RewriteRHS qes rhs' wh, [])
             -- Andreas, 2014-03-05 kill range of copied patterns
             -- since they really do not have a source location.
-            cs       = [A.Clause (A.LHS i (A.LHSHead x (killRange aps)) pats) [] rhs'' outerWhere False]
+            cs       = [A.Clause (A.LHS i (A.LHSHead x (killRange aps)) pats) [] [] rhs'' outerWhere False]
 
         checkWithRHS x qname t lhsResult [withExpr] [withType] cs
 
@@ -627,9 +655,9 @@ checkWithRHS x aux t (LHSResult npars delta ps trhs _ _asb) vs0 as cs = Bench.bi
         let n = size us
             m = size delta
             -- First the variables bound outside this definition
-            (us0, us1') = genericSplitAt (n - m) us
+            (us0, us1') = splitAt (n - m) us
             -- Then permute the rest and grab those needed to for the with arguments
-            (us1, us2)  = genericSplitAt (size delta1) $ permute perm' us1'
+            (us1, us2)  = splitAt (size delta1) $ permute perm' us1'
             -- Now stuff the with arguments in between and finish with the remaining variables
             v    = Def aux $ map Apply $ us0 ++ us1 ++ map defaultArg withArgs ++ us2
         -- Andreas, 2013-02-26 add with-name to signature for printing purposes
@@ -654,11 +682,12 @@ checkWithRHS x aux t (LHSResult npars delta ps trhs _ _asb) vs0 as cs = Bench.bi
         reportSDoc "tc.with.top" 20 $
           text "              body" <+> prettyTCM v
 
-        return (Just v, WithFunction x aux t delta1 delta2 vs as t' ps npars perm' perm finalPerm cs)
+        return (Just v, WithFunction x aux t delta delta1 delta2 vs as t' ps npars perm' perm finalPerm cs)
 
+-- | Invoked in empty context.
 checkWithFunction :: [Name] -> WithFunctionProblem -> TCM ()
 checkWithFunction _ NoWithFunction = return ()
-checkWithFunction cxtNames (WithFunction f aux t delta1 delta2 vs as b qs npars perm' perm finalPerm cs) = do
+checkWithFunction cxtNames (WithFunction f aux t delta delta1 delta2 vs as b qs npars perm' perm finalPerm cs) = do
 
   let -- Δ₁ ws Δ₂ ⊢ withSub : Δ′    (where Δ′ is the context of the parent lhs)
       withSub :: Substitution
@@ -673,7 +702,7 @@ checkWithFunction cxtNames (WithFunction f aux t delta1 delta2 vs as b qs npars
       , text "as     =" <+> addContext delta1 (prettyTCM as)
       , text "vs     =" <+> do addContext delta1 $ prettyTCM vs
       , text "b      =" <+> do addContext delta1 $ addContext delta2 $ prettyTCM b
-      , text "qs     =" <+> prettyList (map pretty qs)
+      , text "qs     =" <+> do addContext delta $ prettyTCMPatternList qs
       , text "perm'  =" <+> text (show perm')
       , text "perm   =" <+> text (show perm)
       , text "fperm  =" <+> text (show finalPerm)
@@ -735,7 +764,7 @@ checkWithFunction cxtNames (WithFunction f aux t delta1 delta2 vs as b qs npars
 
   -- Construct the body for the with function
   cs <- return $ map (A.lhsToSpine) cs
-  cs <- buildWithFunction cxtNames f aux t qs npars withSub finalPerm (size delta1) n cs
+  cs <- buildWithFunction cxtNames f aux t delta qs npars withSub finalPerm (size delta1) n cs
   cs <- return $ map (A.spineToLhs) cs
 
   -- Check the with function
@@ -778,22 +807,6 @@ newSection m tel cont = do
 
     withCurrentModule m cont
 
-
--- | Check if a pattern contains an absurd pattern. For instance, @suc ()@
-containsAbsurdPattern :: A.Pattern -> Bool
-containsAbsurdPattern p = case p of
-    A.AbsurdP _   -> True
-    A.VarP _      -> False
-    A.WildP _     -> False
-    A.DotP _ _ _  -> False
-    A.LitP _      -> False
-    A.AsP _ _ p   -> containsAbsurdPattern p
-    A.ConP _ _ ps -> any (containsAbsurdPattern . namedArg) ps
-    A.RecP _ fs   -> any (containsAbsurdPattern . (^. exprFieldA)) fs
-    A.ProjP{}     -> False
-    A.DefP _ _ ps -> any (containsAbsurdPattern . namedArg) ps
-    A.PatternSynP _ _ _ -> __IMPOSSIBLE__ -- False
-
 -- | Set the current clause number.
 atClause :: QName -> Int -> A.RHS -> TCM a -> TCM a
 atClause name i rhs = local $ \ e -> e { envClause = IPClause name i rhs }
