@@ -29,7 +29,7 @@ import Agda.Syntax.Internal as I
 import Agda.Syntax.Internal.Pattern
 import Agda.Syntax.Abstract (IsProjP(..))
 import qualified Agda.Syntax.Abstract as A
-import Agda.Syntax.Abstract.Views (asView)
+import Agda.Syntax.Abstract.Views (asView, deepUnscope)
 import Agda.Syntax.Common as Common
 import Agda.Syntax.Info as A
 import Agda.Syntax.Position
@@ -69,6 +69,7 @@ import Agda.Utils.ListT
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -79,7 +80,7 @@ import Agda.Utils.Impossible
 flexiblePatterns :: [NamedArg A.Pattern] -> TCM FlexibleVars
 flexiblePatterns nps = do
   forMaybeM (zip (downFrom $ length nps) nps) $ \ (i, Arg ai p) -> do
-    runMaybeT $ (\ f -> FlexibleVar (getHiding ai) f (Just i) i) <$> maybeFlexiblePattern p
+    runMaybeT $ (\ f -> FlexibleVar (getHiding ai) (getOrigin ai) f (Just i) i) <$> maybeFlexiblePattern p
 
 -- | A pattern is flexible if it is dotted or implicit, or a record pattern
 --   with only flexible subpatterns.
@@ -87,18 +88,28 @@ class IsFlexiblePattern a where
   maybeFlexiblePattern :: a -> MaybeT TCM FlexibleVarKind
 
   isFlexiblePattern :: a -> TCM Bool
-  isFlexiblePattern p = isJust <$> runMaybeT (maybeFlexiblePattern p)
+  isFlexiblePattern p =
+    maybe False notOtherFlex <$> runMaybeT (maybeFlexiblePattern p)
+    where
+    notOtherFlex = \case
+      RecordFlex fls -> all notOtherFlex fls
+      ImplicitFlex   -> True
+      DotFlex        -> True
+      OtherFlex      -> False
 
 instance IsFlexiblePattern A.Pattern where
-  maybeFlexiblePattern p =
+  maybeFlexiblePattern p = do
+    reportSDoc "tc.lhs.flex" 30 $ text "maybeFlexiblePattern" <+> prettyA p
+    reportSDoc "tc.lhs.flex" 60 $ text "maybeFlexiblePattern (raw) " <+> (text . show . deepUnscope) p
     case p of
       A.DotP{}  -> return DotFlex
       A.VarP{}  -> return ImplicitFlex
       A.WildP{} -> return ImplicitFlex
       A.AsP _ _ p -> maybeFlexiblePattern p
       A.ConP _ (A.AmbQ [c]) qs
-        -> ifM (isNothing <$> isRecordConstructor c) mzero {-else-}
+        -> ifM (isNothing <$> isRecordConstructor c) (return OtherFlex) {-else-}
              (maybeFlexiblePattern qs)
+      A.LitP{}  -> return OtherFlex
       _ -> mzero
 
 instance IsFlexiblePattern (I.Pattern' a) where
@@ -110,6 +121,7 @@ instance IsFlexiblePattern (I.Pattern' a) where
         | Just _            <- conPRecord i -> maybeFlexiblePattern ps
         | otherwise -> mzero
       I.VarP{}  -> mzero
+      I.AbsurdP{} -> mzero
       I.LitP{}  -> mzero
       I.ProjP{} -> mzero
 
@@ -137,7 +149,7 @@ updateInPatterns as ps qs = do
   reportSDoc "tc.lhs.top" 20 $ text "updateInPatterns" <+> nest 2 (vcat
     [ text "as      =" <+> prettyList (map prettyTCM as)
     , text "ps      =" <+> prettyList (map prettyA ps)
-    , text "qs      =" <+> prettyList (map (text . show) qs)
+    , text "qs      =" <+> prettyList (map pretty qs)
     ])
   first (map snd . IntMap.toDescList) <$> updates as ps qs
   where
@@ -155,28 +167,34 @@ updateInPatterns as ps qs = do
         A.DotP _ _ e -> return (IntMap.empty, [DPI Nothing  (Just e) u a])
         A.WildP _  -> return (IntMap.empty, [DPI Nothing  Nothing  u a])
         A.VarP x   -> return (IntMap.empty, [DPI (Just x) Nothing  u a])
-        A.ConP _ (A.AmbQ [c]) qs -> do
-          Def r es  <- ignoreSharing <$> reduce (unEl $ unDom a)
-          let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
-          (ftel, us) <- etaExpandRecord r vs u
-          qs <- insertImplicitPatterns ExpandLast qs ftel
-          let instTel EmptyTel _                   = []
-              instTel (ExtendTel arg tel) (u : us) = arg : instTel (absApp tel u) us
-              instTel ExtendTel{} []               = __IMPOSSIBLE__
-              bs0 = instTel ftel (map unArg us)
-              -- Andreas, 2012-09-19 propagate relevance info to dot patterns
-              bs  = map (mapRelevance (composeRelevance (getRelevance a))) bs0
-          updates bs qs (map (DotP . unArg) us `withArgsFrom` teleArgNames ftel)
+        p@(A.ConP _ (A.AmbQ [c]) qs) -> ifM (isNothing <$> isRecordConstructor c)
+          (return (IntMap.empty, [DPI Nothing (Just $ A.patternToExpr p) u a]))
+          (do
+            Def r es  <- ignoreSharing <$> reduce (unEl $ unDom a)
+            let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+            (ftel, us) <- etaExpandRecord r vs u
+
+            qs <- insertImplicitPatterns ExpandLast qs ftel
+            reportSDoc "tc.lhs.imp" 20 $
+              text "insertImplicitPatternsT returned" <+> fsep (map prettyA qs)
+
+            let instTel EmptyTel _                   = []
+                instTel (ExtendTel arg tel) (u : us) = arg : instTel (absApp tel u) us
+                instTel ExtendTel{} []               = __IMPOSSIBLE__
+                bs0 = instTel ftel (map unArg us)
+                -- Andreas, 2012-09-19 propagate relevance info to dot patterns
+                bs  = map (mapRelevance (composeRelevance (getRelevance a))) bs0
+            updates bs qs (map (DotP . unArg) us `withArgsFrom` teleArgNames ftel))
+        p@A.ConP{} -> return (IntMap.empty, [DPI Nothing (Just $ A.patternToExpr p) u a])
+        p@A.LitP{} -> return (IntMap.empty, [DPI Nothing (Just $ A.patternToExpr p) u a])
         A.AsP         _ _ _ -> __IMPOSSIBLE__
-        A.ConP        _ _ _ -> __IMPOSSIBLE__
         A.RecP        _ _   -> __IMPOSSIBLE__
         A.ProjP       _ _ _ -> __IMPOSSIBLE__
         A.DefP        _ _ _ -> __IMPOSSIBLE__
         A.AbsurdP     _     -> __IMPOSSIBLE__
-        A.LitP        _     -> __IMPOSSIBLE__
         A.PatternSynP _ _ _ -> __IMPOSSIBLE__
       -- Case: the unifier eta-expanded the variable
-      ConP c cpi qs -> do
+      ConP _c _cpi qs -> do
         Def r es <- ignoreSharing <$> reduce (unEl $ unDom a)
         def      <- theDef <$> getConstInfo r
         let pars = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
@@ -194,15 +212,16 @@ updateInPatterns as ps qs = do
                   _            -> []
         second (dpi++) <$>
           updates as (projectInPat p fs) (map (fmap namedThing) qs)
+      AbsurdP{}  -> __IMPOSSIBLE__
       LitP _     -> __IMPOSSIBLE__
       ProjP{}    -> __IMPOSSIBLE__
 
     projectInPat :: NamedArg A.Pattern -> [Arg QName] -> [NamedArg A.Pattern]
     projectInPat p fs = case namedThing (unArg p) of
-      A.VarP x            -> map (makeDotField (PatRange $ getRange x)) fs
-      A.ConP cpi _ nps    -> nps
+      A.VarP x            -> map (makeWildField (PatRange $ getRange x)) fs
+      A.ConP _ _ nps      -> nps
       A.WildP pi          -> map (makeWildField pi) fs
-      A.DotP pi _ e       -> map (makeDotField pi) fs
+      A.DotP pi o e       -> map (makeDotField pi o) fs
       A.ProjP _ _ _       -> __IMPOSSIBLE__
       A.DefP _ _ _        -> __IMPOSSIBLE__
       A.AsP _ _ _         -> __IMPOSSIBLE__
@@ -212,32 +231,41 @@ updateInPatterns as ps qs = do
       A.RecP _ _          -> __IMPOSSIBLE__
       where
         makeWildField pi (Arg fi f) = Arg fi $ unnamed $ A.WildP pi
-        makeDotField pi (Arg fi f) = Arg fi $ unnamed $
-          A.DotP pi Inserted $ A.Underscore underscoreInfo
+        makeDotField pi o (Arg fi f) = Arg fi $ unnamed $
+          A.DotP pi o $ A.Underscore underscoreInfo
           where
             underscoreInfo = A.MetaInfo
               { A.metaRange          = getRange pi
               , A.metaScope          = emptyScopeInfo
               , A.metaNumber         = Nothing
-              , A.metaNameSuggestion = show $ A.nameConcrete $ qnameName f
+              , A.metaNameSuggestion = prettyShow $ qnameName f
               }
 
 
--- | Check if a problem is solved. That is, if the patterns are all variables.
+-- | Check if a problem is solved.
+--   That is, if the patterns are all variables,
+--   and there is no 'problemRest'.
 isSolvedProblem :: Problem -> Bool
 isSolvedProblem problem = null (restPats $ problemRest problem) &&
-    all (isSolved . snd . asView . namedArg) (problemInPat problem)
+  problemAllVariables problem
+
+-- | Check if a problem consists only of variable patterns.
+--   (Includes the 'problemRest').
+problemAllVariables :: Problem -> Bool
+problemAllVariables problem =
+    all (isSolved . snd . asView . namedArg) $
+      restPats (problemRest problem) ++ problemInPat problem
   where
     -- need further splitting:
     isSolved A.ConP{}        = False
     isSolved A.LitP{}        = False
     isSolved A.ProjP{}       = False
     isSolved A.RecP{}        = False  -- record pattern
+    isSolved A.AbsurdP{}     = False
     -- solved:
     isSolved A.VarP{}        = True
     isSolved A.WildP{}       = True
     isSolved A.DotP{}        = True
-    isSolved A.AbsurdP{}     = True
     -- impossible:
     isSolved A.DefP{}        = __IMPOSSIBLE__
     isSolved A.AsP{}         = __IMPOSSIBLE__  -- removed by asView
@@ -268,15 +296,15 @@ noShadowingOfConstructors mkCall problem =
   noShadowing (A.ProjP     {}) t = return ()  -- projection pattern
   noShadowing (A.DefP      {}) t = __IMPOSSIBLE__
   noShadowing (A.DotP      {}) t = return ()
-  noShadowing (A.AsP       {}) t = __IMPOSSIBLE__
+  noShadowing (A.AsP       {}) t = __IMPOSSIBLE__ -- removed by asView
   noShadowing (A.LitP      {}) t = __IMPOSSIBLE__
   noShadowing (A.PatternSynP {}) t = __IMPOSSIBLE__
   noShadowing (A.VarP x)       t = do
     reportSDoc "tc.lhs.shadow" 30 $ vcat
-      [ text $ "checking whether pattern variable " ++ show x ++ " shadows a constructor"
+      [ text $ "checking whether pattern variable " ++ prettyShow x ++ " shadows a constructor"
       , nest 2 $ text "type of variable =" <+> prettyTCM t
       ]
-    reportSLn "tc.lhs.shadow" 70 $ "  t = " ++ show t
+    reportSDoc "tc.lhs.shadow" 70 $ nest 2 $ text "t =" <+> pretty t
     t <- reduce t
     case t of
       Def t _ -> do
@@ -327,11 +355,11 @@ checkDotPattern (DPI _ (Just e) v (Dom info a)) =
         ]
   applyRelevanceToContext (argInfoRelevance info) $ do
     u <- checkExpr e a
-    reportSDoc "tc.lhs.dot" 30 $
+    reportSDoc "tc.lhs.dot" 50 $
       sep [ text "equalTerm"
-          , nest 2 $ text $ show a
-          , nest 2 $ text $ show u
-          , nest 2 $ text $ show v
+          , nest 2 $ pretty a
+          , nest 2 $ pretty u
+          , nest 2 $ pretty v
           ]
     -- Should be ok to do noConstraints here
     noConstraints $ equalTerm a u v
@@ -397,12 +425,13 @@ checkLeftoverDotPatterns ps vs as dpi = do
       A.VarP _     -> return idv
       A.WildP _    -> return idv
       A.AbsurdP _  -> return idv
+      -- Andreas, 2017-01-18, issue #2413, AsP is not __IMPOSSIBLE__
+      A.AsP _ _ p0 -> checkUserDot (setNamedArg p p0) v a idv
       A.ConP _ _ _ -> __IMPOSSIBLE__
       A.LitP _     -> __IMPOSSIBLE__
       A.ProjP _ _ _-> __IMPOSSIBLE__
       A.DefP _ _ _ -> __IMPOSSIBLE__
       A.RecP _ _   -> __IMPOSSIBLE__
-      A.AsP  _ _ _ -> __IMPOSSIBLE__
       A.PatternSynP _ _ _ -> __IMPOSSIBLE__
 
     gatherImplicitDotVars :: DotPatternInst -> TCM [(Int,Projectns)]
@@ -431,11 +460,11 @@ checkLeftoverDotPatterns ps vs as dpi = do
      reportSDoc "tc.lhs.dot" 40 $ vcat
        [ text "undotImplicitVar"
        , nest 2 $ vcat
-         [ text $ "i  =  " ++ show i
-         , text   "fs = " <+> sep (map (prettyTCM . snd) fs)
-         , text   "a  = " <+> prettyTCM a
-         , text $ "raw=  "  ++ show a
-         , text $ "idv=  "  ++ show idv
+         [ text "i  =" <+> pshow i
+         , text "fs =" <+> sep (map (prettyTCM . snd) fs)
+         , text "a  =" <+> prettyTCM a
+         , text "raw=" <+> pretty a
+         , text "idv=" <+> pshow idv
          ]
        ]
      case lookupImplicitDotVar (i,fs) idv of
@@ -473,9 +502,9 @@ bindLHSVars []        tel@ExtendTel{}  _   = do
   __IMPOSSIBLE__
 bindLHSVars (_ : _)   EmptyTel         _   = __IMPOSSIBLE__
 bindLHSVars []        EmptyTel         ret = ret
-bindLHSVars (p : ps) (ExtendTel a tel) ret = do
+bindLHSVars (p : ps) tel0@(ExtendTel a tel) ret = do
   -- see test/Fail/WronHidingInLHS:
-  unless (getHiding p == getHiding a) $ typeError WrongHidingInLHS
+  unless (sameHiding p a) $ typeError WrongHidingInLHS
 
   case namedArg p of
     A.VarP x      -> addContext (x, a) $ bindLHSVars ps (absBody tel) ret
@@ -484,16 +513,14 @@ bindLHSVars (p : ps) (ExtendTel a tel) ret = do
                  -- introduces unwanted underscores in error messages
                  -- (Andreas, 2015-05-28)
     A.DotP _ _ _  -> bindDummy (absName tel)
-    A.AbsurdP pi  -> do
-      -- Andreas, 2012-03-15: allow postponement of emptyness check
-      isEmptyType (getRange pi) $ unDom a
-      -- OLD CODE: isReallyEmptyType $ unArg a
-      bindDummy (absName tel)
+    A.AbsurdP pi  -> __IMPOSSIBLE__
+    -- Andreas, 2017-01-18, issue #2413
+    -- A.AsP is not __IMPOSSIBLE__
+    A.AsP _ _ p0    -> bindLHSVars (setNamedArg p p0 : ps) tel0 ret
     A.ConP{}        -> __IMPOSSIBLE__
     A.RecP{}        -> __IMPOSSIBLE__
     A.ProjP{}       -> __IMPOSSIBLE__
     A.DefP{}        -> __IMPOSSIBLE__
-    A.AsP{}         -> __IMPOSSIBLE__
     A.LitP{}        -> __IMPOSSIBLE__
     A.PatternSynP{} -> __IMPOSSIBLE__
     where
@@ -560,16 +587,18 @@ checkLeftHandSide
      -- ^ The expected type @a = Γ → b@.
   -> Maybe Substitution
      -- ^ Module parameter substitution from with-abstraction.
+  -> [A.StrippedDotPattern]
+     -- ^ Dot patterns that have been stripped away by with-desugaring.
   -> (LHSResult -> TCM a)
      -- ^ Continuation.
   -> TCM a
-checkLeftHandSide c f ps a withSub' = Bench.billToCPS [Bench.Typing, Bench.CheckLHS] $ \ ret -> do
+checkLeftHandSide c f ps a withSub' strippedDots = Bench.billToCPS [Bench.Typing, Bench.CheckLHS] $ \ ret -> do
 
   -- To allow module parameters to be refined by matching, we're adding the
   -- context arguments as wildcard patterns and extending the type with the
   -- context telescope.
   cxt <- reverse <$> getContext
-  let tel = telFromList' show cxt
+  let tel = telFromList' prettyShow cxt
       cps = [ unnamed . A.VarP . fst <$> setOrigin Inserted (argFromDom d)
             | d <- cxt ]
   problem0 <- problemFromPats (cps ++ ps) (telePi tel a)
@@ -593,8 +622,8 @@ checkLeftHandSide c f ps a withSub' = Bench.billToCPS [Bench.Typing, Bench.Check
 
   -- doing the splits:
   inTopContext $ do
-    LHSState problem@(Problem pxs qs delta rest) dpi
-      <- checkLHS f $ LHSState problem0 []
+    LHSState problem@(Problem pxs qs delta rest) dpi sbe
+      <- checkLHS f $ LHSState problem0 [] []
 
     unless (null $ restPats rest) $ typeError $ TooManyArgumentsInLHS a
 
@@ -627,9 +656,9 @@ checkLeftHandSide c f ps a withSub' = Bench.billToCPS [Bench.Typing, Bench.Check
       reportSDoc "tc.lhs.top" 10 $ text "asb' = " <+> (brackets $ fsep $ punctuate comma $ map prettyTCM asb')
 
       reportSDoc "tc.lhs.top" 10 $ text "bound pattern variables"
-      reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "context = " <+> ((text . show) =<< getContext)
+      reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "context = " <+> (pretty =<< getContextTelescope)
       reportSDoc "tc.lhs.top" 10 $ nest 2 $ text "type  = " <+> prettyTCM b'
-      reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "type  = " <+> text (show b')
+      reportSDoc "tc.lhs.top" 60 $ nest 2 $ text "type  = " <+> pretty b'
 
       let notProj ProjP{} = False
           notProj _       = True
@@ -643,12 +672,12 @@ checkLeftHandSide c f ps a withSub' = Bench.billToCPS [Bench.Typing, Bench.Check
           withSub = fromMaybe (wkS (numPats - length cxt) idS) withSub'
           -- At this point we need to update the module parameters for all
           -- parent modules.
-          patSub   = (map (patternToTerm . namedArg) $ reverse $ take numPats qs) ++# EmptyS
+          patSub   = (map (patternToTerm . namedArg) $ reverse $ take numPats qs) ++# (EmptyS __IMPOSSIBLE__)
           paramSub = composeS patSub withSub
           lhsResult = LHSResult (length cxt) delta qs b' patSub asb'
-      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "patSub   = " <+> text (show patSub)
-      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "withSub  = " <+> text (show withSub)
-      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "paramSub = " <+> text (show paramSub)
+      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "patSub   = " <+> pretty patSub
+      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "withSub  = " <+> pretty withSub
+      reportSDoc "tc.lhs.top" 20 $ nest 2 $ text "paramSub = " <+> pretty paramSub
 
       let newLets = [ AsB x (applySubst paramSub v) (applySubst paramSub $ unDom a) | (x, (v, a)) <- oldLets ]
       reportSDoc "tc.lhs.top" 50 $ text "old let-bindings:" <+> text (show oldLets)
@@ -658,12 +687,16 @@ checkLeftHandSide c f ps a withSub' = Bench.billToCPS [Bench.Typing, Bench.Check
         applyRelevanceToContext (getRelevance b') $ updateModuleParameters paramSub $ do
         bindAsPatterns asb' $ do
 
-          rebindLocalRewriteRules
-
           -- Check dot patterns
           mapM_ checkDotPattern dpi
+          mapM_ (uncurry isEmptyType) sbe
           checkLeftoverDotPatterns pxs (downFrom $ size delta) (flattenTel delta) dpi
 
+          -- Type check dot patterns that have been thrown away by
+          -- with-desugaring.
+          mapM_ checkStrippedDotPattern $ applySubst paramSub strippedDots
+
+
         -- Issue2303: don't bind asb' for the continuation (return in lhsResult instead)
         ret lhsResult
 
@@ -672,7 +705,7 @@ checkLHS
   :: Maybe QName       -- ^ The name of the definition we are checking.
   -> LHSState          -- ^ The current state.
   -> TCM LHSState      -- ^ The final state after all splitting is completed
-checkLHS f st@(LHSState problem dpi) = do
+checkLHS f st@(LHSState problem dpi sbe) = do
 
   problem <- insertImplicitProblem problem
   -- Note: inserting implicits no longer preserve solvedness,
@@ -680,17 +713,18 @@ checkLHS f st@(LHSState problem dpi) = do
   if isSolvedProblem problem then return $ st { lhsProblem = problem } else do
 
     unlessM (optPatternMatching <$> gets getPragmaOptions) $
-      typeError $ GenericError $ "Pattern matching is disabled"
+      unless (problemAllVariables problem) $
+        typeError $ GenericError $ "Pattern matching is disabled"
 
     foldListT trySplit nothingToSplit $ splitProblem f problem
   where
-
+    nothingToSplit :: TCM LHSState
     nothingToSplit = do
       reportSLn "tc.lhs.split" 50 $ "checkLHS: nothing to split in problem " ++ show problem
       nothingToSplitError problem
 
     -- Split problem rest (projection pattern, does not fail as there is no call to unifier)
-
+    trySplit :: SplitProblem -> TCM LHSState -> TCM LHSState
     trySplit (SplitRest projPat o projType) _ = do
 
       -- Compute the new problem
@@ -701,7 +735,7 @@ checkLHS f st@(LHSState problem dpi) = do
           ip'      = ip ++ [fmap (Named Nothing . ProjP o) projPat]
           problem' = Problem ps' ip' delta rest
       -- Jump the trampolin
-      st' <- updateProblemRest (LHSState problem' dpi)
+      st' <- updateProblemRest (LHSState problem' dpi sbe)
       -- If the field is irrelevant, we need to continue in irr. cxt.
       -- (see Issue 939).
       applyRelevanceToContext (getRelevance projPat) $ do
@@ -721,6 +755,7 @@ checkLHS f st@(LHSState problem dpi) = do
           --       ++ [ raise (size delta2) $ Lit lit ]
           --       ++ [ var i | i <- [size delta2 ..] ]
           dpi'     = applyPatSubst rho dpi
+          sbe'     = map (second $ applyPatSubst rho) sbe
           ip'      = applySubst rho ip
           rest'    = applyPatSubst rho (problemRest problem)
 
@@ -728,47 +763,42 @@ checkLHS f st@(LHSState problem dpi) = do
       let ps'      = problemInPat p0 ++ problemInPat (absBody p1)
           delta'   = abstract delta1 delta2
           problem' = Problem ps' ip' delta' rest'
-      st' <- updateProblemRest (LHSState problem' dpi')
+      st' <- updateProblemRest (LHSState problem' dpi' sbe')
       checkLHS f st'
 
+    -- Split on absurd pattern (adding type to list of types that should be empty)
+    trySplit (Split p0 (Arg info (AbsurdFocus pi i a)) p1) _ = do
+      let tel = problemTel problem
+      reportSDoc "tc.lhs.split.absurd" 10 $ sep
+        [ text "splitting on absurd pattern"
+        , nest 2 $ text "tel  =" <+> prettyTCM tel
+        , nest 2 $ text "var  =" <+> addContext tel (prettyTCM $ var i)
+        , nest 2 $ text "type =" <+> addContext tel (prettyTCM a)
+        ]
+      let rho = liftS i $ consS (AbsurdP $ VarP $ DBPatVar "()" 0) $ raiseS 1
+      checkLHS f $ st
+            { lhsProblem = problem
+              { problemInPat  = (problemInPat p0) ++
+                                [Arg info $ unnamed $ A.WildP pi] ++
+                                problemInPat (absBody p1)
+              , problemOutPat = applySubst rho $ problemOutPat problem
+              }
+            , lhsShouldBeEmptyTypes = (getRange pi , a) : lhsShouldBeEmptyTypes st
+            }
+
     -- Split on constructor pattern (unifier might fail)
-
-    trySplit (Split p0 focus@(Arg info Focus{}) p1) tryNextSplit = do
-      res <- trySplitConstructor p0 focus p1
-      case res of
-        -- Success.  Continue checking LHS.
-        Unifies st'    -> checkLHS f st'
-        -- Mismatch.  Report and abort.
-        NoUnify  tcerr -> throwError tcerr
-        -- Unclear situation.  Try next split.
-        -- If no split works, give error from first split.
-        -- This is conservative, but might not be the best behavior.
-        -- It might be better to collect all the errors and print all of them.
-        DontKnow tcerr -> tryNextSplit `catchError` \ _ -> throwError tcerr
-
-    whenUnifies
-      :: UnificationResult' a
-      -> (a -> TCM (UnificationResult' b))
-      -> TCM (UnificationResult' b)
-    whenUnifies res cont = do
-      case res of
-        Unifies a      -> cont a
-        NoUnify  tcerr -> return $ NoUnify  tcerr
-        DontKnow tcerr -> return $ DontKnow tcerr
-
-    trySplitConstructor p0 (Arg info LitFocus{}) p1 = __IMPOSSIBLE__
-    trySplitConstructor p0 (Arg info
-             (Focus { focusCon      = c
-                    , focusPatOrigin= porigin
-                    , focusConArgs  = qs
-                    , focusRange    = r
-                    , focusOutPat   = ip
-                    , focusDatatype = d
-                    , focusParams   = vs
-                    , focusIndices  = ws
-                    , focusType     = a
-                    }
-             )) p1 = do
+    trySplit (Split p0 (Arg info
+               (Focus { focusCon      = c
+                      , focusPatOrigin= porigin
+                      , focusConArgs  = qs
+                      , focusRange    = r
+                      , focusOutPat   = ip
+                      , focusDatatype = d
+                      , focusParams   = vs
+                      , focusIndices  = ws
+                      , focusType     = a
+                      }
+               )) p1) tryNextSplit = do
       traceCall (CheckPattern (A.ConP (ConPatInfo porigin $ PatRange r) (A.AmbQ [c]) qs)
                                        (problemTel p0)
                                        (El Prop $ Def d $ map Apply $ vs ++ ws)) $ do
@@ -788,12 +818,15 @@ checkLHS f st@(LHSState problem dpi) = do
           , nest 2 $ vcat
             [ text "delta1 = " <+> prettyTCM delta1
             , text "typeOfSplitVar =" <+> addContext delta1 (prettyTCM typeOfSplitVar)
-            , text "focusOutPat =" <+> (text . show) ip
+            , text "focusOutPat =" <+> pretty ip
             , text "delta2 = " <+> addContext delta1 (addContext ("x",domFromArg typeOfSplitVar) (prettyTCM delta2))
             ]
           ]
 
-        c <- (`withRangeOf` c) <$> getConForm c
+        c <- either
+               (sigError __IMPOSSIBLE_VERBOSE__ (typeError $ AbstractConstructorNotInScope c))
+               (return . (`withRangeOf` c))
+               =<< getConForm c
         ca <- defType <$> getConInfo c
 
         reportSDoc "tc.lhs.split" 20 $ nest 2 $ vcat
@@ -824,6 +857,8 @@ checkLHS f st@(LHSState problem dpi) = do
 
         -- Insert implicit patterns
         qs' <- insertImplicitPatterns ExpandLast qs gamma'
+        reportSDoc "tc.lhs.imp" 20 $
+          text "insertImplicitPatternsT returned" <+> fsep (map prettyA qs')
 
         unless ((size qs' :: Int) == size gamma') $
           typeError $ WrongNumberOfConstructorArguments (conName c) (size gamma') (size qs')
@@ -832,13 +867,18 @@ checkLHS f st@(LHSState problem dpi) = do
 
         -- Get the type of the datatype.
         da <- (`piApply` vs) . defType <$> getConstInfo d
+        reportSDoc "tc.lhs.split" 30 $ text "  da = " <+> prettyTCM da
 
         -- Compute the flexible variables
         flex <- flexiblePatterns (problemInPat p0 ++ qs')
+        reportSDoc "tc.lhs.split" 30 $ text "computed flexible variables"
 
         -- Compute the constructor indices by dropping the parameters
         let us' = drop (size vs) us
 
+        -- Raise given indices over constructor telescope
+        let ws' = raise (size gamma) ws
+
         reportSDoc "tc.lhs.top" 15 $ addContext delta1 $
           sep [ text "preparing to unify"
               , nest 2 $ vcat
@@ -884,8 +924,22 @@ checkLHS f st@(LHSState problem dpi) = do
                  flex
                  (raise (size gamma) da)
                  us'
-                 (raise (size gamma) ws)
-        whenUnifies res $ \ (delta1',rho0,es) -> do
+                 ws'
+        case res of
+
+         -- Mismatch.  Report and abort.
+         NoUnify neg -> typeError $ ImpossibleConstructor (conName c) neg
+
+         -- Unclear situation.  Try next split.
+         -- If no split works, give error from first split.
+         -- This is conservative, but might not be the best behavior.
+         -- It might be better to collect all the errors and print all of them.
+         DontKnow errs -> tryNextSplit
+           `catchError` \ _ -> typeError $ SplitError $
+             UnificationStuck (conName c) (delta1 `abstract` gamma) us' ws' errs
+
+         -- Success.
+         Unifies (delta1',rho0,es) -> do
 
           reportSDoc "tc.lhs.top" 15 $ text "unification successful"
           reportSDoc "tc.lhs.top" 20 $ nest 2 $ vcat
@@ -910,7 +964,7 @@ checkLHS f st@(LHSState problem dpi) = do
                             newPats
 
           reportSDoc "tc.lhs.top" 20 $ addContext delta1' $ nest 2 $ vcat
-            [ text "p0'     =" <+> text (show p0')
+            [ text "p0'     =" <+> text (show $ deepUnscope p0')
             , text "newDpi  =" <+> brackets (fsep $ punctuate comma $ map prettyTCM newDpi)
             ]
 
@@ -944,9 +998,9 @@ checkLHS f st@(LHSState problem dpi) = do
             ]
 
           reportSDoc "tc.lhs.top" 70 $ addContext delta1' $ nest 2 $ vcat
-            [ text "crho2   =" <+> text (show crho2)
-            , text "rho3    =" <+> text (show rho3)
-            , text "delta2' =" <+> text (show delta2')
+            [ text "crho2   =" <+> pretty crho2
+            , text "rho3    =" <+> pretty rho3
+            , text "delta2' =" <+> pretty delta2'
             ]
 
           reportSDoc "tc.lhs.top" 15 $ nest 2 $ vcat
@@ -964,10 +1018,12 @@ checkLHS f st@(LHSState problem dpi) = do
 
           -- The final dpis are the new ones plus the old ones substituted by ρ
           let dpi' = applyPatSubst rho dpi ++ raise (size delta2') newDpi
+              sbe' = map (second $ applyPatSubst rho) sbe
 
           reportSDoc "tc.lhs.top" 15 $ addContext delta' $
             nest 2 $ vcat
               [ text "dpi'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM dpi')
+              , text "sbe'    =" <+> brackets (fsep $ punctuate comma $ map prettyTCM sbe')
               ]
 
           -- Apply the substitution
@@ -976,25 +1032,25 @@ checkLHS f st@(LHSState problem dpi) = do
 
           reportSDoc "tc.lhs.top" 15 $ addContext delta' $
             nest 2 $ vcat
-              [ text "ip' =" <+> text (show ip) ]
+              [ text "ip' =" <+> pretty ip ]
 
           -- Construct the new problem
           let problem' = Problem ps' ip' delta' rest'
 
           -- if rest type reduces,
           -- extend the split problem by previously not considered patterns
-          st'@(LHSState problem'@(Problem ps' ip' delta' rest') dpi')
-            <- updateProblemRest $ LHSState problem' dpi'
+          st'@(LHSState problem'@(Problem ps' ip' delta' rest') dpi' sbe')
+            <- updateProblemRest $ LHSState problem' dpi' sbe'
 
           reportSDoc "tc.lhs.top" 12 $ sep
             [ text "new problem from rest"
             , nest 2 $ vcat
               [ text "ps'     =" <+> fsep (map prettyA ps')
               , text "delta'  =" <+> prettyTCM delta'
-              , text "ip'     =" <+> text (show ip')
+              , text "ip'     =" <+> pretty ip'
               ]
             ]
-          return $ Unifies st'
+          checkLHS f st'
 
 
 -- | Ensures that we are not performing pattern matching on codata.
@@ -1004,6 +1060,7 @@ noPatternMatchingOnCodata = mapM_ (check . namedArg)
   where
   check (VarP {})   = return ()
   check (DotP {})   = return ()
+  check (AbsurdP{}) = return ()
   check (ProjP{})   = return ()
   check (LitP {})   = return ()  -- Literals are assumed not to be coinductive.
   check (ConP con _ ps) = do
@@ -1014,3 +1071,15 @@ noPatternMatchingOnCodata = mapM_ (check . namedArg)
       Just False -> mapM_ (check . namedArg) ps
       Just True  -> typeError $
         GenericError "Pattern matching on coinductive types is not allowed"
+
+-- | Type check dot pattern stripped from a with function.
+checkStrippedDotPattern :: A.StrippedDotPattern -> TCM ()
+checkStrippedDotPattern (A.StrippedDot e v a) = do
+  reportSDoc "tc.with.dot" 30 $ vcat
+    [ text "Checking stripped dot pattern"
+    , nest 2 $ vcat [ text "e =" <+> prettyTCM e
+                    , text "v =" <+> prettyTCM v
+                    , text "a =" <+> prettyTCM a
+                    , text "Γ =" <+> (inTopContext . prettyTCM =<< getContextTelescope) ] ]
+  u <- checkExpr e a
+  equalTerm a u v
