@@ -11,6 +11,7 @@ import Data.Semigroup (Semigroup, Monoid, (<>), mempty, mappend, mconcat)
 import Data.Traversable
 
 import Agda.Syntax.Common
+import Agda.Syntax.Info
 import Agda.Syntax.Literal
 import Agda.Syntax.Position
 import Agda.Syntax.Internal
@@ -40,12 +41,14 @@ data FlexibleVarKind
       --   Saves the 'FlexibleVarKind' of its subpatterns.
   | ImplicitFlex -- ^ From a hidden formal argument or underscore ('WildP').
   | DotFlex      -- ^ From a dot pattern ('DotP').
+  | OtherFlex    -- ^ From a non-record constructor or literal ('ConP' or 'LitP').
   deriving (Eq, Show)
 
 -- | Flexible variables are equipped with information where they come from,
 --   in order to make a choice which one to assign when two flexibles are unified.
 data FlexibleVar a = FlexibleVar
   { flexHiding :: Hiding
+  , flexOrigin :: Origin
   , flexKind   :: FlexibleVarKind
   , flexPos    :: Maybe Int
   , flexVar    :: a
@@ -55,16 +58,22 @@ instance LensHiding (FlexibleVar a) where
   getHiding     = flexHiding
   mapHiding f x = x { flexHiding = f (flexHiding x) }
 
-defaultFlexibleVar :: a -> FlexibleVar a
-defaultFlexibleVar a = FlexibleVar Hidden ImplicitFlex Nothing a
+instance LensOrigin (FlexibleVar a) where
+  getOrigin     = flexOrigin
+  mapOrigin f x = x { flexOrigin = f (flexOrigin x) }
 
-flexibleVarFromHiding :: Hiding -> a -> FlexibleVar a
-flexibleVarFromHiding h a = FlexibleVar h ImplicitFlex Nothing a
+-- UNUSED
+-- defaultFlexibleVar :: a -> FlexibleVar a
+-- defaultFlexibleVar a = FlexibleVar Hidden Inserted ImplicitFlex Nothing a
+
+-- UNUSED
+-- flexibleVarFromHiding :: Hiding -> a -> FlexibleVar a
+-- flexibleVarFromHiding h a = FlexibleVar h ImplicitFlex Nothing a
 
 allFlexVars :: Telescope -> FlexibleVars
 allFlexVars tel = zipWith makeFlex (downFrom $ size tel) $ telToList tel
   where
-    makeFlex i d = FlexibleVar (getHiding d) ImplicitFlex (Just i) i
+    makeFlex i d = FlexibleVar (getHiding d) (getOrigin d) ImplicitFlex (Just i) i
 
 data FlexChoice = ChooseLeft | ChooseRight | ChooseEither | ExpandBoth
   deriving (Eq, Show)
@@ -80,7 +89,7 @@ instance Semigroup FlexChoice where
   ChooseRight  <> ChooseRight  = ChooseRight
 
 instance Monoid FlexChoice where
-  mempty = ChooseEither
+  mempty  = ChooseEither
   mappend = (<>)
 
 class ChooseFlex a where
@@ -94,6 +103,9 @@ instance ChooseFlex FlexibleVarKind where
   chooseFlex (RecordFlex xs) y               = chooseFlex xs (repeat y)
   chooseFlex x               (RecordFlex ys) = chooseFlex (repeat x) ys
   chooseFlex ImplicitFlex    ImplicitFlex    = ChooseEither
+  chooseFlex ImplicitFlex    _               = ChooseLeft
+  chooseFlex _               ImplicitFlex    = ChooseRight
+  chooseFlex OtherFlex       OtherFlex       = ChooseEither
 
 instance ChooseFlex a => ChooseFlex [a] where
   chooseFlex xs ys = mconcat $ zipWith chooseFlex xs ys
@@ -105,13 +117,22 @@ instance ChooseFlex a => ChooseFlex (Maybe a) where
   chooseFlex (Just x) (Just y) = chooseFlex x y
 
 instance ChooseFlex Hiding where
-  chooseFlex Hidden   Hidden   = ChooseEither
-  chooseFlex Hidden   _        = ChooseLeft
-  chooseFlex _        Hidden   = ChooseRight
-  chooseFlex Instance Instance = ChooseEither
-  chooseFlex Instance _        = ChooseLeft
-  chooseFlex _        Instance = ChooseRight
-  chooseFlex _        _        = ChooseEither
+  chooseFlex Hidden     Hidden     = ChooseEither
+  chooseFlex Hidden     _          = ChooseLeft
+  chooseFlex _          Hidden     = ChooseRight
+  chooseFlex Instance{} Instance{} = ChooseEither
+  chooseFlex Instance{} _          = ChooseLeft
+  chooseFlex _          Instance{} = ChooseRight
+  chooseFlex _          _          = ChooseEither
+
+instance ChooseFlex Origin where
+  chooseFlex Inserted  Inserted  = ChooseEither
+  chooseFlex Inserted  _         = ChooseLeft
+  chooseFlex _         Inserted  = ChooseRight
+  chooseFlex Reflected Reflected = ChooseEither
+  chooseFlex Reflected _         = ChooseLeft
+  chooseFlex _         Reflected = ChooseRight
+  chooseFlex _         _         = ChooseEither
 
 instance ChooseFlex Int where
   chooseFlex x y = case compare x y of
@@ -120,8 +141,8 @@ instance ChooseFlex Int where
     GT -> ChooseRight
 
 instance (ChooseFlex a) => ChooseFlex (FlexibleVar a) where
-  chooseFlex (FlexibleVar h1 f1 p1 i1) (FlexibleVar h2 f2 p2 i2) =
-    firstChoice [ chooseFlex f1 f2, chooseFlex h1 h2
+  chooseFlex (FlexibleVar h1 o1 f1 p1 i1) (FlexibleVar h2 o2 f2 p2 i2) =
+    firstChoice [ chooseFlex f1 f2, chooseFlex o1 o2, chooseFlex h1 h2
                 , chooseFlex p1 p2, chooseFlex i1 i2]
       where
         firstChoice :: [FlexChoice] -> FlexChoice
@@ -193,6 +214,11 @@ data Focus
     , focusType     :: Type -- ^ Type of variable we are splitting, kept for record patterns.
     }
   | LitFocus Literal [NamedArg DeBruijnPattern] Type
+  | AbsurdFocus
+    { absurdFocusRange :: PatInfo
+    , absurdFocusVar   :: Int
+    , absurdFocusType  :: Type
+    }
 
 -- | Result of 'splitProblem':  Determines position for the next split.
 data SplitProblem
@@ -244,6 +270,7 @@ data AsBinding      = AsB Name Term Type
 data LHSState = LHSState
   { lhsProblem :: Problem
   , lhsDPI     :: [DotPatternInst]
+  , lhsShouldBeEmptyTypes :: [(Range,Type)]
   }
 
 instance Subst Term ProblemRest where
@@ -276,8 +303,8 @@ instance PrettyTCM AsBinding where
 
 instance PP.Pretty AsBinding where
   pretty (AsB x v a) =
-    PP.text (show x ++ " =") PP.<+> PP.hang (PP.pretty v PP.<+> PP.text ":") 2
-                                            (PP.pretty a)
+    PP.pretty x PP.<+> PP.text "=" PP.<+>
+      PP.hang (PP.pretty v PP.<+> PP.text ":") 2 (PP.pretty a)
 
 instance InstantiateFull AsBinding where
   instantiateFull' (AsB x v a) = AsB x <$> instantiateFull' v <*> instantiateFull' a
