@@ -1,5 +1,4 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
 
 module Agda.TypeChecking.Rules.LHS.Split
   ( splitProblem
@@ -11,8 +10,9 @@ import Control.Applicative hiding (empty)
 import Control.Monad.Trans ( lift )
 import Control.Monad.Trans.Maybe
 
+import Data.Either
 import Data.Maybe (fromMaybe)
-import Data.List hiding (null)
+import qualified Data.List as List
 import Data.Traversable hiding (mapM, sequence)
 import Data.Foldable (msum)
 
@@ -36,6 +36,7 @@ import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Constraints
 import Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Datatypes
+import Agda.TypeChecking.Errors (dropTopLevelModule)
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Irrelevance
 import Agda.TypeChecking.MetaVars
@@ -48,6 +49,7 @@ import Agda.TypeChecking.Telescope
 
 import Agda.TypeChecking.Rules.LHS.Problem
 
+import Agda.Utils.Either
 import Agda.Utils.Except (catchError)
 import Agda.Utils.Functor ((<.>))
 import Agda.Utils.Lens
@@ -57,8 +59,10 @@ import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Permutation
-import Agda.Utils.Tuple
+import Agda.Utils.Pretty (prettyShow)
 import qualified Agda.Utils.Pretty as P
+import Agda.Utils.Size
+import Agda.Utils.Tuple
 
 #include "undefined.h"
 import Agda.Utils.Impossible
@@ -79,12 +83,17 @@ splitProblem ::
 splitProblem mf (Problem ps qs tel pr) = do
   lift $ do
     reportSLn "tc.lhs.split" 20 $ "initiating splitting"
-      ++ maybe "" ((" for definition " ++) . show) mf
+      ++ maybe "" ((" for definition " ++) . prettyShow) mf
     reportSDoc "tc.lhs.split" 30 $ sep
       [ nest 2 $ text "ps   =" <+> sep (map (P.parens <.> prettyA) ps)
       , nest 2 $ text "qs   =" <+> sep (map (P.parens <.> prettyTCM . namedArg) qs)
       , nest 2 $ text "tel  =" <+> prettyTCM tel
       ]
+    reportSDoc "tc.lhs.split" 60 $ sep
+      [ nest 2 $ text "ps  (raw) =" <+> sep (map (P.parens <.> text . show) ps)
+      , nest 2 $ text "qs  (raw) =" <+> sep (map (P.parens <.> text . show . namedArg) qs)
+      , nest 2 $ text "tel (raw) =" <+> (text . show) tel
+      ]
   splitP ps tel
   where
     -- Result splitting
@@ -109,17 +118,20 @@ splitProblem mf (Problem ps qs tel pr) = do
         caseMaybeM (lift $ isRecordType $ unArg b) notRecord $ \(r, vs, def) -> case def of
           Record{ recFields = fs } -> do
             lift $ reportSDoc "tc.lhs.split" 20 $ sep
-              [ text $ "we are of record type r  = " ++ show r
+              [ text $ "we are of record type r  = " ++ prettyShow r
               , text   "applied to parameters vs = " <+> prettyTCM vs
-              , text $ "and have fields       fs = " ++ show fs
+              , text $ "and have fields       fs = " ++ prettyShow fs
               ]
             -- The record "self" is the definition f applied to the patterns
             let es = patternsToElims qs
             -- Note: the module parameters are already part of qs
             let self = defaultArg $ Def f [] `applyE` es
                 ai   = getArgInfo p
-            -- Try the projection candidates
-            msum $ map (tryProj o ai self fs vs (length projs >= 2)) projs
+            -- Try the projection candidates.
+            -- Fail hard for the last candidate.
+            msum $ mapAwareLast (tryProj o ai self fs r vs $ length ds >= 2) projs
+            -- -- This fails softly on all (if more than one) candidates.
+            -- msum $ map (tryProj o ai self fs r vs (length projs >= 2)) projs
 
           _ -> __IMPOSSIBLE__
       where
@@ -129,16 +141,46 @@ splitProblem mf (Problem ps qs tel pr) = do
       wrongHiding :: MonadTCM tcm => QName -> tcm a
       wrongHiding d = typeError . GenericDocError =<< do
         liftTCM $ text "Wrong hiding used for projection " <+> prettyTCM d
+      -- Issue #2423: error which reports the disambiguation
+      wrongProj :: MonadTCM tcm => QName -> Bool -> tcm a
+      wrongProj d amb = typeError . GenericDocError =<< do
+        liftTCM $ sep
+          [ text "Cannot eliminate type "
+          , prettyTCM (unArg b)
+          , text " with projection "
+          , if amb then text . prettyShow =<< dropTopLevelModule d else prettyTCM d
+          ]
 
-      tryProj :: ProjOrigin -> ArgInfo -> Arg Term -> [Arg QName] -> Args -> Bool -> (QName, Projection) -> ListT TCM SplitProblem
-      tryProj o ai self fs vs amb (d0, proj) = do
+      -- | Pass 'True' unless last element of the list.
+      mapAwareLast :: forall a b. (Bool -> a -> b) -> [a] -> [b]
+      mapAwareLast f []     = []
+      mapAwareLast f [a]    = [f False a]
+      mapAwareLast f (a:as) = f True a : mapAwareLast f as
+
+      tryProj
+        :: ProjOrigin           -- ^ Origin of projection pattern.
+        -> ArgInfo              -- ^ ArgInfo of projection pattern.
+        -> Arg Term             -- ^ Self: value we are eliminating.
+        -> [Arg QName]          -- ^ Fields of record type under consideration.
+        -> QName                -- ^ Name of record type we are eliminating.
+        -> Args                 -- ^ Parameters of record type we are eliminating.
+        -> Bool                 -- ^ Did we start out with an ambiguous projection in the beginning?
+        -> Bool                 -- ^ More than 1 candidates?  If yes, fail softly.
+        -> (QName, Projection)  -- ^ Current candidate.
+        -> ListT TCM SplitProblem
+      tryProj o ai self fs r vs amb soft (d0, proj) = do
         -- Recoverable errors are those coming from the projection.
-        -- If we have several projections (amb) we just try the next one.
-        let ambErr err = if amb then mzero else err
+        -- If we have several projections we fail @soft@ly and just try the next one.
+        let ambErr err = if soft then mzero else err
+            ambTry m
+             | soft = unlessM (liftTCM $ tryConversion m) mzero -- succeed without constraints
+             -- This would leave constraints:
+             -- -- | amb = whenNothingM (liftTCM $ tryMaybe $ disableDestructiveUpdate m) mzero
+             | otherwise = liftTCM $ noConstraints m
         case proj of
           -- Andreas, 2015-05-06 issue 1413 projProper=Nothing is not impossible
-          Projection{projProper = False} -> ambErr notProjP
-          Projection{projProper = True, projOrig = d, projLams = lams} -> do
+          Projection{projProper = Nothing} -> ambErr notProjP
+          Projection{projProper = Just qr, projOrig = d, projLams = lams} -> do
             let ai = projArgInfo proj
             -- If projIndex==0, then the projection is already applied
             -- to the record value (like in @open R r@), and then it
@@ -148,19 +190,26 @@ splitProblem mf (Problem ps qs tel pr) = do
             -- If the target is not a record type, that's an error.
             -- It could be a meta, but since we cannot postpone lhs checking, we crash here.
             lift $ reportSDoc "tc.lhs.split" 20 $ sep
-              [ text $ "proj                  d0 = " ++ show d0
-              , text $ "original proj         d  = " ++ show d
+              [ text $ "proj                  d0 = " ++ prettyShow d0
+              , text $ "original proj         d  = " ++ prettyShow d
               ]
             -- Get the field decoration.
             -- If the projection pattern name @d@ is not a field name,
             -- we have to try the next projection name.
             -- If this was not an ambiguous projection, that's an error.
-            argd <- maybe (ambErr failure) return $ find ((d ==) . unArg) fs
+            argd <- maybe (ambErr $ wrongProj d amb) return $ List.find ((d ==) . unArg) fs
             let ai' = setRelevance (getRelevance argd) ai
 
+            -- Andreas, 2016-12-31, issue #2374:
+            -- We can also disambiguate by hiding info.
+            unless (sameHiding p ai) $ ambErr $ wrongHiding d
+
+            -- Andreas, 2016-12-31, issue #1976:
+            -- Check parameters.
+            ambTry $ checkParameters qr r vs
+
             -- From here, we have the correctly disambiguated projection.
             -- Thus, we no longer catch errors.
-            unless (getHiding p == getHiding ai) $ wrongHiding d
 
             -- For highlighting, we remember which name we disambiguated to.
             -- This is safe here (fingers crossed) as we won't decide on a
@@ -205,7 +254,7 @@ splitProblem mf (Problem ps qs tel pr) = do
         ]
 
       -- Andreas, 2016-06-30, issue #2075: need test here!
-      unless (getHiding p == getHiding ai) $ typeError WrongHidingInLHS
+      unless (sameHiding p ai) $ typeError WrongHidingInLHS
 
       -- Possible reinvokations:
       let -- 1. Redo this argument (after meta instantiation).
@@ -256,7 +305,7 @@ splitProblem mf (Problem ps qs tel pr) = do
             -- Subcase: a record type (d vs)
             Right (d, vs, def) -> do
               let np = recPars def
-              let (pars, ixs) = genericSplitAt np vs
+              let (pars, ixs) = splitAt np vs
               lift $ reportSDoc "tc.lhs.split" 10 $ vcat
                 [ sep [ text "splitting on"
                       , nest 2 $ fsep [ prettyA p, text ":", prettyTCM dom ]
@@ -278,6 +327,16 @@ splitProblem mf (Problem ps qs tel pr) = do
                 , splitRPats   = Abs x  $ Problem ps () tel __IMPOSSIBLE__
                 }) `mplus` keepGoing
 
+        -- Case: absurd pattern.
+        p@(A.AbsurdP info) -> do
+          lift $ reportSDoc "tc.lhs.split.absurd" 30 $ text "split AbsurdP: type is " <+> prettyTCM a
+          let i = size tel
+          (return Split
+            { splitLPats = empty
+            , splitFocus = Arg ai $ AbsurdFocus info i $ raise (i+1) a
+            , splitRPats = Abs x  $ Problem ps () tel __IMPOSSIBLE__
+            }) `mplus` keepGoing
+
         -- Case: constructor pattern.
         p@(A.ConP ci (A.AmbQ cs) args) -> do
           let tryInstantiate a'
@@ -303,12 +362,18 @@ splitProblem mf (Problem ps qs tel pr) = do
             case ignoreSharing $ unEl a' of
 
               -- Subcase: split type is a Def.
-              Def d es    -> do
+              Def d es    -> (liftTCM $ theDef <$> getConstInfo d) >>= \case
 
-                def <- liftTCM $ theDef <$> getConstInfo d
+               -- Issue #2253: the data type could be abstract.
+               AbstractDefn{} -> liftTCM $ traceCall (CheckPattern p EmptyTel a) $ do
+                 typeError . GenericDocError =<< do
+                   text "Cannot split on abstract data type" <+> prettyTCM d
+
+               def -> do
 
                 -- We cannot split on (shape-)irrelevant non-records.
                 -- Andreas, 2011-10-04 unless allowed by option
+                lift $ reportSLn "tc.lhs.split" 30 $ "split ConP: relevance is " ++ show ai
                 unless (defIsRecord def) $
                   when (unusableRelevance $ getRelevance ai) $
                   unlessM (liftTCM $ optExperimentalIrrelevance <$> pragmaOptions) $
@@ -327,24 +392,26 @@ splitProblem mf (Problem ps qs tel pr) = do
                     traceCall (CheckPattern p EmptyTel a) $ do  -- TODO: wrong telescope
                       -- Check that we construct something in the right datatype
                       c <- lift $ do
-                          cs' <- mapM canonicalName cs
+                          -- Andreas, 2017-08-13, issue #2686: ignore abstract constructors
+                          (cs1, cs') <- unzip . snd . partitionEithers <$> do
+                            forM cs $ \ c -> mapRight ((c,) . conName) <$> getConHead c
+                          when (null cs1) $ typeError $ AbstractConstructorNotInScope $ head cs
                           d'  <- canonicalName d
-                          let cons def = case theDef def of
-                                Datatype{dataCons = cs} -> cs
-                                Record{recConHead = c}      -> [conName c]
-                                _                       -> __IMPOSSIBLE__
-                          cs0 <- cons <$> getConstInfo d'
-                          case [ c | (c, c') <- zip cs cs', elem c' cs0 ] of
+                          cs0 <- (theDef <$> getConstInfo d') <&> \case
+                                Datatype{dataCons = cs0} -> cs0
+                                Record{recConHead = c0}  -> [conName c0]
+                                _ -> __IMPOSSIBLE__
+                          case [ c | (c, c') <- zip cs1 cs', elem c' cs0 ] of
                             [c]   -> do
                               -- If constructor pattern was ambiguous,
                               -- remember our choice for highlighting info.
                               when (length cs >= 2) $ storeDisambiguatedName c
                               return c
-                            []    -> typeError $ ConstructorPatternInWrongDatatype (head cs) d
-                            cs    -> -- if there are more than one we give up (they might have different types)
-                              typeError $ CantResolveOverloadedConstructorsTargetingSameDatatype d cs
+                            []    -> typeError $ ConstructorPatternInWrongDatatype (head cs1) d
+                            cs3   -> -- if there are more than one we give up (they might have different types)
+                              typeError $ CantResolveOverloadedConstructorsTargetingSameDatatype d cs3
 
-                      let (pars, ixs) = genericSplitAt np vs
+                      let (pars, ixs) = splitAt np vs
                       lift $ reportSDoc "tc.lhs.split" 10 $ vcat
                         [ sep [ text "splitting on"
                               , nest 2 $ fsep [ prettyA p, text ":", prettyTCM dom ]
@@ -363,7 +430,9 @@ splitProblem mf (Problem ps qs tel pr) = do
                       -- so we add a check here.
                       -- I guess this issue could be solved more systematically,
                       -- but the extra check here is non-invasive to the existing code.
-                      checkParsIfUnambiguous cs d pars
+                      -- Andreas, 2016-12-31 fixing issue #1975
+                      -- Do this also for constructors which were originally ambiguous.
+                      checkConstructorParameters c d pars
 
                       (return Split
                         { splitLPats   = empty
@@ -377,23 +446,34 @@ splitProblem mf (Problem ps qs tel pr) = do
         _ -> keepGoing
 
 
--- | @checkParsIfUnambiguous [c] d pars@ checks that the data/record type
+-- | @checkConstructorParameters c d pars@ checks that the data/record type
 --   behind @c@ is has initial parameters (coming e.g. from a module instantiation)
 --   that coincide with an prefix of @pars@.
-checkParsIfUnambiguous :: MonadTCM tcm => [QName] -> QName -> Args -> tcm ()
-checkParsIfUnambiguous [c] d pars = liftTCM $ do
-  dc <- getConstructorData c
+checkConstructorParameters :: MonadTCM tcm => QName -> QName -> Args -> tcm ()
+checkConstructorParameters c d pars = do
+  dc <- liftTCM $ getConstructorData c
+  checkParameters dc d pars
+
+-- | Check that given parameters match the parameters of the inferred
+--   constructor/projection.
+checkParameters
+  :: MonadTCM tcm
+  => QName  -- ^ The record/data type name of the chosen constructor/projection.
+  -> QName  -- ^ The record/data type name as supplied by the type signature.
+  -> Args   -- ^ The parameters.
+  -> tcm ()
+checkParameters dc d pars = liftTCM $ do
   a  <- reduce (Def dc [])
   case ignoreSharing a of
     Def d0 es -> do -- compare parameters
       let vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       reportSDoc "tc.lhs.split" 40 $
-        vcat [ nest 2 $ text "d                   =" <+> prettyTCM d
-             , nest 2 $ text "d0 (should be == d) =" <+> prettyTCM d0
-             , nest 2 $ text "dc                  =" <+> prettyTCM dc
+        vcat [ nest 2 $ text "d                   =" <+> (text . prettyShow) d
+             , nest 2 $ text "d0 (should be == d) =" <+> (text . prettyShow) d0
+             , nest 2 $ text "dc                  =" <+> (text . prettyShow) dc
+             , nest 2 $ text "vs                  =" <+> prettyTCM vs
              ]
       -- when (d0 /= d) __IMPOSSIBLE__ -- d could have extra qualification
       t <- typeOfConst d
       compareArgs [] t (Def d []) vs (take (length vs) pars)
     _ -> __IMPOSSIBLE__
-checkParsIfUnambiguous _ _ _ = return ()
