@@ -2,7 +2,6 @@
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE NondecreasingIndentation   #-}
-{-# LANGUAGE ScopedTypeVariables        #-}
 {-# LANGUAGE UndecidableInstances       #-}
 {-# LANGUAGE ViewPatterns               #-}
 
@@ -105,8 +104,7 @@
 module Agda.TypeChecking.Rules.LHS.Unify
   ( UnificationResult
   , UnificationResult'(..)
-  , unifyIndices
-  , unifyIndices_ ) where
+  , unifyIndices ) where
 
 import Prelude hiding (null)
 
@@ -122,7 +120,7 @@ import Control.Monad.Writer (WriterT(..), MonadWriter(..), Monoid(..))
 import Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Semigroup hiding (Arg)
-import Data.List hiding (null, sort)
+import qualified Data.List as List
 
 import Data.Typeable (Typeable)
 import Data.Foldable (Foldable)
@@ -139,7 +137,6 @@ import Agda.Syntax.Position
 
 import Agda.TypeChecking.Monad
 import qualified Agda.TypeChecking.Monad.Benchmark as Bench
-import Agda.TypeChecking.Monad.Exception
 import Agda.TypeChecking.Monad.Builtin (constructorForm)
 import Agda.TypeChecking.Conversion -- equalTerm
 import Agda.TypeChecking.Constraints
@@ -161,10 +158,7 @@ import Agda.Interaction.Options (optInjectiveTypeConstructors, optWithoutK)
 import Agda.TypeChecking.Rules.LHS.Problem hiding (Substitution)
 -- import Agda.TypeChecking.SyntacticEquality
 
-import Agda.Utils.Except
-  ( Error(noMsg, strMsg)
-  , MonadError(catchError, throwError)
-  )
+import Agda.Utils.Except ( MonadError(catchError, throwError) )
 import Agda.Utils.Either
 import Agda.Utils.Functor
 import Agda.Utils.List
@@ -186,38 +180,22 @@ type UnificationResult = UnificationResult'
   )
 
 data UnificationResult' a
-  = Unifies  a      -- ^ Unification succeeded.
-  | NoUnify  TCErr  -- ^ Terms are not unifiable.
-  | DontKnow TCErr  -- ^ Some other error happened, unification got stuck.
+  = Unifies  a                    -- ^ Unification succeeded.
+  | NoUnify  NegativeUnification  -- ^ Terms are not unifiable.
+  | DontKnow [UnificationFailure] -- ^ Some other error happened, unification got stuck.
   deriving (Typeable, Show, Functor, Foldable, Traversable)
 
 -- | Unify indices.
 --
---   In @unifyIndices_ flex a us vs@,
---
---   @a@ is the type eliminated by @us@ and @vs@
---     (usally the type of a constructor),
---     need not be reduced,
+--   In @unifyIndices gamma flex us vs@,
 --
 --   @us@ and @vs@ are the argument lists to unify,
 --
+--   @gamma@ is the telescope of free variables in @us@ and @vs@.
+--
 --   @flex@ is the set of flexible (instantiable) variabes in @us@ and @vs@.
 --
 --   The result is the most general unifier of @us@ and @vs@.
-unifyIndices_ :: MonadTCM tcm
-              => Telescope
-              -> FlexibleVars
-              -> Type
-              -> Args
-              -> Args
-              -> tcm (Telescope, PatternSubstitution, [NamedArg DeBruijnPattern])
-unifyIndices_ tel flex a us vs = liftTCM $ do
-  r <- unifyIndices tel flex a us vs
-  case r of
-    Unifies sub   -> return sub
-    DontKnow err  -> throwError err
-    NoUnify  err  -> throwError err
-
 unifyIndices :: MonadTCM tcm
              => Telescope
              -> FlexibleVars
@@ -397,8 +375,8 @@ solveVar k u s = case instantiateTelescope (varTel s) k u of
   where
     permuteFlex :: Permutation -> FlexibleVars -> FlexibleVars
     permuteFlex perm =
-      mapMaybe $ \(FlexibleVar h k p x) ->
-        FlexibleVar h k p <$> findIndex (x==) (permPicks perm)
+      mapMaybe $ \(FlexibleVar h o k p x) ->
+        FlexibleVar h o k p <$> List.findIndex (x==) (permPicks perm)
 
 applyUnder :: Int -> Telescope -> Term -> Telescope
 applyUnder k tel u
@@ -471,8 +449,8 @@ data UnifyStep
     { conflictAt         :: Int
     , conflictDatatype   :: QName
     , conflictParameters :: Args
-    , conflictConLeft    :: ConHead
-    , conflictConRight   :: ConHead
+    , conflictLeft       :: Term
+    , conflictRight      :: Term
     }
   | Cycle
     { cycleAt            :: Int
@@ -516,67 +494,67 @@ instance PrettyTCM UnifyStep where
   prettyTCM step = case step of
     Deletion k a u v -> text "Deletion" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "type:       " <+> text (show a)
-      , text "lhs:        " <+> text (show u)
-      , text "rhs:        " <+> text (show v)
+      , text "type:       " <+> prettyTCM a
+      , text "lhs:        " <+> prettyTCM u
+      , text "rhs:        " <+> prettyTCM v
       ])
     Solution k a i u -> text "Solution" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "type:       " <+> text (show a)
+      , text "type:       " <+> prettyTCM a
       , text "variable:   " <+> text (show i)
-      , text "term:       " <+> text (show u)
+      , text "term:       " <+> prettyTCM u
       ])
     Injectivity k a d pars ixs c -> text "Injectivity" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "type:       " <+> text (show a)
-      , text "datatype:   " <+> text (show d)
-      , text "parameters: " <+> text (show pars)
-      , text "indices:    " <+> text (show ixs)
-      , text "constructor:" <+> text (show c)
+      , text "type:       " <+> prettyTCM a
+      , text "datatype:   " <+> prettyTCM d
+      , text "parameters: " <+> prettyList_ (map prettyTCM pars)
+      , text "indices:    " <+> prettyList_ (map prettyTCM ixs)
+      , text "constructor:" <+> prettyTCM c
       ])
-    Conflict k d pars c1 c2 -> text "Conflict" $$ nest 2 (vcat $
+    Conflict k d pars u v -> text "Conflict" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "datatype:   " <+> text (show d)
-      , text "parameters: " <+> text (show pars)
-      , text "con1:       " <+> text (show c1)
-      , text "con2:       " <+> text (show c2)
+      , text "datatype:   " <+> prettyTCM d
+      , text "parameters: " <+> prettyList_ (map prettyTCM pars)
+      , text "lhs:        " <+> prettyTCM u
+      , text "rhs:        " <+> prettyTCM v
       ])
     Cycle k d pars i u -> text "Cycle" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "datatype:   " <+> text (show d)
-      , text "parameters: " <+> text (show pars)
+      , text "datatype:   " <+> prettyTCM d
+      , text "parameters: " <+> prettyList_ (map prettyTCM pars)
       , text "variable:   " <+> text (show i)
-      , text "term:       " <+> text (show u)
+      , text "term:       " <+> prettyTCM u
       ])
     EtaExpandVar fi r pars -> text "EtaExpandVar" $$ nest 2 (vcat $
       [ text "variable:   " <+> text (show fi)
-      , text "record type:" <+> text (show r)
-      , text "parameters: " <+> text (show pars)
+      , text "record type:" <+> prettyTCM r
+      , text "parameters: " <+> prettyTCM pars
       ])
     EtaExpandEquation k r pars -> text "EtaExpandVar" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "record type:" <+> text (show r)
-      , text "parameters: " <+> text (show pars)
+      , text "record type:" <+> prettyTCM r
+      , text "parameters: " <+> prettyTCM pars
       ])
     LitConflict k a u v -> text "LitConflict" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "type:       " <+> text (show a)
-      , text "lhs:        " <+> text (show u)
-      , text "rhs:        " <+> text (show v)
+      , text "type:       " <+> prettyTCM a
+      , text "lhs:        " <+> prettyTCM u
+      , text "rhs:        " <+> prettyTCM v
       ])
     StripSizeSuc k u v -> text "StripSizeSuc" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "lhs:        " <+> text (show u)
-      , text "rhs:        " <+> text (show v)
+      , text "lhs:        " <+> prettyTCM u
+      , text "rhs:        " <+> prettyTCM v
       ])
     SkipIrrelevantEquation k -> text "SkipIrrelevantEquation" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
       ])
     TypeConInjectivity k d us vs -> text "TypeConInjectivity" $$ nest 2 (vcat $
       [ text "position:   " <+> text (show k)
-      , text "datatype:   " <+> text (show d)
-      , text "lhs:        " <+> text (show us)
-      , text "rhs:        " <+> text (show vs)
+      , text "datatype:   " <+> prettyTCM d
+      , text "lhs:        " <+> prettyList_ (map prettyTCM us)
+      , text "rhs:        " <+> prettyList_ (map prettyTCM vs)
       ])
 
 type UnifyStrategy = UnifyState -> ListT TCM UnifyStep
@@ -600,14 +578,15 @@ completeStrategyAt k s = msum $ map (\strat -> strat k s) $
     , etaExpandVarStrategy
     , etaExpandEquationStrategy
     , injectiveTypeConStrategy
+    , injectivePragmaStrategy
     , simplifySizesStrategy
     , checkEqualityStrategy
     ]
 
 -- | Returns true if the variables 0..k-1 don't occur in x
-isHom :: (Free' a All, Subst Term a) => Int -> a -> Maybe a
+isHom :: (Free a, Subst Term a) => Int -> a -> Maybe a
 isHom n x = do
-  guard $ getAll $ runFree (\ (i,_) -> All (i >= n)) IgnoreNot x
+  guard $ getAll $ runFree (All . (>= n)) IgnoreNot x
   return $ raise (-n) x
 
 -- | Checks whether the given term (of the given type) is beta-eta-equivalent
@@ -680,7 +659,7 @@ findFlexible :: Int -> FlexibleVars -> Maybe (FlexibleVar Nat)
 findFlexible i flex =
   let flex'      = map flexVar flex
       flexible i = i `elem` flex'
-  in find ((i ==) . flexVar) flex
+  in List.find ((i ==) . flexVar) flex
 
 basicUnifyStrategy :: Int -> UnifyStrategy
 basicUnifyStrategy k s = do
@@ -690,19 +669,23 @@ basicUnifyStrategy k s = do
   liftTCM $ reportSDoc "tc.lhs.unify" 30 $ text "isEtaVar results: " <+> text (show [mi,mj])
   case (mi, mj) of
     (Just i, Just j)
-     | i == j -> return $ Deletion k ha (var i) (var i)
+     | i == j -> mzero -- Taken care of by checkEqualityStrategy
     (Just i, Just j)
      | Just fi <- findFlexible i flex
      , Just fj <- findFlexible j flex -> do
        let choice = chooseFlex fi fj
+           firstTryLeft  = msum [ return (Solution k ha i v)
+                                , return (Solution k ha j u)]
+           firstTryRight = msum [ return (Solution k ha j u)
+                                , return (Solution k ha i v)]
        liftTCM $ reportSDoc "tc.lhs.unify" 40 $ text "fi = " <+> text (show fi)
        liftTCM $ reportSDoc "tc.lhs.unify" 40 $ text "fj = " <+> text (show fj)
        liftTCM $ reportSDoc "tc.lhs.unify" 40 $ text "chooseFlex: " <+> text (show choice)
        case choice of
-         ChooseLeft   -> return $ Solution k ha i v
-         ChooseRight  -> return $ Solution k ha j u
+         ChooseLeft   -> firstTryLeft
+         ChooseRight  -> firstTryRight
          ExpandBoth   -> mzero -- This should be taken care of by etaExpandEquationStrategy
-         ChooseEither -> return $ Solution k ha j u
+         ChooseEither -> firstTryRight
     (Just i, _)
      | Just _ <- findFlexible i flex -> return $ Solution k ha i v
     (_, Just j)
@@ -720,18 +703,12 @@ dataStrategy k s = do
       npars <- mcatMaybes $ liftTCM $ getNumberOfParameters d
       let (pars,ixs) = splitAt npars $ fromMaybe __IMPOSSIBLE__ $ allApplyElims es
       hpars <- mfromMaybe $ isHom k pars
-      liftTCM $ reportSDoc "tc.lhs.unify" 40 $ addContext (varTel s `abstract` eqTel s) $
+      liftTCM $ reportSDoc "tc.lhs.unify" 40 $ addContext (varTel s) $
         text "Found equation at datatype " <+> prettyTCM d
          <+> text " with (homogeneous) parameters " <+> prettyTCM hpars
       case (ignoreSharing u, ignoreSharing v) of
-        (MetaV m es, Con c ci _   ) -> do
-          us <- mcatMaybes $ liftTCM $ addContext (varTel s) $ instMetaCon m es d hpars c ci
-          return $ Injectivity k a d hpars ixs c
-        (Con c ci _   , MetaV m es) -> do
-          vs <- mcatMaybes $ liftTCM $ addContext (varTel s) $ instMetaCon m es d hpars c ci
-          return $ Injectivity k a d hpars ixs c
         (Con c _ _   , Con c' _ _  ) | c == c' -> return $ Injectivity k a d hpars ixs c
-        (Con c _ _   , Con c' _ _  ) -> return $ Conflict k d hpars c c'
+        (Con c _ _   , Con c' _ _  ) -> return $ Conflict k d hpars u v
         (Var i []  , v         ) -> ifOccursStronglyRigid i v $ return $ Cycle k d hpars i v
         (u         , Var j []  ) -> ifOccursStronglyRigid j u $ return $ Cycle k d hpars j u
         _ -> mzero
@@ -746,47 +723,6 @@ dataStrategy k s = do
           StronglyRigid -> ret
           _ -> mzero
 
-    -- Instantiate the meta with a constructor applied to fresh metas
-    -- Returns the fresh metas if successful
-    instMetaCon :: MetaId -> Elims -> QName -> Args -> ConHead -> ConInfo -> TCM (Maybe Args)
-    instMetaCon m es d pars c ci = do
-      caseMaybe (allApplyElims es) (return Nothing) $ \ us -> do
-        ifNotM (asks envAssignMetas) (return Nothing) $ {-else-} tryMaybe $ do
-          reportSDoc "tc.lhs.unify" 60 $
-            text "Trying to instantiate the meta" <+> prettyTCM (MetaV m es) <+>
-            text "with the constructor" <+> prettyTCM c <+> text "applied to fresh metas"
-          margs <- do
-            -- The new metas should have the same dependencies as the original meta
-            mv <- lookupMeta m
-
-            ctype <- (`piApply` pars) . defType <$> liftTCM (getConstInfo $ conName c)
-            reportSDoc "tc.lhs.unify" 80 $ text "Type of constructor: " <+> prettyTCM ctype
-            withMetaInfo' mv $ do
-              let perm = mvPermutation mv
-              reportSDoc "tc.lhs.unify" 100 $ vcat
-                [ text "Permutation of meta: " <+> prettyTCM perm
-                ]
-              cxt <- instantiateFull =<< getContextTelescope
-              reportSDoc "tc.lhs.unify" 100 $ do
-                let flat = flattenTel cxt
-                let badRen  :: Substitution = renaming __IMPOSSIBLE__ perm
-                let goodRen :: Substitution = renaming __IMPOSSIBLE__ $ flipP perm
-                vcat
-                  [ text "Context of meta: " <+> (inTopContext $ prettyTCM cxt)
-                  , text "Flattened:       " <+> prettyTCM flat
-                  , text "Flattened (raw): " <+> text (show flat)
-                  , text "Bad renaming:    " <+> text (show badRen)
-                  , text "Good renaming:   " <+> text (show goodRen)
-                  , text "Raw permutation: " <+> prettyTCM (permute perm flat)
-                  ]
-              let tel = permuteTel perm cxt
-              reportSDoc "tc.lhs.unify" 100 $ text "Context tel (for new metas): " <+> prettyTCM tel
-              -- important: create the meta in the same environment as the original meta
-              newArgsMetaCtx ctype tel perm us
-          reportSDoc "tc.lhs.unify" 80 $ text "Generated meta args: " <+> prettyTCM margs
-          noConstraints $ assignV DirEq m us (Con c ci margs)
-          return margs
-
 checkEqualityStrategy :: Int -> UnifyStrategy
 checkEqualityStrategy k s = do
   let Equal a u v = getEquality k s
@@ -796,11 +732,13 @@ checkEqualityStrategy k s = do
 
 literalStrategy :: Int -> UnifyStrategy
 literalStrategy k s = do
-  eq <- liftTCM $ eqUnLevel $ getEquality k s
-  case eq of
-    Equal a u@(Lit l1) v@(Lit l2)
-     | l1 == l2  -> return $ Deletion k a u u -- TODO: wrong context of a, but does it matter?
-     | otherwise -> return $ LitConflict k a l1 l2 -- same problem here
+  let n = eqCount s
+  Equal a u v <- liftTCM $ eqUnLevel $ getEquality k s
+  ha <- mfromMaybe $ isHom n a
+  case (u , v) of
+    (Lit l1 , Lit l2)
+     | l1 == l2  -> return $ Deletion k ha u v
+     | otherwise -> return $ LitConflict k ha l1 l2
     _ -> mzero
 
 etaExpandVarStrategy :: Int -> UnifyStrategy
@@ -881,7 +819,7 @@ injectiveTypeConStrategy k s = do
                 Datatype{} -> True
                 Record{}   -> True
                 Axiom{}    -> True
-                AbstractDefn -> False -- True triggers issue #2250
+                AbstractDefn{} -> False -- True triggers issue #2250
                 Function{}   -> False
                 Primitive{}  -> False
                 Constructor{}-> __IMPOSSIBLE__  -- Never a type!
@@ -890,6 +828,19 @@ injectiveTypeConStrategy k s = do
       return $ TypeConInjectivity k d us vs
     _ -> mzero
 
+injectivePragmaStrategy :: Int -> UnifyStrategy
+injectivePragmaStrategy k s = do
+  eq <- liftTCM $ eqUnLevel $ getEquality k s
+  case eq of
+    Equal a u@(ignoreSharing -> Def d es) v@(ignoreSharing -> Def d' es') | d == d' -> do
+      -- d must have an injective pragma
+      def <- liftTCM $ getConstInfo d
+      guard $ defInjective def
+      let us = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
+          vs = fromMaybe __IMPOSSIBLE__ $ allApplyElims es'
+      return $ TypeConInjectivity k d us vs
+    _ -> mzero
+
 skipIrrelevantStrategy :: Int -> UnifyStrategy
 skipIrrelevantStrategy k s = do
   let i = getEqInfo k s
@@ -941,24 +892,43 @@ runUnifyM = runWriterT
 unifyStep :: UnifyState -> UnifyStep -> UnifyM (UnificationResult' UnifyState)
 
 unifyStep s Deletion{ deleteAt = k , deleteType = a , deleteLeft = u , deleteRight = v } = do
-    liftTCM $ addContext (varTel s) $ noConstraints $ dontAssignMetas $ equalTerm a u v
-    ifM (liftTCM $ optWithoutK <$> pragmaOptions)
-    {-then-} (DontKnow <$> liftTCM withoutKErr)
-    {-else-} (do
-      let (s', sigma) = solveEq k u s
-      tellUnifyProof sigma
-      Unifies <$> liftTCM (reduceEqTel s'))
-  `catchError` \err -> return $ DontKnow err
-  where
-    withoutKErr = addContext (varTel s) $ typeError_ $ WithoutKError a u u
+    -- Check definitional equality of u and v
+    isReflexive <- liftTCM $ addContext (varTel s) $ do
+      dontAssignMetas $ disableDestructiveUpdate $ noConstraints $
+        equalTerm a u v
+      return Nothing
+      `catchError` \err -> return $ Just err
+    withoutK <- liftTCM $ optWithoutK <$> pragmaOptions
+    case isReflexive of
+      Just err     -> return $ DontKnow []
+      _ | withoutK -> return $ DontKnow [UnifyReflexiveEq (varTel s) a u]
+      _            -> do
+        let (s', sigma) = solveEq k u s
+        tellUnifyProof sigma
+        Unifies <$> liftTCM (reduceEqTel s')
 
 unifyStep s Solution{ solutionAt = k , solutionType = a , solutionVar = i , solutionTerm = u } = do
   let m = varCount s
-  caseMaybeM (trySolveVar (m-1-i) u s) (DontKnow <$> err) $ \(s',sub) -> do
-    tellUnifySubst sub
-    let (s'', sigma) = solveEq k (applyPatSubst sub u) s'
-    tellUnifyProof sigma
-    Unifies <$> liftTCM (reduce s'')
+
+  -- Check that the type of the variable is equal to the type of the equation
+  -- (not just a subtype), otherwise we cannot instantiate (see Issue 2407).
+  let a' = getVarType (m-1-i) s
+  equalTypes <- liftTCM $ addContext (varTel s) $ do
+    reportSDoc "tc.lhs.unify" 45 $ text "Equation type: " <+> prettyTCM a
+    reportSDoc "tc.lhs.unify" 45 $ text "Variable type: " <+> prettyTCM a'
+    dontAssignMetas $ disableDestructiveUpdate $ noConstraints $
+      equalType a a'
+    return Nothing
+    `catchError` \err -> return $ Just err
+  case equalTypes of
+    Just err -> return $ DontKnow []
+    Nothing  -> caseMaybeM (trySolveVar (m-1-i) u s)
+      (return $ DontKnow [UnifyRecursiveEq (varTel s) a i u])
+      (\(s',sub) -> do
+        tellUnifySubst sub
+        let (s'', sigma) = solveEq k (applyPatSubst sub u) s'
+        tellUnifyProof sigma
+        Unifies <$> liftTCM (reduce s''))
   where
     trySolveVar i u s = case solveVar i u s of
       Just x  -> return $ Just x
@@ -966,7 +936,6 @@ unifyStep s Solution{ solutionAt = k , solutionType = a , solutionVar = i , solu
         u <- liftTCM $ normalise u
         s <- liftTCM $ normaliseVarTel s
         return $ solveVar i u s
-    err = addContext (varTel s) $ typeError_ $ UnificationRecursiveEq a i u
 
 unifyStep s (Injectivity k a d pars ixs c) = do
   withoutK <- liftTCM $ optWithoutK <$> pragmaOptions
@@ -974,7 +943,8 @@ unifyStep s (Injectivity k a d pars ixs c) = do
 
   -- Get constructor telescope and target indices
   ctype <- (`piApply` pars) . defType <$> liftTCM (getConInfo c)
-  reportSDoc "tc.lhs.unify" 40 $ text "Constructor type: " <+> prettyTCM ctype
+  addContext (varTel s) $ reportSDoc "tc.lhs.unify" 40 $
+    text "Constructor type: " <+> prettyTCM ctype
   TelV ctel ctarget <- liftTCM $ telView ctype
   let cixs = case ignoreSharing $ unEl ctarget of
                Def d' es | d == d' ->
@@ -984,10 +954,11 @@ unifyStep s (Injectivity k a d pars ixs c) = do
 
   -- Get index telescope of the datatype
   dtype    <- (`piApply` pars) . defType <$> liftTCM (getConstInfo d)
-  reportSDoc "tc.lhs.unify" 40 $ text "Datatype type: " <+> prettyTCM dtype
+  addContext (varTel s) $ reportSDoc "tc.lhs.unify" 40 $
+    text "Datatype type: " <+> prettyTCM dtype
 
   -- Split equation telescope into parts before and after current equation
-  let (eqListTel1, _ : eqListTel2) = genericSplitAt k $ telToList $ eqTel s
+  let (eqListTel1, _ : eqListTel2) = splitAt k $ telToList $ eqTel s
       (eqTel1, eqTel2) = (telFromList eqListTel1, telFromList eqListTel2)
 
   -- This is where the magic of higher-dimensional unification happens
@@ -996,10 +967,11 @@ unifyStep s (Injectivity k a d pars ixs c) = do
   -- recursively (this doesn't get stuck in a loop because a type should
   -- never be indexed over itself). Note the similarity with the
   -- computeNeighbourhood function in Agda.TypeChecking.Coverage.
+  let hduTel = eqTel1 `abstract` raise (size eqTel1) ctel
   res <- liftTCM $ addContext (varTel s) $ unifyIndices
-           (eqTel1 `abstract` ctel)
-           (allFlexVars $ eqTel1 `abstract` ctel)
-           dtype
+           hduTel
+           (allFlexVars hduTel)
+           (raise (size hduTel) dtype)
            (raise (size ctel) ixs)
            (raiseFrom (size ctel) (size eqTel1) cixs)
   case res of
@@ -1010,7 +982,11 @@ unifyStep s (Injectivity k a d pars ixs c) = do
 
     -- TODO: we could still make progress here if not --without-K,
     -- but I'm not sure if it's necessary.
-    DontKnow _ -> liftTCM $ DontKnow <$> err
+    DontKnow _ -> let n           = eqCount s
+                      Equal a u v = getEquality k s
+                  in return $ DontKnow [UnifyIndicesNotVars
+                       (varTel s `abstract` eqTel s) a
+                       (raise n u) (raise n v) (raise (n-k) ixs)]
 
     Unifies (eqTel1', rho0, _) -> do
       -- Split ps0 into parts for eqTel1 and ctel
@@ -1040,31 +1016,22 @@ unifyStep s (Injectivity k a d pars ixs c) = do
 
       return $ Unifies $ s { eqTel = eqTel' , eqLHS = lhs' , eqRHS = rhs' }
 
-  where
-    err :: TCM TCErr
-    err = let n           = eqCount s
-              Equal a u v = getEquality k s
-          in addContext (varTel s `abstract` eqTel s) $ typeError_
-               (UnifyIndicesNotVars a (raise n u) (raise n v) (raise (n-k) ixs))
-
 unifyStep s Conflict
-  { conflictConLeft    = c
-  , conflictConRight   = c'
-  } = NoUnify <$> addContext (varTel s)
-        (typeError_ $ UnifyConflict c c')
+  { conflictLeft  = u
+  , conflictRight = v
+  } = return $ NoUnify $ UnifyConflict (varTel s) u v
 
 unifyStep s Cycle
   { cycleVar        = i
   , cycleOccursIn   = u
-  } = NoUnify <$> addContext (varTel s)
-        (typeError_ $ UnifyCycle i u)
+  } = return $ NoUnify $ UnifyCycle (varTel s) i u
 
 unifyStep s EtaExpandVar{ expandVar = fi, expandVarRecordType = d , expandVarParameters = pars } = do
   delta   <- liftTCM $ (`apply` pars) <$> getRecordFieldTypes d
   c       <- liftTCM $ getRecordConstructor d
   let nfields         = size delta
       (varTel', rho)  = expandTelescopeVar (varTel s) (m-1-i) delta c
-      projectFlexible = [ FlexibleVar (flexHiding fi) (projFlexKind j) (flexPos fi) (i+j) | j <- [0..nfields-1] ]
+      projectFlexible = [ FlexibleVar (flexHiding fi) (flexOrigin fi) (projFlexKind j) (flexPos fi) (i+j) | j <- [0..nfields-1] ]
   tellUnifySubst $ rho
   Unifies <$> liftTCM (reduce $ UState
     { varTel   = varTel'
@@ -1083,6 +1050,7 @@ unifyStep s EtaExpandVar{ expandVar = fi, expandVarRecordType = d , expandVarPar
       RecordFlex ks -> fromMaybe ImplicitFlex $ ks !!! j
       ImplicitFlex  -> ImplicitFlex
       DotFlex       -> DotFlex
+      OtherFlex     -> OtherFlex
 
     liftFlexible :: Int -> Int -> Maybe Int
     liftFlexible n j = if j == i then Nothing else Just (if j > i then j + (n-1) else j)
@@ -1113,8 +1081,7 @@ unifyStep s LitConflict
   { litType          = a
   , litConflictLeft  = l
   , litConflictRight = l'
-  } = NoUnify <$> addContext (varTel s)
-        (typeError_ $ UnequalTerms CmpEq (Lit l) (Lit l') a)
+  } = return $ NoUnify $ UnifyConflict (varTel s) (Lit l) (Lit l')
 
 unifyStep s (StripSizeSuc k u v) = do
   sizeTy <- liftTCM sizeType
@@ -1169,7 +1136,8 @@ unify s strategy = if isUnifyStateSolved s
                  -> UnifyM (UnificationResult' UnifyState)
                  -> UnifyM (UnificationResult' UnifyState)
     tryUnifyStep step fallback = do
-      reportSDoc "tc.lhs.unify" 20 $ text "trying unifyStep" <+> prettyTCM step
+      addContext (varTel s) $
+        reportSDoc "tc.lhs.unify" 20 $ text "trying unifyStep" <+> prettyTCM step
       x <- unifyStep s step
       case x of
         Unifies s'   -> do
@@ -1177,15 +1145,12 @@ unify s strategy = if isUnifyStateSolved s
           reportSDoc "tc.lhs.unify" 20 $ text "new unifyState:" <+> prettyTCM s'
           writeUnifyLog $ UnificationStep s step
           return x
-        NoUnify{}    -> return x
-        DontKnow err -> do
+        NoUnify{}     -> return x
+        DontKnow err1 -> do
           y <- fallback
           case y of
-            DontKnow{} -> return x
-            _          -> return y
+            DontKnow err2 -> return $ DontKnow $ err1 ++ err2
+            _             -> return y
 
     failure :: UnifyM (UnificationResult' a)
-    failure = do
-      err <- addContext (varTel s) $ typeError_ $
-               UnificationStuck (eqTel s) (map unArg $ eqLHS s) (map unArg $ eqRHS s)
-      return $ DontKnow err
+    failure = return $ DontKnow []
