@@ -35,6 +35,7 @@ import {-# SOURCE #-} Agda.TypeChecking.Rules.Decl (checkDecl)
 import Agda.Utils.Monad
 import Agda.Utils.Null
 import Agda.Utils.Permutation
+import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -93,11 +94,6 @@ checkRecDef i name ind eta con ps contel fields =
       -- compute the field telescope (does not include record parameters)
       let TelV ftel _ = telView' contype
 
-          -- A record is irrelevant if all of its fields are.
-          -- In this case, the associated module parameter will be irrelevant.
-          -- See issue 392.
-          recordRelevance = minimum $ Irrelevant : (map getRelevance $ telToList ftel)
-
       -- Compute correct type of constructor
 
       -- t = tel -> t0 where t0 must be a sort s
@@ -149,7 +145,24 @@ checkRecDef i name ind eta con ps contel fields =
           -- haveEta      = maybe (Inferred $ conInduction == Inductive && etaenabled) Specified eta
           con = ConHead conName conInduction $ map unArg fs
 
-      reportSDoc "tc.rec" 30 $ text "record constructor is " <+> text (show con)
+          -- A record is irrelevant if all of its fields are.
+          -- In this case, the associated module parameter will be irrelevant.
+          -- See issue 392.
+          -- Unless it's been declared coinductive or no-eta-equality (#2607).
+          recordRelevance
+            | eta          == Just False  = Relevant
+            | conInduction == CoInductive = Relevant
+            | otherwise                   = minimum $ Irrelevant : (map getRelevance $ telToList ftel)
+
+      -- Andreas, 2017-01-26, issue #2436
+      -- Disallow coinductive records with eta-equality
+      when (conInduction == CoInductive && etaEqualityToBool haveEta == True) $ do
+        typeError . GenericDocError =<< do
+          sep [ text "Agda doesn't like coinductive records with eta-equality."
+              , text "If you must, use pragma"
+              , text "{-# ETA" <+> prettyTCM name <+> text "#-}"
+              ]
+      reportSDoc "tc.rec" 30 $ text "record constructor is " <+> prettyTCM con
 
       -- Add the record definition.
 
@@ -176,8 +189,7 @@ checkRecDef i name ind eta con ps contel fields =
                   -- We retain the original user declaration [(co)inductive]
                   -- in case the record turns out to be recursive.
               -- Determined by positivity checker:
-              , recRecursive      = False
-              , recMutual         = []
+              , recMutual         = Nothing
               }
 
         -- Add record constructor to signature
@@ -185,6 +197,7 @@ checkRecDef i name ind eta con ps contel fields =
           defaultDefn defaultArgInfo conName (telh `abstract` contype) $
             Constructor
               { conPars   = npars
+              , conArity  = size fs
               , conSrcCon = con
               , conData   = name
               , conAbstr  = Info.defAbstract conInfo
@@ -197,7 +210,7 @@ checkRecDef i name ind eta con ps contel fields =
         addNamedInstance conName name
 
       -- Check that the fields fit inside the sort
-      contype `fitsIn` s
+      _ <- contype `fitsIn` s
 
       {- Andreas, 2011-04-27 WRONG because field types are checked again
          and then non-stricts should not yet be irrelevant
@@ -251,7 +264,7 @@ checkRecDef i name ind eta con ps contel fields =
           [ text "record section:"
           , nest 2 $ sep
             [ prettyTCM m <+> (inTopContext . prettyTCM =<< getContextTelescope)
-            , fsep $ punctuate comma $ map (text . show . getName) fields
+            , fsep $ punctuate comma $ map (return . P.pretty . getName) fields
             ]
           ]
         reportSDoc "tc.rec.def" 15 $ nest 2 $ vcat
@@ -276,9 +289,6 @@ checkRecDef i name ind eta con ps contel fields =
           setDefaultModuleParameters m
           checkRecordProjections m name hasNamedCon con tel' (raise 1 ftel) fields
 
-        -- Andreas 2012-02-13: postpone polarity computation until after positivity check
-        -- computePolarity name
-
       return ()
 
 {-| @checkRecordProjections m r q tel ftel fs@.
@@ -316,13 +326,14 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
       -- because then meta variables are created again.
       -- Instead, we take the field type t from the field telescope.
       reportSDoc "tc.rec.proj" 5 $ sep
-        [ text "checking projection" <+> text (show x)
+        [ text "checking projection" <+> prettyTCM x
         , nest 2 $ vcat
           [ text "top   =" <+> (inTopContext . prettyTCM =<< getContextTelescope)
           , text "tel   =" <+> (inTopContext . prettyTCM $ tel)
           , text "ftel1 =" <+> prettyTCM ftel1
           , text "t     =" <+> prettyTCM t
           , text "ftel2 =" <+> addContext ftel1 (underAbstraction_ ftel2 prettyTCM)
+          , text "abstr =" <+> (text . show) (Info.defAbstract info)
           ]
         ]
 
@@ -391,19 +402,21 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
             cpo    = if hasNamedCon then ConOCon else ConORec
             cpi    = ConPatternInfo (Just cpo) (Just $ argFromDom $ fmap snd rt)
             conp   = defaultArg $ ConP con cpi $
-                     [ Arg info $ unnamed $ varP "x" | Dom info _ <- telToList ftel ]
+                     [ Arg ai' $ unnamed $ varP "x" | Dom ai' _ <- telToList ftel ]
             body   = Just $ bodyMod $ var (size ftel2)
             cltel  = ftel
-            clause = Clause { clauseRange = getRange info
+            clause = Clause { clauseLHSRange  = getRange info
+                            , clauseFullRange = getRange info
                             , clauseTel       = killRange cltel
                             , namedClausePats = [Named Nothing <$> numberPatVars __IMPOSSIBLE__ (idP $ size ftel) conp]
                             , clauseBody      = body
                             , clauseType      = Just $ Arg ai t
                             , clauseCatchall  = False
+                            , clauseUnreachable = Just False
                             }
 
         let projection = Projection
-              { projProper   = True
+              { projProper   = Just r
               , projOrig     = projname
               -- name of the record type:
               , projFromType = defaultArg r
@@ -433,7 +446,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
               -- functions.
         cc <- compileClauses Nothing [clause]
 
-        reportSDoc "tc.cc" 10 $ do
+        reportSDoc "tc.cc" 60 $ do
           sep [ text "compiled clauses of " <+> prettyTCM projname
               , nest 2 $ text (show cc)
               ]
@@ -449,7 +462,7 @@ checkRecordProjections m r hasNamedCon con tel ftel fs = do
                 , funCopatternLHS   = isCopatternLHS [clause]
                 })
               { defArgOccurrences = [StrictPos] }
-          computePolarity projname
+          computePolarity [projname]
           when (Info.defInstance info == InstanceDef) $
             addTypedInstance projname finalt
 
