@@ -10,10 +10,11 @@ import Agda.TypeChecking.Serialise.Instances.Common ()
 import Agda.TypeChecking.Serialise.Instances.Compilers ()
 import Agda.TypeChecking.Serialise.Instances.Highlighting ()
 import Agda.TypeChecking.Serialise.Instances.Internal ()
+import Agda.TypeChecking.Serialise.Instances.Errors ()
 
 instance EmbPrj Interface where
-  icod_ (Interface a b c d e f g h i j k l m n) = icode14' a b c d e f g h i j k l m n
+  icod_ (Interface a b c d e f g h i j k l m) = icodeN' Interface a b c d e f g h i j k l m
 
   value = vcase valu where
-    valu [a, b, c, d, e, f, g, h, i, j, k, l, m, n] = valu14 Interface a b c d e f g h i j k l m n
-    valu _                                          = malformed
+    valu [a, b, c, d, e, f, g, h, i, j, k, l, m] = valuN Interface a b c d e f g h i j k l m
+    valu _                                       = malformed
