@@ -1,6 +1,5 @@
-{-# LANGUAGE CPP                       #-}
-{-# LANGUAGE DeriveDataTypeable        #-}
-{-# LANGUAGE ScopedTypeVariables       #-}
+{-# LANGUAGE CPP                #-}
+{-# LANGUAGE DeriveDataTypeable #-}
 
 #if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
@@ -9,11 +8,14 @@
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
-module Agda.TypeChecking.Serialise.Instances.Common () where
+module Agda.TypeChecking.Serialise.Instances.Common (SerialisedRange(..)) where
+
+import Prelude hiding (mapM)
 
 import Control.Applicative
-import Control.Monad.Reader
+import Control.Monad.Reader hiding (mapM)
 import Control.Monad.State.Strict (gets, modify)
+import Control.Exception
 
 import Data.Array.IArray
 import Data.Word
@@ -29,9 +31,12 @@ import Data.Set (Set)
 import qualified Data.Set as Set
 import Data.Sequence (Seq)
 import qualified Data.Sequence as Seq
+import Data.Traversable ( mapM )
+
 #if __GLASGOW_HASKELL__ <= 708
-import Data.Typeable (Typeable)
+import Data.Typeable ( Typeable )
 #endif
+
 import Data.Void
 
 import Agda.Syntax.Common
@@ -51,10 +56,15 @@ import qualified Agda.Utils.BiMap as BiMap
 import Agda.Utils.HashMap (HashMap)
 import qualified Agda.Utils.HashMap as HMap
 import Agda.Utils.FileName
+import Agda.Utils.Maybe
 import qualified Agda.Utils.Maybe.Strict as Strict
+import Agda.Utils.Trie
 
 import Agda.Utils.Except
 
+import Agda.Utils.Empty (Empty)
+import qualified Agda.Utils.Empty as Empty
+
 #include "undefined.h"
 import Agda.Utils.Impossible
 
@@ -75,7 +85,7 @@ instance EmbPrj Integer where
   value i = (! i) `fmap` gets integerE
 
 instance EmbPrj Word64 where
-  icod_ i = icode2' (int32 q) (int32 r)
+  icod_ i = icodeN' (undefined :: Int32 -> Int32 -> Int32) (int32 q) (int32 r)
     where (q, r) = quotRem i (2^32)
           int32 :: Word64 -> Int32
           int32 = fromIntegral
@@ -106,42 +116,38 @@ instance EmbPrj Void where
   value = vcase valu where valu _ = malformed
 
 instance EmbPrj () where
-  icod_ () = icode0'
+  icod_ () = icodeN' ()
 
   value = vcase valu where
-    valu [] = valu0 ()
+    valu [] = valuN ()
     valu _  = malformed
 
 instance (EmbPrj a, EmbPrj b) => EmbPrj (a, b) where
-  icod_ (a, b) = icode2' a b
+  icod_ (a, b) = icodeN' (,) a b
 
-  value = vcase valu where
-    valu [a, b] = valu2 (,) a b
-    valu _      = malformed
+  value = valueN (,)
 
 instance (EmbPrj a, EmbPrj b, EmbPrj c) => EmbPrj (a, b, c) where
-  icod_ (a, b, c) = icode3' a b c
+  icod_ (a, b, c) = icodeN' (,,) a b c
 
-  value = vcase valu where
-    valu [a, b, c] = valu3 (,,) a b c
-    valu _         = malformed
+  value = valueN (,,)
 
 instance (EmbPrj a, EmbPrj b) => EmbPrj (Either a b) where
-  icod_ (Left  x) = icode1 0 x
-  icod_ (Right x) = icode1 1 x
+  icod_ (Left  x) = icodeN 0 Left x
+  icod_ (Right x) = icodeN 1 Right x
 
   value = vcase valu where
-    valu [0, x] = valu1 Left  x
-    valu [1, x] = valu1 Right x
+    valu [0, x] = valuN Left  x
+    valu [1, x] = valuN Right x
     valu _   = malformed
 
 instance EmbPrj a => EmbPrj (Maybe a) where
-  icod_ Nothing  = icode0'
-  icod_ (Just x) = icode1' x
+  icod_ Nothing  = icodeN' Nothing
+  icod_ (Just x) = icodeN' Just x
 
   value = vcase valu where
-    valu []  = valu0 Nothing
-    valu [x] = valu1 Just x
+    valu []  = valuN Nothing
+    valu [x] = valuN Just x
     valu _   = malformed
 
 instance EmbPrj a => EmbPrj (Strict.Maybe a) where
@@ -149,18 +155,37 @@ instance EmbPrj a => EmbPrj (Strict.Maybe a) where
   value m = Strict.toStrict `fmap` value m
 
 instance EmbPrj Bool where
-  icod_ True  = icode0'
-  icod_ False = icode0 0
+  icod_ True  = icodeN' True
+  icod_ False = icodeN 0 False
 
   value = vcase valu where
-    valu []  = valu0 True
-    valu [0] = valu0 False
+    valu []  = valuN True
+    valu [0] = valuN False
     valu _   = malformed
 
+instance EmbPrj DataOrRecord where
+  icod_ IsData   = icodeN' IsData
+  icod_ IsRecord = icodeN 0 IsRecord
+
+  value = vcase $ \case
+    []  -> valuN IsData
+    [0] -> valuN IsRecord
+    _   -> malformed
+
 instance EmbPrj AbsolutePath where
   icod_ file = do
-    d <-  asks absPathD
-    liftIO $ fromMaybe __IMPOSSIBLE__ <$> H.lookup d file
+    d <- asks absPathD
+    liftIO $ flip fromMaybeM (H.lookup d file) $ do
+      -- The path @file@ should be cached in the dictionary @d@.
+      -- This seems not to be the case, thus, crash here.
+      -- But leave some hints for the posterity why things could go so wrong.
+      -- reportSLn "impossible" 10 -- does not work here
+      putStrLn $ unlines $
+        [ "Panic while serializing absolute path: " ++ show file
+        , "The path could not be found in the dictionary:"
+        ]
+      putStrLn . show =<< H.toList d
+      __IMPOSSIBLE__
 
   value m = do
     m :: TopLevelModuleName
@@ -174,25 +199,21 @@ instance EmbPrj AbsolutePath where
       Right f  -> return f
 
 instance EmbPrj a => EmbPrj (Position' a) where
-  icod_ (P.Pn file pos line col) = icode4' file pos line col
+  icod_ (P.Pn file pos line col) = icodeN' P.Pn file pos line col
 
-  value = vcase valu where
-    valu [f, p, l, c] = valu4 P.Pn f p l c
-    valu _            = malformed
+  value = valueN P.Pn
 
 instance EmbPrj TopLevelModuleName where
-  icod_ (TopLevelModuleName a) = icode1' a
+  icod_ (TopLevelModuleName a b) = icodeN' TopLevelModuleName a b
 
-  value = vcase valu where
-    valu [a] = valu1 TopLevelModuleName a
-    valu _   = malformed
+  value = valueN TopLevelModuleName
 
 #if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPABLE #-} EmbPrj a => EmbPrj [a] where
 #else
 instance EmbPrj a => EmbPrj [a] where
 #endif
-  icod_ xs = icodeN =<< mapM icode xs
+  icod_ xs = icodeNode =<< mapM icode xs
   value    = vcase (mapM value)
 --   icode []       = icode0'
 --   icode (x : xs) = icode2' x xs
@@ -212,21 +233,24 @@ instance (Ord a, EmbPrj a) => EmbPrj (Set a) where
   icod_ s = icode (Set.toList s)
   value s = Set.fromList `fmap` value s
 
+instance (Ord a, EmbPrj a, EmbPrj b) => EmbPrj (Trie a b) where
+  icod_ (Trie a b)= icodeN' Trie a b
+
+  value = valueN Trie
+
 instance EmbPrj a => EmbPrj (Seq a) where
   icod_ s = icode (Fold.toList s)
   value s = Seq.fromList `fmap` value s
 
 instance EmbPrj a => EmbPrj (P.Interval' a) where
-  icod_ (P.Interval p q) = icode2' p q
+  icod_ (P.Interval p q) = icodeN' P.Interval p q
 
-  value = vcase valu where
-    valu [p, q] = valu2 P.Interval p q
-    valu _      = malformed
+  value = valueN P.Interval
 
 -- | Ranges are always deserialised as 'noRange'.
 
 instance EmbPrj Range where
-  icod_ _ = icode0'
+  icod_ _ = icodeN' ()
   value _ = return noRange
 
 -- | Ranges that should be serialised properly.
@@ -238,84 +262,81 @@ newtype SerialisedRange = SerialisedRange { underlyingRange :: Range }
 
 instance EmbPrj SerialisedRange where
   icod_ (SerialisedRange r) =
-    icode2' (P.rangeFile r) (P.rangeIntervals r)
+    icodeN' (undefined :: SrcFile -> [IntervalWithoutFile] -> SerialisedRange)
+            (P.rangeFile r) (P.rangeIntervals r)
 
   value = vcase valu where
-    valu [a, b] = SerialisedRange <$> valu2 P.intervalsToRange a b
+    valu [a, b] = SerialisedRange <$> valuN P.intervalsToRange a b
     valu _      = malformed
 
 instance EmbPrj C.Name where
-  icod_ (C.NoName a b) = icode2 0 a b
-  icod_ (C.Name r xs)  = icode2' r xs
+  icod_ (C.NoName a b) = icodeN 0 C.NoName a b
+  icod_ (C.Name r xs)  = icodeN' C.Name r xs
 
   value = vcase valu where
-    valu [0, a, b] = valu2 C.NoName a b
-    valu [r, xs]   = valu2 C.Name   r xs
+    valu [0, a, b] = valuN C.NoName a b
+    valu [r, xs]   = valuN C.Name   r xs
     valu _         = malformed
 
 instance EmbPrj NamePart where
-  icod_ Hole   = icode0'
-  icod_ (Id a) = icode1' a
+  icod_ Hole   = icodeN' Hole
+  icod_ (Id a) = icodeN' Id a
 
   value = vcase valu where
-    valu []  = valu0 Hole
-    valu [a] = valu1 Id a
+    valu []  = valuN Hole
+    valu [a] = valuN Id a
     valu _   = malformed
 
 instance EmbPrj C.QName where
-  icod_ (Qual    a b) = icode2' a b
-  icod_ (C.QName a  ) = icode1' a
+  icod_ (Qual    a b) = icodeN' Qual a b
+  icod_ (C.QName a  ) = icodeN' C.QName a
 
   value = vcase valu where
-    valu [a, b] = valu2 Qual    a b
-    valu [a]    = valu1 C.QName a
+    valu [a, b] = valuN Qual    a b
+    valu [a]    = valuN C.QName a
     valu _      = malformed
 
 instance EmbPrj Agda.Syntax.Fixity.Associativity where
-  icod_ LeftAssoc  = icode0'
-  icod_ RightAssoc = icode0 1
-  icod_ NonAssoc   = icode0 2
+  icod_ LeftAssoc  = icodeN' LeftAssoc
+  icod_ RightAssoc = icodeN 1 RightAssoc
+  icod_ NonAssoc   = icodeN 2 NonAssoc
 
   value = vcase valu where
-    valu []  = valu0 LeftAssoc
-    valu [1] = valu0 RightAssoc
-    valu [2] = valu0 NonAssoc
+    valu []  = valuN LeftAssoc
+    valu [1] = valuN RightAssoc
+    valu [2] = valuN NonAssoc
     valu _   = malformed
 
 instance EmbPrj Agda.Syntax.Fixity.PrecedenceLevel where
-  icod_ Unrelated   = icode0'
-  icod_ (Related a) = icode1' a
+  icod_ Unrelated   = icodeN' Unrelated
+  icod_ (Related a) = icodeN' Related a
 
   value = vcase valu where
-    valu []  = valu0 Unrelated
-    valu [a] = valu1 Related a
+    valu []  = valuN Unrelated
+    valu [a] = valuN Related a
     valu _   = malformed
 
 instance EmbPrj Agda.Syntax.Fixity.Fixity where
-  icod_ (Fixity a b c) = icode3' a b c
+  icod_ (Fixity a b c) = icodeN' Fixity a b c
 
-  value = vcase valu where
-    valu [a, b, c] = valu3 Fixity a b c
-    valu _         = malformed
+  value = valueN Fixity
 
 instance EmbPrj Agda.Syntax.Fixity.Fixity' where
-  icod_ (Fixity' a b _) = icode2' a b  -- discard theNameRange
+  icod_ (Fixity' a b r) = icodeN' (\ a b -> Fixity' a b r) a b  -- discard theNameRange
 
-  value = vcase valu where
-    valu [a, b] = valu2 (\ f n -> Fixity' f n noRange) a b
-    valu _      = malformed
+  value = valueN (\ f n -> Fixity' f n noRange)
 
 instance EmbPrj GenPart where
-  icod_ (BindHole a)   = icode1 0 a
-  icod_ (NormalHole a) = icode1 1 a
-  icod_ (WildHole a)   = icode1 2 a
-  icod_ (IdPart a)     = icode1' a
+  icod_ (BindHole a)   = icodeN 0 BindHole a
+  icod_ (NormalHole a) = icodeN 1 NormalHole a
+  icod_ (WildHole a)   = icodeN 2 WildHole a
+  icod_ (IdPart a)     = icodeN' IdPart a
 
   value = vcase valu where
-    valu [0, a] = valu1 BindHole a
-    valu [1, a] = valu1 NormalHole a
-    valu [2, a] = valu1 WildHole a
-    valu [a]    = valu1 IdPart a
+    valu [0, a] = valuN BindHole a
+    valu [1, a] = valuN NormalHole a
+    valu [2, a] = valuN WildHole a
+    valu [a]    = valuN IdPart a
     valu _      = malformed
 
 instance EmbPrj MetaId where
@@ -323,11 +344,9 @@ instance EmbPrj MetaId where
   value i = MetaId <$> value i
 
 instance EmbPrj A.QName where
-  icod_ n@(A.QName a b) = icodeMemo qnameD qnameC (qnameId n) $ icode2' a b
+  icod_ n@(A.QName a b) = icodeMemo qnameD qnameC (qnameId n) $ icodeN' A.QName a b
 
-  value = vcase valu where
-    valu [a, b] = valu2 A.QName a b
-    valu _      = malformed
+  value = valueN A.QName
 
 instance EmbPrj A.AmbiguousQName where
   icod_ (A.AmbQ a) = icode a
@@ -339,90 +358,73 @@ instance EmbPrj A.ModuleName where
 
 instance EmbPrj A.Name where
   icod_ (A.Name a b c d) = icodeMemo nameD nameC a $
-    icode4' a b (SerialisedRange c) d
+    icodeN' (\ a b -> A.Name a b . underlyingRange) a b (SerialisedRange c) d
 
-  value = vcase valu where
-    valu [a, b, c, d] = valu4 (\a b c -> A.Name a b (underlyingRange c))
-                              a b c d
-    valu _            = malformed
+  value = valueN (\a b c -> A.Name a b (underlyingRange c))
 
 instance EmbPrj a => EmbPrj (C.FieldAssignment' a) where
-  icod_ (C.FieldAssignment a b) = icode2' a b
+  icod_ (C.FieldAssignment a b) = icodeN' C.FieldAssignment a b
 
-  value = vcase valu where
-    valu [a, b] = valu2 C.FieldAssignment a b
-    valu _      = malformed
+  value = valueN C.FieldAssignment
 
 instance (EmbPrj s, EmbPrj t) => EmbPrj (Named s t) where
-  icod_ (Named a b) = icode2' a b
+  icod_ (Named a b) = icodeN' Named a b
 
-  value = vcase valu where
-    valu [a, b] = valu2 Named a b
-    valu _      = malformed
+  value = valueN Named
 
 instance EmbPrj a => EmbPrj (Ranged a) where
-  icod_ (Ranged r x) = icode2' r x
+  icod_ (Ranged r x) = icodeN' Ranged r x
 
-  value = vcase valu where
-    valu [r, x] = valu2 Ranged r x
-    valu _      = malformed
+  value = valueN Ranged
 
 instance EmbPrj ArgInfo where
-  icod_ (ArgInfo h r o v) = icode4' h r o v
+  icod_ (ArgInfo h r o) = icodeN' ArgInfo h r o
 
-  value = vcase valu where
-    valu [h, r, o, v] = valu4 ArgInfo h r o v
-    valu _            = malformed
+  value = valueN ArgInfo
 
 instance EmbPrj NameId where
-  icod_ (NameId a b) = icode2' a b
+  icod_ (NameId a b) = icodeN' NameId a b
 
-  value = vcase valu where
-    valu [a, b] = valu2 NameId a b
-    valu _      = malformed
+  value = valueN NameId
 
 instance (Eq k, Hashable k, EmbPrj k, EmbPrj v) => EmbPrj (HashMap k v) where
   icod_ m = icode (HMap.toList m)
   value m = HMap.fromList `fmap` value m
 
 instance EmbPrj a => EmbPrj (WithHiding a) where
-  icod_ (WithHiding a b) = icode2' a b
+  icod_ (WithHiding a b) = icodeN' WithHiding a b
 
-  value = vcase valu where
-    valu [a, b] = valu2 WithHiding a b
-    valu _      = malformed
+  value = valueN WithHiding
 
 instance EmbPrj a => EmbPrj (Arg a) where
-  icod_ (Arg i e) = icode2' i e
+  icod_ (Arg i e) = icodeN' Arg i e
 
-  value = vcase valu where
-    valu [i, e] = valu2 Arg i e
-    valu _      = malformed
+  value = valueN Arg
 
 instance EmbPrj a => EmbPrj (Dom a) where
-  icod_ (Dom i e) = icode2' i e
+  icod_ (Dom i e) = icodeN' Dom i e
 
-  value = vcase valu where
-    valu [i, e] = valu2 Dom i e
-    valu _      = malformed
+  value = valueN Dom
 
 instance EmbPrj Induction where
-  icod_ Inductive   = icode0'
-  icod_ CoInductive = icode0 1
+  icod_ Inductive   = icodeN' Inductive
+  icod_ CoInductive = icodeN 1 CoInductive
 
   value = vcase valu where
-    valu []  = valu0 Inductive
-    valu [1] = valu0 CoInductive
+    valu []  = valuN Inductive
+    valu [1] = valuN CoInductive
     valu _   = malformed
 
 instance EmbPrj Hiding where
-  icod_ Hidden    = return 0
-  icod_ NotHidden = return 1
-  icod_ Instance  = return 2
+  icod_ Hidden                = return 0
+  icod_ NotHidden             = return 1
+  icod_ (Instance NoOverlap)  = return 2
+  icod_ (Instance YesOverlap) = return 3
 
   value 0 = return Hidden
   value 1 = return NotHidden
-  value 2 = return Instance
+  value 2 = return (Instance NoOverlap)
+  value 3 = return (Instance YesOverlap)
   value _ = malformed
 
 instance EmbPrj Relevance where
@@ -431,34 +433,36 @@ instance EmbPrj Relevance where
   icod_ (Forced Small) = return 2
   icod_ (Forced Big)   = return 3
   icod_ NonStrict      = return 4
-  icod_ UnusedArg      = return 5
 
   value 0 = return Relevant
   value 1 = return Irrelevant
   value 2 = return (Forced Small)
   value 3 = return (Forced Big)
   value 4 = return NonStrict
-  value 5 = return UnusedArg
   value _ = malformed
 
 instance EmbPrj Origin where
   icod_ UserWritten = return 0
   icod_ Inserted    = return 1
   icod_ Reflected   = return 2
+  icod_ CaseSplit   = return 3
 
   value 0 = return UserWritten
   value 1 = return Inserted
   value 2 = return Reflected
+  value 3 = return CaseSplit
   value _ = malformed
 
 instance EmbPrj ConOrigin where
   icod_ ConOSystem = return 0
   icod_ ConOCon    = return 1
   icod_ ConORec    = return 2
+  icod_ ConOSplit  = return 3
 
   value 0 = return ConOSystem
   value 1 = return ConOCon
   value 2 = return ConORec
+  value 3 = return ConOSplit
   value _ = malformed
 
 instance EmbPrj ProjOrigin where
@@ -472,36 +476,50 @@ instance EmbPrj ProjOrigin where
   value _ = malformed
 
 instance EmbPrj Agda.Syntax.Literal.Literal where
-  icod_ (LitNat    a b)   = icode2' a b
-  icod_ (LitFloat  a b)   = icode2 1 a b
-  icod_ (LitString a b)   = icode2 2 a b
-  icod_ (LitChar   a b)   = icode2 3 a b
-  icod_ (LitQName  a b)   = icode2 5 a b
-  icod_ (LitMeta   a b c) = icode3 6 a b c
+  icod_ (LitNat    a b)   = icodeN' LitNat a b
+  icod_ (LitFloat  a b)   = icodeN 1 LitFloat a b
+  icod_ (LitString a b)   = icodeN 2 LitString a b
+  icod_ (LitChar   a b)   = icodeN 3 LitChar a b
+  icod_ (LitQName  a b)   = icodeN 5 LitQName a b
+  icod_ (LitMeta   a b c) = icodeN 6 LitMeta a b c
 
   value = vcase valu where
-    valu [a, b]       = valu2 LitNat    a b
-    valu [1, a, b]    = valu2 LitFloat  a b
-    valu [2, a, b]    = valu2 LitString a b
-    valu [3, a, b]    = valu2 LitChar   a b
-    valu [5, a, b]    = valu2 LitQName  a b
-    valu [6, a, b, c] = valu3 LitMeta   a b c
+    valu [a, b]       = valuN LitNat    a b
+    valu [1, a, b]    = valuN LitFloat  a b
+    valu [2, a, b]    = valuN LitString a b
+    valu [3, a, b]    = valuN LitChar   a b
+    valu [5, a, b]    = valuN LitQName  a b
+    valu [6, a, b, c] = valuN LitMeta   a b c
     valu _            = malformed
 
 instance EmbPrj IsAbstract where
-  icod_ AbstractDef = icode0 0
-  icod_ ConcreteDef = icode0'
+  icod_ AbstractDef = icodeN 0 AbstractDef
+  icod_ ConcreteDef = icodeN' ConcreteDef
 
   value = vcase valu where
-    valu [0] = valu0 AbstractDef
-    valu []  = valu0 ConcreteDef
+    valu [0] = valuN AbstractDef
+    valu []  = valuN ConcreteDef
     valu _   = malformed
 
 instance EmbPrj Delayed where
-  icod_ Delayed    = icode0 0
-  icod_ NotDelayed = icode0'
+  icod_ Delayed    = icodeN 0 Delayed
+  icod_ NotDelayed = icodeN' NotDelayed
 
   value = vcase valu where
-    valu [0] = valu0 Delayed
-    valu []  = valu0 NotDelayed
+    valu [0] = valuN Delayed
+    valu []  = valuN NotDelayed
     valu _   = malformed
+
+instance EmbPrj Impossible where
+  icod_ (Impossible a b)  = icodeN 0 Impossible a b
+  icod_ (Unreachable a b) = icodeN 1 Unreachable a b
+
+  value = vcase valu where
+    valu [0, a, b] = valuN Impossible  a b
+    valu [1, a, b] = valuN Unreachable a b
+    valu _         = malformed
+
+instance EmbPrj Empty where
+  icod_ a = icod_ =<< lift (Empty.toImpossible a)
+
+  value = fmap throwImpossible . value
