@@ -7,7 +7,7 @@ import Prelude hiding (null)
 
 import Control.Monad.Writer
 
-import Data.List hiding (null)
+import qualified Data.List as List
 import qualified Data.Map as Map
 
 import Agda.Syntax.Common
@@ -43,18 +43,31 @@ import Agda.Utils.Impossible
 -- | Check whether a type is either not a SIZELT or a SIZELT that is non-empty.
 checkSizeLtSat :: Term -> TCM ()
 checkSizeLtSat t = do
-  reportSDoc "tc.size" 10 $ sep
-    [ text "checking that " <+> prettyTCM t <+> text " is not an empty type of sizes"
-    , text "in context " <+> do inTopContext . prettyTCM =<< getContextTelescope
-    ]
+  reportSDoc "tc.size" 10 $ do
+    tel <- getContextTelescope
+    sep
+      [ text "checking that " <+> prettyTCM t <+> text " is not an empty type of sizes"
+      , if null tel then empty else do
+        text "in context " <+> inTopContext (prettyTCM tel)
+      ]
+  reportSLn "tc.size" 60 $ "- raw type = " ++ show t
   let postpone :: Term -> TCM ()
-      postpone t = addConstraint $ CheckSizeLtSat t
+      postpone t = do
+        reportSDoc "tc.size.lt" 20 $ sep
+          [ text "- postponing `not empty type of sizes' check for " <+> prettyTCM t ]
+        addConstraint $ CheckSizeLtSat t
+  let ok :: TCM ()
+      ok = reportSLn "tc.size.lt" 20 $ "- succeeded: not an empty type of sizes"
   ifBlocked t (const postpone) $ \ t -> do
-    caseMaybeM (isSizeType t) (return ()) $ \ b -> do
+    reportSLn "tc.size.lt" 20 $ "- type is not blocked"
+    caseMaybeM (isSizeType t) ok $ \ b -> do
+      reportSLn "tc.size.lt" 20 $ " - type is a size type"
       case b of
-        BoundedNo -> return ()
+        BoundedNo -> ok
         BoundedLt b -> do
+          reportSDoc "tc.size.lt" 20 $ text " - type is SIZELT" <+> prettyTCM b
           ifBlocked b (\ _ _ -> postpone t) $ \ b -> do
+            reportSLn "tc.size.lt" 20 $ " - size bound is not blocked"
             catchConstraint (CheckSizeLtSat t) $ do
               unlessM (checkSizeNeverZero b) $ do
                 typeError . GenericDocError =<< do
@@ -177,9 +190,9 @@ boundedSizeMetaHook v tel0 a = do
   case res of
     Just (BoundedLt u) -> do
       n <- getContextSize
-      let tel | n > 0     = telFromList $ genericDrop n $ telToList tel0
+      let tel | n > 0     = telFromList $ drop n $ telToList tel0
               | otherwise = tel0
-      addContext tel $ do
+      addContext' tel $ do
         v <- sizeSuc 1 $ raise (size tel) v `apply` teleArgs tel
         -- compareSizes CmpLeq v u
         size <- sizeType
@@ -286,8 +299,14 @@ compareMaxViews cmp us vs = case (cmp, us, vs) of
 
 -- | @compareBelowMax u vs@ checks @u <= max vs@.  Precondition: @size vs >= 2@
 compareBelowMax :: DeepSizeView -> SizeMaxView -> TCM ()
-compareBelowMax u vs =
+compareBelowMax u vs = do
+  reportSDoc "tc.conv.size" 45 $ vcat
+    [ text "compareBelowMax"
+    ]
   alt (dontAssignMetas $ alts $ map (compareSizeViews CmpLeq u) vs) $ do
+    reportSDoc "tc.conv.size" 45 $ vcat
+      [ text "compareBelowMax: giving up"
+      ]
     u <- unDeepSizeView u
     v <- unMaxView vs
     size <- sizeType
@@ -299,6 +318,12 @@ compareBelowMax u vs =
 
 compareSizeViews :: Comparison -> DeepSizeView -> DeepSizeView -> TCM ()
 compareSizeViews cmp s1' s2' = do
+  reportSDoc "tc.conv.size" 45 $ hsep
+    [ text "compareSizeViews"
+    , text (show s1')
+    , text (show cmp)
+    , text (show s2')
+    ]
   size <- sizeType
   let (s1, s2) = removeSucs (s1', s2')
       withUnView cont = do
@@ -359,14 +384,27 @@ isSizeConstraint :: Closure Constraint -> TCM Bool
 isSizeConstraint Closure{ clValue = ValueCmp _ s _ _ } = isJust <$> isSizeType s
 isSizeConstraint _ = return False
 
+-- | Take out all size constraints (DANGER!).
+takeSizeConstraints :: TCM [Closure Constraint]
+takeSizeConstraints = do
+  test <- isSizeTypeTest
+  let sizeConstraint :: Closure Constraint -> Bool
+      sizeConstraint cl@Closure{ clValue = ValueCmp CmpLeq s _ _ }
+              | isJust (test $ unEl s) = True
+      sizeConstraint _ = False
+  cs <- filter sizeConstraint . map theConstraint <$> getAllConstraints
+  dropConstraints $ sizeConstraint . theConstraint
+  return cs
+
 -- | Find the size constraints.
 getSizeConstraints :: TCM [Closure Constraint]
 getSizeConstraints = do
   test <- isSizeTypeTest
-  let sizeConstraint cl@Closure{ clValue = ValueCmp CmpLeq s _ _ }
-              | isJust (test $ unEl s) = Just cl
-      sizeConstraint _ = Nothing
-  mapMaybe (sizeConstraint . theConstraint) <$> getAllConstraints
+  let sizeConstraint :: Closure Constraint -> Bool
+      sizeConstraint cl@Closure{ clValue = ValueCmp CmpLeq s _ _ }
+              | isJust (test $ unEl s) = True
+      sizeConstraint _ = False
+  filter sizeConstraint . map theConstraint <$> getAllConstraints
 
 -- | Return a list of size metas and their context.
 getSizeMetas :: Bool -> TCM [(MetaId, Type, Telescope)]
@@ -524,17 +562,17 @@ oldCanonicalizeSizeConstraint c@(Leq a n b) =
   case (a,b) of
     (Rigid{}, Rigid{})       -> return c
     (SizeMeta m xs, Rigid i) -> do
-      j <- findIndex (==i) xs
+      j <- List.findIndex (==i) xs
       return $ Leq (SizeMeta m [0..size xs-1]) n (Rigid j)
     (Rigid i, SizeMeta m xs) -> do
-      j <- findIndex (==i) xs
+      j <- List.findIndex (==i) xs
       return $ Leq (Rigid j) n (SizeMeta m [0..size xs-1])
     (SizeMeta m xs, SizeMeta l ys)
          -- try to invert xs on ys
-       | Just ys' <- mapM (\ y -> findIndex (==y) xs) ys ->
+       | Just ys' <- mapM (\ y -> List.findIndex (==y) xs) ys ->
            return $ Leq (SizeMeta m [0..size xs-1]) n (SizeMeta l ys')
          -- try to invert ys on xs
-       | Just xs' <- mapM (\ x -> findIndex (==x) ys) xs ->
+       | Just xs' <- mapM (\ x -> List.findIndex (==x) ys) xs ->
            return $ Leq (SizeMeta m xs') n (SizeMeta l [0..size ys-1])
          -- give up
        | otherwise -> Nothing
@@ -562,7 +600,7 @@ oldSolveSizeConstraints = whenM haveSizedTypes $ do
 
         -- Size metas in constraints.
         metas0 :: [(MetaId, Int)]  -- meta id + arity
-        metas0 = nub $ map (mapSnd length) $ concatMap flexibleVariables cs
+        metas0 = List.nub $ map (mapSnd length) $ concatMap flexibleVariables cs
 
         -- Unconstrained size metas that do not occur in constraints.
         metas1 :: [(MetaId, Int)]
