@@ -1,6 +1,5 @@
-{-# LANGUAGE CPP                   #-}
+{-# LANGUAGE CPP                      #-}
 {-# LANGUAGE NondecreasingIndentation #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
 
 -- | Solving size constraints under hypotheses.
 --
@@ -103,12 +102,12 @@ import qualified Agda.Utils.List as List
 import Agda.Utils.Maybe
 import Agda.Utils.Monad
 import Agda.Utils.Null
+import Agda.Utils.Pretty (Pretty, prettyShow)
+import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
 import Agda.Utils.Tuple
 import qualified Agda.Utils.VarSet as VarSet
 
-import qualified Agda.Utils.Either as Either
-
 #include "undefined.h"
 import Agda.Utils.Impossible
 
@@ -125,9 +124,10 @@ data DefaultToInfty
 solveSizeConstraints :: DefaultToInfty -> TCM ()
 solveSizeConstraints flag =  do
 
-  -- 1. Get the constraints normalised.
+  -- 1. Take out the size constraints normalised.
 
-  cs0 <- mapM (mapClosure normalise) =<< S.getSizeConstraints
+  cs0 <- mapM (mapClosure normalise) =<< S.takeSizeConstraints
+    -- NOTE: this deletes the size constraints from the constraint set!
   unless (null cs0) $
     reportSDoc "tc.size.solve" 40 $ vcat $
       [ text $ "Solving constraints (" ++ show flag ++ ")"
@@ -215,6 +215,8 @@ solveSizeConstraints flag =  do
   --       forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
 
   -- 5. Make sure we did not lose any constraints.
+
+  -- This is necessary since we have removed the size constraints.
   forM_ cs0 $ \ cl -> enterClosure cl solveConstraint
 
 
@@ -341,7 +343,7 @@ castConstraintToCurrentContext cl = do
   -- Check that all the free variables of the constraint are contained in
   -- coveredVars.
   -- We ignore the free variables occurring in sorts.
-  guard $ getAll $ runFree (\ (i, _) -> All $ i `VarSet.member` coveredVars) IgnoreAll c
+  guard $ getAll $ runFree (All . (`VarSet.member` coveredVars)) IgnoreAll c
   -- Turn cand into a substitution.
   -- Since we ignored the free variables in sorts, we better patch up
   -- the substitution with some dummy term (Sort Prop) rather than __IMPOSSIBLE__.
@@ -493,10 +495,10 @@ solveCluster flag ccs = do
   -- Running the solver once might result in unsolvable left-over constraints.
   -- We need to iterate the solver to detect this.
   sol :: Solution NamedRigid Int <- either err return $
-    iterateSolver Map.empty hg csF Map.empty
+    iterateSolver Map.empty hg csF emptySolution
 
   -- Convert solution to meta instantiation.
-  forM_ (Map.assocs sol) $ \ (m, a) -> do
+  forM_ (Map.assocs $ theSolution sol) $ \ (m, a) -> do
     unless (validOffset a) __IMPOSSIBLE__
     -- Solution does not contain metas
     u <- unSizeExpr $ fmap __IMPOSSIBLE__ a
@@ -516,7 +518,8 @@ solveCluster flag ccs = do
       [ text $ "  xs = " ++ show xs
       , text $ "  u  = " ++ show u
       ]
-    assignMeta n x t xs u
+    unlessM (isFrozen x) $
+      assignMeta n x t xs u
     -- WRONG:
     -- let partialSubst = List.sort $ zip xs $ map var $ downFrom n
     -- assignMeta' n x t (length xs) partialSubst u
@@ -529,7 +532,7 @@ solveCluster flag ccs = do
 
   --  ms = unsolved size metas from cluster
   let ms = Set.fromList (map sizeMetaId metas) Set.\\  -- Some CPP or ghc does not like trailing backslash, thus, this comment!
-             Set.mapMonotonic MetaId (Map.keysSet sol)
+             Set.mapMonotonic MetaId (Map.keysSet $ theSolution sol)
   --  Make sure they do not contain an interaction point
   let noIP = Set.null $ Set.intersection ims ms
 
@@ -552,7 +555,7 @@ solveCluster flag ccs = do
               reportSDoc "tc.size.solve" 30 $
                 prettyTCM (MetaV m []) <+> text "is frozen, cannot set it to ∞"
               return False
-        ifM (isFrozen m) no $ {-else-} do
+        ifM (isFrozen m `or2M` do not <$> asks envAssignMetas) no $ {-else-} do
           reportSDoc "tc.size.solve" 20 $
             text "solution " <+> prettyTCM (MetaV m []) <+>
             text " := "      <+> prettyTCM inf
@@ -582,7 +585,7 @@ getSizeHypotheses gamma = inTopContext $ modifyContext (const gamma) $ do
       forM (zip [0..] gamma) $ \ (i, ce) -> do
         -- Get name and type of variable i.
         let xt = unDom $ ctxEntry ce
-            x  = show $ fst xt
+            x  = prettyShow $ fst xt
         t <- reduce . raise (1 + i) . unEl . snd $ xt
         case ignoreSharing t of
           Def d [Apply u] | d == sizelt -> do
@@ -656,11 +659,11 @@ canonicalizeSizeConstraint c@(Constraint a cmp b) = Just c
 data NamedRigid = NamedRigid
   { rigidName  :: String   -- ^ Name for printing in debug messages.
   , rigidIndex :: Int      -- ^ De Bruijn index.
-  }
+  } deriving (Show)
 
 instance Eq NamedRigid where (==) = (==) `on` rigidIndex
 instance Ord NamedRigid where compare = compare `on` rigidIndex
-instance Show NamedRigid where show = rigidName
+instance Pretty NamedRigid where pretty = P.text . rigidName
 instance Plus NamedRigid Int NamedRigid where
   plus (NamedRigid x i) j = NamedRigid x (i + j)
 
@@ -671,14 +674,14 @@ data SizeMeta = SizeMeta
   -- , sizeMetaPerm :: Permutation -- ^ Permutation from the current context
   --                               --   to the context of the meta.
   , sizeMetaArgs :: [Int]       -- ^ De Bruijn indices.
-  }
+  } deriving (Show)
 
 -- | An equality which ignores the meta arguments.
 instance Eq  SizeMeta where (==)    = (==)    `on` sizeMetaId
 -- | An order which ignores the meta arguments.
 instance Ord SizeMeta where compare = compare `on` sizeMetaId
 
-instance Show SizeMeta where show = show . sizeMetaId
+instance Pretty SizeMeta where pretty = P.pretty . sizeMetaId
 
 instance PrettyTCM SizeMeta where
   prettyTCM (SizeMeta x es) = prettyTCM (MetaV x $ map (Apply . defaultArg . var) es)
@@ -721,7 +724,7 @@ instance PrettyTCM (SizeConstraint) where
   prettyTCM (Constraint a cmp b) = do
     u <- unSizeExpr a
     v <- unSizeExpr b
-    prettyTCM u <+> text (show cmp) <+> prettyTCM v
+    prettyTCM u <+> pretty cmp <+> prettyTCM v
 
 -- | Size constraint with de Bruijn indices.
 data HypSizeConstraint = HypSizeConstraint
@@ -778,7 +781,7 @@ sizeExpr u = do
     SizeInf     -> return $ Just Infty
     SizeSuc u   -> fmap (`plus` (1 :: Offset)) <$> sizeExpr u
     OtherSize u -> case ignoreSharing u of
-      Var i []    -> (\ x -> Just $ Rigid (NamedRigid x i) 0) . show <$> nameOfBV i
+      Var i []    -> (\ x -> Just $ Rigid (NamedRigid x i) 0) . prettyShow <$> nameOfBV i
 --      MetaV m es  -> return $ Just $ Flex (SizeMeta m es) 0
       MetaV m es | Just xs <- mapM isVar es, List.fastDistinct xs
                   -> return $ Just $ Flex (SizeMeta m xs) 0
