@@ -1,7 +1,6 @@
-{-# LANGUAGE CPP                #-}
-{-# LANGUAGE DeriveDataTypeable #-}
-{-# LANGUAGE ScopedTypeVariables    #-}
-{-# LANGUAGE UndecidableInstances   #-}
+{-# LANGUAGE CPP                  #-}
+{-# LANGUAGE DeriveDataTypeable   #-}
+{-# LANGUAGE UndecidableInstances #-}
 
 #if __GLASGOW_HASKELL__ <= 708
 {-# LANGUAGE OverlappingInstances #-}
@@ -9,6 +8,13 @@
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
+-- | This module contains the definition of hereditary substitution
+-- and application operating on internal syntax which is in β-normal
+-- form (β including projection reductions).
+--
+-- Further, it contains auxiliary functions which rely on substitution
+-- but not on reduction.
+
 module Agda.TypeChecking.Substitute
   ( module Agda.TypeChecking.Substitute
   , module Agda.TypeChecking.Substitute.Class
@@ -19,7 +25,6 @@ module Agda.TypeChecking.Substitute
 import Control.Applicative
 import Data.Function
 import Data.Functor
-import Data.List hiding (sort, drop)
 import qualified Data.List as List
 import Data.Map (Map)
 import Data.Maybe
@@ -78,7 +83,7 @@ canProject :: QName -> Term -> Maybe (Arg Term)
 canProject f v =
   case ignoreSharing v of
     (Con (ConHead _ _ fs) _ vs) -> do
-      i <- elemIndex f fs
+      i <- List.elemIndex f fs
       headMaybe (drop i vs)
     _ -> Nothing
 
@@ -91,7 +96,7 @@ conApp ch@(ConHead c _ fs) ci args (Proj o f : es) =
         "conApp: constructor " ++ show c ++
         " with fields " ++ show fs ++
         " projected by " ++ show f
-      i = maybe failure id            $ elemIndex f fs
+      i = maybe failure id            $ List.elemIndex f fs
       v = maybe failure argToDontCare $ headMaybe $ drop i args
   in  applyE v es
 
@@ -178,19 +183,22 @@ instance Subst Term a => Apply (Tele a) where
   apply (ExtendTel _ tel) (t : ts) = lazyAbsApp tel (unArg t) `apply` ts
 
 instance Apply Definition where
-  apply (Defn info x t pol occ df m c inst copy ma d) args =
-    Defn info x (piApply t args) (apply pol args) (apply occ args) df m c inst copy ma (apply d args)
+  apply (Defn info x t pol occ df m c inst copy ma inj d) args =
+    Defn info x (piApply t args) (apply pol args) (apply occ args) df m c inst copy ma inj (apply d args)
 
 instance Apply RewriteRule where
-  apply r args = RewriteRule
-    { rewName    = rewName r
-    , rewContext = apply (rewContext r) args
-    , rewHead    = rewHead r
-    , rewPats    = applySubst sub (rewPats r)
-    , rewRHS     = applySubst sub (rewRHS r)
-    , rewType    = applySubst sub (rewType r)
-    }
-    where sub = parallelS (map unArg args)
+  apply r args =
+    let newContext = apply (rewContext r) args
+        sub        = liftS (size newContext) $ parallelS $
+                       reverse $ map (PTerm . unArg) args
+    in RewriteRule
+       { rewName    = rewName r
+       , rewContext = newContext
+       , rewHead    = rewHead r
+       , rewPats    = applySubst sub (rewPats r)
+       , rewRHS     = applyNLPatSubst sub (rewRHS r)
+       , rewType    = applyNLPatSubst sub (rewType r)
+       }
 
 #if __GLASGOW_HASKELL__ >= 710
 instance {-# OVERLAPPING #-} Apply [Occ.Occurrence] where
@@ -221,6 +229,7 @@ instance Apply [NamedArg (Pattern' a)] where
       in  case namedArg p of
             VarP{}  -> recurse
             DotP{}  -> __IMPOSSIBLE__
+            AbsurdP{} -> __IMPOSSIBLE__
             LitP{}  -> __IMPOSSIBLE__
             ConP{}  -> __IMPOSSIBLE__
             ProjP{} -> __IMPOSSIBLE__
@@ -238,7 +247,7 @@ instance Apply Defn where
   apply d [] = d
   apply d args = case d of
     Axiom{} -> d
-    AbstractDefn -> d
+    AbstractDefn d -> AbstractDefn $ apply d args
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
             , funProjection = Nothing } ->
       d { funClauses    = apply cs args
@@ -305,15 +314,16 @@ instance Apply Clause where
     -- It is assumed that we only apply a clause to "parameters", i.e.
     -- arguments introduced by lambda lifting. The problem is that these aren't
     -- necessarily the first elements of the clause telescope.
-    apply cls@(Clause r tel ps b t catchall) args
+    apply cls@(Clause rl rf tel ps b t catchall unreachable) args
       | length args > length ps = __IMPOSSIBLE__
       | otherwise =
-      Clause r
+      Clause rl rf
              tel'
              (applySubst rhoP $ drop (length args) ps)
              (applySubst rho b)
              (applySubst rho t)
              catchall
+             unreachable
       where
         -- We have
         --  Γ ⊢ args, for some outer context Γ
@@ -354,29 +364,35 @@ instance Apply Clause where
             VarP (DBPatVar _ i) -> mkSub tm (n - 1) (substP i v' ps) vs `composeS` singletonS i (tm v')
               where v' = raise (n - 1) v
             DotP{}  -> mkSub tm n ps vs
+            AbsurdP p' -> mkSub tm n (setNamedArg p p' : ps) (v : vs)
+            ConP c _ ps' -> mkSub tm n (ps' ++ ps) (projections c v ++ vs)
             LitP{}  -> __IMPOSSIBLE__
-            ConP{}  -> __IMPOSSIBLE__
             ProjP{} -> __IMPOSSIBLE__
         mkSub _ _ _ _ = __IMPOSSIBLE__
 
-        -- The parameter patterns 'ps' are all variables or dot patterns. If they
-        -- are variables they can appear anywhere in the clause telescope. This
-        -- function constructs the new telescope with 'vs' substituted for 'ps'.
+        -- The parameter patterns 'ps' are all variables or dot patterns, or eta
+        -- expanded record patterns (issue #2550). If they are variables they
+        -- can appear anywhere in the clause telescope. This function
+        -- constructs the new telescope with 'vs' substituted for 'ps'.
         -- Example:
         --    tel = (x : A) (y : B) (z : C) (w : D)
         --    ps  = y@3 w@0
         --    vs  = u v
         --    newTel tel ps vs = (x : A) (z : C[u/y])
+        newTel :: Nat -> Telescope -> [NamedArg DeBruijnPattern] -> [Term] -> Telescope
         newTel n tel [] [] = tel
         newTel n tel (p : ps) (v : vs) =
           case namedArg p of
             VarP (DBPatVar _ i) -> newTel (n - 1) (subTel (size tel - 1 - i) v tel) (substP i (raise (n - 1) v) ps) vs
             DotP{}              -> newTel n tel ps vs
+            AbsurdP{}           -> __IMPOSSIBLE__
+            ConP c _ ps'        -> newTel n tel (ps' ++ ps) (projections c v ++ vs)
             LitP{}              -> __IMPOSSIBLE__
-            ConP{}              -> __IMPOSSIBLE__
             ProjP{}             -> __IMPOSSIBLE__
         newTel _ tel _ _ = __IMPOSSIBLE__
 
+        projections c v = [ applyE v [Proj ProjSystem f] | f <- conFields c ]
+
         -- subTel i v (Δ₁ (xᵢ : A) Δ₂) = Δ₁ Δ₂[xᵢ = v]
         subTel i v EmptyTel = __IMPOSSIBLE__
         subTel 0 v (ExtendTel _ tel) = absApp tel v
@@ -465,7 +481,7 @@ instance DoDrop a => Abstract (Drop a) where
 instance Apply Permutation where
   -- The permutation must start with [0..m - 1]
   -- NB: section (- m) not possible (unary minus), hence (flip (-) m)
-  apply (Perm n xs) args = Perm (n - m) $ map (flip (-) m) $ genericDrop m xs
+  apply (Perm n xs) args = Perm (n - m) $ map (flip (-) m) $ drop m xs
     where
       m = size args
 
@@ -506,8 +522,8 @@ instance Abstract Telescope where
   ExtendTel arg xtel `abstract` tel = ExtendTel arg $ xtel <&> (`abstract` tel)
 
 instance Abstract Definition where
-  abstract tel (Defn info x t pol occ df m c inst copy ma d) =
-    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c inst copy ma (abstract tel d)
+  abstract tel (Defn info x t pol occ df m c inst copy ma inj d) =
+    Defn info x (abstract tel t) (abstract tel pol) (abstract tel occ) df m c inst copy ma inj (abstract tel d)
 
 -- | @tel ⊢ (Γ ⊢ lhs ↦ rhs : t)@ becomes @tel, Γ ⊢ lhs ↦ rhs : t)@
 --   we do not need to change lhs, rhs, and t since they live in Γ.
@@ -545,7 +561,7 @@ instance Abstract ProjLams where
 instance Abstract Defn where
   abstract tel d = case d of
     Axiom{} -> d
-    AbstractDefn -> d
+    AbstractDefn d -> AbstractDefn $ abstract tel d
     Function{ funClauses = cs, funCompiled = cc, funInv = inv
             , funProjection = Nothing  } ->
       d { funClauses  = abstract tel cs
@@ -582,23 +598,24 @@ instance Abstract Defn where
       d { primClauses = abstract tel cs }
 
 instance Abstract PrimFun where
-    abstract tel (PrimFun x ar def) = PrimFun x (ar + n) $ \ts -> def $ genericDrop n ts
+    abstract tel (PrimFun x ar def) = PrimFun x (ar + n) $ \ts -> def $ drop n ts
         where n = size tel
 
 instance Abstract Clause where
-  abstract tel (Clause r tel' ps b t catchall) =
-    Clause r (abstract tel tel')
+  abstract tel (Clause rl rf tel' ps b t catchall unreachable) =
+    Clause rl rf (abstract tel tel')
            (namedTelVars m tel ++ ps)
            b
            t -- nothing to do for t, since it lives under the telescope
            catchall
+           unreachable
       where m = size tel + size tel'
 
 instance Abstract CompiledClauses where
   abstract tel Fail = Fail
   abstract tel (Done xs t) = Done (map (argFromDom . fmap fst) (telToList tel) ++ xs) t
   abstract tel (Case n bs) =
-    Case (n <&> \ i -> i + fromIntegral (size tel)) (abstract tel bs)
+    Case (n <&> \ i -> i + size tel) (abstract tel bs)
 
 instance Abstract a => Abstract (WithArity a) where
   abstract tel (WithArity n a) = WithArity n $ abstract tel a
@@ -728,28 +745,65 @@ instance Subst Term Pattern where
     ConP c mt ps -> ConP c (applySubst rho mt) $ applySubst rho ps
     DotP t       -> DotP $ applySubst rho t
     VarP s       -> p
+    AbsurdP p    -> AbsurdP $ applySubst rho p
     LitP l       -> p
     ProjP{}      -> p
 
-instance Subst Term NLPat where
+instance DeBruijn NLPat where
+  deBruijnVar i = PVar i []
+  deBruijnView p = case p of
+    PVar i []   -> Just i
+    PVar{}      -> Nothing
+    PWild{}     -> Nothing
+    PDef{}      -> Nothing
+    PLam{}      -> Nothing
+    PPi{}       -> Nothing
+    PBoundVar{} -> Nothing -- or... ?
+    PTerm{}     -> Nothing -- or... ?
+
+applyNLPatSubst :: (Subst Term a) => Substitution' NLPat -> a -> a
+applyNLPatSubst = applySubst . fmap nlPatToTerm
+  where
+    nlPatToTerm :: NLPat -> Term
+    nlPatToTerm p = case p of
+      PVar i xs      -> Var i $ map (Apply . fmap var) xs
+      PTerm u        -> u
+      PWild          -> __IMPOSSIBLE__
+      PDef f es      -> __IMPOSSIBLE__
+      PLam i u       -> __IMPOSSIBLE__
+      PPi a b        -> __IMPOSSIBLE__
+      PBoundVar i es -> __IMPOSSIBLE__
+
+instance Subst NLPat NLPat where
   applySubst rho p = case p of
-    PVar id i bvs -> p
+    PVar i bvs -> lookupS rho i `applyBV` bvs
     PWild  -> p
     PDef f es -> PDef f $ applySubst rho es
     PLam i u -> PLam i $ applySubst rho u
     PPi a b -> PPi (applySubst rho a) (applySubst rho b)
     PBoundVar i es -> PBoundVar i $ applySubst rho es
-    PTerm u -> PTerm $ applySubst rho u
+    PTerm u -> PTerm $ applyNLPatSubst rho u
 
-instance Subst Term NLPType where
+    where
+      applyBV :: NLPat -> [Arg Int] -> NLPat
+      applyBV p ys = case p of
+        PVar i xs      -> PVar i (xs ++ ys)
+        PTerm u        -> PTerm $ u `apply` map (fmap var) ys
+        PWild          -> __IMPOSSIBLE__
+        PDef f es      -> __IMPOSSIBLE__
+        PLam i u       -> __IMPOSSIBLE__
+        PPi a b        -> __IMPOSSIBLE__
+        PBoundVar i es -> __IMPOSSIBLE__
+
+instance Subst NLPat NLPType where
   applySubst rho (NLPType s a) = NLPType (applySubst rho s) (applySubst rho a)
 
-instance Subst Term RewriteRule where
+instance Subst NLPat RewriteRule where
   applySubst rho (RewriteRule q gamma f ps rhs t) =
-    RewriteRule q (applySubst rho gamma)
+    RewriteRule q (applyNLPatSubst rho gamma)
                 f (applySubst (liftS n rho) ps)
-                  (applySubst (liftS n rho) rhs)
-                  (applySubst (liftS n rho) t)
+                  (applyNLPatSubst (liftS n rho) rhs)
+                  (applyNLPatSubst (liftS n rho) t)
     where n = size gamma
 
 instance Subst t a => Subst t (Blocked a) where
@@ -793,6 +847,9 @@ instance Subst Term ModuleParameters where
 instance Subst Term A.NamedDotPattern where
   applySubst rho (A.NamedDot x v a) = A.NamedDot x (applySubst rho v) (applySubst rho a)
 
+instance Subst Term A.StrippedDotPattern where
+  applySubst rho (A.StrippedDot e v a) = A.StrippedDot e (applySubst rho v) (applySubst rho a)
+
 instance Subst t a => Subst t (Elim' a) where
   applySubst rho e = case e of
     Apply v -> Apply $ applySubst rho v
@@ -841,7 +898,7 @@ instance Subst Term EqualityView where
   applySubst rho (EqualityType s eq l t a b) = EqualityType
     (applySubst rho s)
     eq
-    (applySubst rho l)
+    (map (applySubst rho) l)
     (applySubst rho t)
     (applySubst rho a)
     (applySubst rho b)
@@ -863,6 +920,7 @@ instance Subst DeBruijnPattern DeBruijnPattern where
     VarP x       -> useName (dbPatVarName x) $ lookupS rho $ dbPatVarIndex x
     DotP u       -> DotP $ applyPatSubst rho u
     ConP c ci ps -> ConP c ci $ applySubst rho ps
+    AbsurdP p    -> AbsurdP $ applySubst rho p
     LitP x       -> p
     ProjP{}      -> p
     where
@@ -879,7 +937,7 @@ instance Subst DeBruijnPattern DeBruijnPattern where
 --   This function is an optimization, saving us from construction lambdas we
 --   immediately remove through application.
 projDropParsApply :: Projection -> ProjOrigin -> Args -> Term
-projDropParsApply (Projection proper d r _ lams) o args =
+projDropParsApply (Projection prop d r _ lams) o args =
   case initLast $ getProjLams lams of
     -- If we have no more abstractions, we must be a record field
     -- (projection applied already to record value).
@@ -892,11 +950,14 @@ projDropParsApply (Projection proper d r _ lams) o args =
       -- We only have to abstract over the parameters that exceed the arguments.
       -- We only have to apply to the arguments that exceed the parameters.
       in List.foldr (\ (Arg ai x) -> Lam ai . NoAbs x) (core `apply` args') pars'
+  where proper = isJust prop
 
 ---------------------------------------------------------------------------
 -- * Telescopes
 ---------------------------------------------------------------------------
 
+-- ** Telescope view of a type
+
 type TelView = TelV Type
 data TelV a  = TelV { theTel :: Tele (Dom a), theCore :: a }
   deriving (Typeable, Show, Functor)
@@ -904,25 +965,23 @@ data TelV a  = TelV { theTel :: Tele (Dom a), theCore :: a }
 deriving instance (Subst t a, Eq  a) => Eq  (TelV a)
 deriving instance (Subst t a, Ord a) => Ord (TelV a)
 
-type ListTel' a = [Dom (a, Type)]
-type ListTel = ListTel' ArgName
+-- | Takes off all exposed function domains from the given type.
+--   This means that it does not reduce to expose @Pi@-types.
+telView' :: Type -> TelView
+telView' = telView'UpTo (-1)
 
-telFromList' :: (a -> ArgName) -> ListTel' a -> Telescope
-telFromList' f = foldr extTel EmptyTel
+-- | @telView'UpTo n t@ takes off the first @n@ exposed function types of @t@.
+-- Takes off all (exposed ones) if @n < 0@.
+telView'UpTo :: Int -> Type -> TelView
+telView'UpTo 0 t = TelV EmptyTel t
+telView'UpTo n t = case ignoreSharing $ unEl t of
+  Pi a b  -> absV a (absName b) $ telView'UpTo (n - 1) (absBody b)
+  _       -> TelV EmptyTel t
   where
-    extTel (Dom info (x, a)) = ExtendTel (Dom info a) . Abs (f x)
+    absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
 
-telFromList :: ListTel -> Telescope
-telFromList = telFromList' id
 
-telToList :: Telescope -> ListTel
-telToList EmptyTel            = []
-telToList (ExtendTel arg tel) = fmap (absName tel,) arg : telToList (absBody tel)
-  -- Andreas, 2013-12-14: This would work also for 'NoAbs',
-  -- since 'absBody' raises.
-
-telToArgs :: Telescope -> [Arg ArgName]
-telToArgs tel = [ Arg (domInfo d) (fst $ unDom d) | d <- telToList tel ]
+-- ** Creating telescopes from lists of types
 
 -- | Turn a typed binding @(x1 .. xn : A)@ into a telescope.
 bindsToTel' :: (Name -> a) -> [Name] -> Dom Type -> ListTel' a
@@ -941,20 +1000,8 @@ bindsWithHidingToTel' f (WithHiding h x : xs) t =
 bindsWithHidingToTel :: [WithHiding Name] -> Dom Type -> ListTel
 bindsWithHidingToTel = bindsWithHidingToTel' nameToArgName
 
--- | Takes off all exposed function domains from the given type.
---   This means that it does not reduce to expose @Pi@-types.
-telView' :: Type -> TelView
-telView' = telView'UpTo (-1)
 
--- | @telView'UpTo n t@ takes off the first @n@ exposed function types of @t@.
--- Takes off all (exposed ones) if @n < 0@.
-telView'UpTo :: Int -> Type -> TelView
-telView'UpTo 0 t = TelV EmptyTel t
-telView'UpTo n t = case ignoreSharing $ unEl t of
-  Pi a b  -> absV a (absName b) $ telView'UpTo (n - 1) (absBody b)
-  _       -> TelV EmptyTel t
-  where
-    absV a x (TelV tel t) = TelV (ExtendTel a (Abs x tel)) t
+-- ** Abstracting in terms and types
 
 -- | @mkPi dom t = telePi (telFromList [dom]) t@
 mkPi :: Dom (ArgName, Type) -> Type -> Type
@@ -975,11 +1022,11 @@ telePi' reAbs = telePi where
       s2 = getSort <$> b
       el = El $ dLub s1 s2
 
--- | Uses free variable analysis to introduce 'noAbs' bindings.
+-- | Uses free variable analysis to introduce 'NoAbs' bindings.
 telePi :: Telescope -> Type -> Type
 telePi = telePi' reAbs
 
--- | Everything will be a 'Abs'.
+-- | Everything will be an 'Abs'.
 telePi_ :: Telescope -> Type -> Type
 telePi_ = telePi' id
 
@@ -994,6 +1041,10 @@ telePi_ (ExtendTel u tel) t = el $ Pi u b
     s2 = fmap getSort b
 -}
 
+-- | Abstract over a telescope in a term, producing lambdas.
+--   Dumb abstraction: Always produces 'Abs', never 'NoAbs'.
+--
+--   The implementation is sound because 'Telescope' does not use 'NoAbs'.
 teleLam :: Telescope -> Term -> Term
 teleLam  EmptyTel         t = t
 teleLam (ExtendTel u tel) t = Lam (domInfo u) $ flip teleLam t <$> tel
@@ -1008,6 +1059,21 @@ instance TeleNoAbs ListTel where
 instance TeleNoAbs Telescope where
   teleNoAbs tel = teleNoAbs $ telToList tel
 
+
+-- ** Telescope typing
+
+-- | Given arguments @vs : tel@ (vector typing), extract their individual types.
+--   Returns @Nothing@ is @tel@ is not long enough.
+
+typeArgsWithTel :: Telescope -> [Term] -> Maybe [Dom Type]
+typeArgsWithTel _ []                         = return []
+typeArgsWithTel (ExtendTel dom tel) (v : vs) = (dom :) <$> typeArgsWithTel (absApp tel v) vs
+typeArgsWithTel EmptyTel{} (_:_)             = Nothing
+
+---------------------------------------------------------------------------
+-- * Clauses
+---------------------------------------------------------------------------
+
 -- | In compiled clauses, the variables in the clause body are relative to the
 --   pattern variables (including dot patterns) instead of the clause telescope.
 compiledClauseBody :: Clause -> Maybe Term
@@ -1016,6 +1082,8 @@ compiledClauseBody cl = applySubst (renamingR perm) $ clauseBody cl
 
 ---------------------------------------------------------------------------
 -- * Syntactic equality and order
+--
+-- Needs weakening.
 ---------------------------------------------------------------------------
 
 deriving instance Eq Substitution
@@ -1106,6 +1174,9 @@ instance Ord Term where
   _          `compare` MetaV{}    = GT
   DontCare{} `compare` DontCare{} = EQ
 
+-- | Equality of binders relies on weakening
+--   which is a specical case of renaming
+--   which is a specical case of substitution.
 instance (Subst t a, Eq a) => Eq (Abs a) where
   NoAbs _ a == NoAbs _ b = a == b
   Abs   _ a == Abs   _ b = a == b
@@ -1163,6 +1234,19 @@ dLub :: Sort -> Abs Sort -> Sort
 dLub s1 (NoAbs _ s2) = sLub s1 s2
 dLub s1 b@(Abs _ s2) = case occurrence 0 s2 of
   Flexible _    -> DLub s1 b
+  -- Andreas, 2017-01-18, issue #2408:
+  -- The sort of @.(a : A) → Set (f a)@ in context @f : .A → Level@
+  -- is @dLub Set λ a → Set (lsuc (f a))@, but @DLub@s are not serialized.
+  -- Alternatives:
+  -- 1. -- Irrelevantly -> sLub s1 (absApp b $ DontCare $ Sort Prop)
+  --    We cheat here by simplifying the sort to @Set (lsuc (f *))@
+  --    where * is a dummy value.  The rationale is that @f * = f a@ (irrelevance!)
+  --    and that if we already have a neutral level @f a@
+  --    it should not hurt to have @f *@ even if type @A@ is empty.
+  --    However: sorts are printed in error messages when sorts do not match.
+  --    Also, sorts with a dummy like Prop would be ill-typed.
+  -- 2. We keep the DLub, and serialize it.
+  --    That's clean and principled, even though DLubs make level solving harder.
   Irrelevantly  -> DLub s1 b
   NoOccurrence  -> sLub s1 (noabsApp __IMPOSSIBLE__ b)
 --  Free.Unused   -> sLub s1 (absApp b __IMPOSSIBLE__) -- triggers Issue784
