@@ -27,13 +27,11 @@ import Agda.Syntax.Translation.InternalToAbstract
 import Agda.Syntax.Translation.ReflectedToAbstract
 
 import Agda.TypeChecking.CompiledClause
-import Agda.TypeChecking.Datatypes ( getConHead )
 import Agda.TypeChecking.DropArgs
 import Agda.TypeChecking.Free
 import Agda.TypeChecking.Level
 import Agda.TypeChecking.Monad
 import Agda.TypeChecking.Monad.Builtin
-import Agda.TypeChecking.Monad.Exception
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Reduce
 import Agda.TypeChecking.Reduce.Monad
@@ -49,18 +47,25 @@ import {-# SOURCE #-} Agda.TypeChecking.Rules.Term
 import {-# SOURCE #-} Agda.TypeChecking.Rules.Def
 
 import Agda.Utils.Except
-import Agda.Utils.Impossible
+  ( mkExceptT
+  , MonadError(catchError, throwError)
+  , ExceptT
+  , runExceptT
+  )
+import Agda.Utils.Either
+import Agda.Utils.FileName
+import Agda.Utils.Lens
 import Agda.Utils.Maybe
+import Agda.Utils.Maybe.Strict (toLazy)
 import Agda.Utils.Monad
 import Agda.Utils.Permutation ( Permutation(Perm), compactP )
+import Agda.Utils.Pretty (prettyShow)
 import Agda.Utils.String ( Str(Str), unStr )
 import Agda.Utils.VarSet (VarSet)
 import qualified Agda.Utils.VarSet as Set
-import Agda.Utils.Maybe.Strict (toLazy)
-import Agda.Utils.FileName
-import Agda.Utils.Lens
 
 #include "undefined.h"
+import Agda.Utils.Impossible
 
 agdaTermType :: TCM Type
 agdaTermType = El (mkType 0) <$> primAgdaTerm
@@ -78,15 +83,15 @@ data Dirty = Dirty | Clean
 -- definitions. Also state snapshot from last commit and whether the state is
 -- dirty (definitions have been added).
 type UnquoteState = (Dirty, TCState)
-type UnquoteM = ReaderT Context (StateT UnquoteState (WriterT [QName] (ExceptionT UnquoteError TCM)))
+type UnquoteM = ReaderT Context (StateT UnquoteState (WriterT [QName] (ExceptT UnquoteError TCM)))
 
 type UnquoteRes a = Either UnquoteError ((a, UnquoteState), [QName])
 
 unpackUnquoteM :: UnquoteM a -> Context -> UnquoteState -> TCM (UnquoteRes a)
-unpackUnquoteM m cxt s = runExceptionT $ runWriterT $ runStateT (runReaderT m cxt) s
+unpackUnquoteM m cxt s = runExceptT $ runWriterT $ runStateT (runReaderT m cxt) s
 
 packUnquoteM :: (Context -> UnquoteState -> TCM (UnquoteRes a)) -> UnquoteM a
-packUnquoteM f = ReaderT $ \ cxt -> StateT $ \ s -> WriterT $ ExceptionT $ f cxt s
+packUnquoteM f = ReaderT $ \ cxt -> StateT $ \ s -> WriterT $ mkExceptT $ f cxt s
 
 runUnquoteM :: UnquoteM a -> TCM (Either UnquoteError (a, [QName]))
 runUnquoteM m = do
@@ -100,7 +105,7 @@ runUnquoteM m = do
     isDefined x = do
       def <- theDef <$> getConstInfo x
       case def of
-        Function{funClauses = []} -> genericError $ "Missing definition for " ++ show x
+        Function{funClauses = []} -> genericError $ "Missing definition for " ++ prettyShow x
         _       -> return ()
 
 liftU :: TCM a -> UnquoteM a
@@ -135,16 +140,16 @@ reduceQuotedTerm t = do
   b <- liftU $ ifBlocked t (\ m _ -> pure $ Left  m)
                            (\ t   -> pure $ Right t)
   case b of
-    Left m  -> do s <- gets snd; throwException $ BlockedOnMeta s m
+    Left m  -> do s <- gets snd; throwError $ BlockedOnMeta s m
     Right t -> return t
 
 class Unquote a where
   unquote :: I.Term -> UnquoteM a
 
 unquoteN :: Unquote a => Arg Term -> UnquoteM a
-unquoteN a | notHidden a && isRelevant a =
+unquoteN a | visible a && isRelevant a =
     unquote $ unArg a
-unquoteN a = throwException $ BadVisibility "visible" a
+unquoteN a = throwError $ BadVisibility "visible" a
 
 choice :: Monad m => [(m Bool, m a)] -> m a -> m a
 choice [] dflt = dflt
@@ -152,23 +157,23 @@ choice ((mb, mx) : mxs) dflt = ifM mb mx $ choice mxs dflt
 
 ensureDef :: QName -> UnquoteM QName
 ensureDef x = do
-  i <- liftU $ (theDef <$> getConstInfo x) `catchError` \_ -> return Axiom  -- for recursive unquoteDecl
+  i <- liftU $ either (const Axiom) theDef <$> getConstInfo' x  -- for recursive unquoteDecl
   case i of
     Constructor{} -> do
       def <- liftU $ prettyTCM =<< primAgdaTermDef
       con <- liftU $ prettyTCM =<< primAgdaTermCon
-      throwException $ ConInsteadOfDef x (show def) (show con)
+      throwError $ ConInsteadOfDef x (show def) (show con)
     _ -> return x
 
 ensureCon :: QName -> UnquoteM QName
 ensureCon x = do
-  i <- liftU $ (theDef <$> getConstInfo x) `catchError` \_ -> return Axiom  -- for recursive unquoteDecl
+  i <- liftU $ either (const Axiom) theDef <$> getConstInfo' x  -- for recursive unquoteDecl
   case i of
     Constructor{} -> return x
     _ -> do
       def <- liftU $ prettyTCM =<< primAgdaTermDef
       con <- liftU $ prettyTCM =<< primAgdaTermCon
-      throwException $ DefInsteadOfCon x (show def) (show con)
+      throwError $ DefInsteadOfCon x (show def) (show con)
 
 pickName :: R.Type -> String
 pickName a =
@@ -185,10 +190,10 @@ instance Unquote ArgInfo where
     case ignoreSharing t of
       Con c _ [h,r] -> do
         choice
-          [(c `isCon` primArgArgInfo, ArgInfo <$> unquoteN h <*> unquoteN r <*> pure Reflected <*> pure False)]
+          [(c `isCon` primArgArgInfo, ArgInfo <$> unquoteN h <*> unquoteN r <*> pure Reflected)]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "arg info" t
+      _ -> throwError $ NonCanonical "arg info" t
 
 instance Unquote a => Unquote (Arg a) where
   unquote t = do
@@ -199,7 +204,7 @@ instance Unquote a => Unquote (Arg a) where
           [(c `isCon` primArgArg, Arg <$> unquoteN info <*> unquoteN x)]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "arg" t
+      _ -> throwError $ NonCanonical "arg" t
 
 -- Andreas, 2013-10-20: currently, post-fix projections are not part of the
 -- quoted syntax.
@@ -214,35 +219,35 @@ instance Unquote Bool where
         choice [ (c `isCon` primTrue,  pure True)
                , (c `isCon` primFalse, pure False) ]
                __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "boolean" t
+      _ -> throwError $ NonCanonical "boolean" t
 
 instance Unquote Integer where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
       Lit (LitNat _ n) -> return n
-      _ -> throwException $ NonCanonical "integer" t
+      _ -> throwError $ NonCanonical "integer" t
 
 instance Unquote Double where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
       Lit (LitFloat _ x) -> return x
-      _ -> throwException $ NonCanonical "float" t
+      _ -> throwError $ NonCanonical "float" t
 
 instance Unquote Char where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
       Lit (LitChar _ x) -> return x
-      _ -> throwException $ NonCanonical "char" t
+      _ -> throwError $ NonCanonical "char" t
 
 instance Unquote Str where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
       Lit (LitString _ x) -> return (Str x)
-      _ -> throwException $ NonCanonical "string" t
+      _ -> throwError $ NonCanonical "string" t
 
 unquoteString :: Term -> UnquoteM String
 unquoteString x = unStr <$> unquote x
@@ -266,7 +271,7 @@ instance Unquote ErrorPart where
                , (c `isCon` primAgdaErrorPartTerm,   TermPart <$> unquoteN x)
                , (c `isCon` primAgdaErrorPartName,   NamePart <$> unquoteN x) ]
                __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "error part" t
+      _ -> throwError $ NonCanonical "error part" t
 
 instance Unquote a => Unquote [a] where
   unquote t = do
@@ -281,7 +286,7 @@ instance Unquote a => Unquote [a] where
           [(c `isCon` primNil, return [])]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "list" t
+      _ -> throwError $ NonCanonical "list" t
 
 instance Unquote Hiding where
   unquote t = do
@@ -290,11 +295,11 @@ instance Unquote Hiding where
       Con c _ [] -> do
         choice
           [(c `isCon` primHidden,  return Hidden)
-          ,(c `isCon` primInstance, return Instance)
+          ,(c `isCon` primInstance, return (Instance NoOverlap))
           ,(c `isCon` primVisible, return NotHidden)]
           __IMPOSSIBLE__
       Con c _ vs -> __IMPOSSIBLE__
-      _        -> throwException $ NonCanonical "visibility" t
+      _        -> throwError $ NonCanonical "visibility" t
 
 instance Unquote Relevance where
   unquote t = do
@@ -306,14 +311,14 @@ instance Unquote Relevance where
           ,(c `isCon` primIrrelevant, return Irrelevant)]
           __IMPOSSIBLE__
       Con c _ vs -> __IMPOSSIBLE__
-      _        -> throwException $ NonCanonical "relevance" t
+      _        -> throwError $ NonCanonical "relevance" t
 
 instance Unquote QName where
   unquote t = do
     t <- reduceQuotedTerm t
     case ignoreSharing t of
       Lit (LitQName _ x) -> return x
-      _                  -> throwException $ NonCanonical "name" t
+      _                  -> throwError $ NonCanonical "name" t
 
 instance Unquote a => Unquote (R.Abs a) where
   unquote t = do
@@ -324,7 +329,7 @@ instance Unquote a => Unquote (R.Abs a) where
           [(c `isCon` primAbsAbs, R.Abs <$> (hint <$> unquoteNString x) <*> unquoteN y)]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "abstraction" t
+      _ -> throwError $ NonCanonical "abstraction" t
 
     where hint x | not (null x) = x
                  | otherwise    = "_"
@@ -344,7 +349,7 @@ instance Unquote MetaId where
               sep [ text "Can't unquote stale metavariable"
                   , pretty m <> text "." <> pretty x ]
         return x
-      _ -> throwException $ NonCanonical "meta variable" t
+      _ -> throwError $ NonCanonical "meta variable" t
 
 instance Unquote a => Unquote (Dom a) where
   unquote t = domFromArg <$> unquote t
@@ -363,7 +368,7 @@ instance Unquote R.Sort where
           ,(c `isCon` primAgdaSortLit, R.LitS <$> unquoteN u)]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "sort" t
+      _ -> throwError $ NonCanonical "sort" t
 
 instance Unquote Literal where
   unquote t = do
@@ -382,7 +387,7 @@ instance Unquote Literal where
           , (c `isCon` primAgdaLitMeta,   litMeta   noRange =<< unquoteN x) ]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "literal" t
+      _ -> throwError $ NonCanonical "literal" t
 
 instance Unquote R.Term where
   unquote t = do
@@ -420,7 +425,7 @@ instance Unquote R.Term where
 
       Con{} -> __IMPOSSIBLE__
       Lit{} -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "term" t
+      _ -> throwError $ NonCanonical "term" t
 
 instance Unquote R.Pattern where
   unquote t = do
@@ -442,7 +447,7 @@ instance Unquote R.Pattern where
           [ (c `isCon` primAgdaPatCon, R.ConP <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "pattern" t
+      _ -> throwError $ NonCanonical "pattern" t
 
 instance Unquote R.Clause where
   unquote t = do
@@ -457,7 +462,7 @@ instance Unquote R.Clause where
           [ (c `isCon` primAgdaClauseClause, R.Clause <$> unquoteN x <*> unquoteN y) ]
           __IMPOSSIBLE__
       Con c _ _ -> __IMPOSSIBLE__
-      _ -> throwException $ NonCanonical "clause" t
+      _ -> throwError $ NonCanonical "clause" t
 
 -- Unquoting TCM computations ---------------------------------------------
 
@@ -474,7 +479,7 @@ evalTCM :: I.Term -> UnquoteM I.Term
 evalTCM v = do
   v <- reduceQuotedTerm v
   liftU $ reportSDoc "tc.unquote.eval" 90 $ text "evalTCM" <+> prettyTCM v
-  let failEval = throwException $ NonCanonical "type checking computation" v
+  let failEval = throwError $ NonCanonical "type checking computation" v
 
   case ignoreSharing v of
     I.Def f [] ->
@@ -501,7 +506,9 @@ evalTCM v = do
              , (f `isDef` primAgdaTCMTypeError,   tcFun1 tcTypeError   u)
              , (f `isDef` primAgdaTCMQuoteTerm,   tcQuoteTerm (unElim u))
              , (f `isDef` primAgdaTCMUnquoteTerm, tcFun1 (tcUnquoteTerm (mkT (unElim l) (unElim a))) u)
-             , (f `isDef` primAgdaTCMBlockOnMeta, uqFun1 tcBlockOnMeta u) ]
+             , (f `isDef` primAgdaTCMBlockOnMeta, uqFun1 tcBlockOnMeta u)
+             , (f `isDef` primAgdaTCMDebugPrint,  tcFun3 tcDebugPrint l a u)
+             ]
              failEval
     I.Def f [_, _, u, v] ->
       choice [ (f `isDef` primAgdaTCMCatchError,    tcCatchError    (unElim u) (unElim v))
@@ -550,9 +557,19 @@ evalTCM v = do
       b <- unquote (unElim b)
       fun a b
 
+    uqFun3 :: (Unquote a, Unquote b, Unquote c) => (a -> b -> c -> UnquoteM d) -> Elim -> Elim -> Elim -> UnquoteM d
+    uqFun3 fun a b c = do
+      a <- unquote (unElim a)
+      b <- unquote (unElim b)
+      c <- unquote (unElim c)
+      fun a b c
+
     tcFun2 :: (Unquote a, Unquote b) => (a -> b -> TCM c) -> Elim -> Elim -> UnquoteM c
     tcFun2 fun = uqFun2 (\ x y -> liftU (fun x y))
 
+    tcFun3 :: (Unquote a, Unquote b, Unquote c) => (a -> b -> c -> TCM d) -> Elim -> Elim -> Elim -> UnquoteM d
+    tcFun3 fun = uqFun3 (\ x y z -> liftU (fun x y z))
+
     tcFreshName :: Str -> TCM Term
     tcFreshName s = do
       m <- currentModule
@@ -568,7 +585,7 @@ evalTCM v = do
     tcBlockOnMeta :: MetaId -> UnquoteM Term
     tcBlockOnMeta x = do
       s <- gets snd
-      throwException (BlockedOnMeta s x)
+      throwError (BlockedOnMeta s x)
 
     tcCommit :: UnquoteM Term
     tcCommit = do
@@ -582,6 +599,11 @@ evalTCM v = do
     tcTypeError :: [ErrorPart] -> TCM a
     tcTypeError err = typeError . GenericDocError =<< fsep (map prettyTCM err)
 
+    tcDebugPrint :: Str -> Integer -> [ErrorPart] -> TCM Term
+    tcDebugPrint (Str s) n msg = do
+      reportSDoc s (fromIntegral n) $ fsep (map prettyTCM msg)
+      primUnitUnit
+
     tcInferType :: R.Term -> TCM Term
     tcInferType v = do
       (_, a) <- inferExpr =<< toAbstract_ v
@@ -622,7 +644,7 @@ evalTCM v = do
     extendCxt :: Arg R.Type -> UnquoteM a -> UnquoteM a
     extendCxt a m = do
       a <- liftU $ traverse (isType_ <=< toAbstract_) a
-      liftU1 (addContext (domFromArg a :: Dom Type)) m
+      liftU1 (addContext' (domFromArg a :: Dom Type)) m
 
     tcExtendContext :: Term -> Term -> UnquoteM Term
     tcExtendContext a m = do
@@ -639,8 +661,8 @@ evalTCM v = do
         go (a : as) m = extendCxt a $ go as m
 
     constInfo :: QName -> TCM Definition
-    constInfo x = getConstInfo x `catchError` \ _ ->
-                  genericError $ "Unbound name: " ++ show x
+    constInfo x = either err return =<< getConstInfo' x
+      where err _ = genericError $ "Unbound name: " ++ prettyShow x
 
     tcGetType :: QName -> TCM Term
     tcGetType x = quoteType . defType =<< constInfo x
@@ -662,24 +684,26 @@ evalTCM v = do
     tcDeclareDef :: Arg QName -> R.Type -> UnquoteM Term
     tcDeclareDef (Arg i x) a = inOriginalContext $ do
       setDirty
-      let h = getHiding i
-          r = getRelevance i
-      when (h == Hidden) $ liftU $ typeError . GenericDocError =<< text "Cannot declare hidden function" <+> prettyTCM x
+      let r = getRelevance i
+      when (hidden i) $ liftU $ typeError . GenericDocError =<<
+        text "Cannot declare hidden function" <+> prettyTCM x
       tell [x]
       liftU $ do
-        reportSDoc "tc.unquote.decl" 10 $ sep [ text "declare" <+> prettyTCM x <+> text ":"
-                                              , nest 2 $ prettyTCM a ]
+        reportSDoc "tc.unquote.decl" 10 $ sep
+          [ text "declare" <+> prettyTCM x <+> text ":"
+          , nest 2 $ prettyTCM a
+          ]
         a <- isType_ =<< toAbstract_ a
-        alreadyDefined <- isJust <$> tryMaybe (getConstInfo x)
-        when alreadyDefined $ genericError $ "Multiple declarations of " ++ show x
+        alreadyDefined <- isRight <$> getConstInfo' x
+        when alreadyDefined $ genericError $ "Multiple declarations of " ++ prettyShow x
         addConstant x $ defaultDefn i x a emptyFunction
-        when (h == Instance) $ addTypedInstance x a
+        when (isInstance i) $ addTypedInstance x a
         primUnitUnit
 
     tcDefineFun :: QName -> [R.Clause] -> UnquoteM Term
     tcDefineFun x cs = inOriginalContext $ (setDirty >>) $ liftU $ do
-      _ <- getConstInfo x `catchError` \ _ ->
-        genericError $ "Missing declaration for " ++ show x
+      whenM (isLeft <$> getConstInfo' x) $
+        genericError $ "Missing declaration for " ++ prettyShow x
       cs <- mapM (toAbstract_ . QNamed x) cs
       reportSDoc "tc.unquote.def" 10 $ vcat $ map prettyA cs
       let i = mkDefInfo (nameConcrete $ qnameName x) noFixity' PublicAccess ConcreteDef noRange
