@@ -8,7 +8,8 @@ import Control.Applicative hiding (empty)
 import Control.Monad
 import Control.Monad.Writer (WriterT, runWriterT, tell)
 
-import Data.List
+import Data.Either
+import qualified Data.List as List
 import Data.Maybe
 import Data.Monoid
 import Data.Traversable (traverse)
@@ -33,6 +34,7 @@ import Agda.TypeChecking.Records
 import Agda.TypeChecking.Substitute
 import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.ReconstructParameters
+import Agda.TypeChecking.Rules.Term
 
 import Agda.TypeChecking.Abstract
 import Agda.TypeChecking.Rules.LHS.Implicit
@@ -45,6 +47,7 @@ import Agda.Utils.Monad
 import Agda.Utils.Null (empty)
 import Agda.Utils.Permutation
 import Agda.Utils.Pretty (prettyShow)
+import qualified Agda.Utils.Pretty as P
 import Agda.Utils.Size
 
 #include "undefined.h"
@@ -66,9 +69,9 @@ import Agda.Utils.Impossible
 --
 --   Output:
 --
---   [@Δ₁@]        part of context not needed for with arguments and their types.
+--   [@Δ₁@]        part of context needed for with arguments and their types.
 --
---   [@Δ₂@]        part of context needed for with arguments and their types.
+--   [@Δ₂@]        part of context not needed for with arguments and their types.
 --
 --   [@π@]         permutation from Δ to Δ₁Δ₂ as returned by 'splitTelescope'.
 --
@@ -189,7 +192,7 @@ countWithArgs = sum . map countArgs
 withArguments :: [Term] -> [EqualityView] -> [Term]
 withArguments vs as = concat $ for (zip vs as) $ \case
   (v, OtherType a) -> [v]
-  (prf, eqt@(EqualityType s _eq _l t v _v')) -> [unArg v, prf]
+  (prf, eqt@(EqualityType s _eq _pars _t v _v')) -> [unArg v, prf]
 
 
 -- | Compute the clauses for the with-function given the original patterns.
@@ -198,6 +201,7 @@ buildWithFunction
   -> QName                -- ^ Name of the parent function.
   -> QName                -- ^ Name of the with-function.
   -> Type                 -- ^ Types of the parent function.
+  -> Telescope            -- ^ Context of parent patterns.
   -> [NamedArg DeBruijnPattern] -- ^ Parent patterns.
   -> Nat                  -- ^ Number of module parameters in parent patterns
   -> Substitution         -- ^ Substitution from parent lhs to with function lhs
@@ -206,18 +210,22 @@ buildWithFunction
   -> Nat                  -- ^ Number of with expressions.
   -> [A.SpineClause]      -- ^ With-clauses.
   -> TCM [A.SpineClause]  -- ^ With-clauses flattened wrt. parent patterns.
-buildWithFunction cxtNames f aux t qs npars withSub perm n1 n cs = mapM buildWithClause cs
+buildWithFunction cxtNames f aux t delta qs npars withSub perm n1 n cs = mapM buildWithClause cs
   where
     -- Nested with-functions will iterate this function once for each parent clause.
-    buildWithClause (A.Clause (A.SpineLHS i _ ps wps) inheritedDots rhs wh catchall) = do
-      let (wps0, wps1) = genericSplitAt n wps
+    buildWithClause (A.Clause (A.SpineLHS i _ ps wps) inheritedDots inhStrippedDots rhs wh catchall) = do
+      let (wps0, wps1) = splitAt n wps
           ps0          = map defaultNamedArg wps0
       reportSDoc "tc.with" 50 $ text "inheritedDots:" <+> vcat [ prettyTCM x <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a
                                                                | A.NamedDot x v a <- inheritedDots ]
       rhs <- buildRHS rhs
-      (namedDots, ps') <- stripWithClausePatterns cxtNames f aux t qs npars perm ps
-      let (ps1, ps2) = genericSplitAt n1 ps'
-      let result = A.Clause (A.SpineLHS i aux (ps1 ++ ps0 ++ ps2) wps1) (inheritedDots ++ namedDots) rhs wh catchall
+      (namedDots, strippedDots, ps') <- stripWithClausePatterns cxtNames f aux t delta qs npars perm ps
+      reportSDoc "tc.with" 50 $ hang (text "strippedDots:") 2 $
+                                  vcat [ prettyTCM e <+> text "==" <+> prettyTCM v <+> (text ":" <+> prettyTCM t)
+                                       | A.StrippedDot e v t <- strippedDots ]
+      let (ps1, ps2) = splitAt n1 ps'
+      let result = A.Clause (A.SpineLHS i aux (ps1 ++ ps0 ++ ps2) wps1) (inheritedDots ++ namedDots)
+                                                                        (inhStrippedDots ++ strippedDots) rhs wh catchall
       reportSDoc "tc.with" 20 $ vcat
         [ text "buildWithClause returns" <+> prettyA result
         ]
@@ -235,14 +243,15 @@ buildWithFunction cxtNames f aux t qs npars withSub perm n1 n cs = mapM buildWit
     -- of the innermost parent clause. Note that this makes them live in the
     -- context of the with-function arguments before any pattern matching. We
     -- need to update again once the with-clause patterns have been checked.
-    -- This happens in Rules.Def.checkClause before calling checkRHS.
+    -- This happens in Rules.Def.checkClause before calling checkRHS. The same
+    -- goes for stripped dots.
     permuteNamedDots :: A.SpineClause -> A.SpineClause
-    permuteNamedDots (A.Clause lhs dots rhs wh catchall) =
-      A.Clause lhs (applySubst withSub dots) rhs wh catchall
+    permuteNamedDots (A.Clause lhs dots sdots rhs wh catchall) =
+      A.Clause lhs (applySubst withSub dots) (applySubst withSub sdots) rhs wh catchall
 
-{-| @stripWithClausePatterns cxtNames parent f t qs np π ps = ps'@
+{-| @stripWithClausePatterns cxtNames parent f t Δ qs np π ps = ps'@
 
-[@Δ@]   context bound by lhs of original function (not an argument).
+[@Δ@]   context bound by lhs of original function.
 
 [@f@]   name of @with@-function.
 
@@ -312,12 +321,13 @@ stripWithClausePatterns
   -> QName                    -- ^ Name of the parent function.
   -> QName                    -- ^ Name of with-function.
   -> Type                     -- ^ __@t@__   top-level type of the original function.
+  -> Telescope                -- ^ __@Δ@__   context of patterns of parent function.
   -> [NamedArg DeBruijnPattern] -- ^ __@qs@__  internal patterns for original function.
   -> Nat                      -- ^ __@npars@__ number of module parameters is @qs@.
   -> Permutation              -- ^ __@π@__   permutation taking @vars(qs)@ to @support(Δ)@.
   -> [NamedArg A.Pattern]     -- ^ __@ps@__  patterns in with clause (eliminating type @t@).
-  -> TCM ([A.NamedDotPattern], [NamedArg A.Pattern]) -- ^ __@ps'@__ patterns for with function (presumably of type @Δ@).
-stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
+  -> TCM ([A.NamedDotPattern], [A.StrippedDotPattern], [NamedArg A.Pattern]) -- ^ __@ps'@__ patterns for with function (presumably of type @Δ@).
+stripWithClausePatterns cxtNames parent f t delta qs npars perm ps = do
   -- Andreas, 2014-03-05 expand away pattern synoyms (issue 1074)
   ps <- expandPatternSynonyms ps
   -- Ulf, 2016-11-16 Issue 2303: We need the module parameter
@@ -338,7 +348,8 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
     ]
 
   -- Andreas, 2015-11-09 Issue 1710: self starts with parent-function, not with-function!
-  (ps', namedDots) <- runWriterT $ strip (Def parent []) t psi qs
+  (ps', out) <- runWriterT $ strip (Def parent []) t psi qs
+  let (strippedDots, namedDots) = partitionEithers out
   reportSDoc "tc.with.strip" 50 $ nest 2 $
     text "namedDots:" <+> vcat [ prettyTCM x <+> text "=" <+> prettyTCM v <+> text ":" <+> prettyTCM a | A.NamedDot x v a <- namedDots ]
   let psp = permute perm ps'
@@ -349,7 +360,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
   -- Andreas, 2014-03-05 Issue 142:
   -- In some cases, permute throws away some dot patterns of ps'
   -- which are then never checked.
-  if True then return (namedDots, psp) else do
+  if True then return (namedDots, strippedDots, psp) else do
     -- Andreas, 2014-03-05 Disabled the fix for issue 142, the following is dead code:
     forM_ (permute (droppedP perm) ps') $ \ p -> setCurrentRange p $ do
       reportSDoc "tc.with.strip" 10 $ text "warning: dropped pattern " <+> prettyA p
@@ -361,7 +372,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
             "This inaccessible pattern is never checked, so only _ allowed here"
           _ -> return ()
         _ -> return ()
-    return (namedDots, psp)
+    return (namedDots, strippedDots, psp)
   where
 
     strip
@@ -369,7 +380,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
       -> Type                         -- ^ The type to be eliminated.
       -> [NamedArg A.Pattern]       -- ^ With-clause patterns.
       -> [NamedArg DeBruijnPattern] -- ^ Parent-clause patterns with de Bruijn indices relative to Δ.
-      -> WriterT [A.NamedDotPattern] TCM [NamedArg A.Pattern]
+      -> WriterT [Either A.StrippedDotPattern A.NamedDotPattern] TCM [NamedArg A.Pattern]
             -- ^ With-clause patterns decomposed by parent-clause patterns.
             --   Also outputs named dot patterns from the parent clause that
             --   we need to add let-bindings for.
@@ -407,7 +418,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
       | A.AsP _ x p <- namedArg p0 = do
         (a, _) <- mustBePi t
         let v = patternToTerm (namedArg q)
-        tell [A.NamedDot x v (unDom a)]
+        tell [Right $ A.NamedDot x v (unDom a)]
         strip self t (fmap (p <$) p0 : ps) qs
     strip self t ps0@(p0 : ps) qs0@(q : qs) = do
       p <- liftTCM $ expandLitPattern p0
@@ -429,13 +440,14 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
       case namedArg q of
         ProjP o d -> case A.maybePostfixProjP p of
           Just (o', AmbQ ds) -> do
-            let d' = head ds
-            when (length ds /= 1) __IMPOSSIBLE__
+            -- Andreas, 2016-12-28, issue #2360:
+            -- We disambiguate the projection in the with clause
+            -- to the projection in the parent clause.
             d  <- liftTCM $ getOriginalProjection d
-            d' <- liftTCM $ getOriginalProjection d'
+            found <- anyM ds ((d ==) <.> (liftTCM . getOriginalProjection))
             -- We assume here that neither @o@ nor @o'@ can be @ProjSystem@.
             if o /= o' then liftTCM $ mismatchOrigin o o' else do
-            if d /= d' then mismatch else do
+            if not found then mismatch else do
               (self1, t1, ps) <- liftTCM $ do
                 t <- reduce t
                 (_, self1, t1) <- fromMaybe __IMPOSSIBLE__ <$> projectTyped self t o d
@@ -447,14 +459,15 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
               strip self1 t1 ps qs
           Nothing -> mismatch
 
-        VarP x  -> do
-          let v = var (dbPatVarIndex x)
-          t  <- piApply1 t v
-          ps <- strip (self `apply1` v) t ps qs
-          return $ p : ps
+        VarP x  -> (p :) <$> recurse (var (dbPatVarIndex x))
+
+        AbsurdP p -> __IMPOSSIBLE__
 
         DotP v  -> case namedArg p of
-          A.DotP r o _  -> ok p
+          A.DotP r o e  -> do
+            (a, _) <- mustBePi t
+            tell [Left $ A.StrippedDot e v (unDom a)]
+            ok p
           A.WildP _     -> ok p
           -- Ulf, 2016-05-30: dot patterns are no longer mandatory so a parent
           -- dot pattern can appear as a variable in the child clause. Indeed
@@ -463,7 +476,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
           -- insert a let for it.
           A.VarP x -> do
             (a, _) <- mustBePi t
-            tell [A.NamedDot x v (unDom a)]
+            tell [Right $ A.NamedDot x v (unDom a)]
             ok p
           -- Andreas, 2013-03-21 in case the implicit A.pattern has already been eta-expanded
           -- we just fold it back.  This fixes issues 665 and 824.
@@ -480,9 +493,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
           _ -> failDotPat
           where
             okFlex = ok . makeImplicitP
-            ok p = do
-              t' <- piApply1 t v
-              (p :) <$> strip (self `apply1` v) t' ps qs
+            ok p   = (p :) <$> recurse v
 
         q'@(ConP c ci qs') -> do
          reportSDoc "tc.with.strip" 60 $
@@ -492,7 +503,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
          Def d es <- liftTCM $ ignoreSharing <$> normalise (unEl $ unDom a)
          let us = fromMaybe __IMPOSSIBLE__ $ allApplyElims es
          -- Get the original constructor and field names.
-         c <- (`withRangeOf` c) <$> do liftTCM $ getConForm $ conName c
+         c <- either __IMPOSSIBLE__ (`withRangeOf` c) <$> do liftTCM $ getConForm $ conName c
 
          case namedArg p of
 
@@ -503,20 +514,17 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
             maybe __IMPOSSIBLE__ (\ p -> strip self t (p : ps) qs0) =<< do
               liftTCM $ expandImplicitPattern' (unDom a) $ makeImplicitP p
 
-          -- Andreas, 2013-03-21 if we encounter an implicit pattern
-          -- in the with-clause, we expand it and restart
-          -- Andreas, 2015-07-07 Issue 1606 do this whenever the parent
-          -- is a record pattern, regardless of whether it came from an implicit
-          -- or not.  This allows to drop hidden flexible record patterns from
-          -- the with clauses even when they were present in the parent clause.
-          A.WildP{} | Just _ <- conPRecord ci -> do
-            maybe __IMPOSSIBLE__ (\ p -> strip self t (p : ps) qs0) =<< do
-              liftTCM $ expandImplicitPattern' (unDom a) p
+          -- Andreas, 2016-12-29, issue #2363.
+          -- Allow _ to stand for the corresponding parent pattern.
+          A.WildP{} -> do
+            let ps' = map (updateNamedArg $ const $ A.WildP empty) qs'
+            stripConP d us b c ConOCon qs' ps'
 
           A.ConP _ (A.AmbQ cs') ps' -> do
             -- Check whether the with-clause constructor can be (possibly trivially)
             -- disambiguated to be equal to the parent-clause constructor.
-            cs' <- liftTCM $ mapM getConForm cs'
+            -- Andreas, 2017-08-13, herein, ignore abstract constructors.
+            cs' <- liftTCM $ do snd . partitionEithers <$> mapM getConForm cs'
             unless (elem c cs') mismatch
             -- Strip the subpatterns ps' and then continue.
             stripConP d us b c ConOCon qs' ps'
@@ -536,10 +544,8 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
            mismatch
 
         LitP lit -> case namedArg p of
-          A.LitP lit' | lit == lit' -> do
-            (a, b) <- mustBePi t
-            let v = Lit lit
-            strip (self `apply1` v) (b `absApp` v) ps qs
+          A.LitP lit' | lit == lit' -> recurse $ Lit lit
+          A.WildP{}                 -> recurse $ Lit lit
 
           p@(A.PatternSynP pi' c' [ps']) -> do
              reportSDoc "impossible" 10 $
@@ -548,13 +554,17 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
 
           _ -> mismatch
       where
-        mismatch = typeError $
-          WithClausePatternMismatch (namedArg p0) (dbPatVarName <$> namedArg q)
-        mismatchOrigin o o' = typeError . GenericDocError =<< fsep
+        recurse v = do
+          t' <- piApply1 t v
+          strip (self `apply1` v) t' ps qs
+
+        mismatch = addContext delta $ typeError $
+          WithClausePatternMismatch (namedArg p0) q
+        mismatchOrigin o o' = addContext delta . typeError . GenericDocError =<< fsep
           [ text "With clause pattern"
           , prettyA p0
           , text "is not an instance of its parent pattern"
-          , prettyTCM $ dbPatVarName <$> namedArg q
+          , P.fsep <$> prettyTCMPatterns [q]
           , text $ "since the parent pattern is " ++ prettyProjOrigin o ++
                    " and the with clause pattern is " ++ prettyProjOrigin o'
           ]
@@ -582,14 +592,14 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
              -- ^ Argument patterns (parent clause).
           -> [NamedArg A.Pattern]
              -- ^ Argument patterns (with clause).
-          -> WriterT [A.NamedDotPattern] TCM [NamedArg A.Pattern]
+          -> WriterT [Either A.StrippedDotPattern A.NamedDotPattern] TCM [NamedArg A.Pattern]
              -- ^ Stripped patterns.
         stripConP d us b c ci qs' ps' = do
 
           -- Get the type and number of parameters of the constructor.
           Defn {defType = ct, theDef = Constructor{conPars = np}}  <- getConInfo c
           -- Compute the argument telescope for the constructor
-          let ct' = ct `piApply` genericTake np us
+          let ct' = ct `piApply` take np us
           TelV tel' _ <- liftTCM $ telView ct'
 
           reportSDoc "tc.with.strip" 20 $
@@ -597,7 +607,7 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
                  , text "ct' = " <+> prettyTCM ct'
                  , text "np  = " <+> text (show np)
                  , text "us  = " <+> prettyList (map prettyTCM us)
-                 , text "us' = " <+> prettyList (map prettyTCM $ genericTake np us)
+                 , text "us' = " <+> prettyList (map prettyTCM $ take np us)
                  ]
 
           -- Compute the new type
@@ -638,14 +648,22 @@ stripWithClausePatterns cxtNames parent f t qs npars perm ps = do
 --     f (suc a) (suc b) | c
 --   @
 withDisplayForm
-  :: QName                -- ^ The name of parent function.
-  -> QName                -- ^ The name of the @with@-function.
-  -> Telescope            -- ^ __@Δ₁@__      The arguments of the @with@ function before the @with@ expressions.
-  -> Telescope            -- ^ __@Δ₂@__      The arguments of the @with@ function after the @with@ expressions.
-  -> Nat                  -- ^ __@n@__       The number of @with@ expressions.
-  -> [NamedArg DeBruijnPattern]   -- ^ __@qs@__      The parent patterns.
-  -> Permutation          -- ^ __@perm@__    Permutation to split into needed and unneeded vars.
-  -> Permutation          -- ^ __@lhsPerm@__ Permutation reordering the variables in parent patterns.
+  :: QName
+       -- ^ The name of parent function.
+  -> QName
+       -- ^ The name of the @with@-function.
+  -> Telescope
+       -- ^ __@Δ₁@__     The arguments of the @with@ function before the @with@ expressions.
+  -> Telescope
+       -- ^ __@Δ₂@__     The arguments of the @with@ function after the @with@ expressions.
+  -> Nat
+       -- ^ __@n@__      The number of @with@ expressions.
+  -> [NamedArg DeBruijnPattern]
+      -- ^ __@qs@__      The parent patterns.
+  -> Permutation
+      -- ^ __@perm@__    Permutation to split into needed and unneeded vars.
+  -> Permutation
+      -- ^ __@lhsPerm@__ Permutation reordering the variables in parent patterns.
   -> TCM DisplayForm
 withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) lhsPerm = do
 
@@ -683,14 +701,14 @@ withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) lhsPerm = do
   reportSDoc "tc.with.display" 20 $ vcat
     [ text "withDisplayForm"
     , nest 2 $ vcat
-      [ text "f      =" <+> text (show f)
-      , text "aux    =" <+> text (show aux)
+      [ text "f      =" <+> text (prettyShow f)
+      , text "aux    =" <+> text (prettyShow aux)
       , text "delta1 =" <+> prettyTCM delta1
       , text "delta2 =" <+> do addContext delta1 $ prettyTCM delta2
       , text "n      =" <+> text (show n)
       , text "perm   =" <+> text (show perm)
       , text "top    =" <+> do addFullCtx $ prettyTCM topArgs
-      , text "qs     =" <+> sep (map (prettyTCM . namedArg) qs)
+      , text "qs     =" <+> prettyList (map pretty qs)
       , text "qsToTm =" <+> prettyTCM tqs0 -- ctx would be permuted form of delta1 ++ delta2
       , text "ys     =" <+> text (show ys)
       , text "rho    =" <+> text (prettyShow rho)
@@ -709,7 +727,7 @@ withDisplayForm f aux delta1 delta2 n qs perm@(Perm m _) lhsPerm = do
     -- Andreas, 2015-10-28: Yes, but properly! (Issue 1407)
     sub top ys wild = parallelS $ map term [0 .. m + top - 1]
       where
-        term i = maybe wild var $ findIndex (Just i ==) ys
+        term i = maybe wild var $ List.findIndex (Just i ==) ys
     -- -- OLD
     -- sub top rho wild = parallelS $ map term [0 .. m - 1] ++ topTerms
     --   where
@@ -741,5 +759,6 @@ patsToElims = map $ toElim . fmap namedThing
       ProjP _ d   -> DDef d [] -- WRONG. TODO: convert spine to non-spine ... DDef d . defaultArg
       VarP x      -> DTerm  $ var $ dbPatVarIndex x
       DotP t      -> DDot   $ t
+      AbsurdP p   -> toTerm p
       ConP c cpi ps -> DCon c (fromConPatternInfo cpi) $ toTerms ps
       LitP l      -> DTerm  $ Lit l
