@@ -17,7 +17,7 @@ import Control.Applicative
 import System.Directory
 import System.FilePath
 
-#if mingw32_HOST_OS
+#ifdef mingw32_HOST_OS
 import Control.Exception (bracket)
 import System.Win32 (findFirstFile, findClose, getFindDataFileName)
 #endif
@@ -25,7 +25,8 @@ import System.Win32 (findFirstFile, findClose, getFindDataFileName)
 import Data.Text (Text)
 import qualified Data.Text as Text
 import Data.Function
-import Data.Hashable
+import Data.Hashable (Hashable)
+import Data.Data (Data)
 import Data.Typeable (Typeable)
 
 import Agda.Utils.Monad
@@ -40,7 +41,7 @@ import Agda.Utils.Impossible
 -- paths point to the same files or directories.
 
 newtype AbsolutePath = AbsolutePath { byteStringPath :: Text }
-  deriving (Eq, Ord, Typeable, Hashable)
+  deriving (Eq, Ord, Typeable, Data, Hashable)
 
 -- | Extract the 'AbsolutePath' to be used as 'FilePath'.
 filePath :: AbsolutePath -> FilePath
@@ -67,7 +68,7 @@ mkAbsolute f
   | otherwise    = __IMPOSSIBLE__
 
 rootPath :: FilePath
-#if mingw32_HOST_OS
+#ifdef mingw32_HOST_OS
 rootPath = joinDrive "C:" [pathSeparator]
 #else
 rootPath = [pathSeparator]
@@ -113,7 +114,7 @@ infix 4 ===
 --  case insensitively.)
 
 doesFileExistCaseSensitive :: FilePath -> IO Bool
-#if mingw32_HOST_OS
+#ifdef mingw32_HOST_OS
 doesFileExistCaseSensitive f = do
   doesFileExist f `and2M` do
     bracket (findFirstFile f) (findClose . fst) $
