@@ -1,14 +1,12 @@
-{-# LANGUAGE CPP #-}
+-- | Utility functions for lists.
 
-{-| Utitlity functions on lists.
--}
 module Agda.Utils.List where
 
 import Control.Arrow (first)
 
 import Data.Functor ((<$>))
 import Data.Function
-import Data.List
+import qualified Data.List as List
 import Data.Maybe
 import qualified Data.Map as Map
 import qualified Data.Set as Set
@@ -19,9 +17,6 @@ import qualified Agda.Utils.Bag as Bag
 
 import Agda.Utils.Tuple
 
-#include "undefined.h"
-import Agda.Utils.Impossible
-
 -- | Case distinction for lists, with list first.
 --   Cf. 'Agda.Utils.Null.ifNull'.
 caseList :: [a] -> b -> (a -> [a] -> b) -> b
@@ -49,6 +44,14 @@ lastMaybe :: [a] -> Maybe a
 lastMaybe [] = Nothing
 lastMaybe xs = Just $ last xs
 
+-- | Last two elements (safe).
+last2 :: [a] -> Maybe (a, a)
+last2 (x : y : xs) = Just $ loop x y xs
+  where
+  loop x y []     = (x, y)
+  loop x y (z:xs) = loop y z xs
+last2 _ = Nothing
+
 -- | Opposite of cons @(:)@, safe.
 uncons :: [a] -> Maybe (a, [a])
 uncons []     = Nothing
@@ -60,11 +63,10 @@ mcons ma as = maybe as (:as) ma
 
 -- | 'init' and 'last' in one go, safe.
 initLast :: [a] -> Maybe ([a],a)
-initLast [] = Nothing
-initLast as = Just $ loop as where
-  loop []       = __IMPOSSIBLE__
-  loop [a]      = ([], a)
-  loop (a : as) = mapFst (a:) $ loop as
+initLast []     = Nothing
+initLast (a:as) = Just $ loop a as where
+  loop a []      = ([], a)
+  loop a (b : bs) = mapFst (a:) $ loop b bs
 
 -- | Lookup function (partially safe).
 (!!!) :: [a] -> Int -> Maybe a
@@ -261,7 +263,7 @@ duplicates = mapMaybe dup . Bag.groups . Bag.fromList
     dup (a : _ : _) = Just a
     dup _           = Nothing
 
--- | A variant of 'groupBy' which applies the predicate to consecutive
+-- | A variant of 'List.groupBy' which applies the predicate to consecutive
 -- pairs.
 
 groupBy' :: (a -> a -> Bool) -> [a] -> [[a]]
@@ -274,10 +276,10 @@ groupBy' p xxs@(x : xs) = grp x $ zipWith (\x y -> (p x y, y)) xxs xs
                    []            -> []
                    ((_, z) : zs) -> grp z zs
 
--- | @'groupOn' f = 'groupBy' (('==') \`on\` f) '.' 'sortBy' ('compare' \`on\` f)@.
+-- | @'groupOn' f = 'groupBy' (('==') \`on\` f) '.' 'List.sortBy' ('compare' \`on\` f)@.
 
 groupOn :: Ord b => (a -> b) -> [a] -> [[a]]
-groupOn f = groupBy ((==) `on` f) . sortBy (compare `on` f)
+groupOn f = List.groupBy ((==) `on` f) . List.sortBy (compare `on` f)
 
 -- | @splitExactlyAt n xs = Just (ys, zs)@ iff @xs = ys ++ zs@
 --   and @genericLength ys = n@.
@@ -297,23 +299,27 @@ genericElemIndex x xs =
   map (== x) xs
 
 -- | Requires both lists to have the same length.
+--
+--   Otherwise, @Nothing@ is returned.
 
-zipWith' :: (a -> b -> c) -> [a] -> [b] -> [c]
-zipWith' f []        []      = []
-zipWith' f (x : xs) (y : ys) = f x y : zipWith' f xs ys
-zipWith' f []       (_ : _)  = {- ' -} __IMPOSSIBLE__
-zipWith' f (_ : _)  []       = {- ' -} __IMPOSSIBLE__
+zipWith' :: (a -> b -> c) -> [a] -> [b] -> Maybe [c]
+zipWith' f = loop
+  where
+  loop []        []      = Just []
+  loop (x : xs) (y : ys) = (f x y :) <$> loop xs ys
+  loop []       (_ : _)  = Nothing
+  loop (_ : _)  []       = Nothing
 
-{- UNUSED; a better type would be
-   zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], Either [a] [b])
+-- -- UNUSED; a better type would be
+-- -- zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], Either [a] [b])
+
+-- -- | Like zipWith, but returns the leftover elements of the input lists.
+-- zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], [a] , [b])
+-- zipWithTails f xs       []       = ([], xs, [])
+-- zipWithTails f []       ys       = ([], [] , ys)
+-- zipWithTails f (x : xs) (y : ys) = (f x y : zs , as , bs)
+--   where (zs , as , bs) = zipWithTails f xs ys
 
--- | Like zipWith, but returns the leftover elements of the input lists.
-zipWithTails :: (a -> b -> c) -> [a] -> [b] -> ([c], [a] , [b])
-zipWithTails f xs       []       = ([], xs, [])
-zipWithTails f []       ys       = ([], [] , ys)
-zipWithTails f (x : xs) (y : ys) = (f x y : zs , as , bs)
-  where (zs , as , bs) = zipWithTails f xs ys
--}
 
 -- | Efficient variant of 'nubBy' for finite lists.
 --
@@ -323,10 +329,10 @@ zipWithTails f (x : xs) (y : ys) = (f x y : zs , as , bs)
 nubOn :: Ord b => (a -> b) -> [a] -> [a]
 nubOn tag =
   map snd
-  . sortBy (compare `on` fst)
+  . List.sortBy (compare `on` fst)
   . map (snd . head)
-  . groupBy ((==) `on` fst)
-  . sortBy (compare `on` fst)
+  . List.groupBy ((==) `on` fst)
+  . List.sortBy (compare `on` fst)
   . map (\p@(_, x) -> (tag x, p))
   . zip [1..]
 
@@ -339,7 +345,7 @@ nubOn tag =
 --
 -- Furthermore
 --
--- > sortBy (compare `on` f) (uniqOn f xs) == uniqOn f xs.
+-- > List.sortBy (compare `on` f) (uniqOn f xs) == uniqOn f xs.
 uniqOn :: Ord b => (a -> b) -> [a] -> [a]
 uniqOn key = Map.elems . Map.fromList . map (\ a -> (key a, a))
 
@@ -366,8 +372,8 @@ editDistanceSpec (x : xs) (y : ys)
 
 editDistance :: Eq a => [a] -> [a] -> Int
 editDistance xs ys = editD 0 0
-  where xss = tails xs
-        yss = tails ys
+  where xss = List.tails xs
+        yss = List.tails ys
         tbl = Map.fromList [ ((i, j), editD' i j) | i <- [0..length xss - 1], j <- [0..length yss - 1] ]
         editD i j = tbl Map.! (i, j)
         editD' i j =
