@@ -1,9 +1,4 @@
 {-# LANGUAGE CPP                   #-}
-{-# LANGUAGE DeriveFunctor         #-}
-{-# LANGUAGE FlexibleInstances     #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE RankNTypes            #-}
-{-# LANGUAGE TupleSections         #-}
 {-# LANGUAGE UndecidableInstances  #-}  -- Due to limitations of funct.dep.
 
 -- | @ListT@ done right,
@@ -20,7 +15,7 @@ import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State
 
-import Data.Monoid
+import Data.Semigroup
 
 import Agda.Utils.Maybe
 
@@ -108,10 +103,12 @@ liftListT lift xs = runMListT $ caseMaybeM (lift $ runListT xs) (return nilListT
 
 -- Instances
 
+instance Monad m => Semigroup (ListT m a) where
+  l1 <> l2 = ListT $ foldListT cons (runListT l2) l1
+    where cons a = runListT . consListT a . ListT
 instance Monad m => Monoid (ListT m a) where
   mempty        = nilListT
-  mappend l1 l2 = ListT $ foldListT cons (runListT l2) l1
-    where cons a = runListT . consListT a . ListT
+  mappend = (<>)
 
 instance (Functor m, Applicative m, Monad m) => Alternative (ListT m) where
   empty = mempty
