@@ -5,9 +5,6 @@ module Agda.Utils.Monad
     , when, unless, MonadPlus(..)
     , (<$>), (<*>)
     , (<$)
-#if MIN_VERSION_mtl(2,2,0)
-    , Control.Monad.State.modify'
-#endif
     )
     where
 
@@ -20,6 +17,7 @@ import Data.Traversable as Trav hiding (for, sequence)
 import Data.Foldable as Fold
 import Data.Maybe
 
+import Agda.Utils.Either
 import Agda.Utils.Except
   ( Error(strMsg)
   , MonadError(catchError, throwError)
@@ -36,19 +34,11 @@ k ==<< (ma, mb) = ma >>= \ a -> k a =<< mb
 
 -- Conditionals and monads ------------------------------------------------
 
--- | @when_@ is just @Control.Monad.when@ with a more general type.
-when_ :: Monad m => Bool -> m a -> m ()
-when_ b m = when b $ m >> return ()
+whenM :: Monad m => m Bool -> m () -> m ()
+whenM c m = c >>= (`when` m)
 
--- | @unless_@ is just @Control.Monad.unless@ with a more general type.
-unless_ :: Monad m => Bool -> m a -> m ()
-unless_ b m = unless b $ m >> return ()
-
-whenM :: Monad m => m Bool -> m a -> m ()
-whenM c m = c >>= (`when_` m)
-
-unlessM :: Monad m => m Bool -> m a -> m ()
-unlessM c m = c >>= (`unless_` m)
+unlessM :: Monad m => m Bool -> m () -> m ()
+unlessM c m = c >>= (`unless` m)
 
 -- | Monadic guard.
 guardM :: (Monad m, MonadPlus m) => m Bool -> m ()
@@ -85,11 +75,19 @@ anyM :: (Functor f, Foldable f, Monad m) => f a -> (a -> m Bool) -> m Bool
 anyM xs f = orM $ fmap f xs
 
 -- | Lazy monadic disjunction with @Either@  truth values.
+--   Returns the last error message if all fail.
 altM1 :: Monad m => (a -> m (Either err b)) -> [a] -> m (Either err b)
 altM1 f []       = __IMPOSSIBLE__
 altM1 f [a]      = f a
 altM1 f (a : as) = either (const $ altM1 f as) (return . Right) =<< f a
 
+-- | Lazy monadic disjunction with accumulation of errors in a monoid.
+--   Errors are discarded if we succeed.
+orEitherM :: (Monoid e, Monad m, Functor m) => [m (Either e b)] -> m (Either e b)
+orEitherM []       = return $ Left mempty
+orEitherM (m : ms) = caseEitherM m (\e -> mapLeft (e `mappend`) <$> orEitherM ms)
+                                   (return . Right)
+
 -- Loops gathering results in a Monoid ------------------------------------
 
 -- | Generalized version of @mapM_ :: Monad m => (a -> m ()) -> [a] -> m ()@
@@ -112,20 +110,19 @@ forM' = flip mapM'
 
 -- Continuation monad -----------------------------------------------------
 
-type Cont r a = (a -> r) -> r
-
--- | 'Control.Monad.mapM' for the continuation monad. Terribly useful.
-thread :: (a -> Cont r b) -> [a] -> Cont r [b]
-thread f [] ret = ret []
-thread f (x:xs) ret =
-    f x $ \y -> thread f xs $ \ys -> ret (y:ys)
+-- Andreas, 2017-04-11, issue #2543
+-- The terribly useful thread function is now UNUSED.  [Sadistic laughter :)]
+--
+-- type Cont r a = (a -> r) -> r
+--
+-- -- | 'Control.Monad.mapM' for the continuation monad. Terribly useful.
+-- thread :: (a -> Cont r b) -> [a] -> Cont r [b]
+-- thread f [] ret = ret []
+-- thread f (x:xs) ret =
+--     f x $ \y -> thread f xs $ \ys -> ret (y:ys)
 
 -- Lists and monads -------------------------------------------------------
 
--- | Requires both lists to have the same lengths.
-zipWithM' :: Monad m => (a -> b -> m c) -> [a] -> [b] -> m [c]
-zipWithM' f xs ys = sequence (zipWith' f xs ys)
-
 -- | A monadic version of @'mapMaybe' :: (a -> Maybe b) -> [a] -> [b]@.
 mapMaybeM
 #if __GLASGOW_HASKELL__ <= 708
@@ -163,10 +160,10 @@ partitionM f (x:xs) =
 -- | Finally for the 'Error' class. Errors in the finally part take
 -- precedence over prior errors.
 
-finally :: MonadError e m => m a -> m b -> m a
+finally :: MonadError e m => m a -> m () -> m a
 first `finally` after = do
   r <- catchError (liftM Right first) (return . Left)
-  _ <- after
+  after
   case r of
     Left e  -> throwError e
     Right r -> return r
@@ -181,26 +178,19 @@ tryMaybe m = (Just <$> m) `catchError` \ _ -> return Nothing
 -- | Bracket without failure.  Typically used to preserve state.
 bracket_ :: Monad m
          => m a         -- ^ Acquires resource. Run first.
-         -> (a -> m c)  -- ^ Releases resource. Run last.
+         -> (a -> m ())  -- ^ Releases resource. Run last.
          -> m b         -- ^ Computes result. Run in-between.
          -> m b
 bracket_ acquire release compute = do
   resource <- acquire
   result <- compute
-  _ <- release resource
+  release resource
   return result
 
 -- | Restore state after computation.
 localState :: MonadState s m => m a -> m a
 localState = bracket_ get put
 
-#if !MIN_VERSION_mtl(2,2,0)
-modify' :: MonadState s m => (s -> s) -> m ()
-modify' f = do
-  x <- get
-  put $! f x
-#endif
-
 -- Read -------------------------------------------------------------------
 
 readM :: (Error e, MonadError e m, Read a) => String -> m a
