@@ -1,5 +1,9 @@
 {-# LANGUAGE CPP #-}
 
+#if __GLASGOW_HASKELL__ <= 708
+{-# LANGUAGE DeriveDataTypeable #-}
+#endif
+
 {-| Pretty printing functions.
 -}
 module Agda.Utils.Pretty
@@ -8,9 +12,17 @@ module Agda.Utils.Pretty
     ) where
 
 import Data.Int ( Int32 )
+import Data.Data (Data(..))
+
+#if __GLASGOW_HASKELL__ <= 708
+import Data.Typeable ( Typeable )
+#endif
 
 import Text.PrettyPrint hiding (TextDetails(Str), empty)
 
+#include "undefined.h"
+import Agda.Utils.Impossible
+
 -- * Pretty class
 
 -- | While 'Show' is for rendering data in Haskell syntax,
@@ -23,20 +35,17 @@ import Text.PrettyPrint hiding (TextDetails(Str), empty)
 class Pretty a where
   pretty      :: a -> Doc
   prettyPrec  :: Int -> a -> Doc
+  prettyList  :: [a] -> Doc
 
   pretty      = prettyPrec 0
   prettyPrec  = const pretty
+  prettyList  = brackets . prettyList_
 
 -- | Use instead of 'show' when printing to world.
 
 prettyShow :: Pretty a => a -> String
 prettyShow = render . pretty
 
--- | Space separated list of pretty things.
-
-prettyList :: Pretty a => [a] -> Doc
-prettyList = sep . map (prettyPrec 10000)
-
 -- * Pretty instances
 
 instance Pretty Bool    where pretty = text . show
@@ -46,12 +55,17 @@ instance Pretty Integer where pretty = text . show
 
 instance Pretty Char where
   pretty c = text [c]
+  prettyList = text
 
 instance Pretty Doc where
   pretty = id
 
-instance Pretty String where
-  pretty = text
+instance Pretty a => Pretty (Maybe a) where
+  prettyPrec p Nothing  = text "Nothing"
+  prettyPrec p (Just x) = mparens (p > 0) $ text "Just" <+> prettyPrec 10 x
+
+instance Pretty a => Pretty [a] where
+  pretty = prettyList
 
 -- * 'Doc' utilities
 
@@ -61,6 +75,15 @@ pwords = map text . words
 fwords :: String -> Doc
 fwords = fsep . pwords
 
+-- | Comma separated list, without the brackets.
+prettyList_ :: Pretty a => [a] -> Doc
+prettyList_ = fsep . punctuate comma . map pretty
+
+-- ASR (2016-12-13): In pretty >= 1.1.2.0 the below function 'mparens'
+-- is called 'maybeParens'. I didn't use that name due to the issue
+-- https://github.com/haskell/pretty/issues/40.
+
+-- | Apply 'parens' to 'Doc' if boolean is true.
 mparens :: Bool -> Doc -> Doc
 mparens True  = parens
 mparens False = id
@@ -81,3 +104,22 @@ align max rows =
 multiLineText :: String -> Doc
 multiLineText = vcat . map text . lines
 
+
+#if __GLASGOW_HASKELL__ <= 708
+deriving instance Typeable Doc
+#endif
+
+-- cheating because you shouldn't be digging this far anyway
+instance Data Doc where
+  gunfold _ _ _ = __IMPOSSIBLE__
+  toConstr      = __IMPOSSIBLE__
+  dataTypeOf    = __IMPOSSIBLE__
+
+infixl 6 <?>
+-- | @a <?> b = hang a 2 b@
+(<?>) :: Doc -> Doc -> Doc
+a <?> b = hang a 2 b
+
+-- | @pshow = text . pretty@
+pshow :: Show a => a -> Doc
+pshow = text . show
